package speiger.src.collections.bytes.maps.impl.misc;

import java.util.Arrays;
import java.util.Comparator;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Objects;
import java.util.function.Consumer;
import java.util.function.BiFunction;

import speiger.src.collections.bytes.collections.ByteBidirectionalIterator;
import speiger.src.collections.bytes.functions.ByteComparator;
import speiger.src.collections.bytes.functions.ByteConsumer;
import speiger.src.collections.objects.functions.consumer.ObjectByteConsumer;
import speiger.src.collections.bytes.functions.function.Byte2BooleanFunction;
import speiger.src.collections.bytes.functions.consumer.ByteObjectConsumer;
import speiger.src.collections.bytes.functions.function.Byte2ObjectFunction;
import speiger.src.collections.bytes.functions.function.ByteObjectUnaryOperator;
import speiger.src.collections.bytes.functions.function.ByteByteUnaryOperator;
import speiger.src.collections.bytes.lists.ByteListIterator;
import speiger.src.collections.bytes.maps.abstracts.AbstractByte2ObjectMap;
import speiger.src.collections.bytes.maps.interfaces.Byte2ObjectMap;
import speiger.src.collections.bytes.maps.interfaces.Byte2ObjectSortedMap;
import speiger.src.collections.bytes.sets.AbstractByteSet;
import speiger.src.collections.bytes.sets.ByteSortedSet;
import speiger.src.collections.bytes.sets.ByteSet;
import speiger.src.collections.bytes.utils.maps.Byte2ObjectMaps;
import speiger.src.collections.objects.collections.AbstractObjectCollection;
import speiger.src.collections.objects.collections.ObjectCollection;
import speiger.src.collections.objects.collections.ObjectIterator;
import speiger.src.collections.objects.functions.ObjectSupplier;
import speiger.src.collections.objects.functions.function.ObjectObjectUnaryOperator;

import speiger.src.collections.objects.functions.consumer.ObjectObjectConsumer;
import speiger.src.collections.objects.functions.function.Object2BooleanFunction;
import speiger.src.collections.objects.collections.ObjectBidirectionalIterator;
import speiger.src.collections.objects.lists.ObjectListIterator;
import speiger.src.collections.objects.sets.AbstractObjectSet;
import speiger.src.collections.objects.sets.ObjectSortedSet;
import speiger.src.collections.objects.sets.ObjectSet;
import speiger.src.collections.utils.HashUtil;

/**
 * A Very Specific Type Specific implementation of a ArrayMap.
 * This type of map is for very specific use cases that usaully would have lead to Tupled Lists otherwise.
 * It also does not allow duplication (except for array constructors) and checks from last to first.
 * It is not designed to be used as a HashMap replacement due to the poor performance it would cause.
 * @note in this implementation SubMaps do NOT keep track of parent changes fully. For performance reasons it will just have a start/end index and not values
 * Anything within that range will be updated appropiatly a shrink/growth of elements will break SubMaps in some ways. This can be useful but be careful
 * @note this implementation does not shrink and only grows.
 * @param <V> the type of elements maintained by this Collection
 */
public class Byte2ObjectArrayMap<V> extends AbstractByte2ObjectMap<V> implements Byte2ObjectSortedMap<V>
{
	/** The Backing keys array */
	protected transient byte[] keys;
	/** The Backing values array */
	protected transient V[] values;
	/** Amount of Elements stored in the ArrayMap */
	protected int size = 0;
	/** KeySet cache */
	protected ByteSet keySet;
	/** Values cache */
	protected ObjectCollection<V> valuesC;
	/** EntrySet cache */
	protected FastSortedSet<V> entrySet;
	
	/**
	 * Default Constructor
	 */
	public Byte2ObjectArrayMap() {
		this(HashUtil.DEFAULT_MIN_CAPACITY);
	}
	
	/**
	 * Constructor that defines the minimum capacity
	 * @param minCapacity the minimum capacity the HashMap is allowed to be.
	 * @throws IllegalStateException if the minimum capacity is negative
	 */
	public Byte2ObjectArrayMap(int minCapacity) {
		if(minCapacity < 0)	throw new IllegalStateException("Minimum Capacity is negative. This is not allowed");
		keys = new byte[minCapacity];
		values = (V[])new Object[minCapacity];
	}
	
	/**
	 * Helper constructor that allow to create a map from boxed values (it will unbox them)
	 * @param keys the keys that should be put into the map
	 * @param values the values that should be put into the map.
	 * @throws IllegalStateException if the keys and values do not match in length
	 */
	public Byte2ObjectArrayMap(Byte[] keys, V[] values) {
		this(keys, values, keys.length);
	}
	
	/**
	 * Helper constructor that allow to create a map from boxed values (it will unbox them) with a custom length
	 * @param keys the keys that should be put into the map
	 * @param values the values that should be put into the map.
	 * @param length the amount of values that should be pulled from the array
	 * @throws IllegalStateException if the keys and values do not match in length
	 */
	public Byte2ObjectArrayMap(Byte[] keys, V[] values, int length) {
		this(length);
		if(keys.length != values.length) throw new IllegalStateException("Input Arrays are not equal size");
		putAll(keys, values, 0, length);
	}
	
	/**
	 * Helper constructor that allow to create a map from unboxed values
	 * @param keys the keys that should be put into the map
	 * @param values the values that should be put into the map.
	 * @throws IllegalStateException if the keys and values do not match in lenght
	 */
	public Byte2ObjectArrayMap(byte[] keys, V[] values) {
		this(keys, values, keys.length);
	}
	
	/**
	 * Helper constructor that allow to create a map from unboxed values
	 * @param keys the keys that should be put into the map
	 * @param values the values that should be put into the map.
	 * @param length the amount of values that should be pulled from the array
	 * @throws IllegalStateException if the keys and values do not match in lenght
	 */
	public Byte2ObjectArrayMap(byte[] keys, V[] values, int length) {
		this(length);
		if(keys.length != values.length) throw new IllegalStateException("Input Arrays are not equal size");
		putAll(keys, values, 0, length);
	}
	
	/**
	 * A Helper constructor that allows to create a Map with exactly the same values as the provided map.
	 * @param map the values that should be present in the map
	 */
	public Byte2ObjectArrayMap(Map<? extends Byte, ? extends V> map) {
		this(map.size());
		putAll(map);
	}
	
	/**
	 * A Type Specific Helper function that allows to create a new Map with exactly the same values as the provided map.
	 * @param map the values that should be present in the map
 	 */
	public Byte2ObjectArrayMap(Byte2ObjectMap<V> map) {
		this(map.size());
		for(ObjectIterator<Byte2ObjectMap.Entry<V>> iter = Byte2ObjectMaps.fastIterator(map);iter.hasNext();size++) {
			Byte2ObjectMap.Entry<V> entry = iter.next();
			keys[size] = entry.getByteKey();
			values[size] = entry.getValue();
		}
	}
	
	@Override
	public V put(byte key, V value) {
		int index = findIndex(key);
		if(index < 0) {
			insertIndex(size++, key, value);
			return getDefaultReturnValue();
		}
		V oldValue = values[index];
		values[index] = value;
		return oldValue;
	}
	
	@Override
	public V putIfAbsent(byte key, V value) {
		int index = findIndex(key);
		if(index < 0) {
			insertIndex(size++, key, value);
			return getDefaultReturnValue();
		}
		return values[index];
	}
	
	@Override
	public V putAndMoveToFirst(byte key, V value) {
		int index = findIndex(key);
		if(index < 0) {
			insertIndex(0, key, value);
			size++;
			return getDefaultReturnValue();
		}
		V lastValue = values[index];
		values[index] = value;
		moveIndexToFirst(index);
		return lastValue;
	}
	
	@Override
	public V putAndMoveToLast(byte key, V value) {
		int index = findIndex(key);
		if(index < 0) {
			insertIndex(size++, key, value);
			return getDefaultReturnValue();
		}
		V lastValue = values[index];
		values[index] = value;
		moveIndexToLast(index);
		return lastValue;
	}
	
	@Override
	public boolean moveToFirst(byte key) {
		int index = findIndex(key);
		if(index > 0) {
			moveIndexToFirst(index);
			return true;
		}
		return false;
	}
	
	@Override
	public boolean moveToLast(byte key) {
		int index = findIndex(key);
		if(index < size-1) {
			moveIndexToLast(index);
			return true;
		}
		return false;
	}
	
	@Override
	public boolean containsKey(byte key) {
		return findIndex(key) >= 0;
	}
	
	@Override
	public boolean containsKey(Object key) {
		return findIndex(key) >= 0;
	}
	
	@Override
	public boolean containsValue(Object value) {
		return findValue(value) >= 0;
	}
	
	@Override
	public V get(byte key) {
		int index = findIndex(key);
		return index < 0 ? getDefaultReturnValue() : values[index];
	}
	
	@Override
	public V getOrDefault(byte key, V defaultValue) {
		int index = findIndex(key);
		return index < 0 ? defaultValue : values[index];
	}
	
	@Override
	public V getAndMoveToFirst(byte key) {
		int index = findIndex(key);
		if(index >= 0) {
			V value = values[index];
			moveIndexToFirst(index);
			return value;
		}
		return getDefaultReturnValue();
	}
	
	@Override
	public V getAndMoveToLast(byte key) {
		int index = findIndex(key);
		if(index >= 0) {
			V value = values[index];
			moveIndexToLast(index);
			return value;
		}
		return getDefaultReturnValue();
	}
	
	@Override
	public byte firstByteKey() {
		if(size <= 0) throw new NoSuchElementException();
		return keys[0];
	}
	
	@Override
	public byte lastByteKey() {
		if(size <= 0) throw new NoSuchElementException();
		return keys[size-1];
	}
	
	@Override
	public V firstValue() {
		if(size <= 0) throw new NoSuchElementException();
		return values[0];
	}
	
	@Override
	public V lastValue() {
		if(size <= 0) throw new NoSuchElementException();
		return values[size-1];
	}
	
	@Override
	public byte pollFirstByteKey() {
		if(size == 0) throw new NoSuchElementException();
		byte result = keys[0];
		removeIndex(0);
		return result;
	}
	
	@Override
	public byte pollLastByteKey() {
		if(size == 0) throw new NoSuchElementException();
		byte result = keys[size-1];
		removeIndex(size-1);
		return result;
	}
	
	@Override
	public V remove(byte key) {
		int index = findIndex(key);
		if(index < 0) return getDefaultReturnValue();
		V value = values[index];
		removeIndex(index);
		return value;
	}
	
	@Override
	public V removeOrDefault(byte key, V defaultValue) {
		int index = findIndex(key);
		if(index < 0) return defaultValue;
		V value = values[index];
		removeIndex(index);
		return value;
	}
	
	@Override
	public boolean remove(byte key, V value) {
		int index = findIndex(key, value);
		if(index < 0) return false;
		removeIndex(index);
		return true;
	}
	
	@Override
	public V remove(Object key) {
		int index = findIndex(key);
		if(index < 0) return getDefaultReturnValue();
		V value = values[index];
		removeIndex(index);
		return value;
	}
	
	@Override
	public boolean remove(Object key, Object value) {
		int index = findIndex(key, value);
		if(index < 0) return false;
		removeIndex(index);
		return true;
	}
	
	@Override
	public void forEach(ByteObjectConsumer<V> action) {
		if(size() <= 0) return;
		for(int i = 0;i<size;i++)
			action.accept(keys[i], values[i]);
	}
	
	@Override
	public ByteSet keySet() {
		if(keySet == null) keySet = new KeySet();
		return keySet;
	}

	@Override
	public ObjectCollection<V> values() {
		if(valuesC == null) valuesC = new Values();
		return valuesC;
	}

	@Override
	public ObjectSortedSet<Byte2ObjectMap.Entry<V>> byte2ObjectEntrySet() {
		if(entrySet == null) entrySet = new MapEntrySet();
		return entrySet;
	}
	
	@Override
	public boolean replace(byte key, V oldValue, V newValue) {
		int index = findIndex(key);
		if(index < 0 || values[index] != oldValue) return false;
		values[index] = newValue;
		return true;
	}
	
	@Override
	public V replace(byte key, V value) {
		int index = findIndex(key);
		if(index < 0) return getDefaultReturnValue();
		V oldValue = values[index];
		values[index] = value;
		return oldValue;
	}
	
	@Override
	public V compute(byte key, ByteObjectUnaryOperator<V> mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		int index = findIndex(key);
		if(index == -1) {
			V newValue = mappingFunction.apply(key, getDefaultReturnValue());
			if(Objects.equals(newValue, getDefaultReturnValue())) return newValue;
			insertIndex(size++, key, newValue);
			return newValue;
		}
		V newValue = mappingFunction.apply(key, values[index]);
		if(Objects.equals(newValue, getDefaultReturnValue())) {
			removeIndex(index);
			return newValue;
		}
		values[index] = newValue;
		return newValue;
	}
	
	@Override
	public V computeIfAbsent(byte key, Byte2ObjectFunction<V> mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		int index = findIndex(key);
		if(index == -1) {
			V newValue = mappingFunction.get(key);
			if(Objects.equals(newValue, getDefaultReturnValue())) return newValue;
			insertIndex(size++, key, newValue);
			return newValue;
		}
		V newValue = values[index];
		if(Objects.equals(newValue, getDefaultReturnValue())) {
			newValue = mappingFunction.get(key);
			if(Objects.equals(newValue, getDefaultReturnValue())) return newValue;
			values[index] = newValue;
		}
		return newValue;
	}
	
	@Override
	public V supplyIfAbsent(byte key, ObjectSupplier<V> valueProvider) {
		Objects.requireNonNull(valueProvider);
		int index = findIndex(key);
		if(index == -1) {
			V newValue = valueProvider.get();
			if(Objects.equals(newValue, getDefaultReturnValue())) return newValue;
			insertIndex(size++, key, newValue);
			return newValue;
		}
		V newValue = values[index];
		if(Objects.equals(newValue, getDefaultReturnValue())) {
			newValue = valueProvider.get();
			if(Objects.equals(newValue, getDefaultReturnValue())) return newValue;
			values[index] = newValue;
		}
		return newValue;
	}
	
	@Override
	public V computeIfPresent(byte key, ByteObjectUnaryOperator<V> mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		int index = findIndex(key);
		if(index == -1 || Objects.equals(values[index], getDefaultReturnValue())) return getDefaultReturnValue();
		V newValue = mappingFunction.apply(key, values[index]);
		if(Objects.equals(newValue, getDefaultReturnValue())) {
			removeIndex(index);
			return newValue;
		}
		values[index] = newValue;
		return newValue;
	}
	
	@Override
	public V merge(byte key, V value, ObjectObjectUnaryOperator<V, V> mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		Objects.requireNonNull(value);
		int index = findIndex(key);
		V newValue = index == -1 || Objects.equals(values[index], getDefaultReturnValue()) ? value : mappingFunction.apply(values[index], value);
		if(Objects.equals(newValue, getDefaultReturnValue())) {
			if(index >= 0)
				removeIndex(index);
		}
		else if(index == -1) insertIndex(size++, key, newValue);
		else values[index] = newValue;
		return newValue;
	}
	
	@Override
	public void mergeAll(Byte2ObjectMap<V> m, ObjectObjectUnaryOperator<V, V> mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		for(Byte2ObjectMap.Entry<V> entry : Byte2ObjectMaps.fastIterable(m)) {
			byte key = entry.getByteKey();
			int index = findIndex(key);
			V newValue = index == -1 || Objects.equals(values[index], getDefaultReturnValue()) ? entry.getValue() : mappingFunction.apply(values[index], entry.getValue());
			if(Objects.equals(newValue, getDefaultReturnValue())) {
				if(index >= 0)
					removeIndex(index);
			}
			else if(index == -1) insertIndex(size++, key, newValue);
			else values[index] = newValue;
		}
	}
	
	@Override
	public int size() {
		return size;
	}
	
	@Override
	public void clear() {
		Arrays.fill(keys, 0, size, (byte)0);
		Arrays.fill(values, 0, size, null);
		size = 0;
	}
	
	public Byte2ObjectArrayMap<V> copy() {
		Byte2ObjectArrayMap<V> map = new Byte2ObjectArrayMap<>();
		map.size = size;
		map.keys = Arrays.copyOf(keys, keys.length);
		map.values = Arrays.copyOf(values, keys.length);
		return map;
	}
	
	@Override
	public ByteComparator comparator() {
		return null;
	}
	
	@Override
	public Byte2ObjectSortedMap<V> subMap(byte fromKey, byte toKey) {
		int fromIndex = findIndex(fromKey);
		int toIndex = findIndex(toKey);
		if(fromIndex == -1 || toIndex == -1) throw new NoSuchElementException();
		return new SubMap(fromIndex, toIndex - fromIndex + 1);
	}
	
	@Override
	public Byte2ObjectSortedMap<V> headMap(byte toKey) {
		int toIndex = findIndex(toKey);
		if(toIndex == -1) throw new NoSuchElementException();
		return new SubMap(0, toIndex + 1);
	}
	
	@Override
	public Byte2ObjectSortedMap<V> tailMap(byte fromKey) {
		int fromIndex = findIndex(fromKey);
		if(fromIndex == -1) throw new NoSuchElementException();
		return new SubMap(fromIndex, size - fromIndex);
	}
	
	protected void moveIndexToFirst(int index) {
		if(index == 0) return;
		byte key = keys[index];
		V value = values[index];
		System.arraycopy(keys, 0, keys, 1, index);
		System.arraycopy(values, 0, values, 1, index);
		keys[0] = key;
		values[0] = value;
	}
	
	protected void moveIndexToLast(int index) {
		if(index == size-1) return;
		byte key = keys[index];
		V value = values[index];
		System.arraycopy(keys, index+1, keys, index, size-index-1);
		System.arraycopy(values, index+1, values, index, size-index-1);
		keys[size-1] = key;
		values[size-1] = value;
	}
	
	protected void grow(int newSize) {
		if(newSize < keys.length) return;
		newSize = Math.min(newSize, keys.length == 0 ? 2 : keys.length * 2);
		keys = Arrays.copyOf(keys, newSize);
		values = Arrays.copyOf(values, newSize);
	}
	
	protected void insertIndex(int index, byte key, V value) {
		grow(size+1);
		if(index != size) {
			System.arraycopy(keys, index, keys, index+1, size-index);
			System.arraycopy(values, index, values, index+1, size-index);		
		}
		keys[index] = key;
		values[index] = value;
	}
	
	protected void removeRange(int from, int to) {
		if(from < 0 || from >= size) throw new IllegalStateException("From Element ");
		int length = to - from;
		if(length <= 0) return;
		if(to != size) {
			System.arraycopy(keys, to, keys, from, size - to);
			System.arraycopy(values, to, values, from, size - to);
		}
		for(int i = 0;i<length;i++) {
			keys[i+to] = (byte)0;
			values[i+to] = null;
		}
		size -= length;
	}
	
	protected void removeIndex(int index) {
		if(index == size-1) {
			size--;
			keys[size] = (byte)0;
			values[size] = null;
			return;
		}
		System.arraycopy(keys, index+1, keys, index, size-index-1);
		System.arraycopy(values, index+1, values, index, size-index-1);		
		size--;
		keys[size] = (byte)0;
		values[size] = null;
	}
	
	protected int findIndex(byte key, V value) {
		for(int i = size-1;i>=0;i--)
			if(keys[i] == key && Objects.equals(values[i], value)) return i;
		return -1;		
	}
	
	protected int findIndex(byte key) {
		for(int i = size-1;i>=0;i--)
			if(keys[i] == key) return i;
		return -1;
	}
	
	protected int findIndex(Object key, Object value) {
		for(int i = size-1;i>=0;i--)
			if(Objects.equals(key, Byte.valueOf(keys[i])) && Objects.equals(value, values[i])) return i;
		return -1;		
	}
	
	protected int findIndex(Object key) {
		for(int i = size-1;i>=0;i--)
			if(Objects.equals(key, Byte.valueOf(keys[i]))) return i;
		return -1;
	}
	
	protected int findValue(Object value) {
		for(int i = size-1;i>=0;i--)
			if(Objects.equals(value, values[i])) return i;
		return -1;
	}
	
	private class SubMap extends AbstractByte2ObjectMap<V> implements Byte2ObjectSortedMap<V> {
		int offset;
		int length;
		
		ByteSet keySet;
		ObjectCollection<V> valuesC;
		FastSortedSet<V> entrySet;
		
		public SubMap(int offset, int length) {
			this.offset = offset;
			this.length = length;
		}
		
		int end() { return offset+length; }
		
		@Override
		public V put(byte key, V value) {
			int index = findIndex(key);
			if(index < 0) {
				insertIndex(end(), key, value);
				length++;
				return getDefaultReturnValue();
			}
			V oldValue = values[index];
			values[index] = value;
			return oldValue;
		}
		
		@Override
		public V putIfAbsent(byte key, V value) {
			int index = findIndex(key);
			if(index >= 0) {
				insertIndex(end(), key, value);
				length++;
			}
			return getDefaultReturnValue();
		}
		
		@Override
		public V putAndMoveToFirst(byte key, V value) {
			int index = findIndex(key);
			if(index < 0) {
				insertIndex(offset, key, value);
				length++;
				return getDefaultReturnValue();
			}
			V lastValue = values[index];
			values[index] = value;
			moveIndexToFirst(index);
			return lastValue;
		}
		
		@Override
		public V putAndMoveToLast(byte key, V value) {
			int index = findIndex(key);
			if(index < 0) {
				insertIndex(end(), key, value);
				length++;
				return getDefaultReturnValue();
			}
			V lastValue = values[index];
			values[index] = value;
			moveIndexToLast(index);
			return lastValue;
		}
		
		@Override
		public boolean moveToFirst(byte key) {
			int index = findIndex(key);
			if(index > 0) {
				moveIndexToFirst(index);
				return true;
			}
			return false;
		}
		
		@Override
		public boolean moveToLast(byte key) {
			int index = findIndex(key);
			if(index < size()-1) {
				moveIndexToLast(index);
				return true;
			}
			return false;
		}
		
		@Override
		public boolean containsKey(byte key) {
			return findIndex(key) >= 0;
		}
		
		@Override
		public boolean containsKey(Object key) {
			return findIndex(key) >= 0;
		}
		
		@Override
		public boolean containsValue(Object value) {
			return findValue(value) >= 0;
		}
		
		@Override
		public V get(byte key) {
			int index = findIndex(key);
			return index < 0 ? getDefaultReturnValue() : values[index];
		}
		
		@Override
		public V getOrDefault(byte key, V defaultValue) {
			int index = findIndex(key);
			return index < 0 ? defaultValue : values[index];
		}
		
		@Override
		public V getAndMoveToFirst(byte key) {
			int index = findIndex(key);
			if(index >= 0) {
				V value = values[index];
				moveIndexToFirst(index);
				return value;
			}
			return getDefaultReturnValue();
		}
		
		@Override
		public V getAndMoveToLast(byte key) {
			int index = findIndex(key);
			if(index >= 0) {
				V value = values[index];
				moveIndexToLast(index);
				return value;
			}
			return getDefaultReturnValue();
		}
		
		@Override
		public byte firstByteKey() {
			if(length <= 0) throw new NoSuchElementException();
			return keys[offset];
		}
		
		@Override
		public byte lastByteKey() {
			if(length <= 0) throw new NoSuchElementException();
			return keys[end()-1];
		}
		
		@Override
		public V firstValue() {
			if(length <= 0) throw new NoSuchElementException();
			return values[end()-1];
		}
		
		@Override
		public V lastValue() {
			if(length <= 0) throw new NoSuchElementException();
			return values[offset];
		}
		
		@Override
		public byte pollFirstByteKey() {
			if(length == 0) throw new NoSuchElementException();
			byte result = keys[offset];
			removeIndex(offset);
			length--;
			return result;
		}
		
		@Override
		public byte pollLastByteKey() {
			if(length == 0) throw new NoSuchElementException();
			byte result = keys[end()-1];
			removeIndex(end()-1);
			length--;
			return result;
		}
		
		@Override
		public V remove(byte key) {
			int index = findIndex(key);
			if(index < 0) return getDefaultReturnValue();
			V value = values[index];
			removeIndex(index);
			length--;
			return value;
		}
		
		@Override
		public V removeOrDefault(byte key, V defaultValue) {
			int index = findIndex(key);
			if(index < 0) return defaultValue;
			V value = values[index];
			removeIndex(index);
			length--;
			return value;
		}
		
		@Override
		public boolean remove(byte key, V value) {
			int index = findIndex(key, value);
			if(index < 0) return false;
			removeIndex(index);
			length--;
			return true;
		}
		
		@Override
		public V remove(Object key) {
			int index = findIndex(key);
			if(index < 0) return getDefaultReturnValue();
			V value = values[index];
			removeIndex(index);
			length--;
			return value;
		}
		
		@Override
		public boolean remove(Object key, Object value) {
			int index = findIndex(key, value);
			if(index < 0) return false;
			removeIndex(index);
			length--;
			return true;
		}
		
		@Override
		public int size() {
			return length;
		}
		
		@Override
		public void clear() {
			removeRange(offset, offset+length);
			length = 0;
		}
		
		@Override
		public ByteSet keySet() {
			if(keySet == null) keySet = new SubKeySet();
			return keySet;
		}
		
		@Override
		public ObjectCollection<V> values() {
			if(valuesC == null) valuesC = new SubValues();
			return valuesC;
		}

		@Override
		public ObjectSet<Byte2ObjectMap.Entry<V>> byte2ObjectEntrySet() {
			if(entrySet == null) entrySet = new SubMapEntrySet();
			return entrySet;
		}
		
		@Override
		public SubMap copy() { 
			throw new UnsupportedOperationException(); 
		}
		
		@Override
		public ByteComparator comparator() {
			return null;
		}
		
		@Override
		public Byte2ObjectSortedMap<V> subMap(byte fromKey, byte toKey) {
			int fromIndex = findIndex(fromKey);
			int toIndex = findIndex(toKey);
			if(fromIndex == -1 || toIndex == -1) throw new NoSuchElementException();
			return new SubMap(fromIndex, toIndex - fromIndex + 1);
		}
		
		@Override
		public Byte2ObjectSortedMap<V> headMap(byte toKey) {
			int toIndex = findIndex(toKey);
			if(toIndex == -1) throw new NoSuchElementException();
			return new SubMap(offset, toIndex + 1);
		}
		
		@Override
		public Byte2ObjectSortedMap<V> tailMap(byte fromKey) {
			int fromIndex = findIndex(fromKey);
			if(fromIndex == -1) throw new NoSuchElementException();
			return new SubMap(fromIndex, size - fromIndex);
		}
		
		protected void moveIndexToFirst(int index) {
			if(index == 0) return;
			byte key = keys[index];
			V value = values[index];
			System.arraycopy(keys, offset, keys, offset+1, index);
			System.arraycopy(values, offset, values, offset+1, index);
			keys[offset] = key;
			values[offset] = value;
		}
		
		protected void moveIndexToLast(int index) {
			if(index == length-1) return;
			byte key = keys[index];
			V value = values[index];
			System.arraycopy(keys, index+1, keys, index, end()-index-1);
			System.arraycopy(values, index+1, values, index, end()-index-1);
			keys[end()-1] = key;
			values[end()-1] = value;
		}
		
		protected int findIndex(byte key, V value) {
			for(int i = length-1;i>=0;i--)
				if(keys[offset+i] == key && Objects.equals(values[offset+i], value)) return i;
			return -1;		
		}
		
		protected int findIndex(byte key) {
			for(int i = length-1;i>=0;i--)
				if(keys[offset+i] == key) return i+offset;
			return -1;
		}
		
		protected int findIndex(Object key, Object value) {
			for(int i = length-1;i>=0;i--)
				if(Objects.equals(key, Byte.valueOf(keys[offset+i])) && Objects.equals(value, values[offset+i])) return i+offset;
			return -1;		
		}
		
		protected int findIndex(Object key) {
			for(int i = length-1;i>=0;i--)
				if(Objects.equals(key, Byte.valueOf(keys[offset+i]))) return i+offset;
			return -1;
		}
		
		protected int findValue(Object value) {
			for(int i = length-1;i>=0;i--)
				if(Objects.equals(value, values[offset+i])) return i+offset;
			return -1;
		}
		
		private class SubMapEntrySet extends AbstractObjectSet<Byte2ObjectMap.Entry<V>> implements Byte2ObjectSortedMap.FastSortedSet<V> {
			@Override
			public boolean addAndMoveToFirst(Byte2ObjectMap.Entry<V> o) { throw new UnsupportedOperationException(); }
			@Override
			public boolean addAndMoveToLast(Byte2ObjectMap.Entry<V> o) { throw new UnsupportedOperationException(); }
			
			@Override
			public boolean moveToFirst(Byte2ObjectMap.Entry<V> o) {
				return SubMap.this.moveToFirst(o.getByteKey());
			}
			
			@Override
			public boolean moveToLast(Byte2ObjectMap.Entry<V> o) {
				return SubMap.this.moveToLast(o.getByteKey());
			}
			
			@Override
			public Byte2ObjectMap.Entry<V> first() {
				return new BasicEntry<>(firstByteKey(), firstValue());
			}
			
			@Override
			public Byte2ObjectMap.Entry<V> last() {
				return new BasicEntry<>(lastByteKey(), lastValue());
			}
			
			@Override
			public Byte2ObjectMap.Entry<V> pollFirst() {
				BasicEntry<V> entry = new BasicEntry<>(firstByteKey(), firstValue());
				pollFirstByteKey();
				return entry;
			}
			
			@Override
			public Byte2ObjectMap.Entry<V> pollLast() {
				BasicEntry<V> entry = new BasicEntry<>(lastByteKey(), lastValue());
				pollLastByteKey();
				return entry;
			}
			
			@Override
			public ObjectBidirectionalIterator<Byte2ObjectMap.Entry<V>> iterator() {
				return new SubEntryIterator();
			}
			
			@Override
			public ObjectBidirectionalIterator<Byte2ObjectMap.Entry<V>> iterator(Byte2ObjectMap.Entry<V> fromElement) {
				return new SubEntryIterator(fromElement.getByteKey());
			}
			
			@Override
			public ObjectBidirectionalIterator<Byte2ObjectMap.Entry<V>> fastIterator() {
				return new SubFastEntryIterator();
			}
			
			@Override
			public ObjectBidirectionalIterator<Byte2ObjectMap.Entry<V>> fastIterator(byte fromElement) {
				return new SubFastEntryIterator(fromElement);
			}
			
			@Override
			public SubMapEntrySet copy() { throw new UnsupportedOperationException(); }
			
			@Override
			public void forEach(Consumer<? super Byte2ObjectMap.Entry<V>> action) {
				Objects.requireNonNull(action);
				for(int i = 0;i<length;i++)
					action.accept(new BasicEntry<>(keys[offset+i], values[offset+i]));
			}
			
			@Override
			public void fastForEach(Consumer<? super Byte2ObjectMap.Entry<V>> action) {
				Objects.requireNonNull(action);
				if(size() <= 0) return;
				BasicEntry<V> entry = new BasicEntry<>();
				for(int i = 0;i<length;i++) {
					entry.set(keys[offset+i], values[offset+i]);
					action.accept(entry);
				}
			}
			
			@Override
			public <E> void forEach(E input, ObjectObjectConsumer<E, Byte2ObjectMap.Entry<V>> action) {
				Objects.requireNonNull(action);
				for(int i = 0;i<length;i++)
					action.accept(input, new BasicEntry<>(keys[offset+i], values[offset+i]));
			}
			
			@Override
			public boolean matchesAny(Object2BooleanFunction<Byte2ObjectMap.Entry<V>> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return false;
				BasicEntry<V> entry = new BasicEntry<>();
				for(int i = 0;i<length;i++) {
					entry.set(keys[offset+i], values[offset+i]);
					if(filter.getBoolean(entry)) return true;
				}
				return false;
			}
			
			@Override
			public boolean matchesNone(Object2BooleanFunction<Byte2ObjectMap.Entry<V>> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return true;
				BasicEntry<V> entry = new BasicEntry<>();
				for(int i = 0;i<length;i++) {
					entry.set(keys[offset+i], values[offset+i]);
					if(filter.getBoolean(entry)) return false;
				}
				return true;
			}
			
			@Override
			public boolean matchesAll(Object2BooleanFunction<Byte2ObjectMap.Entry<V>> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return true;
				BasicEntry<V> entry = new BasicEntry<>();
				for(int i = 0;i<length;i++) {
					entry.set(keys[offset+i], values[offset+i]);
					if(!filter.getBoolean(entry)) return false;
				}
				return true;
			}
			
			@Override
			public <E> E reduce(E identity, BiFunction<E, Byte2ObjectMap.Entry<V>, E> operator) {
				Objects.requireNonNull(operator);
				E state = identity;
				for(int i = 0;i<length;i++) {
					state = operator.apply(state, new BasicEntry<>(keys[offset+i], values[offset+i]));
				}
				return state;
			}
			
			@Override
			public Byte2ObjectMap.Entry<V> reduce(ObjectObjectUnaryOperator<Byte2ObjectMap.Entry<V>, Byte2ObjectMap.Entry<V>> operator) {
				Objects.requireNonNull(operator);
				Byte2ObjectMap.Entry<V> state = null;
				boolean empty = true;
				for(int i = 0;i<length;i++) {
					if(empty) {
						empty = false;
						state = new BasicEntry<>(keys[offset+i], values[offset+i]);
						continue;
					}
					state = operator.apply(state, new BasicEntry<>(keys[offset+i], values[offset+i]));
				}
				return state;
			}
			
			@Override
			public Byte2ObjectMap.Entry<V> findFirst(Object2BooleanFunction<Byte2ObjectMap.Entry<V>> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return null;
				BasicEntry<V> entry = new BasicEntry<>();
				for(int i = 0;i<length;i++) {
					entry.set(keys[offset+i], values[offset+i]);
					if(filter.getBoolean(entry)) return entry;
				}
				return null;
			}
			
			@Override
			public int count(Object2BooleanFunction<Byte2ObjectMap.Entry<V>> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return 0;
				int result = 0;
				BasicEntry<V> entry = new BasicEntry<>();
				for(int i = 0;i<length;i++) {
					entry.set(keys[offset+i], values[offset+i]);
					if(filter.getBoolean(entry)) result++;
				}
				return result;
			}
			
			@Override
			@Deprecated
			public boolean contains(Object o) {
				if(o instanceof Map.Entry) {
					if(o instanceof Byte2ObjectMap.Entry) {
						Byte2ObjectMap.Entry<V> entry = (Byte2ObjectMap.Entry<V>)o;
						return SubMap.this.findIndex(entry.getByteKey(), entry.getValue()) >= 0;
					}
					else {
						Map.Entry<?, ?> entry = (Map.Entry<?, ?>)o;
						return Byte2ObjectArrayMap.this.findIndex(entry.getKey(), entry.getValue()) >= 0;
					}
				}
				return false;
			}
			
			@Override
			@Deprecated
			public boolean remove(Object o) {
				if(o instanceof Map.Entry) {
					if(o instanceof Byte2ObjectMap.Entry) {
						Byte2ObjectMap.Entry<V> entry = (Byte2ObjectMap.Entry<V>)o;
						return SubMap.this.remove(entry.getByteKey(), entry.getValue());
					}
					Map.Entry<?, ?> entry = (Map.Entry<?, ?>)o;
					return SubMap.this.remove(entry.getKey(), entry.getValue());
				}
				return false;
			}
			
			@Override
			public int size() {
				return SubMap.this.size();
			}
			
			@Override
			public void clear() {
				SubMap.this.clear();
			}
			
			@Override
			public Comparator<Byte2ObjectMap.Entry<V>> comparator() {
				return null;
			}
			@Override
			public ObjectSortedSet<Byte2ObjectMap.Entry<V>> subSet(Byte2ObjectMap.Entry<V> fromElement, Byte2ObjectMap.Entry<V> toElement) { throw new UnsupportedOperationException(); }
			@Override
			public ObjectSortedSet<Byte2ObjectMap.Entry<V>> headSet(Byte2ObjectMap.Entry<V> toElement) { throw new UnsupportedOperationException(); }
			@Override
			public ObjectSortedSet<Byte2ObjectMap.Entry<V>> tailSet(Byte2ObjectMap.Entry<V> fromElement) { throw new UnsupportedOperationException(); }
		}
		
		private class SubKeySet extends AbstractByteSet implements ByteSortedSet {
			@Override
			public boolean contains(byte e) { return containsKey(e); }
			
			@Override
			public boolean remove(byte o) {
				int oldSize = length;
				SubMap.this.remove(o);
				return length != oldSize;
			}
			
			@Override
			public boolean add(byte o) { throw new UnsupportedOperationException(); }
			@Override
			public boolean addAndMoveToFirst(byte o) { throw new UnsupportedOperationException(); }
			@Override
			public boolean addAndMoveToLast(byte o) { throw new UnsupportedOperationException(); }
			@Override
			public boolean moveToFirst(byte o) { return SubMap.this.moveToFirst(o); }
			@Override
			public boolean moveToLast(byte o) { return SubMap.this.moveToLast(o); }
			@Override
			public ByteListIterator iterator() { return new SubKeyIterator(); }
			@Override
			public ByteBidirectionalIterator iterator(byte fromElement) { return new SubKeyIterator(fromElement); } 
			@Override
			public int size() { return SubMap.this.size(); }
			@Override
			public void clear() { SubMap.this.clear(); }
			@Override
			public byte firstByte() { return firstByteKey(); }
			@Override
			public byte pollFirstByte() { return pollFirstByteKey(); }
			@Override
			public byte lastByte() { return lastByteKey(); }
			@Override
			public byte pollLastByte() { return pollLastByteKey(); }
			
			@Override
			public SubKeySet copy() { throw new UnsupportedOperationException(); }
			
			@Override
			public void forEach(ByteConsumer action) { 
				Objects.requireNonNull(action);
				for(int i = 0;i<length;action.accept(keys[offset+i])); 
			}
			
			@Override
			public <E> void forEach(E input, ObjectByteConsumer<E> action) {
				Objects.requireNonNull(action);
				if(size() <= 0) return;
				for(int i = 0;i<length;action.accept(input, keys[offset+i])); 

			}
			
			@Override
			public boolean matchesAny(Byte2BooleanFunction filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return false;
				for(int i = 0;i<length;i++) {
					if(filter.get(keys[i+offset])) return true;					
				}
				return false;
			}
			
			@Override
			public boolean matchesNone(Byte2BooleanFunction filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return true;
				for(int i = 0;i<length;i++) {
					if(filter.get(keys[i+offset])) return false;					
				}
				return true;
			}
			
			@Override
			public boolean matchesAll(Byte2BooleanFunction filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return true;
				for(int i = 0;i<length;i++) {
					if(!filter.get(keys[i+offset])) return false;					
				}
				return true;
			}
			
			@Override
			public byte reduce(byte identity, ByteByteUnaryOperator operator) {
				Objects.requireNonNull(operator);
				byte state = identity;
				for(int i = 0;i<length;i++) {
					state = operator.applyAsByte(state, keys[i+offset]);
				}
				return state;
			}
			
			@Override
			public byte reduce(ByteByteUnaryOperator operator) {
				Objects.requireNonNull(operator);
				byte state = (byte)0;
				boolean empty = true;
				for(int i = 0;i<length;i++) {
					if(empty) {
						empty = false;
						state = keys[i+offset];
						continue;
					}
					state = operator.applyAsByte(state, keys[i+offset]);
				}
				return state;
			}
			
			@Override
			public byte findFirst(Byte2BooleanFunction filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return (byte)0;
				for(int i = 0;i<length;i++) {
					if(filter.get(keys[i+offset])) return keys[i+offset];					
				}
				return (byte)0;
			}
			
			@Override
			public int count(Byte2BooleanFunction filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return 0;
				int result = 0;
				for(int i = 0;i<length;i++) {
					if(filter.get(keys[i+offset])) result++;					
				}
				return result;
			}
			
			@Override
			public ByteComparator comparator() { return null; }
			@Override
			public ByteSortedSet subSet(byte fromElement, byte toElement) { throw new UnsupportedOperationException(); }
			@Override
			public ByteSortedSet headSet(byte toElement) { throw new UnsupportedOperationException(); }
			@Override
			public ByteSortedSet tailSet(byte fromElement) { throw new UnsupportedOperationException(); }
		}
		
		private class SubValues extends AbstractObjectCollection<V> {
			@Override
			public boolean contains(Object e) {
				return containsValue(e);
			}

			@Override
			public boolean add(V o) {
				throw new UnsupportedOperationException();
			}
			
			@Override
			public ObjectIterator<V> iterator() {
				return new SubValueIterator();
			}
			
			@Override
			public int size() {
				return SubMap.this.size();
			}
			
			@Override
			public void clear() {
				SubMap.this.clear();
			}
			
			@Override
			public void forEach(Consumer<? super V> action) {
				Objects.requireNonNull(action);
				for(int i = 0;i<length;action.accept(values[offset+i]));
			}
			
			@Override
			public <E> void forEach(E input, ObjectObjectConsumer<E, V> action) {
				Objects.requireNonNull(action);
				for(int i = 0;i<length;action.accept(input, values[offset+i]));

			}
			
			@Override
			public boolean matchesAny(Object2BooleanFunction<V> filter) {
				Objects.requireNonNull(filter);
				for(int i = 0;i<length;i++) {
					if(filter.getBoolean(values[offset+i])) return true;
				}
				return false;
			}
			
			@Override
			public boolean matchesNone(Object2BooleanFunction<V> filter) {
				Objects.requireNonNull(filter);
				for(int i = 0;i<length;i++) {
					if(filter.getBoolean(values[offset+i])) return false;
				}
				return true;
			}
			
			@Override
			public boolean matchesAll(Object2BooleanFunction<V> filter) {
				Objects.requireNonNull(filter);
				for(int i = 0;i<length;i++) {
					if(!filter.getBoolean(values[offset+i])) return false;
				}
				return true;
			}
			
			@Override
			public <E> E reduce(E identity, BiFunction<E, V, E> operator) {
				Objects.requireNonNull(operator);
				E state = identity;
				for(int i = 0;i<length;i++) {
					state = operator.apply(state, values[offset+i]);
				}
				return state;
			}
			
			@Override
			public V reduce(ObjectObjectUnaryOperator<V, V> operator) {
				Objects.requireNonNull(operator);
				V state = null;
				boolean empty = true;
				for(int i = 0;i<length;i++) {
					if(empty) {
						empty = false;
						state = values[offset+i];
						continue;
					}
					state = operator.apply(state, values[offset+i]);
				}
				return state;
			}
			
			@Override
			public V findFirst(Object2BooleanFunction<V> filter) {
				Objects.requireNonNull(filter);
				for(int i = 0;i<length;i++) {
					if(filter.getBoolean(values[offset+i])) return values[offset+i];
				}
				return null;
			}
			
			@Override
			public int count(Object2BooleanFunction<V> filter) {
				Objects.requireNonNull(filter);
				int result = 0;
				for(int i = 0;i<length;i++) {
					if(filter.getBoolean(values[offset+i])) result++;
				}
				return result;
			}
		}
		
		private class SubFastEntryIterator extends SubMapIterator implements ObjectListIterator<Byte2ObjectMap.Entry<V>> {
			MapEntry entry = new MapEntry();
			
			public SubFastEntryIterator() {}
			public SubFastEntryIterator(byte from) {
				index = findIndex(from);
			}
			
			@Override
			public Byte2ObjectMap.Entry<V> next() {
				entry.index = nextEntry();
				return entry;
			}
			
			@Override
			public Byte2ObjectMap.Entry<V> previous() {
				entry.index = previousEntry();
				return entry;
			}
			
			@Override
			public void set(Byte2ObjectMap.Entry<V> e) { throw new UnsupportedOperationException(); }
			@Override
			public void add(Byte2ObjectMap.Entry<V> e) { throw new UnsupportedOperationException(); }
		}
		
		private class SubEntryIterator extends SubMapIterator implements ObjectListIterator<Byte2ObjectMap.Entry<V>> {
			MapEntry entry = null;
			
			public SubEntryIterator() {}
			public SubEntryIterator(byte from) {
				index = findIndex(from);
			}
			
			@Override
			public Byte2ObjectMap.Entry<V> next() {
				return entry = new MapEntry(nextEntry());
			}
			
			@Override
			public Byte2ObjectMap.Entry<V> previous() {
				return entry = new MapEntry(previousEntry());
			}
			
			@Override
			public void remove() {
				super.remove();
				entry.index = -1;
			}
			
			@Override
			public void set(Byte2ObjectMap.Entry<V> e) { throw new UnsupportedOperationException(); }
			@Override
			public void add(Byte2ObjectMap.Entry<V> e) { throw new UnsupportedOperationException(); }
		}
		
		private class SubKeyIterator extends SubMapIterator implements ByteListIterator {
			public SubKeyIterator() {}
			public SubKeyIterator(byte element) {
				index = findIndex(element);
			}
			@Override
			public byte previousByte() {
				return keys[previousEntry()];
			}

			@Override
			public byte nextByte() {
				return keys[nextEntry()];
			}

			@Override
			public void set(byte e) { throw new UnsupportedOperationException(); }

			@Override
			public void add(byte e) { throw new UnsupportedOperationException(); }	
		}
		
		private class SubValueIterator extends SubMapIterator implements ObjectListIterator<V> {
			@Override
			public V previous() {
				return values[previousEntry()];
			}

			@Override
			public V next() {
				return values[nextEntry()];
			}

			@Override
			public void set(V e) { throw new UnsupportedOperationException(); }

			@Override
			public void add(V e) { throw new UnsupportedOperationException(); }	
		}
		
		private class SubMapIterator {
			int index;
			int lastReturned = -1;

			public boolean hasNext() {
				return index < length;
			}
			
			public boolean hasPrevious() {
				return index > 0;
			}
			
			public int nextIndex() {
				return index;
			}
			
			public int previousIndex() {
				return index-1;
			}
			
			public void remove() {
				if(lastReturned == -1)
					throw new IllegalStateException();
				removeIndex(lastReturned);
				if(lastReturned < index)
					index--;
				lastReturned = -1;
			}
			
			public int previousEntry() {
				if(!hasPrevious()) throw new NoSuchElementException();
				int returnIndex = offset+index;
				lastReturned = index--;
				return returnIndex;
			}
			
			public int nextEntry() {
				if(!hasNext()) throw new NoSuchElementException();
				int returnIndex = offset+index;
				lastReturned = index++;
				return returnIndex;
			}
			
			public int skip(int amount) {
				if(amount < 0) throw new IllegalStateException("Negative Numbers are not allowed");
				int steps = Math.min(amount, (size() - 1) - index);
				index += steps;
				return steps;
			}
			
			public int back(int amount) {
				if(amount < 0) throw new IllegalStateException("Negative Numbers are not allowed");
				int steps = Math.min(amount, index);
				index -= steps;
				return steps;
			}
		}
	}
	
	private class MapEntrySet extends AbstractObjectSet<Byte2ObjectMap.Entry<V>> implements Byte2ObjectSortedMap.FastSortedSet<V> {
		@Override
		public boolean addAndMoveToFirst(Byte2ObjectMap.Entry<V> o) { throw new UnsupportedOperationException(); }
		@Override
		public boolean addAndMoveToLast(Byte2ObjectMap.Entry<V> o) { throw new UnsupportedOperationException(); }
		
		@Override
		public boolean moveToFirst(Byte2ObjectMap.Entry<V> o) {
			return Byte2ObjectArrayMap.this.moveToFirst(o.getByteKey());
		}
		
		@Override
		public boolean moveToLast(Byte2ObjectMap.Entry<V> o) {
			return Byte2ObjectArrayMap.this.moveToLast(o.getByteKey());
		}
		
		@Override
		public Byte2ObjectMap.Entry<V> first() {
			return new BasicEntry<>(firstByteKey(), firstValue());
		}
		
		@Override
		public Byte2ObjectMap.Entry<V> last() {
			return new BasicEntry<>(lastByteKey(), lastValue());
		}
		
		@Override
		public Byte2ObjectMap.Entry<V> pollFirst() {
			BasicEntry<V> entry = new BasicEntry<>(firstByteKey(), firstValue());
			pollFirstByteKey();
			return entry;
		}
		
		@Override
		public Byte2ObjectMap.Entry<V> pollLast() {
			BasicEntry<V> entry = new BasicEntry<>(lastByteKey(), lastValue());
			pollLastByteKey();
			return entry;
		}
		
		@Override
		public ObjectBidirectionalIterator<Byte2ObjectMap.Entry<V>> iterator() {
			return new EntryIterator();
		}
		
		@Override
		public ObjectBidirectionalIterator<Byte2ObjectMap.Entry<V>> iterator(Byte2ObjectMap.Entry<V> fromElement) {
			return new EntryIterator(fromElement.getByteKey());
		}
		
		@Override
		public ObjectBidirectionalIterator<Byte2ObjectMap.Entry<V>> fastIterator() {
			return new FastEntryIterator();
		}
		
		@Override
		public ObjectBidirectionalIterator<Byte2ObjectMap.Entry<V>> fastIterator(byte fromElement) {
			return new FastEntryIterator(fromElement);
		}
		
		@Override
		public MapEntrySet copy() { throw new UnsupportedOperationException(); }
		
		@Override
		public void forEach(Consumer<? super Byte2ObjectMap.Entry<V>> action) {
			Objects.requireNonNull(action);
			for(int i = 0;i<size;i++)
				action.accept(new BasicEntry<>(keys[i], values[i]));
		}
		
		@Override
		public void fastForEach(Consumer<? super Byte2ObjectMap.Entry<V>> action) {
			Objects.requireNonNull(action);
			if(size() <= 0) return;
			BasicEntry<V> entry = new BasicEntry<>();
			for(int i = 0;i<size;i++) {
				entry.set(keys[i], values[i]);
				action.accept(entry);
			}
		}
		
		@Override
		public <E> void forEach(E input, ObjectObjectConsumer<E, Byte2ObjectMap.Entry<V>> action) {
			Objects.requireNonNull(action);
			for(int i = 0;i<size;i++) {
				action.accept(input, new BasicEntry<>(keys[i], values[i]));
			}
		}
		
		@Override
		public boolean matchesAny(Object2BooleanFunction<Byte2ObjectMap.Entry<V>> filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return false;
			BasicEntry<V> entry = new BasicEntry<>();
			for(int i = 0;i<size;i++) {
				entry.set(keys[i], values[i]);
				if(filter.getBoolean(entry)) return true;
			}
			return false;
		}
		
		@Override
		public boolean matchesNone(Object2BooleanFunction<Byte2ObjectMap.Entry<V>> filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return true;
			BasicEntry<V> entry = new BasicEntry<>();
			for(int i = 0;i<size;i++) {
				entry.set(keys[i], values[i]);
				if(filter.getBoolean(entry)) return false;
			}
			return true;
		}
		
		@Override
		public boolean matchesAll(Object2BooleanFunction<Byte2ObjectMap.Entry<V>> filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return true;
			BasicEntry<V> entry = new BasicEntry<>();
			for(int i = 0;i<size;i++) {
				entry.set(keys[i], values[i]);
				if(!filter.getBoolean(entry)) return false;
			}
			return true;
		}
		
		@Override
		public <E> E reduce(E identity, BiFunction<E, Byte2ObjectMap.Entry<V>, E> operator) {
			Objects.requireNonNull(operator);
			E state = identity;
			for(int i = 0;i<size;i++) {
				state = operator.apply(state, new BasicEntry<>(keys[i], values[i]));
			}
			return state;
		}
		
		@Override
		public Byte2ObjectMap.Entry<V> reduce(ObjectObjectUnaryOperator<Byte2ObjectMap.Entry<V>, Byte2ObjectMap.Entry<V>> operator) {
			Objects.requireNonNull(operator);
			Byte2ObjectMap.Entry<V> state = null;
			boolean empty = true;
			for(int i = 0;i<size;i++) {
				if(empty) {
					empty = false;
					state = new BasicEntry<>(keys[i], values[i]);
					continue;
				}
				state = operator.apply(state, new BasicEntry<>(keys[i], values[i]));
			}
			return state;
		}
		
		@Override
		public Byte2ObjectMap.Entry<V> findFirst(Object2BooleanFunction<Byte2ObjectMap.Entry<V>> filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return null;
			BasicEntry<V> entry = new BasicEntry<>();
			for(int i = 0;i<size;i++) {
				entry.set(keys[i], values[i]);
				if(filter.getBoolean(entry)) return entry;
			}
			return null;
		}
		
		@Override
		public int count(Object2BooleanFunction<Byte2ObjectMap.Entry<V>> filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return 0;
			int result = 0;
			BasicEntry<V> entry = new BasicEntry<>();
			for(int i = 0;i<size;i++) {
				entry.set(keys[i], values[i]);
				if(filter.getBoolean(entry)) result++;
			}
			return result;
		}
		
		@Override
		@Deprecated
		public boolean contains(Object o) {
			if(o instanceof Map.Entry) {
				if(o instanceof Byte2ObjectMap.Entry) {
					Byte2ObjectMap.Entry<V> entry = (Byte2ObjectMap.Entry<V>)o;
					int index = Byte2ObjectArrayMap.this.findIndex(entry.getByteKey());
					if(index >= 0) return Objects.equals(entry.getValue(), Byte2ObjectArrayMap.this.values[index]);
				}
				else {
					Map.Entry<?, ?> entry = (Map.Entry<?, ?>)o;
					int index = Byte2ObjectArrayMap.this.findIndex(entry.getKey());
					if(index >= 0) return Objects.equals(entry.getValue(), Byte2ObjectArrayMap.this.values[index]);
				}
			}
			return false;
		}
		
		@Override
		@Deprecated
		public boolean remove(Object o) {
			if(o instanceof Map.Entry) {
				if(o instanceof Byte2ObjectMap.Entry) {
					Byte2ObjectMap.Entry<V> entry = (Byte2ObjectMap.Entry<V>)o;
					return Byte2ObjectArrayMap.this.remove(entry.getByteKey(), entry.getValue());
				}
				Map.Entry<?, ?> entry = (Map.Entry<?, ?>)o;
				return Byte2ObjectArrayMap.this.remove(entry.getKey(), entry.getValue());
			}
			return false;
		}
		
		@Override
		public int size() {
			return Byte2ObjectArrayMap.this.size();
		}
		
		@Override
		public void clear() {
			Byte2ObjectArrayMap.this.clear();
		}
		
		@Override
		public Comparator<Byte2ObjectMap.Entry<V>> comparator() {
			return null;
		}
		@Override
		public ObjectSortedSet<Byte2ObjectMap.Entry<V>> subSet(Byte2ObjectMap.Entry<V> fromElement, Byte2ObjectMap.Entry<V> toElement) { throw new UnsupportedOperationException(); }
		@Override
		public ObjectSortedSet<Byte2ObjectMap.Entry<V>> headSet(Byte2ObjectMap.Entry<V> toElement) { throw new UnsupportedOperationException(); }
		@Override
		public ObjectSortedSet<Byte2ObjectMap.Entry<V>> tailSet(Byte2ObjectMap.Entry<V> fromElement) { throw new UnsupportedOperationException(); }
	}
	
	private class KeySet extends AbstractByteSet implements ByteSortedSet {
		@Override
		public boolean contains(byte e) { return containsKey(e); }
		
		@Override
		public boolean remove(byte o) {
			int oldSize = size;
			Byte2ObjectArrayMap.this.remove(o);
			return size != oldSize;
		}
		
		@Override
		public boolean add(byte o) { throw new UnsupportedOperationException(); }
		@Override
		public boolean addAndMoveToFirst(byte o) { throw new UnsupportedOperationException(); }
		@Override
		public boolean addAndMoveToLast(byte o) { throw new UnsupportedOperationException(); }
		@Override
		public boolean moveToFirst(byte o) { return Byte2ObjectArrayMap.this.moveToFirst(o); }
		@Override
		public boolean moveToLast(byte o) { return Byte2ObjectArrayMap.this.moveToLast(o); }
		@Override
		public ByteListIterator iterator() { return new KeyIterator(); }
		@Override
		public ByteBidirectionalIterator iterator(byte fromElement) { return new KeyIterator(fromElement); } 
		@Override
		public int size() { return Byte2ObjectArrayMap.this.size(); }
		@Override
		public void clear() { Byte2ObjectArrayMap.this.clear(); }
		@Override
		public byte firstByte() { return firstByteKey(); }
		@Override
		public byte pollFirstByte() { return pollFirstByteKey(); }
		@Override
		public byte lastByte() { return lastByteKey(); }
		@Override
		public byte pollLastByte() { return pollLastByteKey(); }
		
		@Override
		public KeySet copy() { throw new UnsupportedOperationException(); }
		
		@Override
		public void forEach(ByteConsumer action) { 
			Objects.requireNonNull(action);
			for(int i = 0;i<size;action.accept(keys[i++])); 
		}
		
		@Override
		public <E> void forEach(E input, ObjectByteConsumer<E> action) {
			Objects.requireNonNull(action);
			for(int i = 0;i<size;action.accept(input, keys[i++])); 

		}
		
		@Override
		public boolean matchesAny(Byte2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return false;
			for(int i = 0;i<size;i++) {
				if(filter.get(keys[i])) return true;
			}
			return false;
		}
		
		@Override
		public boolean matchesNone(Byte2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(int i = 0;i<size;i++) {
				if(filter.get(keys[i])) return false;
			}
			return true;
		}
		
		@Override
		public boolean matchesAll(Byte2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(int i = 0;i<size;i++) {
				if(!filter.get(keys[i])) return false;
			}
			return true;
		}
		
		@Override
		public byte reduce(byte identity, ByteByteUnaryOperator operator) {
			Objects.requireNonNull(operator);
			byte state = identity;
			for(int i = 0;i<size;i++) {
				state = operator.applyAsByte(state, keys[i]);
			}
			return state;
		}
		
		@Override
		public byte reduce(ByteByteUnaryOperator operator) {
			Objects.requireNonNull(operator);
			byte state = (byte)0;
			boolean empty = true;
			for(int i = 0;i<size;i++) {
				if(empty) {
					empty = false;
					state = keys[i];
					continue;
				}
				state = operator.applyAsByte(state, keys[i]);
			}
			return state;
		}
		
		@Override
		public byte findFirst(Byte2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(int i = 0;i<size;i++) {
				if(filter.get(keys[i])) return keys[i];
			}
			return (byte)0;
		}
		
		@Override
		public int count(Byte2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			int result = 0;
			for(int i = 0;i<size;i++) {
				if(filter.get(keys[i])) result++;
			}
			return result;
		}
		
		@Override
		public ByteComparator comparator() { return null; }
		@Override
		public ByteSortedSet subSet(byte fromElement, byte toElement) { throw new UnsupportedOperationException(); }
		@Override
		public ByteSortedSet headSet(byte toElement) { throw new UnsupportedOperationException(); }
		@Override
		public ByteSortedSet tailSet(byte fromElement) { throw new UnsupportedOperationException(); }
	}
	
	private class Values extends AbstractObjectCollection<V> {
		@Override
		public boolean contains(Object e) {
			return containsValue(e);
		}
		
		@Override
		public boolean add(V o) {
			throw new UnsupportedOperationException();
		}

		@Override
		public ObjectIterator<V> iterator() {
			return new ValueIterator();
		}
		
		@Override
		public int size() {
			return Byte2ObjectArrayMap.this.size();
		}
		
		@Override
		public void clear() {
			Byte2ObjectArrayMap.this.clear();
		}
		
		@Override
		public void forEach(Consumer<? super V> action) {
			Objects.requireNonNull(action);
			for(int i = 0;i<size;action.accept(values[i++]));
		}
		
		@Override
		public boolean matchesAny(Object2BooleanFunction<V> filter) {
			Objects.requireNonNull(filter);
			for(int i = 0;i<size;i++) {
				if(filter.getBoolean(values[i])) return true;
			}
			return false;
		}
		
		@Override
		public boolean matchesNone(Object2BooleanFunction<V> filter) {
			Objects.requireNonNull(filter);
			for(int i = 0;i<size;i++) {
				if(filter.getBoolean(values[i])) return false;
			}
			return true;
		}
		
		@Override
		public boolean matchesAll(Object2BooleanFunction<V> filter) {
			Objects.requireNonNull(filter);
			for(int i = 0;i<size;i++) {
				if(!filter.getBoolean(values[i])) return false;
			}
			return true;
		}
		
		@Override
		public <E> E reduce(E identity, BiFunction<E, V, E> operator) {
			Objects.requireNonNull(operator);
			E state = identity;
			for(int i = 0;i<size;i++) {
				state = operator.apply(state, values[i]);
			}
			return state;
		}
		
		@Override
		public V reduce(ObjectObjectUnaryOperator<V, V> operator) {
			Objects.requireNonNull(operator);
			V state = null;
			boolean empty = true;
			for(int i = 0;i<size;i++) {
				if(empty) {
					empty = false;
					state = values[i];
					continue;
				}
				state = operator.apply(state, values[i]);
			}
			return state;
		}
		
		@Override
		public V findFirst(Object2BooleanFunction<V> filter) {
			Objects.requireNonNull(filter);
			for(int i = 0;i<size;i++) {
				if(filter.getBoolean(values[i])) return values[i];
			}
			return null;
		}
		
		@Override
		public int count(Object2BooleanFunction<V> filter) {
			Objects.requireNonNull(filter);
			int result = 0;
			for(int i = 0;i<size;i++) {
				if(filter.getBoolean(values[i])) result++;
			}
			return result;
		}
	}
	
	private class FastEntryIterator extends MapIterator implements ObjectListIterator<Byte2ObjectMap.Entry<V>> {
		MapEntry entry = new MapEntry();
		
		public FastEntryIterator() {}
		public FastEntryIterator(byte from) {
			index = findIndex(from);
		}
		
		@Override
		public Byte2ObjectMap.Entry<V> next() {
			entry.index = nextEntry();
			return entry;
		}
		
		@Override
		public Byte2ObjectMap.Entry<V> previous() {
			entry.index = previousEntry();
			return entry;
		}
		
		@Override
		public void set(Byte2ObjectMap.Entry<V> e) { throw new UnsupportedOperationException(); }
		@Override
		public void add(Byte2ObjectMap.Entry<V> e) { throw new UnsupportedOperationException(); }
	}
	
	private class EntryIterator extends MapIterator implements ObjectListIterator<Byte2ObjectMap.Entry<V>> {
		MapEntry entry = null;
		
		public EntryIterator() {}
		public EntryIterator(byte from) {
			index = findIndex(from);
		}
		
		@Override
		public Byte2ObjectMap.Entry<V> next() {
			return entry = new MapEntry(nextEntry());
		}
		
		@Override
		public Byte2ObjectMap.Entry<V> previous() {
			return entry = new MapEntry(previousEntry());
		}
		
		@Override
		public void remove() {
			super.remove();
			entry.index = -1;
		}
		
		@Override
		public void set(Byte2ObjectMap.Entry<V> e) { throw new UnsupportedOperationException(); }
		@Override
		public void add(Byte2ObjectMap.Entry<V> e) { throw new UnsupportedOperationException(); }
	}
	
	private class KeyIterator extends MapIterator implements ByteListIterator {
		public KeyIterator() {}
		public KeyIterator(byte element) {
			index = findIndex(element);
		}
		@Override
		public byte previousByte() {
			return keys[previousEntry()];
		}

		@Override
		public byte nextByte() {
			return keys[nextEntry()];
		}

		@Override
		public void set(byte e) { throw new UnsupportedOperationException(); }

		@Override
		public void add(byte e) { throw new UnsupportedOperationException(); }	
	}
	
	private class ValueIterator extends MapIterator implements ObjectListIterator<V> {
		@Override
		public V previous() {
			return values[previousEntry()];
		}

		@Override
		public V next() {
			return values[nextEntry()];
		}

		@Override
		public void set(V e) { throw new UnsupportedOperationException(); }

		@Override
		public void add(V e) { throw new UnsupportedOperationException(); }	
	}
	
	private class MapIterator {
		int index;
		int lastReturned = -1;

		public boolean hasNext() {
			return index < size;
		}
		
		public boolean hasPrevious() {
			return index > 0;
		}
		
		public int nextIndex() {
			return index;
		}
		
		public int previousIndex() {
			return index-1;
		}
		
		public void remove() {
			if(lastReturned == -1)
				throw new IllegalStateException();
			removeIndex(lastReturned);
			if(lastReturned < index)
				index--;
			lastReturned = -1;
		}
		
		public int previousEntry() {
			if(!hasPrevious()) throw new NoSuchElementException();
			lastReturned = index;
			return index--;
		}
		
		public int nextEntry() {
			if(!hasNext()) throw new NoSuchElementException();
			lastReturned = index;
			return index++;
		}
		
		public int skip(int amount) {
			if(amount < 0) throw new IllegalStateException("Negative Numbers are not allowed");
			int steps = Math.min(amount, (size() - 1) - index);
			index += steps;
			return steps;
		}
		
		public int back(int amount) {
			if(amount < 0) throw new IllegalStateException("Negative Numbers are not allowed");
			int steps = Math.min(amount, index);
			index -= steps;
			return steps;
		}
	}
	
	private class MapEntry implements Byte2ObjectMap.Entry<V>, Map.Entry<Byte, V> {
		int index = -1;
		
		public MapEntry() {}
		public MapEntry(int index) {
			this.index = index;
		}

		@Override
		public byte getByteKey() {
			return keys[index];
		}

		@Override
		public V getValue() {
			return values[index];
		}

		@Override
		public V setValue(V value) {
			V oldValue = values[index];
			values[index] = value;
			return oldValue;
		}
		
		@Override
		public boolean equals(Object obj) {
			if(obj instanceof Map.Entry) {
				if(obj instanceof Byte2ObjectMap.Entry) {
					Byte2ObjectMap.Entry<V> entry = (Byte2ObjectMap.Entry<V>)obj;
					return keys[index] == entry.getByteKey() && Objects.equals(values[index], entry.getValue());
				}
				Map.Entry<?, ?> entry = (Map.Entry<?, ?>)obj;
				Object key = entry.getKey();
				Object value = entry.getValue();
				return key instanceof Byte && keys[index] == ((Byte)key).byteValue() && Objects.equals(values[index], value);
			}
			return false;
		}
		
		@Override
		public int hashCode() {
			return Byte.hashCode(keys[index]) ^ Objects.hashCode(values[index]);
		}
		
		@Override
		public String toString() {
			return Byte.toString(keys[index]) + "=" + Objects.toString(values[index]);
		}
	}
}