package speiger.src.collections.bytes.maps.impl.tree;

import java.util.Map;
import java.util.Objects;
import java.util.NoSuchElementException;
import java.util.function.Consumer;
import java.util.function.BiFunction;

import speiger.src.collections.bytes.collections.ByteBidirectionalIterator;
import speiger.src.collections.bytes.functions.ByteComparator;
import speiger.src.collections.bytes.functions.ByteConsumer;
import speiger.src.collections.objects.functions.consumer.ObjectByteConsumer;
import speiger.src.collections.bytes.functions.function.Byte2BooleanFunction;
import speiger.src.collections.bytes.functions.consumer.ByteDoubleConsumer;
import speiger.src.collections.bytes.functions.function.Byte2DoubleFunction;
import speiger.src.collections.bytes.functions.function.ByteDoubleUnaryOperator;
import speiger.src.collections.bytes.functions.function.ByteByteUnaryOperator;
import speiger.src.collections.bytes.lists.ByteListIterator;
import speiger.src.collections.bytes.maps.abstracts.AbstractByte2DoubleMap;
import speiger.src.collections.bytes.maps.interfaces.Byte2DoubleMap;
import speiger.src.collections.bytes.maps.interfaces.Byte2DoubleNavigableMap;
import speiger.src.collections.bytes.sets.AbstractByteSet;
import speiger.src.collections.bytes.sets.ByteNavigableSet;
import speiger.src.collections.bytes.sets.ByteSet;
import speiger.src.collections.bytes.sets.ByteSortedSet;
import speiger.src.collections.bytes.utils.ByteIterators;
import speiger.src.collections.bytes.utils.maps.Byte2DoubleMaps;
import speiger.src.collections.doubles.collections.AbstractDoubleCollection;
import speiger.src.collections.doubles.collections.DoubleCollection;
import speiger.src.collections.doubles.collections.DoubleIterator;
import speiger.src.collections.doubles.functions.DoubleSupplier;
import speiger.src.collections.doubles.functions.function.DoubleDoubleUnaryOperator;
import speiger.src.collections.doubles.lists.DoubleListIterator;
import speiger.src.collections.doubles.utils.DoubleIterators;
import speiger.src.collections.doubles.functions.DoubleConsumer;
import speiger.src.collections.objects.functions.consumer.ObjectObjectConsumer;
import speiger.src.collections.objects.functions.function.Object2BooleanFunction;
import speiger.src.collections.objects.functions.consumer.ObjectDoubleConsumer;
import speiger.src.collections.doubles.functions.function.Double2BooleanFunction;
import speiger.src.collections.objects.lists.ObjectListIterator;
import speiger.src.collections.objects.utils.ObjectIterators;
import speiger.src.collections.objects.collections.ObjectIterator;
import speiger.src.collections.objects.functions.function.ObjectObjectUnaryOperator;
import speiger.src.collections.objects.sets.AbstractObjectSet;
import speiger.src.collections.objects.sets.ObjectSet;

/**
 * A Simple Type Specific RB TreeMap implementation that reduces boxing/unboxing.
 * It is using a bit more memory then <a href="https://github.com/vigna/fastutil">FastUtil</a>,
 * but it saves a lot of Performance on the Optimized removal and iteration logic.
 * Which makes the implementation actually useable and does not get outperformed by Javas default implementation.
 */
public class Byte2DoubleRBTreeMap extends AbstractByte2DoubleMap implements Byte2DoubleNavigableMap
{
	/** The center of the Tree */
	protected transient Entry tree;
	/** The Lowest possible Node */
	protected transient Entry first;
	/** The Highest possible Node */
	protected transient Entry last;
	/** The amount of elements stored in the Map */
	protected int size = 0;
	/** The Sorter of the Tree */
	protected transient ByteComparator comparator;
	
	/** the default return value for max searches */
	protected byte defaultMaxNotFound = Byte.MIN_VALUE;
	/** the default return value for min searches */
	protected byte defaultMinNotFound = Byte.MAX_VALUE;
	
	/** KeySet Cache */
	protected ByteNavigableSet keySet;
	/** Values Cache */
	protected DoubleCollection values;
	/** EntrySet Cache */
	protected ObjectSet<Byte2DoubleMap.Entry> entrySet;
	
	/**
	 * Default Constructor
	 */
	public Byte2DoubleRBTreeMap() {
	}
	
	/**
	 * Constructor that allows to define the sorter
	 * @param comp the function that decides how the tree is sorted, can be null
	 */
	public Byte2DoubleRBTreeMap(ByteComparator comp) {
		comparator = comp;
	}
	
	/**
	 * Helper constructor that allow to create a map from boxed values (it will unbox them)
	 * @param keys the keys that should be put into the map
	 * @param values the values that should be put into the map.
	 * @throws IllegalStateException if the keys and values do not match in lenght
	 */
	public Byte2DoubleRBTreeMap(Byte[] keys, Double[] values) {
		this(keys, values, null);
	}
	
	/**
	 * Helper constructor that has a custom sorter and allow to create a map from boxed values (it will unbox them)
	 * @param keys the keys that should be put into the map
	 * @param values the values that should be put into the map.
	 * @param comp the function that decides how the tree is sorted, can be null
	 * @throws IllegalStateException if the keys and values do not match in lenght
	 */
	public Byte2DoubleRBTreeMap(Byte[] keys, Double[] values, ByteComparator comp) {
		comparator = comp;
		if(keys.length != values.length) throw new IllegalStateException("Input Arrays are not equal size");
		for(int i = 0,m=keys.length;i<m;i++) put(keys[i].byteValue(), values[i].doubleValue());
	}
	
	/**
	 * Helper constructor that allow to create a map from unboxed values
	 * @param keys the keys that should be put into the map
	 * @param values the values that should be put into the map.
	 * @throws IllegalStateException if the keys and values do not match in lenght
	 */
	public Byte2DoubleRBTreeMap(byte[] keys, double[] values) {
		this(keys, values, null);
	}
	
	/**
	 * Helper constructor that has a custom sorter and allow to create a map from unboxed values
	 * @param keys the keys that should be put into the map
	 * @param values the values that should be put into the map.
	 * @param comp the function that decides how the tree is sorted, can be null
	 * @throws IllegalStateException if the keys and values do not match in lenght
	 */
	public Byte2DoubleRBTreeMap(byte[] keys, double[] values, ByteComparator comp) {
		comparator = comp;
		if(keys.length != values.length) throw new IllegalStateException("Input Arrays are not equal size");
		for(int i = 0,m=keys.length;i<m;i++) put(keys[i], values[i]);
	}
	
	/**
	 * A Helper constructor that allows to create a Map with exactly the same values as the provided map.
	 * @param map the values that should be present in the map
	 */
	public Byte2DoubleRBTreeMap(Map<? extends Byte, ? extends Double> map) {
		this(map, null);
	}
	
	/**
	 * A Helper constructor that has a custom sorter and allows to create a Map with exactly the same values as the provided map.
	 * @param map the values that should be present in the map
	 * @param comp the function that decides how the tree is sorted, can be null
	 */
	public Byte2DoubleRBTreeMap(Map<? extends Byte, ? extends Double> map, ByteComparator comp) {
		comparator = comp;
		putAll(map);
	}
	
	/**
	 * A Type Specific Helper function that allows to create a new Map with exactly the same values as the provided map.
	 * @param map the values that should be present in the map
 	 */
	public Byte2DoubleRBTreeMap(Byte2DoubleMap map) {
		this(map, null);
	}
	
	/**
	 * A Type Specific Helper function that has a custom sorter and allows to create a new Map with exactly the same values as the provided map.
	 * @param map the values that should be present in the map
	 * @param comp the function that decides how the tree is sorted, can be null
 	 */
	public Byte2DoubleRBTreeMap(Byte2DoubleMap map, ByteComparator comp) {
		comparator = comp;
		putAll(map);
	}
	
	@Override
	public void setDefaultMaxValue(byte value) { defaultMaxNotFound = value; }
	@Override
	public byte getDefaultMaxValue() { return defaultMaxNotFound; }
	@Override
	public void setDefaultMinValue(byte value) { defaultMinNotFound = value; }
	@Override
	public byte getDefaultMinValue() { return defaultMinNotFound; }
	
	@Override
	public double put(byte key, double value) {
		if(tree == null) {
			tree = first = last = new Entry(key, value, null);
			size++;
			return getDefaultReturnValue();
		}
		int compare = 0;
		Entry parent = tree;
		while(true) {
			if((compare = compare(key, parent.key)) == 0) return parent.setValue(value);
			if(compare < 0) {
				if(parent.left == null) break;
				parent = parent.left;
			}
			else if(compare > 0) {
				if(parent.right == null) break;
				parent = parent.right;
			}
		}
		Entry adding = new Entry(key, value, parent);
		if(compare < 0)  {
			parent.left = adding;
			if(parent == first)	first = adding;
		}
		else {
			parent.right = adding;
			if(parent == last) last = adding;
		}
		fixAfterInsertion(adding);
		size++;
		return getDefaultReturnValue();
	}
	
	@Override
	public double putIfAbsent(byte key, double value) {

		if(tree == null) {
			tree = first = last = new Entry(key, value, null);
			size++;
			return getDefaultReturnValue();
		}
		int compare = 0;
		Entry parent = tree;
		while(true) {
			if((compare = compare(key, parent.key)) == 0) return parent.value;
			if(compare < 0) {
				if(parent.left == null) break;
				parent = parent.left;
			}
			else if(compare > 0) {
				if(parent.right == null) break;
				parent = parent.right;
			}
		}
		Entry adding = new Entry(key, value, parent);
		if(compare < 0)  {
			parent.left = adding;
			if(parent == first)	first = adding;
		}
		else  {
			parent.right = adding;
			if(parent == last) last = adding;
		}
		fixAfterInsertion(adding);
		size++;
		return getDefaultReturnValue();
	}
	
	@Override
	public double addTo(byte key, double value) {
		if(tree == null) {
			tree = first = last = new Entry(key, value, null);
			size++;
			return getDefaultReturnValue();
		}
		int compare = 0;
		Entry parent = tree;
		while(true) {
			if((compare = compare(key, parent.key)) == 0) return parent.addTo(value);
			if(compare < 0) {
				if(parent.left == null) break;
				parent = parent.left;
			}
			else if(compare > 0) {
				if(parent.right == null) break;
				parent = parent.right;
			}
		}
		Entry adding = new Entry(key, value, parent);
		if(compare < 0)  {
			parent.left = adding;
			if(parent == first)	first = adding;
		}
		else  {
			parent.right = adding;
			if(parent == last) last = adding;
		}
		fixAfterInsertion(adding);
		size++;
		return getDefaultReturnValue();
	}
	
	@Override
	public double putAndMoveToFirst(byte key, double value) { throw new UnsupportedOperationException(); }
	@Override
	public double putAndMoveToLast(byte key, double value) { throw new UnsupportedOperationException(); }
	@Override
	public boolean moveToFirst(byte key) { throw new UnsupportedOperationException(); }
	@Override
	public boolean moveToLast(byte key) { throw new UnsupportedOperationException(); }
	@Override
	public double getAndMoveToFirst(byte key) { throw new UnsupportedOperationException(); }
	@Override
	public double getAndMoveToLast(byte key) { throw new UnsupportedOperationException(); }
	
	@Override
	public ByteComparator comparator() { return comparator; }
	
	@Override
	public boolean containsKey(byte key) {
		return findNode(key) != null;
	}
	
	@Override
	public double get(byte key) {
		Entry node = findNode(key);
		return node == null ? getDefaultReturnValue() : node.value;
	}
	
	@Override
	public double getOrDefault(byte key, double defaultValue) {
		Entry node = findNode(key);
		return node == null ? defaultValue : node.value;
	}
	
	@Override
	public byte firstByteKey() {
		if(tree == null) throw new NoSuchElementException();
		return first.key;
	}
	
	@Override
	public byte pollFirstByteKey() {
		if(tree == null) throw new NoSuchElementException();
		byte result = first.key;
		removeNode(first);
		return result;
	}
	
	@Override
	public byte lastByteKey() {
		if(tree == null) throw new NoSuchElementException();
		return last.key;
	}
	
	@Override
	public byte pollLastByteKey() {
		if(tree == null) throw new NoSuchElementException();
		byte result = last.key;
		removeNode(last);
		return result;
	}
	
	@Override
	public Byte2DoubleMap.Entry firstEntry() {
		if(tree == null) throw new NoSuchElementException();
		return new BasicEntry(first.key, first.value);
	}
	
	@Override
	public Byte2DoubleMap.Entry lastEntry() {
		if(tree == null) throw new NoSuchElementException();
		return new BasicEntry(last.key, last.value);
	}
	
	@Override
	public Byte2DoubleMap.Entry pollFirstEntry() {
		if(tree == null) throw new NoSuchElementException();
		BasicEntry entry = new BasicEntry(first.key, first.value);
		removeNode(first);
		return entry;
	}
	
	@Override
	public Byte2DoubleMap.Entry pollLastEntry() {
		if(tree == null) throw new NoSuchElementException();
		BasicEntry entry = new BasicEntry(last.key, last.value);
		removeNode(last);
		return entry;
	}
	
	@Override
	public double firstDoubleValue() {
		if(tree == null) throw new NoSuchElementException();
		return first.value;
	}
	
	@Override
	public double lastDoubleValue() {
		if(tree == null) throw new NoSuchElementException();
		return last.value;
	}
	
	@Override
	public double remove(byte key) {
		Entry entry = findNode(key);
		if(entry == null) return getDefaultReturnValue();
		double value = entry.value;
		removeNode(entry);
		return value;
	}
	
	@Override
	public double removeOrDefault(byte key, double defaultValue) {
		Entry entry = findNode(key);
		if(entry == null) return defaultValue;
		double value = entry.value;
		removeNode(entry);
		return value;
	}
	
	@Override
	public boolean remove(byte key, double value) {
		Entry entry = findNode(key);
		if(entry == null || entry.value != value) return false;
		removeNode(entry);
		return true;
	}
	
	@Override
	public boolean replace(byte key, double oldValue, double newValue) {
		Entry entry = findNode(key);
		if(entry == null || entry.value != oldValue) return false;
		entry.value = newValue;
		return true;
	}
	
	@Override
	public double replace(byte key, double value) {
		Entry entry = findNode(key);
		if(entry == null) return getDefaultReturnValue();
		double oldValue = entry.value;
		entry.value = value;
		return oldValue;
	}
	
	@Override
	public double computeDouble(byte key, ByteDoubleUnaryOperator mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		Entry entry = findNode(key);
		if(entry == null) {
			double newValue = mappingFunction.applyAsDouble(key, getDefaultReturnValue());
			if(Double.doubleToLongBits(newValue) == Double.doubleToLongBits(getDefaultReturnValue())) return newValue;
			put(key, newValue);
			return newValue;
		}
		double newValue = mappingFunction.applyAsDouble(key, entry.value);
		if(Double.doubleToLongBits(newValue) == Double.doubleToLongBits(getDefaultReturnValue())) {
			removeNode(entry);
			return newValue;
		}
		entry.value = newValue;
		return newValue;
	}
	
	@Override
	public double computeDoubleIfAbsent(byte key, Byte2DoubleFunction mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		Entry entry = findNode(key);
		if(entry == null) {
			double newValue = mappingFunction.get(key);
			if(Double.doubleToLongBits(newValue) == Double.doubleToLongBits(getDefaultReturnValue())) return newValue;
			put(key, newValue);
			return newValue;
		}
		if(Double.doubleToLongBits(entry.value) == Double.doubleToLongBits(getDefaultReturnValue())) {
			double newValue = mappingFunction.get(key);
			if(Double.doubleToLongBits(newValue) == Double.doubleToLongBits(getDefaultReturnValue())) return newValue;
			entry.value = newValue;
		}
		return entry.value;
	}
	
	@Override
	public double supplyDoubleIfAbsent(byte key, DoubleSupplier valueProvider) {
		Objects.requireNonNull(valueProvider);
		Entry entry = findNode(key);
		if(entry == null) {
			double newValue = valueProvider.getDouble();
			if(Double.doubleToLongBits(newValue) == Double.doubleToLongBits(getDefaultReturnValue())) return newValue;
			put(key, newValue);
			return newValue;
		}
		if(Double.doubleToLongBits(entry.value) == Double.doubleToLongBits(getDefaultReturnValue())) {
			double newValue = valueProvider.getDouble();
			if(Double.doubleToLongBits(newValue) == Double.doubleToLongBits(getDefaultReturnValue())) return newValue;
			entry.value = newValue;
		}
		return entry.value;
	}
	
	@Override
	public double computeDoubleIfPresent(byte key, ByteDoubleUnaryOperator mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		Entry entry = findNode(key);
		if(entry == null || Double.doubleToLongBits(entry.value) == Double.doubleToLongBits(getDefaultReturnValue())) return getDefaultReturnValue();
		double newValue = mappingFunction.applyAsDouble(key, entry.value);
		if(Double.doubleToLongBits(newValue) == Double.doubleToLongBits(getDefaultReturnValue())) {
			removeNode(entry);
			return newValue;
		}
		entry.value = newValue;
		return newValue;
	}
	
	@Override
	public double mergeDouble(byte key, double value, DoubleDoubleUnaryOperator mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		Entry entry = findNode(key);
		double newValue = entry == null || Double.doubleToLongBits(entry.value) == Double.doubleToLongBits(getDefaultReturnValue()) ? value : mappingFunction.applyAsDouble(entry.value, value);
		if(Double.doubleToLongBits(newValue) == Double.doubleToLongBits(getDefaultReturnValue())) {
			if(entry != null)
				removeNode(entry);
		}
		else if(entry == null) put(key, newValue);
		else entry.value = newValue;
		return newValue;
	}
	
	@Override
	public void mergeAllDouble(Byte2DoubleMap m, DoubleDoubleUnaryOperator mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		for(Byte2DoubleMap.Entry entry : Byte2DoubleMaps.fastIterable(m)) {
			byte key = entry.getByteKey();
			Entry subEntry = findNode(key);
			double newValue = subEntry == null || Double.doubleToLongBits(subEntry.value) == Double.doubleToLongBits(getDefaultReturnValue()) ? entry.getDoubleValue() : mappingFunction.applyAsDouble(subEntry.value, entry.getDoubleValue());
			if(Double.doubleToLongBits(newValue) == Double.doubleToLongBits(getDefaultReturnValue())) {
				if(subEntry != null)
					removeNode(subEntry);
			}
			else if(subEntry == null) put(key, newValue);
			else subEntry.value = newValue;
		}
	}
	
	@Override
	public void forEach(ByteDoubleConsumer action) {
		for(Entry entry = first;entry != null;entry = entry.next())
			action.accept(entry.key, entry.value);
	}
	
	@Override
	public int size() { return size; }
	
	@Override
	public void clear() {
		size = 0;
		first = null;
		last = null;
		tree = null;
	}
	
	ByteListIterator keyIterator(boolean descending) {
		ByteListIterator iter = new KeyIterator(descending);
		return descending ? ByteIterators.invert(iter) : iter;
	}
	
	ByteListIterator keyIterator(byte key) {
		return new KeyIterator(findNode(key));
	}
	
	public Byte2DoubleRBTreeMap copy() {
		Byte2DoubleRBTreeMap set = new Byte2DoubleRBTreeMap();
		set.size = size;
		if(tree != null) {
			set.tree = tree.copy();
			Entry lastFound = null;
			for(Entry entry = tree;entry != null;entry = entry.left) lastFound = entry;
			set.first = lastFound;
			lastFound = null;
			for(Entry entry = tree;entry != null;entry = entry.right) lastFound = entry;
			set.last = lastFound;
		}
		return set;
	}
	
	@Override
	public ByteSortedSet keySet() {
		return navigableKeySet();
	}
	
	@Override
	public ObjectSet<Byte2DoubleMap.Entry> byte2DoubleEntrySet() {
		if(entrySet == null) entrySet = new EntrySet();
		return entrySet;
	}
	
	@Override
	public DoubleCollection values() {
		if(values == null) values = new Values();
		return values;
	}
	
	@Override
	public ByteNavigableSet navigableKeySet() {
		if(keySet == null) keySet = new KeySet(this);
		return keySet;
	}
	
	@Override
	public Byte2DoubleNavigableMap descendingMap() {
		return new DescendingSubMap(this, true, (byte)0, true, true, (byte)0, true);
	}
	
	@Override
	public ByteNavigableSet descendingKeySet() {
		return descendingMap().navigableKeySet();
	}
	
	@Override
	public Byte2DoubleNavigableMap subMap(byte fromKey, boolean fromInclusive, byte toKey, boolean toInclusive) {
		return new AscendingSubMap(this, false, fromKey, fromInclusive, false, toKey, toInclusive);
	}
	
	@Override
	public Byte2DoubleNavigableMap headMap(byte toKey, boolean inclusive) {
		return new AscendingSubMap(this, true, (byte)0, true, false, toKey, inclusive);
	}
	
	@Override
	public Byte2DoubleNavigableMap tailMap(byte fromKey, boolean inclusive) {
		return new AscendingSubMap(this, false, fromKey, inclusive, true, (byte)0, true);
	}
	
	@Override
	public byte lowerKey(byte e) {
		Entry node = findLowerNode(e);
		return node != null ? node.key : getDefaultMinValue();
	}

	@Override
	public byte floorKey(byte e) {
		Entry node = findFloorNode(e);
		return node != null ? node.key : getDefaultMinValue();
	}
	
	@Override
	public byte higherKey(byte e) {
		Entry node = findHigherNode(e);
		return node != null ? node.key : getDefaultMaxValue();
	}

	@Override
	public byte ceilingKey(byte e) {
		Entry node = findCeilingNode(e);
		return node != null ? node.key : getDefaultMaxValue();
	}
	
	@Override
	public Byte2DoubleMap.Entry lowerEntry(byte key) {
		Entry node = findLowerNode(key);
		return node != null ? new BasicEntry(node.key, node.value) : null;
	}
	
	@Override
	public Byte2DoubleMap.Entry higherEntry(byte key) {
		Entry node = findHigherNode(key);
		return node != null ? new BasicEntry(node.key, node.value) : null;
	}
	
	@Override
	public Byte2DoubleMap.Entry floorEntry(byte key) {
		Entry node = findFloorNode(key);
		return node != null ? new BasicEntry(node.key, node.value) : null;
	}
	
	@Override
	public Byte2DoubleMap.Entry ceilingEntry(byte key) {
		Entry node = findCeilingNode(key);
		return node != null ? new BasicEntry(node.key, node.value) : null;
	}
	
	protected Entry findLowerNode(byte key) {
		Entry entry = tree;
		while(entry != null) {
			if(compare(key, entry.key) > 0) {
				if(entry.right != null) entry = entry.right;
				else return entry;
			}
			else {
				if(entry.left != null) entry = entry.left;
				else {
					Entry parent = entry.parent;
					while(parent != null && parent.left == entry) {
						entry = parent;
						parent = parent.parent;
					}
					return parent;
				}
			}
		}
		return null;
	}
	
	protected Entry findFloorNode(byte key) {
		Entry entry = tree;
		int compare;
		while(entry != null) {
			if((compare = compare(key, entry.key)) > 0) {
				if(entry.right == null) break;
				entry = entry.right;
				continue;
			}
			else if(compare < 0) {
				if(entry.left != null) entry = entry.left;
				else {
					Entry parent = entry.parent;
					while(parent != null && parent.left == entry) {
						entry = parent;
						parent = parent.parent;
					}
					return parent;
				}
				continue;
			}
			break;
		}
		return entry;
	}
	
	protected Entry findCeilingNode(byte key) {
		Entry entry = tree;
		int compare;
		while(entry != null) {
			if((compare = compare(key, entry.key)) < 0) {
				if(entry.left == null) break;
				entry = entry.left;
				continue;
			}
			else if(compare > 0) {
				if(entry.right != null) entry = entry.right;
				else {
					Entry parent = entry.parent;
					while(parent != null && parent.right == entry) {
						entry = parent;
						parent = parent.parent;
					}
					return parent;
				}
				continue;
			}
			break;
		}
		return entry;
	}
	
	protected Entry findHigherNode(byte key) {
		Entry entry = tree;
		while(entry != null) {
			if(compare(key, entry.key) < 0) {
				if(entry.left != null) entry = entry.left;
				else return entry;
			}
			else {
				if(entry.right != null) entry = entry.right;
				else {
					Entry parent = entry.parent;
					while(parent != null && parent.right == entry) {
						entry = parent;
						parent = parent.parent;
					}
					return parent;
				}
			}
		}
		return null;
	}
	
	protected Entry findNode(byte key) {
		Entry node = tree;
		int compare;
		while(node != null) {
			if((compare = compare(key, node.key)) == 0) return node;
			if(compare < 0) node = node.left;
			else node = node.right;
		}
		return null;
	}
	
	protected void removeNode(Entry entry) {
		size--;
		if(entry.needsSuccessor()) {
			Entry successor = entry.next();
			entry.key = successor.key;
			entry.value = successor.value;
			entry = successor;
		}
		Entry replacement = entry.left != null ? entry.left : entry.right;
		if(replacement != null) {
			if(entry.replace(replacement)) tree = replacement;
			if(entry == first) first = replacement;
			if(entry == last) last = entry.right != null ? entry.right : replacement;
			entry.left = entry.right = entry.parent = null;
			if(entry.isBlack()) fixAfterDeletion(replacement);
		}
		else if(entry.parent == null) tree = first = last = null;
		else {
			if(entry.isBlack())
				fixAfterDeletion(entry);
			entry.replace(null);
			if(entry.parent != null) {
				Entry parent = entry.parent;
				if(entry == first) first = parent.left != null ? parent.left : parent;
				if(entry == last) last = entry.right != null ? parent.right : parent;
			}
			entry.parent = null;
		}
	}
	
	protected void validate(byte k) { compare(k, k); }
	protected int compare(byte k, byte v) { return comparator != null ? comparator.compare(k, v) : Byte.compare(k, v);}
	protected static boolean isBlack(Entry p) { return p == null || p.isBlack(); }
	protected static Entry parentOf(Entry p) { return (p == null ? null : p.parent); }
	protected static void setBlack(Entry p, boolean c) { if(p != null) p.setBlack(c); }
	protected static Entry leftOf(Entry p) { return p == null ? null : p.left; }
	protected static Entry rightOf(Entry p) { return (p == null) ? null : p.right; }
	
	protected void rotateLeft(Entry entry) {
		if(entry != null) {
			Entry right = entry.right;
			entry.right = right.left;
			if(right.left != null) right.left.parent = entry;
			right.parent = entry.parent;
			if(entry.parent == null) tree = right;
			else if(entry.parent.left == entry) entry.parent.left = right;
			else entry.parent.right = right;
			right.left = entry;
			entry.parent = right;
		}
	}
	
	protected void rotateRight(Entry entry) {
		if(entry != null) {
			Entry left = entry.left;
			entry.left = left.right;
			if(left.right != null) left.right.parent = entry;
			left.parent = entry.parent;
			if(entry.parent == null) tree = left;
			else if(entry.parent.right == entry) entry.parent.right = left;
			else entry.parent.left = left;
			left.right = entry;
			entry.parent = left;
		}
	}
	
	protected void fixAfterInsertion(Entry entry) {
		entry.setBlack(false);
		while(entry != null && entry != tree && !entry.parent.isBlack()) {
			if(parentOf(entry) == leftOf(parentOf(parentOf(entry)))) {
				Entry y = rightOf(parentOf(parentOf(entry)));
				if(!isBlack(y)) {
					setBlack(parentOf(entry), true);
					setBlack(y, true);
					setBlack(parentOf(parentOf(entry)), false);
					entry = parentOf(parentOf(entry));
				}
				else {
					if(entry == rightOf(parentOf(entry))) {
						entry = parentOf(entry);
						rotateLeft(entry);
					}
					setBlack(parentOf(entry), true);
					setBlack(parentOf(parentOf(entry)), false);
					rotateRight(parentOf(parentOf(entry)));
				}
			}
			else {
				Entry y = leftOf(parentOf(parentOf(entry)));
				if(!isBlack(y)) {
					setBlack(parentOf(entry), true);
					setBlack(y, true);
					setBlack(parentOf(parentOf(entry)), false);
					entry = parentOf(parentOf(entry));
				}
				else {
					if(entry == leftOf(parentOf(entry))) {
						entry = parentOf(entry);
						rotateRight(entry);
					}
					setBlack(parentOf(entry), true);
					setBlack(parentOf(parentOf(entry)), false);
					rotateLeft(parentOf(parentOf(entry)));
				}
			}
		}
		tree.setBlack(true);
	}
	
	protected void fixAfterDeletion(Entry entry) {
		while(entry != tree && isBlack(entry)) {
			if(entry == leftOf(parentOf(entry))) {
				Entry sib = rightOf(parentOf(entry));
				if(!isBlack(sib)) {
					setBlack(sib, true);
					setBlack(parentOf(entry), false);
					rotateLeft(parentOf(entry));
					sib = rightOf(parentOf(entry));
				}
				if(isBlack(leftOf(sib)) && isBlack(rightOf(sib))) {
					setBlack(sib, false);
					entry = parentOf(entry);
				}
				else {
					if(isBlack(rightOf(sib))) {
						setBlack(leftOf(sib), true);
						setBlack(sib, false);
						rotateRight(sib);
						sib = rightOf(parentOf(entry));
					}
					setBlack(sib, isBlack(parentOf(entry)));
					setBlack(parentOf(entry), true);
					setBlack(rightOf(sib), true);
					rotateLeft(parentOf(entry));
					entry = tree;
				}
			}
			else {
				Entry sib = leftOf(parentOf(entry));
				if(!isBlack(sib)) {
					setBlack(sib, true);
					setBlack(parentOf(entry), false);
					rotateRight(parentOf(entry));
					sib = leftOf(parentOf(entry));
				}
				if(isBlack(rightOf(sib)) && isBlack(leftOf(sib))) {
					setBlack(sib, false);
					entry = parentOf(entry);
				}
				else {
					if(isBlack(leftOf(sib))) {
						setBlack(rightOf(sib), true);
						setBlack(sib, false);
						rotateLeft(sib);
						sib = leftOf(parentOf(entry));
					}
					setBlack(sib, isBlack(parentOf(entry)));
					setBlack(parentOf(entry), true);
					setBlack(leftOf(sib), true);
					rotateRight(parentOf(entry));
					entry = tree;
				}
			}
		}
		setBlack(entry, true);
	}
	
	static final class AscendingSubMap extends NavigableSubMap {

		public AscendingSubMap(Byte2DoubleRBTreeMap m, boolean fromStart, byte low, boolean loInclusive, boolean toEnd, byte high, boolean hiInclusive) {
			super(m, fromStart, low, loInclusive, toEnd, high, hiInclusive);
		}
		
		@Override
		ByteListIterator keyIterator(boolean descending) { return new SubMapKeyIterator(descending); }
		@Override
		ByteListIterator keyIterator(byte key) {
			Byte2DoubleRBTreeMap.Entry entry = m.findNode(key);
			return entry == null || !inClosedRange(key) ? null : new SubMapKeyIterator(entry, fromStart ? null : findLowest(), toEnd ? null : findHighest());
		}
		@Override
		DoubleListIterator valueIterator() { return new SubMapValueIterator(false); }
		@Override
		ObjectListIterator<Byte2DoubleMap.Entry> entryIterator() { return new SubMapEntrySetIterator(false); }
		
		@Override
		public Byte2DoubleNavigableMap descendingMap() {
			if(descendingMap == null) descendingMap = new DescendingSubMap(m, fromStart, low, loInclusive, toEnd, high, hiInclusive);
			return descendingMap;
		}
		
		Byte2DoubleRBTreeMap.Entry next(Byte2DoubleRBTreeMap.Entry entry) {
			return entry.next();
		}
		
		Byte2DoubleRBTreeMap.Entry previous(Byte2DoubleRBTreeMap.Entry entry) {
			return entry.previous();
		}
		
		@Override
		public Byte2DoubleNavigableMap subMap(byte fromKey, boolean fromInclusive, byte toKey, boolean toInclusive) {
			if (!inRange(fromKey, fromInclusive)) throw new IllegalArgumentException("fromKey out of range");
			if (!inRange(toKey, toInclusive)) throw new IllegalArgumentException("toKey out of range");
			return new AscendingSubMap(m, false, fromKey, fromInclusive, false, toKey, toInclusive);
		}
		
		@Override
		public Byte2DoubleNavigableMap headMap(byte toKey, boolean inclusive) {
			if (!inRange(toKey, inclusive)) throw new IllegalArgumentException("toKey out of range");
			return new AscendingSubMap(m, fromStart, low, loInclusive, false, toKey, inclusive);
		}
		
		@Override
		public Byte2DoubleNavigableMap tailMap(byte fromKey, boolean inclusive) {
			if (!inRange(fromKey, inclusive)) throw new IllegalArgumentException("fromKey out of range");
			return new AscendingSubMap(m, false, fromKey, inclusive, toEnd, high, hiInclusive);
		}
	}
	
	static final class DescendingSubMap extends NavigableSubMap {

		public DescendingSubMap(Byte2DoubleRBTreeMap m, boolean fromStart, byte low, boolean loInclusive, boolean toEnd, byte high, boolean hiInclusive) {
			super(m, fromStart, low, loInclusive, toEnd, high, hiInclusive);
		}
		
		@Override
		public byte firstByteKey() { return super.lastByteKey(); }
		
		@Override
		public byte pollFirstByteKey() { return super.pollLastByteKey(); }
		
		@Override
		public byte lastByteKey() { return super.firstByteKey(); }
		
		@Override
		public byte pollLastByteKey() { return super.pollFirstByteKey(); }
		
		@Override
		public double firstDoubleValue() { return super.lastDoubleValue(); }
		
		@Override
		public double lastDoubleValue() { return super.firstDoubleValue(); }
		
		@Override
		public Byte2DoubleMap.Entry firstEntry() { return super.lastEntry(); }
		
		@Override
		public Byte2DoubleMap.Entry lastEntry() { return super.firstEntry(); }
		
		@Override
		public Byte2DoubleMap.Entry pollFirstEntry() { return super.pollLastEntry(); }
		
		@Override
		public Byte2DoubleMap.Entry pollLastEntry() { return super.pollFirstEntry(); }
		
		@Override
		public byte lowerKey(byte e) { return super.higherKey(e); }
		
		@Override
		public byte floorKey(byte e) { return super.ceilingKey(e); }
		
		@Override
		public byte ceilingKey(byte e) { return super.floorKey(e); }
		
		@Override
		public byte higherKey(byte e) { return super.lowerKey(e); }
		
		@Override
		public Byte2DoubleMap.Entry lowerEntry(byte e) { return super.higherEntry(e); }
		
		@Override
		public Byte2DoubleMap.Entry floorEntry(byte e) { return super.ceilingEntry(e); }
		
		@Override
		public Byte2DoubleMap.Entry ceilingEntry(byte e) { return super.floorEntry(e); }
		
		@Override
		public Byte2DoubleMap.Entry higherEntry(byte e) { return super.lowerEntry(e); }
		
		protected Byte2DoubleRBTreeMap.Entry findLowest() {
			if(fromStart) return m.first;
			Byte2DoubleRBTreeMap.Entry entry = loInclusive ? m.findCeilingNode(low) : m.findHigherNode(low);
			return entry == null || tooHigh(entry.key) ? null : entry;
		}

		protected Byte2DoubleRBTreeMap.Entry findHighest() {
			if(toEnd) return m.last;
			Byte2DoubleRBTreeMap.Entry entry = hiInclusive ? m.findFloorNode(high) : m.findLowerNode(high);
			return entry == null || tooLow(entry.key) ? null : entry;
		}
		
		@Override
		ByteListIterator keyIterator(boolean descending) {
			ByteListIterator iter = new SubMapKeyIterator(!descending);
			return descending ? iter : ByteIterators.invert(iter);
		}
		
		@Override
		ByteListIterator keyIterator(byte key) {
			Byte2DoubleRBTreeMap.Entry entry = m.findNode(key);
			return entry == null || !inClosedRange(key) ? null : ByteIterators.invert(new SubMapKeyIterator(entry, fromStart ? null : findLowest(), toEnd ? null : findHighest()));
		}
		
		Byte2DoubleRBTreeMap.Entry next(Byte2DoubleRBTreeMap.Entry entry) {
			return entry.previous();
		}
		
		Byte2DoubleRBTreeMap.Entry previous(Byte2DoubleRBTreeMap.Entry entry) {
			return entry.next();
		}
		
		@Override
		public ByteComparator comparator() { return m.comparator() == null ? null : m.comparator().reversed(); }
		
		@Override
		DoubleListIterator valueIterator() { return DoubleIterators.invert(new SubMapValueIterator(true)); }
		
		@Override
		ObjectListIterator<Byte2DoubleMap.Entry> entryIterator() { return ObjectIterators.invert(new SubMapEntrySetIterator(true)); }
		
		@Override
		public Byte2DoubleNavigableMap descendingMap() {
			if(descendingMap == null) descendingMap = new AscendingSubMap(m, fromStart, low, loInclusive, toEnd, high, hiInclusive);
			return descendingMap;
		}

		@Override
		public Byte2DoubleNavigableMap subMap(byte fromKey, boolean fromInclusive, byte toKey, boolean toInclusive) {
			if (!inRange(fromKey, fromInclusive)) throw new IllegalArgumentException("fromKey out of range");
			if (!inRange(toKey, toInclusive)) throw new IllegalArgumentException("toKey out of range");
			return new DescendingSubMap(m, false, fromKey, fromInclusive, false, toKey, toInclusive);
		}
		
		@Override
		public Byte2DoubleNavigableMap headMap(byte toKey, boolean inclusive) {
			if (!inRange(toKey, inclusive)) throw new IllegalArgumentException("toKey out of range");
			return new DescendingSubMap(m, fromStart, low, loInclusive, false, toKey, inclusive);
		}
		
		@Override
		public Byte2DoubleNavigableMap tailMap(byte fromKey, boolean inclusive) {
			if (!inRange(fromKey, inclusive)) throw new IllegalArgumentException("fromKey out of range");
			return new DescendingSubMap(m, false, fromKey, inclusive, toEnd, high, hiInclusive);
		}
	}
	
	static abstract class NavigableSubMap extends AbstractByte2DoubleMap implements Byte2DoubleNavigableMap {
		final Byte2DoubleRBTreeMap m;
		final byte low;
		final byte high;
		final boolean fromStart;
		final boolean toEnd;
		final boolean loInclusive;
		final boolean hiInclusive;
		transient Byte2DoubleNavigableMap descendingMap;
		transient SubMapEntrySet entrySet;
		transient KeySet keySet;
		transient SubMapValues values;
		
		NavigableSubMap(Byte2DoubleRBTreeMap m, boolean fromStart, byte low, boolean loInclusive, boolean toEnd, byte high, boolean hiInclusive) {
			this.m = m;
			this.low = low;
			this.high = high;
			this.fromStart = fromStart;
			this.toEnd = toEnd;
			this.loInclusive = loInclusive;
			this.hiInclusive = hiInclusive;
		}
		
		abstract ByteListIterator keyIterator(boolean descending);
		abstract ByteListIterator keyIterator(byte key);
		abstract DoubleListIterator valueIterator();
		abstract ObjectListIterator<Byte2DoubleMap.Entry> entryIterator();
		abstract Byte2DoubleRBTreeMap.Entry next(Byte2DoubleRBTreeMap.Entry entry);
		abstract Byte2DoubleRBTreeMap.Entry previous(Byte2DoubleRBTreeMap.Entry entry);
		
		@Override
		public NavigableSubMap copy() { throw new UnsupportedOperationException(); }
		@Override
		public abstract Byte2DoubleNavigableMap descendingMap();
		@Override
		public abstract Byte2DoubleNavigableMap subMap(byte fromKey, boolean fromInclusive, byte toKey, boolean toInclusive);
		@Override
		public abstract Byte2DoubleNavigableMap headMap(byte toKey, boolean inclusive);
		@Override
		public abstract Byte2DoubleNavigableMap tailMap(byte fromKey, boolean inclusive);
		
		boolean tooLow(byte key) { return !fromStart && (loInclusive ? m.compare(key, low) < 0 : m.compare(key, low) <= 0); }
		boolean tooHigh(byte key) { return !toEnd && (hiInclusive ? m.compare(key, high) > 0 : m.compare(key, high) >= 0); }
		boolean inRange(byte key) { return !tooLow(key) && !tooHigh(key); }
		boolean inClosedRange(byte key) { return (fromStart || m.compare(key, low) >= 0) && (toEnd || m.compare(high, key) >= 0); }
		boolean inRange(byte key, boolean inclusive) { return inclusive ? inRange(key) : inClosedRange(key); }
		
		@Override
		public void setDefaultMaxValue(byte e) { m.setDefaultMaxValue(e); }
		@Override
		public byte getDefaultMaxValue() { return m.getDefaultMaxValue(); }
		@Override
		public void setDefaultMinValue(byte e) { m.setDefaultMinValue(e); }
		@Override
		public byte getDefaultMinValue() { return m.getDefaultMinValue(); }
		
		@Override
		public AbstractByte2DoubleMap setDefaultReturnValue(double v) { 
			m.setDefaultReturnValue(v);
			return this;
		}
		
		@Override
		public double getDefaultReturnValue() { return m.getDefaultReturnValue(); }
		
		@Override
		public double putAndMoveToFirst(byte key, double value) { throw new UnsupportedOperationException(); }
		@Override
		public double putAndMoveToLast(byte key, double value) { throw new UnsupportedOperationException(); }
		@Override
		public boolean moveToFirst(byte key) { throw new UnsupportedOperationException(); }
		@Override
		public boolean moveToLast(byte key) { throw new UnsupportedOperationException(); }
		@Override
		public double getAndMoveToFirst(byte key) { throw new UnsupportedOperationException(); }
		@Override
		public double getAndMoveToLast(byte key) { throw new UnsupportedOperationException(); }
		@Override
		public ByteComparator comparator() { return m.comparator(); }
		@Override
		public boolean containsKey(byte key) { return inRange(key) && m.containsKey(key); }
		
		@Override
		public byte firstByteKey() { 
			Byte2DoubleRBTreeMap.Entry entry = findLowest();
			return entry == null ? getDefaultMaxValue() : entry.key;
		}
		
		@Override
		public byte pollFirstByteKey() {
			if(fromStart) return m.pollFirstByteKey();
			Byte2DoubleRBTreeMap.Entry entry = loInclusive ? m.findCeilingNode(low) : m.findHigherNode(low);
			if(entry != null && !tooHigh(entry.key)) {
				byte value = entry.key;
				m.removeNode(entry);
				return value;
			}
			return getDefaultMaxValue();
		}
		
		@Override
		public byte lastByteKey() {
			Byte2DoubleRBTreeMap.Entry entry = findHighest();
			return entry == null ? getDefaultMinValue() : entry.key;	
		}
		
		@Override
		public byte pollLastByteKey() {
			if(toEnd) return m.pollLastByteKey();
			Byte2DoubleRBTreeMap.Entry entry = hiInclusive ? m.findFloorNode(high) : m.findLowerNode(high);
			if(entry != null && !tooLow(entry.key)) {
				byte value = entry.key;
				m.removeNode(entry);
				return value;
			}
			return getDefaultMinValue();
		}
		
		@Override
		public double firstDoubleValue() {
			Byte2DoubleRBTreeMap.Entry entry = findLowest();
			return entry == null ? getDefaultReturnValue() : entry.value;
		}
		
		@Override
		public double lastDoubleValue() {
			Byte2DoubleRBTreeMap.Entry entry = findHighest();
			return entry == null ? getDefaultReturnValue() : entry.value;
		}
		
		protected Byte2DoubleRBTreeMap.Entry findLowest() {
			if(fromStart) return m.first;
			Byte2DoubleRBTreeMap.Entry entry = loInclusive ? m.findCeilingNode(low) : m.findHigherNode(low);
			return entry == null || tooHigh(entry.key) ? null : entry;
		}
		
		protected Byte2DoubleRBTreeMap.Entry findHighest() {
			if(toEnd) return m.last;
			Byte2DoubleRBTreeMap.Entry entry = hiInclusive ? m.findFloorNode(high) : m.findLowerNode(high);
			return entry == null || tooLow(entry.key) ? null : entry;
		}
		
		@Override
		public double put(byte key, double value) {
			if(!inRange(key)) throw new IllegalArgumentException("key out of range");
			return m.put(key, value);
		}
		
		@Override
		public double putIfAbsent(byte key, double value) {
			if(!inRange(key)) throw new IllegalArgumentException("key out of range");
			return m.putIfAbsent(key, value);
		}
		
		@Override
		public double addTo(byte key, double value) {
			if(!inRange(key)) throw new IllegalArgumentException("key out of range");
			return m.addTo(key, value);
		}
		
		@Override
		public double get(byte key) {
			return inRange(key) ? m.get(key) : getDefaultReturnValue();
		}
		
		@Override
		public double getOrDefault(byte key, double defaultValue) {
			return inRange(key) ? m.getOrDefault(key, defaultValue) : getDefaultReturnValue();
		}
		
		@Override
		public double remove(byte key) {
			return inRange(key) ? m.remove(key) : getDefaultReturnValue();
		}
		
		@Override
		public double removeOrDefault(byte key, double defaultValue) {
			return inRange(key) ? m.remove(key) : defaultValue;
		}
		
		@Override
		public boolean remove(byte key, double value) {
			return inRange(key) && m.remove(key, value);
		}
		
		@Override
		public ObjectSet<Byte2DoubleMap.Entry> byte2DoubleEntrySet() {
			if(entrySet == null) entrySet = new SubMapEntrySet();
			return entrySet;
		}
		
		@Override
		public ByteSet keySet() { return navigableKeySet(); }
		
		@Override
		public ByteNavigableSet navigableKeySet() {
			if(keySet == null) keySet = new KeySet(this);
			return keySet;
		}
		
		@Override
		public ByteNavigableSet descendingKeySet() {
			return descendingMap().navigableKeySet();
		}
		
		@Override
		public DoubleCollection values() {
			if(values == null) values = new SubMapValues();
			return values;
		}
		
		@Override
		public void forEach(ByteDoubleConsumer action) {
			for(Byte2DoubleRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
				action.accept(entry.key, entry.value);
		}
		
		@Override
		public int size() {
			return (fromStart && toEnd) ? m.size() : entrySet().size();
		}
		
		@Override
		public Byte2DoubleMap.Entry firstEntry() {
			Byte2DoubleRBTreeMap.Entry entry = findLowest();
			return entry == null ? null : new BasicEntry(entry.key, entry.value);
		}
		
		@Override
		public Byte2DoubleMap.Entry lastEntry() {
			Byte2DoubleRBTreeMap.Entry entry = findHighest();
			return entry == null ? null : new BasicEntry(entry.key, entry.value);
		}
		
		@Override
		public Byte2DoubleMap.Entry pollFirstEntry() {
			Byte2DoubleRBTreeMap.Entry entry = findLowest();
			if(entry == null) return null;
			BasicEntry result = new BasicEntry(entry.key, entry.value);
			m.removeNode(entry);
			return result;
		}
		
		@Override
		public Byte2DoubleMap.Entry pollLastEntry() {
			Byte2DoubleRBTreeMap.Entry entry = findHighest();
			if(entry == null) return null;
			BasicEntry result = new BasicEntry(entry.key, entry.value);
			m.removeNode(entry);
			return result;
		}
		
		@Override
		public byte lowerKey(byte e) {
			if(tooHigh(e)) {
				Byte2DoubleRBTreeMap.Entry entry = findHighest();
				return entry == null ? getDefaultMinValue() : entry.key;
			}
			Byte2DoubleRBTreeMap.Entry entry = m.findLowerNode(e);
			return entry == null || tooHigh(entry.key) ? getDefaultMaxValue() : entry.key;
		}
		
		@Override
		public byte floorKey(byte e) {
			if(tooHigh(e)) {
				Byte2DoubleRBTreeMap.Entry entry = findHighest();
				return entry == null ? getDefaultMinValue() : entry.key;
			}
			Byte2DoubleRBTreeMap.Entry entry = m.findFloorNode(e);
			return entry == null || tooHigh(entry.key) ? getDefaultMaxValue() : entry.key;
		}
		
		@Override
		public byte ceilingKey(byte e) {
			if(tooLow(e)) {
				Byte2DoubleRBTreeMap.Entry entry = findLowest();
				return entry == null ? getDefaultMaxValue() : entry.key;
			}
			Byte2DoubleRBTreeMap.Entry entry = m.findCeilingNode(e);
			return entry == null || tooLow(entry.key) ? getDefaultMinValue() : entry.key;
		}
		
		@Override
		public byte higherKey(byte e) {
			if(tooLow(e)) {
				Byte2DoubleRBTreeMap.Entry entry = findLowest();
				return entry == null ? getDefaultMaxValue() : entry.key;
			}
			Byte2DoubleRBTreeMap.Entry entry = m.findHigherNode(e);
			return entry == null || tooLow(entry.key) ? getDefaultMinValue() : entry.key;
		}
		
		@Override
		public Byte2DoubleMap.Entry lowerEntry(byte e) {
			if(tooHigh(e)) {
				Byte2DoubleRBTreeMap.Entry entry = findHighest();
				return entry == null ? null : new BasicEntry(entry.key, entry.value);
			}
			Byte2DoubleRBTreeMap.Entry entry = m.findLowerNode(e);
			return entry == null || tooHigh(entry.key) ? null : new BasicEntry(entry.key, entry.value);
		}
		
		@Override
		public Byte2DoubleMap.Entry floorEntry(byte e) {
			if(tooHigh(e)) {
				Byte2DoubleRBTreeMap.Entry entry = findHighest();
				return entry == null ? null : new BasicEntry(entry.key, entry.value);
			}
			Byte2DoubleRBTreeMap.Entry entry = m.findFloorNode(e);
			return entry == null || tooHigh(entry.key) ? null : new BasicEntry(entry.key, entry.value);
		}
		
		@Override
		public Byte2DoubleMap.Entry ceilingEntry(byte e) {
			if(tooLow(e)) {
				Byte2DoubleRBTreeMap.Entry entry = findLowest();
				return entry == null ? null : new BasicEntry(entry.key, entry.value);
			}
			Byte2DoubleRBTreeMap.Entry entry = m.findCeilingNode(e);
			return entry == null || tooLow(entry.key) ? null : new BasicEntry(entry.key, entry.value);
		}
		
		@Override
		public Byte2DoubleMap.Entry higherEntry(byte e) {
			if(tooLow(e)) {
				Byte2DoubleRBTreeMap.Entry entry = findLowest();
				return entry == null ? null : new BasicEntry(entry.key, entry.value);
			}
			Byte2DoubleRBTreeMap.Entry entry = m.findHigherNode(e);
			return entry == null || tooLow(entry.key) ? null : new BasicEntry(entry.key, entry.value);
		}
		
		class SubMapEntrySet extends AbstractObjectSet<Byte2DoubleMap.Entry> implements ObjectSet<Byte2DoubleMap.Entry> {
			@Override
			@Deprecated
			public boolean contains(Object o) {
				if(o instanceof Map.Entry) {
					if(o instanceof Byte2DoubleMap.Entry) {
						Byte2DoubleMap.Entry entry = (Byte2DoubleMap.Entry)o;
						if(entry.getKey() == null && comparator() == null) return false;
						Byte2DoubleRBTreeMap.Entry subEntry = m.findNode(entry.getByteKey());
						if(subEntry != null) return Double.doubleToLongBits(entry.getDoubleValue()) == Double.doubleToLongBits(subEntry.value);
					}
					else {
						Map.Entry<?, ?> entry = (Map.Entry<?, ?>)o;
						if(entry.getKey() == null && comparator() == null) return false;
						if(!(entry.getKey() instanceof Byte)) return false;
						Byte2DoubleRBTreeMap.Entry subEntry = m.findNode((Byte)entry.getKey());
						if(subEntry != null) return Objects.equals(entry.getValue(), Double.valueOf(subEntry.value));
					}
				}
				return false;
			}
			
			@Override
			@Deprecated
			public boolean remove(Object o) {
				if(o instanceof Map.Entry) {
					if(o instanceof Byte2DoubleMap.Entry) {
						Byte2DoubleMap.Entry entry = (Byte2DoubleMap.Entry)o;
						return NavigableSubMap.this.remove(entry.getByteKey(), entry.getDoubleValue());
					}
					Map.Entry<?, ?> entry = (Map.Entry<?, ?>)o;
					return NavigableSubMap.this.remove(entry.getKey(), entry.getValue());
				}
				return false;
			}
			@Override
			public ObjectIterator<Byte2DoubleMap.Entry> iterator() { return entryIterator(); }
			@Override
			public int size() { return NavigableSubMap.this.size(); }
			@Override
			public void clear() { NavigableSubMap.this.clear(); }
			
			@Override
			public void forEach(Consumer<? super Byte2DoubleMap.Entry> action) {
				Objects.requireNonNull(action);
				for(Byte2DoubleRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					action.accept(new BasicEntry(entry.key, entry.value));
			}
			
			@Override
			public <E> void forEach(E input, ObjectObjectConsumer<E, Byte2DoubleMap.Entry> action) {
				Objects.requireNonNull(action);
				for(Byte2DoubleRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					action.accept(input, new BasicEntry(entry.key, entry.value));
			}
			
			@Override
			public boolean matchesAny(Object2BooleanFunction<Byte2DoubleMap.Entry> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return false;
				BasicEntry subEntry = new BasicEntry();
				for(Byte2DoubleRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					subEntry.set(entry.key, entry.value);
					if(filter.getBoolean(subEntry)) return true;
				}
				return false;
			}
			
			@Override
			public boolean matchesNone(Object2BooleanFunction<Byte2DoubleMap.Entry> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return true;
				BasicEntry subEntry = new BasicEntry();
				for(Byte2DoubleRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					subEntry.set(entry.key, entry.value);
					if(filter.getBoolean(subEntry)) return false;
				}
				return true;
			}
			
			@Override
			public boolean matchesAll(Object2BooleanFunction<Byte2DoubleMap.Entry> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return true;
				BasicEntry subEntry = new BasicEntry();
				for(Byte2DoubleRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					subEntry.set(entry.key, entry.value);
					if(!filter.getBoolean(subEntry)) return false;
				}
				return true;
			}
			
			@Override
			public <E> E reduce(E identity, BiFunction<E, Byte2DoubleMap.Entry, E> operator) {
				Objects.requireNonNull(operator);
				E state = identity;
				for(Byte2DoubleRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					state = operator.apply(state, new BasicEntry(entry.key, entry.value));
				}
				return state;
			}
			
			@Override
			public Byte2DoubleMap.Entry reduce(ObjectObjectUnaryOperator<Byte2DoubleMap.Entry, Byte2DoubleMap.Entry> operator) {
				Objects.requireNonNull(operator);
				Byte2DoubleMap.Entry state = null;
				boolean empty = true;
				for(Byte2DoubleRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					if(empty) {
						empty = false;
						state = new BasicEntry(entry.key, entry.value);
						continue;
					}
					state = operator.apply(state, new BasicEntry(entry.key, entry.value));
				}
				return state;
			}
			
			@Override
			public Byte2DoubleMap.Entry findFirst(Object2BooleanFunction<Byte2DoubleMap.Entry> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return null;
				BasicEntry subEntry = new BasicEntry();
				for(Byte2DoubleRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					subEntry.set(entry.key, entry.value);
					if(filter.getBoolean(subEntry)) return subEntry;
				}
				return null;
			}
			
			@Override
			public int count(Object2BooleanFunction<Byte2DoubleMap.Entry> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return 0;
				int result = 0;
				BasicEntry subEntry = new BasicEntry();
				for(Byte2DoubleRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					subEntry.set(entry.key, entry.value);
					if(filter.getBoolean(subEntry)) result++;
				}
				return result;
			}
		}
		
		final class SubMapValues extends AbstractDoubleCollection {
			@Override
			public boolean add(double o) { throw new UnsupportedOperationException(); }
			
			@Override
			public boolean contains(double e) {
				return containsValue(e);
			}
			
			@Override
			public DoubleIterator iterator() { return valueIterator(); }
			
			@Override
			public int size() {
				return NavigableSubMap.this.size();
			}
			
			@Override
			public void clear() {
				NavigableSubMap.this.clear();
			}
			
			@Override
			public void forEach(DoubleConsumer action) {
				Objects.requireNonNull(action);
				for(Byte2DoubleRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					action.accept(entry.value);
			}
			
			@Override
			public <E> void forEach(E input, ObjectDoubleConsumer<E> action) {
				Objects.requireNonNull(action);
				for(Byte2DoubleRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					action.accept(input, entry.value);
			}
			
			@Override
			public boolean matchesAny(Double2BooleanFunction filter) {
				Objects.requireNonNull(filter);
				for(Byte2DoubleRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					if(filter.get(entry.value)) return true;
				return false;
			}
			
			@Override
			public boolean matchesNone(Double2BooleanFunction filter) {
				Objects.requireNonNull(filter);
				for(Byte2DoubleRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					if(filter.get(entry.value)) return false;
				return true;
			}
			
			@Override
			public boolean matchesAll(Double2BooleanFunction filter) {
				Objects.requireNonNull(filter);
				for(Byte2DoubleRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					if(!filter.get(entry.value)) return false;
				return true;
			}
			
			@Override
			public double reduce(double identity, DoubleDoubleUnaryOperator operator) {
				Objects.requireNonNull(operator);
				double state = identity;
				for(Byte2DoubleRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					state = operator.applyAsDouble(state, entry.value);
				return state;
			}
			
			@Override
			public double reduce(DoubleDoubleUnaryOperator operator) {
				Objects.requireNonNull(operator);
				double state = 0D;
				boolean empty = true;
				for(Byte2DoubleRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					if(empty) {
						empty = false;
						state = entry.value;
						continue;
					}
					state = operator.applyAsDouble(state, entry.value);
				}
				return state;
			}
			
			@Override
			public double findFirst(Double2BooleanFunction filter) {
				Objects.requireNonNull(filter);
				for(Byte2DoubleRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					if(filter.get(entry.value)) return entry.value;
				return 0D;
			}
			
			@Override
			public int count(Double2BooleanFunction filter) {
				Objects.requireNonNull(filter);
				int result = 0;
				for(Byte2DoubleRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					if(filter.get(entry.value)) result++;
				return result;
			}
		}
		
		final class SubMapEntrySetIterator extends SubMapEntryIterator implements ObjectListIterator<Byte2DoubleMap.Entry> {
			public SubMapEntrySetIterator(boolean descending) {
				super(descending);
			}
			
			public SubMapEntrySetIterator(Byte2DoubleRBTreeMap.Entry entry, Byte2DoubleRBTreeMap.Entry lowerFence, Byte2DoubleRBTreeMap.Entry upperFence) {
				super(entry, lowerFence, upperFence);
			}
			
			@Override
			public Byte2DoubleMap.Entry next() { return nextEntry(); }
			@Override
			public Byte2DoubleMap.Entry previous() { return previousEntry(); }
			
			@Override
			public void set(Byte2DoubleMap.Entry e) { throw new UnsupportedOperationException(); }
			@Override
			public void add(Byte2DoubleMap.Entry e) { throw new UnsupportedOperationException(); }
		}
		
		final class SubMapKeyIterator extends SubMapEntryIterator implements ByteListIterator {
			public SubMapKeyIterator(boolean descending) {
				super(descending);
			}
			
			public SubMapKeyIterator(Byte2DoubleRBTreeMap.Entry entry, Byte2DoubleRBTreeMap.Entry lowerFence, Byte2DoubleRBTreeMap.Entry upperFence) {
				super(entry, lowerFence, upperFence);
			}
			
			@Override
			public byte nextByte() { return nextEntry().key; }
			@Override
			public byte previousByte() { return previousEntry().key; }
			
			@Override
			public void set(byte e) { throw new UnsupportedOperationException(); }
			@Override
			public void add(byte e) { throw new UnsupportedOperationException(); }
		}
		
		final class SubMapValueIterator extends SubMapEntryIterator implements DoubleListIterator {
			public SubMapValueIterator(boolean descending) {
				super(descending);
			}
			
			public SubMapValueIterator(Byte2DoubleRBTreeMap.Entry entry, Byte2DoubleRBTreeMap.Entry lowerFence, Byte2DoubleRBTreeMap.Entry upperFence) {
				super(entry, lowerFence, upperFence);
			}
			
			@Override
			public double nextDouble() { return nextEntry().value; }
			@Override
			public double previousDouble() { return previousEntry().value; }
			
			@Override
			public void set(double e) { throw new UnsupportedOperationException(); }
			@Override
			public void add(double e) { throw new UnsupportedOperationException(); }
		}
		
		abstract class SubMapEntryIterator {
			Byte lowerFence;
			Byte upperFence;
			Byte2DoubleRBTreeMap.Entry next;
			Byte2DoubleRBTreeMap.Entry previous;
			Byte2DoubleRBTreeMap.Entry current;
			int index = 0;
			
			public SubMapEntryIterator(boolean descending) {
				this(descending ? findHighest() : findLowest(), fromStart ? null : findLowest(), toEnd ? null : findHighest());
			}
			
			public SubMapEntryIterator(Byte2DoubleRBTreeMap.Entry entry, Byte2DoubleRBTreeMap.Entry lowerFence, Byte2DoubleRBTreeMap.Entry upperFence) {
				next = entry;
				previous = entry.previous();
				this.lowerFence = lowerFence != null ? Byte.valueOf(lowerFence.key) : null;
				this.upperFence = upperFence != null ? Byte.valueOf(upperFence.key) : null;
			}
			
			public boolean hasNext() {
				return next != null && (upperFence == null || next.key == upperFence.byteValue());
			}
			
			public boolean hasPrevious() {
				return previous != null && (lowerFence == null || next.key == lowerFence.byteValue());
			}
			
			public int nextIndex() {
				return index;
			}
			
			public int previousIndex() {
				return index - 1;
			}
			
			protected void updateNext() {
				next = current.next();
			}
			
			protected void updatePrevious() {
				previous = current.previous();
			}
			
			public Byte2DoubleRBTreeMap.Entry nextEntry() {
				if(!hasNext()) throw new NoSuchElementException();
				current = previous = next;
				updateNext();
				index++;
				return current;
			}
			
			public Byte2DoubleRBTreeMap.Entry previousEntry() {
				if(!hasPrevious()) throw new NoSuchElementException();
				current = next = previous;
				updatePrevious();
				index--;
				return current;
			}
			
			public void remove() {
				if(current == null) throw new IllegalStateException();
				if(current == previous) index--;
				updateNext();
				updatePrevious();
				if(current.needsSuccessor()) next = current;
				m.removeNode(current);
				current = null;
			}
		}
	}
	
	private class EntrySet extends AbstractObjectSet<Byte2DoubleMap.Entry> {
		
		@Override
		public ObjectIterator<Byte2DoubleMap.Entry> iterator() {
			return new EntryIterator(false);
		}
		
		@Override
		@Deprecated
		public boolean contains(Object o) {
			if(o instanceof Map.Entry) {
				if(o instanceof Byte2DoubleMap.Entry) {
					Byte2DoubleMap.Entry entry = (Byte2DoubleMap.Entry)o;
					if(entry.getKey() == null && comparator() == null) return false;
					Byte2DoubleRBTreeMap.Entry subEntry = Byte2DoubleRBTreeMap.this.findNode(entry.getByteKey());
					if(subEntry != null) return Double.doubleToLongBits(entry.getDoubleValue()) == Double.doubleToLongBits(subEntry.value);
				}
				else {
					Map.Entry<?, ?> entry = (Map.Entry<?, ?>)o;
					if(entry.getKey() == null && comparator() == null) return false;
					if(!(entry.getKey() instanceof Byte)) return false;
					Byte2DoubleRBTreeMap.Entry subEntry = Byte2DoubleRBTreeMap.this.findNode((Byte)entry.getKey());
					if(subEntry != null) return Objects.equals(entry.getValue(), Double.valueOf(subEntry.value));
				}
			}
			return false;
		}
		
		@Override
		@Deprecated
		public boolean remove(Object o) {
			if(o instanceof Map.Entry) {
				if(o instanceof Byte2DoubleMap.Entry) {
					Byte2DoubleMap.Entry entry = (Byte2DoubleMap.Entry)o;
					return Byte2DoubleRBTreeMap.this.remove(entry.getByteKey(), entry.getDoubleValue());
				}
				Map.Entry<?, ?> entry = (Map.Entry<?, ?>)o;
				return Byte2DoubleRBTreeMap.this.remove(entry.getKey(), entry.getValue());
			}
			return false;
		}
		
		@Override
		public int size() {
			return Byte2DoubleRBTreeMap.this.size();
		}
		
		@Override
		public void clear() {
			Byte2DoubleRBTreeMap.this.clear();
		}
		
		@Override
		public void forEach(Consumer<? super Byte2DoubleMap.Entry> action) {
			Objects.requireNonNull(action);
			for(Byte2DoubleRBTreeMap.Entry entry = first;entry != null;entry = entry.next())
				action.accept(new BasicEntry(entry.key, entry.value));
		}
		
		@Override
		public <E> void forEach(E input, ObjectObjectConsumer<E, Byte2DoubleMap.Entry> action) {
			Objects.requireNonNull(action);
			for(Byte2DoubleRBTreeMap.Entry entry = first;entry != null;entry = entry.next())
				action.accept(input, new BasicEntry(entry.key, entry.value));
		}
		
		@Override
		public boolean matchesAny(Object2BooleanFunction<Byte2DoubleMap.Entry> filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return false;
			BasicEntry subEntry = new BasicEntry();
			for(Byte2DoubleRBTreeMap.Entry entry = first;entry != null;entry = entry.next()) {
				subEntry.set(entry.key, entry.value);
				if(filter.getBoolean(subEntry)) return true;
			}
			return false;
		}
		
		@Override
		public boolean matchesNone(Object2BooleanFunction<Byte2DoubleMap.Entry> filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return true;
			BasicEntry subEntry = new BasicEntry();
			for(Byte2DoubleRBTreeMap.Entry entry = first;entry != null;entry = entry.next()) {
				subEntry.set(entry.key, entry.value);
				if(filter.getBoolean(subEntry)) return false;
			}
			return true;
		}
		
		@Override
		public boolean matchesAll(Object2BooleanFunction<Byte2DoubleMap.Entry> filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return true;
			BasicEntry subEntry = new BasicEntry();
			for(Byte2DoubleRBTreeMap.Entry entry = first;entry != null;entry = entry.next()) {
				subEntry.set(entry.key, entry.value);
				if(!filter.getBoolean(subEntry)) return false;
			}
			return true;
		}
		
		@Override
		public <E> E reduce(E identity, BiFunction<E, Byte2DoubleMap.Entry, E> operator) {
			Objects.requireNonNull(operator);
			E state = identity;
			for(Byte2DoubleRBTreeMap.Entry entry = first;entry != null;entry = entry.next()) {
				state = operator.apply(state, new BasicEntry(entry.key, entry.value));
			}
			return state;
		}
		
		@Override
		public Byte2DoubleMap.Entry reduce(ObjectObjectUnaryOperator<Byte2DoubleMap.Entry, Byte2DoubleMap.Entry> operator) {
			Objects.requireNonNull(operator);
			Byte2DoubleMap.Entry state = null;
			boolean empty = true;
			for(Byte2DoubleRBTreeMap.Entry entry = first;entry != null;entry = entry.next()) {
				if(empty) {
					empty = false;
					state = new BasicEntry(entry.key, entry.value);
					continue;
				}
				state = operator.apply(state, new BasicEntry(entry.key, entry.value));
			}
			return state;
		}
		
		@Override
		public Byte2DoubleMap.Entry findFirst(Object2BooleanFunction<Byte2DoubleMap.Entry> filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return null;
			BasicEntry subEntry = new BasicEntry();
			for(Byte2DoubleRBTreeMap.Entry entry = first;entry != null;entry = entry.next()) {
				subEntry.set(entry.key, entry.value);
				if(filter.getBoolean(subEntry)) return subEntry;
			}
			return null;
		}
		
		@Override
		public int count(Object2BooleanFunction<Byte2DoubleMap.Entry> filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return 0;
			int result = 0;
			BasicEntry subEntry = new BasicEntry();
			for(Byte2DoubleRBTreeMap.Entry entry = first;entry != null;entry = entry.next()) {
				subEntry.set(entry.key, entry.value);
				if(filter.getBoolean(subEntry)) result++;
			}
			return result;
		}
	}
	
	static final class KeySet extends AbstractByteSet implements ByteNavigableSet {
		Byte2DoubleNavigableMap m;
		
		KeySet(Byte2DoubleNavigableMap m) {
			this.m = m;
		}
		
		@Override
		public void setDefaultMaxValue(byte e) { m.setDefaultMaxValue(e); }
		@Override
		public byte getDefaultMaxValue() { return m.getDefaultMaxValue(); }
		@Override
		public void setDefaultMinValue(byte e) { m.setDefaultMinValue(e); }
		@Override
		public byte getDefaultMinValue() { return m.getDefaultMinValue(); }
		@Override
		public boolean add(byte o) { throw new UnsupportedOperationException(); }
		@Override
		public boolean addAndMoveToFirst(byte o) { throw new UnsupportedOperationException(); }
		@Override
		public boolean addAndMoveToLast(byte o) { throw new UnsupportedOperationException(); }
		@Override
		public boolean moveToFirst(byte o) { throw new UnsupportedOperationException(); }
		@Override
		public boolean moveToLast(byte o) { throw new UnsupportedOperationException(); }
		@Override
		public ByteComparator comparator() { return m.comparator(); }
		@Override
		public byte lower(byte e) { return m.lowerKey(e); }
		@Override
		public byte floor(byte e) { return m.floorKey(e); }
		@Override
		public byte ceiling(byte e) { return m.ceilingKey(e); }
		@Override
		public byte higher(byte e) { return m.higherKey(e); }
		@Override
		public byte firstByte() { return m.firstByteKey(); }
		@Override
		public byte pollFirstByte() { return m.pollFirstByteKey(); }
		@Override
		public byte lastByte() { return m.lastByteKey(); }
		@Override
		public byte pollLastByte() { return m.pollLastByteKey(); }		
		@Override
		public boolean remove(byte o) { 
			int oldSize = m.size();
			m.remove(o); 
			return oldSize != m.size();
		}
		
		@Override
		public KeySet copy() { throw new UnsupportedOperationException(); }
		
		@Override
		public ByteNavigableSet subSet(byte fromElement, boolean fromInclusive, byte toElement, boolean toInclusive) {
			return new KeySet(m.subMap(fromElement, fromInclusive, toElement, toInclusive));
		}
		
		@Override
		public ByteNavigableSet headSet(byte toElement, boolean inclusive) {
			return new KeySet(m.headMap(toElement, inclusive));
		}
		
		@Override
		public ByteNavigableSet tailSet(byte fromElement, boolean inclusive) {
			return new KeySet(m.tailMap(fromElement, inclusive));
		}
		
		@Override
		public ByteNavigableSet descendingSet() {
			return new KeySet(m.descendingMap());
		}
		
		@Override
		public ByteBidirectionalIterator iterator() {
			if(m instanceof Byte2DoubleRBTreeMap) return ((Byte2DoubleRBTreeMap)m).keyIterator(false);
			return ((NavigableSubMap)m).keyIterator(false);
		}
		
		@Override
		public ByteBidirectionalIterator iterator(byte fromElement) {
			if(m instanceof Byte2DoubleRBTreeMap) return ((Byte2DoubleRBTreeMap)m).keyIterator(fromElement);
			return ((NavigableSubMap)m).keyIterator(fromElement);
		}
		
		@Override
		public ByteBidirectionalIterator descendingIterator() {
			if(m instanceof Byte2DoubleRBTreeMap) return ((Byte2DoubleRBTreeMap)m).keyIterator(true);
			return ((NavigableSubMap)m).keyIterator(true);
		}
		
		@Override
		public int size() {
			return m.size();
		}
		
		@Override
		public void clear() {
			m.clear();
		}
		
		protected Byte2DoubleRBTreeMap.Entry start() {
			if(m instanceof Byte2DoubleRBTreeMap) return ((Byte2DoubleRBTreeMap)m).first;
			return ((NavigableSubMap)m).findLowest();
		}
		
		protected Byte2DoubleRBTreeMap.Entry end() {
			if(m instanceof Byte2DoubleRBTreeMap) return null;
			return ((NavigableSubMap)m).findLowest();
		}
		
		protected Byte2DoubleRBTreeMap.Entry next(Byte2DoubleRBTreeMap.Entry entry) {
			if(m instanceof Byte2DoubleRBTreeMap) return entry.next();
			return ((NavigableSubMap)m).next(entry);
		}
		
		protected Byte2DoubleRBTreeMap.Entry previous(Byte2DoubleRBTreeMap.Entry entry) {
			if(m instanceof Byte2DoubleRBTreeMap) return entry.previous();
			return ((NavigableSubMap)m).previous(entry);
		}
		
		@Override
		public void forEach(ByteConsumer action) {
			Objects.requireNonNull(action);
			for(Byte2DoubleRBTreeMap.Entry entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				action.accept(entry.key);
		}
		
		@Override
		public <E> void forEach(E input, ObjectByteConsumer<E> action) {
			Objects.requireNonNull(action);
			for(Byte2DoubleRBTreeMap.Entry entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				action.accept(input, entry.key);
		}
		
		@Override
		public boolean matchesAny(Byte2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(Byte2DoubleRBTreeMap.Entry entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				if(filter.get(entry.key)) return true;
			return false;
		}
		
		@Override
		public boolean matchesNone(Byte2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(Byte2DoubleRBTreeMap.Entry entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				if(filter.get(entry.key)) return false;
			return true;
		}
		
		@Override
		public boolean matchesAll(Byte2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(Byte2DoubleRBTreeMap.Entry entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				if(!filter.get(entry.key)) return false;
			return true;
		}
		
		@Override
		public byte reduce(byte identity, ByteByteUnaryOperator operator) {
			Objects.requireNonNull(operator);
			byte state = identity;
			for(Byte2DoubleRBTreeMap.Entry entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				state = operator.applyAsByte(state, entry.key);
			return state;
		}
		
		@Override
		public byte reduce(ByteByteUnaryOperator operator) {
			Objects.requireNonNull(operator);
			byte state = (byte)0;
			boolean empty = true;
			for(Byte2DoubleRBTreeMap.Entry entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry)) {
				if(empty) {
					empty = false;
					state = entry.key;
					continue;
				}
				state = operator.applyAsByte(state, entry.key);
			}
			return state;
		}
		
		@Override
		public byte findFirst(Byte2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(Byte2DoubleRBTreeMap.Entry entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				if(filter.get(entry.key)) return entry.key;
			return (byte)0;
		}
		
		@Override
		public int count(Byte2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			int result = 0;
			for(Byte2DoubleRBTreeMap.Entry entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				if(filter.get(entry.key)) result++;
			return result;
		}
	}
	
	final class Values extends AbstractDoubleCollection {

		@Override
		public boolean add(double o) { throw new UnsupportedOperationException(); }
		
		@Override
		public boolean contains(double e) {
			return containsValue(e);
		}
		
		@Override
		public DoubleIterator iterator() { return new ValueIterator(false); }
		
		@Override
		public int size() {
			return Byte2DoubleRBTreeMap.this.size();
		}
		
		@Override
		public void clear() {
			Byte2DoubleRBTreeMap.this.clear();
		}
		
		@Override
		public void forEach(DoubleConsumer action) {
			Objects.requireNonNull(action);
			for(Byte2DoubleRBTreeMap.Entry entry = first;entry != null;entry = entry.next())
				action.accept(entry.value);
		}
		
		@Override
		public <E> void forEach(E input, ObjectDoubleConsumer<E> action) {
			Objects.requireNonNull(action);
			for(Byte2DoubleRBTreeMap.Entry entry = first;entry != null;entry = entry.next())
				action.accept(input, entry.value);
		}
		
		@Override
		public boolean matchesAny(Double2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(Byte2DoubleRBTreeMap.Entry entry = first;entry != null;entry = entry.next())
				if(filter.get(entry.value)) return true;
			return false;
		}
		
		@Override
		public boolean matchesNone(Double2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(Byte2DoubleRBTreeMap.Entry entry = first;entry != null;entry = entry.next())
				if(filter.get(entry.value)) return false;
			return true;
		}
		
		@Override
		public boolean matchesAll(Double2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(Byte2DoubleRBTreeMap.Entry entry = first;entry != null;entry = entry.next())
				if(!filter.get(entry.value)) return false;
			return true;
		}
		
		@Override
		public double reduce(double identity, DoubleDoubleUnaryOperator operator) {
			Objects.requireNonNull(operator);
			double state = identity;
			for(Byte2DoubleRBTreeMap.Entry entry = first;entry != null;entry = entry.next())
				state = operator.applyAsDouble(state, entry.value);
			return state;
		}
		
		@Override
		public double reduce(DoubleDoubleUnaryOperator operator) {
			Objects.requireNonNull(operator);
			double state = 0D;
			boolean empty = true;
			for(Byte2DoubleRBTreeMap.Entry entry = first;entry != null;entry = entry.next()) {
				if(empty) {
					empty = false;
					state = entry.value;
					continue;
				}
				state = operator.applyAsDouble(state, entry.value);
			}
			return state;
		}
		
		@Override
		public double findFirst(Double2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(Byte2DoubleRBTreeMap.Entry entry = first;entry != null;entry = entry.next())
				if(filter.get(entry.value)) return entry.value;
			return 0D;
		}
		
		@Override
		public int count(Double2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			int result = 0;
			for(Byte2DoubleRBTreeMap.Entry entry = first;entry != null;entry = entry.next())
				if(filter.get(entry.value)) result++;
			return result;
		}
	}
	
	final class EntryIterator extends MapEntryIterator implements ObjectListIterator<Byte2DoubleMap.Entry> {
		
		public EntryIterator(boolean descending) {
			super(descending);
		}
		
		public EntryIterator(Entry entry) {
			super(entry);
		}
		
		@Override
		public Byte2DoubleMap.Entry next() {
			return nextEntry();
		}
		
		@Override
		public Byte2DoubleMap.Entry previous() {
			return previousEntry();
		}
		
		@Override
		public void set(Byte2DoubleMap.Entry e) { throw new UnsupportedOperationException(); }
		
		@Override
		public void add(Byte2DoubleMap.Entry e) { throw new UnsupportedOperationException(); }
	}
	
	final class KeyIterator extends MapEntryIterator implements ByteListIterator {
		
		public KeyIterator(boolean descending) {
			super(descending);
		}
		
		public KeyIterator(Entry entry) {
			super(entry);
		}
		
		@Override
		public byte previousByte() { return previousEntry().key; }
		@Override
		public byte nextByte() { return nextEntry().key; }
		
		@Override
		public void set(byte e) { throw new UnsupportedOperationException(); }
		@Override
		public void add(byte e) { throw new UnsupportedOperationException(); }
	}
	
	final class ValueIterator extends MapEntryIterator implements DoubleListIterator {
		
		public ValueIterator(boolean descending) {
			super(descending);
		}
		
		public ValueIterator(Entry entry) {
			super(entry);
		}
		
		@Override
		public double previousDouble() {
			return previousEntry().value;
		}
		
		@Override
		public double nextDouble() {
			return nextEntry().value;
		}
		
		@Override
		public void set(double e) { throw new UnsupportedOperationException(); }
		
		@Override
		public void add(double e) { throw new UnsupportedOperationException(); }
	}
	
	abstract class MapEntryIterator {
		Entry next;
		Entry previous;
		Entry current;
		int index = 0;
		
		public MapEntryIterator(boolean descending) {
			if(descending) previous = last;
			else next = first;
		}
		
		public MapEntryIterator(Entry entry) {
			next = entry;
			previous = entry.previous();
		}
		
		public boolean hasNext() {
			return next != null;
		}
		
		public boolean hasPrevious() {
			return previous != null;
		}
		
		public int nextIndex() {
			return index;
		}
		
		public int previousIndex() {
			return index - 1;
		}
		
		protected void updateNext() {
			next = current.next();
		}
		
		protected void updatePrevious() {
			previous = current.previous();
		}
		
		public Entry nextEntry() {
			if(!hasNext()) throw new NoSuchElementException();
			current = previous = next;
			updateNext();
			index++;
			return current;
		}
		
		public Entry previousEntry() {
			if(!hasPrevious()) throw new NoSuchElementException();
			current = next = previous;
			updatePrevious();
			index--;
			return current;
		}
		
		public void remove() {
			if(current == null) throw new IllegalStateException();
			if(current == previous) index--;
			updateNext();
			updatePrevious();
			if(current.needsSuccessor()) next = current;
			removeNode(current);
			current = null;
		}
	}
	
	private static final class Entry implements Byte2DoubleMap.Entry
	{
		static final int BLACK = 1;
		
		byte key;
		double value;
		int state;
		Entry parent;
		Entry left;
		Entry right;
		
		Entry(byte key, double value, Entry parent) {
			this.key = key;
			this.value = value;
			this.parent = parent;
		}
		
		Entry copy() {
			Entry entry = new Entry(key, value, null);
			entry.state = state;
			if(left != null) {
				Entry newLeft = left.copy();
				entry.left = newLeft;
				newLeft.parent = entry;
			}
			if(right != null) {
				Entry newRight = right.copy();
				entry.right = newRight;
				newRight.parent = entry;
			}
			return entry;
		}
		
		@Override
		public byte getByteKey() {
			return key;
		}
		
		@Override
		public double getDoubleValue() {
			return value;
		}
		
		@Override
		public double setValue(double value) {
			double oldValue = this.value;
			this.value = value;
			return oldValue;
		}
		
		double addTo(double value) {
			double oldValue = this.value;
			this.value += value;
			return oldValue;
		}
		
		@Override
		public boolean equals(Object obj) {
			if(obj instanceof Map.Entry) {
				if(obj instanceof Byte2DoubleMap.Entry) {
					Byte2DoubleMap.Entry entry = (Byte2DoubleMap.Entry)obj;
					return key == entry.getByteKey() && Double.doubleToLongBits(value) == Double.doubleToLongBits(entry.getDoubleValue());
				}
				Map.Entry<?, ?> entry = (Map.Entry<?, ?>)obj;
				Object otherKey = entry.getKey();
				if(otherKey == null) return false;
				Object otherValue = entry.getValue();
				return otherKey instanceof Byte && otherValue instanceof Double && key == ((Byte)otherKey).byteValue() && Double.doubleToLongBits(value) == Double.doubleToLongBits(((Double)otherValue).doubleValue());
			}
			return false;
		}
		
		@Override
		public int hashCode() {
			return Byte.hashCode(key) ^ Double.hashCode(value);
		}
		
		@Override
		public String toString() {
			return Byte.toString(key) + "=" + Double.toString(value);
		}
		
		boolean isBlack() {
			return (state & BLACK) != 0;
		}
		
		void setBlack(boolean value) {
			if(value) state |= BLACK;
			else state &= ~BLACK;
		}
		
		boolean needsSuccessor() { return left != null && right != null; }
		
		boolean replace(Entry entry) {
			if(entry != null) entry.parent = parent;
			if(parent != null) {
				if(parent.left == this) parent.left = entry;
				else parent.right = entry;
			}
			return parent == null;
		}
		
		Entry next() {
			if(right != null) {
				Entry parent = right;
				while(parent.left != null) parent = parent.left;
				return parent;
			}
			Entry parent = this.parent;
			Entry control = this;
			while(parent != null && control == parent.right) {
				control = parent;
				parent = parent.parent;
			}
			return parent;
		}
		
		Entry previous() {
			if(left != null) {
				Entry parent = left;
				while(parent.right != null) parent = parent.right;
				return parent;
			}
			Entry parent = this.parent;
			Entry control = this;
			while(parent != null && control == parent.left) {
				control = parent;
				parent = parent.parent;
			}
			return parent;
		}
	}
}