package speiger.src.collections.bytes.maps.impl.tree;

import java.util.Map;
import java.util.function.Consumer;
import java.util.function.BiFunction;
import java.util.Objects;
import java.util.NoSuchElementException;

import speiger.src.collections.bytes.collections.ByteBidirectionalIterator;
import speiger.src.collections.bytes.functions.ByteComparator;
import speiger.src.collections.bytes.functions.ByteConsumer;
import speiger.src.collections.objects.functions.consumer.ObjectByteConsumer;
import speiger.src.collections.bytes.functions.function.Byte2BooleanFunction;
import speiger.src.collections.bytes.functions.consumer.ByteObjectConsumer;
import speiger.src.collections.bytes.functions.function.Byte2ObjectFunction;
import speiger.src.collections.bytes.functions.function.ByteObjectUnaryOperator;
import speiger.src.collections.bytes.functions.function.ByteByteUnaryOperator;
import speiger.src.collections.bytes.lists.ByteListIterator;
import speiger.src.collections.bytes.maps.abstracts.AbstractByte2ObjectMap;
import speiger.src.collections.bytes.maps.interfaces.Byte2ObjectMap;
import speiger.src.collections.bytes.maps.interfaces.Byte2ObjectNavigableMap;
import speiger.src.collections.bytes.sets.AbstractByteSet;
import speiger.src.collections.bytes.sets.ByteNavigableSet;
import speiger.src.collections.bytes.sets.ByteSet;
import speiger.src.collections.bytes.sets.ByteSortedSet;
import speiger.src.collections.bytes.utils.ByteIterators;
import speiger.src.collections.bytes.utils.maps.Byte2ObjectMaps;
import speiger.src.collections.objects.collections.AbstractObjectCollection;
import speiger.src.collections.objects.collections.ObjectCollection;
import speiger.src.collections.objects.collections.ObjectIterator;
import speiger.src.collections.objects.functions.ObjectSupplier;
import speiger.src.collections.objects.functions.function.ObjectObjectUnaryOperator;
import speiger.src.collections.objects.lists.ObjectListIterator;
import speiger.src.collections.objects.utils.ObjectIterators;

import speiger.src.collections.objects.functions.consumer.ObjectObjectConsumer;
import speiger.src.collections.objects.functions.function.Object2BooleanFunction;
import speiger.src.collections.objects.sets.AbstractObjectSet;
import speiger.src.collections.objects.sets.ObjectSet;

/**
 * A Simple Type Specific AVL TreeMap implementation that reduces boxing/unboxing.
 * It is using a bit more memory then <a href="https://github.com/vigna/fastutil">FastUtil</a>,
 * but it saves a lot of Performance on the Optimized removal and iteration logic.
 * Which makes the implementation actually useable and does not get outperformed by Javas default implementation.
 * @param <V> the type of elements maintained by this Collection
 */
public class Byte2ObjectAVLTreeMap<V> extends AbstractByte2ObjectMap<V> implements Byte2ObjectNavigableMap<V>
{
	/** The center of the Tree */
	protected transient Entry<V> tree;
	/** The Lowest possible Node */
	protected transient Entry<V> first;
	/** The Highest possible Node */
	protected transient Entry<V> last;
	/** The amount of elements stored in the Map */
	protected int size = 0;
	/** The Sorter of the Tree */
	protected transient ByteComparator comparator;
	
	/** the default return value for max searches */
	protected byte defaultMaxNotFound = Byte.MIN_VALUE;
	/** the default return value for min searches */
	protected byte defaultMinNotFound = Byte.MAX_VALUE;
	
	/** KeySet Cache */
	protected ByteNavigableSet keySet;
	/** Values Cache */
	protected ObjectCollection<V> values;
	/** EntrySet Cache */
	protected ObjectSet<Byte2ObjectMap.Entry<V>> entrySet;
	
	/**
	 * Default Constructor
	 */
	public Byte2ObjectAVLTreeMap() {
	}
	
	/**
	 * Constructor that allows to define the sorter
	 * @param comp the function that decides how the tree is sorted, can be null
	 */
	public Byte2ObjectAVLTreeMap(ByteComparator comp) {
		comparator = comp;
	}
	
	/**
	 * Helper constructor that allow to create a map from boxed values (it will unbox them)
	 * @param keys the keys that should be put into the map
	 * @param values the values that should be put into the map.
	 * @throws IllegalStateException if the keys and values do not match in lenght
	 */
	public Byte2ObjectAVLTreeMap(Byte[] keys, V[] values) {
		this(keys, values, null);
	}
	
	/**
	 * Helper constructor that has a custom sorter and allow to create a map from boxed values (it will unbox them)
	 * @param keys the keys that should be put into the map
	 * @param values the values that should be put into the map.
	 * @param comp the function that decides how the tree is sorted, can be null
	 * @throws IllegalStateException if the keys and values do not match in lenght
	 */
	public Byte2ObjectAVLTreeMap(Byte[] keys, V[] values, ByteComparator comp) {
		comparator = comp;
		if(keys.length != values.length) throw new IllegalStateException("Input Arrays are not equal size");
		for(int i = 0,m=keys.length;i<m;i++) put(keys[i].byteValue(), values[i]);
	}
	
	/**
	 * Helper constructor that allow to create a map from unboxed values
	 * @param keys the keys that should be put into the map
	 * @param values the values that should be put into the map.
	 * @throws IllegalStateException if the keys and values do not match in lenght
	 */
	public Byte2ObjectAVLTreeMap(byte[] keys, V[] values) {
		this(keys, values, null);
	}
	
	/**
	 * Helper constructor that has a custom sorter and allow to create a map from unboxed values
	 * @param keys the keys that should be put into the map
	 * @param values the values that should be put into the map.
	 * @param comp the function that decides how the tree is sorted, can be null
	 * @throws IllegalStateException if the keys and values do not match in lenght
	 */
	public Byte2ObjectAVLTreeMap(byte[] keys, V[] values, ByteComparator comp) {
		comparator = comp;
		if(keys.length != values.length) throw new IllegalStateException("Input Arrays are not equal size");
		for(int i = 0,m=keys.length;i<m;i++) put(keys[i], values[i]);
	}
	
	/**
	 * A Helper constructor that allows to create a Map with exactly the same values as the provided map.
	 * @param map the values that should be present in the map
	 */
	public Byte2ObjectAVLTreeMap(Map<? extends Byte, ? extends V> map) {
		this(map, null);
	}
	
	/**
	 * A Helper constructor that has a custom sorter and allows to create a Map with exactly the same values as the provided map.
	 * @param map the values that should be present in the map
	 * @param comp the function that decides how the tree is sorted, can be null
	 */
	public Byte2ObjectAVLTreeMap(Map<? extends Byte, ? extends V> map, ByteComparator comp) {
		comparator = comp;
		putAll(map);
	}
	
	/**
	 * A Type Specific Helper function that allows to create a new Map with exactly the same values as the provided map.
	 * @param map the values that should be present in the map
 	 */
	public Byte2ObjectAVLTreeMap(Byte2ObjectMap<V> map) {
		this(map, null);
	}
	
	/**
	 * A Type Specific Helper function that has a custom sorter and allows to create a new Map with exactly the same values as the provided map.
	 * @param map the values that should be present in the map
	 * @param comp the function that decides how the tree is sorted, can be null
 	 */
	public Byte2ObjectAVLTreeMap(Byte2ObjectMap<V> map, ByteComparator comp) {
		comparator = comp;
		putAll(map);
	}
	
	@Override
	public void setDefaultMaxValue(byte value) { defaultMaxNotFound = value; }
	@Override
	public byte getDefaultMaxValue() { return defaultMaxNotFound; }
	@Override
	public void setDefaultMinValue(byte value) { defaultMinNotFound = value; }
	@Override
	public byte getDefaultMinValue() { return defaultMinNotFound; }
	
	@Override
	public V put(byte key, V value) {
		if(tree == null) {
			tree = first = last = new Entry<>(key, value, null);
			size++;
			return getDefaultReturnValue();
		}
		int compare = 0;
		Entry<V> parent = tree;
		while(true) {
			if((compare = compare(key, parent.key)) == 0) return parent.setValue(value);
			if(compare < 0) {
				if(parent.left == null) break;
				parent = parent.left;
			}
			else if(compare > 0) {
				if(parent.right == null) break;
				parent = parent.right;
			}
		}
		Entry<V> adding = new Entry<>(key, value, parent);
		if(compare < 0)  {
			parent.left = adding;
			if(parent == first)	first = adding;
		}
		else {
			parent.right = adding;
			if(parent == last) last = adding;
		}
		fixAfterInsertion(adding);
		size++;
		return getDefaultReturnValue();
	}
	
	@Override
	public V putIfAbsent(byte key, V value) {
		if(tree == null) {
			tree = first = last = new Entry<>(key, value, null);
			size++;
			return getDefaultReturnValue();
		}
		int compare = 0;
		Entry<V> parent = tree;
		while(true) {
			if((compare = compare(key, parent.key)) == 0) return parent.value;
			if(compare < 0) {
				if(parent.left == null) break;
				parent = parent.left;
			}
			else if(compare > 0) {
				if(parent.right == null) break;
				parent = parent.right;
			}
		}
		Entry<V> adding = new Entry<>(key, value, parent);
		if(compare < 0)  {
			parent.left = adding;
			if(parent == first)	first = adding;
		}
		else  {
			parent.right = adding;
			if(parent == last) last = adding;
		}
		fixAfterInsertion(adding);
		size++;
		return getDefaultReturnValue();
	}
	
	@Override
	public V putAndMoveToFirst(byte key, V value) { throw new UnsupportedOperationException(); }
	@Override
	public V putAndMoveToLast(byte key, V value) { throw new UnsupportedOperationException(); }
	@Override
	public boolean moveToFirst(byte key) { throw new UnsupportedOperationException(); }
	@Override
	public boolean moveToLast(byte key) { throw new UnsupportedOperationException(); }
	@Override
	public V getAndMoveToFirst(byte key) { throw new UnsupportedOperationException(); }
	@Override
	public V getAndMoveToLast(byte key) { throw new UnsupportedOperationException(); }
	
	@Override
	public ByteComparator comparator() { return comparator; }
	
	@Override
	public boolean containsKey(byte key) {
		return findNode(key) != null;
	}
	

	@Override
	public V get(byte key) {
		Entry<V> node = findNode(key);
		return node == null ? getDefaultReturnValue() : node.value;
	}
	
	@Override
	public V getOrDefault(byte key, V defaultValue) {
		Entry<V> node = findNode(key);
		return node == null ? defaultValue : node.value;
	}
	
	@Override
	public byte firstByteKey() {
		if(tree == null) throw new NoSuchElementException();
		return first.key;
	}
	
	@Override
	public byte pollFirstByteKey() {
		if(tree == null) throw new NoSuchElementException();
		byte result = first.key;
		removeNode(first);
		return result;
	}
	
	@Override
	public byte lastByteKey() {
		if(tree == null) throw new NoSuchElementException();
		return last.key;
	}
	
	@Override
	public byte pollLastByteKey() {
		if(tree == null) throw new NoSuchElementException();
		byte result = last.key;
		removeNode(last);
		return result;
	}
	
	@Override
	public Byte2ObjectMap.Entry<V> firstEntry() {
		if(tree == null) throw new NoSuchElementException();
		return new BasicEntry<>(first.key, first.value);
	}
	
	@Override
	public Byte2ObjectMap.Entry<V> lastEntry() {
		if(tree == null) throw new NoSuchElementException();
		return new BasicEntry<>(last.key, last.value);
	}
	
	@Override
	public Byte2ObjectMap.Entry<V> pollFirstEntry() {
		if(tree == null) throw new NoSuchElementException();
		BasicEntry<V> entry = new BasicEntry<>(first.key, first.value);
		removeNode(first);
		return entry;
	}
	
	@Override
	public Byte2ObjectMap.Entry<V> pollLastEntry() {
		if(tree == null) throw new NoSuchElementException();
		BasicEntry<V> entry = new BasicEntry<>(last.key, last.value);
		removeNode(last);
		return entry;
	}
	
	@Override
	public V firstValue() {
		if(tree == null) throw new NoSuchElementException();
		return first.value;
	}
	
	@Override
	public V lastValue() {
		if(tree == null) throw new NoSuchElementException();
		return last.value;
	}
	
	@Override
	public V remove(byte key) {
		Entry<V> entry = findNode(key);
		if(entry == null) return getDefaultReturnValue();
		V value = entry.value;
		removeNode(entry);
		return value;
	}
	
	@Override
	public V removeOrDefault(byte key, V defaultValue) {
		Entry<V> entry = findNode(key);
		if(entry == null) return defaultValue;
		V value = entry.value;
		removeNode(entry);
		return value;
	}
	
	@Override
	public boolean remove(byte key, V value) {
		Entry<V> entry = findNode(key);
		if(entry == null || entry.value != value) return false;
		removeNode(entry);
		return true;
	}
	
	@Override
	public boolean replace(byte key, V oldValue, V newValue) {
		Entry<V> entry = findNode(key);
		if(entry == null || entry.value != oldValue) return false;
		entry.value = newValue;
		return true;
	}
	
	@Override
	public V replace(byte key, V value) {
		Entry<V> entry = findNode(key);
		if(entry == null) return getDefaultReturnValue();
		V oldValue = entry.value;
		entry.value = value;
		return oldValue;
	}
	
	@Override
	public V compute(byte key, ByteObjectUnaryOperator<V> mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		Entry<V> entry = findNode(key);
		if(entry == null) {
			V newValue = mappingFunction.apply(key, getDefaultReturnValue());
			if(Objects.equals(newValue, getDefaultReturnValue())) return newValue;
			put(key, newValue);
			return newValue;
		}
		V newValue = mappingFunction.apply(key, entry.value);
		if(Objects.equals(newValue, getDefaultReturnValue())) {
			removeNode(entry);
			return newValue;
		}
		entry.value = newValue;
		return newValue;
	}
	
	@Override
	public V computeIfAbsent(byte key, Byte2ObjectFunction<V> mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		Entry<V> entry = findNode(key);
		if(entry == null) {
			V newValue = mappingFunction.get(key);
			if(Objects.equals(newValue, getDefaultReturnValue())) return newValue;
			put(key, newValue);
			return newValue;
		}
		if(Objects.equals(entry.value, getDefaultReturnValue())) {
			V newValue = mappingFunction.get(key);
			if(Objects.equals(newValue, getDefaultReturnValue())) return newValue;
			entry.value = newValue;
		}
		return entry.value;
	}
	
	@Override
	public V supplyIfAbsent(byte key, ObjectSupplier<V> valueProvider) {
		Objects.requireNonNull(valueProvider);
		Entry<V> entry = findNode(key);
		if(entry == null) {
			V newValue = valueProvider.get();
			if(Objects.equals(newValue, getDefaultReturnValue())) return newValue;
			put(key, newValue);
			return newValue;
		}
		if(Objects.equals(entry.value, getDefaultReturnValue())) {
			V newValue = valueProvider.get();
			if(Objects.equals(newValue, getDefaultReturnValue())) return newValue;
			entry.value = newValue;
		}
		return entry.value;
	}
	
	@Override
	public V computeIfPresent(byte key, ByteObjectUnaryOperator<V> mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		Entry<V> entry = findNode(key);
		if(entry == null || Objects.equals(entry.value, getDefaultReturnValue())) return getDefaultReturnValue();
		V newValue = mappingFunction.apply(key, entry.value);
		if(Objects.equals(newValue, getDefaultReturnValue())) {
			removeNode(entry);
			return newValue;
		}
		entry.value = newValue;
		return newValue;
	}
	
	@Override
	public V merge(byte key, V value, ObjectObjectUnaryOperator<V, V> mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		Entry<V> entry = findNode(key);
		V newValue = entry == null || Objects.equals(entry.value, getDefaultReturnValue()) ? value : mappingFunction.apply(entry.value, value);
		if(Objects.equals(newValue, getDefaultReturnValue())) {
			if(entry != null)
				removeNode(entry);
		}
		else if(entry == null) put(key, newValue);
		else entry.value = newValue;
		return newValue;
	}
	
	@Override
	public void mergeAll(Byte2ObjectMap<V> m, ObjectObjectUnaryOperator<V, V> mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		for(Byte2ObjectMap.Entry<V> entry : Byte2ObjectMaps.fastIterable(m)) {
			byte key = entry.getByteKey();
			Entry<V> subEntry = findNode(key);
			V newValue = subEntry == null || Objects.equals(subEntry.value, getDefaultReturnValue()) ? entry.getValue() : mappingFunction.apply(subEntry.value, entry.getValue());
			if(Objects.equals(newValue, getDefaultReturnValue())) {
				if(subEntry != null)
					removeNode(subEntry);
			}
			else if(subEntry == null) put(key, newValue);
			else subEntry.value = newValue;
		}
	}
	
	@Override
	public void forEach(ByteObjectConsumer<V> action) {
		for(Entry<V> entry = first;entry != null;entry = entry.next())
			action.accept(entry.key, entry.value);
	}
	
	@Override
	public int size() { return size; }
	
	@Override
	public void clear() {
		size = 0;
		first = null;
		last = null;
		tree = null;
	}
	
	ByteListIterator keyIterator(boolean descending) {
		ByteListIterator iter = new KeyIterator(descending);
		return descending ? ByteIterators.invert(iter) : iter;
	}
	
	ByteListIterator keyIterator(byte key) {
		return new KeyIterator(findNode(key));
	}
	
	public Byte2ObjectAVLTreeMap<V> copy() {
		Byte2ObjectAVLTreeMap<V> set = new Byte2ObjectAVLTreeMap<>();
		set.size = size;
		if(tree != null) {
			set.tree = tree.copy();
			Entry<V> lastFound = null;
			for(Entry<V> entry = tree;entry != null;entry = entry.left) lastFound = entry;
			set.first = lastFound;
			lastFound = null;
			for(Entry<V> entry = tree;entry != null;entry = entry.right) lastFound = entry;
			set.last = lastFound;
		}
		return set;
	}
	
	@Override
	public ByteSortedSet keySet() {
		return navigableKeySet();
	}
	
	@Override
	public ObjectSet<Byte2ObjectMap.Entry<V>> byte2ObjectEntrySet() {
		if(entrySet == null) entrySet = new EntrySet();
		return entrySet;
	}
	
	@Override
	public ObjectCollection<V> values() {
		if(values == null) values = new Values();
		return values;
	}
	
	@Override
	public ByteNavigableSet navigableKeySet() {
		if(keySet == null) keySet = new KeySet<>(this);
		return keySet;
	}
	
	@Override
	public Byte2ObjectNavigableMap<V> descendingMap() {
		return new DescendingSubMap<>(this, true, (byte)0, true, true, (byte)0, true);
	}
	
	@Override
	public ByteNavigableSet descendingKeySet() {
		return descendingMap().navigableKeySet();
	}
	
	@Override
	public Byte2ObjectNavigableMap<V> subMap(byte fromKey, boolean fromInclusive, byte toKey, boolean toInclusive) {
		return new AscendingSubMap<>(this, false, fromKey, fromInclusive, false, toKey, toInclusive);
	}
	
	@Override
	public Byte2ObjectNavigableMap<V> headMap(byte toKey, boolean inclusive) {
		return new AscendingSubMap<>(this, true, (byte)0, true, false, toKey, inclusive);
	}
	
	@Override
	public Byte2ObjectNavigableMap<V> tailMap(byte fromKey, boolean inclusive) {
		return new AscendingSubMap<>(this, false, fromKey, inclusive, true, (byte)0, true);
	}
	
	@Override
	public byte lowerKey(byte e) {
		Entry<V> node = findLowerNode(e);
		return node != null ? node.key : getDefaultMinValue();
	}

	@Override
	public byte floorKey(byte e) {
		Entry<V> node = findFloorNode(e);
		return node != null ? node.key : getDefaultMinValue();
	}
	
	@Override
	public byte higherKey(byte e) {
		Entry<V> node = findHigherNode(e);
		return node != null ? node.key : getDefaultMaxValue();
	}

	@Override
	public byte ceilingKey(byte e) {
		Entry<V> node = findCeilingNode(e);
		return node != null ? node.key : getDefaultMaxValue();
	}
	
	@Override
	public Byte2ObjectMap.Entry<V> lowerEntry(byte key) {
		Entry<V> node = findLowerNode(key);
		return node != null ? new BasicEntry<>(node.key, node.value) : null;
	}
	
	@Override
	public Byte2ObjectMap.Entry<V> higherEntry(byte key) {
		Entry<V> node = findHigherNode(key);
		return node != null ? new BasicEntry<>(node.key, node.value) : null;
	}
	
	@Override
	public Byte2ObjectMap.Entry<V> floorEntry(byte key) {
		Entry<V> node = findFloorNode(key);
		return node != null ? new BasicEntry<>(node.key, node.value) : null;
	}
	
	@Override
	public Byte2ObjectMap.Entry<V> ceilingEntry(byte key) {
		Entry<V> node = findCeilingNode(key);
		return node != null ? new BasicEntry<>(node.key, node.value) : null;
	}
	
	protected Entry<V> findLowerNode(byte key) {
		Entry<V> entry = tree;
		while(entry != null) {
			if(compare(key, entry.key) > 0) {
				if(entry.right != null) entry = entry.right;
				else return entry;
			}
			else {
				if(entry.left != null) entry = entry.left;
				else {
					Entry<V> parent = entry.parent;
					while(parent != null && parent.left == entry) {
						entry = parent;
						parent = parent.parent;
					}
					return parent;
				}
			}
		}
		return null;
	}
	
	protected Entry<V> findFloorNode(byte key) {
		Entry<V> entry = tree;
		int compare;
		while(entry != null) {
			if((compare = compare(key, entry.key)) > 0) {
				if(entry.right == null) break;
				entry = entry.right;
				continue;
			}
			else if(compare < 0) {
				if(entry.left != null) entry = entry.left;
				else {
					Entry<V> parent = entry.parent;
					while(parent != null && parent.left == entry) {
						entry = parent;
						parent = parent.parent;
					}
					return parent;
				}
				continue;
			}
			break;
		}
		return entry;
	}
	
	protected Entry<V> findCeilingNode(byte key) {
		Entry<V> entry = tree;
		int compare;
		while(entry != null) {
			if((compare = compare(key, entry.key)) < 0) {
				if(entry.left == null) break;
				entry = entry.left;
				continue;
			}
			else if(compare > 0) {
				if(entry.right != null) entry = entry.right;
				else {
					Entry<V> parent = entry.parent;
					while(parent != null && parent.right == entry) {
						entry = parent;
						parent = parent.parent;
					}
					return parent;
				}
				continue;
			}
			break;
		}
		return entry;
	}
	
	protected Entry<V> findHigherNode(byte key) {
		Entry<V> entry = tree;
		while(entry != null) {
			if(compare(key, entry.key) < 0) {
				if(entry.left != null) entry = entry.left;
				else return entry;
			}
			else {
				if(entry.right != null) entry = entry.right;
				else {
					Entry<V> parent = entry.parent;
					while(parent != null && parent.right == entry) {
						entry = parent;
						parent = parent.parent;
					}
					return parent;
				}
			}
		}
		return null;
	}
	
	protected Entry<V> findNode(byte key) {
		Entry<V> node = tree;
		int compare;
		while(node != null) {
			if((compare = compare(key, node.key)) == 0) return node;
			if(compare < 0) node = node.left;
			else node = node.right;
		}
		return null;
	}
	
	protected void removeNode(Entry<V> entry) {
		size--;
		if(entry.needsSuccessor()) {
			Entry<V> successor = entry.next();
			entry.key = successor.key;
			entry.value = successor.value;
			entry = successor;
		}
		if(entry.previous() == null) first = entry.next();
		if(entry.next() == null) last = entry.previous();
		Entry<V> replacement = entry.left != null ? entry.left : entry.right;
		if(replacement != null) {
			if(entry.replace(replacement)) tree = replacement;
			entry.left = entry.right = entry.parent = null;
			fixAfterDeletion(replacement);
		}
		else if(entry.parent == null) tree = first = last = null;
		else {
			fixAfterDeletion(entry);
			entry.replace(null);
			entry.parent = null;
		}
	}
	
	protected void validate(byte k) { compare(k, k); }
	protected int compare(byte k, byte v) { return comparator != null ? comparator.compare(k, v) : Byte.compare(k, v);}
	
	/** From CLR */
	protected void rotateLeft(Entry<V> entry) {
		if(entry != null) {
			Entry<V> right = entry.right;
			entry.right = right.left;
			if(right.left != null) right.left.parent = entry;
			right.parent = entry.parent;
			if(entry.parent == null) tree = right;
			else if(entry.parent.left == entry) entry.parent.left = right;
			else entry.parent.right = right;
			right.left = entry;
			entry.parent = right;
			entry.updateHeight();
			right.updateHeight();
		}
	}
	
	/** From CLR */
	protected void rotateRight(Entry<V> entry) {
		if(entry != null) {
			Entry<V> left = entry.left;
			entry.left = left.right;
			if(left.right != null) left.right.parent = entry;
			left.parent = entry.parent;
			if(entry.parent == null) tree = left;
			else if(entry.parent.right == entry) entry.parent.right = left;
			else entry.parent.left = left;
			left.right = entry;
			entry.parent = left;
			entry.updateHeight();
			left.updateHeight();
		}
	}
	
	/** From CLR */
	protected void fixAfterInsertion(Entry<V> entry) {
		while(entry != null) {
			entry.updateHeight();
			int balance = entry.getBalance();
			if(balance > 1) {
				int compare = entry.left.getBalance();
				if(compare > 0) rotateRight(entry);
				else if(compare < 0) {
					rotateLeft(entry.left);
					rotateRight(entry);
				}
			}
			else if(balance < -1) {
				int compare = entry.right.getBalance();
				if(compare < 0) rotateLeft(entry);
				else if(compare > 0) {
					rotateRight(entry.right);
					rotateLeft(entry);
				}
			}
			entry = entry.parent;
		}
	}
	
	/** From CLR */
	protected void fixAfterDeletion(Entry<V> entry) {
		if(entry != null) {
			entry.updateHeight();
			int balance = entry.getBalance();
			if(balance > 1) {
				int subBalance = entry.left.getBalance();
				if(subBalance >= 0) rotateRight(entry);
				else {
					rotateLeft(entry.left);
					rotateRight(entry);
				}
			}
			else if(balance < -1)
			{
				int subBalance = entry.right.getBalance();
				if(subBalance <= 0) rotateLeft(entry);
				else {
					rotateRight(entry.right);
					rotateLeft(entry);
				}
			}
			entry = entry.parent;
		}
	}
	
	static final class AscendingSubMap<V> extends NavigableSubMap<V> {

		public AscendingSubMap(Byte2ObjectAVLTreeMap<V> m, boolean fromStart, byte low, boolean loInclusive, boolean toEnd, byte high, boolean hiInclusive) {
			super(m, fromStart, low, loInclusive, toEnd, high, hiInclusive);
		}
		
		@Override
		ByteListIterator keyIterator(boolean descending) { return new SubMapKeyIterator(descending); }
		@Override
		ByteListIterator keyIterator(byte key) {
			Byte2ObjectAVLTreeMap.Entry<V> entry = m.findNode(key);
			return entry == null || !inClosedRange(key) ? null : new SubMapKeyIterator(entry, fromStart ? null : findLowest(), toEnd ? null : findHighest());
		}
		@Override
		ObjectListIterator<V> valueIterator() { return new SubMapValueIterator(false); }
		@Override
		ObjectListIterator<Byte2ObjectMap.Entry<V>> entryIterator() { return new SubMapEntrySetIterator(false); }
		
		@Override
		public Byte2ObjectNavigableMap<V> descendingMap() {
			if(descendingMap == null) descendingMap = new DescendingSubMap<>(m, fromStart, low, loInclusive, toEnd, high, hiInclusive);
			return descendingMap;
		}
		
		Byte2ObjectAVLTreeMap.Entry<V> next(Byte2ObjectAVLTreeMap.Entry<V> entry) {
			return entry.next();
		}
		
		Byte2ObjectAVLTreeMap.Entry<V> previous(Byte2ObjectAVLTreeMap.Entry<V> entry) {
			return entry.previous();
		}
		
		@Override
		public Byte2ObjectNavigableMap<V> subMap(byte fromKey, boolean fromInclusive, byte toKey, boolean toInclusive) {
			if (!inRange(fromKey, fromInclusive)) throw new IllegalArgumentException("fromKey out of range");
			if (!inRange(toKey, toInclusive)) throw new IllegalArgumentException("toKey out of range");
			return new AscendingSubMap<>(m, false, fromKey, fromInclusive, false, toKey, toInclusive);
		}
		
		@Override
		public Byte2ObjectNavigableMap<V> headMap(byte toKey, boolean inclusive) {
			if (!inRange(toKey, inclusive)) throw new IllegalArgumentException("toKey out of range");
			return new AscendingSubMap<>(m, fromStart, low, loInclusive, false, toKey, inclusive);
		}
		
		@Override
		public Byte2ObjectNavigableMap<V> tailMap(byte fromKey, boolean inclusive) {
			if (!inRange(fromKey, inclusive)) throw new IllegalArgumentException("fromKey out of range");
			return new AscendingSubMap<>(m, false, fromKey, inclusive, toEnd, high, hiInclusive);
		}
	}
	
	static final class DescendingSubMap<V> extends NavigableSubMap<V> {

		public DescendingSubMap(Byte2ObjectAVLTreeMap<V> m, boolean fromStart, byte low, boolean loInclusive, boolean toEnd, byte high, boolean hiInclusive) {
			super(m, fromStart, low, loInclusive, toEnd, high, hiInclusive);
		}
		
		@Override
		public byte firstByteKey() { return super.lastByteKey(); }
		
		@Override
		public byte pollFirstByteKey() { return super.pollLastByteKey(); }
		
		@Override
		public byte lastByteKey() { return super.firstByteKey(); }
		
		@Override
		public byte pollLastByteKey() { return super.pollFirstByteKey(); }
		
		@Override
		public V firstValue() { return super.lastValue(); }
		
		@Override
		public V lastValue() { return super.firstValue(); }
		
		@Override
		public Byte2ObjectMap.Entry<V> firstEntry() { return super.lastEntry(); }
		
		@Override
		public Byte2ObjectMap.Entry<V> lastEntry() { return super.firstEntry(); }
		
		@Override
		public Byte2ObjectMap.Entry<V> pollFirstEntry() { return super.pollLastEntry(); }
		
		@Override
		public Byte2ObjectMap.Entry<V> pollLastEntry() { return super.pollFirstEntry(); }
		
		@Override
		public byte lowerKey(byte e) { return super.higherKey(e); }
		
		@Override
		public byte floorKey(byte e) { return super.ceilingKey(e); }
		
		@Override
		public byte ceilingKey(byte e) { return super.floorKey(e); }
		
		@Override
		public byte higherKey(byte e) { return super.lowerKey(e); }
		
		@Override
		public Byte2ObjectMap.Entry<V> lowerEntry(byte e) { return super.higherEntry(e); }
		
		@Override
		public Byte2ObjectMap.Entry<V> floorEntry(byte e) { return super.ceilingEntry(e); }
		
		@Override
		public Byte2ObjectMap.Entry<V> ceilingEntry(byte e) { return super.floorEntry(e); }
		
		@Override
		public Byte2ObjectMap.Entry<V> higherEntry(byte e) { return super.lowerEntry(e); }
		
		protected Byte2ObjectAVLTreeMap.Entry<V> findLowest() {
			if(fromStart) return m.first;
			Byte2ObjectAVLTreeMap.Entry<V> entry = loInclusive ? m.findFloorNode(high) : m.findLowerNode(high);
			return entry == null || tooHigh(entry.key) ? null : entry;
		}
		
		protected Byte2ObjectAVLTreeMap.Entry<V> findHighest() {
			if(toEnd) return m.last;
			Byte2ObjectAVLTreeMap.Entry<V> entry = hiInclusive ? m.findCeilingNode(low) : m.findHigherNode(low);
			return entry == null || tooLow(entry.key) ? null : entry;
		}
		
		@Override
		ByteListIterator keyIterator(boolean descending) {
			ByteListIterator iter = new SubMapKeyIterator(!descending);
			return descending ? iter : ByteIterators.invert(iter);
		}

		@Override
		ByteListIterator keyIterator(byte key) {
			Byte2ObjectAVLTreeMap.Entry<V> entry = m.findNode(key);
			return entry == null || !inClosedRange(key) ? null : ByteIterators.invert(new SubMapKeyIterator(entry, fromStart ? null : findLowest(), toEnd ? null : findHighest()));
		}
		
		Byte2ObjectAVLTreeMap.Entry<V> next(Byte2ObjectAVLTreeMap.Entry<V> entry) {
			return entry.previous();
		}
		
		Byte2ObjectAVLTreeMap.Entry<V> previous(Byte2ObjectAVLTreeMap.Entry<V> entry) {
			return entry.next();
		}
		
		@Override
		public ByteComparator comparator() { return m.comparator() == null ? null : m.comparator().reversed(); }
		
		@Override
		ObjectListIterator<V> valueIterator() { return ObjectIterators.invert(new SubMapValueIterator(true)); }
		
		@Override
		ObjectListIterator<Byte2ObjectMap.Entry<V>> entryIterator() { return ObjectIterators.invert(new SubMapEntrySetIterator(true)); }
		
		@Override
		public Byte2ObjectNavigableMap<V> descendingMap() {
			if(descendingMap == null) descendingMap = new AscendingSubMap<>(m, fromStart, low, loInclusive, toEnd, high, hiInclusive);
			return descendingMap;
		}

		@Override
		public Byte2ObjectNavigableMap<V> subMap(byte fromKey, boolean fromInclusive, byte toKey, boolean toInclusive) {
			if (!inRange(fromKey, fromInclusive)) throw new IllegalArgumentException("fromKey out of range");
			if (!inRange(toKey, toInclusive)) throw new IllegalArgumentException("toKey out of range");
			return new DescendingSubMap<>(m, false, fromKey, fromInclusive, false, toKey, toInclusive);
		}
		
		@Override
		public Byte2ObjectNavigableMap<V> headMap(byte toKey, boolean inclusive) {
			if (!inRange(toKey, inclusive)) throw new IllegalArgumentException("toKey out of range");
			return new DescendingSubMap<>(m, fromStart, low, loInclusive, false, toKey, inclusive);
		}
		
		@Override
		public Byte2ObjectNavigableMap<V> tailMap(byte fromKey, boolean inclusive) {
			if (!inRange(fromKey, inclusive)) throw new IllegalArgumentException("fromKey out of range");
			return new DescendingSubMap<>(m, false, fromKey, inclusive, toEnd, high, hiInclusive);
		}
	}
	
	static abstract class NavigableSubMap<V> extends AbstractByte2ObjectMap<V> implements Byte2ObjectNavigableMap<V> {
		final Byte2ObjectAVLTreeMap<V> m;
		final byte low;
		final byte high;
		final boolean fromStart;
		final boolean toEnd;
		final boolean loInclusive;
		final boolean hiInclusive;
		transient Byte2ObjectNavigableMap<V> descendingMap;
		transient SubMapEntrySet entrySet;
		transient KeySet<V> keySet;
		transient SubMapValues values;
		
		NavigableSubMap(Byte2ObjectAVLTreeMap<V> m, boolean fromStart, byte low, boolean loInclusive, boolean toEnd, byte high, boolean hiInclusive) {
			this.m = m;
			this.low = low;
			this.high = high;
			this.fromStart = fromStart;
			this.toEnd = toEnd;
			this.loInclusive = loInclusive;
			this.hiInclusive = hiInclusive;
		}
		
		abstract ByteListIterator keyIterator(boolean descending);
		abstract ByteListIterator keyIterator(byte key);
		abstract ObjectListIterator<V> valueIterator();
		abstract ObjectListIterator<Byte2ObjectMap.Entry<V>> entryIterator();
		abstract Byte2ObjectAVLTreeMap.Entry<V> next(Byte2ObjectAVLTreeMap.Entry<V> entry);
		abstract Byte2ObjectAVLTreeMap.Entry<V> previous(Byte2ObjectAVLTreeMap.Entry<V> entry);
		
		@Override
		public NavigableSubMap<V> copy() { throw new UnsupportedOperationException(); }
		@Override
		public abstract Byte2ObjectNavigableMap<V> descendingMap();
		@Override
		public abstract Byte2ObjectNavigableMap<V> subMap(byte fromKey, boolean fromInclusive, byte toKey, boolean toInclusive);
		@Override
		public abstract Byte2ObjectNavigableMap<V> headMap(byte toKey, boolean inclusive);
		@Override
		public abstract Byte2ObjectNavigableMap<V> tailMap(byte fromKey, boolean inclusive); 
		
		boolean tooLow(byte key) { return !fromStart && (loInclusive ? m.compare(key, low) < 0 : m.compare(key, low) <= 0); }
		boolean tooHigh(byte key) { return !toEnd && (hiInclusive ? m.compare(key, high) > 0 : m.compare(key, high) >= 0); }
		boolean inRange(byte key) { return !tooLow(key) && !tooHigh(key); }
		boolean inClosedRange(byte key) { return (fromStart || m.compare(key, low) >= 0) && (toEnd || m.compare(high, key) >= 0); }
		boolean inRange(byte key, boolean inclusive) { return inclusive ? inRange(key) : inClosedRange(key); }
		
		@Override
		public void setDefaultMaxValue(byte e) { m.setDefaultMaxValue(e); }
		@Override
		public byte getDefaultMaxValue() { return m.getDefaultMaxValue(); }
		@Override
		public void setDefaultMinValue(byte e) { m.setDefaultMinValue(e); }
		@Override
		public byte getDefaultMinValue() { return m.getDefaultMinValue(); }
		
		@Override
		public AbstractByte2ObjectMap<V> setDefaultReturnValue(V v) { 
			m.setDefaultReturnValue(v);
			return this;
		}
		
		@Override
		public V getDefaultReturnValue() { return m.getDefaultReturnValue(); }
		
		@Override
		public V putAndMoveToFirst(byte key, V value) { throw new UnsupportedOperationException(); }
		@Override
		public V putAndMoveToLast(byte key, V value) { throw new UnsupportedOperationException(); }
		@Override
		public boolean moveToFirst(byte key) { throw new UnsupportedOperationException(); }
		@Override
		public boolean moveToLast(byte key) { throw new UnsupportedOperationException(); }
		@Override
		public V getAndMoveToFirst(byte key) { throw new UnsupportedOperationException(); }
		@Override
		public V getAndMoveToLast(byte key) { throw new UnsupportedOperationException(); }
		@Override
		public ByteComparator comparator() { return m.comparator(); }
		@Override
		public boolean containsKey(byte key) { return inRange(key) && m.containsKey(key); }
		
		@Override
		public byte firstByteKey() { 
			Byte2ObjectAVLTreeMap.Entry<V> entry = findLowest();
			return entry == null ? getDefaultMaxValue() : entry.key;
		}
		
		@Override
		public byte pollFirstByteKey() {
			if(fromStart) return m.pollFirstByteKey();
			Byte2ObjectAVLTreeMap.Entry<V> entry = loInclusive ? m.findCeilingNode(low) : m.findHigherNode(low);
			if(entry != null && !tooHigh(entry.key)) {
				byte value = entry.key;
				m.removeNode(entry);
				return value;
			}
			return getDefaultMaxValue();
		}
		
		@Override
		public byte lastByteKey() {
			Byte2ObjectAVLTreeMap.Entry<V> entry = findHighest();
			return entry == null ? getDefaultMinValue() : entry.key;	
		}
		
		@Override
		public byte pollLastByteKey() {
			if(toEnd) return m.pollLastByteKey();
			Byte2ObjectAVLTreeMap.Entry<V> entry = hiInclusive ? m.findFloorNode(high) : m.findLowerNode(high);
			if(entry != null && !tooLow(entry.key)) {
				byte value = entry.key;
				m.removeNode(entry);
				return value;
			}
			return getDefaultMinValue();
		}
		
		@Override
		public V firstValue() {
			Byte2ObjectAVLTreeMap.Entry<V> entry = findLowest();
			return entry == null ? getDefaultReturnValue() : entry.value;
		}
		
		@Override
		public V lastValue() {
			Byte2ObjectAVLTreeMap.Entry<V> entry = findHighest();
			return entry == null ? getDefaultReturnValue() : entry.value;
		}
		
		protected Byte2ObjectAVLTreeMap.Entry<V> findLowest() {
			if(fromStart) return m.first;
			Byte2ObjectAVLTreeMap.Entry<V> entry = loInclusive ? m.findCeilingNode(low) : m.findHigherNode(low);
			return entry == null || tooHigh(entry.key) ? null : entry;
		}
		
		protected Byte2ObjectAVLTreeMap.Entry<V> findHighest() {
			if(toEnd) return m.last;
			Byte2ObjectAVLTreeMap.Entry<V> entry = hiInclusive ? m.findFloorNode(high) : m.findLowerNode(high);
			return entry == null || tooLow(entry.key) ? null : entry;
		}
		
		@Override
		public V put(byte key, V value) {
			if(!inRange(key)) throw new IllegalArgumentException("key out of range");
			return m.put(key, value);
		}
		
		@Override
		public V putIfAbsent(byte key, V value) {
			if(!inRange(key)) throw new IllegalArgumentException("key out of range");
			return m.putIfAbsent(key, value);
		}
		
		@Override
		public V get(byte key) {
			return inRange(key) ? m.get(key) : getDefaultReturnValue();
		}
		
		@Override
		public V getOrDefault(byte key, V defaultValue) {
			return inRange(key) ? m.getOrDefault(key, defaultValue) : getDefaultReturnValue();
		}
		
		@Override
		public V remove(byte key) {
			return inRange(key) ? m.remove(key) : getDefaultReturnValue();
		}
		
		@Override
		public V removeOrDefault(byte key, V defaultValue) {
			return inRange(key) ? m.remove(key) : defaultValue;
		}
		
		@Override
		public boolean remove(byte key, V value) {
			return inRange(key) && m.remove(key, value);
		}
		
		@Override
		public ObjectSet<Byte2ObjectMap.Entry<V>> byte2ObjectEntrySet() {
			if(entrySet == null) entrySet = new SubMapEntrySet();
			return entrySet;
		}
		
		@Override
		public ByteSet keySet() { return navigableKeySet(); }
		
		@Override
		public ByteNavigableSet navigableKeySet() {
			if(keySet == null) keySet = new KeySet<>(this);
			return keySet;
		}
		
		@Override
		public ByteNavigableSet descendingKeySet() {
			return descendingMap().navigableKeySet();
		}
		
		@Override
		public ObjectCollection<V> values() {
			if(values == null) values = new SubMapValues();
			return values;
		}
		
		@Override
		public void forEach(ByteObjectConsumer<V> action) {
			for(Byte2ObjectAVLTreeMap.Entry<V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
				action.accept(entry.key, entry.value);
		}
		
		@Override
		public int size() {
			return (fromStart && toEnd) ? m.size() : entrySet().size();
		}
		
		@Override
		public Byte2ObjectMap.Entry<V> firstEntry() {
			Byte2ObjectAVLTreeMap.Entry<V> entry = findLowest();
			return entry == null ? null : new BasicEntry<>(entry.key, entry.value);
		}
		
		@Override
		public Byte2ObjectMap.Entry<V> lastEntry() {
			Byte2ObjectAVLTreeMap.Entry<V> entry = findHighest();
			return entry == null ? null : new BasicEntry<>(entry.key, entry.value);
		}
		
		@Override
		public Byte2ObjectMap.Entry<V> pollFirstEntry() {
			Byte2ObjectAVLTreeMap.Entry<V> entry = findLowest();
			if(entry == null) return null;
			BasicEntry<V> result = new BasicEntry<>(entry.key, entry.value);
			m.removeNode(entry);
			return result;
		}
		
		@Override
		public Byte2ObjectMap.Entry<V> pollLastEntry() {
			Byte2ObjectAVLTreeMap.Entry<V> entry = findHighest();
			if(entry == null) return null;
			BasicEntry<V> result = new BasicEntry<>(entry.key, entry.value);
			m.removeNode(entry);
			return result;
		}
		
		@Override
		public byte lowerKey(byte e) {
			if(tooHigh(e)) {
				Byte2ObjectAVLTreeMap.Entry<V> entry = findHighest();
				return entry == null ? getDefaultMinValue() : entry.key;
			}
			Byte2ObjectAVLTreeMap.Entry<V> entry = m.findLowerNode(e);
			return entry == null || tooHigh(entry.key) ? getDefaultMaxValue() : entry.key;
		}
		
		@Override
		public byte floorKey(byte e) {
			if(tooHigh(e)) {
				Byte2ObjectAVLTreeMap.Entry<V> entry = findHighest();
				return entry == null ? getDefaultMinValue() : entry.key;
			}
			Byte2ObjectAVLTreeMap.Entry<V> entry = m.findFloorNode(e);
			return entry == null || tooHigh(entry.key) ? getDefaultMaxValue() : entry.key;
		}
		
		@Override
		public byte ceilingKey(byte e) {
			if(tooLow(e)) {
				Byte2ObjectAVLTreeMap.Entry<V> entry = findLowest();
				return entry == null ? getDefaultMaxValue() : entry.key;
			}
			Byte2ObjectAVLTreeMap.Entry<V> entry = m.findCeilingNode(e);
			return entry == null || tooLow(entry.key) ? getDefaultMinValue() : entry.key;
		}
		
		@Override
		public byte higherKey(byte e) {
			if(tooLow(e)) {
				Byte2ObjectAVLTreeMap.Entry<V> entry = findLowest();
				return entry == null ? getDefaultMaxValue() : entry.key;
			}
			Byte2ObjectAVLTreeMap.Entry<V> entry = m.findHigherNode(e);
			return entry == null || tooLow(entry.key) ? getDefaultMinValue() : entry.key;
		}
		
		@Override
		public Byte2ObjectMap.Entry<V> lowerEntry(byte e) {
			if(tooHigh(e)) {
				Byte2ObjectAVLTreeMap.Entry<V> entry = findHighest();
				return entry == null ? null : new BasicEntry<>(entry.key, entry.value);
			}
			Byte2ObjectAVLTreeMap.Entry<V> entry = m.findLowerNode(e);
			return entry == null || tooHigh(entry.key) ? null : new BasicEntry<>(entry.key, entry.value);
		}
		
		@Override
		public Byte2ObjectMap.Entry<V> floorEntry(byte e) {
			if(tooHigh(e)) {
				Byte2ObjectAVLTreeMap.Entry<V> entry = findHighest();
				return entry == null ? null : new BasicEntry<>(entry.key, entry.value);
			}
			Byte2ObjectAVLTreeMap.Entry<V> entry = m.findFloorNode(e);
			return entry == null || tooHigh(entry.key) ? null : new BasicEntry<>(entry.key, entry.value);
		}
		
		@Override
		public Byte2ObjectMap.Entry<V> ceilingEntry(byte e) {
			if(tooLow(e)) {
				Byte2ObjectAVLTreeMap.Entry<V> entry = findLowest();
				return entry == null ? null : new BasicEntry<>(entry.key, entry.value);
			}
			Byte2ObjectAVLTreeMap.Entry<V> entry = m.findCeilingNode(e);
			return entry == null || tooLow(entry.key) ? null : new BasicEntry<>(entry.key, entry.value);
		}
		
		@Override
		public Byte2ObjectMap.Entry<V> higherEntry(byte e) {
			if(tooLow(e)) {
				Byte2ObjectAVLTreeMap.Entry<V> entry = findLowest();
				return entry == null ? null : new BasicEntry<>(entry.key, entry.value);
			}
			Byte2ObjectAVLTreeMap.Entry<V> entry = m.findHigherNode(e);
			return entry == null || tooLow(entry.key) ? null : new BasicEntry<>(entry.key, entry.value);
		}
		
		class SubMapEntrySet extends AbstractObjectSet<Byte2ObjectMap.Entry<V>> implements ObjectSet<Byte2ObjectMap.Entry<V>> {
			@Override
			@Deprecated
			public boolean contains(Object o) {
				if(o instanceof Map.Entry) {
					if(o instanceof Byte2ObjectMap.Entry) {
						Byte2ObjectMap.Entry<V> entry = (Byte2ObjectMap.Entry<V>)o;
						if(entry.getKey() == null && comparator() == null) return false;
						Byte2ObjectAVLTreeMap.Entry<V> subEntry = m.findNode(entry.getByteKey());
						if(subEntry != null) return Objects.equals(entry.getValue(), subEntry.value);
					}
					else {
						Map.Entry<?, ?> entry = (Map.Entry<?, ?>)o;
						if(entry.getKey() == null && comparator() == null) return false;
						if(!(entry.getKey() instanceof Byte)) return false;
						Byte2ObjectAVLTreeMap.Entry<V> subEntry = m.findNode((Byte)entry.getKey());
						if(subEntry != null) return Objects.equals(entry.getValue(), subEntry.value);
					}
				}
				return false;
			}
			
			@Override
			@Deprecated
			public boolean remove(Object o) {
				if(o instanceof Map.Entry) {
					if(o instanceof Byte2ObjectMap.Entry) {
						Byte2ObjectMap.Entry<V> entry = (Byte2ObjectMap.Entry<V>)o;
						return NavigableSubMap.this.remove(entry.getByteKey(), entry.getValue());
					}
					Map.Entry<?, ?> entry = (Map.Entry<?, ?>)o;
					return NavigableSubMap.this.remove(entry.getKey(), entry.getValue());
				}
				return false;
			}
			@Override
			public ObjectIterator<Byte2ObjectMap.Entry<V>> iterator() { return entryIterator(); }
			@Override
			public int size() { return NavigableSubMap.this.size(); }
			@Override
			public void clear() { NavigableSubMap.this.clear(); }
			
			@Override
			public void forEach(Consumer<? super Byte2ObjectMap.Entry<V>> action) {
				Objects.requireNonNull(action);
				for(Byte2ObjectAVLTreeMap.Entry<V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					action.accept(new BasicEntry<>(entry.key, entry.value));
			}
			
			@Override
			public <E> void forEach(E input, ObjectObjectConsumer<E, Byte2ObjectMap.Entry<V>> action) {
				Objects.requireNonNull(action);
				for(Byte2ObjectAVLTreeMap.Entry<V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					action.accept(input, new BasicEntry<>(entry.key, entry.value));
			}
			
			@Override
			public boolean matchesAny(Object2BooleanFunction<Byte2ObjectMap.Entry<V>> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return false;
				BasicEntry<V> subEntry = new BasicEntry<>();
				for(Byte2ObjectAVLTreeMap.Entry<V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					subEntry.set(entry.key, entry.value);
					if(filter.getBoolean(subEntry)) return true;
				}
				return false;
			}
			
			@Override
			public boolean matchesNone(Object2BooleanFunction<Byte2ObjectMap.Entry<V>> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return true;
				BasicEntry<V> subEntry = new BasicEntry<>();
				for(Byte2ObjectAVLTreeMap.Entry<V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					subEntry.set(entry.key, entry.value);
					if(filter.getBoolean(subEntry)) return false;
				}
				return true;
			}
			
			@Override
			public boolean matchesAll(Object2BooleanFunction<Byte2ObjectMap.Entry<V>> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return true;
				BasicEntry<V> subEntry = new BasicEntry<>();
				for(Byte2ObjectAVLTreeMap.Entry<V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					subEntry.set(entry.key, entry.value);
					if(!filter.getBoolean(subEntry)) return false;
				}
				return true;
			}
			
			@Override
			public <E> E reduce(E identity, BiFunction<E, Byte2ObjectMap.Entry<V>, E> operator) {
				Objects.requireNonNull(operator);
				E state = identity;
				for(Byte2ObjectAVLTreeMap.Entry<V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					state = operator.apply(state, new BasicEntry<>(entry.key, entry.value));
				}
				return state;
			}
			
			@Override
			public Byte2ObjectMap.Entry<V> reduce(ObjectObjectUnaryOperator<Byte2ObjectMap.Entry<V>, Byte2ObjectMap.Entry<V>> operator) {
				Objects.requireNonNull(operator);
				Byte2ObjectMap.Entry<V> state = null;
				boolean empty = true;
				for(Byte2ObjectAVLTreeMap.Entry<V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					if(empty) {
						empty = false;
						state = new BasicEntry<>(entry.key, entry.value);
						continue;
					}
					state = operator.apply(state, new BasicEntry<>(entry.key, entry.value));
				}
				return state;
			}
			
			@Override
			public Byte2ObjectMap.Entry<V> findFirst(Object2BooleanFunction<Byte2ObjectMap.Entry<V>> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return null;
				BasicEntry<V> subEntry = new BasicEntry<>();
				for(Byte2ObjectAVLTreeMap.Entry<V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					subEntry.set(entry.key, entry.value);
					if(filter.getBoolean(subEntry)) return subEntry;
				}
				return null;
			}
			
			@Override
			public int count(Object2BooleanFunction<Byte2ObjectMap.Entry<V>> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return 0;
				int result = 0;
				BasicEntry<V> subEntry = new BasicEntry<>();
				for(Byte2ObjectAVLTreeMap.Entry<V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					subEntry.set(entry.key, entry.value);
					if(filter.getBoolean(subEntry)) result++;
				}
				return result;
			}
		}
		
		final class SubMapValues extends AbstractObjectCollection<V> {
			@Override
			public boolean add(V o) { throw new UnsupportedOperationException(); }
			
			@Override
			public boolean contains(Object e) {
				return containsValue(e);
			}
			
			@Override
			public ObjectIterator<V> iterator() { return valueIterator(); }
			
			@Override
			public int size() {
				return NavigableSubMap.this.size();
			}
			
			@Override
			public void clear() {
				NavigableSubMap.this.clear();
			}
			
			@Override
			public void forEach(Consumer<? super V> action) {
				Objects.requireNonNull(action);
				for(Byte2ObjectAVLTreeMap.Entry<V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					action.accept(entry.value);
			}
			
			@Override
			public <E> void forEach(E input, ObjectObjectConsumer<E, V> action) {
				Objects.requireNonNull(action);
				for(Byte2ObjectAVLTreeMap.Entry<V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					action.accept(input, entry.value);
			}
			
			@Override
			public boolean matchesAny(Object2BooleanFunction<V> filter) {
				Objects.requireNonNull(filter);
				for(Byte2ObjectAVLTreeMap.Entry<V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					if(filter.getBoolean(entry.value)) return true;
				return false;
			}
			
			@Override
			public boolean matchesNone(Object2BooleanFunction<V> filter) {
				Objects.requireNonNull(filter);
				for(Byte2ObjectAVLTreeMap.Entry<V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					if(filter.getBoolean(entry.value)) return false;
				return true;
			}
			
			@Override
			public boolean matchesAll(Object2BooleanFunction<V> filter) {
				Objects.requireNonNull(filter);
				for(Byte2ObjectAVLTreeMap.Entry<V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					if(!filter.getBoolean(entry.value)) return false;
				return true;
			}
			
			@Override
			public <E> E reduce(E identity, BiFunction<E, V, E> operator) {
				Objects.requireNonNull(operator);
				E state = identity;
				for(Byte2ObjectAVLTreeMap.Entry<V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					state = operator.apply(state, entry.value);
				return state;
			}
			
			@Override
			public V reduce(ObjectObjectUnaryOperator<V, V> operator) {
				Objects.requireNonNull(operator);
				V state = null;
				boolean empty = true;
				for(Byte2ObjectAVLTreeMap.Entry<V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					if(empty) {
						empty = false;
						state = entry.value;
						continue;
					}
					state = operator.apply(state, entry.value);
				}
				return state;
			}
			
			@Override
			public V findFirst(Object2BooleanFunction<V> filter) {
				Objects.requireNonNull(filter);
				for(Byte2ObjectAVLTreeMap.Entry<V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					if(filter.getBoolean(entry.value)) return entry.value;
				return null;
			}
			
			@Override
			public int count(Object2BooleanFunction<V> filter) {
				Objects.requireNonNull(filter);
				int result = 0;
				for(Byte2ObjectAVLTreeMap.Entry<V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					if(filter.getBoolean(entry.value)) result++;
				return result;
			}
		}
		
		final class SubMapEntrySetIterator extends SubMapEntryIterator implements ObjectListIterator<Byte2ObjectMap.Entry<V>> {
			public SubMapEntrySetIterator(boolean descending) {
				super(descending);
			}
			
			public SubMapEntrySetIterator(Byte2ObjectAVLTreeMap.Entry<V> entry, Byte2ObjectAVLTreeMap.Entry<V> lowerFence, Byte2ObjectAVLTreeMap.Entry<V> upperFence) {
				super(entry, lowerFence, upperFence);
			}
			
			@Override
			public Byte2ObjectMap.Entry<V> next() { return nextEntry(); }
			@Override
			public Byte2ObjectMap.Entry<V> previous() { return previousEntry(); }
			
			@Override
			public void set(Byte2ObjectMap.Entry<V> e) { throw new UnsupportedOperationException(); }
			@Override
			public void add(Byte2ObjectMap.Entry<V> e) { throw new UnsupportedOperationException(); }
		}
		
		final class SubMapKeyIterator extends SubMapEntryIterator implements ByteListIterator {
			public SubMapKeyIterator(boolean descending) {
				super(descending);
			}
			
			public SubMapKeyIterator(Byte2ObjectAVLTreeMap.Entry<V> entry, Byte2ObjectAVLTreeMap.Entry<V> lowerFence, Byte2ObjectAVLTreeMap.Entry<V> upperFence) {
				super(entry, lowerFence, upperFence);
			}
			
			@Override
			public byte nextByte() { return nextEntry().key; }
			@Override
			public byte previousByte() { return previousEntry().key; }
			
			@Override
			public void set(byte e) { throw new UnsupportedOperationException(); }
			@Override
			public void add(byte e) { throw new UnsupportedOperationException(); }
		}
		
		final class SubMapValueIterator extends SubMapEntryIterator implements ObjectListIterator<V> {
			public SubMapValueIterator(boolean descending) {
				super(descending);
			}
			
			public SubMapValueIterator(Byte2ObjectAVLTreeMap.Entry<V> entry, Byte2ObjectAVLTreeMap.Entry<V> lowerFence, Byte2ObjectAVLTreeMap.Entry<V> upperFence) {
				super(entry, lowerFence, upperFence);
			}
			
			@Override
			public V next() { return nextEntry().value; }
			@Override
			public V previous() { return previousEntry().value; }
			
			@Override
			public void set(V e) { throw new UnsupportedOperationException(); }
			@Override
			public void add(V e) { throw new UnsupportedOperationException(); }
		}
		
		abstract class SubMapEntryIterator {
			Byte lowerFence;
			Byte upperFence;
			Byte2ObjectAVLTreeMap.Entry<V> next;
			Byte2ObjectAVLTreeMap.Entry<V> previous;
			Byte2ObjectAVLTreeMap.Entry<V> current;
			int index = 0;
			
			public SubMapEntryIterator(boolean descending) {
				this(descending ? findHighest() : findLowest(), fromStart ? null : findLowest(), toEnd ? null : findHighest());
			}
			
			public SubMapEntryIterator(Byte2ObjectAVLTreeMap.Entry<V> entry, Byte2ObjectAVLTreeMap.Entry<V> lowerFence, Byte2ObjectAVLTreeMap.Entry<V> upperFence) {
				next = entry;
				previous = entry.previous();
				this.lowerFence = lowerFence != null ? Byte.valueOf(lowerFence.key) : null;
				this.upperFence = upperFence != null ? Byte.valueOf(upperFence.key) : null;
			}
			
			public boolean hasNext() {
				return next != null && (upperFence == null || next.key == upperFence.byteValue());
			}
			
			public boolean hasPrevious() {
				return previous != null && (lowerFence == null || next.key == lowerFence.byteValue());
			}
			
			public int nextIndex() {
				return index;
			}
			
			public int previousIndex() {
				return index - 1;
			}
			
			protected void updateNext() {
				next = current.next();
			}
			
			protected void updatePrevious() {
				previous = current.previous();
			}
			
			public Byte2ObjectAVLTreeMap.Entry<V> nextEntry() {
				if(!hasNext()) throw new NoSuchElementException();
				current = previous = next;
				updateNext();
				index++;
				return current;
			}
			
			public Byte2ObjectAVLTreeMap.Entry<V> previousEntry() {
				if(!hasPrevious()) throw new NoSuchElementException();
				current = next = previous;
				updatePrevious();
				index--;
				return current;
			}
			
			public void remove() {
				if(current == null) throw new IllegalStateException();
				if(current == previous) index--;
				updateNext();
				updatePrevious();
				if(current.needsSuccessor()) next = current;
				m.removeNode(current);
				current = null;
			}
		}
	}
	
	private class EntrySet extends AbstractObjectSet<Byte2ObjectMap.Entry<V>> {
		
		@Override
		public ObjectIterator<Byte2ObjectMap.Entry<V>> iterator() {
			return new EntryIterator(false);
		}
		
		@Override
		@Deprecated
		public boolean contains(Object o) {
			if(o instanceof Map.Entry) {
				if(o instanceof Byte2ObjectMap.Entry) {
					Byte2ObjectMap.Entry<V> entry = (Byte2ObjectMap.Entry<V>)o;
					if(entry.getKey() == null && comparator() == null) return false;
					Byte2ObjectAVLTreeMap.Entry<V> subEntry = Byte2ObjectAVLTreeMap.this.findNode(entry.getByteKey());
					if(subEntry != null) return Objects.equals(entry.getValue(), subEntry.value);
				}
				else {
					Map.Entry<?, ?> entry = (Map.Entry<?, ?>)o;
					if(entry.getKey() == null && comparator() == null) return false;
					if(!(entry.getKey() instanceof Byte)) return false;
					Byte2ObjectAVLTreeMap.Entry<V> subEntry = Byte2ObjectAVLTreeMap.this.findNode((Byte)entry.getKey());
					if(subEntry != null) return Objects.equals(entry.getValue(), subEntry.value);
				}
			}
			return false;
		}
		
		@Override
		@Deprecated
		public boolean remove(Object o) {
			if(o instanceof Map.Entry) {
				if(o instanceof Byte2ObjectMap.Entry) {
					Byte2ObjectMap.Entry<V> entry = (Byte2ObjectMap.Entry<V>)o;
					return Byte2ObjectAVLTreeMap.this.remove(entry.getByteKey(), entry.getValue());
				}
				Map.Entry<?, ?> entry = (Map.Entry<?, ?>)o;
				return Byte2ObjectAVLTreeMap.this.remove(entry.getKey(), entry.getValue());
			}
			return false;
		}
		
		@Override
		public int size() {
			return Byte2ObjectAVLTreeMap.this.size();
		}
		
		@Override
		public void clear() {
			Byte2ObjectAVLTreeMap.this.clear();
		}
		
		@Override
		public void forEach(Consumer<? super Byte2ObjectMap.Entry<V>> action) {
			Objects.requireNonNull(action);
			for(Byte2ObjectAVLTreeMap.Entry<V> entry = first;entry != null;entry = entry.next())
				action.accept(new BasicEntry<>(entry.key, entry.value));
		}
		
		@Override
		public <E> void forEach(E input, ObjectObjectConsumer<E, Byte2ObjectMap.Entry<V>> action) {
			Objects.requireNonNull(action);
			for(Byte2ObjectAVLTreeMap.Entry<V> entry = first;entry != null;entry = entry.next())
				action.accept(input, new BasicEntry<>(entry.key, entry.value));
		}
		
		@Override
		public boolean matchesAny(Object2BooleanFunction<Byte2ObjectMap.Entry<V>> filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return false;
			BasicEntry<V> subEntry = new BasicEntry<>();
			for(Byte2ObjectAVLTreeMap.Entry<V> entry = first;entry != null;entry = entry.next()) {
				subEntry.set(entry.key, entry.value);
				if(filter.getBoolean(subEntry)) return true;
			}
			return false;
		}
		
		@Override
		public boolean matchesNone(Object2BooleanFunction<Byte2ObjectMap.Entry<V>> filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return true;
			BasicEntry<V> subEntry = new BasicEntry<>();
			for(Byte2ObjectAVLTreeMap.Entry<V> entry = first;entry != null;entry = entry.next()) {
				subEntry.set(entry.key, entry.value);
				if(filter.getBoolean(subEntry)) return false;
			}
			return true;
		}
		
		@Override
		public boolean matchesAll(Object2BooleanFunction<Byte2ObjectMap.Entry<V>> filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return true;
			BasicEntry<V> subEntry = new BasicEntry<>();
			for(Byte2ObjectAVLTreeMap.Entry<V> entry = first;entry != null;entry = entry.next()) {
				subEntry.set(entry.key, entry.value);
				if(!filter.getBoolean(subEntry)) return false;
			}
			return true;
		}
		
		@Override
		public <E> E reduce(E identity, BiFunction<E, Byte2ObjectMap.Entry<V>, E> operator) {
			Objects.requireNonNull(operator);
			E state = identity;
			for(Byte2ObjectAVLTreeMap.Entry<V> entry = first;entry != null;entry = entry.next()) {
				state = operator.apply(state, new BasicEntry<>(entry.key, entry.value));
			}
			return state;
		}
		
		@Override
		public Byte2ObjectMap.Entry<V> reduce(ObjectObjectUnaryOperator<Byte2ObjectMap.Entry<V>, Byte2ObjectMap.Entry<V>> operator) {
			Objects.requireNonNull(operator);
			Byte2ObjectMap.Entry<V> state = null;
			boolean empty = true;
			for(Byte2ObjectAVLTreeMap.Entry<V> entry = first;entry != null;entry = entry.next()) {
				if(empty) {
					empty = false;
					state = new BasicEntry<>(entry.key, entry.value);
					continue;
				}
				state = operator.apply(state, new BasicEntry<>(entry.key, entry.value));
			}
			return state;
		}
		
		@Override
		public Byte2ObjectMap.Entry<V> findFirst(Object2BooleanFunction<Byte2ObjectMap.Entry<V>> filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return null;
			BasicEntry<V> subEntry = new BasicEntry<>();
			for(Byte2ObjectAVLTreeMap.Entry<V> entry = first;entry != null;entry = entry.next()) {
				subEntry.set(entry.key, entry.value);
				if(filter.getBoolean(subEntry)) return subEntry;
			}
			return null;
		}
	}
	
	static final class KeySet<V> extends AbstractByteSet implements ByteNavigableSet {
		Byte2ObjectNavigableMap<V> m;
		
		KeySet(Byte2ObjectNavigableMap<V> m) {
			this.m = m;
		}
		
		@Override
		public void setDefaultMaxValue(byte e) { m.setDefaultMaxValue(e); }
		@Override
		public byte getDefaultMaxValue() { return m.getDefaultMaxValue(); }
		@Override
		public void setDefaultMinValue(byte e) { m.setDefaultMinValue(e); }
		@Override
		public byte getDefaultMinValue() { return m.getDefaultMinValue(); }
		@Override
		public boolean add(byte o) { throw new UnsupportedOperationException(); }
		@Override
		public boolean addAndMoveToFirst(byte o) { throw new UnsupportedOperationException(); }
		@Override
		public boolean addAndMoveToLast(byte o) { throw new UnsupportedOperationException(); }
		@Override
		public boolean moveToFirst(byte o) { throw new UnsupportedOperationException(); }
		@Override
		public boolean moveToLast(byte o) { throw new UnsupportedOperationException(); }
		@Override
		public ByteComparator comparator() { return m.comparator(); }
		@Override
		public byte lower(byte e) { return m.lowerKey(e); }
		@Override
		public byte floor(byte e) { return m.floorKey(e); }
		@Override
		public byte ceiling(byte e) { return m.ceilingKey(e); }
		@Override
		public byte higher(byte e) { return m.higherKey(e); }
		@Override
		public byte firstByte() { return m.firstByteKey(); }
		@Override
		public byte pollFirstByte() { return m.pollFirstByteKey(); }
		@Override
		public byte lastByte() { return m.lastByteKey(); }
		@Override
		public byte pollLastByte() { return m.pollLastByteKey(); }		
		@Override
		public boolean remove(byte o) { 
			int oldSize = m.size();
			m.remove(o); 
			return oldSize != m.size();
		}
		
		@Override
		public KeySet<V> copy() { throw new UnsupportedOperationException(); }
		
		@Override
		public ByteNavigableSet subSet(byte fromElement, boolean fromInclusive, byte toElement, boolean toInclusive) {
			return new KeySet<>(m.subMap(fromElement, fromInclusive, toElement, toInclusive));
		}
		
		@Override
		public ByteNavigableSet headSet(byte toElement, boolean inclusive) {
			return new KeySet<>(m.headMap(toElement, inclusive));
		}
		
		@Override
		public ByteNavigableSet tailSet(byte fromElement, boolean inclusive) {
			return new KeySet<>(m.tailMap(fromElement, inclusive));
		}
		
		@Override
		public ByteNavigableSet descendingSet() {
			return new KeySet<>(m.descendingMap());
		}
		
		@Override
		public ByteBidirectionalIterator iterator() {
			if(m instanceof Byte2ObjectAVLTreeMap) return ((Byte2ObjectAVLTreeMap<V>)m).keyIterator(false);
			return ((NavigableSubMap<V>)m).keyIterator(false);
		}
		
		@Override
		public ByteBidirectionalIterator iterator(byte fromElement) {
			if(m instanceof Byte2ObjectAVLTreeMap) return ((Byte2ObjectAVLTreeMap<V>)m).keyIterator(fromElement);
			return ((NavigableSubMap<V>)m).keyIterator(fromElement);
		}
		
		@Override
		public ByteBidirectionalIterator descendingIterator() {
			if(m instanceof Byte2ObjectAVLTreeMap) return ((Byte2ObjectAVLTreeMap<V>)m).keyIterator(true);
			return ((NavigableSubMap<V>)m).keyIterator(true);
		}
		
		@Override
		public int size() {
			return m.size();
		}
		
		@Override
		public void clear() {
			m.clear();
		}
		
		protected Byte2ObjectAVLTreeMap.Entry<V> start() {
			if(m instanceof Byte2ObjectAVLTreeMap) return ((Byte2ObjectAVLTreeMap<V>)m).first;
			return ((NavigableSubMap<V>)m).findLowest();
		}
		
		protected Byte2ObjectAVLTreeMap.Entry<V> end() {
			if(m instanceof Byte2ObjectAVLTreeMap) return null;
			return ((NavigableSubMap<V>)m).findLowest();
		}
		
		protected Byte2ObjectAVLTreeMap.Entry<V> next(Byte2ObjectAVLTreeMap.Entry<V> entry) {
			if(m instanceof Byte2ObjectAVLTreeMap) return entry.next();
			return ((NavigableSubMap<V>)m).next(entry);
		}
		
		protected Byte2ObjectAVLTreeMap.Entry<V> previous(Byte2ObjectAVLTreeMap.Entry<V> entry) {
			if(m instanceof Byte2ObjectAVLTreeMap) return entry.previous();
			return ((NavigableSubMap<V>)m).previous(entry);
		}
		
		@Override
		public void forEach(ByteConsumer action) {
			Objects.requireNonNull(action);
			for(Byte2ObjectAVLTreeMap.Entry<V> entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				action.accept(entry.key);
		}
		
		@Override
		public <E> void forEach(E input, ObjectByteConsumer<E> action) {
			Objects.requireNonNull(action);
			for(Byte2ObjectAVLTreeMap.Entry<V> entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				action.accept(input, entry.key);
		}
		
		@Override
		public boolean matchesAny(Byte2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(Byte2ObjectAVLTreeMap.Entry<V> entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				if(filter.get(entry.key)) return true;
			return false;
		}
		
		@Override
		public boolean matchesNone(Byte2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(Byte2ObjectAVLTreeMap.Entry<V> entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				if(filter.get(entry.key)) return false;
			return true;
		}
		
		@Override
		public boolean matchesAll(Byte2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(Byte2ObjectAVLTreeMap.Entry<V> entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				if(!filter.get(entry.key)) return false;
			return true;
		}
		
		@Override
		public byte reduce(byte identity, ByteByteUnaryOperator operator) {
			Objects.requireNonNull(operator);
			byte state = identity;
			for(Byte2ObjectAVLTreeMap.Entry<V> entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				state = operator.applyAsByte(state, entry.key);
			return state;
		}
		
		@Override
		public byte reduce(ByteByteUnaryOperator operator) {
			Objects.requireNonNull(operator);
			byte state = (byte)0;
			boolean empty = true;
			for(Byte2ObjectAVLTreeMap.Entry<V> entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry)) {
				if(empty) {
					empty = false;
					state = entry.key;
					continue;
				}
				state = operator.applyAsByte(state, entry.key);
			}
			return state;
		}
		
		@Override
		public byte findFirst(Byte2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(Byte2ObjectAVLTreeMap.Entry<V> entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				if(filter.get(entry.key)) return entry.key;
			return (byte)0;
		}
		
		@Override
		public int count(Byte2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			int result = 0;
			for(Byte2ObjectAVLTreeMap.Entry<V> entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				if(filter.get(entry.key)) result++;
			return result;
		}
	}
	
	final class Values extends AbstractObjectCollection<V> {

		@Override
		public boolean add(V o) { throw new UnsupportedOperationException(); }
		
		@Override
		public boolean contains(Object e) {
			return containsValue(e);
		}
		
		@Override
		public ObjectIterator<V> iterator() { return new ValueIterator(false); }
		
		@Override
		public int size() {
			return Byte2ObjectAVLTreeMap.this.size();
		}
		
		@Override
		public void clear() {
			Byte2ObjectAVLTreeMap.this.clear();
		}
		
		@Override
		public void forEach(Consumer<? super V> action) {
			Objects.requireNonNull(action);
			for(Byte2ObjectAVLTreeMap.Entry<V> entry = first;entry != null;entry = entry.next())
				action.accept(entry.value);
		}
		
		@Override
		public <E> void forEach(E input, ObjectObjectConsumer<E, V> action) {
			Objects.requireNonNull(action);
			for(Byte2ObjectAVLTreeMap.Entry<V> entry = first;entry != null;entry = entry.next())
				action.accept(input, entry.value);
		}
		
		@Override
		public boolean matchesAny(Object2BooleanFunction<V> filter) {
			Objects.requireNonNull(filter);
			for(Byte2ObjectAVLTreeMap.Entry<V> entry = first;entry != null;entry = entry.next())
				if(filter.getBoolean(entry.value)) return true;
			return false;
		}
		
		@Override
		public boolean matchesNone(Object2BooleanFunction<V> filter) {
			Objects.requireNonNull(filter);
			for(Byte2ObjectAVLTreeMap.Entry<V> entry = first;entry != null;entry = entry.next())
				if(filter.getBoolean(entry.value)) return false;
			return true;
		}
		
		@Override
		public boolean matchesAll(Object2BooleanFunction<V> filter) {
			Objects.requireNonNull(filter);
			for(Byte2ObjectAVLTreeMap.Entry<V> entry = first;entry != null;entry = entry.next())
				if(!filter.getBoolean(entry.value)) return false;
			return true;
		}
		
		@Override
		public <E> E reduce(E identity, BiFunction<E, V, E> operator) {
			Objects.requireNonNull(operator);
			E state = identity;
			for(Byte2ObjectAVLTreeMap.Entry<V> entry = first;entry != null;entry = entry.next())
				state = operator.apply(state, entry.value);
			return state;
		}
		
		@Override
		public V reduce(ObjectObjectUnaryOperator<V, V> operator) {
			Objects.requireNonNull(operator);
			V state = null;
			boolean empty = true;
			for(Byte2ObjectAVLTreeMap.Entry<V> entry = first;entry != null;entry = entry.next()) {
				if(empty) {
					empty = false;
					state = entry.value;
					continue;
				}
				state = operator.apply(state, entry.value);
			}
			return state;
		}
		
		@Override
		public V findFirst(Object2BooleanFunction<V> filter) {
			Objects.requireNonNull(filter);
			for(Byte2ObjectAVLTreeMap.Entry<V> entry = first;entry != null;entry = entry.next())
				if(filter.getBoolean(entry.value)) return entry.value;
			return null;
		}
		
		@Override
		public int count(Object2BooleanFunction<V> filter) {
			Objects.requireNonNull(filter);
			int result = 0;
			for(Byte2ObjectAVLTreeMap.Entry<V> entry = first;entry != null;entry = entry.next())
				if(filter.getBoolean(entry.value)) result++;
			return result;
		}
	}
	
	final class EntryIterator extends MapEntryIterator implements ObjectListIterator<Byte2ObjectMap.Entry<V>> {
		
		public EntryIterator(boolean descending) {
			super(descending);
		}
		
		public EntryIterator(Entry<V> entry) {
			super(entry);
		}
		
		@Override
		public Byte2ObjectMap.Entry<V> next() {
			return nextEntry();
		}
		
		@Override
		public Byte2ObjectMap.Entry<V> previous() {
			return previousEntry();
		}
		
		@Override
		public void set(Byte2ObjectMap.Entry<V> e) { throw new UnsupportedOperationException(); }
		
		@Override
		public void add(Byte2ObjectMap.Entry<V> e) { throw new UnsupportedOperationException(); }
	}
	
	final class KeyIterator extends MapEntryIterator implements ByteListIterator {
		
		public KeyIterator(boolean descending) {
			super(descending);
		}
		
		public KeyIterator(Entry<V> entry) {
			super(entry);
		}
		
		@Override
		public byte previousByte() { return previousEntry().key; }
		@Override
		public byte nextByte() { return nextEntry().key; }
		
		@Override
		public void set(byte e) { throw new UnsupportedOperationException(); }
		@Override
		public void add(byte e) { throw new UnsupportedOperationException(); }
	}
	
	final class ValueIterator extends MapEntryIterator implements ObjectListIterator<V> {
		
		public ValueIterator(boolean descending) {
			super(descending);
		}
		
		public ValueIterator(Entry<V> entry) {
			super(entry);
		}
		
		@Override
		public V previous() {
			return previousEntry().value;
		}
		
		@Override
		public V next() {
			return nextEntry().value;
		}
		
		@Override
		public void set(V e) { throw new UnsupportedOperationException(); }
		
		@Override
		public void add(V e) { throw new UnsupportedOperationException(); }
	}
	
	abstract class MapEntryIterator {
		Entry<V> next;
		Entry<V> previous;
		Entry<V> current;
		int index = 0;
		
		public MapEntryIterator(boolean descending) {
			if(descending) previous = last;
			else next = first;
		}
		
		public MapEntryIterator(Entry<V> entry) {
			next = entry;
			previous = entry.previous();
		}
		
		public boolean hasNext() {
			return next != null;
		}
		
		public boolean hasPrevious() {
			return previous != null;
		}
		
		public int nextIndex() {
			return index;
		}
		
		public int previousIndex() {
			return index - 1;
		}
		
		protected void updateNext() {
			next = current.next();
		}
		
		protected void updatePrevious() {
			previous = current.previous();
		}
		
		public Entry<V> nextEntry() {
			if(!hasNext()) throw new NoSuchElementException();
			current = previous = next;
			updateNext();
			index++;
			return current;
		}
		
		public Entry<V> previousEntry() {
			if(!hasPrevious()) throw new NoSuchElementException();
			current = next = previous;
			updatePrevious();
			index--;
			return current;
		}
		
		public void remove() {
			if(current == null) throw new IllegalStateException();
			if(current == previous) index--;
			updateNext();
			updatePrevious();
			if(current.needsSuccessor()) next = current;
			removeNode(current);
			current = null;
		}
	}
	
	private static final class Entry<V> implements Byte2ObjectMap.Entry<V>
	{
		byte key;
		V value;
		int state;
		Entry<V> parent;
		Entry<V> left;
		Entry<V> right;
		
		Entry(byte key, V value, Entry<V> parent) {
			this.key = key;
			this.value = value;
			this.parent = parent;
		}
		
		Entry<V> copy() {
			Entry<V> entry = new Entry<>(key, value, null);
			entry.state = state;
			if(left != null) {
				Entry<V> newLeft = left.copy();
				entry.left = newLeft;
				newLeft.parent = entry;
			}
			if(right != null) {
				Entry<V> newRight = right.copy();
				entry.right = newRight;
				newRight.parent = entry;
			}
			return entry;
		}
		
		@Override
		public byte getByteKey() {
			return key;
		}
		
		@Override
		public V getValue() {
			return value;
		}
		
		@Override
		public V setValue(V value) {
			V oldValue = this.value;
			this.value = value;
			return oldValue;
		}
		
		@Override
		public boolean equals(Object obj) {
			if(obj instanceof Map.Entry) {
				if(obj instanceof Byte2ObjectMap.Entry) {
					Byte2ObjectMap.Entry<V> entry = (Byte2ObjectMap.Entry<V>)obj;
					return key == entry.getByteKey() && Objects.equals(value, entry.getValue());
				}
				Map.Entry<?, ?> entry = (Map.Entry<?, ?>)obj;
				Object otherKey = entry.getKey();
				if(otherKey == null) return false;
				Object otherValue = entry.getValue();
				return otherKey instanceof Byte && key == ((Byte)otherKey).byteValue() && Objects.equals(value, otherValue);
			}
			return false;
		}
		
		@Override
		public int hashCode() {
			return Byte.hashCode(key) ^ Objects.hashCode(value);
		}
		
		@Override
		public String toString() {
			return Byte.toString(key) + "=" + Objects.toString(value);
		}
		
		int getHeight() { return state; }
		
		void updateHeight() { state = (1 + Math.max(left == null ? -1 : left.getHeight(), right == null ? -1 : right.getHeight())); }
		
		int getBalance() { return (left == null ? -1 : left.getHeight()) - (right == null ? -1 : right.getHeight()); }
		
		boolean needsSuccessor() { return left != null && right != null; }
		
		boolean replace(Entry<V> entry) {
			if(entry != null) entry.parent = parent;
			if(parent != null) {
				if(parent.left == this) parent.left = entry;
				else parent.right = entry;
			}
			return parent == null;
		}
		
		Entry<V> next() {
			if(right != null) {
				Entry<V> parent = right;
				while(parent.left != null) parent = parent.left;
				return parent;
			}
			Entry<V> parent = this.parent;
			Entry<V> control = this;
			while(parent != null && control == parent.right) {
				control = parent;
				parent = parent.parent;
			}
			return parent;
		}
		
		Entry<V> previous() {
			if(left != null) {
				Entry<V> parent = left;
				while(parent.right != null) parent = parent.right;
				return parent;
			}
			Entry<V> parent = this.parent;
			Entry<V> control = this;
			while(parent != null && control == parent.left) {
				control = parent;
				parent = parent.parent;
			}
			return parent;
		}
	}
}