package speiger.src.collections.chars.maps.impl.misc;

import java.util.Arrays;
import java.util.Comparator;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Objects;
import java.util.function.Consumer;
import java.util.function.BiFunction;

import speiger.src.collections.chars.collections.CharBidirectionalIterator;
import speiger.src.collections.chars.functions.CharComparator;
import speiger.src.collections.chars.functions.CharConsumer;
import speiger.src.collections.objects.functions.consumer.ObjectCharConsumer;
import speiger.src.collections.chars.functions.function.Char2BooleanFunction;
import speiger.src.collections.chars.functions.consumer.CharIntConsumer;
import speiger.src.collections.chars.functions.function.Char2IntFunction;
import speiger.src.collections.chars.functions.function.CharIntUnaryOperator;
import speiger.src.collections.chars.functions.function.CharCharUnaryOperator;
import speiger.src.collections.chars.lists.CharListIterator;
import speiger.src.collections.chars.maps.abstracts.AbstractChar2IntMap;
import speiger.src.collections.chars.maps.interfaces.Char2IntMap;
import speiger.src.collections.chars.maps.interfaces.Char2IntSortedMap;
import speiger.src.collections.chars.sets.AbstractCharSet;
import speiger.src.collections.chars.sets.CharSortedSet;
import speiger.src.collections.chars.sets.CharSet;
import speiger.src.collections.chars.utils.maps.Char2IntMaps;
import speiger.src.collections.ints.collections.AbstractIntCollection;
import speiger.src.collections.ints.collections.IntCollection;
import speiger.src.collections.ints.collections.IntIterator;
import speiger.src.collections.ints.functions.IntSupplier;
import speiger.src.collections.ints.functions.function.IntIntUnaryOperator;
import speiger.src.collections.ints.functions.IntConsumer;
import speiger.src.collections.ints.lists.IntListIterator;
import speiger.src.collections.objects.functions.consumer.ObjectObjectConsumer;
import speiger.src.collections.objects.functions.function.Object2BooleanFunction;
import speiger.src.collections.objects.functions.consumer.ObjectIntConsumer;
import speiger.src.collections.ints.functions.function.Int2BooleanFunction;
import speiger.src.collections.objects.collections.ObjectIterator;
import speiger.src.collections.objects.functions.function.ObjectObjectUnaryOperator;
import speiger.src.collections.objects.collections.ObjectBidirectionalIterator;
import speiger.src.collections.objects.lists.ObjectListIterator;
import speiger.src.collections.objects.sets.AbstractObjectSet;
import speiger.src.collections.objects.sets.ObjectSortedSet;
import speiger.src.collections.objects.sets.ObjectSet;
import speiger.src.collections.utils.HashUtil;

/**
 * A Very Specific Type Specific implementation of a ArrayMap.
 * This type of map is for very specific use cases that usaully would have lead to Tupled Lists otherwise.
 * It also does not allow duplication (except for array constructors) and checks from last to first.
 * It is not designed to be used as a HashMap replacement due to the poor performance it would cause.
 * @note in this implementation SubMaps do NOT keep track of parent changes fully. For performance reasons it will just have a start/end index and not values
 * Anything within that range will be updated appropiatly a shrink/growth of elements will break SubMaps in some ways. This can be useful but be careful
 * @note this implementation does not shrink and only grows.
 */
public class Char2IntArrayMap extends AbstractChar2IntMap implements Char2IntSortedMap
{
	/** The Backing keys array */
	protected transient char[] keys;
	/** The Backing values array */
	protected transient int[] values;
	/** Amount of Elements stored in the ArrayMap */
	protected int size = 0;
	/** KeySet cache */
	protected CharSet keySet;
	/** Values cache */
	protected IntCollection valuesC;
	/** EntrySet cache */
	protected FastSortedSet entrySet;
	
	/**
	 * Default Constructor
	 */
	public Char2IntArrayMap() {
		this(HashUtil.DEFAULT_MIN_CAPACITY);
	}
	
	/**
	 * Constructor that defines the minimum capacity
	 * @param minCapacity the minimum capacity the HashMap is allowed to be.
	 * @throws IllegalStateException if the minimum capacity is negative
	 */
	public Char2IntArrayMap(int minCapacity) {
		if(minCapacity < 0)	throw new IllegalStateException("Minimum Capacity is negative. This is not allowed");
		keys = new char[minCapacity];
		values = new int[minCapacity];
	}
	
	/**
	 * Helper constructor that allow to create a map from boxed values (it will unbox them)
	 * @param keys the keys that should be put into the map
	 * @param values the values that should be put into the map.
	 * @throws IllegalStateException if the keys and values do not match in length
	 */
	public Char2IntArrayMap(Character[] keys, Integer[] values) {
		this(keys, values, keys.length);
	}
	
	/**
	 * Helper constructor that allow to create a map from boxed values (it will unbox them) with a custom length
	 * @param keys the keys that should be put into the map
	 * @param values the values that should be put into the map.
	 * @param length the amount of values that should be pulled from the array
	 * @throws IllegalStateException if the keys and values do not match in length
	 */
	public Char2IntArrayMap(Character[] keys, Integer[] values, int length) {
		this(length);
		if(keys.length != values.length) throw new IllegalStateException("Input Arrays are not equal size");
		putAll(keys, values, 0, length);
	}
	
	/**
	 * Helper constructor that allow to create a map from unboxed values
	 * @param keys the keys that should be put into the map
	 * @param values the values that should be put into the map.
	 * @throws IllegalStateException if the keys and values do not match in lenght
	 */
	public Char2IntArrayMap(char[] keys, int[] values) {
		this(keys, values, keys.length);
	}
	
	/**
	 * Helper constructor that allow to create a map from unboxed values
	 * @param keys the keys that should be put into the map
	 * @param values the values that should be put into the map.
	 * @param length the amount of values that should be pulled from the array
	 * @throws IllegalStateException if the keys and values do not match in lenght
	 */
	public Char2IntArrayMap(char[] keys, int[] values, int length) {
		this(length);
		if(keys.length != values.length) throw new IllegalStateException("Input Arrays are not equal size");
		putAll(keys, values, 0, length);
	}
	
	/**
	 * A Helper constructor that allows to create a Map with exactly the same values as the provided map.
	 * @param map the values that should be present in the map
	 */
	public Char2IntArrayMap(Map<? extends Character, ? extends Integer> map) {
		this(map.size());
		putAll(map);
	}
	
	/**
	 * A Type Specific Helper function that allows to create a new Map with exactly the same values as the provided map.
	 * @param map the values that should be present in the map
 	 */
	public Char2IntArrayMap(Char2IntMap map) {
		this(map.size());
		for(ObjectIterator<Char2IntMap.Entry> iter = Char2IntMaps.fastIterator(map);iter.hasNext();size++) {
			Char2IntMap.Entry entry = iter.next();
			keys[size] = entry.getCharKey();
			values[size] = entry.getIntValue();
		}
	}
	
	@Override
	public int put(char key, int value) {
		int index = findIndex(key);
		if(index < 0) {
			insertIndex(size++, key, value);
			return getDefaultReturnValue();
		}
		int oldValue = values[index];
		values[index] = value;
		return oldValue;
	}
	
	@Override
	public int putIfAbsent(char key, int value) {
		int index = findIndex(key);
		if(index < 0) {
			insertIndex(size++, key, value);
			return getDefaultReturnValue();
		}
		return values[index];
	}
	
	@Override
	public int addTo(char key, int value) {
		int index = findIndex(key);
		if(index < 0) {
			insertIndex(size++, key, value);
			return getDefaultReturnValue();
		}
		int oldValue = values[index];
		values[index] += value;
		return oldValue;
	}
	
	@Override
	public int putAndMoveToFirst(char key, int value) {
		int index = findIndex(key);
		if(index < 0) {
			insertIndex(0, key, value);
			size++;
			return getDefaultReturnValue();
		}
		int lastValue = values[index];
		values[index] = value;
		moveIndexToFirst(index);
		return lastValue;
	}
	
	@Override
	public int putAndMoveToLast(char key, int value) {
		int index = findIndex(key);
		if(index < 0) {
			insertIndex(size++, key, value);
			return getDefaultReturnValue();
		}
		int lastValue = values[index];
		values[index] = value;
		moveIndexToLast(index);
		return lastValue;
	}
	
	@Override
	public boolean moveToFirst(char key) {
		int index = findIndex(key);
		if(index > 0) {
			moveIndexToFirst(index);
			return true;
		}
		return false;
	}
	
	@Override
	public boolean moveToLast(char key) {
		int index = findIndex(key);
		if(index < size-1) {
			moveIndexToLast(index);
			return true;
		}
		return false;
	}
	
	@Override
	public boolean containsKey(char key) {
		return findIndex(key) >= 0;
	}
	
	@Override
	public boolean containsValue(int value) {
		return findValue(value) >= 0;
	}
	
	@Override
	public boolean containsKey(Object key) {
		return findIndex(key) >= 0;
	}
	
	@Override
	public boolean containsValue(Object value) {
		return findValue(value) >= 0;
	}
	
	@Override
	public int get(char key) {
		int index = findIndex(key);
		return index < 0 ? getDefaultReturnValue() : values[index];
	}
	
	@Override
	public int getOrDefault(char key, int defaultValue) {
		int index = findIndex(key);
		return index < 0 ? defaultValue : values[index];
	}
	
	@Override
	public int getAndMoveToFirst(char key) {
		int index = findIndex(key);
		if(index >= 0) {
			int value = values[index];
			moveIndexToFirst(index);
			return value;
		}
		return getDefaultReturnValue();
	}
	
	@Override
	public int getAndMoveToLast(char key) {
		int index = findIndex(key);
		if(index >= 0) {
			int value = values[index];
			moveIndexToLast(index);
			return value;
		}
		return getDefaultReturnValue();
	}
	
	@Override
	public char firstCharKey() {
		if(size <= 0) throw new NoSuchElementException();
		return keys[0];
	}
	
	@Override
	public char lastCharKey() {
		if(size <= 0) throw new NoSuchElementException();
		return keys[size-1];
	}
	
	@Override
	public int firstIntValue() {
		if(size <= 0) throw new NoSuchElementException();
		return values[0];
	}
	
	@Override
	public int lastIntValue() {
		if(size <= 0) throw new NoSuchElementException();
		return values[size-1];
	}
	
	@Override
	public char pollFirstCharKey() {
		if(size == 0) throw new NoSuchElementException();
		char result = keys[0];
		removeIndex(0);
		return result;
	}
	
	@Override
	public char pollLastCharKey() {
		if(size == 0) throw new NoSuchElementException();
		char result = keys[size-1];
		removeIndex(size-1);
		return result;
	}
	
	@Override
	public int remove(char key) {
		int index = findIndex(key);
		if(index < 0) return getDefaultReturnValue();
		int value = values[index];
		removeIndex(index);
		return value;
	}
	
	@Override
	public int removeOrDefault(char key, int defaultValue) {
		int index = findIndex(key);
		if(index < 0) return defaultValue;
		int value = values[index];
		removeIndex(index);
		return value;
	}
	
	@Override
	public boolean remove(char key, int value) {
		int index = findIndex(key, value);
		if(index < 0) return false;
		removeIndex(index);
		return true;
	}
	
	@Override
	public Integer remove(Object key) {
		int index = findIndex(key);
		if(index < 0) return Integer.valueOf(getDefaultReturnValue());
		int value = values[index];
		removeIndex(index);
		return Integer.valueOf(value);
	}
	
	@Override
	public boolean remove(Object key, Object value) {
		int index = findIndex(key, value);
		if(index < 0) return false;
		removeIndex(index);
		return true;
	}
	
	@Override
	public void forEach(CharIntConsumer action) {
		if(size() <= 0) return;
		for(int i = 0;i<size;i++)
			action.accept(keys[i], values[i]);
	}
	
	@Override
	public CharSet keySet() {
		if(keySet == null) keySet = new KeySet();
		return keySet;
	}

	@Override
	public IntCollection values() {
		if(valuesC == null) valuesC = new Values();
		return valuesC;
	}

	@Override
	public ObjectSortedSet<Char2IntMap.Entry> char2IntEntrySet() {
		if(entrySet == null) entrySet = new MapEntrySet();
		return entrySet;
	}
	
	@Override
	public boolean replace(char key, int oldValue, int newValue) {
		int index = findIndex(key);
		if(index < 0 || values[index] != oldValue) return false;
		values[index] = newValue;
		return true;
	}
	
	@Override
	public int replace(char key, int value) {
		int index = findIndex(key);
		if(index < 0) return getDefaultReturnValue();
		int oldValue = values[index];
		values[index] = value;
		return oldValue;
	}
	
	@Override
	public int computeInt(char key, CharIntUnaryOperator mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		int index = findIndex(key);
		if(index == -1) {
			int newValue = mappingFunction.applyAsInt(key, getDefaultReturnValue());
			if(newValue == getDefaultReturnValue()) return newValue;
			insertIndex(size++, key, newValue);
			return newValue;
		}
		int newValue = mappingFunction.applyAsInt(key, values[index]);
		if(newValue == getDefaultReturnValue()) {
			removeIndex(index);
			return newValue;
		}
		values[index] = newValue;
		return newValue;
	}
	
	@Override
	public int computeIntIfAbsent(char key, Char2IntFunction mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		int index = findIndex(key);
		if(index == -1) {
			int newValue = mappingFunction.get(key);
			if(newValue == getDefaultReturnValue()) return newValue;
			insertIndex(size++, key, newValue);
			return newValue;
		}
		int newValue = values[index];
		if(newValue == getDefaultReturnValue()) {
			newValue = mappingFunction.get(key);
			if(newValue == getDefaultReturnValue()) return newValue;
			values[index] = newValue;
		}
		return newValue;
	}
	
	@Override
	public int supplyIntIfAbsent(char key, IntSupplier valueProvider) {
		Objects.requireNonNull(valueProvider);
		int index = findIndex(key);
		if(index == -1) {
			int newValue = valueProvider.getInt();
			if(newValue == getDefaultReturnValue()) return newValue;
			insertIndex(size++, key, newValue);
			return newValue;
		}
		int newValue = values[index];
		if(newValue == getDefaultReturnValue()) {
			newValue = valueProvider.getInt();
			if(newValue == getDefaultReturnValue()) return newValue;
			values[index] = newValue;
		}
		return newValue;
	}
	
	@Override
	public int computeIntIfPresent(char key, CharIntUnaryOperator mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		int index = findIndex(key);
		if(index == -1 || values[index] == getDefaultReturnValue()) return getDefaultReturnValue();
		int newValue = mappingFunction.applyAsInt(key, values[index]);
		if(newValue == getDefaultReturnValue()) {
			removeIndex(index);
			return newValue;
		}
		values[index] = newValue;
		return newValue;
	}
	
	@Override
	public int mergeInt(char key, int value, IntIntUnaryOperator mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		int index = findIndex(key);
		int newValue = index == -1 || values[index] == getDefaultReturnValue() ? value : mappingFunction.applyAsInt(values[index], value);
		if(newValue == getDefaultReturnValue()) {
			if(index >= 0)
				removeIndex(index);
		}
		else if(index == -1) insertIndex(size++, key, newValue);
		else values[index] = newValue;
		return newValue;
	}
	
	@Override
	public void mergeAllInt(Char2IntMap m, IntIntUnaryOperator mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		for(Char2IntMap.Entry entry : Char2IntMaps.fastIterable(m)) {
			char key = entry.getCharKey();
			int index = findIndex(key);
			int newValue = index == -1 || values[index] == getDefaultReturnValue() ? entry.getIntValue() : mappingFunction.applyAsInt(values[index], entry.getIntValue());
			if(newValue == getDefaultReturnValue()) {
				if(index >= 0)
					removeIndex(index);
			}
			else if(index == -1) insertIndex(size++, key, newValue);
			else values[index] = newValue;
		}
	}
	
	@Override
	public int size() {
		return size;
	}
	
	@Override
	public void clear() {
		Arrays.fill(keys, 0, size, (char)0);
		Arrays.fill(values, 0, size, 0);
		size = 0;
	}
	
	public Char2IntArrayMap copy() {
		Char2IntArrayMap map = new Char2IntArrayMap();
		map.size = size;
		map.keys = Arrays.copyOf(keys, keys.length);
		map.values = Arrays.copyOf(values, keys.length);
		return map;
	}
	
	@Override
	public CharComparator comparator() {
		return null;
	}
	
	@Override
	public Char2IntSortedMap subMap(char fromKey, char toKey) {
		int fromIndex = findIndex(fromKey);
		int toIndex = findIndex(toKey);
		if(fromIndex == -1 || toIndex == -1) throw new NoSuchElementException();
		return new SubMap(fromIndex, toIndex - fromIndex + 1);
	}
	
	@Override
	public Char2IntSortedMap headMap(char toKey) {
		int toIndex = findIndex(toKey);
		if(toIndex == -1) throw new NoSuchElementException();
		return new SubMap(0, toIndex + 1);
	}
	
	@Override
	public Char2IntSortedMap tailMap(char fromKey) {
		int fromIndex = findIndex(fromKey);
		if(fromIndex == -1) throw new NoSuchElementException();
		return new SubMap(fromIndex, size - fromIndex);
	}
	
	protected void moveIndexToFirst(int index) {
		if(index == 0) return;
		char key = keys[index];
		int value = values[index];
		System.arraycopy(keys, 0, keys, 1, index);
		System.arraycopy(values, 0, values, 1, index);
		keys[0] = key;
		values[0] = value;
	}
	
	protected void moveIndexToLast(int index) {
		if(index == size-1) return;
		char key = keys[index];
		int value = values[index];
		System.arraycopy(keys, index+1, keys, index, size-index-1);
		System.arraycopy(values, index+1, values, index, size-index-1);
		keys[size-1] = key;
		values[size-1] = value;
	}
	
	protected void grow(int newSize) {
		if(newSize < keys.length) return;
		newSize = Math.min(newSize, keys.length == 0 ? 2 : keys.length * 2);
		keys = Arrays.copyOf(keys, newSize);
		values = Arrays.copyOf(values, newSize);
	}
	
	protected void insertIndex(int index, char key, int value) {
		grow(size+1);
		if(index != size) {
			System.arraycopy(keys, index, keys, index+1, size-index);
			System.arraycopy(values, index, values, index+1, size-index);		
		}
		keys[index] = key;
		values[index] = value;
	}
	
	protected void removeRange(int from, int to) {
		if(from < 0 || from >= size) throw new IllegalStateException("From Element ");
		int length = to - from;
		if(length <= 0) return;
		if(to != size) {
			System.arraycopy(keys, to, keys, from, size - to);
			System.arraycopy(values, to, values, from, size - to);
		}
		for(int i = 0;i<length;i++) {
			keys[i+to] = (char)0;
			values[i+to] = 0;
		}
		size -= length;
	}
	
	protected void removeIndex(int index) {
		if(index == size-1) {
			size--;
			keys[size] = (char)0;
			values[size] = 0;
			return;
		}
		System.arraycopy(keys, index+1, keys, index, size-index-1);
		System.arraycopy(values, index+1, values, index, size-index-1);		
		size--;
		keys[size] = (char)0;
		values[size] = 0;
	}
	
	protected int findIndex(char key, int value) {
		for(int i = size-1;i>=0;i--)
			if(keys[i] == key && values[i] == value) return i;
		return -1;		
	}
	
	protected int findIndex(char key) {
		for(int i = size-1;i>=0;i--)
			if(keys[i] == key) return i;
		return -1;
	}
	
	protected int findValue(int value) {
		for(int i = size-1;i>=0;i--)
			if(values[i] == value) return i;
		return -1;
	}
	
	protected int findIndex(Object key, Object value) {
		for(int i = size-1;i>=0;i--)
			if(Objects.equals(key, Character.valueOf(keys[i])) && Objects.equals(value, Integer.valueOf(values[i]))) return i;
		return -1;		
	}
	
	protected int findIndex(Object key) {
		for(int i = size-1;i>=0;i--)
			if(Objects.equals(key, Character.valueOf(keys[i]))) return i;
		return -1;
	}
	
	protected int findValue(Object value) {
		for(int i = size-1;i>=0;i--)
			if(Objects.equals(value, Integer.valueOf(values[i]))) return i;
		return -1;
	}
	
	private class SubMap extends AbstractChar2IntMap implements Char2IntSortedMap {
		int offset;
		int length;
		
		CharSet keySet;
		IntCollection valuesC;
		FastSortedSet entrySet;
		
		public SubMap(int offset, int length) {
			this.offset = offset;
			this.length = length;
		}
		
		int end() { return offset+length; }
		
		@Override
		public int put(char key, int value) {
			int index = findIndex(key);
			if(index < 0) {
				insertIndex(end(), key, value);
				length++;
				return getDefaultReturnValue();
			}
			int oldValue = values[index];
			values[index] = value;
			return oldValue;
		}
		
		@Override
		public int putIfAbsent(char key, int value) {
			int index = findIndex(key);
			if(index >= 0) {
				insertIndex(end(), key, value);
				length++;
			}
			return getDefaultReturnValue();
		}
		
		@Override
		public int addTo(char key, int value) {
			int index = findIndex(key);
			if(index < 0) {
				insertIndex(end(), key, value);
				length++;
				return getDefaultReturnValue();
			}
			int oldValue = values[index];
			values[index] += value;
			return oldValue;
		}
		
		@Override
		public int putAndMoveToFirst(char key, int value) {
			int index = findIndex(key);
			if(index < 0) {
				insertIndex(offset, key, value);
				length++;
				return getDefaultReturnValue();
			}
			int lastValue = values[index];
			values[index] = value;
			moveIndexToFirst(index);
			return lastValue;
		}
		
		@Override
		public int putAndMoveToLast(char key, int value) {
			int index = findIndex(key);
			if(index < 0) {
				insertIndex(end(), key, value);
				length++;
				return getDefaultReturnValue();
			}
			int lastValue = values[index];
			values[index] = value;
			moveIndexToLast(index);
			return lastValue;
		}
		
		@Override
		public boolean moveToFirst(char key) {
			int index = findIndex(key);
			if(index > 0) {
				moveIndexToFirst(index);
				return true;
			}
			return false;
		}
		
		@Override
		public boolean moveToLast(char key) {
			int index = findIndex(key);
			if(index < size()-1) {
				moveIndexToLast(index);
				return true;
			}
			return false;
		}
		
		@Override
		public boolean containsKey(char key) {
			return findIndex(key) >= 0;
		}
		
		@Override
		public boolean containsValue(int value) {
			return findValue(value) >= 0;
		}
		
		@Override
		public boolean containsKey(Object key) {
			return findIndex(key) >= 0;
		}
		
		@Override
		public boolean containsValue(Object value) {
			return findValue(value) >= 0;
		}
		
		@Override
		public int get(char key) {
			int index = findIndex(key);
			return index < 0 ? getDefaultReturnValue() : values[index];
		}
		
		@Override
		public int getOrDefault(char key, int defaultValue) {
			int index = findIndex(key);
			return index < 0 ? defaultValue : values[index];
		}
		
		@Override
		public int getAndMoveToFirst(char key) {
			int index = findIndex(key);
			if(index >= 0) {
				int value = values[index];
				moveIndexToFirst(index);
				return value;
			}
			return getDefaultReturnValue();
		}
		
		@Override
		public int getAndMoveToLast(char key) {
			int index = findIndex(key);
			if(index >= 0) {
				int value = values[index];
				moveIndexToLast(index);
				return value;
			}
			return getDefaultReturnValue();
		}
		
		@Override
		public char firstCharKey() {
			if(length <= 0) throw new NoSuchElementException();
			return keys[offset];
		}
		
		@Override
		public char lastCharKey() {
			if(length <= 0) throw new NoSuchElementException();
			return keys[end()-1];
		}
		
		@Override
		public int firstIntValue() {
			if(length <= 0) throw new NoSuchElementException();
			return values[end()-1];
		}
		
		@Override
		public int lastIntValue() {
			if(length <= 0) throw new NoSuchElementException();
			return values[offset];
		}
		
		@Override
		public char pollFirstCharKey() {
			if(length == 0) throw new NoSuchElementException();
			char result = keys[offset];
			removeIndex(offset);
			length--;
			return result;
		}
		
		@Override
		public char pollLastCharKey() {
			if(length == 0) throw new NoSuchElementException();
			char result = keys[end()-1];
			removeIndex(end()-1);
			length--;
			return result;
		}
		
		@Override
		public int remove(char key) {
			int index = findIndex(key);
			if(index < 0) return getDefaultReturnValue();
			int value = values[index];
			removeIndex(index);
			length--;
			return value;
		}
		
		@Override
		public int removeOrDefault(char key, int defaultValue) {
			int index = findIndex(key);
			if(index < 0) return defaultValue;
			int value = values[index];
			removeIndex(index);
			length--;
			return value;
		}
		
		@Override
		public boolean remove(char key, int value) {
			int index = findIndex(key, value);
			if(index < 0) return false;
			removeIndex(index);
			length--;
			return true;
		}
		
		@Override
		public Integer remove(Object key) {
			int index = findIndex(key);
			if(index < 0) return Integer.valueOf(getDefaultReturnValue());
			int value = values[index];
			removeIndex(index);
			length--;
			return Integer.valueOf(value);
		}
		
		@Override
		public boolean remove(Object key, Object value) {
			int index = findIndex(key, value);
			if(index < 0) return false;
			removeIndex(index);
			length--;
			return true;
		}
		
		@Override
		public int size() {
			return length;
		}
		
		@Override
		public void clear() {
			removeRange(offset, offset+length);
			length = 0;
		}
		
		@Override
		public CharSet keySet() {
			if(keySet == null) keySet = new SubKeySet();
			return keySet;
		}
		
		@Override
		public IntCollection values() {
			if(valuesC == null) valuesC = new SubValues();
			return valuesC;
		}

		@Override
		public ObjectSet<Char2IntMap.Entry> char2IntEntrySet() {
			if(entrySet == null) entrySet = new SubMapEntrySet();
			return entrySet;
		}
		
		@Override
		public SubMap copy() { 
			throw new UnsupportedOperationException(); 
		}
		
		@Override
		public CharComparator comparator() {
			return null;
		}
		
		@Override
		public Char2IntSortedMap subMap(char fromKey, char toKey) {
			int fromIndex = findIndex(fromKey);
			int toIndex = findIndex(toKey);
			if(fromIndex == -1 || toIndex == -1) throw new NoSuchElementException();
			return new SubMap(fromIndex, toIndex - fromIndex + 1);
		}
		
		@Override
		public Char2IntSortedMap headMap(char toKey) {
			int toIndex = findIndex(toKey);
			if(toIndex == -1) throw new NoSuchElementException();
			return new SubMap(offset, toIndex + 1);
		}
		
		@Override
		public Char2IntSortedMap tailMap(char fromKey) {
			int fromIndex = findIndex(fromKey);
			if(fromIndex == -1) throw new NoSuchElementException();
			return new SubMap(fromIndex, size - fromIndex);
		}
		
		protected void moveIndexToFirst(int index) {
			if(index == 0) return;
			char key = keys[index];
			int value = values[index];
			System.arraycopy(keys, offset, keys, offset+1, index);
			System.arraycopy(values, offset, values, offset+1, index);
			keys[offset] = key;
			values[offset] = value;
		}
		
		protected void moveIndexToLast(int index) {
			if(index == length-1) return;
			char key = keys[index];
			int value = values[index];
			System.arraycopy(keys, index+1, keys, index, end()-index-1);
			System.arraycopy(values, index+1, values, index, end()-index-1);
			keys[end()-1] = key;
			values[end()-1] = value;
		}
		
		protected int findIndex(char key, int value) {
			for(int i = length-1;i>=0;i--)
				if(keys[offset+i] == key && values[offset+i] == value) return i;
			return -1;		
		}
		
		protected int findIndex(char key) {
			for(int i = length-1;i>=0;i--)
				if(keys[offset+i] == key) return i+offset;
			return -1;
		}
		
		protected int findValue(int value) {
			for(int i = length-1;i>=0;i--)
				if(values[offset+i] == value) return i+offset;
			return -1;
		}
		
		protected int findIndex(Object key, Object value) {
			for(int i = length-1;i>=0;i--)
				if(Objects.equals(key, Character.valueOf(keys[offset+i])) && Objects.equals(value, Integer.valueOf(values[offset+i]))) return i+offset;
			return -1;		
		}
		
		protected int findIndex(Object key) {
			for(int i = length-1;i>=0;i--)
				if(Objects.equals(key, Character.valueOf(keys[offset+i]))) return i+offset;
			return -1;
		}
		
		protected int findValue(Object value) {
			for(int i = length-1;i>=0;i--)
				if(Objects.equals(value, Integer.valueOf(values[offset+i]))) return i+offset;
			return -1;
		}
		
		private class SubMapEntrySet extends AbstractObjectSet<Char2IntMap.Entry> implements Char2IntSortedMap.FastSortedSet {
			@Override
			public boolean addAndMoveToFirst(Char2IntMap.Entry o) { throw new UnsupportedOperationException(); }
			@Override
			public boolean addAndMoveToLast(Char2IntMap.Entry o) { throw new UnsupportedOperationException(); }
			
			@Override
			public boolean moveToFirst(Char2IntMap.Entry o) {
				return SubMap.this.moveToFirst(o.getCharKey());
			}
			
			@Override
			public boolean moveToLast(Char2IntMap.Entry o) {
				return SubMap.this.moveToLast(o.getCharKey());
			}
			
			@Override
			public Char2IntMap.Entry first() {
				return new BasicEntry(firstCharKey(), firstIntValue());
			}
			
			@Override
			public Char2IntMap.Entry last() {
				return new BasicEntry(lastCharKey(), lastIntValue());
			}
			
			@Override
			public Char2IntMap.Entry pollFirst() {
				BasicEntry entry = new BasicEntry(firstCharKey(), firstIntValue());
				pollFirstCharKey();
				return entry;
			}
			
			@Override
			public Char2IntMap.Entry pollLast() {
				BasicEntry entry = new BasicEntry(lastCharKey(), lastIntValue());
				pollLastCharKey();
				return entry;
			}
			
			@Override
			public ObjectBidirectionalIterator<Char2IntMap.Entry> iterator() {
				return new SubEntryIterator();
			}
			
			@Override
			public ObjectBidirectionalIterator<Char2IntMap.Entry> iterator(Char2IntMap.Entry fromElement) {
				return new SubEntryIterator(fromElement.getCharKey());
			}
			
			@Override
			public ObjectBidirectionalIterator<Char2IntMap.Entry> fastIterator() {
				return new SubFastEntryIterator();
			}
			
			@Override
			public ObjectBidirectionalIterator<Char2IntMap.Entry> fastIterator(char fromElement) {
				return new SubFastEntryIterator(fromElement);
			}
			
			@Override
			public SubMapEntrySet copy() { throw new UnsupportedOperationException(); }
			
			@Override
			public void forEach(Consumer<? super Char2IntMap.Entry> action) {
				Objects.requireNonNull(action);
				for(int i = 0;i<length;i++)
					action.accept(new BasicEntry(keys[offset+i], values[offset+i]));
			}
			
			@Override
			public void fastForEach(Consumer<? super Char2IntMap.Entry> action) {
				Objects.requireNonNull(action);
				if(size() <= 0) return;
				BasicEntry entry = new BasicEntry();
				for(int i = 0;i<length;i++) {
					entry.set(keys[offset+i], values[offset+i]);
					action.accept(entry);
				}
			}
			
			@Override
			public <E> void forEach(E input, ObjectObjectConsumer<E, Char2IntMap.Entry> action) {
				Objects.requireNonNull(action);
				for(int i = 0;i<length;i++)
					action.accept(input, new BasicEntry(keys[offset+i], values[offset+i]));
			}
			
			@Override
			public boolean matchesAny(Object2BooleanFunction<Char2IntMap.Entry> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return false;
				BasicEntry entry = new BasicEntry();
				for(int i = 0;i<length;i++) {
					entry.set(keys[offset+i], values[offset+i]);
					if(filter.getBoolean(entry)) return true;
				}
				return false;
			}
			
			@Override
			public boolean matchesNone(Object2BooleanFunction<Char2IntMap.Entry> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return true;
				BasicEntry entry = new BasicEntry();
				for(int i = 0;i<length;i++) {
					entry.set(keys[offset+i], values[offset+i]);
					if(filter.getBoolean(entry)) return false;
				}
				return true;
			}
			
			@Override
			public boolean matchesAll(Object2BooleanFunction<Char2IntMap.Entry> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return true;
				BasicEntry entry = new BasicEntry();
				for(int i = 0;i<length;i++) {
					entry.set(keys[offset+i], values[offset+i]);
					if(!filter.getBoolean(entry)) return false;
				}
				return true;
			}
			
			@Override
			public <E> E reduce(E identity, BiFunction<E, Char2IntMap.Entry, E> operator) {
				Objects.requireNonNull(operator);
				E state = identity;
				for(int i = 0;i<length;i++) {
					state = operator.apply(state, new BasicEntry(keys[offset+i], values[offset+i]));
				}
				return state;
			}
			
			@Override
			public Char2IntMap.Entry reduce(ObjectObjectUnaryOperator<Char2IntMap.Entry, Char2IntMap.Entry> operator) {
				Objects.requireNonNull(operator);
				Char2IntMap.Entry state = null;
				boolean empty = true;
				for(int i = 0;i<length;i++) {
					if(empty) {
						empty = false;
						state = new BasicEntry(keys[offset+i], values[offset+i]);
						continue;
					}
					state = operator.apply(state, new BasicEntry(keys[offset+i], values[offset+i]));
				}
				return state;
			}
			
			@Override
			public Char2IntMap.Entry findFirst(Object2BooleanFunction<Char2IntMap.Entry> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return null;
				BasicEntry entry = new BasicEntry();
				for(int i = 0;i<length;i++) {
					entry.set(keys[offset+i], values[offset+i]);
					if(filter.getBoolean(entry)) return entry;
				}
				return null;
			}
			
			@Override
			public int count(Object2BooleanFunction<Char2IntMap.Entry> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return 0;
				int result = 0;
				BasicEntry entry = new BasicEntry();
				for(int i = 0;i<length;i++) {
					entry.set(keys[offset+i], values[offset+i]);
					if(filter.getBoolean(entry)) result++;
				}
				return result;
			}
			
			@Override
			@Deprecated
			public boolean contains(Object o) {
				if(o instanceof Map.Entry) {
					if(o instanceof Char2IntMap.Entry) {
						Char2IntMap.Entry entry = (Char2IntMap.Entry)o;
						return SubMap.this.findIndex(entry.getCharKey(), entry.getIntValue()) >= 0;
					}
					else {
						Map.Entry<?, ?> entry = (Map.Entry<?, ?>)o;
						return Char2IntArrayMap.this.findIndex(entry.getKey(), entry.getValue()) >= 0;
					}
				}
				return false;
			}
			
			@Override
			@Deprecated
			public boolean remove(Object o) {
				if(o instanceof Map.Entry) {
					if(o instanceof Char2IntMap.Entry) {
						Char2IntMap.Entry entry = (Char2IntMap.Entry)o;
						return SubMap.this.remove(entry.getCharKey(), entry.getIntValue());
					}
					Map.Entry<?, ?> entry = (Map.Entry<?, ?>)o;
					return SubMap.this.remove(entry.getKey(), entry.getValue());
				}
				return false;
			}
			
			@Override
			public int size() {
				return SubMap.this.size();
			}
			
			@Override
			public void clear() {
				SubMap.this.clear();
			}
			
			@Override
			public Comparator<Char2IntMap.Entry> comparator() {
				return null;
			}
			@Override
			public ObjectSortedSet<Char2IntMap.Entry> subSet(Char2IntMap.Entry fromElement, Char2IntMap.Entry toElement) { throw new UnsupportedOperationException(); }
			@Override
			public ObjectSortedSet<Char2IntMap.Entry> headSet(Char2IntMap.Entry toElement) { throw new UnsupportedOperationException(); }
			@Override
			public ObjectSortedSet<Char2IntMap.Entry> tailSet(Char2IntMap.Entry fromElement) { throw new UnsupportedOperationException(); }
		}
		
		private class SubKeySet extends AbstractCharSet implements CharSortedSet {
			@Override
			public boolean contains(char e) { return containsKey(e); }
			
			@Override
			public boolean remove(char o) {
				int oldSize = length;
				SubMap.this.remove(o);
				return length != oldSize;
			}
			
			@Override
			public boolean add(char o) { throw new UnsupportedOperationException(); }
			@Override
			public boolean addAndMoveToFirst(char o) { throw new UnsupportedOperationException(); }
			@Override
			public boolean addAndMoveToLast(char o) { throw new UnsupportedOperationException(); }
			@Override
			public boolean moveToFirst(char o) { return SubMap.this.moveToFirst(o); }
			@Override
			public boolean moveToLast(char o) { return SubMap.this.moveToLast(o); }
			@Override
			public CharListIterator iterator() { return new SubKeyIterator(); }
			@Override
			public CharBidirectionalIterator iterator(char fromElement) { return new SubKeyIterator(fromElement); } 
			@Override
			public int size() { return SubMap.this.size(); }
			@Override
			public void clear() { SubMap.this.clear(); }
			@Override
			public char firstChar() { return firstCharKey(); }
			@Override
			public char pollFirstChar() { return pollFirstCharKey(); }
			@Override
			public char lastChar() { return lastCharKey(); }
			@Override
			public char pollLastChar() { return pollLastCharKey(); }
			
			@Override
			public SubKeySet copy() { throw new UnsupportedOperationException(); }
			
			@Override
			public void forEach(CharConsumer action) { 
				Objects.requireNonNull(action);
				for(int i = 0;i<length;action.accept(keys[offset+i])); 
			}
			
			@Override
			public <E> void forEach(E input, ObjectCharConsumer<E> action) {
				Objects.requireNonNull(action);
				if(size() <= 0) return;
				for(int i = 0;i<length;action.accept(input, keys[offset+i])); 

			}
			
			@Override
			public boolean matchesAny(Char2BooleanFunction filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return false;
				for(int i = 0;i<length;i++) {
					if(filter.get(keys[i+offset])) return true;					
				}
				return false;
			}
			
			@Override
			public boolean matchesNone(Char2BooleanFunction filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return true;
				for(int i = 0;i<length;i++) {
					if(filter.get(keys[i+offset])) return false;					
				}
				return true;
			}
			
			@Override
			public boolean matchesAll(Char2BooleanFunction filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return true;
				for(int i = 0;i<length;i++) {
					if(!filter.get(keys[i+offset])) return false;					
				}
				return true;
			}
			
			@Override
			public char reduce(char identity, CharCharUnaryOperator operator) {
				Objects.requireNonNull(operator);
				char state = identity;
				for(int i = 0;i<length;i++) {
					state = operator.applyAsChar(state, keys[i+offset]);
				}
				return state;
			}
			
			@Override
			public char reduce(CharCharUnaryOperator operator) {
				Objects.requireNonNull(operator);
				char state = (char)0;
				boolean empty = true;
				for(int i = 0;i<length;i++) {
					if(empty) {
						empty = false;
						state = keys[i+offset];
						continue;
					}
					state = operator.applyAsChar(state, keys[i+offset]);
				}
				return state;
			}
			
			@Override
			public char findFirst(Char2BooleanFunction filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return (char)0;
				for(int i = 0;i<length;i++) {
					if(filter.get(keys[i+offset])) return keys[i+offset];					
				}
				return (char)0;
			}
			
			@Override
			public int count(Char2BooleanFunction filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return 0;
				int result = 0;
				for(int i = 0;i<length;i++) {
					if(filter.get(keys[i+offset])) result++;					
				}
				return result;
			}
			
			@Override
			public CharComparator comparator() { return null; }
			@Override
			public CharSortedSet subSet(char fromElement, char toElement) { throw new UnsupportedOperationException(); }
			@Override
			public CharSortedSet headSet(char toElement) { throw new UnsupportedOperationException(); }
			@Override
			public CharSortedSet tailSet(char fromElement) { throw new UnsupportedOperationException(); }
		}
		
		private class SubValues extends AbstractIntCollection {
			@Override
			public boolean contains(int e) {
				return containsValue(e);
			}
			
			@Override
			public boolean add(int o) {
				throw new UnsupportedOperationException();
			}
			
			@Override
			public IntIterator iterator() {
				return new SubValueIterator();
			}
			
			@Override
			public int size() {
				return SubMap.this.size();
			}
			
			@Override
			public void clear() {
				SubMap.this.clear();
			}
			
			@Override
			public void forEach(IntConsumer action) {
				Objects.requireNonNull(action);
				for(int i = 0;i<length;action.accept(values[offset+i]));
			}
			
			@Override
			public <E> void forEach(E input, ObjectIntConsumer<E> action) {
				Objects.requireNonNull(action);
				for(int i = 0;i<length;action.accept(input, values[offset+i]));

			}
			
			@Override
			public boolean matchesAny(Int2BooleanFunction filter) {
				Objects.requireNonNull(filter);
				for(int i = 0;i<length;i++) {
					if(filter.get(values[offset+i])) return true;
				}
				return false;
			}
			
			@Override
			public boolean matchesNone(Int2BooleanFunction filter) {
				Objects.requireNonNull(filter);
				for(int i = 0;i<length;i++) {
					if(filter.get(values[offset+i])) return false;
				}
				return true;
			}
			
			@Override
			public boolean matchesAll(Int2BooleanFunction filter) {
				Objects.requireNonNull(filter);
				for(int i = 0;i<length;i++) {
					if(!filter.get(values[offset+i])) return false;
				}
				return true;
			}
			
			@Override
			public int reduce(int identity, IntIntUnaryOperator operator) {
				Objects.requireNonNull(operator);
				int state = identity;
				for(int i = 0;i<length;i++) {
					state = operator.applyAsInt(state, values[offset+i]);
				}
				return state;
			}
			
			@Override
			public int reduce(IntIntUnaryOperator operator) {
				Objects.requireNonNull(operator);
				int state = 0;
				boolean empty = true;
				for(int i = 0;i<length;i++) {
					if(empty) {
						empty = false;
						state = values[offset+i];
						continue;
					}
					state = operator.applyAsInt(state, values[offset+i]);
				}
				return state;
			}
			
			@Override
			public int findFirst(Int2BooleanFunction filter) {
				Objects.requireNonNull(filter);
				for(int i = 0;i<length;i++) {
					if(filter.get(values[offset+i])) return values[offset+i];
				}
				return 0;
			}
			
			@Override
			public int count(Int2BooleanFunction filter) {
				Objects.requireNonNull(filter);
				int result = 0;
				for(int i = 0;i<length;i++) {
					if(filter.get(values[offset+i])) result++;
				}
				return result;
			}
		}
		
		private class SubFastEntryIterator extends SubMapIterator implements ObjectListIterator<Char2IntMap.Entry> {
			MapEntry entry = new MapEntry();
			
			public SubFastEntryIterator() {}
			public SubFastEntryIterator(char from) {
				index = findIndex(from);
			}
			
			@Override
			public Char2IntMap.Entry next() {
				entry.index = nextEntry();
				return entry;
			}
			
			@Override
			public Char2IntMap.Entry previous() {
				entry.index = previousEntry();
				return entry;
			}
			
			@Override
			public void set(Char2IntMap.Entry e) { throw new UnsupportedOperationException(); }
			@Override
			public void add(Char2IntMap.Entry e) { throw new UnsupportedOperationException(); }
		}
		
		private class SubEntryIterator extends SubMapIterator implements ObjectListIterator<Char2IntMap.Entry> {
			MapEntry entry = null;
			
			public SubEntryIterator() {}
			public SubEntryIterator(char from) {
				index = findIndex(from);
			}
			
			@Override
			public Char2IntMap.Entry next() {
				return entry = new MapEntry(nextEntry());
			}
			
			@Override
			public Char2IntMap.Entry previous() {
				return entry = new MapEntry(previousEntry());
			}
			
			@Override
			public void remove() {
				super.remove();
				entry.index = -1;
			}
			
			@Override
			public void set(Char2IntMap.Entry e) { throw new UnsupportedOperationException(); }
			@Override
			public void add(Char2IntMap.Entry e) { throw new UnsupportedOperationException(); }
		}
		
		private class SubKeyIterator extends SubMapIterator implements CharListIterator {
			public SubKeyIterator() {}
			public SubKeyIterator(char element) {
				index = findIndex(element);
			}
			@Override
			public char previousChar() {
				return keys[previousEntry()];
			}

			@Override
			public char nextChar() {
				return keys[nextEntry()];
			}

			@Override
			public void set(char e) { throw new UnsupportedOperationException(); }

			@Override
			public void add(char e) { throw new UnsupportedOperationException(); }	
		}
		
		private class SubValueIterator extends SubMapIterator implements IntListIterator {
			@Override
			public int previousInt() {
				return values[previousEntry()];
			}

			@Override
			public int nextInt() {
				return values[nextEntry()];
			}

			@Override
			public void set(int e) { throw new UnsupportedOperationException(); }

			@Override
			public void add(int e) { throw new UnsupportedOperationException(); }	
		}
		
		private class SubMapIterator {
			int index;
			int lastReturned = -1;

			public boolean hasNext() {
				return index < length;
			}
			
			public boolean hasPrevious() {
				return index > 0;
			}
			
			public int nextIndex() {
				return index;
			}
			
			public int previousIndex() {
				return index-1;
			}
			
			public void remove() {
				if(lastReturned == -1)
					throw new IllegalStateException();
				removeIndex(lastReturned);
				if(lastReturned < index)
					index--;
				lastReturned = -1;
			}
			
			public int previousEntry() {
				if(!hasPrevious()) throw new NoSuchElementException();
				int returnIndex = offset+index;
				lastReturned = index--;
				return returnIndex;
			}
			
			public int nextEntry() {
				if(!hasNext()) throw new NoSuchElementException();
				int returnIndex = offset+index;
				lastReturned = index++;
				return returnIndex;
			}
			
			public int skip(int amount) {
				if(amount < 0) throw new IllegalStateException("Negative Numbers are not allowed");
				int steps = Math.min(amount, (size() - 1) - index);
				index += steps;
				return steps;
			}
			
			public int back(int amount) {
				if(amount < 0) throw new IllegalStateException("Negative Numbers are not allowed");
				int steps = Math.min(amount, index);
				index -= steps;
				return steps;
			}
		}
	}
	
	private class MapEntrySet extends AbstractObjectSet<Char2IntMap.Entry> implements Char2IntSortedMap.FastSortedSet {
		@Override
		public boolean addAndMoveToFirst(Char2IntMap.Entry o) { throw new UnsupportedOperationException(); }
		@Override
		public boolean addAndMoveToLast(Char2IntMap.Entry o) { throw new UnsupportedOperationException(); }
		
		@Override
		public boolean moveToFirst(Char2IntMap.Entry o) {
			return Char2IntArrayMap.this.moveToFirst(o.getCharKey());
		}
		
		@Override
		public boolean moveToLast(Char2IntMap.Entry o) {
			return Char2IntArrayMap.this.moveToLast(o.getCharKey());
		}
		
		@Override
		public Char2IntMap.Entry first() {
			return new BasicEntry(firstCharKey(), firstIntValue());
		}
		
		@Override
		public Char2IntMap.Entry last() {
			return new BasicEntry(lastCharKey(), lastIntValue());
		}
		
		@Override
		public Char2IntMap.Entry pollFirst() {
			BasicEntry entry = new BasicEntry(firstCharKey(), firstIntValue());
			pollFirstCharKey();
			return entry;
		}
		
		@Override
		public Char2IntMap.Entry pollLast() {
			BasicEntry entry = new BasicEntry(lastCharKey(), lastIntValue());
			pollLastCharKey();
			return entry;
		}
		
		@Override
		public ObjectBidirectionalIterator<Char2IntMap.Entry> iterator() {
			return new EntryIterator();
		}
		
		@Override
		public ObjectBidirectionalIterator<Char2IntMap.Entry> iterator(Char2IntMap.Entry fromElement) {
			return new EntryIterator(fromElement.getCharKey());
		}
		
		@Override
		public ObjectBidirectionalIterator<Char2IntMap.Entry> fastIterator() {
			return new FastEntryIterator();
		}
		
		@Override
		public ObjectBidirectionalIterator<Char2IntMap.Entry> fastIterator(char fromElement) {
			return new FastEntryIterator(fromElement);
		}
		
		@Override
		public MapEntrySet copy() { throw new UnsupportedOperationException(); }
		
		@Override
		public void forEach(Consumer<? super Char2IntMap.Entry> action) {
			Objects.requireNonNull(action);
			for(int i = 0;i<size;i++)
				action.accept(new BasicEntry(keys[i], values[i]));
		}
		
		@Override
		public void fastForEach(Consumer<? super Char2IntMap.Entry> action) {
			Objects.requireNonNull(action);
			if(size() <= 0) return;
			BasicEntry entry = new BasicEntry();
			for(int i = 0;i<size;i++) {
				entry.set(keys[i], values[i]);
				action.accept(entry);
			}
		}
		
		@Override
		public <E> void forEach(E input, ObjectObjectConsumer<E, Char2IntMap.Entry> action) {
			Objects.requireNonNull(action);
			for(int i = 0;i<size;i++) {
				action.accept(input, new BasicEntry(keys[i], values[i]));
			}
		}
		
		@Override
		public boolean matchesAny(Object2BooleanFunction<Char2IntMap.Entry> filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return false;
			BasicEntry entry = new BasicEntry();
			for(int i = 0;i<size;i++) {
				entry.set(keys[i], values[i]);
				if(filter.getBoolean(entry)) return true;
			}
			return false;
		}
		
		@Override
		public boolean matchesNone(Object2BooleanFunction<Char2IntMap.Entry> filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return true;
			BasicEntry entry = new BasicEntry();
			for(int i = 0;i<size;i++) {
				entry.set(keys[i], values[i]);
				if(filter.getBoolean(entry)) return false;
			}
			return true;
		}
		
		@Override
		public boolean matchesAll(Object2BooleanFunction<Char2IntMap.Entry> filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return true;
			BasicEntry entry = new BasicEntry();
			for(int i = 0;i<size;i++) {
				entry.set(keys[i], values[i]);
				if(!filter.getBoolean(entry)) return false;
			}
			return true;
		}
		
		@Override
		public <E> E reduce(E identity, BiFunction<E, Char2IntMap.Entry, E> operator) {
			Objects.requireNonNull(operator);
			E state = identity;
			for(int i = 0;i<size;i++) {
				state = operator.apply(state, new BasicEntry(keys[i], values[i]));
			}
			return state;
		}
		
		@Override
		public Char2IntMap.Entry reduce(ObjectObjectUnaryOperator<Char2IntMap.Entry, Char2IntMap.Entry> operator) {
			Objects.requireNonNull(operator);
			Char2IntMap.Entry state = null;
			boolean empty = true;
			for(int i = 0;i<size;i++) {
				if(empty) {
					empty = false;
					state = new BasicEntry(keys[i], values[i]);
					continue;
				}
				state = operator.apply(state, new BasicEntry(keys[i], values[i]));
			}
			return state;
		}
		
		@Override
		public Char2IntMap.Entry findFirst(Object2BooleanFunction<Char2IntMap.Entry> filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return null;
			BasicEntry entry = new BasicEntry();
			for(int i = 0;i<size;i++) {
				entry.set(keys[i], values[i]);
				if(filter.getBoolean(entry)) return entry;
			}
			return null;
		}
		
		@Override
		public int count(Object2BooleanFunction<Char2IntMap.Entry> filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return 0;
			int result = 0;
			BasicEntry entry = new BasicEntry();
			for(int i = 0;i<size;i++) {
				entry.set(keys[i], values[i]);
				if(filter.getBoolean(entry)) result++;
			}
			return result;
		}
		
		@Override
		@Deprecated
		public boolean contains(Object o) {
			if(o instanceof Map.Entry) {
				if(o instanceof Char2IntMap.Entry) {
					Char2IntMap.Entry entry = (Char2IntMap.Entry)o;
					int index = Char2IntArrayMap.this.findIndex(entry.getCharKey());
					if(index >= 0) return entry.getIntValue() == Char2IntArrayMap.this.values[index];
				}
				else {
					Map.Entry<?, ?> entry = (Map.Entry<?, ?>)o;
					int index = Char2IntArrayMap.this.findIndex(entry.getKey());
					if(index >= 0) return Objects.equals(entry.getValue(), Integer.valueOf(Char2IntArrayMap.this.values[index]));
				}
			}
			return false;
		}
		
		@Override
		@Deprecated
		public boolean remove(Object o) {
			if(o instanceof Map.Entry) {
				if(o instanceof Char2IntMap.Entry) {
					Char2IntMap.Entry entry = (Char2IntMap.Entry)o;
					return Char2IntArrayMap.this.remove(entry.getCharKey(), entry.getIntValue());
				}
				Map.Entry<?, ?> entry = (Map.Entry<?, ?>)o;
				return Char2IntArrayMap.this.remove(entry.getKey(), entry.getValue());
			}
			return false;
		}
		
		@Override
		public int size() {
			return Char2IntArrayMap.this.size();
		}
		
		@Override
		public void clear() {
			Char2IntArrayMap.this.clear();
		}
		
		@Override
		public Comparator<Char2IntMap.Entry> comparator() {
			return null;
		}
		@Override
		public ObjectSortedSet<Char2IntMap.Entry> subSet(Char2IntMap.Entry fromElement, Char2IntMap.Entry toElement) { throw new UnsupportedOperationException(); }
		@Override
		public ObjectSortedSet<Char2IntMap.Entry> headSet(Char2IntMap.Entry toElement) { throw new UnsupportedOperationException(); }
		@Override
		public ObjectSortedSet<Char2IntMap.Entry> tailSet(Char2IntMap.Entry fromElement) { throw new UnsupportedOperationException(); }
	}
	
	private class KeySet extends AbstractCharSet implements CharSortedSet {
		@Override
		public boolean contains(char e) { return containsKey(e); }
		
		@Override
		public boolean remove(char o) {
			int oldSize = size;
			Char2IntArrayMap.this.remove(o);
			return size != oldSize;
		}
		
		@Override
		public boolean add(char o) { throw new UnsupportedOperationException(); }
		@Override
		public boolean addAndMoveToFirst(char o) { throw new UnsupportedOperationException(); }
		@Override
		public boolean addAndMoveToLast(char o) { throw new UnsupportedOperationException(); }
		@Override
		public boolean moveToFirst(char o) { return Char2IntArrayMap.this.moveToFirst(o); }
		@Override
		public boolean moveToLast(char o) { return Char2IntArrayMap.this.moveToLast(o); }
		@Override
		public CharListIterator iterator() { return new KeyIterator(); }
		@Override
		public CharBidirectionalIterator iterator(char fromElement) { return new KeyIterator(fromElement); } 
		@Override
		public int size() { return Char2IntArrayMap.this.size(); }
		@Override
		public void clear() { Char2IntArrayMap.this.clear(); }
		@Override
		public char firstChar() { return firstCharKey(); }
		@Override
		public char pollFirstChar() { return pollFirstCharKey(); }
		@Override
		public char lastChar() { return lastCharKey(); }
		@Override
		public char pollLastChar() { return pollLastCharKey(); }
		
		@Override
		public KeySet copy() { throw new UnsupportedOperationException(); }
		
		@Override
		public void forEach(CharConsumer action) { 
			Objects.requireNonNull(action);
			for(int i = 0;i<size;action.accept(keys[i++])); 
		}
		
		@Override
		public <E> void forEach(E input, ObjectCharConsumer<E> action) {
			Objects.requireNonNull(action);
			for(int i = 0;i<size;action.accept(input, keys[i++])); 

		}
		
		@Override
		public boolean matchesAny(Char2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return false;
			for(int i = 0;i<size;i++) {
				if(filter.get(keys[i])) return true;
			}
			return false;
		}
		
		@Override
		public boolean matchesNone(Char2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(int i = 0;i<size;i++) {
				if(filter.get(keys[i])) return false;
			}
			return true;
		}
		
		@Override
		public boolean matchesAll(Char2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(int i = 0;i<size;i++) {
				if(!filter.get(keys[i])) return false;
			}
			return true;
		}
		
		@Override
		public char reduce(char identity, CharCharUnaryOperator operator) {
			Objects.requireNonNull(operator);
			char state = identity;
			for(int i = 0;i<size;i++) {
				state = operator.applyAsChar(state, keys[i]);
			}
			return state;
		}
		
		@Override
		public char reduce(CharCharUnaryOperator operator) {
			Objects.requireNonNull(operator);
			char state = (char)0;
			boolean empty = true;
			for(int i = 0;i<size;i++) {
				if(empty) {
					empty = false;
					state = keys[i];
					continue;
				}
				state = operator.applyAsChar(state, keys[i]);
			}
			return state;
		}
		
		@Override
		public char findFirst(Char2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(int i = 0;i<size;i++) {
				if(filter.get(keys[i])) return keys[i];
			}
			return (char)0;
		}
		
		@Override
		public int count(Char2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			int result = 0;
			for(int i = 0;i<size;i++) {
				if(filter.get(keys[i])) result++;
			}
			return result;
		}
		
		@Override
		public CharComparator comparator() { return null; }
		@Override
		public CharSortedSet subSet(char fromElement, char toElement) { throw new UnsupportedOperationException(); }
		@Override
		public CharSortedSet headSet(char toElement) { throw new UnsupportedOperationException(); }
		@Override
		public CharSortedSet tailSet(char fromElement) { throw new UnsupportedOperationException(); }
	}
	
	private class Values extends AbstractIntCollection {
		@Override
		public boolean contains(int e) {
			return containsValue(e);
		}
		
		@Override
		public boolean add(int o) {
			throw new UnsupportedOperationException();
		}

		@Override
		public IntIterator iterator() {
			return new ValueIterator();
		}
		
		@Override
		public int size() {
			return Char2IntArrayMap.this.size();
		}
		
		@Override
		public void clear() {
			Char2IntArrayMap.this.clear();
		}
		
		@Override
		public void forEach(IntConsumer action) {
			Objects.requireNonNull(action);
			for(int i = 0;i<size;action.accept(values[i++]));
		}
		
		@Override
		public boolean matchesAny(Int2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(int i = 0;i<size;i++) {
				if(filter.get(values[i])) return true;
			}
			return false;
		}
		
		@Override
		public boolean matchesNone(Int2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(int i = 0;i<size;i++) {
				if(filter.get(values[i])) return false;
			}
			return true;
		}
		
		@Override
		public boolean matchesAll(Int2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(int i = 0;i<size;i++) {
				if(!filter.get(values[i])) return false;
			}
			return true;
		}
		
		@Override
		public int reduce(int identity, IntIntUnaryOperator operator) {
			Objects.requireNonNull(operator);
			int state = identity;
			for(int i = 0;i<size;i++) {
				state = operator.applyAsInt(state, values[i]);
			}
			return state;
		}
		
		@Override
		public int reduce(IntIntUnaryOperator operator) {
			Objects.requireNonNull(operator);
			int state = 0;
			boolean empty = true;
			for(int i = 0;i<size;i++) {
				if(empty) {
					empty = false;
					state = values[i];
					continue;
				}
				state = operator.applyAsInt(state, values[i]);
			}
			return state;
		}
		
		@Override
		public int findFirst(Int2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(int i = 0;i<size;i++) {
				if(filter.get(values[i])) return values[i];
			}
			return 0;
		}
		
		@Override
		public int count(Int2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			int result = 0;
			for(int i = 0;i<size;i++) {
				if(filter.get(values[i])) result++;
			}
			return result;
		}
	}
	
	private class FastEntryIterator extends MapIterator implements ObjectListIterator<Char2IntMap.Entry> {
		MapEntry entry = new MapEntry();
		
		public FastEntryIterator() {}
		public FastEntryIterator(char from) {
			index = findIndex(from);
		}
		
		@Override
		public Char2IntMap.Entry next() {
			entry.index = nextEntry();
			return entry;
		}
		
		@Override
		public Char2IntMap.Entry previous() {
			entry.index = previousEntry();
			return entry;
		}
		
		@Override
		public void set(Char2IntMap.Entry e) { throw new UnsupportedOperationException(); }
		@Override
		public void add(Char2IntMap.Entry e) { throw new UnsupportedOperationException(); }
	}
	
	private class EntryIterator extends MapIterator implements ObjectListIterator<Char2IntMap.Entry> {
		MapEntry entry = null;
		
		public EntryIterator() {}
		public EntryIterator(char from) {
			index = findIndex(from);
		}
		
		@Override
		public Char2IntMap.Entry next() {
			return entry = new MapEntry(nextEntry());
		}
		
		@Override
		public Char2IntMap.Entry previous() {
			return entry = new MapEntry(previousEntry());
		}
		
		@Override
		public void remove() {
			super.remove();
			entry.index = -1;
		}
		
		@Override
		public void set(Char2IntMap.Entry e) { throw new UnsupportedOperationException(); }
		@Override
		public void add(Char2IntMap.Entry e) { throw new UnsupportedOperationException(); }
	}
	
	private class KeyIterator extends MapIterator implements CharListIterator {
		public KeyIterator() {}
		public KeyIterator(char element) {
			index = findIndex(element);
		}
		@Override
		public char previousChar() {
			return keys[previousEntry()];
		}

		@Override
		public char nextChar() {
			return keys[nextEntry()];
		}

		@Override
		public void set(char e) { throw new UnsupportedOperationException(); }

		@Override
		public void add(char e) { throw new UnsupportedOperationException(); }	
	}
	
	private class ValueIterator extends MapIterator implements IntListIterator {
		@Override
		public int previousInt() {
			return values[previousEntry()];
		}

		@Override
		public int nextInt() {
			return values[nextEntry()];
		}

		@Override
		public void set(int e) { throw new UnsupportedOperationException(); }

		@Override
		public void add(int e) { throw new UnsupportedOperationException(); }	
	}
	
	private class MapIterator {
		int index;
		int lastReturned = -1;

		public boolean hasNext() {
			return index < size;
		}
		
		public boolean hasPrevious() {
			return index > 0;
		}
		
		public int nextIndex() {
			return index;
		}
		
		public int previousIndex() {
			return index-1;
		}
		
		public void remove() {
			if(lastReturned == -1)
				throw new IllegalStateException();
			removeIndex(lastReturned);
			if(lastReturned < index)
				index--;
			lastReturned = -1;
		}
		
		public int previousEntry() {
			if(!hasPrevious()) throw new NoSuchElementException();
			lastReturned = index;
			return index--;
		}
		
		public int nextEntry() {
			if(!hasNext()) throw new NoSuchElementException();
			lastReturned = index;
			return index++;
		}
		
		public int skip(int amount) {
			if(amount < 0) throw new IllegalStateException("Negative Numbers are not allowed");
			int steps = Math.min(amount, (size() - 1) - index);
			index += steps;
			return steps;
		}
		
		public int back(int amount) {
			if(amount < 0) throw new IllegalStateException("Negative Numbers are not allowed");
			int steps = Math.min(amount, index);
			index -= steps;
			return steps;
		}
	}
	
	private class MapEntry implements Char2IntMap.Entry, Map.Entry<Character, Integer> {
		int index = -1;
		
		public MapEntry() {}
		public MapEntry(int index) {
			this.index = index;
		}

		@Override
		public char getCharKey() {
			return keys[index];
		}

		@Override
		public int getIntValue() {
			return values[index];
		}

		@Override
		public int setValue(int value) {
			int oldValue = values[index];
			values[index] = value;
			return oldValue;
		}
		
		@Override
		public boolean equals(Object obj) {
			if(obj instanceof Map.Entry) {
				if(obj instanceof Char2IntMap.Entry) {
					Char2IntMap.Entry entry = (Char2IntMap.Entry)obj;
					return keys[index] == entry.getCharKey() && values[index] == entry.getIntValue();
				}
				Map.Entry<?, ?> entry = (Map.Entry<?, ?>)obj;
				Object key = entry.getKey();
				Object value = entry.getValue();
				return key instanceof Character && value instanceof Integer && keys[index] == ((Character)key).charValue() && values[index] == ((Integer)value).intValue();
			}
			return false;
		}
		
		@Override
		public int hashCode() {
			return Character.hashCode(keys[index]) ^ Integer.hashCode(values[index]);
		}
		
		@Override
		public String toString() {
			return Character.toString(keys[index]) + "=" + Integer.toString(values[index]);
		}
	}
}