package speiger.src.collections.doubles.maps.impl.tree;

import java.util.Map;
import java.util.Objects;
import java.util.NoSuchElementException;
import java.util.function.Consumer;
import java.util.function.BiFunction;

import speiger.src.collections.doubles.collections.DoubleBidirectionalIterator;
import speiger.src.collections.doubles.functions.DoubleComparator;
import speiger.src.collections.doubles.functions.DoubleConsumer;
import speiger.src.collections.objects.functions.consumer.ObjectDoubleConsumer;
import speiger.src.collections.doubles.functions.function.Double2BooleanFunction;
import speiger.src.collections.doubles.functions.consumer.DoubleCharConsumer;
import speiger.src.collections.doubles.functions.function.Double2CharFunction;
import speiger.src.collections.doubles.functions.function.DoubleCharUnaryOperator;
import speiger.src.collections.doubles.functions.function.DoubleDoubleUnaryOperator;
import speiger.src.collections.doubles.lists.DoubleListIterator;
import speiger.src.collections.doubles.maps.abstracts.AbstractDouble2CharMap;
import speiger.src.collections.doubles.maps.interfaces.Double2CharMap;
import speiger.src.collections.doubles.maps.interfaces.Double2CharNavigableMap;
import speiger.src.collections.doubles.sets.AbstractDoubleSet;
import speiger.src.collections.doubles.sets.DoubleNavigableSet;
import speiger.src.collections.doubles.sets.DoubleSet;
import speiger.src.collections.doubles.sets.DoubleSortedSet;
import speiger.src.collections.doubles.utils.DoubleIterators;
import speiger.src.collections.doubles.utils.maps.Double2CharMaps;
import speiger.src.collections.chars.collections.AbstractCharCollection;
import speiger.src.collections.chars.collections.CharCollection;
import speiger.src.collections.chars.collections.CharIterator;
import speiger.src.collections.chars.functions.CharSupplier;
import speiger.src.collections.chars.functions.function.CharCharUnaryOperator;
import speiger.src.collections.chars.lists.CharListIterator;
import speiger.src.collections.chars.utils.CharIterators;
import speiger.src.collections.chars.functions.CharConsumer;
import speiger.src.collections.objects.functions.consumer.ObjectObjectConsumer;
import speiger.src.collections.objects.functions.function.Object2BooleanFunction;
import speiger.src.collections.objects.functions.consumer.ObjectCharConsumer;
import speiger.src.collections.chars.functions.function.Char2BooleanFunction;
import speiger.src.collections.objects.lists.ObjectListIterator;
import speiger.src.collections.objects.utils.ObjectIterators;
import speiger.src.collections.objects.collections.ObjectIterator;
import speiger.src.collections.objects.functions.function.ObjectObjectUnaryOperator;
import speiger.src.collections.objects.sets.AbstractObjectSet;
import speiger.src.collections.objects.sets.ObjectSet;

/**
 * A Simple Type Specific RB TreeMap implementation that reduces boxing/unboxing.
 * It is using a bit more memory then <a href="https://github.com/vigna/fastutil">FastUtil</a>,
 * but it saves a lot of Performance on the Optimized removal and iteration logic.
 * Which makes the implementation actually useable and does not get outperformed by Javas default implementation.
 */
public class Double2CharRBTreeMap extends AbstractDouble2CharMap implements Double2CharNavigableMap
{
	/** The center of the Tree */
	protected transient Entry tree;
	/** The Lowest possible Node */
	protected transient Entry first;
	/** The Highest possible Node */
	protected transient Entry last;
	/** The amount of elements stored in the Map */
	protected int size = 0;
	/** The Sorter of the Tree */
	protected transient DoubleComparator comparator;
	
	/** the default return value for max searches */
	protected double defaultMaxNotFound = Double.MIN_VALUE;
	/** the default return value for min searches */
	protected double defaultMinNotFound = Double.MAX_VALUE;
	
	/** KeySet Cache */
	protected DoubleNavigableSet keySet;
	/** Values Cache */
	protected CharCollection values;
	/** EntrySet Cache */
	protected ObjectSet<Double2CharMap.Entry> entrySet;
	
	/**
	 * Default Constructor
	 */
	public Double2CharRBTreeMap() {
	}
	
	/**
	 * Constructor that allows to define the sorter
	 * @param comp the function that decides how the tree is sorted, can be null
	 */
	public Double2CharRBTreeMap(DoubleComparator comp) {
		comparator = comp;
	}
	
	/**
	 * Helper constructor that allow to create a map from boxed values (it will unbox them)
	 * @param keys the keys that should be put into the map
	 * @param values the values that should be put into the map.
	 * @throws IllegalStateException if the keys and values do not match in lenght
	 */
	public Double2CharRBTreeMap(Double[] keys, Character[] values) {
		this(keys, values, null);
	}
	
	/**
	 * Helper constructor that has a custom sorter and allow to create a map from boxed values (it will unbox them)
	 * @param keys the keys that should be put into the map
	 * @param values the values that should be put into the map.
	 * @param comp the function that decides how the tree is sorted, can be null
	 * @throws IllegalStateException if the keys and values do not match in lenght
	 */
	public Double2CharRBTreeMap(Double[] keys, Character[] values, DoubleComparator comp) {
		comparator = comp;
		if(keys.length != values.length) throw new IllegalStateException("Input Arrays are not equal size");
		for(int i = 0,m=keys.length;i<m;i++) put(keys[i].doubleValue(), values[i].charValue());
	}
	
	/**
	 * Helper constructor that allow to create a map from unboxed values
	 * @param keys the keys that should be put into the map
	 * @param values the values that should be put into the map.
	 * @throws IllegalStateException if the keys and values do not match in lenght
	 */
	public Double2CharRBTreeMap(double[] keys, char[] values) {
		this(keys, values, null);
	}
	
	/**
	 * Helper constructor that has a custom sorter and allow to create a map from unboxed values
	 * @param keys the keys that should be put into the map
	 * @param values the values that should be put into the map.
	 * @param comp the function that decides how the tree is sorted, can be null
	 * @throws IllegalStateException if the keys and values do not match in lenght
	 */
	public Double2CharRBTreeMap(double[] keys, char[] values, DoubleComparator comp) {
		comparator = comp;
		if(keys.length != values.length) throw new IllegalStateException("Input Arrays are not equal size");
		for(int i = 0,m=keys.length;i<m;i++) put(keys[i], values[i]);
	}
	
	/**
	 * A Helper constructor that allows to create a Map with exactly the same values as the provided map.
	 * @param map the values that should be present in the map
	 */
	public Double2CharRBTreeMap(Map<? extends Double, ? extends Character> map) {
		this(map, null);
	}
	
	/**
	 * A Helper constructor that has a custom sorter and allows to create a Map with exactly the same values as the provided map.
	 * @param map the values that should be present in the map
	 * @param comp the function that decides how the tree is sorted, can be null
	 */
	public Double2CharRBTreeMap(Map<? extends Double, ? extends Character> map, DoubleComparator comp) {
		comparator = comp;
		putAll(map);
	}
	
	/**
	 * A Type Specific Helper function that allows to create a new Map with exactly the same values as the provided map.
	 * @param map the values that should be present in the map
 	 */
	public Double2CharRBTreeMap(Double2CharMap map) {
		this(map, null);
	}
	
	/**
	 * A Type Specific Helper function that has a custom sorter and allows to create a new Map with exactly the same values as the provided map.
	 * @param map the values that should be present in the map
	 * @param comp the function that decides how the tree is sorted, can be null
 	 */
	public Double2CharRBTreeMap(Double2CharMap map, DoubleComparator comp) {
		comparator = comp;
		putAll(map);
	}
	
	@Override
	public void setDefaultMaxValue(double value) { defaultMaxNotFound = value; }
	@Override
	public double getDefaultMaxValue() { return defaultMaxNotFound; }
	@Override
	public void setDefaultMinValue(double value) { defaultMinNotFound = value; }
	@Override
	public double getDefaultMinValue() { return defaultMinNotFound; }
	
	@Override
	public char put(double key, char value) {
		if(tree == null) {
			tree = first = last = new Entry(key, value, null);
			size++;
			return getDefaultReturnValue();
		}
		int compare = 0;
		Entry parent = tree;
		while(true) {
			if((compare = compare(key, parent.key)) == 0) return parent.setValue(value);
			if(compare < 0) {
				if(parent.left == null) break;
				parent = parent.left;
			}
			else if(compare > 0) {
				if(parent.right == null) break;
				parent = parent.right;
			}
		}
		Entry adding = new Entry(key, value, parent);
		if(compare < 0)  {
			parent.left = adding;
			if(parent == first)	first = adding;
		}
		else {
			parent.right = adding;
			if(parent == last) last = adding;
		}
		fixAfterInsertion(adding);
		size++;
		return getDefaultReturnValue();
	}
	
	@Override
	public char putIfAbsent(double key, char value) {

		if(tree == null) {
			tree = first = last = new Entry(key, value, null);
			size++;
			return getDefaultReturnValue();
		}
		int compare = 0;
		Entry parent = tree;
		while(true) {
			if((compare = compare(key, parent.key)) == 0) return parent.value;
			if(compare < 0) {
				if(parent.left == null) break;
				parent = parent.left;
			}
			else if(compare > 0) {
				if(parent.right == null) break;
				parent = parent.right;
			}
		}
		Entry adding = new Entry(key, value, parent);
		if(compare < 0)  {
			parent.left = adding;
			if(parent == first)	first = adding;
		}
		else  {
			parent.right = adding;
			if(parent == last) last = adding;
		}
		fixAfterInsertion(adding);
		size++;
		return getDefaultReturnValue();
	}
	
	@Override
	public char addTo(double key, char value) {
		if(tree == null) {
			tree = first = last = new Entry(key, value, null);
			size++;
			return getDefaultReturnValue();
		}
		int compare = 0;
		Entry parent = tree;
		while(true) {
			if((compare = compare(key, parent.key)) == 0) return parent.addTo(value);
			if(compare < 0) {
				if(parent.left == null) break;
				parent = parent.left;
			}
			else if(compare > 0) {
				if(parent.right == null) break;
				parent = parent.right;
			}
		}
		Entry adding = new Entry(key, value, parent);
		if(compare < 0)  {
			parent.left = adding;
			if(parent == first)	first = adding;
		}
		else  {
			parent.right = adding;
			if(parent == last) last = adding;
		}
		fixAfterInsertion(adding);
		size++;
		return getDefaultReturnValue();
	}
	
	@Override
	public char putAndMoveToFirst(double key, char value) { throw new UnsupportedOperationException(); }
	@Override
	public char putAndMoveToLast(double key, char value) { throw new UnsupportedOperationException(); }
	@Override
	public boolean moveToFirst(double key) { throw new UnsupportedOperationException(); }
	@Override
	public boolean moveToLast(double key) { throw new UnsupportedOperationException(); }
	@Override
	public char getAndMoveToFirst(double key) { throw new UnsupportedOperationException(); }
	@Override
	public char getAndMoveToLast(double key) { throw new UnsupportedOperationException(); }
	
	@Override
	public DoubleComparator comparator() { return comparator; }
	
	@Override
	public boolean containsKey(double key) {
		return findNode(key) != null;
	}
	
	@Override
	public char get(double key) {
		Entry node = findNode(key);
		return node == null ? getDefaultReturnValue() : node.value;
	}
	
	@Override
	public char getOrDefault(double key, char defaultValue) {
		Entry node = findNode(key);
		return node == null ? defaultValue : node.value;
	}
	
	@Override
	public double firstDoubleKey() {
		if(tree == null) throw new NoSuchElementException();
		return first.key;
	}
	
	@Override
	public double pollFirstDoubleKey() {
		if(tree == null) throw new NoSuchElementException();
		double result = first.key;
		removeNode(first);
		return result;
	}
	
	@Override
	public double lastDoubleKey() {
		if(tree == null) throw new NoSuchElementException();
		return last.key;
	}
	
	@Override
	public double pollLastDoubleKey() {
		if(tree == null) throw new NoSuchElementException();
		double result = last.key;
		removeNode(last);
		return result;
	}
	
	@Override
	public Double2CharMap.Entry firstEntry() {
		if(tree == null) throw new NoSuchElementException();
		return new BasicEntry(first.key, first.value);
	}
	
	@Override
	public Double2CharMap.Entry lastEntry() {
		if(tree == null) throw new NoSuchElementException();
		return new BasicEntry(last.key, last.value);
	}
	
	@Override
	public Double2CharMap.Entry pollFirstEntry() {
		if(tree == null) throw new NoSuchElementException();
		BasicEntry entry = new BasicEntry(first.key, first.value);
		removeNode(first);
		return entry;
	}
	
	@Override
	public Double2CharMap.Entry pollLastEntry() {
		if(tree == null) throw new NoSuchElementException();
		BasicEntry entry = new BasicEntry(last.key, last.value);
		removeNode(last);
		return entry;
	}
	
	@Override
	public char firstCharValue() {
		if(tree == null) throw new NoSuchElementException();
		return first.value;
	}
	
	@Override
	public char lastCharValue() {
		if(tree == null) throw new NoSuchElementException();
		return last.value;
	}
	
	@Override
	public char remove(double key) {
		Entry entry = findNode(key);
		if(entry == null) return getDefaultReturnValue();
		char value = entry.value;
		removeNode(entry);
		return value;
	}
	
	@Override
	public char removeOrDefault(double key, char defaultValue) {
		Entry entry = findNode(key);
		if(entry == null) return defaultValue;
		char value = entry.value;
		removeNode(entry);
		return value;
	}
	
	@Override
	public boolean remove(double key, char value) {
		Entry entry = findNode(key);
		if(entry == null || entry.value != value) return false;
		removeNode(entry);
		return true;
	}
	
	@Override
	public boolean replace(double key, char oldValue, char newValue) {
		Entry entry = findNode(key);
		if(entry == null || entry.value != oldValue) return false;
		entry.value = newValue;
		return true;
	}
	
	@Override
	public char replace(double key, char value) {
		Entry entry = findNode(key);
		if(entry == null) return getDefaultReturnValue();
		char oldValue = entry.value;
		entry.value = value;
		return oldValue;
	}
	
	@Override
	public char computeChar(double key, DoubleCharUnaryOperator mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		Entry entry = findNode(key);
		if(entry == null) {
			char newValue = mappingFunction.applyAsChar(key, getDefaultReturnValue());
			if(newValue == getDefaultReturnValue()) return newValue;
			put(key, newValue);
			return newValue;
		}
		char newValue = mappingFunction.applyAsChar(key, entry.value);
		if(newValue == getDefaultReturnValue()) {
			removeNode(entry);
			return newValue;
		}
		entry.value = newValue;
		return newValue;
	}
	
	@Override
	public char computeCharIfAbsent(double key, Double2CharFunction mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		Entry entry = findNode(key);
		if(entry == null) {
			char newValue = mappingFunction.get(key);
			if(newValue == getDefaultReturnValue()) return newValue;
			put(key, newValue);
			return newValue;
		}
		if(entry.value == getDefaultReturnValue()) {
			char newValue = mappingFunction.get(key);
			if(newValue == getDefaultReturnValue()) return newValue;
			entry.value = newValue;
		}
		return entry.value;
	}
	
	@Override
	public char supplyCharIfAbsent(double key, CharSupplier valueProvider) {
		Objects.requireNonNull(valueProvider);
		Entry entry = findNode(key);
		if(entry == null) {
			char newValue = valueProvider.getChar();
			if(newValue == getDefaultReturnValue()) return newValue;
			put(key, newValue);
			return newValue;
		}
		if(entry.value == getDefaultReturnValue()) {
			char newValue = valueProvider.getChar();
			if(newValue == getDefaultReturnValue()) return newValue;
			entry.value = newValue;
		}
		return entry.value;
	}
	
	@Override
	public char computeCharIfPresent(double key, DoubleCharUnaryOperator mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		Entry entry = findNode(key);
		if(entry == null || entry.value == getDefaultReturnValue()) return getDefaultReturnValue();
		char newValue = mappingFunction.applyAsChar(key, entry.value);
		if(newValue == getDefaultReturnValue()) {
			removeNode(entry);
			return newValue;
		}
		entry.value = newValue;
		return newValue;
	}
	
	@Override
	public char mergeChar(double key, char value, CharCharUnaryOperator mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		Entry entry = findNode(key);
		char newValue = entry == null || entry.value == getDefaultReturnValue() ? value : mappingFunction.applyAsChar(entry.value, value);
		if(newValue == getDefaultReturnValue()) {
			if(entry != null)
				removeNode(entry);
		}
		else if(entry == null) put(key, newValue);
		else entry.value = newValue;
		return newValue;
	}
	
	@Override
	public void mergeAllChar(Double2CharMap m, CharCharUnaryOperator mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		for(Double2CharMap.Entry entry : Double2CharMaps.fastIterable(m)) {
			double key = entry.getDoubleKey();
			Entry subEntry = findNode(key);
			char newValue = subEntry == null || subEntry.value == getDefaultReturnValue() ? entry.getCharValue() : mappingFunction.applyAsChar(subEntry.value, entry.getCharValue());
			if(newValue == getDefaultReturnValue()) {
				if(subEntry != null)
					removeNode(subEntry);
			}
			else if(subEntry == null) put(key, newValue);
			else subEntry.value = newValue;
		}
	}
	
	@Override
	public void forEach(DoubleCharConsumer action) {
		for(Entry entry = first;entry != null;entry = entry.next())
			action.accept(entry.key, entry.value);
	}
	
	@Override
	public int size() { return size; }
	
	@Override
	public void clear() {
		size = 0;
		first = null;
		last = null;
		tree = null;
	}
	
	DoubleListIterator keyIterator(boolean descending) {
		DoubleListIterator iter = new KeyIterator(descending);
		return descending ? DoubleIterators.invert(iter) : iter;
	}
	
	DoubleListIterator keyIterator(double key) {
		return new KeyIterator(findNode(key));
	}
	
	public Double2CharRBTreeMap copy() {
		Double2CharRBTreeMap set = new Double2CharRBTreeMap();
		set.size = size;
		if(tree != null) {
			set.tree = tree.copy();
			Entry lastFound = null;
			for(Entry entry = tree;entry != null;entry = entry.left) lastFound = entry;
			set.first = lastFound;
			lastFound = null;
			for(Entry entry = tree;entry != null;entry = entry.right) lastFound = entry;
			set.last = lastFound;
		}
		return set;
	}
	
	@Override
	public DoubleSortedSet keySet() {
		return navigableKeySet();
	}
	
	@Override
	public ObjectSet<Double2CharMap.Entry> double2CharEntrySet() {
		if(entrySet == null) entrySet = new EntrySet();
		return entrySet;
	}
	
	@Override
	public CharCollection values() {
		if(values == null) values = new Values();
		return values;
	}
	
	@Override
	public DoubleNavigableSet navigableKeySet() {
		if(keySet == null) keySet = new KeySet(this);
		return keySet;
	}
	
	@Override
	public Double2CharNavigableMap descendingMap() {
		return new DescendingSubMap(this, true, 0D, true, true, 0D, true);
	}
	
	@Override
	public DoubleNavigableSet descendingKeySet() {
		return descendingMap().navigableKeySet();
	}
	
	@Override
	public Double2CharNavigableMap subMap(double fromKey, boolean fromInclusive, double toKey, boolean toInclusive) {
		return new AscendingSubMap(this, false, fromKey, fromInclusive, false, toKey, toInclusive);
	}
	
	@Override
	public Double2CharNavigableMap headMap(double toKey, boolean inclusive) {
		return new AscendingSubMap(this, true, 0D, true, false, toKey, inclusive);
	}
	
	@Override
	public Double2CharNavigableMap tailMap(double fromKey, boolean inclusive) {
		return new AscendingSubMap(this, false, fromKey, inclusive, true, 0D, true);
	}
	
	@Override
	public double lowerKey(double e) {
		Entry node = findLowerNode(e);
		return node != null ? node.key : getDefaultMinValue();
	}

	@Override
	public double floorKey(double e) {
		Entry node = findFloorNode(e);
		return node != null ? node.key : getDefaultMinValue();
	}
	
	@Override
	public double higherKey(double e) {
		Entry node = findHigherNode(e);
		return node != null ? node.key : getDefaultMaxValue();
	}

	@Override
	public double ceilingKey(double e) {
		Entry node = findCeilingNode(e);
		return node != null ? node.key : getDefaultMaxValue();
	}
	
	@Override
	public Double2CharMap.Entry lowerEntry(double key) {
		Entry node = findLowerNode(key);
		return node != null ? new BasicEntry(node.key, node.value) : null;
	}
	
	@Override
	public Double2CharMap.Entry higherEntry(double key) {
		Entry node = findHigherNode(key);
		return node != null ? new BasicEntry(node.key, node.value) : null;
	}
	
	@Override
	public Double2CharMap.Entry floorEntry(double key) {
		Entry node = findFloorNode(key);
		return node != null ? new BasicEntry(node.key, node.value) : null;
	}
	
	@Override
	public Double2CharMap.Entry ceilingEntry(double key) {
		Entry node = findCeilingNode(key);
		return node != null ? new BasicEntry(node.key, node.value) : null;
	}
	
	protected Entry findLowerNode(double key) {
		Entry entry = tree;
		while(entry != null) {
			if(compare(key, entry.key) > 0) {
				if(entry.right != null) entry = entry.right;
				else return entry;
			}
			else {
				if(entry.left != null) entry = entry.left;
				else {
					Entry parent = entry.parent;
					while(parent != null && parent.left == entry) {
						entry = parent;
						parent = parent.parent;
					}
					return parent;
				}
			}
		}
		return null;
	}
	
	protected Entry findFloorNode(double key) {
		Entry entry = tree;
		int compare;
		while(entry != null) {
			if((compare = compare(key, entry.key)) > 0) {
				if(entry.right == null) break;
				entry = entry.right;
				continue;
			}
			else if(compare < 0) {
				if(entry.left != null) entry = entry.left;
				else {
					Entry parent = entry.parent;
					while(parent != null && parent.left == entry) {
						entry = parent;
						parent = parent.parent;
					}
					return parent;
				}
				continue;
			}
			break;
		}
		return entry;
	}
	
	protected Entry findCeilingNode(double key) {
		Entry entry = tree;
		int compare;
		while(entry != null) {
			if((compare = compare(key, entry.key)) < 0) {
				if(entry.left == null) break;
				entry = entry.left;
				continue;
			}
			else if(compare > 0) {
				if(entry.right != null) entry = entry.right;
				else {
					Entry parent = entry.parent;
					while(parent != null && parent.right == entry) {
						entry = parent;
						parent = parent.parent;
					}
					return parent;
				}
				continue;
			}
			break;
		}
		return entry;
	}
	
	protected Entry findHigherNode(double key) {
		Entry entry = tree;
		while(entry != null) {
			if(compare(key, entry.key) < 0) {
				if(entry.left != null) entry = entry.left;
				else return entry;
			}
			else {
				if(entry.right != null) entry = entry.right;
				else {
					Entry parent = entry.parent;
					while(parent != null && parent.right == entry) {
						entry = parent;
						parent = parent.parent;
					}
					return parent;
				}
			}
		}
		return null;
	}
	
	protected Entry findNode(double key) {
		Entry node = tree;
		int compare;
		while(node != null) {
			if((compare = compare(key, node.key)) == 0) return node;
			if(compare < 0) node = node.left;
			else node = node.right;
		}
		return null;
	}
	
	protected void removeNode(Entry entry) {
		size--;
		if(entry.needsSuccessor()) {
			Entry successor = entry.next();
			entry.key = successor.key;
			entry.value = successor.value;
			entry = successor;
		}
		Entry replacement = entry.left != null ? entry.left : entry.right;
		if(replacement != null) {
			if(entry.replace(replacement)) tree = replacement;
			if(entry == first) first = replacement;
			if(entry == last) last = entry.right != null ? entry.right : replacement;
			entry.left = entry.right = entry.parent = null;
			if(entry.isBlack()) fixAfterDeletion(replacement);
		}
		else if(entry.parent == null) tree = first = last = null;
		else {
			if(entry.isBlack())
				fixAfterDeletion(entry);
			entry.replace(null);
			if(entry.parent != null) {
				Entry parent = entry.parent;
				if(entry == first) first = parent.left != null ? parent.left : parent;
				if(entry == last) last = entry.right != null ? parent.right : parent;
			}
			entry.parent = null;
		}
	}
	
	protected void validate(double k) { compare(k, k); }
	protected int compare(double k, double v) { return comparator != null ? comparator.compare(k, v) : Double.compare(k, v);}
	protected static boolean isBlack(Entry p) { return p == null || p.isBlack(); }
	protected static Entry parentOf(Entry p) { return (p == null ? null : p.parent); }
	protected static void setBlack(Entry p, boolean c) { if(p != null) p.setBlack(c); }
	protected static Entry leftOf(Entry p) { return p == null ? null : p.left; }
	protected static Entry rightOf(Entry p) { return (p == null) ? null : p.right; }
	
	protected void rotateLeft(Entry entry) {
		if(entry != null) {
			Entry right = entry.right;
			entry.right = right.left;
			if(right.left != null) right.left.parent = entry;
			right.parent = entry.parent;
			if(entry.parent == null) tree = right;
			else if(entry.parent.left == entry) entry.parent.left = right;
			else entry.parent.right = right;
			right.left = entry;
			entry.parent = right;
		}
	}
	
	protected void rotateRight(Entry entry) {
		if(entry != null) {
			Entry left = entry.left;
			entry.left = left.right;
			if(left.right != null) left.right.parent = entry;
			left.parent = entry.parent;
			if(entry.parent == null) tree = left;
			else if(entry.parent.right == entry) entry.parent.right = left;
			else entry.parent.left = left;
			left.right = entry;
			entry.parent = left;
		}
	}
	
	protected void fixAfterInsertion(Entry entry) {
		entry.setBlack(false);
		while(entry != null && entry != tree && !entry.parent.isBlack()) {
			if(parentOf(entry) == leftOf(parentOf(parentOf(entry)))) {
				Entry y = rightOf(parentOf(parentOf(entry)));
				if(!isBlack(y)) {
					setBlack(parentOf(entry), true);
					setBlack(y, true);
					setBlack(parentOf(parentOf(entry)), false);
					entry = parentOf(parentOf(entry));
				}
				else {
					if(entry == rightOf(parentOf(entry))) {
						entry = parentOf(entry);
						rotateLeft(entry);
					}
					setBlack(parentOf(entry), true);
					setBlack(parentOf(parentOf(entry)), false);
					rotateRight(parentOf(parentOf(entry)));
				}
			}
			else {
				Entry y = leftOf(parentOf(parentOf(entry)));
				if(!isBlack(y)) {
					setBlack(parentOf(entry), true);
					setBlack(y, true);
					setBlack(parentOf(parentOf(entry)), false);
					entry = parentOf(parentOf(entry));
				}
				else {
					if(entry == leftOf(parentOf(entry))) {
						entry = parentOf(entry);
						rotateRight(entry);
					}
					setBlack(parentOf(entry), true);
					setBlack(parentOf(parentOf(entry)), false);
					rotateLeft(parentOf(parentOf(entry)));
				}
			}
		}
		tree.setBlack(true);
	}
	
	protected void fixAfterDeletion(Entry entry) {
		while(entry != tree && isBlack(entry)) {
			if(entry == leftOf(parentOf(entry))) {
				Entry sib = rightOf(parentOf(entry));
				if(!isBlack(sib)) {
					setBlack(sib, true);
					setBlack(parentOf(entry), false);
					rotateLeft(parentOf(entry));
					sib = rightOf(parentOf(entry));
				}
				if(isBlack(leftOf(sib)) && isBlack(rightOf(sib))) {
					setBlack(sib, false);
					entry = parentOf(entry);
				}
				else {
					if(isBlack(rightOf(sib))) {
						setBlack(leftOf(sib), true);
						setBlack(sib, false);
						rotateRight(sib);
						sib = rightOf(parentOf(entry));
					}
					setBlack(sib, isBlack(parentOf(entry)));
					setBlack(parentOf(entry), true);
					setBlack(rightOf(sib), true);
					rotateLeft(parentOf(entry));
					entry = tree;
				}
			}
			else {
				Entry sib = leftOf(parentOf(entry));
				if(!isBlack(sib)) {
					setBlack(sib, true);
					setBlack(parentOf(entry), false);
					rotateRight(parentOf(entry));
					sib = leftOf(parentOf(entry));
				}
				if(isBlack(rightOf(sib)) && isBlack(leftOf(sib))) {
					setBlack(sib, false);
					entry = parentOf(entry);
				}
				else {
					if(isBlack(leftOf(sib))) {
						setBlack(rightOf(sib), true);
						setBlack(sib, false);
						rotateLeft(sib);
						sib = leftOf(parentOf(entry));
					}
					setBlack(sib, isBlack(parentOf(entry)));
					setBlack(parentOf(entry), true);
					setBlack(leftOf(sib), true);
					rotateRight(parentOf(entry));
					entry = tree;
				}
			}
		}
		setBlack(entry, true);
	}
	
	static final class AscendingSubMap extends NavigableSubMap {

		public AscendingSubMap(Double2CharRBTreeMap m, boolean fromStart, double low, boolean loInclusive, boolean toEnd, double high, boolean hiInclusive) {
			super(m, fromStart, low, loInclusive, toEnd, high, hiInclusive);
		}
		
		@Override
		DoubleListIterator keyIterator(boolean descending) { return new SubMapKeyIterator(descending); }
		@Override
		DoubleListIterator keyIterator(double key) {
			Double2CharRBTreeMap.Entry entry = m.findNode(key);
			return entry == null || !inClosedRange(key) ? null : new SubMapKeyIterator(entry, fromStart ? null : findLowest(), toEnd ? null : findHighest());
		}
		@Override
		CharListIterator valueIterator() { return new SubMapValueIterator(false); }
		@Override
		ObjectListIterator<Double2CharMap.Entry> entryIterator() { return new SubMapEntrySetIterator(false); }
		
		@Override
		public Double2CharNavigableMap descendingMap() {
			if(descendingMap == null) descendingMap = new DescendingSubMap(m, fromStart, low, loInclusive, toEnd, high, hiInclusive);
			return descendingMap;
		}
		
		Double2CharRBTreeMap.Entry next(Double2CharRBTreeMap.Entry entry) {
			return entry.next();
		}
		
		Double2CharRBTreeMap.Entry previous(Double2CharRBTreeMap.Entry entry) {
			return entry.previous();
		}
		
		@Override
		public Double2CharNavigableMap subMap(double fromKey, boolean fromInclusive, double toKey, boolean toInclusive) {
			if (!inRange(fromKey, fromInclusive)) throw new IllegalArgumentException("fromKey out of range");
			if (!inRange(toKey, toInclusive)) throw new IllegalArgumentException("toKey out of range");
			return new AscendingSubMap(m, false, fromKey, fromInclusive, false, toKey, toInclusive);
		}
		
		@Override
		public Double2CharNavigableMap headMap(double toKey, boolean inclusive) {
			if (!inRange(toKey, inclusive)) throw new IllegalArgumentException("toKey out of range");
			return new AscendingSubMap(m, fromStart, low, loInclusive, false, toKey, inclusive);
		}
		
		@Override
		public Double2CharNavigableMap tailMap(double fromKey, boolean inclusive) {
			if (!inRange(fromKey, inclusive)) throw new IllegalArgumentException("fromKey out of range");
			return new AscendingSubMap(m, false, fromKey, inclusive, toEnd, high, hiInclusive);
		}
	}
	
	static final class DescendingSubMap extends NavigableSubMap {

		public DescendingSubMap(Double2CharRBTreeMap m, boolean fromStart, double low, boolean loInclusive, boolean toEnd, double high, boolean hiInclusive) {
			super(m, fromStart, low, loInclusive, toEnd, high, hiInclusive);
		}
		
		@Override
		public double firstDoubleKey() { return super.lastDoubleKey(); }
		
		@Override
		public double pollFirstDoubleKey() { return super.pollLastDoubleKey(); }
		
		@Override
		public double lastDoubleKey() { return super.firstDoubleKey(); }
		
		@Override
		public double pollLastDoubleKey() { return super.pollFirstDoubleKey(); }
		
		@Override
		public char firstCharValue() { return super.lastCharValue(); }
		
		@Override
		public char lastCharValue() { return super.firstCharValue(); }
		
		@Override
		public Double2CharMap.Entry firstEntry() { return super.lastEntry(); }
		
		@Override
		public Double2CharMap.Entry lastEntry() { return super.firstEntry(); }
		
		@Override
		public Double2CharMap.Entry pollFirstEntry() { return super.pollLastEntry(); }
		
		@Override
		public Double2CharMap.Entry pollLastEntry() { return super.pollFirstEntry(); }
		
		@Override
		public double lowerKey(double e) { return super.higherKey(e); }
		
		@Override
		public double floorKey(double e) { return super.ceilingKey(e); }
		
		@Override
		public double ceilingKey(double e) { return super.floorKey(e); }
		
		@Override
		public double higherKey(double e) { return super.lowerKey(e); }
		
		@Override
		public Double2CharMap.Entry lowerEntry(double e) { return super.higherEntry(e); }
		
		@Override
		public Double2CharMap.Entry floorEntry(double e) { return super.ceilingEntry(e); }
		
		@Override
		public Double2CharMap.Entry ceilingEntry(double e) { return super.floorEntry(e); }
		
		@Override
		public Double2CharMap.Entry higherEntry(double e) { return super.lowerEntry(e); }
		
		protected Double2CharRBTreeMap.Entry findLowest() {
			if(fromStart) return m.first;
			Double2CharRBTreeMap.Entry entry = loInclusive ? m.findCeilingNode(low) : m.findHigherNode(low);
			return entry == null || tooHigh(entry.key) ? null : entry;
		}

		protected Double2CharRBTreeMap.Entry findHighest() {
			if(toEnd) return m.last;
			Double2CharRBTreeMap.Entry entry = hiInclusive ? m.findFloorNode(high) : m.findLowerNode(high);
			return entry == null || tooLow(entry.key) ? null : entry;
		}
		
		@Override
		DoubleListIterator keyIterator(boolean descending) {
			DoubleListIterator iter = new SubMapKeyIterator(!descending);
			return descending ? iter : DoubleIterators.invert(iter);
		}
		
		@Override
		DoubleListIterator keyIterator(double key) {
			Double2CharRBTreeMap.Entry entry = m.findNode(key);
			return entry == null || !inClosedRange(key) ? null : DoubleIterators.invert(new SubMapKeyIterator(entry, fromStart ? null : findLowest(), toEnd ? null : findHighest()));
		}
		
		Double2CharRBTreeMap.Entry next(Double2CharRBTreeMap.Entry entry) {
			return entry.previous();
		}
		
		Double2CharRBTreeMap.Entry previous(Double2CharRBTreeMap.Entry entry) {
			return entry.next();
		}
		
		@Override
		public DoubleComparator comparator() { return m.comparator() == null ? null : m.comparator().reversed(); }
		
		@Override
		CharListIterator valueIterator() { return CharIterators.invert(new SubMapValueIterator(true)); }
		
		@Override
		ObjectListIterator<Double2CharMap.Entry> entryIterator() { return ObjectIterators.invert(new SubMapEntrySetIterator(true)); }
		
		@Override
		public Double2CharNavigableMap descendingMap() {
			if(descendingMap == null) descendingMap = new AscendingSubMap(m, fromStart, low, loInclusive, toEnd, high, hiInclusive);
			return descendingMap;
		}

		@Override
		public Double2CharNavigableMap subMap(double fromKey, boolean fromInclusive, double toKey, boolean toInclusive) {
			if (!inRange(fromKey, fromInclusive)) throw new IllegalArgumentException("fromKey out of range");
			if (!inRange(toKey, toInclusive)) throw new IllegalArgumentException("toKey out of range");
			return new DescendingSubMap(m, false, fromKey, fromInclusive, false, toKey, toInclusive);
		}
		
		@Override
		public Double2CharNavigableMap headMap(double toKey, boolean inclusive) {
			if (!inRange(toKey, inclusive)) throw new IllegalArgumentException("toKey out of range");
			return new DescendingSubMap(m, fromStart, low, loInclusive, false, toKey, inclusive);
		}
		
		@Override
		public Double2CharNavigableMap tailMap(double fromKey, boolean inclusive) {
			if (!inRange(fromKey, inclusive)) throw new IllegalArgumentException("fromKey out of range");
			return new DescendingSubMap(m, false, fromKey, inclusive, toEnd, high, hiInclusive);
		}
	}
	
	static abstract class NavigableSubMap extends AbstractDouble2CharMap implements Double2CharNavigableMap {
		final Double2CharRBTreeMap m;
		final double low;
		final double high;
		final boolean fromStart;
		final boolean toEnd;
		final boolean loInclusive;
		final boolean hiInclusive;
		transient Double2CharNavigableMap descendingMap;
		transient SubMapEntrySet entrySet;
		transient KeySet keySet;
		transient SubMapValues values;
		
		NavigableSubMap(Double2CharRBTreeMap m, boolean fromStart, double low, boolean loInclusive, boolean toEnd, double high, boolean hiInclusive) {
			this.m = m;
			this.low = low;
			this.high = high;
			this.fromStart = fromStart;
			this.toEnd = toEnd;
			this.loInclusive = loInclusive;
			this.hiInclusive = hiInclusive;
		}
		
		abstract DoubleListIterator keyIterator(boolean descending);
		abstract DoubleListIterator keyIterator(double key);
		abstract CharListIterator valueIterator();
		abstract ObjectListIterator<Double2CharMap.Entry> entryIterator();
		abstract Double2CharRBTreeMap.Entry next(Double2CharRBTreeMap.Entry entry);
		abstract Double2CharRBTreeMap.Entry previous(Double2CharRBTreeMap.Entry entry);
		
		@Override
		public NavigableSubMap copy() { throw new UnsupportedOperationException(); }
		@Override
		public abstract Double2CharNavigableMap descendingMap();
		@Override
		public abstract Double2CharNavigableMap subMap(double fromKey, boolean fromInclusive, double toKey, boolean toInclusive);
		@Override
		public abstract Double2CharNavigableMap headMap(double toKey, boolean inclusive);
		@Override
		public abstract Double2CharNavigableMap tailMap(double fromKey, boolean inclusive);
		
		boolean tooLow(double key) { return !fromStart && (loInclusive ? m.compare(key, low) < 0 : m.compare(key, low) <= 0); }
		boolean tooHigh(double key) { return !toEnd && (hiInclusive ? m.compare(key, high) > 0 : m.compare(key, high) >= 0); }
		boolean inRange(double key) { return !tooLow(key) && !tooHigh(key); }
		boolean inClosedRange(double key) { return (fromStart || m.compare(key, low) >= 0) && (toEnd || m.compare(high, key) >= 0); }
		boolean inRange(double key, boolean inclusive) { return inclusive ? inRange(key) : inClosedRange(key); }
		
		@Override
		public void setDefaultMaxValue(double e) { m.setDefaultMaxValue(e); }
		@Override
		public double getDefaultMaxValue() { return m.getDefaultMaxValue(); }
		@Override
		public void setDefaultMinValue(double e) { m.setDefaultMinValue(e); }
		@Override
		public double getDefaultMinValue() { return m.getDefaultMinValue(); }
		
		@Override
		public AbstractDouble2CharMap setDefaultReturnValue(char v) { 
			m.setDefaultReturnValue(v);
			return this;
		}
		
		@Override
		public char getDefaultReturnValue() { return m.getDefaultReturnValue(); }
		
		@Override
		public char putAndMoveToFirst(double key, char value) { throw new UnsupportedOperationException(); }
		@Override
		public char putAndMoveToLast(double key, char value) { throw new UnsupportedOperationException(); }
		@Override
		public boolean moveToFirst(double key) { throw new UnsupportedOperationException(); }
		@Override
		public boolean moveToLast(double key) { throw new UnsupportedOperationException(); }
		@Override
		public char getAndMoveToFirst(double key) { throw new UnsupportedOperationException(); }
		@Override
		public char getAndMoveToLast(double key) { throw new UnsupportedOperationException(); }
		@Override
		public DoubleComparator comparator() { return m.comparator(); }
		@Override
		public boolean containsKey(double key) { return inRange(key) && m.containsKey(key); }
		
		@Override
		public double firstDoubleKey() { 
			Double2CharRBTreeMap.Entry entry = findLowest();
			return entry == null ? getDefaultMaxValue() : entry.key;
		}
		
		@Override
		public double pollFirstDoubleKey() {
			if(fromStart) return m.pollFirstDoubleKey();
			Double2CharRBTreeMap.Entry entry = loInclusive ? m.findCeilingNode(low) : m.findHigherNode(low);
			if(entry != null && !tooHigh(entry.key)) {
				double value = entry.key;
				m.removeNode(entry);
				return value;
			}
			return getDefaultMaxValue();
		}
		
		@Override
		public double lastDoubleKey() {
			Double2CharRBTreeMap.Entry entry = findHighest();
			return entry == null ? getDefaultMinValue() : entry.key;	
		}
		
		@Override
		public double pollLastDoubleKey() {
			if(toEnd) return m.pollLastDoubleKey();
			Double2CharRBTreeMap.Entry entry = hiInclusive ? m.findFloorNode(high) : m.findLowerNode(high);
			if(entry != null && !tooLow(entry.key)) {
				double value = entry.key;
				m.removeNode(entry);
				return value;
			}
			return getDefaultMinValue();
		}
		
		@Override
		public char firstCharValue() {
			Double2CharRBTreeMap.Entry entry = findLowest();
			return entry == null ? getDefaultReturnValue() : entry.value;
		}
		
		@Override
		public char lastCharValue() {
			Double2CharRBTreeMap.Entry entry = findHighest();
			return entry == null ? getDefaultReturnValue() : entry.value;
		}
		
		protected Double2CharRBTreeMap.Entry findLowest() {
			if(fromStart) return m.first;
			Double2CharRBTreeMap.Entry entry = loInclusive ? m.findCeilingNode(low) : m.findHigherNode(low);
			return entry == null || tooHigh(entry.key) ? null : entry;
		}
		
		protected Double2CharRBTreeMap.Entry findHighest() {
			if(toEnd) return m.last;
			Double2CharRBTreeMap.Entry entry = hiInclusive ? m.findFloorNode(high) : m.findLowerNode(high);
			return entry == null || tooLow(entry.key) ? null : entry;
		}
		
		@Override
		public char put(double key, char value) {
			if(!inRange(key)) throw new IllegalArgumentException("key out of range");
			return m.put(key, value);
		}
		
		@Override
		public char putIfAbsent(double key, char value) {
			if(!inRange(key)) throw new IllegalArgumentException("key out of range");
			return m.putIfAbsent(key, value);
		}
		
		@Override
		public char addTo(double key, char value) {
			if(!inRange(key)) throw new IllegalArgumentException("key out of range");
			return m.addTo(key, value);
		}
		
		@Override
		public char get(double key) {
			return inRange(key) ? m.get(key) : getDefaultReturnValue();
		}
		
		@Override
		public char getOrDefault(double key, char defaultValue) {
			return inRange(key) ? m.getOrDefault(key, defaultValue) : getDefaultReturnValue();
		}
		
		@Override
		public char remove(double key) {
			return inRange(key) ? m.remove(key) : getDefaultReturnValue();
		}
		
		@Override
		public char removeOrDefault(double key, char defaultValue) {
			return inRange(key) ? m.remove(key) : defaultValue;
		}
		
		@Override
		public boolean remove(double key, char value) {
			return inRange(key) && m.remove(key, value);
		}
		
		@Override
		public ObjectSet<Double2CharMap.Entry> double2CharEntrySet() {
			if(entrySet == null) entrySet = new SubMapEntrySet();
			return entrySet;
		}
		
		@Override
		public DoubleSet keySet() { return navigableKeySet(); }
		
		@Override
		public DoubleNavigableSet navigableKeySet() {
			if(keySet == null) keySet = new KeySet(this);
			return keySet;
		}
		
		@Override
		public DoubleNavigableSet descendingKeySet() {
			return descendingMap().navigableKeySet();
		}
		
		@Override
		public CharCollection values() {
			if(values == null) values = new SubMapValues();
			return values;
		}
		
		@Override
		public void forEach(DoubleCharConsumer action) {
			for(Double2CharRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
				action.accept(entry.key, entry.value);
		}
		
		@Override
		public int size() {
			return (fromStart && toEnd) ? m.size() : entrySet().size();
		}
		
		@Override
		public Double2CharMap.Entry firstEntry() {
			Double2CharRBTreeMap.Entry entry = findLowest();
			return entry == null ? null : new BasicEntry(entry.key, entry.value);
		}
		
		@Override
		public Double2CharMap.Entry lastEntry() {
			Double2CharRBTreeMap.Entry entry = findHighest();
			return entry == null ? null : new BasicEntry(entry.key, entry.value);
		}
		
		@Override
		public Double2CharMap.Entry pollFirstEntry() {
			Double2CharRBTreeMap.Entry entry = findLowest();
			if(entry == null) return null;
			BasicEntry result = new BasicEntry(entry.key, entry.value);
			m.removeNode(entry);
			return result;
		}
		
		@Override
		public Double2CharMap.Entry pollLastEntry() {
			Double2CharRBTreeMap.Entry entry = findHighest();
			if(entry == null) return null;
			BasicEntry result = new BasicEntry(entry.key, entry.value);
			m.removeNode(entry);
			return result;
		}
		
		@Override
		public double lowerKey(double e) {
			if(tooHigh(e)) {
				Double2CharRBTreeMap.Entry entry = findHighest();
				return entry == null ? getDefaultMinValue() : entry.key;
			}
			Double2CharRBTreeMap.Entry entry = m.findLowerNode(e);
			return entry == null || tooHigh(entry.key) ? getDefaultMaxValue() : entry.key;
		}
		
		@Override
		public double floorKey(double e) {
			if(tooHigh(e)) {
				Double2CharRBTreeMap.Entry entry = findHighest();
				return entry == null ? getDefaultMinValue() : entry.key;
			}
			Double2CharRBTreeMap.Entry entry = m.findFloorNode(e);
			return entry == null || tooHigh(entry.key) ? getDefaultMaxValue() : entry.key;
		}
		
		@Override
		public double ceilingKey(double e) {
			if(tooLow(e)) {
				Double2CharRBTreeMap.Entry entry = findLowest();
				return entry == null ? getDefaultMaxValue() : entry.key;
			}
			Double2CharRBTreeMap.Entry entry = m.findCeilingNode(e);
			return entry == null || tooLow(entry.key) ? getDefaultMinValue() : entry.key;
		}
		
		@Override
		public double higherKey(double e) {
			if(tooLow(e)) {
				Double2CharRBTreeMap.Entry entry = findLowest();
				return entry == null ? getDefaultMaxValue() : entry.key;
			}
			Double2CharRBTreeMap.Entry entry = m.findHigherNode(e);
			return entry == null || tooLow(entry.key) ? getDefaultMinValue() : entry.key;
		}
		
		@Override
		public Double2CharMap.Entry lowerEntry(double e) {
			if(tooHigh(e)) {
				Double2CharRBTreeMap.Entry entry = findHighest();
				return entry == null ? null : new BasicEntry(entry.key, entry.value);
			}
			Double2CharRBTreeMap.Entry entry = m.findLowerNode(e);
			return entry == null || tooHigh(entry.key) ? null : new BasicEntry(entry.key, entry.value);
		}
		
		@Override
		public Double2CharMap.Entry floorEntry(double e) {
			if(tooHigh(e)) {
				Double2CharRBTreeMap.Entry entry = findHighest();
				return entry == null ? null : new BasicEntry(entry.key, entry.value);
			}
			Double2CharRBTreeMap.Entry entry = m.findFloorNode(e);
			return entry == null || tooHigh(entry.key) ? null : new BasicEntry(entry.key, entry.value);
		}
		
		@Override
		public Double2CharMap.Entry ceilingEntry(double e) {
			if(tooLow(e)) {
				Double2CharRBTreeMap.Entry entry = findLowest();
				return entry == null ? null : new BasicEntry(entry.key, entry.value);
			}
			Double2CharRBTreeMap.Entry entry = m.findCeilingNode(e);
			return entry == null || tooLow(entry.key) ? null : new BasicEntry(entry.key, entry.value);
		}
		
		@Override
		public Double2CharMap.Entry higherEntry(double e) {
			if(tooLow(e)) {
				Double2CharRBTreeMap.Entry entry = findLowest();
				return entry == null ? null : new BasicEntry(entry.key, entry.value);
			}
			Double2CharRBTreeMap.Entry entry = m.findHigherNode(e);
			return entry == null || tooLow(entry.key) ? null : new BasicEntry(entry.key, entry.value);
		}
		
		class SubMapEntrySet extends AbstractObjectSet<Double2CharMap.Entry> implements ObjectSet<Double2CharMap.Entry> {
			@Override
			@Deprecated
			public boolean contains(Object o) {
				if(o instanceof Map.Entry) {
					if(o instanceof Double2CharMap.Entry) {
						Double2CharMap.Entry entry = (Double2CharMap.Entry)o;
						if(entry.getKey() == null && comparator() == null) return false;
						Double2CharRBTreeMap.Entry subEntry = m.findNode(entry.getDoubleKey());
						if(subEntry != null) return entry.getCharValue() == subEntry.value;
					}
					else {
						Map.Entry<?, ?> entry = (Map.Entry<?, ?>)o;
						if(entry.getKey() == null && comparator() == null) return false;
						if(!(entry.getKey() instanceof Double)) return false;
						Double2CharRBTreeMap.Entry subEntry = m.findNode((Double)entry.getKey());
						if(subEntry != null) return Objects.equals(entry.getValue(), Character.valueOf(subEntry.value));
					}
				}
				return false;
			}
			
			@Override
			@Deprecated
			public boolean remove(Object o) {
				if(o instanceof Map.Entry) {
					if(o instanceof Double2CharMap.Entry) {
						Double2CharMap.Entry entry = (Double2CharMap.Entry)o;
						return NavigableSubMap.this.remove(entry.getDoubleKey(), entry.getCharValue());
					}
					Map.Entry<?, ?> entry = (Map.Entry<?, ?>)o;
					return NavigableSubMap.this.remove(entry.getKey(), entry.getValue());
				}
				return false;
			}
			@Override
			public ObjectIterator<Double2CharMap.Entry> iterator() { return entryIterator(); }
			@Override
			public int size() { return NavigableSubMap.this.size(); }
			@Override
			public void clear() { NavigableSubMap.this.clear(); }
			
			@Override
			public void forEach(Consumer<? super Double2CharMap.Entry> action) {
				Objects.requireNonNull(action);
				for(Double2CharRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					action.accept(new BasicEntry(entry.key, entry.value));
			}
			
			@Override
			public <E> void forEach(E input, ObjectObjectConsumer<E, Double2CharMap.Entry> action) {
				Objects.requireNonNull(action);
				for(Double2CharRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					action.accept(input, new BasicEntry(entry.key, entry.value));
			}
			
			@Override
			public boolean matchesAny(Object2BooleanFunction<Double2CharMap.Entry> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return false;
				BasicEntry subEntry = new BasicEntry();
				for(Double2CharRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					subEntry.set(entry.key, entry.value);
					if(filter.getBoolean(subEntry)) return true;
				}
				return false;
			}
			
			@Override
			public boolean matchesNone(Object2BooleanFunction<Double2CharMap.Entry> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return true;
				BasicEntry subEntry = new BasicEntry();
				for(Double2CharRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					subEntry.set(entry.key, entry.value);
					if(filter.getBoolean(subEntry)) return false;
				}
				return true;
			}
			
			@Override
			public boolean matchesAll(Object2BooleanFunction<Double2CharMap.Entry> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return true;
				BasicEntry subEntry = new BasicEntry();
				for(Double2CharRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					subEntry.set(entry.key, entry.value);
					if(!filter.getBoolean(subEntry)) return false;
				}
				return true;
			}
			
			@Override
			public <E> E reduce(E identity, BiFunction<E, Double2CharMap.Entry, E> operator) {
				Objects.requireNonNull(operator);
				E state = identity;
				for(Double2CharRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					state = operator.apply(state, new BasicEntry(entry.key, entry.value));
				}
				return state;
			}
			
			@Override
			public Double2CharMap.Entry reduce(ObjectObjectUnaryOperator<Double2CharMap.Entry, Double2CharMap.Entry> operator) {
				Objects.requireNonNull(operator);
				Double2CharMap.Entry state = null;
				boolean empty = true;
				for(Double2CharRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					if(empty) {
						empty = false;
						state = new BasicEntry(entry.key, entry.value);
						continue;
					}
					state = operator.apply(state, new BasicEntry(entry.key, entry.value));
				}
				return state;
			}
			
			@Override
			public Double2CharMap.Entry findFirst(Object2BooleanFunction<Double2CharMap.Entry> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return null;
				BasicEntry subEntry = new BasicEntry();
				for(Double2CharRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					subEntry.set(entry.key, entry.value);
					if(filter.getBoolean(subEntry)) return subEntry;
				}
				return null;
			}
			
			@Override
			public int count(Object2BooleanFunction<Double2CharMap.Entry> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return 0;
				int result = 0;
				BasicEntry subEntry = new BasicEntry();
				for(Double2CharRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					subEntry.set(entry.key, entry.value);
					if(filter.getBoolean(subEntry)) result++;
				}
				return result;
			}
		}
		
		final class SubMapValues extends AbstractCharCollection {
			@Override
			public boolean add(char o) { throw new UnsupportedOperationException(); }
			
			@Override
			public boolean contains(char e) {
				return containsValue(e);
			}
			
			@Override
			public CharIterator iterator() { return valueIterator(); }
			
			@Override
			public int size() {
				return NavigableSubMap.this.size();
			}
			
			@Override
			public void clear() {
				NavigableSubMap.this.clear();
			}
			
			@Override
			public void forEach(CharConsumer action) {
				Objects.requireNonNull(action);
				for(Double2CharRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					action.accept(entry.value);
			}
			
			@Override
			public <E> void forEach(E input, ObjectCharConsumer<E> action) {
				Objects.requireNonNull(action);
				for(Double2CharRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					action.accept(input, entry.value);
			}
			
			@Override
			public boolean matchesAny(Char2BooleanFunction filter) {
				Objects.requireNonNull(filter);
				for(Double2CharRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					if(filter.get(entry.value)) return true;
				return false;
			}
			
			@Override
			public boolean matchesNone(Char2BooleanFunction filter) {
				Objects.requireNonNull(filter);
				for(Double2CharRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					if(filter.get(entry.value)) return false;
				return true;
			}
			
			@Override
			public boolean matchesAll(Char2BooleanFunction filter) {
				Objects.requireNonNull(filter);
				for(Double2CharRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					if(!filter.get(entry.value)) return false;
				return true;
			}
			
			@Override
			public char reduce(char identity, CharCharUnaryOperator operator) {
				Objects.requireNonNull(operator);
				char state = identity;
				for(Double2CharRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					state = operator.applyAsChar(state, entry.value);
				return state;
			}
			
			@Override
			public char reduce(CharCharUnaryOperator operator) {
				Objects.requireNonNull(operator);
				char state = (char)0;
				boolean empty = true;
				for(Double2CharRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					if(empty) {
						empty = false;
						state = entry.value;
						continue;
					}
					state = operator.applyAsChar(state, entry.value);
				}
				return state;
			}
			
			@Override
			public char findFirst(Char2BooleanFunction filter) {
				Objects.requireNonNull(filter);
				for(Double2CharRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					if(filter.get(entry.value)) return entry.value;
				return (char)0;
			}
			
			@Override
			public int count(Char2BooleanFunction filter) {
				Objects.requireNonNull(filter);
				int result = 0;
				for(Double2CharRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					if(filter.get(entry.value)) result++;
				return result;
			}
		}
		
		final class SubMapEntrySetIterator extends SubMapEntryIterator implements ObjectListIterator<Double2CharMap.Entry> {
			public SubMapEntrySetIterator(boolean descending) {
				super(descending);
			}
			
			public SubMapEntrySetIterator(Double2CharRBTreeMap.Entry entry, Double2CharRBTreeMap.Entry lowerFence, Double2CharRBTreeMap.Entry upperFence) {
				super(entry, lowerFence, upperFence);
			}
			
			@Override
			public Double2CharMap.Entry next() { return nextEntry(); }
			@Override
			public Double2CharMap.Entry previous() { return previousEntry(); }
			
			@Override
			public void set(Double2CharMap.Entry e) { throw new UnsupportedOperationException(); }
			@Override
			public void add(Double2CharMap.Entry e) { throw new UnsupportedOperationException(); }
		}
		
		final class SubMapKeyIterator extends SubMapEntryIterator implements DoubleListIterator {
			public SubMapKeyIterator(boolean descending) {
				super(descending);
			}
			
			public SubMapKeyIterator(Double2CharRBTreeMap.Entry entry, Double2CharRBTreeMap.Entry lowerFence, Double2CharRBTreeMap.Entry upperFence) {
				super(entry, lowerFence, upperFence);
			}
			
			@Override
			public double nextDouble() { return nextEntry().key; }
			@Override
			public double previousDouble() { return previousEntry().key; }
			
			@Override
			public void set(double e) { throw new UnsupportedOperationException(); }
			@Override
			public void add(double e) { throw new UnsupportedOperationException(); }
		}
		
		final class SubMapValueIterator extends SubMapEntryIterator implements CharListIterator {
			public SubMapValueIterator(boolean descending) {
				super(descending);
			}
			
			public SubMapValueIterator(Double2CharRBTreeMap.Entry entry, Double2CharRBTreeMap.Entry lowerFence, Double2CharRBTreeMap.Entry upperFence) {
				super(entry, lowerFence, upperFence);
			}
			
			@Override
			public char nextChar() { return nextEntry().value; }
			@Override
			public char previousChar() { return previousEntry().value; }
			
			@Override
			public void set(char e) { throw new UnsupportedOperationException(); }
			@Override
			public void add(char e) { throw new UnsupportedOperationException(); }
		}
		
		abstract class SubMapEntryIterator {
			Double lowerFence;
			Double upperFence;
			Double2CharRBTreeMap.Entry next;
			Double2CharRBTreeMap.Entry previous;
			Double2CharRBTreeMap.Entry current;
			int index = 0;
			
			public SubMapEntryIterator(boolean descending) {
				this(descending ? findHighest() : findLowest(), fromStart ? null : findLowest(), toEnd ? null : findHighest());
			}
			
			public SubMapEntryIterator(Double2CharRBTreeMap.Entry entry, Double2CharRBTreeMap.Entry lowerFence, Double2CharRBTreeMap.Entry upperFence) {
				next = entry;
				previous = entry.previous();
				this.lowerFence = lowerFence != null ? Double.valueOf(lowerFence.key) : null;
				this.upperFence = upperFence != null ? Double.valueOf(upperFence.key) : null;
			}
			
			public boolean hasNext() {
				return next != null && (upperFence == null || Double.doubleToLongBits(next.key) == Double.doubleToLongBits(upperFence.doubleValue()));
			}
			
			public boolean hasPrevious() {
				return previous != null && (lowerFence == null || Double.doubleToLongBits(next.key) == Double.doubleToLongBits(lowerFence.doubleValue()));
			}
			
			public int nextIndex() {
				return index;
			}
			
			public int previousIndex() {
				return index - 1;
			}
			
			protected void updateNext() {
				next = current.next();
			}
			
			protected void updatePrevious() {
				previous = current.previous();
			}
			
			public Double2CharRBTreeMap.Entry nextEntry() {
				if(!hasNext()) throw new NoSuchElementException();
				current = previous = next;
				updateNext();
				index++;
				return current;
			}
			
			public Double2CharRBTreeMap.Entry previousEntry() {
				if(!hasPrevious()) throw new NoSuchElementException();
				current = next = previous;
				updatePrevious();
				index--;
				return current;
			}
			
			public void remove() {
				if(current == null) throw new IllegalStateException();
				if(current == previous) index--;
				updateNext();
				updatePrevious();
				if(current.needsSuccessor()) next = current;
				m.removeNode(current);
				current = null;
			}
		}
	}
	
	private class EntrySet extends AbstractObjectSet<Double2CharMap.Entry> {
		
		@Override
		public ObjectIterator<Double2CharMap.Entry> iterator() {
			return new EntryIterator(false);
		}
		
		@Override
		@Deprecated
		public boolean contains(Object o) {
			if(o instanceof Map.Entry) {
				if(o instanceof Double2CharMap.Entry) {
					Double2CharMap.Entry entry = (Double2CharMap.Entry)o;
					if(entry.getKey() == null && comparator() == null) return false;
					Double2CharRBTreeMap.Entry subEntry = Double2CharRBTreeMap.this.findNode(entry.getDoubleKey());
					if(subEntry != null) return entry.getCharValue() == subEntry.value;
				}
				else {
					Map.Entry<?, ?> entry = (Map.Entry<?, ?>)o;
					if(entry.getKey() == null && comparator() == null) return false;
					if(!(entry.getKey() instanceof Double)) return false;
					Double2CharRBTreeMap.Entry subEntry = Double2CharRBTreeMap.this.findNode((Double)entry.getKey());
					if(subEntry != null) return Objects.equals(entry.getValue(), Character.valueOf(subEntry.value));
				}
			}
			return false;
		}
		
		@Override
		@Deprecated
		public boolean remove(Object o) {
			if(o instanceof Map.Entry) {
				if(o instanceof Double2CharMap.Entry) {
					Double2CharMap.Entry entry = (Double2CharMap.Entry)o;
					return Double2CharRBTreeMap.this.remove(entry.getDoubleKey(), entry.getCharValue());
				}
				Map.Entry<?, ?> entry = (Map.Entry<?, ?>)o;
				return Double2CharRBTreeMap.this.remove(entry.getKey(), entry.getValue());
			}
			return false;
		}
		
		@Override
		public int size() {
			return Double2CharRBTreeMap.this.size();
		}
		
		@Override
		public void clear() {
			Double2CharRBTreeMap.this.clear();
		}
		
		@Override
		public void forEach(Consumer<? super Double2CharMap.Entry> action) {
			Objects.requireNonNull(action);
			for(Double2CharRBTreeMap.Entry entry = first;entry != null;entry = entry.next())
				action.accept(new BasicEntry(entry.key, entry.value));
		}
		
		@Override
		public <E> void forEach(E input, ObjectObjectConsumer<E, Double2CharMap.Entry> action) {
			Objects.requireNonNull(action);
			for(Double2CharRBTreeMap.Entry entry = first;entry != null;entry = entry.next())
				action.accept(input, new BasicEntry(entry.key, entry.value));
		}
		
		@Override
		public boolean matchesAny(Object2BooleanFunction<Double2CharMap.Entry> filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return false;
			BasicEntry subEntry = new BasicEntry();
			for(Double2CharRBTreeMap.Entry entry = first;entry != null;entry = entry.next()) {
				subEntry.set(entry.key, entry.value);
				if(filter.getBoolean(subEntry)) return true;
			}
			return false;
		}
		
		@Override
		public boolean matchesNone(Object2BooleanFunction<Double2CharMap.Entry> filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return true;
			BasicEntry subEntry = new BasicEntry();
			for(Double2CharRBTreeMap.Entry entry = first;entry != null;entry = entry.next()) {
				subEntry.set(entry.key, entry.value);
				if(filter.getBoolean(subEntry)) return false;
			}
			return true;
		}
		
		@Override
		public boolean matchesAll(Object2BooleanFunction<Double2CharMap.Entry> filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return true;
			BasicEntry subEntry = new BasicEntry();
			for(Double2CharRBTreeMap.Entry entry = first;entry != null;entry = entry.next()) {
				subEntry.set(entry.key, entry.value);
				if(!filter.getBoolean(subEntry)) return false;
			}
			return true;
		}
		
		@Override
		public <E> E reduce(E identity, BiFunction<E, Double2CharMap.Entry, E> operator) {
			Objects.requireNonNull(operator);
			E state = identity;
			for(Double2CharRBTreeMap.Entry entry = first;entry != null;entry = entry.next()) {
				state = operator.apply(state, new BasicEntry(entry.key, entry.value));
			}
			return state;
		}
		
		@Override
		public Double2CharMap.Entry reduce(ObjectObjectUnaryOperator<Double2CharMap.Entry, Double2CharMap.Entry> operator) {
			Objects.requireNonNull(operator);
			Double2CharMap.Entry state = null;
			boolean empty = true;
			for(Double2CharRBTreeMap.Entry entry = first;entry != null;entry = entry.next()) {
				if(empty) {
					empty = false;
					state = new BasicEntry(entry.key, entry.value);
					continue;
				}
				state = operator.apply(state, new BasicEntry(entry.key, entry.value));
			}
			return state;
		}
		
		@Override
		public Double2CharMap.Entry findFirst(Object2BooleanFunction<Double2CharMap.Entry> filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return null;
			BasicEntry subEntry = new BasicEntry();
			for(Double2CharRBTreeMap.Entry entry = first;entry != null;entry = entry.next()) {
				subEntry.set(entry.key, entry.value);
				if(filter.getBoolean(subEntry)) return subEntry;
			}
			return null;
		}
		
		@Override
		public int count(Object2BooleanFunction<Double2CharMap.Entry> filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return 0;
			int result = 0;
			BasicEntry subEntry = new BasicEntry();
			for(Double2CharRBTreeMap.Entry entry = first;entry != null;entry = entry.next()) {
				subEntry.set(entry.key, entry.value);
				if(filter.getBoolean(subEntry)) result++;
			}
			return result;
		}
	}
	
	static final class KeySet extends AbstractDoubleSet implements DoubleNavigableSet {
		Double2CharNavigableMap m;
		
		KeySet(Double2CharNavigableMap m) {
			this.m = m;
		}
		
		@Override
		public void setDefaultMaxValue(double e) { m.setDefaultMaxValue(e); }
		@Override
		public double getDefaultMaxValue() { return m.getDefaultMaxValue(); }
		@Override
		public void setDefaultMinValue(double e) { m.setDefaultMinValue(e); }
		@Override
		public double getDefaultMinValue() { return m.getDefaultMinValue(); }
		@Override
		public boolean add(double o) { throw new UnsupportedOperationException(); }
		@Override
		public boolean addAndMoveToFirst(double o) { throw new UnsupportedOperationException(); }
		@Override
		public boolean addAndMoveToLast(double o) { throw new UnsupportedOperationException(); }
		@Override
		public boolean moveToFirst(double o) { throw new UnsupportedOperationException(); }
		@Override
		public boolean moveToLast(double o) { throw new UnsupportedOperationException(); }
		@Override
		public DoubleComparator comparator() { return m.comparator(); }
		@Override
		public double lower(double e) { return m.lowerKey(e); }
		@Override
		public double floor(double e) { return m.floorKey(e); }
		@Override
		public double ceiling(double e) { return m.ceilingKey(e); }
		@Override
		public double higher(double e) { return m.higherKey(e); }
		@Override
		public double firstDouble() { return m.firstDoubleKey(); }
		@Override
		public double pollFirstDouble() { return m.pollFirstDoubleKey(); }
		@Override
		public double lastDouble() { return m.lastDoubleKey(); }
		@Override
		public double pollLastDouble() { return m.pollLastDoubleKey(); }		
		@Override
		public boolean remove(double o) { 
			int oldSize = m.size();
			m.remove(o); 
			return oldSize != m.size();
		}
		
		@Override
		public KeySet copy() { throw new UnsupportedOperationException(); }
		
		@Override
		public DoubleNavigableSet subSet(double fromElement, boolean fromInclusive, double toElement, boolean toInclusive) {
			return new KeySet(m.subMap(fromElement, fromInclusive, toElement, toInclusive));
		}
		
		@Override
		public DoubleNavigableSet headSet(double toElement, boolean inclusive) {
			return new KeySet(m.headMap(toElement, inclusive));
		}
		
		@Override
		public DoubleNavigableSet tailSet(double fromElement, boolean inclusive) {
			return new KeySet(m.tailMap(fromElement, inclusive));
		}
		
		@Override
		public DoubleNavigableSet descendingSet() {
			return new KeySet(m.descendingMap());
		}
		
		@Override
		public DoubleBidirectionalIterator iterator() {
			if(m instanceof Double2CharRBTreeMap) return ((Double2CharRBTreeMap)m).keyIterator(false);
			return ((NavigableSubMap)m).keyIterator(false);
		}
		
		@Override
		public DoubleBidirectionalIterator iterator(double fromElement) {
			if(m instanceof Double2CharRBTreeMap) return ((Double2CharRBTreeMap)m).keyIterator(fromElement);
			return ((NavigableSubMap)m).keyIterator(fromElement);
		}
		
		@Override
		public DoubleBidirectionalIterator descendingIterator() {
			if(m instanceof Double2CharRBTreeMap) return ((Double2CharRBTreeMap)m).keyIterator(true);
			return ((NavigableSubMap)m).keyIterator(true);
		}
		
		@Override
		public int size() {
			return m.size();
		}
		
		@Override
		public void clear() {
			m.clear();
		}
		
		protected Double2CharRBTreeMap.Entry start() {
			if(m instanceof Double2CharRBTreeMap) return ((Double2CharRBTreeMap)m).first;
			return ((NavigableSubMap)m).findLowest();
		}
		
		protected Double2CharRBTreeMap.Entry end() {
			if(m instanceof Double2CharRBTreeMap) return null;
			return ((NavigableSubMap)m).findLowest();
		}
		
		protected Double2CharRBTreeMap.Entry next(Double2CharRBTreeMap.Entry entry) {
			if(m instanceof Double2CharRBTreeMap) return entry.next();
			return ((NavigableSubMap)m).next(entry);
		}
		
		protected Double2CharRBTreeMap.Entry previous(Double2CharRBTreeMap.Entry entry) {
			if(m instanceof Double2CharRBTreeMap) return entry.previous();
			return ((NavigableSubMap)m).previous(entry);
		}
		
		@Override
		public void forEach(DoubleConsumer action) {
			Objects.requireNonNull(action);
			for(Double2CharRBTreeMap.Entry entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				action.accept(entry.key);
		}
		
		@Override
		public <E> void forEach(E input, ObjectDoubleConsumer<E> action) {
			Objects.requireNonNull(action);
			for(Double2CharRBTreeMap.Entry entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				action.accept(input, entry.key);
		}
		
		@Override
		public boolean matchesAny(Double2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(Double2CharRBTreeMap.Entry entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				if(filter.get(entry.key)) return true;
			return false;
		}
		
		@Override
		public boolean matchesNone(Double2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(Double2CharRBTreeMap.Entry entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				if(filter.get(entry.key)) return false;
			return true;
		}
		
		@Override
		public boolean matchesAll(Double2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(Double2CharRBTreeMap.Entry entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				if(!filter.get(entry.key)) return false;
			return true;
		}
		
		@Override
		public double reduce(double identity, DoubleDoubleUnaryOperator operator) {
			Objects.requireNonNull(operator);
			double state = identity;
			for(Double2CharRBTreeMap.Entry entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				state = operator.applyAsDouble(state, entry.key);
			return state;
		}
		
		@Override
		public double reduce(DoubleDoubleUnaryOperator operator) {
			Objects.requireNonNull(operator);
			double state = 0D;
			boolean empty = true;
			for(Double2CharRBTreeMap.Entry entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry)) {
				if(empty) {
					empty = false;
					state = entry.key;
					continue;
				}
				state = operator.applyAsDouble(state, entry.key);
			}
			return state;
		}
		
		@Override
		public double findFirst(Double2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(Double2CharRBTreeMap.Entry entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				if(filter.get(entry.key)) return entry.key;
			return 0D;
		}
		
		@Override
		public int count(Double2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			int result = 0;
			for(Double2CharRBTreeMap.Entry entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				if(filter.get(entry.key)) result++;
			return result;
		}
	}
	
	final class Values extends AbstractCharCollection {

		@Override
		public boolean add(char o) { throw new UnsupportedOperationException(); }
		
		@Override
		public boolean contains(char e) {
			return containsValue(e);
		}
		
		@Override
		public CharIterator iterator() { return new ValueIterator(false); }
		
		@Override
		public int size() {
			return Double2CharRBTreeMap.this.size();
		}
		
		@Override
		public void clear() {
			Double2CharRBTreeMap.this.clear();
		}
		
		@Override
		public void forEach(CharConsumer action) {
			Objects.requireNonNull(action);
			for(Double2CharRBTreeMap.Entry entry = first;entry != null;entry = entry.next())
				action.accept(entry.value);
		}
		
		@Override
		public <E> void forEach(E input, ObjectCharConsumer<E> action) {
			Objects.requireNonNull(action);
			for(Double2CharRBTreeMap.Entry entry = first;entry != null;entry = entry.next())
				action.accept(input, entry.value);
		}
		
		@Override
		public boolean matchesAny(Char2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(Double2CharRBTreeMap.Entry entry = first;entry != null;entry = entry.next())
				if(filter.get(entry.value)) return true;
			return false;
		}
		
		@Override
		public boolean matchesNone(Char2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(Double2CharRBTreeMap.Entry entry = first;entry != null;entry = entry.next())
				if(filter.get(entry.value)) return false;
			return true;
		}
		
		@Override
		public boolean matchesAll(Char2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(Double2CharRBTreeMap.Entry entry = first;entry != null;entry = entry.next())
				if(!filter.get(entry.value)) return false;
			return true;
		}
		
		@Override
		public char reduce(char identity, CharCharUnaryOperator operator) {
			Objects.requireNonNull(operator);
			char state = identity;
			for(Double2CharRBTreeMap.Entry entry = first;entry != null;entry = entry.next())
				state = operator.applyAsChar(state, entry.value);
			return state;
		}
		
		@Override
		public char reduce(CharCharUnaryOperator operator) {
			Objects.requireNonNull(operator);
			char state = (char)0;
			boolean empty = true;
			for(Double2CharRBTreeMap.Entry entry = first;entry != null;entry = entry.next()) {
				if(empty) {
					empty = false;
					state = entry.value;
					continue;
				}
				state = operator.applyAsChar(state, entry.value);
			}
			return state;
		}
		
		@Override
		public char findFirst(Char2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(Double2CharRBTreeMap.Entry entry = first;entry != null;entry = entry.next())
				if(filter.get(entry.value)) return entry.value;
			return (char)0;
		}
		
		@Override
		public int count(Char2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			int result = 0;
			for(Double2CharRBTreeMap.Entry entry = first;entry != null;entry = entry.next())
				if(filter.get(entry.value)) result++;
			return result;
		}
	}
	
	final class EntryIterator extends MapEntryIterator implements ObjectListIterator<Double2CharMap.Entry> {
		
		public EntryIterator(boolean descending) {
			super(descending);
		}
		
		public EntryIterator(Entry entry) {
			super(entry);
		}
		
		@Override
		public Double2CharMap.Entry next() {
			return nextEntry();
		}
		
		@Override
		public Double2CharMap.Entry previous() {
			return previousEntry();
		}
		
		@Override
		public void set(Double2CharMap.Entry e) { throw new UnsupportedOperationException(); }
		
		@Override
		public void add(Double2CharMap.Entry e) { throw new UnsupportedOperationException(); }
	}
	
	final class KeyIterator extends MapEntryIterator implements DoubleListIterator {
		
		public KeyIterator(boolean descending) {
			super(descending);
		}
		
		public KeyIterator(Entry entry) {
			super(entry);
		}
		
		@Override
		public double previousDouble() { return previousEntry().key; }
		@Override
		public double nextDouble() { return nextEntry().key; }
		
		@Override
		public void set(double e) { throw new UnsupportedOperationException(); }
		@Override
		public void add(double e) { throw new UnsupportedOperationException(); }
	}
	
	final class ValueIterator extends MapEntryIterator implements CharListIterator {
		
		public ValueIterator(boolean descending) {
			super(descending);
		}
		
		public ValueIterator(Entry entry) {
			super(entry);
		}
		
		@Override
		public char previousChar() {
			return previousEntry().value;
		}
		
		@Override
		public char nextChar() {
			return nextEntry().value;
		}
		
		@Override
		public void set(char e) { throw new UnsupportedOperationException(); }
		
		@Override
		public void add(char e) { throw new UnsupportedOperationException(); }
	}
	
	abstract class MapEntryIterator {
		Entry next;
		Entry previous;
		Entry current;
		int index = 0;
		
		public MapEntryIterator(boolean descending) {
			if(descending) previous = last;
			else next = first;
		}
		
		public MapEntryIterator(Entry entry) {
			next = entry;
			previous = entry.previous();
		}
		
		public boolean hasNext() {
			return next != null;
		}
		
		public boolean hasPrevious() {
			return previous != null;
		}
		
		public int nextIndex() {
			return index;
		}
		
		public int previousIndex() {
			return index - 1;
		}
		
		protected void updateNext() {
			next = current.next();
		}
		
		protected void updatePrevious() {
			previous = current.previous();
		}
		
		public Entry nextEntry() {
			if(!hasNext()) throw new NoSuchElementException();
			current = previous = next;
			updateNext();
			index++;
			return current;
		}
		
		public Entry previousEntry() {
			if(!hasPrevious()) throw new NoSuchElementException();
			current = next = previous;
			updatePrevious();
			index--;
			return current;
		}
		
		public void remove() {
			if(current == null) throw new IllegalStateException();
			if(current == previous) index--;
			updateNext();
			updatePrevious();
			if(current.needsSuccessor()) next = current;
			removeNode(current);
			current = null;
		}
	}
	
	private static final class Entry implements Double2CharMap.Entry
	{
		static final int BLACK = 1;
		
		double key;
		char value;
		int state;
		Entry parent;
		Entry left;
		Entry right;
		
		Entry(double key, char value, Entry parent) {
			this.key = key;
			this.value = value;
			this.parent = parent;
		}
		
		Entry copy() {
			Entry entry = new Entry(key, value, null);
			entry.state = state;
			if(left != null) {
				Entry newLeft = left.copy();
				entry.left = newLeft;
				newLeft.parent = entry;
			}
			if(right != null) {
				Entry newRight = right.copy();
				entry.right = newRight;
				newRight.parent = entry;
			}
			return entry;
		}
		
		@Override
		public double getDoubleKey() {
			return key;
		}
		
		@Override
		public char getCharValue() {
			return value;
		}
		
		@Override
		public char setValue(char value) {
			char oldValue = this.value;
			this.value = value;
			return oldValue;
		}
		
		char addTo(char value) {
			char oldValue = this.value;
			this.value += value;
			return oldValue;
		}
		
		@Override
		public boolean equals(Object obj) {
			if(obj instanceof Map.Entry) {
				if(obj instanceof Double2CharMap.Entry) {
					Double2CharMap.Entry entry = (Double2CharMap.Entry)obj;
					return Double.doubleToLongBits(key) == Double.doubleToLongBits(entry.getDoubleKey()) && value == entry.getCharValue();
				}
				Map.Entry<?, ?> entry = (Map.Entry<?, ?>)obj;
				Object otherKey = entry.getKey();
				if(otherKey == null) return false;
				Object otherValue = entry.getValue();
				return otherKey instanceof Double && otherValue instanceof Character && Double.doubleToLongBits(key) == Double.doubleToLongBits(((Double)otherKey).doubleValue()) && value == ((Character)otherValue).charValue();
			}
			return false;
		}
		
		@Override
		public int hashCode() {
			return Double.hashCode(key) ^ Character.hashCode(value);
		}
		
		@Override
		public String toString() {
			return Double.toString(key) + "=" + Character.toString(value);
		}
		
		boolean isBlack() {
			return (state & BLACK) != 0;
		}
		
		void setBlack(boolean value) {
			if(value) state |= BLACK;
			else state &= ~BLACK;
		}
		
		boolean needsSuccessor() { return left != null && right != null; }
		
		boolean replace(Entry entry) {
			if(entry != null) entry.parent = parent;
			if(parent != null) {
				if(parent.left == this) parent.left = entry;
				else parent.right = entry;
			}
			return parent == null;
		}
		
		Entry next() {
			if(right != null) {
				Entry parent = right;
				while(parent.left != null) parent = parent.left;
				return parent;
			}
			Entry parent = this.parent;
			Entry control = this;
			while(parent != null && control == parent.right) {
				control = parent;
				parent = parent.parent;
			}
			return parent;
		}
		
		Entry previous() {
			if(left != null) {
				Entry parent = left;
				while(parent.right != null) parent = parent.right;
				return parent;
			}
			Entry parent = this.parent;
			Entry control = this;
			while(parent != null && control == parent.left) {
				control = parent;
				parent = parent.parent;
			}
			return parent;
		}
	}
}