package speiger.src.collections.doubles.maps.impl.tree;

import java.util.Map;
import java.util.Objects;
import java.util.NoSuchElementException;
import java.util.function.Consumer;
import java.util.function.BiFunction;

import speiger.src.collections.doubles.collections.DoubleBidirectionalIterator;
import speiger.src.collections.doubles.functions.DoubleComparator;
import speiger.src.collections.doubles.functions.DoubleConsumer;
import speiger.src.collections.objects.functions.consumer.ObjectDoubleConsumer;
import speiger.src.collections.doubles.functions.function.Double2BooleanFunction;
import speiger.src.collections.doubles.functions.consumer.DoubleObjectConsumer;
import speiger.src.collections.doubles.functions.function.Double2ObjectFunction;
import speiger.src.collections.doubles.functions.function.DoubleObjectUnaryOperator;
import speiger.src.collections.doubles.functions.function.DoubleDoubleUnaryOperator;
import speiger.src.collections.doubles.lists.DoubleListIterator;
import speiger.src.collections.doubles.maps.abstracts.AbstractDouble2ObjectMap;
import speiger.src.collections.doubles.maps.interfaces.Double2ObjectMap;
import speiger.src.collections.doubles.maps.interfaces.Double2ObjectNavigableMap;
import speiger.src.collections.doubles.sets.AbstractDoubleSet;
import speiger.src.collections.doubles.sets.DoubleNavigableSet;
import speiger.src.collections.doubles.sets.DoubleSet;
import speiger.src.collections.doubles.sets.DoubleSortedSet;
import speiger.src.collections.doubles.utils.DoubleIterators;
import speiger.src.collections.doubles.utils.maps.Double2ObjectMaps;
import speiger.src.collections.objects.collections.AbstractObjectCollection;
import speiger.src.collections.objects.collections.ObjectCollection;
import speiger.src.collections.objects.collections.ObjectIterator;
import speiger.src.collections.objects.functions.ObjectSupplier;
import speiger.src.collections.objects.functions.function.ObjectObjectUnaryOperator;
import speiger.src.collections.objects.lists.ObjectListIterator;
import speiger.src.collections.objects.utils.ObjectIterators;

import speiger.src.collections.objects.functions.consumer.ObjectObjectConsumer;
import speiger.src.collections.objects.functions.function.Object2BooleanFunction;
import speiger.src.collections.objects.sets.AbstractObjectSet;
import speiger.src.collections.objects.sets.ObjectSet;

/**
 * A Simple Type Specific RB TreeMap implementation that reduces boxing/unboxing.
 * It is using a bit more memory then <a href="https://github.com/vigna/fastutil">FastUtil</a>,
 * but it saves a lot of Performance on the Optimized removal and iteration logic.
 * Which makes the implementation actually useable and does not get outperformed by Javas default implementation.
 * @param <V> the type of elements maintained by this Collection
 */
public class Double2ObjectRBTreeMap<V> extends AbstractDouble2ObjectMap<V> implements Double2ObjectNavigableMap<V>
{
	/** The center of the Tree */
	protected transient Entry<V> tree;
	/** The Lowest possible Node */
	protected transient Entry<V> first;
	/** The Highest possible Node */
	protected transient Entry<V> last;
	/** The amount of elements stored in the Map */
	protected int size = 0;
	/** The Sorter of the Tree */
	protected transient DoubleComparator comparator;
	
	/** the default return value for max searches */
	protected double defaultMaxNotFound = Double.MIN_VALUE;
	/** the default return value for min searches */
	protected double defaultMinNotFound = Double.MAX_VALUE;
	
	/** KeySet Cache */
	protected DoubleNavigableSet keySet;
	/** Values Cache */
	protected ObjectCollection<V> values;
	/** EntrySet Cache */
	protected ObjectSet<Double2ObjectMap.Entry<V>> entrySet;
	
	/**
	 * Default Constructor
	 */
	public Double2ObjectRBTreeMap() {
	}
	
	/**
	 * Constructor that allows to define the sorter
	 * @param comp the function that decides how the tree is sorted, can be null
	 */
	public Double2ObjectRBTreeMap(DoubleComparator comp) {
		comparator = comp;
	}
	
	/**
	 * Helper constructor that allow to create a map from boxed values (it will unbox them)
	 * @param keys the keys that should be put into the map
	 * @param values the values that should be put into the map.
	 * @throws IllegalStateException if the keys and values do not match in lenght
	 */
	public Double2ObjectRBTreeMap(Double[] keys, V[] values) {
		this(keys, values, null);
	}
	
	/**
	 * Helper constructor that has a custom sorter and allow to create a map from boxed values (it will unbox them)
	 * @param keys the keys that should be put into the map
	 * @param values the values that should be put into the map.
	 * @param comp the function that decides how the tree is sorted, can be null
	 * @throws IllegalStateException if the keys and values do not match in lenght
	 */
	public Double2ObjectRBTreeMap(Double[] keys, V[] values, DoubleComparator comp) {
		comparator = comp;
		if(keys.length != values.length) throw new IllegalStateException("Input Arrays are not equal size");
		for(int i = 0,m=keys.length;i<m;i++) put(keys[i].doubleValue(), values[i]);
	}
	
	/**
	 * Helper constructor that allow to create a map from unboxed values
	 * @param keys the keys that should be put into the map
	 * @param values the values that should be put into the map.
	 * @throws IllegalStateException if the keys and values do not match in lenght
	 */
	public Double2ObjectRBTreeMap(double[] keys, V[] values) {
		this(keys, values, null);
	}
	
	/**
	 * Helper constructor that has a custom sorter and allow to create a map from unboxed values
	 * @param keys the keys that should be put into the map
	 * @param values the values that should be put into the map.
	 * @param comp the function that decides how the tree is sorted, can be null
	 * @throws IllegalStateException if the keys and values do not match in lenght
	 */
	public Double2ObjectRBTreeMap(double[] keys, V[] values, DoubleComparator comp) {
		comparator = comp;
		if(keys.length != values.length) throw new IllegalStateException("Input Arrays are not equal size");
		for(int i = 0,m=keys.length;i<m;i++) put(keys[i], values[i]);
	}
	
	/**
	 * A Helper constructor that allows to create a Map with exactly the same values as the provided map.
	 * @param map the values that should be present in the map
	 */
	public Double2ObjectRBTreeMap(Map<? extends Double, ? extends V> map) {
		this(map, null);
	}
	
	/**
	 * A Helper constructor that has a custom sorter and allows to create a Map with exactly the same values as the provided map.
	 * @param map the values that should be present in the map
	 * @param comp the function that decides how the tree is sorted, can be null
	 */
	public Double2ObjectRBTreeMap(Map<? extends Double, ? extends V> map, DoubleComparator comp) {
		comparator = comp;
		putAll(map);
	}
	
	/**
	 * A Type Specific Helper function that allows to create a new Map with exactly the same values as the provided map.
	 * @param map the values that should be present in the map
 	 */
	public Double2ObjectRBTreeMap(Double2ObjectMap<V> map) {
		this(map, null);
	}
	
	/**
	 * A Type Specific Helper function that has a custom sorter and allows to create a new Map with exactly the same values as the provided map.
	 * @param map the values that should be present in the map
	 * @param comp the function that decides how the tree is sorted, can be null
 	 */
	public Double2ObjectRBTreeMap(Double2ObjectMap<V> map, DoubleComparator comp) {
		comparator = comp;
		putAll(map);
	}
	
	@Override
	public void setDefaultMaxValue(double value) { defaultMaxNotFound = value; }
	@Override
	public double getDefaultMaxValue() { return defaultMaxNotFound; }
	@Override
	public void setDefaultMinValue(double value) { defaultMinNotFound = value; }
	@Override
	public double getDefaultMinValue() { return defaultMinNotFound; }
	
	@Override
	public V put(double key, V value) {
		if(tree == null) {
			tree = first = last = new Entry<>(key, value, null);
			size++;
			return getDefaultReturnValue();
		}
		int compare = 0;
		Entry<V> parent = tree;
		while(true) {
			if((compare = compare(key, parent.key)) == 0) return parent.setValue(value);
			if(compare < 0) {
				if(parent.left == null) break;
				parent = parent.left;
			}
			else if(compare > 0) {
				if(parent.right == null) break;
				parent = parent.right;
			}
		}
		Entry<V> adding = new Entry<>(key, value, parent);
		if(compare < 0)  {
			parent.left = adding;
			if(parent == first)	first = adding;
		}
		else {
			parent.right = adding;
			if(parent == last) last = adding;
		}
		fixAfterInsertion(adding);
		size++;
		return getDefaultReturnValue();
	}
	
	@Override
	public V putIfAbsent(double key, V value) {

		if(tree == null) {
			tree = first = last = new Entry<>(key, value, null);
			size++;
			return getDefaultReturnValue();
		}
		int compare = 0;
		Entry<V> parent = tree;
		while(true) {
			if((compare = compare(key, parent.key)) == 0) return parent.value;
			if(compare < 0) {
				if(parent.left == null) break;
				parent = parent.left;
			}
			else if(compare > 0) {
				if(parent.right == null) break;
				parent = parent.right;
			}
		}
		Entry<V> adding = new Entry<>(key, value, parent);
		if(compare < 0)  {
			parent.left = adding;
			if(parent == first)	first = adding;
		}
		else  {
			parent.right = adding;
			if(parent == last) last = adding;
		}
		fixAfterInsertion(adding);
		size++;
		return getDefaultReturnValue();
	}
	
	@Override
	public V putAndMoveToFirst(double key, V value) { throw new UnsupportedOperationException(); }
	@Override
	public V putAndMoveToLast(double key, V value) { throw new UnsupportedOperationException(); }
	@Override
	public boolean moveToFirst(double key) { throw new UnsupportedOperationException(); }
	@Override
	public boolean moveToLast(double key) { throw new UnsupportedOperationException(); }
	@Override
	public V getAndMoveToFirst(double key) { throw new UnsupportedOperationException(); }
	@Override
	public V getAndMoveToLast(double key) { throw new UnsupportedOperationException(); }
	
	@Override
	public DoubleComparator comparator() { return comparator; }
	
	@Override
	public boolean containsKey(double key) {
		return findNode(key) != null;
	}
	
	@Override
	public V get(double key) {
		Entry<V> node = findNode(key);
		return node == null ? getDefaultReturnValue() : node.value;
	}
	
	@Override
	public V getOrDefault(double key, V defaultValue) {
		Entry<V> node = findNode(key);
		return node == null ? defaultValue : node.value;
	}
	
	@Override
	public double firstDoubleKey() {
		if(tree == null) throw new NoSuchElementException();
		return first.key;
	}
	
	@Override
	public double pollFirstDoubleKey() {
		if(tree == null) throw new NoSuchElementException();
		double result = first.key;
		removeNode(first);
		return result;
	}
	
	@Override
	public double lastDoubleKey() {
		if(tree == null) throw new NoSuchElementException();
		return last.key;
	}
	
	@Override
	public double pollLastDoubleKey() {
		if(tree == null) throw new NoSuchElementException();
		double result = last.key;
		removeNode(last);
		return result;
	}
	
	@Override
	public Double2ObjectMap.Entry<V> firstEntry() {
		if(tree == null) throw new NoSuchElementException();
		return new BasicEntry<>(first.key, first.value);
	}
	
	@Override
	public Double2ObjectMap.Entry<V> lastEntry() {
		if(tree == null) throw new NoSuchElementException();
		return new BasicEntry<>(last.key, last.value);
	}
	
	@Override
	public Double2ObjectMap.Entry<V> pollFirstEntry() {
		if(tree == null) throw new NoSuchElementException();
		BasicEntry<V> entry = new BasicEntry<>(first.key, first.value);
		removeNode(first);
		return entry;
	}
	
	@Override
	public Double2ObjectMap.Entry<V> pollLastEntry() {
		if(tree == null) throw new NoSuchElementException();
		BasicEntry<V> entry = new BasicEntry<>(last.key, last.value);
		removeNode(last);
		return entry;
	}
	
	@Override
	public V firstValue() {
		if(tree == null) throw new NoSuchElementException();
		return first.value;
	}
	
	@Override
	public V lastValue() {
		if(tree == null) throw new NoSuchElementException();
		return last.value;
	}
	
	@Override
	public V remove(double key) {
		Entry<V> entry = findNode(key);
		if(entry == null) return getDefaultReturnValue();
		V value = entry.value;
		removeNode(entry);
		return value;
	}
	
	@Override
	public V removeOrDefault(double key, V defaultValue) {
		Entry<V> entry = findNode(key);
		if(entry == null) return defaultValue;
		V value = entry.value;
		removeNode(entry);
		return value;
	}
	
	@Override
	public boolean remove(double key, V value) {
		Entry<V> entry = findNode(key);
		if(entry == null || entry.value != value) return false;
		removeNode(entry);
		return true;
	}
	
	@Override
	public boolean replace(double key, V oldValue, V newValue) {
		Entry<V> entry = findNode(key);
		if(entry == null || entry.value != oldValue) return false;
		entry.value = newValue;
		return true;
	}
	
	@Override
	public V replace(double key, V value) {
		Entry<V> entry = findNode(key);
		if(entry == null) return getDefaultReturnValue();
		V oldValue = entry.value;
		entry.value = value;
		return oldValue;
	}
	
	@Override
	public V compute(double key, DoubleObjectUnaryOperator<V> mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		Entry<V> entry = findNode(key);
		if(entry == null) {
			V newValue = mappingFunction.apply(key, getDefaultReturnValue());
			if(Objects.equals(newValue, getDefaultReturnValue())) return newValue;
			put(key, newValue);
			return newValue;
		}
		V newValue = mappingFunction.apply(key, entry.value);
		if(Objects.equals(newValue, getDefaultReturnValue())) {
			removeNode(entry);
			return newValue;
		}
		entry.value = newValue;
		return newValue;
	}
	
	@Override
	public V computeIfAbsent(double key, Double2ObjectFunction<V> mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		Entry<V> entry = findNode(key);
		if(entry == null) {
			V newValue = mappingFunction.get(key);
			if(Objects.equals(newValue, getDefaultReturnValue())) return newValue;
			put(key, newValue);
			return newValue;
		}
		if(Objects.equals(entry.value, getDefaultReturnValue())) {
			V newValue = mappingFunction.get(key);
			if(Objects.equals(newValue, getDefaultReturnValue())) return newValue;
			entry.value = newValue;
		}
		return entry.value;
	}
	
	@Override
	public V supplyIfAbsent(double key, ObjectSupplier<V> valueProvider) {
		Objects.requireNonNull(valueProvider);
		Entry<V> entry = findNode(key);
		if(entry == null) {
			V newValue = valueProvider.get();
			if(Objects.equals(newValue, getDefaultReturnValue())) return newValue;
			put(key, newValue);
			return newValue;
		}
		if(Objects.equals(entry.value, getDefaultReturnValue())) {
			V newValue = valueProvider.get();
			if(Objects.equals(newValue, getDefaultReturnValue())) return newValue;
			entry.value = newValue;
		}
		return entry.value;
	}
	
	@Override
	public V computeIfPresent(double key, DoubleObjectUnaryOperator<V> mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		Entry<V> entry = findNode(key);
		if(entry == null || Objects.equals(entry.value, getDefaultReturnValue())) return getDefaultReturnValue();
		V newValue = mappingFunction.apply(key, entry.value);
		if(Objects.equals(newValue, getDefaultReturnValue())) {
			removeNode(entry);
			return newValue;
		}
		entry.value = newValue;
		return newValue;
	}
	
	@Override
	public V merge(double key, V value, ObjectObjectUnaryOperator<V, V> mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		Entry<V> entry = findNode(key);
		V newValue = entry == null || Objects.equals(entry.value, getDefaultReturnValue()) ? value : mappingFunction.apply(entry.value, value);
		if(Objects.equals(newValue, getDefaultReturnValue())) {
			if(entry != null)
				removeNode(entry);
		}
		else if(entry == null) put(key, newValue);
		else entry.value = newValue;
		return newValue;
	}
	
	@Override
	public void mergeAll(Double2ObjectMap<V> m, ObjectObjectUnaryOperator<V, V> mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		for(Double2ObjectMap.Entry<V> entry : Double2ObjectMaps.fastIterable(m)) {
			double key = entry.getDoubleKey();
			Entry<V> subEntry = findNode(key);
			V newValue = subEntry == null || Objects.equals(subEntry.value, getDefaultReturnValue()) ? entry.getValue() : mappingFunction.apply(subEntry.value, entry.getValue());
			if(Objects.equals(newValue, getDefaultReturnValue())) {
				if(subEntry != null)
					removeNode(subEntry);
			}
			else if(subEntry == null) put(key, newValue);
			else subEntry.value = newValue;
		}
	}
	
	@Override
	public void forEach(DoubleObjectConsumer<V> action) {
		for(Entry<V> entry = first;entry != null;entry = entry.next())
			action.accept(entry.key, entry.value);
	}
	
	@Override
	public int size() { return size; }
	
	@Override
	public void clear() {
		size = 0;
		first = null;
		last = null;
		tree = null;
	}
	
	DoubleListIterator keyIterator(boolean descending) {
		DoubleListIterator iter = new KeyIterator(descending);
		return descending ? DoubleIterators.invert(iter) : iter;
	}
	
	DoubleListIterator keyIterator(double key) {
		return new KeyIterator(findNode(key));
	}
	
	public Double2ObjectRBTreeMap<V> copy() {
		Double2ObjectRBTreeMap<V> set = new Double2ObjectRBTreeMap<>();
		set.size = size;
		if(tree != null) {
			set.tree = tree.copy();
			Entry<V> lastFound = null;
			for(Entry<V> entry = tree;entry != null;entry = entry.left) lastFound = entry;
			set.first = lastFound;
			lastFound = null;
			for(Entry<V> entry = tree;entry != null;entry = entry.right) lastFound = entry;
			set.last = lastFound;
		}
		return set;
	}
	
	@Override
	public DoubleSortedSet keySet() {
		return navigableKeySet();
	}
	
	@Override
	public ObjectSet<Double2ObjectMap.Entry<V>> double2ObjectEntrySet() {
		if(entrySet == null) entrySet = new EntrySet();
		return entrySet;
	}
	
	@Override
	public ObjectCollection<V> values() {
		if(values == null) values = new Values();
		return values;
	}
	
	@Override
	public DoubleNavigableSet navigableKeySet() {
		if(keySet == null) keySet = new KeySet<>(this);
		return keySet;
	}
	
	@Override
	public Double2ObjectNavigableMap<V> descendingMap() {
		return new DescendingSubMap<>(this, true, 0D, true, true, 0D, true);
	}
	
	@Override
	public DoubleNavigableSet descendingKeySet() {
		return descendingMap().navigableKeySet();
	}
	
	@Override
	public Double2ObjectNavigableMap<V> subMap(double fromKey, boolean fromInclusive, double toKey, boolean toInclusive) {
		return new AscendingSubMap<>(this, false, fromKey, fromInclusive, false, toKey, toInclusive);
	}
	
	@Override
	public Double2ObjectNavigableMap<V> headMap(double toKey, boolean inclusive) {
		return new AscendingSubMap<>(this, true, 0D, true, false, toKey, inclusive);
	}
	
	@Override
	public Double2ObjectNavigableMap<V> tailMap(double fromKey, boolean inclusive) {
		return new AscendingSubMap<>(this, false, fromKey, inclusive, true, 0D, true);
	}
	
	@Override
	public double lowerKey(double e) {
		Entry<V> node = findLowerNode(e);
		return node != null ? node.key : getDefaultMinValue();
	}

	@Override
	public double floorKey(double e) {
		Entry<V> node = findFloorNode(e);
		return node != null ? node.key : getDefaultMinValue();
	}
	
	@Override
	public double higherKey(double e) {
		Entry<V> node = findHigherNode(e);
		return node != null ? node.key : getDefaultMaxValue();
	}

	@Override
	public double ceilingKey(double e) {
		Entry<V> node = findCeilingNode(e);
		return node != null ? node.key : getDefaultMaxValue();
	}
	
	@Override
	public Double2ObjectMap.Entry<V> lowerEntry(double key) {
		Entry<V> node = findLowerNode(key);
		return node != null ? new BasicEntry<>(node.key, node.value) : null;
	}
	
	@Override
	public Double2ObjectMap.Entry<V> higherEntry(double key) {
		Entry<V> node = findHigherNode(key);
		return node != null ? new BasicEntry<>(node.key, node.value) : null;
	}
	
	@Override
	public Double2ObjectMap.Entry<V> floorEntry(double key) {
		Entry<V> node = findFloorNode(key);
		return node != null ? new BasicEntry<>(node.key, node.value) : null;
	}
	
	@Override
	public Double2ObjectMap.Entry<V> ceilingEntry(double key) {
		Entry<V> node = findCeilingNode(key);
		return node != null ? new BasicEntry<>(node.key, node.value) : null;
	}
	
	protected Entry<V> findLowerNode(double key) {
		Entry<V> entry = tree;
		while(entry != null) {
			if(compare(key, entry.key) > 0) {
				if(entry.right != null) entry = entry.right;
				else return entry;
			}
			else {
				if(entry.left != null) entry = entry.left;
				else {
					Entry<V> parent = entry.parent;
					while(parent != null && parent.left == entry) {
						entry = parent;
						parent = parent.parent;
					}
					return parent;
				}
			}
		}
		return null;
	}
	
	protected Entry<V> findFloorNode(double key) {
		Entry<V> entry = tree;
		int compare;
		while(entry != null) {
			if((compare = compare(key, entry.key)) > 0) {
				if(entry.right == null) break;
				entry = entry.right;
				continue;
			}
			else if(compare < 0) {
				if(entry.left != null) entry = entry.left;
				else {
					Entry<V> parent = entry.parent;
					while(parent != null && parent.left == entry) {
						entry = parent;
						parent = parent.parent;
					}
					return parent;
				}
				continue;
			}
			break;
		}
		return entry;
	}
	
	protected Entry<V> findCeilingNode(double key) {
		Entry<V> entry = tree;
		int compare;
		while(entry != null) {
			if((compare = compare(key, entry.key)) < 0) {
				if(entry.left == null) break;
				entry = entry.left;
				continue;
			}
			else if(compare > 0) {
				if(entry.right != null) entry = entry.right;
				else {
					Entry<V> parent = entry.parent;
					while(parent != null && parent.right == entry) {
						entry = parent;
						parent = parent.parent;
					}
					return parent;
				}
				continue;
			}
			break;
		}
		return entry;
	}
	
	protected Entry<V> findHigherNode(double key) {
		Entry<V> entry = tree;
		while(entry != null) {
			if(compare(key, entry.key) < 0) {
				if(entry.left != null) entry = entry.left;
				else return entry;
			}
			else {
				if(entry.right != null) entry = entry.right;
				else {
					Entry<V> parent = entry.parent;
					while(parent != null && parent.right == entry) {
						entry = parent;
						parent = parent.parent;
					}
					return parent;
				}
			}
		}
		return null;
	}
	
	protected Entry<V> findNode(double key) {
		Entry<V> node = tree;
		int compare;
		while(node != null) {
			if((compare = compare(key, node.key)) == 0) return node;
			if(compare < 0) node = node.left;
			else node = node.right;
		}
		return null;
	}
	
	protected void removeNode(Entry<V> entry) {
		size--;
		if(entry.needsSuccessor()) {
			Entry<V> successor = entry.next();
			entry.key = successor.key;
			entry.value = successor.value;
			entry = successor;
		}
		Entry<V> replacement = entry.left != null ? entry.left : entry.right;
		if(replacement != null) {
			if(entry.replace(replacement)) tree = replacement;
			if(entry == first) first = replacement;
			if(entry == last) last = entry.right != null ? entry.right : replacement;
			entry.left = entry.right = entry.parent = null;
			if(entry.isBlack()) fixAfterDeletion(replacement);
		}
		else if(entry.parent == null) tree = first = last = null;
		else {
			if(entry.isBlack())
				fixAfterDeletion(entry);
			entry.replace(null);
			if(entry.parent != null) {
				Entry<V> parent = entry.parent;
				if(entry == first) first = parent.left != null ? parent.left : parent;
				if(entry == last) last = entry.right != null ? parent.right : parent;
			}
			entry.parent = null;
		}
	}
	
	protected void validate(double k) { compare(k, k); }
	protected int compare(double k, double v) { return comparator != null ? comparator.compare(k, v) : Double.compare(k, v);}
	protected static <V> boolean isBlack(Entry<V> p) { return p == null || p.isBlack(); }
	protected static <V> Entry<V> parentOf(Entry<V> p) { return (p == null ? null : p.parent); }
	protected static <V> void setBlack(Entry<V> p, boolean c) { if(p != null) p.setBlack(c); }
	protected static <V> Entry<V> leftOf(Entry<V> p) { return p == null ? null : p.left; }
	protected static <V> Entry<V> rightOf(Entry<V> p) { return (p == null) ? null : p.right; }
	
	protected void rotateLeft(Entry<V> entry) {
		if(entry != null) {
			Entry<V> right = entry.right;
			entry.right = right.left;
			if(right.left != null) right.left.parent = entry;
			right.parent = entry.parent;
			if(entry.parent == null) tree = right;
			else if(entry.parent.left == entry) entry.parent.left = right;
			else entry.parent.right = right;
			right.left = entry;
			entry.parent = right;
		}
	}
	
	protected void rotateRight(Entry<V> entry) {
		if(entry != null) {
			Entry<V> left = entry.left;
			entry.left = left.right;
			if(left.right != null) left.right.parent = entry;
			left.parent = entry.parent;
			if(entry.parent == null) tree = left;
			else if(entry.parent.right == entry) entry.parent.right = left;
			else entry.parent.left = left;
			left.right = entry;
			entry.parent = left;
		}
	}
	
	protected void fixAfterInsertion(Entry<V> entry) {
		entry.setBlack(false);
		while(entry != null && entry != tree && !entry.parent.isBlack()) {
			if(parentOf(entry) == leftOf(parentOf(parentOf(entry)))) {
				Entry<V> y = rightOf(parentOf(parentOf(entry)));
				if(!isBlack(y)) {
					setBlack(parentOf(entry), true);
					setBlack(y, true);
					setBlack(parentOf(parentOf(entry)), false);
					entry = parentOf(parentOf(entry));
				}
				else {
					if(entry == rightOf(parentOf(entry))) {
						entry = parentOf(entry);
						rotateLeft(entry);
					}
					setBlack(parentOf(entry), true);
					setBlack(parentOf(parentOf(entry)), false);
					rotateRight(parentOf(parentOf(entry)));
				}
			}
			else {
				Entry<V> y = leftOf(parentOf(parentOf(entry)));
				if(!isBlack(y)) {
					setBlack(parentOf(entry), true);
					setBlack(y, true);
					setBlack(parentOf(parentOf(entry)), false);
					entry = parentOf(parentOf(entry));
				}
				else {
					if(entry == leftOf(parentOf(entry))) {
						entry = parentOf(entry);
						rotateRight(entry);
					}
					setBlack(parentOf(entry), true);
					setBlack(parentOf(parentOf(entry)), false);
					rotateLeft(parentOf(parentOf(entry)));
				}
			}
		}
		tree.setBlack(true);
	}
	
	protected void fixAfterDeletion(Entry<V> entry) {
		while(entry != tree && isBlack(entry)) {
			if(entry == leftOf(parentOf(entry))) {
				Entry<V> sib = rightOf(parentOf(entry));
				if(!isBlack(sib)) {
					setBlack(sib, true);
					setBlack(parentOf(entry), false);
					rotateLeft(parentOf(entry));
					sib = rightOf(parentOf(entry));
				}
				if(isBlack(leftOf(sib)) && isBlack(rightOf(sib))) {
					setBlack(sib, false);
					entry = parentOf(entry);
				}
				else {
					if(isBlack(rightOf(sib))) {
						setBlack(leftOf(sib), true);
						setBlack(sib, false);
						rotateRight(sib);
						sib = rightOf(parentOf(entry));
					}
					setBlack(sib, isBlack(parentOf(entry)));
					setBlack(parentOf(entry), true);
					setBlack(rightOf(sib), true);
					rotateLeft(parentOf(entry));
					entry = tree;
				}
			}
			else {
				Entry<V> sib = leftOf(parentOf(entry));
				if(!isBlack(sib)) {
					setBlack(sib, true);
					setBlack(parentOf(entry), false);
					rotateRight(parentOf(entry));
					sib = leftOf(parentOf(entry));
				}
				if(isBlack(rightOf(sib)) && isBlack(leftOf(sib))) {
					setBlack(sib, false);
					entry = parentOf(entry);
				}
				else {
					if(isBlack(leftOf(sib))) {
						setBlack(rightOf(sib), true);
						setBlack(sib, false);
						rotateLeft(sib);
						sib = leftOf(parentOf(entry));
					}
					setBlack(sib, isBlack(parentOf(entry)));
					setBlack(parentOf(entry), true);
					setBlack(leftOf(sib), true);
					rotateRight(parentOf(entry));
					entry = tree;
				}
			}
		}
		setBlack(entry, true);
	}
	
	static final class AscendingSubMap<V> extends NavigableSubMap<V> {

		public AscendingSubMap(Double2ObjectRBTreeMap<V> m, boolean fromStart, double low, boolean loInclusive, boolean toEnd, double high, boolean hiInclusive) {
			super(m, fromStart, low, loInclusive, toEnd, high, hiInclusive);
		}
		
		@Override
		DoubleListIterator keyIterator(boolean descending) { return new SubMapKeyIterator(descending); }
		@Override
		DoubleListIterator keyIterator(double key) {
			Double2ObjectRBTreeMap.Entry<V> entry = m.findNode(key);
			return entry == null || !inClosedRange(key) ? null : new SubMapKeyIterator(entry, fromStart ? null : findLowest(), toEnd ? null : findHighest());
		}
		@Override
		ObjectListIterator<V> valueIterator() { return new SubMapValueIterator(false); }
		@Override
		ObjectListIterator<Double2ObjectMap.Entry<V>> entryIterator() { return new SubMapEntrySetIterator(false); }
		
		@Override
		public Double2ObjectNavigableMap<V> descendingMap() {
			if(descendingMap == null) descendingMap = new DescendingSubMap<>(m, fromStart, low, loInclusive, toEnd, high, hiInclusive);
			return descendingMap;
		}
		
		Double2ObjectRBTreeMap.Entry<V> next(Double2ObjectRBTreeMap.Entry<V> entry) {
			return entry.next();
		}
		
		Double2ObjectRBTreeMap.Entry<V> previous(Double2ObjectRBTreeMap.Entry<V> entry) {
			return entry.previous();
		}
		
		@Override
		public Double2ObjectNavigableMap<V> subMap(double fromKey, boolean fromInclusive, double toKey, boolean toInclusive) {
			if (!inRange(fromKey, fromInclusive)) throw new IllegalArgumentException("fromKey out of range");
			if (!inRange(toKey, toInclusive)) throw new IllegalArgumentException("toKey out of range");
			return new AscendingSubMap<>(m, false, fromKey, fromInclusive, false, toKey, toInclusive);
		}
		
		@Override
		public Double2ObjectNavigableMap<V> headMap(double toKey, boolean inclusive) {
			if (!inRange(toKey, inclusive)) throw new IllegalArgumentException("toKey out of range");
			return new AscendingSubMap<>(m, fromStart, low, loInclusive, false, toKey, inclusive);
		}
		
		@Override
		public Double2ObjectNavigableMap<V> tailMap(double fromKey, boolean inclusive) {
			if (!inRange(fromKey, inclusive)) throw new IllegalArgumentException("fromKey out of range");
			return new AscendingSubMap<>(m, false, fromKey, inclusive, toEnd, high, hiInclusive);
		}
	}
	
	static final class DescendingSubMap<V> extends NavigableSubMap<V> {

		public DescendingSubMap(Double2ObjectRBTreeMap<V> m, boolean fromStart, double low, boolean loInclusive, boolean toEnd, double high, boolean hiInclusive) {
			super(m, fromStart, low, loInclusive, toEnd, high, hiInclusive);
		}
		
		@Override
		public double firstDoubleKey() { return super.lastDoubleKey(); }
		
		@Override
		public double pollFirstDoubleKey() { return super.pollLastDoubleKey(); }
		
		@Override
		public double lastDoubleKey() { return super.firstDoubleKey(); }
		
		@Override
		public double pollLastDoubleKey() { return super.pollFirstDoubleKey(); }
		
		@Override
		public V firstValue() { return super.lastValue(); }
		
		@Override
		public V lastValue() { return super.firstValue(); }
		
		@Override
		public Double2ObjectMap.Entry<V> firstEntry() { return super.lastEntry(); }
		
		@Override
		public Double2ObjectMap.Entry<V> lastEntry() { return super.firstEntry(); }
		
		@Override
		public Double2ObjectMap.Entry<V> pollFirstEntry() { return super.pollLastEntry(); }
		
		@Override
		public Double2ObjectMap.Entry<V> pollLastEntry() { return super.pollFirstEntry(); }
		
		@Override
		public double lowerKey(double e) { return super.higherKey(e); }
		
		@Override
		public double floorKey(double e) { return super.ceilingKey(e); }
		
		@Override
		public double ceilingKey(double e) { return super.floorKey(e); }
		
		@Override
		public double higherKey(double e) { return super.lowerKey(e); }
		
		@Override
		public Double2ObjectMap.Entry<V> lowerEntry(double e) { return super.higherEntry(e); }
		
		@Override
		public Double2ObjectMap.Entry<V> floorEntry(double e) { return super.ceilingEntry(e); }
		
		@Override
		public Double2ObjectMap.Entry<V> ceilingEntry(double e) { return super.floorEntry(e); }
		
		@Override
		public Double2ObjectMap.Entry<V> higherEntry(double e) { return super.lowerEntry(e); }
		
		protected Double2ObjectRBTreeMap.Entry<V> findLowest() {
			if(fromStart) return m.first;
			Double2ObjectRBTreeMap.Entry<V> entry = loInclusive ? m.findCeilingNode(low) : m.findHigherNode(low);
			return entry == null || tooHigh(entry.key) ? null : entry;
		}

		protected Double2ObjectRBTreeMap.Entry<V> findHighest() {
			if(toEnd) return m.last;
			Double2ObjectRBTreeMap.Entry<V> entry = hiInclusive ? m.findFloorNode(high) : m.findLowerNode(high);
			return entry == null || tooLow(entry.key) ? null : entry;
		}
		
		@Override
		DoubleListIterator keyIterator(boolean descending) {
			DoubleListIterator iter = new SubMapKeyIterator(!descending);
			return descending ? iter : DoubleIterators.invert(iter);
		}
		
		@Override
		DoubleListIterator keyIterator(double key) {
			Double2ObjectRBTreeMap.Entry<V> entry = m.findNode(key);
			return entry == null || !inClosedRange(key) ? null : DoubleIterators.invert(new SubMapKeyIterator(entry, fromStart ? null : findLowest(), toEnd ? null : findHighest()));
		}
		
		Double2ObjectRBTreeMap.Entry<V> next(Double2ObjectRBTreeMap.Entry<V> entry) {
			return entry.previous();
		}
		
		Double2ObjectRBTreeMap.Entry<V> previous(Double2ObjectRBTreeMap.Entry<V> entry) {
			return entry.next();
		}
		
		@Override
		public DoubleComparator comparator() { return m.comparator() == null ? null : m.comparator().reversed(); }
		
		@Override
		ObjectListIterator<V> valueIterator() { return ObjectIterators.invert(new SubMapValueIterator(true)); }
		
		@Override
		ObjectListIterator<Double2ObjectMap.Entry<V>> entryIterator() { return ObjectIterators.invert(new SubMapEntrySetIterator(true)); }
		
		@Override
		public Double2ObjectNavigableMap<V> descendingMap() {
			if(descendingMap == null) descendingMap = new AscendingSubMap<>(m, fromStart, low, loInclusive, toEnd, high, hiInclusive);
			return descendingMap;
		}

		@Override
		public Double2ObjectNavigableMap<V> subMap(double fromKey, boolean fromInclusive, double toKey, boolean toInclusive) {
			if (!inRange(fromKey, fromInclusive)) throw new IllegalArgumentException("fromKey out of range");
			if (!inRange(toKey, toInclusive)) throw new IllegalArgumentException("toKey out of range");
			return new DescendingSubMap<>(m, false, fromKey, fromInclusive, false, toKey, toInclusive);
		}
		
		@Override
		public Double2ObjectNavigableMap<V> headMap(double toKey, boolean inclusive) {
			if (!inRange(toKey, inclusive)) throw new IllegalArgumentException("toKey out of range");
			return new DescendingSubMap<>(m, fromStart, low, loInclusive, false, toKey, inclusive);
		}
		
		@Override
		public Double2ObjectNavigableMap<V> tailMap(double fromKey, boolean inclusive) {
			if (!inRange(fromKey, inclusive)) throw new IllegalArgumentException("fromKey out of range");
			return new DescendingSubMap<>(m, false, fromKey, inclusive, toEnd, high, hiInclusive);
		}
	}
	
	static abstract class NavigableSubMap<V> extends AbstractDouble2ObjectMap<V> implements Double2ObjectNavigableMap<V> {
		final Double2ObjectRBTreeMap<V> m;
		final double low;
		final double high;
		final boolean fromStart;
		final boolean toEnd;
		final boolean loInclusive;
		final boolean hiInclusive;
		transient Double2ObjectNavigableMap<V> descendingMap;
		transient SubMapEntrySet entrySet;
		transient KeySet<V> keySet;
		transient SubMapValues values;
		
		NavigableSubMap(Double2ObjectRBTreeMap<V> m, boolean fromStart, double low, boolean loInclusive, boolean toEnd, double high, boolean hiInclusive) {
			this.m = m;
			this.low = low;
			this.high = high;
			this.fromStart = fromStart;
			this.toEnd = toEnd;
			this.loInclusive = loInclusive;
			this.hiInclusive = hiInclusive;
		}
		
		abstract DoubleListIterator keyIterator(boolean descending);
		abstract DoubleListIterator keyIterator(double key);
		abstract ObjectListIterator<V> valueIterator();
		abstract ObjectListIterator<Double2ObjectMap.Entry<V>> entryIterator();
		abstract Double2ObjectRBTreeMap.Entry<V> next(Double2ObjectRBTreeMap.Entry<V> entry);
		abstract Double2ObjectRBTreeMap.Entry<V> previous(Double2ObjectRBTreeMap.Entry<V> entry);
		
		@Override
		public NavigableSubMap<V> copy() { throw new UnsupportedOperationException(); }
		@Override
		public abstract Double2ObjectNavigableMap<V> descendingMap();
		@Override
		public abstract Double2ObjectNavigableMap<V> subMap(double fromKey, boolean fromInclusive, double toKey, boolean toInclusive);
		@Override
		public abstract Double2ObjectNavigableMap<V> headMap(double toKey, boolean inclusive);
		@Override
		public abstract Double2ObjectNavigableMap<V> tailMap(double fromKey, boolean inclusive);
		
		boolean tooLow(double key) { return !fromStart && (loInclusive ? m.compare(key, low) < 0 : m.compare(key, low) <= 0); }
		boolean tooHigh(double key) { return !toEnd && (hiInclusive ? m.compare(key, high) > 0 : m.compare(key, high) >= 0); }
		boolean inRange(double key) { return !tooLow(key) && !tooHigh(key); }
		boolean inClosedRange(double key) { return (fromStart || m.compare(key, low) >= 0) && (toEnd || m.compare(high, key) >= 0); }
		boolean inRange(double key, boolean inclusive) { return inclusive ? inRange(key) : inClosedRange(key); }
		
		@Override
		public void setDefaultMaxValue(double e) { m.setDefaultMaxValue(e); }
		@Override
		public double getDefaultMaxValue() { return m.getDefaultMaxValue(); }
		@Override
		public void setDefaultMinValue(double e) { m.setDefaultMinValue(e); }
		@Override
		public double getDefaultMinValue() { return m.getDefaultMinValue(); }
		
		@Override
		public AbstractDouble2ObjectMap<V> setDefaultReturnValue(V v) { 
			m.setDefaultReturnValue(v);
			return this;
		}
		
		@Override
		public V getDefaultReturnValue() { return m.getDefaultReturnValue(); }
		
		@Override
		public V putAndMoveToFirst(double key, V value) { throw new UnsupportedOperationException(); }
		@Override
		public V putAndMoveToLast(double key, V value) { throw new UnsupportedOperationException(); }
		@Override
		public boolean moveToFirst(double key) { throw new UnsupportedOperationException(); }
		@Override
		public boolean moveToLast(double key) { throw new UnsupportedOperationException(); }
		@Override
		public V getAndMoveToFirst(double key) { throw new UnsupportedOperationException(); }
		@Override
		public V getAndMoveToLast(double key) { throw new UnsupportedOperationException(); }
		@Override
		public DoubleComparator comparator() { return m.comparator(); }
		@Override
		public boolean containsKey(double key) { return inRange(key) && m.containsKey(key); }
		
		@Override
		public double firstDoubleKey() { 
			Double2ObjectRBTreeMap.Entry<V> entry = findLowest();
			return entry == null ? getDefaultMaxValue() : entry.key;
		}
		
		@Override
		public double pollFirstDoubleKey() {
			if(fromStart) return m.pollFirstDoubleKey();
			Double2ObjectRBTreeMap.Entry<V> entry = loInclusive ? m.findCeilingNode(low) : m.findHigherNode(low);
			if(entry != null && !tooHigh(entry.key)) {
				double value = entry.key;
				m.removeNode(entry);
				return value;
			}
			return getDefaultMaxValue();
		}
		
		@Override
		public double lastDoubleKey() {
			Double2ObjectRBTreeMap.Entry<V> entry = findHighest();
			return entry == null ? getDefaultMinValue() : entry.key;	
		}
		
		@Override
		public double pollLastDoubleKey() {
			if(toEnd) return m.pollLastDoubleKey();
			Double2ObjectRBTreeMap.Entry<V> entry = hiInclusive ? m.findFloorNode(high) : m.findLowerNode(high);
			if(entry != null && !tooLow(entry.key)) {
				double value = entry.key;
				m.removeNode(entry);
				return value;
			}
			return getDefaultMinValue();
		}
		
		@Override
		public V firstValue() {
			Double2ObjectRBTreeMap.Entry<V> entry = findLowest();
			return entry == null ? getDefaultReturnValue() : entry.value;
		}
		
		@Override
		public V lastValue() {
			Double2ObjectRBTreeMap.Entry<V> entry = findHighest();
			return entry == null ? getDefaultReturnValue() : entry.value;
		}
		
		protected Double2ObjectRBTreeMap.Entry<V> findLowest() {
			if(fromStart) return m.first;
			Double2ObjectRBTreeMap.Entry<V> entry = loInclusive ? m.findCeilingNode(low) : m.findHigherNode(low);
			return entry == null || tooHigh(entry.key) ? null : entry;
		}
		
		protected Double2ObjectRBTreeMap.Entry<V> findHighest() {
			if(toEnd) return m.last;
			Double2ObjectRBTreeMap.Entry<V> entry = hiInclusive ? m.findFloorNode(high) : m.findLowerNode(high);
			return entry == null || tooLow(entry.key) ? null : entry;
		}
		
		@Override
		public V put(double key, V value) {
			if(!inRange(key)) throw new IllegalArgumentException("key out of range");
			return m.put(key, value);
		}
		
		@Override
		public V putIfAbsent(double key, V value) {
			if(!inRange(key)) throw new IllegalArgumentException("key out of range");
			return m.putIfAbsent(key, value);
		}
		
		@Override
		public V get(double key) {
			return inRange(key) ? m.get(key) : getDefaultReturnValue();
		}
		
		@Override
		public V getOrDefault(double key, V defaultValue) {
			return inRange(key) ? m.getOrDefault(key, defaultValue) : getDefaultReturnValue();
		}
		
		@Override
		public V remove(double key) {
			return inRange(key) ? m.remove(key) : getDefaultReturnValue();
		}
		
		@Override
		public V removeOrDefault(double key, V defaultValue) {
			return inRange(key) ? m.remove(key) : defaultValue;
		}
		
		@Override
		public boolean remove(double key, V value) {
			return inRange(key) && m.remove(key, value);
		}
		
		@Override
		public ObjectSet<Double2ObjectMap.Entry<V>> double2ObjectEntrySet() {
			if(entrySet == null) entrySet = new SubMapEntrySet();
			return entrySet;
		}
		
		@Override
		public DoubleSet keySet() { return navigableKeySet(); }
		
		@Override
		public DoubleNavigableSet navigableKeySet() {
			if(keySet == null) keySet = new KeySet<>(this);
			return keySet;
		}
		
		@Override
		public DoubleNavigableSet descendingKeySet() {
			return descendingMap().navigableKeySet();
		}
		
		@Override
		public ObjectCollection<V> values() {
			if(values == null) values = new SubMapValues();
			return values;
		}
		
		@Override
		public void forEach(DoubleObjectConsumer<V> action) {
			for(Double2ObjectRBTreeMap.Entry<V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
				action.accept(entry.key, entry.value);
		}
		
		@Override
		public int size() {
			return (fromStart && toEnd) ? m.size() : entrySet().size();
		}
		
		@Override
		public Double2ObjectMap.Entry<V> firstEntry() {
			Double2ObjectRBTreeMap.Entry<V> entry = findLowest();
			return entry == null ? null : new BasicEntry<>(entry.key, entry.value);
		}
		
		@Override
		public Double2ObjectMap.Entry<V> lastEntry() {
			Double2ObjectRBTreeMap.Entry<V> entry = findHighest();
			return entry == null ? null : new BasicEntry<>(entry.key, entry.value);
		}
		
		@Override
		public Double2ObjectMap.Entry<V> pollFirstEntry() {
			Double2ObjectRBTreeMap.Entry<V> entry = findLowest();
			if(entry == null) return null;
			BasicEntry<V> result = new BasicEntry<>(entry.key, entry.value);
			m.removeNode(entry);
			return result;
		}
		
		@Override
		public Double2ObjectMap.Entry<V> pollLastEntry() {
			Double2ObjectRBTreeMap.Entry<V> entry = findHighest();
			if(entry == null) return null;
			BasicEntry<V> result = new BasicEntry<>(entry.key, entry.value);
			m.removeNode(entry);
			return result;
		}
		
		@Override
		public double lowerKey(double e) {
			if(tooHigh(e)) {
				Double2ObjectRBTreeMap.Entry<V> entry = findHighest();
				return entry == null ? getDefaultMinValue() : entry.key;
			}
			Double2ObjectRBTreeMap.Entry<V> entry = m.findLowerNode(e);
			return entry == null || tooHigh(entry.key) ? getDefaultMaxValue() : entry.key;
		}
		
		@Override
		public double floorKey(double e) {
			if(tooHigh(e)) {
				Double2ObjectRBTreeMap.Entry<V> entry = findHighest();
				return entry == null ? getDefaultMinValue() : entry.key;
			}
			Double2ObjectRBTreeMap.Entry<V> entry = m.findFloorNode(e);
			return entry == null || tooHigh(entry.key) ? getDefaultMaxValue() : entry.key;
		}
		
		@Override
		public double ceilingKey(double e) {
			if(tooLow(e)) {
				Double2ObjectRBTreeMap.Entry<V> entry = findLowest();
				return entry == null ? getDefaultMaxValue() : entry.key;
			}
			Double2ObjectRBTreeMap.Entry<V> entry = m.findCeilingNode(e);
			return entry == null || tooLow(entry.key) ? getDefaultMinValue() : entry.key;
		}
		
		@Override
		public double higherKey(double e) {
			if(tooLow(e)) {
				Double2ObjectRBTreeMap.Entry<V> entry = findLowest();
				return entry == null ? getDefaultMaxValue() : entry.key;
			}
			Double2ObjectRBTreeMap.Entry<V> entry = m.findHigherNode(e);
			return entry == null || tooLow(entry.key) ? getDefaultMinValue() : entry.key;
		}
		
		@Override
		public Double2ObjectMap.Entry<V> lowerEntry(double e) {
			if(tooHigh(e)) {
				Double2ObjectRBTreeMap.Entry<V> entry = findHighest();
				return entry == null ? null : new BasicEntry<>(entry.key, entry.value);
			}
			Double2ObjectRBTreeMap.Entry<V> entry = m.findLowerNode(e);
			return entry == null || tooHigh(entry.key) ? null : new BasicEntry<>(entry.key, entry.value);
		}
		
		@Override
		public Double2ObjectMap.Entry<V> floorEntry(double e) {
			if(tooHigh(e)) {
				Double2ObjectRBTreeMap.Entry<V> entry = findHighest();
				return entry == null ? null : new BasicEntry<>(entry.key, entry.value);
			}
			Double2ObjectRBTreeMap.Entry<V> entry = m.findFloorNode(e);
			return entry == null || tooHigh(entry.key) ? null : new BasicEntry<>(entry.key, entry.value);
		}
		
		@Override
		public Double2ObjectMap.Entry<V> ceilingEntry(double e) {
			if(tooLow(e)) {
				Double2ObjectRBTreeMap.Entry<V> entry = findLowest();
				return entry == null ? null : new BasicEntry<>(entry.key, entry.value);
			}
			Double2ObjectRBTreeMap.Entry<V> entry = m.findCeilingNode(e);
			return entry == null || tooLow(entry.key) ? null : new BasicEntry<>(entry.key, entry.value);
		}
		
		@Override
		public Double2ObjectMap.Entry<V> higherEntry(double e) {
			if(tooLow(e)) {
				Double2ObjectRBTreeMap.Entry<V> entry = findLowest();
				return entry == null ? null : new BasicEntry<>(entry.key, entry.value);
			}
			Double2ObjectRBTreeMap.Entry<V> entry = m.findHigherNode(e);
			return entry == null || tooLow(entry.key) ? null : new BasicEntry<>(entry.key, entry.value);
		}
		
		class SubMapEntrySet extends AbstractObjectSet<Double2ObjectMap.Entry<V>> implements ObjectSet<Double2ObjectMap.Entry<V>> {
			@Override
			@Deprecated
			public boolean contains(Object o) {
				if(o instanceof Map.Entry) {
					if(o instanceof Double2ObjectMap.Entry) {
						Double2ObjectMap.Entry<V> entry = (Double2ObjectMap.Entry<V>)o;
						if(entry.getKey() == null && comparator() == null) return false;
						Double2ObjectRBTreeMap.Entry<V> subEntry = m.findNode(entry.getDoubleKey());
						if(subEntry != null) return Objects.equals(entry.getValue(), subEntry.value);
					}
					else {
						Map.Entry<?, ?> entry = (Map.Entry<?, ?>)o;
						if(entry.getKey() == null && comparator() == null) return false;
						if(!(entry.getKey() instanceof Double)) return false;
						Double2ObjectRBTreeMap.Entry<V> subEntry = m.findNode((Double)entry.getKey());
						if(subEntry != null) return Objects.equals(entry.getValue(), subEntry.value);
					}
				}
				return false;
			}
			
			@Override
			@Deprecated
			public boolean remove(Object o) {
				if(o instanceof Map.Entry) {
					if(o instanceof Double2ObjectMap.Entry) {
						Double2ObjectMap.Entry<V> entry = (Double2ObjectMap.Entry<V>)o;
						return NavigableSubMap.this.remove(entry.getDoubleKey(), entry.getValue());
					}
					Map.Entry<?, ?> entry = (Map.Entry<?, ?>)o;
					return NavigableSubMap.this.remove(entry.getKey(), entry.getValue());
				}
				return false;
			}
			@Override
			public ObjectIterator<Double2ObjectMap.Entry<V>> iterator() { return entryIterator(); }
			@Override
			public int size() { return NavigableSubMap.this.size(); }
			@Override
			public void clear() { NavigableSubMap.this.clear(); }
			
			@Override
			public void forEach(Consumer<? super Double2ObjectMap.Entry<V>> action) {
				Objects.requireNonNull(action);
				for(Double2ObjectRBTreeMap.Entry<V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					action.accept(new BasicEntry<>(entry.key, entry.value));
			}
			
			@Override
			public <E> void forEach(E input, ObjectObjectConsumer<E, Double2ObjectMap.Entry<V>> action) {
				Objects.requireNonNull(action);
				for(Double2ObjectRBTreeMap.Entry<V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					action.accept(input, new BasicEntry<>(entry.key, entry.value));
			}
			
			@Override
			public boolean matchesAny(Object2BooleanFunction<Double2ObjectMap.Entry<V>> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return false;
				BasicEntry<V> subEntry = new BasicEntry<>();
				for(Double2ObjectRBTreeMap.Entry<V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					subEntry.set(entry.key, entry.value);
					if(filter.getBoolean(subEntry)) return true;
				}
				return false;
			}
			
			@Override
			public boolean matchesNone(Object2BooleanFunction<Double2ObjectMap.Entry<V>> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return true;
				BasicEntry<V> subEntry = new BasicEntry<>();
				for(Double2ObjectRBTreeMap.Entry<V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					subEntry.set(entry.key, entry.value);
					if(filter.getBoolean(subEntry)) return false;
				}
				return true;
			}
			
			@Override
			public boolean matchesAll(Object2BooleanFunction<Double2ObjectMap.Entry<V>> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return true;
				BasicEntry<V> subEntry = new BasicEntry<>();
				for(Double2ObjectRBTreeMap.Entry<V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					subEntry.set(entry.key, entry.value);
					if(!filter.getBoolean(subEntry)) return false;
				}
				return true;
			}
			
			@Override
			public <E> E reduce(E identity, BiFunction<E, Double2ObjectMap.Entry<V>, E> operator) {
				Objects.requireNonNull(operator);
				E state = identity;
				for(Double2ObjectRBTreeMap.Entry<V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					state = operator.apply(state, new BasicEntry<>(entry.key, entry.value));
				}
				return state;
			}
			
			@Override
			public Double2ObjectMap.Entry<V> reduce(ObjectObjectUnaryOperator<Double2ObjectMap.Entry<V>, Double2ObjectMap.Entry<V>> operator) {
				Objects.requireNonNull(operator);
				Double2ObjectMap.Entry<V> state = null;
				boolean empty = true;
				for(Double2ObjectRBTreeMap.Entry<V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					if(empty) {
						empty = false;
						state = new BasicEntry<>(entry.key, entry.value);
						continue;
					}
					state = operator.apply(state, new BasicEntry<>(entry.key, entry.value));
				}
				return state;
			}
			
			@Override
			public Double2ObjectMap.Entry<V> findFirst(Object2BooleanFunction<Double2ObjectMap.Entry<V>> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return null;
				BasicEntry<V> subEntry = new BasicEntry<>();
				for(Double2ObjectRBTreeMap.Entry<V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					subEntry.set(entry.key, entry.value);
					if(filter.getBoolean(subEntry)) return subEntry;
				}
				return null;
			}
			
			@Override
			public int count(Object2BooleanFunction<Double2ObjectMap.Entry<V>> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return 0;
				int result = 0;
				BasicEntry<V> subEntry = new BasicEntry<>();
				for(Double2ObjectRBTreeMap.Entry<V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					subEntry.set(entry.key, entry.value);
					if(filter.getBoolean(subEntry)) result++;
				}
				return result;
			}
		}
		
		final class SubMapValues extends AbstractObjectCollection<V> {
			@Override
			public boolean add(V o) { throw new UnsupportedOperationException(); }
			
			@Override
			public boolean contains(Object e) {
				return containsValue(e);
			}
			
			@Override
			public ObjectIterator<V> iterator() { return valueIterator(); }
			
			@Override
			public int size() {
				return NavigableSubMap.this.size();
			}
			
			@Override
			public void clear() {
				NavigableSubMap.this.clear();
			}
			
			@Override
			public void forEach(Consumer<? super V> action) {
				Objects.requireNonNull(action);
				for(Double2ObjectRBTreeMap.Entry<V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					action.accept(entry.value);
			}
			
			@Override
			public <E> void forEach(E input, ObjectObjectConsumer<E, V> action) {
				Objects.requireNonNull(action);
				for(Double2ObjectRBTreeMap.Entry<V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					action.accept(input, entry.value);
			}
			
			@Override
			public boolean matchesAny(Object2BooleanFunction<V> filter) {
				Objects.requireNonNull(filter);
				for(Double2ObjectRBTreeMap.Entry<V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					if(filter.getBoolean(entry.value)) return true;
				return false;
			}
			
			@Override
			public boolean matchesNone(Object2BooleanFunction<V> filter) {
				Objects.requireNonNull(filter);
				for(Double2ObjectRBTreeMap.Entry<V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					if(filter.getBoolean(entry.value)) return false;
				return true;
			}
			
			@Override
			public boolean matchesAll(Object2BooleanFunction<V> filter) {
				Objects.requireNonNull(filter);
				for(Double2ObjectRBTreeMap.Entry<V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					if(!filter.getBoolean(entry.value)) return false;
				return true;
			}
			
			@Override
			public <E> E reduce(E identity, BiFunction<E, V, E> operator) {
				Objects.requireNonNull(operator);
				E state = identity;
				for(Double2ObjectRBTreeMap.Entry<V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					state = operator.apply(state, entry.value);
				return state;
			}
			
			@Override
			public V reduce(ObjectObjectUnaryOperator<V, V> operator) {
				Objects.requireNonNull(operator);
				V state = null;
				boolean empty = true;
				for(Double2ObjectRBTreeMap.Entry<V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					if(empty) {
						empty = false;
						state = entry.value;
						continue;
					}
					state = operator.apply(state, entry.value);
				}
				return state;
			}
			
			@Override
			public V findFirst(Object2BooleanFunction<V> filter) {
				Objects.requireNonNull(filter);
				for(Double2ObjectRBTreeMap.Entry<V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					if(filter.getBoolean(entry.value)) return entry.value;
				return null;
			}
			
			@Override
			public int count(Object2BooleanFunction<V> filter) {
				Objects.requireNonNull(filter);
				int result = 0;
				for(Double2ObjectRBTreeMap.Entry<V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					if(filter.getBoolean(entry.value)) result++;
				return result;
			}
		}
		
		final class SubMapEntrySetIterator extends SubMapEntryIterator implements ObjectListIterator<Double2ObjectMap.Entry<V>> {
			public SubMapEntrySetIterator(boolean descending) {
				super(descending);
			}
			
			public SubMapEntrySetIterator(Double2ObjectRBTreeMap.Entry<V> entry, Double2ObjectRBTreeMap.Entry<V> lowerFence, Double2ObjectRBTreeMap.Entry<V> upperFence) {
				super(entry, lowerFence, upperFence);
			}
			
			@Override
			public Double2ObjectMap.Entry<V> next() { return nextEntry(); }
			@Override
			public Double2ObjectMap.Entry<V> previous() { return previousEntry(); }
			
			@Override
			public void set(Double2ObjectMap.Entry<V> e) { throw new UnsupportedOperationException(); }
			@Override
			public void add(Double2ObjectMap.Entry<V> e) { throw new UnsupportedOperationException(); }
		}
		
		final class SubMapKeyIterator extends SubMapEntryIterator implements DoubleListIterator {
			public SubMapKeyIterator(boolean descending) {
				super(descending);
			}
			
			public SubMapKeyIterator(Double2ObjectRBTreeMap.Entry<V> entry, Double2ObjectRBTreeMap.Entry<V> lowerFence, Double2ObjectRBTreeMap.Entry<V> upperFence) {
				super(entry, lowerFence, upperFence);
			}
			
			@Override
			public double nextDouble() { return nextEntry().key; }
			@Override
			public double previousDouble() { return previousEntry().key; }
			
			@Override
			public void set(double e) { throw new UnsupportedOperationException(); }
			@Override
			public void add(double e) { throw new UnsupportedOperationException(); }
		}
		
		final class SubMapValueIterator extends SubMapEntryIterator implements ObjectListIterator<V> {
			public SubMapValueIterator(boolean descending) {
				super(descending);
			}
			
			public SubMapValueIterator(Double2ObjectRBTreeMap.Entry<V> entry, Double2ObjectRBTreeMap.Entry<V> lowerFence, Double2ObjectRBTreeMap.Entry<V> upperFence) {
				super(entry, lowerFence, upperFence);
			}
			
			@Override
			public V next() { return nextEntry().value; }
			@Override
			public V previous() { return previousEntry().value; }
			
			@Override
			public void set(V e) { throw new UnsupportedOperationException(); }
			@Override
			public void add(V e) { throw new UnsupportedOperationException(); }
		}
		
		abstract class SubMapEntryIterator {
			Double lowerFence;
			Double upperFence;
			Double2ObjectRBTreeMap.Entry<V> next;
			Double2ObjectRBTreeMap.Entry<V> previous;
			Double2ObjectRBTreeMap.Entry<V> current;
			int index = 0;
			
			public SubMapEntryIterator(boolean descending) {
				this(descending ? findHighest() : findLowest(), fromStart ? null : findLowest(), toEnd ? null : findHighest());
			}
			
			public SubMapEntryIterator(Double2ObjectRBTreeMap.Entry<V> entry, Double2ObjectRBTreeMap.Entry<V> lowerFence, Double2ObjectRBTreeMap.Entry<V> upperFence) {
				next = entry;
				previous = entry.previous();
				this.lowerFence = lowerFence != null ? Double.valueOf(lowerFence.key) : null;
				this.upperFence = upperFence != null ? Double.valueOf(upperFence.key) : null;
			}
			
			public boolean hasNext() {
				return next != null && (upperFence == null || Double.doubleToLongBits(next.key) == Double.doubleToLongBits(upperFence.doubleValue()));
			}
			
			public boolean hasPrevious() {
				return previous != null && (lowerFence == null || Double.doubleToLongBits(next.key) == Double.doubleToLongBits(lowerFence.doubleValue()));
			}
			
			public int nextIndex() {
				return index;
			}
			
			public int previousIndex() {
				return index - 1;
			}
			
			protected void updateNext() {
				next = current.next();
			}
			
			protected void updatePrevious() {
				previous = current.previous();
			}
			
			public Double2ObjectRBTreeMap.Entry<V> nextEntry() {
				if(!hasNext()) throw new NoSuchElementException();
				current = previous = next;
				updateNext();
				index++;
				return current;
			}
			
			public Double2ObjectRBTreeMap.Entry<V> previousEntry() {
				if(!hasPrevious()) throw new NoSuchElementException();
				current = next = previous;
				updatePrevious();
				index--;
				return current;
			}
			
			public void remove() {
				if(current == null) throw new IllegalStateException();
				if(current == previous) index--;
				updateNext();
				updatePrevious();
				if(current.needsSuccessor()) next = current;
				m.removeNode(current);
				current = null;
			}
		}
	}
	
	private class EntrySet extends AbstractObjectSet<Double2ObjectMap.Entry<V>> {
		
		@Override
		public ObjectIterator<Double2ObjectMap.Entry<V>> iterator() {
			return new EntryIterator(false);
		}
		
		@Override
		@Deprecated
		public boolean contains(Object o) {
			if(o instanceof Map.Entry) {
				if(o instanceof Double2ObjectMap.Entry) {
					Double2ObjectMap.Entry<V> entry = (Double2ObjectMap.Entry<V>)o;
					if(entry.getKey() == null && comparator() == null) return false;
					Double2ObjectRBTreeMap.Entry<V> subEntry = Double2ObjectRBTreeMap.this.findNode(entry.getDoubleKey());
					if(subEntry != null) return Objects.equals(entry.getValue(), subEntry.value);
				}
				else {
					Map.Entry<?, ?> entry = (Map.Entry<?, ?>)o;
					if(entry.getKey() == null && comparator() == null) return false;
					if(!(entry.getKey() instanceof Double)) return false;
					Double2ObjectRBTreeMap.Entry<V> subEntry = Double2ObjectRBTreeMap.this.findNode((Double)entry.getKey());
					if(subEntry != null) return Objects.equals(entry.getValue(), subEntry.value);
				}
			}
			return false;
		}
		
		@Override
		@Deprecated
		public boolean remove(Object o) {
			if(o instanceof Map.Entry) {
				if(o instanceof Double2ObjectMap.Entry) {
					Double2ObjectMap.Entry<V> entry = (Double2ObjectMap.Entry<V>)o;
					return Double2ObjectRBTreeMap.this.remove(entry.getDoubleKey(), entry.getValue());
				}
				Map.Entry<?, ?> entry = (Map.Entry<?, ?>)o;
				return Double2ObjectRBTreeMap.this.remove(entry.getKey(), entry.getValue());
			}
			return false;
		}
		
		@Override
		public int size() {
			return Double2ObjectRBTreeMap.this.size();
		}
		
		@Override
		public void clear() {
			Double2ObjectRBTreeMap.this.clear();
		}
		
		@Override
		public void forEach(Consumer<? super Double2ObjectMap.Entry<V>> action) {
			Objects.requireNonNull(action);
			for(Double2ObjectRBTreeMap.Entry<V> entry = first;entry != null;entry = entry.next())
				action.accept(new BasicEntry<>(entry.key, entry.value));
		}
		
		@Override
		public <E> void forEach(E input, ObjectObjectConsumer<E, Double2ObjectMap.Entry<V>> action) {
			Objects.requireNonNull(action);
			for(Double2ObjectRBTreeMap.Entry<V> entry = first;entry != null;entry = entry.next())
				action.accept(input, new BasicEntry<>(entry.key, entry.value));
		}
		
		@Override
		public boolean matchesAny(Object2BooleanFunction<Double2ObjectMap.Entry<V>> filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return false;
			BasicEntry<V> subEntry = new BasicEntry<>();
			for(Double2ObjectRBTreeMap.Entry<V> entry = first;entry != null;entry = entry.next()) {
				subEntry.set(entry.key, entry.value);
				if(filter.getBoolean(subEntry)) return true;
			}
			return false;
		}
		
		@Override
		public boolean matchesNone(Object2BooleanFunction<Double2ObjectMap.Entry<V>> filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return true;
			BasicEntry<V> subEntry = new BasicEntry<>();
			for(Double2ObjectRBTreeMap.Entry<V> entry = first;entry != null;entry = entry.next()) {
				subEntry.set(entry.key, entry.value);
				if(filter.getBoolean(subEntry)) return false;
			}
			return true;
		}
		
		@Override
		public boolean matchesAll(Object2BooleanFunction<Double2ObjectMap.Entry<V>> filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return true;
			BasicEntry<V> subEntry = new BasicEntry<>();
			for(Double2ObjectRBTreeMap.Entry<V> entry = first;entry != null;entry = entry.next()) {
				subEntry.set(entry.key, entry.value);
				if(!filter.getBoolean(subEntry)) return false;
			}
			return true;
		}
		
		@Override
		public <E> E reduce(E identity, BiFunction<E, Double2ObjectMap.Entry<V>, E> operator) {
			Objects.requireNonNull(operator);
			E state = identity;
			for(Double2ObjectRBTreeMap.Entry<V> entry = first;entry != null;entry = entry.next()) {
				state = operator.apply(state, new BasicEntry<>(entry.key, entry.value));
			}
			return state;
		}
		
		@Override
		public Double2ObjectMap.Entry<V> reduce(ObjectObjectUnaryOperator<Double2ObjectMap.Entry<V>, Double2ObjectMap.Entry<V>> operator) {
			Objects.requireNonNull(operator);
			Double2ObjectMap.Entry<V> state = null;
			boolean empty = true;
			for(Double2ObjectRBTreeMap.Entry<V> entry = first;entry != null;entry = entry.next()) {
				if(empty) {
					empty = false;
					state = new BasicEntry<>(entry.key, entry.value);
					continue;
				}
				state = operator.apply(state, new BasicEntry<>(entry.key, entry.value));
			}
			return state;
		}
		
		@Override
		public Double2ObjectMap.Entry<V> findFirst(Object2BooleanFunction<Double2ObjectMap.Entry<V>> filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return null;
			BasicEntry<V> subEntry = new BasicEntry<>();
			for(Double2ObjectRBTreeMap.Entry<V> entry = first;entry != null;entry = entry.next()) {
				subEntry.set(entry.key, entry.value);
				if(filter.getBoolean(subEntry)) return subEntry;
			}
			return null;
		}
		
		@Override
		public int count(Object2BooleanFunction<Double2ObjectMap.Entry<V>> filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return 0;
			int result = 0;
			BasicEntry<V> subEntry = new BasicEntry<>();
			for(Double2ObjectRBTreeMap.Entry<V> entry = first;entry != null;entry = entry.next()) {
				subEntry.set(entry.key, entry.value);
				if(filter.getBoolean(subEntry)) result++;
			}
			return result;
		}
	}
	
	static final class KeySet<V> extends AbstractDoubleSet implements DoubleNavigableSet {
		Double2ObjectNavigableMap<V> m;
		
		KeySet(Double2ObjectNavigableMap<V> m) {
			this.m = m;
		}
		
		@Override
		public void setDefaultMaxValue(double e) { m.setDefaultMaxValue(e); }
		@Override
		public double getDefaultMaxValue() { return m.getDefaultMaxValue(); }
		@Override
		public void setDefaultMinValue(double e) { m.setDefaultMinValue(e); }
		@Override
		public double getDefaultMinValue() { return m.getDefaultMinValue(); }
		@Override
		public boolean add(double o) { throw new UnsupportedOperationException(); }
		@Override
		public boolean addAndMoveToFirst(double o) { throw new UnsupportedOperationException(); }
		@Override
		public boolean addAndMoveToLast(double o) { throw new UnsupportedOperationException(); }
		@Override
		public boolean moveToFirst(double o) { throw new UnsupportedOperationException(); }
		@Override
		public boolean moveToLast(double o) { throw new UnsupportedOperationException(); }
		@Override
		public DoubleComparator comparator() { return m.comparator(); }
		@Override
		public double lower(double e) { return m.lowerKey(e); }
		@Override
		public double floor(double e) { return m.floorKey(e); }
		@Override
		public double ceiling(double e) { return m.ceilingKey(e); }
		@Override
		public double higher(double e) { return m.higherKey(e); }
		@Override
		public double firstDouble() { return m.firstDoubleKey(); }
		@Override
		public double pollFirstDouble() { return m.pollFirstDoubleKey(); }
		@Override
		public double lastDouble() { return m.lastDoubleKey(); }
		@Override
		public double pollLastDouble() { return m.pollLastDoubleKey(); }		
		@Override
		public boolean remove(double o) { 
			int oldSize = m.size();
			m.remove(o); 
			return oldSize != m.size();
		}
		
		@Override
		public KeySet<V> copy() { throw new UnsupportedOperationException(); }
		
		@Override
		public DoubleNavigableSet subSet(double fromElement, boolean fromInclusive, double toElement, boolean toInclusive) {
			return new KeySet<>(m.subMap(fromElement, fromInclusive, toElement, toInclusive));
		}
		
		@Override
		public DoubleNavigableSet headSet(double toElement, boolean inclusive) {
			return new KeySet<>(m.headMap(toElement, inclusive));
		}
		
		@Override
		public DoubleNavigableSet tailSet(double fromElement, boolean inclusive) {
			return new KeySet<>(m.tailMap(fromElement, inclusive));
		}
		
		@Override
		public DoubleNavigableSet descendingSet() {
			return new KeySet<>(m.descendingMap());
		}
		
		@Override
		public DoubleBidirectionalIterator iterator() {
			if(m instanceof Double2ObjectRBTreeMap) return ((Double2ObjectRBTreeMap<V>)m).keyIterator(false);
			return ((NavigableSubMap<V>)m).keyIterator(false);
		}
		
		@Override
		public DoubleBidirectionalIterator iterator(double fromElement) {
			if(m instanceof Double2ObjectRBTreeMap) return ((Double2ObjectRBTreeMap<V>)m).keyIterator(fromElement);
			return ((NavigableSubMap<V>)m).keyIterator(fromElement);
		}
		
		@Override
		public DoubleBidirectionalIterator descendingIterator() {
			if(m instanceof Double2ObjectRBTreeMap) return ((Double2ObjectRBTreeMap<V>)m).keyIterator(true);
			return ((NavigableSubMap<V>)m).keyIterator(true);
		}
		
		@Override
		public int size() {
			return m.size();
		}
		
		@Override
		public void clear() {
			m.clear();
		}
		
		protected Double2ObjectRBTreeMap.Entry<V> start() {
			if(m instanceof Double2ObjectRBTreeMap) return ((Double2ObjectRBTreeMap<V>)m).first;
			return ((NavigableSubMap<V>)m).findLowest();
		}
		
		protected Double2ObjectRBTreeMap.Entry<V> end() {
			if(m instanceof Double2ObjectRBTreeMap) return null;
			return ((NavigableSubMap<V>)m).findLowest();
		}
		
		protected Double2ObjectRBTreeMap.Entry<V> next(Double2ObjectRBTreeMap.Entry<V> entry) {
			if(m instanceof Double2ObjectRBTreeMap) return entry.next();
			return ((NavigableSubMap<V>)m).next(entry);
		}
		
		protected Double2ObjectRBTreeMap.Entry<V> previous(Double2ObjectRBTreeMap.Entry<V> entry) {
			if(m instanceof Double2ObjectRBTreeMap) return entry.previous();
			return ((NavigableSubMap<V>)m).previous(entry);
		}
		
		@Override
		public void forEach(DoubleConsumer action) {
			Objects.requireNonNull(action);
			for(Double2ObjectRBTreeMap.Entry<V> entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				action.accept(entry.key);
		}
		
		@Override
		public <E> void forEach(E input, ObjectDoubleConsumer<E> action) {
			Objects.requireNonNull(action);
			for(Double2ObjectRBTreeMap.Entry<V> entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				action.accept(input, entry.key);
		}
		
		@Override
		public boolean matchesAny(Double2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(Double2ObjectRBTreeMap.Entry<V> entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				if(filter.get(entry.key)) return true;
			return false;
		}
		
		@Override
		public boolean matchesNone(Double2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(Double2ObjectRBTreeMap.Entry<V> entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				if(filter.get(entry.key)) return false;
			return true;
		}
		
		@Override
		public boolean matchesAll(Double2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(Double2ObjectRBTreeMap.Entry<V> entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				if(!filter.get(entry.key)) return false;
			return true;
		}
		
		@Override
		public double reduce(double identity, DoubleDoubleUnaryOperator operator) {
			Objects.requireNonNull(operator);
			double state = identity;
			for(Double2ObjectRBTreeMap.Entry<V> entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				state = operator.applyAsDouble(state, entry.key);
			return state;
		}
		
		@Override
		public double reduce(DoubleDoubleUnaryOperator operator) {
			Objects.requireNonNull(operator);
			double state = 0D;
			boolean empty = true;
			for(Double2ObjectRBTreeMap.Entry<V> entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry)) {
				if(empty) {
					empty = false;
					state = entry.key;
					continue;
				}
				state = operator.applyAsDouble(state, entry.key);
			}
			return state;
		}
		
		@Override
		public double findFirst(Double2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(Double2ObjectRBTreeMap.Entry<V> entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				if(filter.get(entry.key)) return entry.key;
			return 0D;
		}
		
		@Override
		public int count(Double2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			int result = 0;
			for(Double2ObjectRBTreeMap.Entry<V> entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				if(filter.get(entry.key)) result++;
			return result;
		}
	}
	
	final class Values extends AbstractObjectCollection<V> {

		@Override
		public boolean add(V o) { throw new UnsupportedOperationException(); }
		
		@Override
		public boolean contains(Object e) {
			return containsValue(e);
		}
		
		@Override
		public ObjectIterator<V> iterator() { return new ValueIterator(false); }
		
		@Override
		public int size() {
			return Double2ObjectRBTreeMap.this.size();
		}
		
		@Override
		public void clear() {
			Double2ObjectRBTreeMap.this.clear();
		}
		
		@Override
		public void forEach(Consumer<? super V> action) {
			Objects.requireNonNull(action);
			for(Double2ObjectRBTreeMap.Entry<V> entry = first;entry != null;entry = entry.next())
				action.accept(entry.value);
		}
		
		@Override
		public <E> void forEach(E input, ObjectObjectConsumer<E, V> action) {
			Objects.requireNonNull(action);
			for(Double2ObjectRBTreeMap.Entry<V> entry = first;entry != null;entry = entry.next())
				action.accept(input, entry.value);
		}
		
		@Override
		public boolean matchesAny(Object2BooleanFunction<V> filter) {
			Objects.requireNonNull(filter);
			for(Double2ObjectRBTreeMap.Entry<V> entry = first;entry != null;entry = entry.next())
				if(filter.getBoolean(entry.value)) return true;
			return false;
		}
		
		@Override
		public boolean matchesNone(Object2BooleanFunction<V> filter) {
			Objects.requireNonNull(filter);
			for(Double2ObjectRBTreeMap.Entry<V> entry = first;entry != null;entry = entry.next())
				if(filter.getBoolean(entry.value)) return false;
			return true;
		}
		
		@Override
		public boolean matchesAll(Object2BooleanFunction<V> filter) {
			Objects.requireNonNull(filter);
			for(Double2ObjectRBTreeMap.Entry<V> entry = first;entry != null;entry = entry.next())
				if(!filter.getBoolean(entry.value)) return false;
			return true;
		}
		
		@Override
		public <E> E reduce(E identity, BiFunction<E, V, E> operator) {
			Objects.requireNonNull(operator);
			E state = identity;
			for(Double2ObjectRBTreeMap.Entry<V> entry = first;entry != null;entry = entry.next())
				state = operator.apply(state, entry.value);
			return state;
		}
		
		@Override
		public V reduce(ObjectObjectUnaryOperator<V, V> operator) {
			Objects.requireNonNull(operator);
			V state = null;
			boolean empty = true;
			for(Double2ObjectRBTreeMap.Entry<V> entry = first;entry != null;entry = entry.next()) {
				if(empty) {
					empty = false;
					state = entry.value;
					continue;
				}
				state = operator.apply(state, entry.value);
			}
			return state;
		}
		
		@Override
		public V findFirst(Object2BooleanFunction<V> filter) {
			Objects.requireNonNull(filter);
			for(Double2ObjectRBTreeMap.Entry<V> entry = first;entry != null;entry = entry.next())
				if(filter.getBoolean(entry.value)) return entry.value;
			return null;
		}
		
		@Override
		public int count(Object2BooleanFunction<V> filter) {
			Objects.requireNonNull(filter);
			int result = 0;
			for(Double2ObjectRBTreeMap.Entry<V> entry = first;entry != null;entry = entry.next())
				if(filter.getBoolean(entry.value)) result++;
			return result;
		}
	}
	
	final class EntryIterator extends MapEntryIterator implements ObjectListIterator<Double2ObjectMap.Entry<V>> {
		
		public EntryIterator(boolean descending) {
			super(descending);
		}
		
		public EntryIterator(Entry<V> entry) {
			super(entry);
		}
		
		@Override
		public Double2ObjectMap.Entry<V> next() {
			return nextEntry();
		}
		
		@Override
		public Double2ObjectMap.Entry<V> previous() {
			return previousEntry();
		}
		
		@Override
		public void set(Double2ObjectMap.Entry<V> e) { throw new UnsupportedOperationException(); }
		
		@Override
		public void add(Double2ObjectMap.Entry<V> e) { throw new UnsupportedOperationException(); }
	}
	
	final class KeyIterator extends MapEntryIterator implements DoubleListIterator {
		
		public KeyIterator(boolean descending) {
			super(descending);
		}
		
		public KeyIterator(Entry<V> entry) {
			super(entry);
		}
		
		@Override
		public double previousDouble() { return previousEntry().key; }
		@Override
		public double nextDouble() { return nextEntry().key; }
		
		@Override
		public void set(double e) { throw new UnsupportedOperationException(); }
		@Override
		public void add(double e) { throw new UnsupportedOperationException(); }
	}
	
	final class ValueIterator extends MapEntryIterator implements ObjectListIterator<V> {
		
		public ValueIterator(boolean descending) {
			super(descending);
		}
		
		public ValueIterator(Entry<V> entry) {
			super(entry);
		}
		
		@Override
		public V previous() {
			return previousEntry().value;
		}
		
		@Override
		public V next() {
			return nextEntry().value;
		}
		
		@Override
		public void set(V e) { throw new UnsupportedOperationException(); }
		
		@Override
		public void add(V e) { throw new UnsupportedOperationException(); }
	}
	
	abstract class MapEntryIterator {
		Entry<V> next;
		Entry<V> previous;
		Entry<V> current;
		int index = 0;
		
		public MapEntryIterator(boolean descending) {
			if(descending) previous = last;
			else next = first;
		}
		
		public MapEntryIterator(Entry<V> entry) {
			next = entry;
			previous = entry.previous();
		}
		
		public boolean hasNext() {
			return next != null;
		}
		
		public boolean hasPrevious() {
			return previous != null;
		}
		
		public int nextIndex() {
			return index;
		}
		
		public int previousIndex() {
			return index - 1;
		}
		
		protected void updateNext() {
			next = current.next();
		}
		
		protected void updatePrevious() {
			previous = current.previous();
		}
		
		public Entry<V> nextEntry() {
			if(!hasNext()) throw new NoSuchElementException();
			current = previous = next;
			updateNext();
			index++;
			return current;
		}
		
		public Entry<V> previousEntry() {
			if(!hasPrevious()) throw new NoSuchElementException();
			current = next = previous;
			updatePrevious();
			index--;
			return current;
		}
		
		public void remove() {
			if(current == null) throw new IllegalStateException();
			if(current == previous) index--;
			updateNext();
			updatePrevious();
			if(current.needsSuccessor()) next = current;
			removeNode(current);
			current = null;
		}
	}
	
	private static final class Entry<V> implements Double2ObjectMap.Entry<V>
	{
		static final int BLACK = 1;
		
		double key;
		V value;
		int state;
		Entry<V> parent;
		Entry<V> left;
		Entry<V> right;
		
		Entry(double key, V value, Entry<V> parent) {
			this.key = key;
			this.value = value;
			this.parent = parent;
		}
		
		Entry<V> copy() {
			Entry<V> entry = new Entry<>(key, value, null);
			entry.state = state;
			if(left != null) {
				Entry<V> newLeft = left.copy();
				entry.left = newLeft;
				newLeft.parent = entry;
			}
			if(right != null) {
				Entry<V> newRight = right.copy();
				entry.right = newRight;
				newRight.parent = entry;
			}
			return entry;
		}
		
		@Override
		public double getDoubleKey() {
			return key;
		}
		
		@Override
		public V getValue() {
			return value;
		}
		
		@Override
		public V setValue(V value) {
			V oldValue = this.value;
			this.value = value;
			return oldValue;
		}
		
		@Override
		public boolean equals(Object obj) {
			if(obj instanceof Map.Entry) {
				if(obj instanceof Double2ObjectMap.Entry) {
					Double2ObjectMap.Entry<V> entry = (Double2ObjectMap.Entry<V>)obj;
					return Double.doubleToLongBits(key) == Double.doubleToLongBits(entry.getDoubleKey()) && Objects.equals(value, entry.getValue());
				}
				Map.Entry<?, ?> entry = (Map.Entry<?, ?>)obj;
				Object otherKey = entry.getKey();
				if(otherKey == null) return false;
				Object otherValue = entry.getValue();
				return otherKey instanceof Double && Double.doubleToLongBits(key) == Double.doubleToLongBits(((Double)otherKey).doubleValue()) && Objects.equals(value, otherValue);
			}
			return false;
		}
		
		@Override
		public int hashCode() {
			return Double.hashCode(key) ^ Objects.hashCode(value);
		}
		
		@Override
		public String toString() {
			return Double.toString(key) + "=" + Objects.toString(value);
		}
		
		boolean isBlack() {
			return (state & BLACK) != 0;
		}
		
		void setBlack(boolean value) {
			if(value) state |= BLACK;
			else state &= ~BLACK;
		}
		
		boolean needsSuccessor() { return left != null && right != null; }
		
		boolean replace(Entry<V> entry) {
			if(entry != null) entry.parent = parent;
			if(parent != null) {
				if(parent.left == this) parent.left = entry;
				else parent.right = entry;
			}
			return parent == null;
		}
		
		Entry<V> next() {
			if(right != null) {
				Entry<V> parent = right;
				while(parent.left != null) parent = parent.left;
				return parent;
			}
			Entry<V> parent = this.parent;
			Entry<V> control = this;
			while(parent != null && control == parent.right) {
				control = parent;
				parent = parent.parent;
			}
			return parent;
		}
		
		Entry<V> previous() {
			if(left != null) {
				Entry<V> parent = left;
				while(parent.right != null) parent = parent.right;
				return parent;
			}
			Entry<V> parent = this.parent;
			Entry<V> control = this;
			while(parent != null && control == parent.left) {
				control = parent;
				parent = parent.parent;
			}
			return parent;
		}
	}
}