package speiger.src.collections.ints.maps.impl.tree;

import java.util.Map;
import java.util.Objects;
import java.util.NoSuchElementException;
import java.util.function.Consumer;
import java.util.function.BiFunction;

import speiger.src.collections.ints.collections.IntBidirectionalIterator;
import speiger.src.collections.ints.functions.IntComparator;
import speiger.src.collections.ints.functions.IntConsumer;
import speiger.src.collections.objects.functions.consumer.ObjectIntConsumer;
import speiger.src.collections.ints.functions.function.Int2BooleanFunction;
import speiger.src.collections.ints.functions.consumer.IntLongConsumer;
import speiger.src.collections.ints.functions.function.Int2LongFunction;
import speiger.src.collections.ints.functions.function.IntLongUnaryOperator;
import speiger.src.collections.ints.functions.function.IntIntUnaryOperator;
import speiger.src.collections.ints.lists.IntListIterator;
import speiger.src.collections.ints.maps.abstracts.AbstractInt2LongMap;
import speiger.src.collections.ints.maps.interfaces.Int2LongMap;
import speiger.src.collections.ints.maps.interfaces.Int2LongNavigableMap;
import speiger.src.collections.ints.sets.AbstractIntSet;
import speiger.src.collections.ints.sets.IntNavigableSet;
import speiger.src.collections.ints.sets.IntSet;
import speiger.src.collections.ints.sets.IntSortedSet;
import speiger.src.collections.ints.utils.IntIterators;
import speiger.src.collections.ints.utils.maps.Int2LongMaps;
import speiger.src.collections.longs.collections.AbstractLongCollection;
import speiger.src.collections.longs.collections.LongCollection;
import speiger.src.collections.longs.collections.LongIterator;
import speiger.src.collections.longs.functions.LongSupplier;
import speiger.src.collections.longs.functions.function.LongLongUnaryOperator;
import speiger.src.collections.longs.lists.LongListIterator;
import speiger.src.collections.longs.utils.LongIterators;
import speiger.src.collections.longs.functions.LongConsumer;
import speiger.src.collections.objects.functions.consumer.ObjectObjectConsumer;
import speiger.src.collections.objects.functions.function.Object2BooleanFunction;
import speiger.src.collections.objects.functions.consumer.ObjectLongConsumer;
import speiger.src.collections.longs.functions.function.Long2BooleanFunction;
import speiger.src.collections.objects.lists.ObjectListIterator;
import speiger.src.collections.objects.utils.ObjectIterators;
import speiger.src.collections.objects.collections.ObjectIterator;
import speiger.src.collections.objects.functions.function.ObjectObjectUnaryOperator;
import speiger.src.collections.objects.sets.AbstractObjectSet;
import speiger.src.collections.objects.sets.ObjectSet;

/**
 * A Simple Type Specific RB TreeMap implementation that reduces boxing/unboxing.
 * It is using a bit more memory then <a href="https://github.com/vigna/fastutil">FastUtil</a>,
 * but it saves a lot of Performance on the Optimized removal and iteration logic.
 * Which makes the implementation actually useable and does not get outperformed by Javas default implementation.
 */
public class Int2LongRBTreeMap extends AbstractInt2LongMap implements Int2LongNavigableMap
{
	/** The center of the Tree */
	protected transient Entry tree;
	/** The Lowest possible Node */
	protected transient Entry first;
	/** The Highest possible Node */
	protected transient Entry last;
	/** The amount of elements stored in the Map */
	protected int size = 0;
	/** The Sorter of the Tree */
	protected transient IntComparator comparator;
	
	/** the default return value for max searches */
	protected int defaultMaxNotFound = Integer.MIN_VALUE;
	/** the default return value for min searches */
	protected int defaultMinNotFound = Integer.MAX_VALUE;
	
	/** KeySet Cache */
	protected IntNavigableSet keySet;
	/** Values Cache */
	protected LongCollection values;
	/** EntrySet Cache */
	protected ObjectSet<Int2LongMap.Entry> entrySet;
	
	/**
	 * Default Constructor
	 */
	public Int2LongRBTreeMap() {
	}
	
	/**
	 * Constructor that allows to define the sorter
	 * @param comp the function that decides how the tree is sorted, can be null
	 */
	public Int2LongRBTreeMap(IntComparator comp) {
		comparator = comp;
	}
	
	/**
	 * Helper constructor that allow to create a map from boxed values (it will unbox them)
	 * @param keys the keys that should be put into the map
	 * @param values the values that should be put into the map.
	 * @throws IllegalStateException if the keys and values do not match in lenght
	 */
	public Int2LongRBTreeMap(Integer[] keys, Long[] values) {
		this(keys, values, null);
	}
	
	/**
	 * Helper constructor that has a custom sorter and allow to create a map from boxed values (it will unbox them)
	 * @param keys the keys that should be put into the map
	 * @param values the values that should be put into the map.
	 * @param comp the function that decides how the tree is sorted, can be null
	 * @throws IllegalStateException if the keys and values do not match in lenght
	 */
	public Int2LongRBTreeMap(Integer[] keys, Long[] values, IntComparator comp) {
		comparator = comp;
		if(keys.length != values.length) throw new IllegalStateException("Input Arrays are not equal size");
		for(int i = 0,m=keys.length;i<m;i++) put(keys[i].intValue(), values[i].longValue());
	}
	
	/**
	 * Helper constructor that allow to create a map from unboxed values
	 * @param keys the keys that should be put into the map
	 * @param values the values that should be put into the map.
	 * @throws IllegalStateException if the keys and values do not match in lenght
	 */
	public Int2LongRBTreeMap(int[] keys, long[] values) {
		this(keys, values, null);
	}
	
	/**
	 * Helper constructor that has a custom sorter and allow to create a map from unboxed values
	 * @param keys the keys that should be put into the map
	 * @param values the values that should be put into the map.
	 * @param comp the function that decides how the tree is sorted, can be null
	 * @throws IllegalStateException if the keys and values do not match in lenght
	 */
	public Int2LongRBTreeMap(int[] keys, long[] values, IntComparator comp) {
		comparator = comp;
		if(keys.length != values.length) throw new IllegalStateException("Input Arrays are not equal size");
		for(int i = 0,m=keys.length;i<m;i++) put(keys[i], values[i]);
	}
	
	/**
	 * A Helper constructor that allows to create a Map with exactly the same values as the provided map.
	 * @param map the values that should be present in the map
	 */
	public Int2LongRBTreeMap(Map<? extends Integer, ? extends Long> map) {
		this(map, null);
	}
	
	/**
	 * A Helper constructor that has a custom sorter and allows to create a Map with exactly the same values as the provided map.
	 * @param map the values that should be present in the map
	 * @param comp the function that decides how the tree is sorted, can be null
	 */
	public Int2LongRBTreeMap(Map<? extends Integer, ? extends Long> map, IntComparator comp) {
		comparator = comp;
		putAll(map);
	}
	
	/**
	 * A Type Specific Helper function that allows to create a new Map with exactly the same values as the provided map.
	 * @param map the values that should be present in the map
 	 */
	public Int2LongRBTreeMap(Int2LongMap map) {
		this(map, null);
	}
	
	/**
	 * A Type Specific Helper function that has a custom sorter and allows to create a new Map with exactly the same values as the provided map.
	 * @param map the values that should be present in the map
	 * @param comp the function that decides how the tree is sorted, can be null
 	 */
	public Int2LongRBTreeMap(Int2LongMap map, IntComparator comp) {
		comparator = comp;
		putAll(map);
	}
	
	@Override
	public void setDefaultMaxValue(int value) { defaultMaxNotFound = value; }
	@Override
	public int getDefaultMaxValue() { return defaultMaxNotFound; }
	@Override
	public void setDefaultMinValue(int value) { defaultMinNotFound = value; }
	@Override
	public int getDefaultMinValue() { return defaultMinNotFound; }
	
	@Override
	public long put(int key, long value) {
		if(tree == null) {
			tree = first = last = new Entry(key, value, null);
			size++;
			return getDefaultReturnValue();
		}
		int compare = 0;
		Entry parent = tree;
		while(true) {
			if((compare = compare(key, parent.key)) == 0) return parent.setValue(value);
			if(compare < 0) {
				if(parent.left == null) break;
				parent = parent.left;
			}
			else if(compare > 0) {
				if(parent.right == null) break;
				parent = parent.right;
			}
		}
		Entry adding = new Entry(key, value, parent);
		if(compare < 0)  {
			parent.left = adding;
			if(parent == first)	first = adding;
		}
		else {
			parent.right = adding;
			if(parent == last) last = adding;
		}
		fixAfterInsertion(adding);
		size++;
		return getDefaultReturnValue();
	}
	
	@Override
	public long putIfAbsent(int key, long value) {

		if(tree == null) {
			tree = first = last = new Entry(key, value, null);
			size++;
			return getDefaultReturnValue();
		}
		int compare = 0;
		Entry parent = tree;
		while(true) {
			if((compare = compare(key, parent.key)) == 0) return parent.value;
			if(compare < 0) {
				if(parent.left == null) break;
				parent = parent.left;
			}
			else if(compare > 0) {
				if(parent.right == null) break;
				parent = parent.right;
			}
		}
		Entry adding = new Entry(key, value, parent);
		if(compare < 0)  {
			parent.left = adding;
			if(parent == first)	first = adding;
		}
		else  {
			parent.right = adding;
			if(parent == last) last = adding;
		}
		fixAfterInsertion(adding);
		size++;
		return getDefaultReturnValue();
	}
	
	@Override
	public long addTo(int key, long value) {
		if(tree == null) {
			tree = first = last = new Entry(key, value, null);
			size++;
			return getDefaultReturnValue();
		}
		int compare = 0;
		Entry parent = tree;
		while(true) {
			if((compare = compare(key, parent.key)) == 0) return parent.addTo(value);
			if(compare < 0) {
				if(parent.left == null) break;
				parent = parent.left;
			}
			else if(compare > 0) {
				if(parent.right == null) break;
				parent = parent.right;
			}
		}
		Entry adding = new Entry(key, value, parent);
		if(compare < 0)  {
			parent.left = adding;
			if(parent == first)	first = adding;
		}
		else  {
			parent.right = adding;
			if(parent == last) last = adding;
		}
		fixAfterInsertion(adding);
		size++;
		return getDefaultReturnValue();
	}
	
	@Override
	public long putAndMoveToFirst(int key, long value) { throw new UnsupportedOperationException(); }
	@Override
	public long putAndMoveToLast(int key, long value) { throw new UnsupportedOperationException(); }
	@Override
	public boolean moveToFirst(int key) { throw new UnsupportedOperationException(); }
	@Override
	public boolean moveToLast(int key) { throw new UnsupportedOperationException(); }
	@Override
	public long getAndMoveToFirst(int key) { throw new UnsupportedOperationException(); }
	@Override
	public long getAndMoveToLast(int key) { throw new UnsupportedOperationException(); }
	
	@Override
	public IntComparator comparator() { return comparator; }
	
	@Override
	public boolean containsKey(int key) {
		return findNode(key) != null;
	}
	
	@Override
	public long get(int key) {
		Entry node = findNode(key);
		return node == null ? getDefaultReturnValue() : node.value;
	}
	
	@Override
	public long getOrDefault(int key, long defaultValue) {
		Entry node = findNode(key);
		return node == null ? defaultValue : node.value;
	}
	
	@Override
	public int firstIntKey() {
		if(tree == null) throw new NoSuchElementException();
		return first.key;
	}
	
	@Override
	public int pollFirstIntKey() {
		if(tree == null) throw new NoSuchElementException();
		int result = first.key;
		removeNode(first);
		return result;
	}
	
	@Override
	public int lastIntKey() {
		if(tree == null) throw new NoSuchElementException();
		return last.key;
	}
	
	@Override
	public int pollLastIntKey() {
		if(tree == null) throw new NoSuchElementException();
		int result = last.key;
		removeNode(last);
		return result;
	}
	
	@Override
	public Int2LongMap.Entry firstEntry() {
		if(tree == null) throw new NoSuchElementException();
		return new BasicEntry(first.key, first.value);
	}
	
	@Override
	public Int2LongMap.Entry lastEntry() {
		if(tree == null) throw new NoSuchElementException();
		return new BasicEntry(last.key, last.value);
	}
	
	@Override
	public Int2LongMap.Entry pollFirstEntry() {
		if(tree == null) throw new NoSuchElementException();
		BasicEntry entry = new BasicEntry(first.key, first.value);
		removeNode(first);
		return entry;
	}
	
	@Override
	public Int2LongMap.Entry pollLastEntry() {
		if(tree == null) throw new NoSuchElementException();
		BasicEntry entry = new BasicEntry(last.key, last.value);
		removeNode(last);
		return entry;
	}
	
	@Override
	public long firstLongValue() {
		if(tree == null) throw new NoSuchElementException();
		return first.value;
	}
	
	@Override
	public long lastLongValue() {
		if(tree == null) throw new NoSuchElementException();
		return last.value;
	}
	
	@Override
	public long remove(int key) {
		Entry entry = findNode(key);
		if(entry == null) return getDefaultReturnValue();
		long value = entry.value;
		removeNode(entry);
		return value;
	}
	
	@Override
	public long removeOrDefault(int key, long defaultValue) {
		Entry entry = findNode(key);
		if(entry == null) return defaultValue;
		long value = entry.value;
		removeNode(entry);
		return value;
	}
	
	@Override
	public boolean remove(int key, long value) {
		Entry entry = findNode(key);
		if(entry == null || entry.value != value) return false;
		removeNode(entry);
		return true;
	}
	
	@Override
	public boolean replace(int key, long oldValue, long newValue) {
		Entry entry = findNode(key);
		if(entry == null || entry.value != oldValue) return false;
		entry.value = newValue;
		return true;
	}
	
	@Override
	public long replace(int key, long value) {
		Entry entry = findNode(key);
		if(entry == null) return getDefaultReturnValue();
		long oldValue = entry.value;
		entry.value = value;
		return oldValue;
	}
	
	@Override
	public long computeLong(int key, IntLongUnaryOperator mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		Entry entry = findNode(key);
		if(entry == null) {
			long newValue = mappingFunction.applyAsLong(key, getDefaultReturnValue());
			if(newValue == getDefaultReturnValue()) return newValue;
			put(key, newValue);
			return newValue;
		}
		long newValue = mappingFunction.applyAsLong(key, entry.value);
		if(newValue == getDefaultReturnValue()) {
			removeNode(entry);
			return newValue;
		}
		entry.value = newValue;
		return newValue;
	}
	
	@Override
	public long computeLongIfAbsent(int key, Int2LongFunction mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		Entry entry = findNode(key);
		if(entry == null) {
			long newValue = mappingFunction.get(key);
			if(newValue == getDefaultReturnValue()) return newValue;
			put(key, newValue);
			return newValue;
		}
		if(entry.value == getDefaultReturnValue()) {
			long newValue = mappingFunction.get(key);
			if(newValue == getDefaultReturnValue()) return newValue;
			entry.value = newValue;
		}
		return entry.value;
	}
	
	@Override
	public long supplyLongIfAbsent(int key, LongSupplier valueProvider) {
		Objects.requireNonNull(valueProvider);
		Entry entry = findNode(key);
		if(entry == null) {
			long newValue = valueProvider.getLong();
			if(newValue == getDefaultReturnValue()) return newValue;
			put(key, newValue);
			return newValue;
		}
		if(entry.value == getDefaultReturnValue()) {
			long newValue = valueProvider.getLong();
			if(newValue == getDefaultReturnValue()) return newValue;
			entry.value = newValue;
		}
		return entry.value;
	}
	
	@Override
	public long computeLongIfPresent(int key, IntLongUnaryOperator mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		Entry entry = findNode(key);
		if(entry == null || entry.value == getDefaultReturnValue()) return getDefaultReturnValue();
		long newValue = mappingFunction.applyAsLong(key, entry.value);
		if(newValue == getDefaultReturnValue()) {
			removeNode(entry);
			return newValue;
		}
		entry.value = newValue;
		return newValue;
	}
	
	@Override
	public long mergeLong(int key, long value, LongLongUnaryOperator mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		Entry entry = findNode(key);
		long newValue = entry == null || entry.value == getDefaultReturnValue() ? value : mappingFunction.applyAsLong(entry.value, value);
		if(newValue == getDefaultReturnValue()) {
			if(entry != null)
				removeNode(entry);
		}
		else if(entry == null) put(key, newValue);
		else entry.value = newValue;
		return newValue;
	}
	
	@Override
	public void mergeAllLong(Int2LongMap m, LongLongUnaryOperator mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		for(Int2LongMap.Entry entry : Int2LongMaps.fastIterable(m)) {
			int key = entry.getIntKey();
			Entry subEntry = findNode(key);
			long newValue = subEntry == null || subEntry.value == getDefaultReturnValue() ? entry.getLongValue() : mappingFunction.applyAsLong(subEntry.value, entry.getLongValue());
			if(newValue == getDefaultReturnValue()) {
				if(subEntry != null)
					removeNode(subEntry);
			}
			else if(subEntry == null) put(key, newValue);
			else subEntry.value = newValue;
		}
	}
	
	@Override
	public void forEach(IntLongConsumer action) {
		for(Entry entry = first;entry != null;entry = entry.next())
			action.accept(entry.key, entry.value);
	}
	
	@Override
	public int size() { return size; }
	
	@Override
	public void clear() {
		size = 0;
		first = null;
		last = null;
		tree = null;
	}
	
	IntListIterator keyIterator(boolean descending) {
		IntListIterator iter = new KeyIterator(descending);
		return descending ? IntIterators.invert(iter) : iter;
	}
	
	IntListIterator keyIterator(int key) {
		return new KeyIterator(findNode(key));
	}
	
	public Int2LongRBTreeMap copy() {
		Int2LongRBTreeMap set = new Int2LongRBTreeMap();
		set.size = size;
		if(tree != null) {
			set.tree = tree.copy();
			Entry lastFound = null;
			for(Entry entry = tree;entry != null;entry = entry.left) lastFound = entry;
			set.first = lastFound;
			lastFound = null;
			for(Entry entry = tree;entry != null;entry = entry.right) lastFound = entry;
			set.last = lastFound;
		}
		return set;
	}
	
	@Override
	public IntSortedSet keySet() {
		return navigableKeySet();
	}
	
	@Override
	public ObjectSet<Int2LongMap.Entry> int2LongEntrySet() {
		if(entrySet == null) entrySet = new EntrySet();
		return entrySet;
	}
	
	@Override
	public LongCollection values() {
		if(values == null) values = new Values();
		return values;
	}
	
	@Override
	public IntNavigableSet navigableKeySet() {
		if(keySet == null) keySet = new KeySet(this);
		return keySet;
	}
	
	@Override
	public Int2LongNavigableMap descendingMap() {
		return new DescendingSubMap(this, true, 0, true, true, 0, true);
	}
	
	@Override
	public IntNavigableSet descendingKeySet() {
		return descendingMap().navigableKeySet();
	}
	
	@Override
	public Int2LongNavigableMap subMap(int fromKey, boolean fromInclusive, int toKey, boolean toInclusive) {
		return new AscendingSubMap(this, false, fromKey, fromInclusive, false, toKey, toInclusive);
	}
	
	@Override
	public Int2LongNavigableMap headMap(int toKey, boolean inclusive) {
		return new AscendingSubMap(this, true, 0, true, false, toKey, inclusive);
	}
	
	@Override
	public Int2LongNavigableMap tailMap(int fromKey, boolean inclusive) {
		return new AscendingSubMap(this, false, fromKey, inclusive, true, 0, true);
	}
	
	@Override
	public int lowerKey(int e) {
		Entry node = findLowerNode(e);
		return node != null ? node.key : getDefaultMinValue();
	}

	@Override
	public int floorKey(int e) {
		Entry node = findFloorNode(e);
		return node != null ? node.key : getDefaultMinValue();
	}
	
	@Override
	public int higherKey(int e) {
		Entry node = findHigherNode(e);
		return node != null ? node.key : getDefaultMaxValue();
	}

	@Override
	public int ceilingKey(int e) {
		Entry node = findCeilingNode(e);
		return node != null ? node.key : getDefaultMaxValue();
	}
	
	@Override
	public Int2LongMap.Entry lowerEntry(int key) {
		Entry node = findLowerNode(key);
		return node != null ? new BasicEntry(node.key, node.value) : null;
	}
	
	@Override
	public Int2LongMap.Entry higherEntry(int key) {
		Entry node = findHigherNode(key);
		return node != null ? new BasicEntry(node.key, node.value) : null;
	}
	
	@Override
	public Int2LongMap.Entry floorEntry(int key) {
		Entry node = findFloorNode(key);
		return node != null ? new BasicEntry(node.key, node.value) : null;
	}
	
	@Override
	public Int2LongMap.Entry ceilingEntry(int key) {
		Entry node = findCeilingNode(key);
		return node != null ? new BasicEntry(node.key, node.value) : null;
	}
	
	protected Entry findLowerNode(int key) {
		Entry entry = tree;
		while(entry != null) {
			if(compare(key, entry.key) > 0) {
				if(entry.right != null) entry = entry.right;
				else return entry;
			}
			else {
				if(entry.left != null) entry = entry.left;
				else {
					Entry parent = entry.parent;
					while(parent != null && parent.left == entry) {
						entry = parent;
						parent = parent.parent;
					}
					return parent;
				}
			}
		}
		return null;
	}
	
	protected Entry findFloorNode(int key) {
		Entry entry = tree;
		int compare;
		while(entry != null) {
			if((compare = compare(key, entry.key)) > 0) {
				if(entry.right == null) break;
				entry = entry.right;
				continue;
			}
			else if(compare < 0) {
				if(entry.left != null) entry = entry.left;
				else {
					Entry parent = entry.parent;
					while(parent != null && parent.left == entry) {
						entry = parent;
						parent = parent.parent;
					}
					return parent;
				}
				continue;
			}
			break;
		}
		return entry;
	}
	
	protected Entry findCeilingNode(int key) {
		Entry entry = tree;
		int compare;
		while(entry != null) {
			if((compare = compare(key, entry.key)) < 0) {
				if(entry.left == null) break;
				entry = entry.left;
				continue;
			}
			else if(compare > 0) {
				if(entry.right != null) entry = entry.right;
				else {
					Entry parent = entry.parent;
					while(parent != null && parent.right == entry) {
						entry = parent;
						parent = parent.parent;
					}
					return parent;
				}
				continue;
			}
			break;
		}
		return entry;
	}
	
	protected Entry findHigherNode(int key) {
		Entry entry = tree;
		while(entry != null) {
			if(compare(key, entry.key) < 0) {
				if(entry.left != null) entry = entry.left;
				else return entry;
			}
			else {
				if(entry.right != null) entry = entry.right;
				else {
					Entry parent = entry.parent;
					while(parent != null && parent.right == entry) {
						entry = parent;
						parent = parent.parent;
					}
					return parent;
				}
			}
		}
		return null;
	}
	
	protected Entry findNode(int key) {
		Entry node = tree;
		int compare;
		while(node != null) {
			if((compare = compare(key, node.key)) == 0) return node;
			if(compare < 0) node = node.left;
			else node = node.right;
		}
		return null;
	}
	
	protected void removeNode(Entry entry) {
		size--;
		if(entry.needsSuccessor()) {
			Entry successor = entry.next();
			entry.key = successor.key;
			entry.value = successor.value;
			entry = successor;
		}
		Entry replacement = entry.left != null ? entry.left : entry.right;
		if(replacement != null) {
			if(entry.replace(replacement)) tree = replacement;
			if(entry == first) first = replacement;
			if(entry == last) last = entry.right != null ? entry.right : replacement;
			entry.left = entry.right = entry.parent = null;
			if(entry.isBlack()) fixAfterDeletion(replacement);
		}
		else if(entry.parent == null) tree = first = last = null;
		else {
			if(entry.isBlack())
				fixAfterDeletion(entry);
			entry.replace(null);
			if(entry.parent != null) {
				Entry parent = entry.parent;
				if(entry == first) first = parent.left != null ? parent.left : parent;
				if(entry == last) last = entry.right != null ? parent.right : parent;
			}
			entry.parent = null;
		}
	}
	
	protected void validate(int k) { compare(k, k); }
	protected int compare(int k, int v) { return comparator != null ? comparator.compare(k, v) : Integer.compare(k, v);}
	protected static boolean isBlack(Entry p) { return p == null || p.isBlack(); }
	protected static Entry parentOf(Entry p) { return (p == null ? null : p.parent); }
	protected static void setBlack(Entry p, boolean c) { if(p != null) p.setBlack(c); }
	protected static Entry leftOf(Entry p) { return p == null ? null : p.left; }
	protected static Entry rightOf(Entry p) { return (p == null) ? null : p.right; }
	
	protected void rotateLeft(Entry entry) {
		if(entry != null) {
			Entry right = entry.right;
			entry.right = right.left;
			if(right.left != null) right.left.parent = entry;
			right.parent = entry.parent;
			if(entry.parent == null) tree = right;
			else if(entry.parent.left == entry) entry.parent.left = right;
			else entry.parent.right = right;
			right.left = entry;
			entry.parent = right;
		}
	}
	
	protected void rotateRight(Entry entry) {
		if(entry != null) {
			Entry left = entry.left;
			entry.left = left.right;
			if(left.right != null) left.right.parent = entry;
			left.parent = entry.parent;
			if(entry.parent == null) tree = left;
			else if(entry.parent.right == entry) entry.parent.right = left;
			else entry.parent.left = left;
			left.right = entry;
			entry.parent = left;
		}
	}
	
	protected void fixAfterInsertion(Entry entry) {
		entry.setBlack(false);
		while(entry != null && entry != tree && !entry.parent.isBlack()) {
			if(parentOf(entry) == leftOf(parentOf(parentOf(entry)))) {
				Entry y = rightOf(parentOf(parentOf(entry)));
				if(!isBlack(y)) {
					setBlack(parentOf(entry), true);
					setBlack(y, true);
					setBlack(parentOf(parentOf(entry)), false);
					entry = parentOf(parentOf(entry));
				}
				else {
					if(entry == rightOf(parentOf(entry))) {
						entry = parentOf(entry);
						rotateLeft(entry);
					}
					setBlack(parentOf(entry), true);
					setBlack(parentOf(parentOf(entry)), false);
					rotateRight(parentOf(parentOf(entry)));
				}
			}
			else {
				Entry y = leftOf(parentOf(parentOf(entry)));
				if(!isBlack(y)) {
					setBlack(parentOf(entry), true);
					setBlack(y, true);
					setBlack(parentOf(parentOf(entry)), false);
					entry = parentOf(parentOf(entry));
				}
				else {
					if(entry == leftOf(parentOf(entry))) {
						entry = parentOf(entry);
						rotateRight(entry);
					}
					setBlack(parentOf(entry), true);
					setBlack(parentOf(parentOf(entry)), false);
					rotateLeft(parentOf(parentOf(entry)));
				}
			}
		}
		tree.setBlack(true);
	}
	
	protected void fixAfterDeletion(Entry entry) {
		while(entry != tree && isBlack(entry)) {
			if(entry == leftOf(parentOf(entry))) {
				Entry sib = rightOf(parentOf(entry));
				if(!isBlack(sib)) {
					setBlack(sib, true);
					setBlack(parentOf(entry), false);
					rotateLeft(parentOf(entry));
					sib = rightOf(parentOf(entry));
				}
				if(isBlack(leftOf(sib)) && isBlack(rightOf(sib))) {
					setBlack(sib, false);
					entry = parentOf(entry);
				}
				else {
					if(isBlack(rightOf(sib))) {
						setBlack(leftOf(sib), true);
						setBlack(sib, false);
						rotateRight(sib);
						sib = rightOf(parentOf(entry));
					}
					setBlack(sib, isBlack(parentOf(entry)));
					setBlack(parentOf(entry), true);
					setBlack(rightOf(sib), true);
					rotateLeft(parentOf(entry));
					entry = tree;
				}
			}
			else {
				Entry sib = leftOf(parentOf(entry));
				if(!isBlack(sib)) {
					setBlack(sib, true);
					setBlack(parentOf(entry), false);
					rotateRight(parentOf(entry));
					sib = leftOf(parentOf(entry));
				}
				if(isBlack(rightOf(sib)) && isBlack(leftOf(sib))) {
					setBlack(sib, false);
					entry = parentOf(entry);
				}
				else {
					if(isBlack(leftOf(sib))) {
						setBlack(rightOf(sib), true);
						setBlack(sib, false);
						rotateLeft(sib);
						sib = leftOf(parentOf(entry));
					}
					setBlack(sib, isBlack(parentOf(entry)));
					setBlack(parentOf(entry), true);
					setBlack(leftOf(sib), true);
					rotateRight(parentOf(entry));
					entry = tree;
				}
			}
		}
		setBlack(entry, true);
	}
	
	static final class AscendingSubMap extends NavigableSubMap {

		public AscendingSubMap(Int2LongRBTreeMap m, boolean fromStart, int low, boolean loInclusive, boolean toEnd, int high, boolean hiInclusive) {
			super(m, fromStart, low, loInclusive, toEnd, high, hiInclusive);
		}
		
		@Override
		IntListIterator keyIterator(boolean descending) { return new SubMapKeyIterator(descending); }
		@Override
		IntListIterator keyIterator(int key) {
			Int2LongRBTreeMap.Entry entry = m.findNode(key);
			return entry == null || !inClosedRange(key) ? null : new SubMapKeyIterator(entry, fromStart ? null : findLowest(), toEnd ? null : findHighest());
		}
		@Override
		LongListIterator valueIterator() { return new SubMapValueIterator(false); }
		@Override
		ObjectListIterator<Int2LongMap.Entry> entryIterator() { return new SubMapEntrySetIterator(false); }
		
		@Override
		public Int2LongNavigableMap descendingMap() {
			if(descendingMap == null) descendingMap = new DescendingSubMap(m, fromStart, low, loInclusive, toEnd, high, hiInclusive);
			return descendingMap;
		}
		
		Int2LongRBTreeMap.Entry next(Int2LongRBTreeMap.Entry entry) {
			return entry.next();
		}
		
		Int2LongRBTreeMap.Entry previous(Int2LongRBTreeMap.Entry entry) {
			return entry.previous();
		}
		
		@Override
		public Int2LongNavigableMap subMap(int fromKey, boolean fromInclusive, int toKey, boolean toInclusive) {
			if (!inRange(fromKey, fromInclusive)) throw new IllegalArgumentException("fromKey out of range");
			if (!inRange(toKey, toInclusive)) throw new IllegalArgumentException("toKey out of range");
			return new AscendingSubMap(m, false, fromKey, fromInclusive, false, toKey, toInclusive);
		}
		
		@Override
		public Int2LongNavigableMap headMap(int toKey, boolean inclusive) {
			if (!inRange(toKey, inclusive)) throw new IllegalArgumentException("toKey out of range");
			return new AscendingSubMap(m, fromStart, low, loInclusive, false, toKey, inclusive);
		}
		
		@Override
		public Int2LongNavigableMap tailMap(int fromKey, boolean inclusive) {
			if (!inRange(fromKey, inclusive)) throw new IllegalArgumentException("fromKey out of range");
			return new AscendingSubMap(m, false, fromKey, inclusive, toEnd, high, hiInclusive);
		}
	}
	
	static final class DescendingSubMap extends NavigableSubMap {

		public DescendingSubMap(Int2LongRBTreeMap m, boolean fromStart, int low, boolean loInclusive, boolean toEnd, int high, boolean hiInclusive) {
			super(m, fromStart, low, loInclusive, toEnd, high, hiInclusive);
		}
		
		@Override
		public int firstIntKey() { return super.lastIntKey(); }
		
		@Override
		public int pollFirstIntKey() { return super.pollLastIntKey(); }
		
		@Override
		public int lastIntKey() { return super.firstIntKey(); }
		
		@Override
		public int pollLastIntKey() { return super.pollFirstIntKey(); }
		
		@Override
		public long firstLongValue() { return super.lastLongValue(); }
		
		@Override
		public long lastLongValue() { return super.firstLongValue(); }
		
		@Override
		public Int2LongMap.Entry firstEntry() { return super.lastEntry(); }
		
		@Override
		public Int2LongMap.Entry lastEntry() { return super.firstEntry(); }
		
		@Override
		public Int2LongMap.Entry pollFirstEntry() { return super.pollLastEntry(); }
		
		@Override
		public Int2LongMap.Entry pollLastEntry() { return super.pollFirstEntry(); }
		
		@Override
		public int lowerKey(int e) { return super.higherKey(e); }
		
		@Override
		public int floorKey(int e) { return super.ceilingKey(e); }
		
		@Override
		public int ceilingKey(int e) { return super.floorKey(e); }
		
		@Override
		public int higherKey(int e) { return super.lowerKey(e); }
		
		@Override
		public Int2LongMap.Entry lowerEntry(int e) { return super.higherEntry(e); }
		
		@Override
		public Int2LongMap.Entry floorEntry(int e) { return super.ceilingEntry(e); }
		
		@Override
		public Int2LongMap.Entry ceilingEntry(int e) { return super.floorEntry(e); }
		
		@Override
		public Int2LongMap.Entry higherEntry(int e) { return super.lowerEntry(e); }
		
		protected Int2LongRBTreeMap.Entry findLowest() {
			if(fromStart) return m.first;
			Int2LongRBTreeMap.Entry entry = loInclusive ? m.findCeilingNode(low) : m.findHigherNode(low);
			return entry == null || tooHigh(entry.key) ? null : entry;
		}

		protected Int2LongRBTreeMap.Entry findHighest() {
			if(toEnd) return m.last;
			Int2LongRBTreeMap.Entry entry = hiInclusive ? m.findFloorNode(high) : m.findLowerNode(high);
			return entry == null || tooLow(entry.key) ? null : entry;
		}
		
		@Override
		IntListIterator keyIterator(boolean descending) {
			IntListIterator iter = new SubMapKeyIterator(!descending);
			return descending ? iter : IntIterators.invert(iter);
		}
		
		@Override
		IntListIterator keyIterator(int key) {
			Int2LongRBTreeMap.Entry entry = m.findNode(key);
			return entry == null || !inClosedRange(key) ? null : IntIterators.invert(new SubMapKeyIterator(entry, fromStart ? null : findLowest(), toEnd ? null : findHighest()));
		}
		
		Int2LongRBTreeMap.Entry next(Int2LongRBTreeMap.Entry entry) {
			return entry.previous();
		}
		
		Int2LongRBTreeMap.Entry previous(Int2LongRBTreeMap.Entry entry) {
			return entry.next();
		}
		
		@Override
		public IntComparator comparator() { return m.comparator() == null ? null : m.comparator().reversed(); }
		
		@Override
		LongListIterator valueIterator() { return LongIterators.invert(new SubMapValueIterator(true)); }
		
		@Override
		ObjectListIterator<Int2LongMap.Entry> entryIterator() { return ObjectIterators.invert(new SubMapEntrySetIterator(true)); }
		
		@Override
		public Int2LongNavigableMap descendingMap() {
			if(descendingMap == null) descendingMap = new AscendingSubMap(m, fromStart, low, loInclusive, toEnd, high, hiInclusive);
			return descendingMap;
		}

		@Override
		public Int2LongNavigableMap subMap(int fromKey, boolean fromInclusive, int toKey, boolean toInclusive) {
			if (!inRange(fromKey, fromInclusive)) throw new IllegalArgumentException("fromKey out of range");
			if (!inRange(toKey, toInclusive)) throw new IllegalArgumentException("toKey out of range");
			return new DescendingSubMap(m, false, fromKey, fromInclusive, false, toKey, toInclusive);
		}
		
		@Override
		public Int2LongNavigableMap headMap(int toKey, boolean inclusive) {
			if (!inRange(toKey, inclusive)) throw new IllegalArgumentException("toKey out of range");
			return new DescendingSubMap(m, fromStart, low, loInclusive, false, toKey, inclusive);
		}
		
		@Override
		public Int2LongNavigableMap tailMap(int fromKey, boolean inclusive) {
			if (!inRange(fromKey, inclusive)) throw new IllegalArgumentException("fromKey out of range");
			return new DescendingSubMap(m, false, fromKey, inclusive, toEnd, high, hiInclusive);
		}
	}
	
	static abstract class NavigableSubMap extends AbstractInt2LongMap implements Int2LongNavigableMap {
		final Int2LongRBTreeMap m;
		final int low;
		final int high;
		final boolean fromStart;
		final boolean toEnd;
		final boolean loInclusive;
		final boolean hiInclusive;
		transient Int2LongNavigableMap descendingMap;
		transient SubMapEntrySet entrySet;
		transient KeySet keySet;
		transient SubMapValues values;
		
		NavigableSubMap(Int2LongRBTreeMap m, boolean fromStart, int low, boolean loInclusive, boolean toEnd, int high, boolean hiInclusive) {
			this.m = m;
			this.low = low;
			this.high = high;
			this.fromStart = fromStart;
			this.toEnd = toEnd;
			this.loInclusive = loInclusive;
			this.hiInclusive = hiInclusive;
		}
		
		abstract IntListIterator keyIterator(boolean descending);
		abstract IntListIterator keyIterator(int key);
		abstract LongListIterator valueIterator();
		abstract ObjectListIterator<Int2LongMap.Entry> entryIterator();
		abstract Int2LongRBTreeMap.Entry next(Int2LongRBTreeMap.Entry entry);
		abstract Int2LongRBTreeMap.Entry previous(Int2LongRBTreeMap.Entry entry);
		
		@Override
		public NavigableSubMap copy() { throw new UnsupportedOperationException(); }
		@Override
		public abstract Int2LongNavigableMap descendingMap();
		@Override
		public abstract Int2LongNavigableMap subMap(int fromKey, boolean fromInclusive, int toKey, boolean toInclusive);
		@Override
		public abstract Int2LongNavigableMap headMap(int toKey, boolean inclusive);
		@Override
		public abstract Int2LongNavigableMap tailMap(int fromKey, boolean inclusive);
		
		boolean tooLow(int key) { return !fromStart && (loInclusive ? m.compare(key, low) < 0 : m.compare(key, low) <= 0); }
		boolean tooHigh(int key) { return !toEnd && (hiInclusive ? m.compare(key, high) > 0 : m.compare(key, high) >= 0); }
		boolean inRange(int key) { return !tooLow(key) && !tooHigh(key); }
		boolean inClosedRange(int key) { return (fromStart || m.compare(key, low) >= 0) && (toEnd || m.compare(high, key) >= 0); }
		boolean inRange(int key, boolean inclusive) { return inclusive ? inRange(key) : inClosedRange(key); }
		
		@Override
		public void setDefaultMaxValue(int e) { m.setDefaultMaxValue(e); }
		@Override
		public int getDefaultMaxValue() { return m.getDefaultMaxValue(); }
		@Override
		public void setDefaultMinValue(int e) { m.setDefaultMinValue(e); }
		@Override
		public int getDefaultMinValue() { return m.getDefaultMinValue(); }
		
		@Override
		public AbstractInt2LongMap setDefaultReturnValue(long v) { 
			m.setDefaultReturnValue(v);
			return this;
		}
		
		@Override
		public long getDefaultReturnValue() { return m.getDefaultReturnValue(); }
		
		@Override
		public long putAndMoveToFirst(int key, long value) { throw new UnsupportedOperationException(); }
		@Override
		public long putAndMoveToLast(int key, long value) { throw new UnsupportedOperationException(); }
		@Override
		public boolean moveToFirst(int key) { throw new UnsupportedOperationException(); }
		@Override
		public boolean moveToLast(int key) { throw new UnsupportedOperationException(); }
		@Override
		public long getAndMoveToFirst(int key) { throw new UnsupportedOperationException(); }
		@Override
		public long getAndMoveToLast(int key) { throw new UnsupportedOperationException(); }
		@Override
		public IntComparator comparator() { return m.comparator(); }
		@Override
		public boolean containsKey(int key) { return inRange(key) && m.containsKey(key); }
		
		@Override
		public int firstIntKey() { 
			Int2LongRBTreeMap.Entry entry = findLowest();
			return entry == null ? getDefaultMaxValue() : entry.key;
		}
		
		@Override
		public int pollFirstIntKey() {
			if(fromStart) return m.pollFirstIntKey();
			Int2LongRBTreeMap.Entry entry = loInclusive ? m.findCeilingNode(low) : m.findHigherNode(low);
			if(entry != null && !tooHigh(entry.key)) {
				int value = entry.key;
				m.removeNode(entry);
				return value;
			}
			return getDefaultMaxValue();
		}
		
		@Override
		public int lastIntKey() {
			Int2LongRBTreeMap.Entry entry = findHighest();
			return entry == null ? getDefaultMinValue() : entry.key;	
		}
		
		@Override
		public int pollLastIntKey() {
			if(toEnd) return m.pollLastIntKey();
			Int2LongRBTreeMap.Entry entry = hiInclusive ? m.findFloorNode(high) : m.findLowerNode(high);
			if(entry != null && !tooLow(entry.key)) {
				int value = entry.key;
				m.removeNode(entry);
				return value;
			}
			return getDefaultMinValue();
		}
		
		@Override
		public long firstLongValue() {
			Int2LongRBTreeMap.Entry entry = findLowest();
			return entry == null ? getDefaultReturnValue() : entry.value;
		}
		
		@Override
		public long lastLongValue() {
			Int2LongRBTreeMap.Entry entry = findHighest();
			return entry == null ? getDefaultReturnValue() : entry.value;
		}
		
		protected Int2LongRBTreeMap.Entry findLowest() {
			if(fromStart) return m.first;
			Int2LongRBTreeMap.Entry entry = loInclusive ? m.findCeilingNode(low) : m.findHigherNode(low);
			return entry == null || tooHigh(entry.key) ? null : entry;
		}
		
		protected Int2LongRBTreeMap.Entry findHighest() {
			if(toEnd) return m.last;
			Int2LongRBTreeMap.Entry entry = hiInclusive ? m.findFloorNode(high) : m.findLowerNode(high);
			return entry == null || tooLow(entry.key) ? null : entry;
		}
		
		@Override
		public long put(int key, long value) {
			if(!inRange(key)) throw new IllegalArgumentException("key out of range");
			return m.put(key, value);
		}
		
		@Override
		public long putIfAbsent(int key, long value) {
			if(!inRange(key)) throw new IllegalArgumentException("key out of range");
			return m.putIfAbsent(key, value);
		}
		
		@Override
		public long addTo(int key, long value) {
			if(!inRange(key)) throw new IllegalArgumentException("key out of range");
			return m.addTo(key, value);
		}
		
		@Override
		public long get(int key) {
			return inRange(key) ? m.get(key) : getDefaultReturnValue();
		}
		
		@Override
		public long getOrDefault(int key, long defaultValue) {
			return inRange(key) ? m.getOrDefault(key, defaultValue) : getDefaultReturnValue();
		}
		
		@Override
		public long remove(int key) {
			return inRange(key) ? m.remove(key) : getDefaultReturnValue();
		}
		
		@Override
		public long removeOrDefault(int key, long defaultValue) {
			return inRange(key) ? m.remove(key) : defaultValue;
		}
		
		@Override
		public boolean remove(int key, long value) {
			return inRange(key) && m.remove(key, value);
		}
		
		@Override
		public ObjectSet<Int2LongMap.Entry> int2LongEntrySet() {
			if(entrySet == null) entrySet = new SubMapEntrySet();
			return entrySet;
		}
		
		@Override
		public IntSet keySet() { return navigableKeySet(); }
		
		@Override
		public IntNavigableSet navigableKeySet() {
			if(keySet == null) keySet = new KeySet(this);
			return keySet;
		}
		
		@Override
		public IntNavigableSet descendingKeySet() {
			return descendingMap().navigableKeySet();
		}
		
		@Override
		public LongCollection values() {
			if(values == null) values = new SubMapValues();
			return values;
		}
		
		@Override
		public void forEach(IntLongConsumer action) {
			for(Int2LongRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
				action.accept(entry.key, entry.value);
		}
		
		@Override
		public int size() {
			return (fromStart && toEnd) ? m.size() : entrySet().size();
		}
		
		@Override
		public Int2LongMap.Entry firstEntry() {
			Int2LongRBTreeMap.Entry entry = findLowest();
			return entry == null ? null : new BasicEntry(entry.key, entry.value);
		}
		
		@Override
		public Int2LongMap.Entry lastEntry() {
			Int2LongRBTreeMap.Entry entry = findHighest();
			return entry == null ? null : new BasicEntry(entry.key, entry.value);
		}
		
		@Override
		public Int2LongMap.Entry pollFirstEntry() {
			Int2LongRBTreeMap.Entry entry = findLowest();
			if(entry == null) return null;
			BasicEntry result = new BasicEntry(entry.key, entry.value);
			m.removeNode(entry);
			return result;
		}
		
		@Override
		public Int2LongMap.Entry pollLastEntry() {
			Int2LongRBTreeMap.Entry entry = findHighest();
			if(entry == null) return null;
			BasicEntry result = new BasicEntry(entry.key, entry.value);
			m.removeNode(entry);
			return result;
		}
		
		@Override
		public int lowerKey(int e) {
			if(tooHigh(e)) {
				Int2LongRBTreeMap.Entry entry = findHighest();
				return entry == null ? getDefaultMinValue() : entry.key;
			}
			Int2LongRBTreeMap.Entry entry = m.findLowerNode(e);
			return entry == null || tooHigh(entry.key) ? getDefaultMaxValue() : entry.key;
		}
		
		@Override
		public int floorKey(int e) {
			if(tooHigh(e)) {
				Int2LongRBTreeMap.Entry entry = findHighest();
				return entry == null ? getDefaultMinValue() : entry.key;
			}
			Int2LongRBTreeMap.Entry entry = m.findFloorNode(e);
			return entry == null || tooHigh(entry.key) ? getDefaultMaxValue() : entry.key;
		}
		
		@Override
		public int ceilingKey(int e) {
			if(tooLow(e)) {
				Int2LongRBTreeMap.Entry entry = findLowest();
				return entry == null ? getDefaultMaxValue() : entry.key;
			}
			Int2LongRBTreeMap.Entry entry = m.findCeilingNode(e);
			return entry == null || tooLow(entry.key) ? getDefaultMinValue() : entry.key;
		}
		
		@Override
		public int higherKey(int e) {
			if(tooLow(e)) {
				Int2LongRBTreeMap.Entry entry = findLowest();
				return entry == null ? getDefaultMaxValue() : entry.key;
			}
			Int2LongRBTreeMap.Entry entry = m.findHigherNode(e);
			return entry == null || tooLow(entry.key) ? getDefaultMinValue() : entry.key;
		}
		
		@Override
		public Int2LongMap.Entry lowerEntry(int e) {
			if(tooHigh(e)) {
				Int2LongRBTreeMap.Entry entry = findHighest();
				return entry == null ? null : new BasicEntry(entry.key, entry.value);
			}
			Int2LongRBTreeMap.Entry entry = m.findLowerNode(e);
			return entry == null || tooHigh(entry.key) ? null : new BasicEntry(entry.key, entry.value);
		}
		
		@Override
		public Int2LongMap.Entry floorEntry(int e) {
			if(tooHigh(e)) {
				Int2LongRBTreeMap.Entry entry = findHighest();
				return entry == null ? null : new BasicEntry(entry.key, entry.value);
			}
			Int2LongRBTreeMap.Entry entry = m.findFloorNode(e);
			return entry == null || tooHigh(entry.key) ? null : new BasicEntry(entry.key, entry.value);
		}
		
		@Override
		public Int2LongMap.Entry ceilingEntry(int e) {
			if(tooLow(e)) {
				Int2LongRBTreeMap.Entry entry = findLowest();
				return entry == null ? null : new BasicEntry(entry.key, entry.value);
			}
			Int2LongRBTreeMap.Entry entry = m.findCeilingNode(e);
			return entry == null || tooLow(entry.key) ? null : new BasicEntry(entry.key, entry.value);
		}
		
		@Override
		public Int2LongMap.Entry higherEntry(int e) {
			if(tooLow(e)) {
				Int2LongRBTreeMap.Entry entry = findLowest();
				return entry == null ? null : new BasicEntry(entry.key, entry.value);
			}
			Int2LongRBTreeMap.Entry entry = m.findHigherNode(e);
			return entry == null || tooLow(entry.key) ? null : new BasicEntry(entry.key, entry.value);
		}
		
		class SubMapEntrySet extends AbstractObjectSet<Int2LongMap.Entry> implements ObjectSet<Int2LongMap.Entry> {
			@Override
			@Deprecated
			public boolean contains(Object o) {
				if(o instanceof Map.Entry) {
					if(o instanceof Int2LongMap.Entry) {
						Int2LongMap.Entry entry = (Int2LongMap.Entry)o;
						if(entry.getKey() == null && comparator() == null) return false;
						Int2LongRBTreeMap.Entry subEntry = m.findNode(entry.getIntKey());
						if(subEntry != null) return entry.getLongValue() == subEntry.value;
					}
					else {
						Map.Entry<?, ?> entry = (Map.Entry<?, ?>)o;
						if(entry.getKey() == null && comparator() == null) return false;
						if(!(entry.getKey() instanceof Integer)) return false;
						Int2LongRBTreeMap.Entry subEntry = m.findNode((Integer)entry.getKey());
						if(subEntry != null) return Objects.equals(entry.getValue(), Long.valueOf(subEntry.value));
					}
				}
				return false;
			}
			
			@Override
			@Deprecated
			public boolean remove(Object o) {
				if(o instanceof Map.Entry) {
					if(o instanceof Int2LongMap.Entry) {
						Int2LongMap.Entry entry = (Int2LongMap.Entry)o;
						return NavigableSubMap.this.remove(entry.getIntKey(), entry.getLongValue());
					}
					Map.Entry<?, ?> entry = (Map.Entry<?, ?>)o;
					return NavigableSubMap.this.remove(entry.getKey(), entry.getValue());
				}
				return false;
			}
			@Override
			public ObjectIterator<Int2LongMap.Entry> iterator() { return entryIterator(); }
			@Override
			public int size() { return NavigableSubMap.this.size(); }
			@Override
			public void clear() { NavigableSubMap.this.clear(); }
			
			@Override
			public void forEach(Consumer<? super Int2LongMap.Entry> action) {
				Objects.requireNonNull(action);
				for(Int2LongRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					action.accept(new BasicEntry(entry.key, entry.value));
			}
			
			@Override
			public <E> void forEach(E input, ObjectObjectConsumer<E, Int2LongMap.Entry> action) {
				Objects.requireNonNull(action);
				for(Int2LongRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					action.accept(input, new BasicEntry(entry.key, entry.value));
			}
			
			@Override
			public boolean matchesAny(Object2BooleanFunction<Int2LongMap.Entry> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return false;
				BasicEntry subEntry = new BasicEntry();
				for(Int2LongRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					subEntry.set(entry.key, entry.value);
					if(filter.getBoolean(subEntry)) return true;
				}
				return false;
			}
			
			@Override
			public boolean matchesNone(Object2BooleanFunction<Int2LongMap.Entry> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return true;
				BasicEntry subEntry = new BasicEntry();
				for(Int2LongRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					subEntry.set(entry.key, entry.value);
					if(filter.getBoolean(subEntry)) return false;
				}
				return true;
			}
			
			@Override
			public boolean matchesAll(Object2BooleanFunction<Int2LongMap.Entry> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return true;
				BasicEntry subEntry = new BasicEntry();
				for(Int2LongRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					subEntry.set(entry.key, entry.value);
					if(!filter.getBoolean(subEntry)) return false;
				}
				return true;
			}
			
			@Override
			public <E> E reduce(E identity, BiFunction<E, Int2LongMap.Entry, E> operator) {
				Objects.requireNonNull(operator);
				E state = identity;
				for(Int2LongRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					state = operator.apply(state, new BasicEntry(entry.key, entry.value));
				}
				return state;
			}
			
			@Override
			public Int2LongMap.Entry reduce(ObjectObjectUnaryOperator<Int2LongMap.Entry, Int2LongMap.Entry> operator) {
				Objects.requireNonNull(operator);
				Int2LongMap.Entry state = null;
				boolean empty = true;
				for(Int2LongRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					if(empty) {
						empty = false;
						state = new BasicEntry(entry.key, entry.value);
						continue;
					}
					state = operator.apply(state, new BasicEntry(entry.key, entry.value));
				}
				return state;
			}
			
			@Override
			public Int2LongMap.Entry findFirst(Object2BooleanFunction<Int2LongMap.Entry> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return null;
				BasicEntry subEntry = new BasicEntry();
				for(Int2LongRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					subEntry.set(entry.key, entry.value);
					if(filter.getBoolean(subEntry)) return subEntry;
				}
				return null;
			}
			
			@Override
			public int count(Object2BooleanFunction<Int2LongMap.Entry> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return 0;
				int result = 0;
				BasicEntry subEntry = new BasicEntry();
				for(Int2LongRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					subEntry.set(entry.key, entry.value);
					if(filter.getBoolean(subEntry)) result++;
				}
				return result;
			}
		}
		
		final class SubMapValues extends AbstractLongCollection {
			@Override
			public boolean add(long o) { throw new UnsupportedOperationException(); }
			
			@Override
			public boolean contains(long e) {
				return containsValue(e);
			}
			
			@Override
			public LongIterator iterator() { return valueIterator(); }
			
			@Override
			public int size() {
				return NavigableSubMap.this.size();
			}
			
			@Override
			public void clear() {
				NavigableSubMap.this.clear();
			}
			
			@Override
			public void forEach(LongConsumer action) {
				Objects.requireNonNull(action);
				for(Int2LongRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					action.accept(entry.value);
			}
			
			@Override
			public <E> void forEach(E input, ObjectLongConsumer<E> action) {
				Objects.requireNonNull(action);
				for(Int2LongRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					action.accept(input, entry.value);
			}
			
			@Override
			public boolean matchesAny(Long2BooleanFunction filter) {
				Objects.requireNonNull(filter);
				for(Int2LongRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					if(filter.get(entry.value)) return true;
				return false;
			}
			
			@Override
			public boolean matchesNone(Long2BooleanFunction filter) {
				Objects.requireNonNull(filter);
				for(Int2LongRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					if(filter.get(entry.value)) return false;
				return true;
			}
			
			@Override
			public boolean matchesAll(Long2BooleanFunction filter) {
				Objects.requireNonNull(filter);
				for(Int2LongRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					if(!filter.get(entry.value)) return false;
				return true;
			}
			
			@Override
			public long reduce(long identity, LongLongUnaryOperator operator) {
				Objects.requireNonNull(operator);
				long state = identity;
				for(Int2LongRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					state = operator.applyAsLong(state, entry.value);
				return state;
			}
			
			@Override
			public long reduce(LongLongUnaryOperator operator) {
				Objects.requireNonNull(operator);
				long state = 0L;
				boolean empty = true;
				for(Int2LongRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					if(empty) {
						empty = false;
						state = entry.value;
						continue;
					}
					state = operator.applyAsLong(state, entry.value);
				}
				return state;
			}
			
			@Override
			public long findFirst(Long2BooleanFunction filter) {
				Objects.requireNonNull(filter);
				for(Int2LongRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					if(filter.get(entry.value)) return entry.value;
				return 0L;
			}
			
			@Override
			public int count(Long2BooleanFunction filter) {
				Objects.requireNonNull(filter);
				int result = 0;
				for(Int2LongRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					if(filter.get(entry.value)) result++;
				return result;
			}
		}
		
		final class SubMapEntrySetIterator extends SubMapEntryIterator implements ObjectListIterator<Int2LongMap.Entry> {
			public SubMapEntrySetIterator(boolean descending) {
				super(descending);
			}
			
			public SubMapEntrySetIterator(Int2LongRBTreeMap.Entry entry, Int2LongRBTreeMap.Entry lowerFence, Int2LongRBTreeMap.Entry upperFence) {
				super(entry, lowerFence, upperFence);
			}
			
			@Override
			public Int2LongMap.Entry next() { return nextEntry(); }
			@Override
			public Int2LongMap.Entry previous() { return previousEntry(); }
			
			@Override
			public void set(Int2LongMap.Entry e) { throw new UnsupportedOperationException(); }
			@Override
			public void add(Int2LongMap.Entry e) { throw new UnsupportedOperationException(); }
		}
		
		final class SubMapKeyIterator extends SubMapEntryIterator implements IntListIterator {
			public SubMapKeyIterator(boolean descending) {
				super(descending);
			}
			
			public SubMapKeyIterator(Int2LongRBTreeMap.Entry entry, Int2LongRBTreeMap.Entry lowerFence, Int2LongRBTreeMap.Entry upperFence) {
				super(entry, lowerFence, upperFence);
			}
			
			@Override
			public int nextInt() { return nextEntry().key; }
			@Override
			public int previousInt() { return previousEntry().key; }
			
			@Override
			public void set(int e) { throw new UnsupportedOperationException(); }
			@Override
			public void add(int e) { throw new UnsupportedOperationException(); }
		}
		
		final class SubMapValueIterator extends SubMapEntryIterator implements LongListIterator {
			public SubMapValueIterator(boolean descending) {
				super(descending);
			}
			
			public SubMapValueIterator(Int2LongRBTreeMap.Entry entry, Int2LongRBTreeMap.Entry lowerFence, Int2LongRBTreeMap.Entry upperFence) {
				super(entry, lowerFence, upperFence);
			}
			
			@Override
			public long nextLong() { return nextEntry().value; }
			@Override
			public long previousLong() { return previousEntry().value; }
			
			@Override
			public void set(long e) { throw new UnsupportedOperationException(); }
			@Override
			public void add(long e) { throw new UnsupportedOperationException(); }
		}
		
		abstract class SubMapEntryIterator {
			Integer lowerFence;
			Integer upperFence;
			Int2LongRBTreeMap.Entry next;
			Int2LongRBTreeMap.Entry previous;
			Int2LongRBTreeMap.Entry current;
			int index = 0;
			
			public SubMapEntryIterator(boolean descending) {
				this(descending ? findHighest() : findLowest(), fromStart ? null : findLowest(), toEnd ? null : findHighest());
			}
			
			public SubMapEntryIterator(Int2LongRBTreeMap.Entry entry, Int2LongRBTreeMap.Entry lowerFence, Int2LongRBTreeMap.Entry upperFence) {
				next = entry;
				previous = entry.previous();
				this.lowerFence = lowerFence != null ? Integer.valueOf(lowerFence.key) : null;
				this.upperFence = upperFence != null ? Integer.valueOf(upperFence.key) : null;
			}
			
			public boolean hasNext() {
				return next != null && (upperFence == null || next.key == upperFence.intValue());
			}
			
			public boolean hasPrevious() {
				return previous != null && (lowerFence == null || next.key == lowerFence.intValue());
			}
			
			public int nextIndex() {
				return index;
			}
			
			public int previousIndex() {
				return index - 1;
			}
			
			protected void updateNext() {
				next = current.next();
			}
			
			protected void updatePrevious() {
				previous = current.previous();
			}
			
			public Int2LongRBTreeMap.Entry nextEntry() {
				if(!hasNext()) throw new NoSuchElementException();
				current = previous = next;
				updateNext();
				index++;
				return current;
			}
			
			public Int2LongRBTreeMap.Entry previousEntry() {
				if(!hasPrevious()) throw new NoSuchElementException();
				current = next = previous;
				updatePrevious();
				index--;
				return current;
			}
			
			public void remove() {
				if(current == null) throw new IllegalStateException();
				if(current == previous) index--;
				updateNext();
				updatePrevious();
				if(current.needsSuccessor()) next = current;
				m.removeNode(current);
				current = null;
			}
		}
	}
	
	private class EntrySet extends AbstractObjectSet<Int2LongMap.Entry> {
		
		@Override
		public ObjectIterator<Int2LongMap.Entry> iterator() {
			return new EntryIterator(false);
		}
		
		@Override
		@Deprecated
		public boolean contains(Object o) {
			if(o instanceof Map.Entry) {
				if(o instanceof Int2LongMap.Entry) {
					Int2LongMap.Entry entry = (Int2LongMap.Entry)o;
					if(entry.getKey() == null && comparator() == null) return false;
					Int2LongRBTreeMap.Entry subEntry = Int2LongRBTreeMap.this.findNode(entry.getIntKey());
					if(subEntry != null) return entry.getLongValue() == subEntry.value;
				}
				else {
					Map.Entry<?, ?> entry = (Map.Entry<?, ?>)o;
					if(entry.getKey() == null && comparator() == null) return false;
					if(!(entry.getKey() instanceof Integer)) return false;
					Int2LongRBTreeMap.Entry subEntry = Int2LongRBTreeMap.this.findNode((Integer)entry.getKey());
					if(subEntry != null) return Objects.equals(entry.getValue(), Long.valueOf(subEntry.value));
				}
			}
			return false;
		}
		
		@Override
		@Deprecated
		public boolean remove(Object o) {
			if(o instanceof Map.Entry) {
				if(o instanceof Int2LongMap.Entry) {
					Int2LongMap.Entry entry = (Int2LongMap.Entry)o;
					return Int2LongRBTreeMap.this.remove(entry.getIntKey(), entry.getLongValue());
				}
				Map.Entry<?, ?> entry = (Map.Entry<?, ?>)o;
				return Int2LongRBTreeMap.this.remove(entry.getKey(), entry.getValue());
			}
			return false;
		}
		
		@Override
		public int size() {
			return Int2LongRBTreeMap.this.size();
		}
		
		@Override
		public void clear() {
			Int2LongRBTreeMap.this.clear();
		}
		
		@Override
		public void forEach(Consumer<? super Int2LongMap.Entry> action) {
			Objects.requireNonNull(action);
			for(Int2LongRBTreeMap.Entry entry = first;entry != null;entry = entry.next())
				action.accept(new BasicEntry(entry.key, entry.value));
		}
		
		@Override
		public <E> void forEach(E input, ObjectObjectConsumer<E, Int2LongMap.Entry> action) {
			Objects.requireNonNull(action);
			for(Int2LongRBTreeMap.Entry entry = first;entry != null;entry = entry.next())
				action.accept(input, new BasicEntry(entry.key, entry.value));
		}
		
		@Override
		public boolean matchesAny(Object2BooleanFunction<Int2LongMap.Entry> filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return false;
			BasicEntry subEntry = new BasicEntry();
			for(Int2LongRBTreeMap.Entry entry = first;entry != null;entry = entry.next()) {
				subEntry.set(entry.key, entry.value);
				if(filter.getBoolean(subEntry)) return true;
			}
			return false;
		}
		
		@Override
		public boolean matchesNone(Object2BooleanFunction<Int2LongMap.Entry> filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return true;
			BasicEntry subEntry = new BasicEntry();
			for(Int2LongRBTreeMap.Entry entry = first;entry != null;entry = entry.next()) {
				subEntry.set(entry.key, entry.value);
				if(filter.getBoolean(subEntry)) return false;
			}
			return true;
		}
		
		@Override
		public boolean matchesAll(Object2BooleanFunction<Int2LongMap.Entry> filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return true;
			BasicEntry subEntry = new BasicEntry();
			for(Int2LongRBTreeMap.Entry entry = first;entry != null;entry = entry.next()) {
				subEntry.set(entry.key, entry.value);
				if(!filter.getBoolean(subEntry)) return false;
			}
			return true;
		}
		
		@Override
		public <E> E reduce(E identity, BiFunction<E, Int2LongMap.Entry, E> operator) {
			Objects.requireNonNull(operator);
			E state = identity;
			for(Int2LongRBTreeMap.Entry entry = first;entry != null;entry = entry.next()) {
				state = operator.apply(state, new BasicEntry(entry.key, entry.value));
			}
			return state;
		}
		
		@Override
		public Int2LongMap.Entry reduce(ObjectObjectUnaryOperator<Int2LongMap.Entry, Int2LongMap.Entry> operator) {
			Objects.requireNonNull(operator);
			Int2LongMap.Entry state = null;
			boolean empty = true;
			for(Int2LongRBTreeMap.Entry entry = first;entry != null;entry = entry.next()) {
				if(empty) {
					empty = false;
					state = new BasicEntry(entry.key, entry.value);
					continue;
				}
				state = operator.apply(state, new BasicEntry(entry.key, entry.value));
			}
			return state;
		}
		
		@Override
		public Int2LongMap.Entry findFirst(Object2BooleanFunction<Int2LongMap.Entry> filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return null;
			BasicEntry subEntry = new BasicEntry();
			for(Int2LongRBTreeMap.Entry entry = first;entry != null;entry = entry.next()) {
				subEntry.set(entry.key, entry.value);
				if(filter.getBoolean(subEntry)) return subEntry;
			}
			return null;
		}
		
		@Override
		public int count(Object2BooleanFunction<Int2LongMap.Entry> filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return 0;
			int result = 0;
			BasicEntry subEntry = new BasicEntry();
			for(Int2LongRBTreeMap.Entry entry = first;entry != null;entry = entry.next()) {
				subEntry.set(entry.key, entry.value);
				if(filter.getBoolean(subEntry)) result++;
			}
			return result;
		}
	}
	
	static final class KeySet extends AbstractIntSet implements IntNavigableSet {
		Int2LongNavigableMap m;
		
		KeySet(Int2LongNavigableMap m) {
			this.m = m;
		}
		
		@Override
		public void setDefaultMaxValue(int e) { m.setDefaultMaxValue(e); }
		@Override
		public int getDefaultMaxValue() { return m.getDefaultMaxValue(); }
		@Override
		public void setDefaultMinValue(int e) { m.setDefaultMinValue(e); }
		@Override
		public int getDefaultMinValue() { return m.getDefaultMinValue(); }
		@Override
		public boolean add(int o) { throw new UnsupportedOperationException(); }
		@Override
		public boolean addAndMoveToFirst(int o) { throw new UnsupportedOperationException(); }
		@Override
		public boolean addAndMoveToLast(int o) { throw new UnsupportedOperationException(); }
		@Override
		public boolean moveToFirst(int o) { throw new UnsupportedOperationException(); }
		@Override
		public boolean moveToLast(int o) { throw new UnsupportedOperationException(); }
		@Override
		public IntComparator comparator() { return m.comparator(); }
		@Override
		public int lower(int e) { return m.lowerKey(e); }
		@Override
		public int floor(int e) { return m.floorKey(e); }
		@Override
		public int ceiling(int e) { return m.ceilingKey(e); }
		@Override
		public int higher(int e) { return m.higherKey(e); }
		@Override
		public int firstInt() { return m.firstIntKey(); }
		@Override
		public int pollFirstInt() { return m.pollFirstIntKey(); }
		@Override
		public int lastInt() { return m.lastIntKey(); }
		@Override
		public int pollLastInt() { return m.pollLastIntKey(); }		
		@Override
		public boolean remove(int o) { 
			int oldSize = m.size();
			m.remove(o); 
			return oldSize != m.size();
		}
		
		@Override
		public KeySet copy() { throw new UnsupportedOperationException(); }
		
		@Override
		public IntNavigableSet subSet(int fromElement, boolean fromInclusive, int toElement, boolean toInclusive) {
			return new KeySet(m.subMap(fromElement, fromInclusive, toElement, toInclusive));
		}
		
		@Override
		public IntNavigableSet headSet(int toElement, boolean inclusive) {
			return new KeySet(m.headMap(toElement, inclusive));
		}
		
		@Override
		public IntNavigableSet tailSet(int fromElement, boolean inclusive) {
			return new KeySet(m.tailMap(fromElement, inclusive));
		}
		
		@Override
		public IntNavigableSet descendingSet() {
			return new KeySet(m.descendingMap());
		}
		
		@Override
		public IntBidirectionalIterator iterator() {
			if(m instanceof Int2LongRBTreeMap) return ((Int2LongRBTreeMap)m).keyIterator(false);
			return ((NavigableSubMap)m).keyIterator(false);
		}
		
		@Override
		public IntBidirectionalIterator iterator(int fromElement) {
			if(m instanceof Int2LongRBTreeMap) return ((Int2LongRBTreeMap)m).keyIterator(fromElement);
			return ((NavigableSubMap)m).keyIterator(fromElement);
		}
		
		@Override
		public IntBidirectionalIterator descendingIterator() {
			if(m instanceof Int2LongRBTreeMap) return ((Int2LongRBTreeMap)m).keyIterator(true);
			return ((NavigableSubMap)m).keyIterator(true);
		}
		
		@Override
		public int size() {
			return m.size();
		}
		
		@Override
		public void clear() {
			m.clear();
		}
		
		protected Int2LongRBTreeMap.Entry start() {
			if(m instanceof Int2LongRBTreeMap) return ((Int2LongRBTreeMap)m).first;
			return ((NavigableSubMap)m).findLowest();
		}
		
		protected Int2LongRBTreeMap.Entry end() {
			if(m instanceof Int2LongRBTreeMap) return null;
			return ((NavigableSubMap)m).findLowest();
		}
		
		protected Int2LongRBTreeMap.Entry next(Int2LongRBTreeMap.Entry entry) {
			if(m instanceof Int2LongRBTreeMap) return entry.next();
			return ((NavigableSubMap)m).next(entry);
		}
		
		protected Int2LongRBTreeMap.Entry previous(Int2LongRBTreeMap.Entry entry) {
			if(m instanceof Int2LongRBTreeMap) return entry.previous();
			return ((NavigableSubMap)m).previous(entry);
		}
		
		@Override
		public void forEach(IntConsumer action) {
			Objects.requireNonNull(action);
			for(Int2LongRBTreeMap.Entry entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				action.accept(entry.key);
		}
		
		@Override
		public <E> void forEach(E input, ObjectIntConsumer<E> action) {
			Objects.requireNonNull(action);
			for(Int2LongRBTreeMap.Entry entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				action.accept(input, entry.key);
		}
		
		@Override
		public boolean matchesAny(Int2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(Int2LongRBTreeMap.Entry entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				if(filter.get(entry.key)) return true;
			return false;
		}
		
		@Override
		public boolean matchesNone(Int2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(Int2LongRBTreeMap.Entry entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				if(filter.get(entry.key)) return false;
			return true;
		}
		
		@Override
		public boolean matchesAll(Int2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(Int2LongRBTreeMap.Entry entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				if(!filter.get(entry.key)) return false;
			return true;
		}
		
		@Override
		public int reduce(int identity, IntIntUnaryOperator operator) {
			Objects.requireNonNull(operator);
			int state = identity;
			for(Int2LongRBTreeMap.Entry entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				state = operator.applyAsInt(state, entry.key);
			return state;
		}
		
		@Override
		public int reduce(IntIntUnaryOperator operator) {
			Objects.requireNonNull(operator);
			int state = 0;
			boolean empty = true;
			for(Int2LongRBTreeMap.Entry entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry)) {
				if(empty) {
					empty = false;
					state = entry.key;
					continue;
				}
				state = operator.applyAsInt(state, entry.key);
			}
			return state;
		}
		
		@Override
		public int findFirst(Int2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(Int2LongRBTreeMap.Entry entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				if(filter.get(entry.key)) return entry.key;
			return 0;
		}
		
		@Override
		public int count(Int2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			int result = 0;
			for(Int2LongRBTreeMap.Entry entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				if(filter.get(entry.key)) result++;
			return result;
		}
	}
	
	final class Values extends AbstractLongCollection {

		@Override
		public boolean add(long o) { throw new UnsupportedOperationException(); }
		
		@Override
		public boolean contains(long e) {
			return containsValue(e);
		}
		
		@Override
		public LongIterator iterator() { return new ValueIterator(false); }
		
		@Override
		public int size() {
			return Int2LongRBTreeMap.this.size();
		}
		
		@Override
		public void clear() {
			Int2LongRBTreeMap.this.clear();
		}
		
		@Override
		public void forEach(LongConsumer action) {
			Objects.requireNonNull(action);
			for(Int2LongRBTreeMap.Entry entry = first;entry != null;entry = entry.next())
				action.accept(entry.value);
		}
		
		@Override
		public <E> void forEach(E input, ObjectLongConsumer<E> action) {
			Objects.requireNonNull(action);
			for(Int2LongRBTreeMap.Entry entry = first;entry != null;entry = entry.next())
				action.accept(input, entry.value);
		}
		
		@Override
		public boolean matchesAny(Long2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(Int2LongRBTreeMap.Entry entry = first;entry != null;entry = entry.next())
				if(filter.get(entry.value)) return true;
			return false;
		}
		
		@Override
		public boolean matchesNone(Long2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(Int2LongRBTreeMap.Entry entry = first;entry != null;entry = entry.next())
				if(filter.get(entry.value)) return false;
			return true;
		}
		
		@Override
		public boolean matchesAll(Long2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(Int2LongRBTreeMap.Entry entry = first;entry != null;entry = entry.next())
				if(!filter.get(entry.value)) return false;
			return true;
		}
		
		@Override
		public long reduce(long identity, LongLongUnaryOperator operator) {
			Objects.requireNonNull(operator);
			long state = identity;
			for(Int2LongRBTreeMap.Entry entry = first;entry != null;entry = entry.next())
				state = operator.applyAsLong(state, entry.value);
			return state;
		}
		
		@Override
		public long reduce(LongLongUnaryOperator operator) {
			Objects.requireNonNull(operator);
			long state = 0L;
			boolean empty = true;
			for(Int2LongRBTreeMap.Entry entry = first;entry != null;entry = entry.next()) {
				if(empty) {
					empty = false;
					state = entry.value;
					continue;
				}
				state = operator.applyAsLong(state, entry.value);
			}
			return state;
		}
		
		@Override
		public long findFirst(Long2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(Int2LongRBTreeMap.Entry entry = first;entry != null;entry = entry.next())
				if(filter.get(entry.value)) return entry.value;
			return 0L;
		}
		
		@Override
		public int count(Long2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			int result = 0;
			for(Int2LongRBTreeMap.Entry entry = first;entry != null;entry = entry.next())
				if(filter.get(entry.value)) result++;
			return result;
		}
	}
	
	final class EntryIterator extends MapEntryIterator implements ObjectListIterator<Int2LongMap.Entry> {
		
		public EntryIterator(boolean descending) {
			super(descending);
		}
		
		public EntryIterator(Entry entry) {
			super(entry);
		}
		
		@Override
		public Int2LongMap.Entry next() {
			return nextEntry();
		}
		
		@Override
		public Int2LongMap.Entry previous() {
			return previousEntry();
		}
		
		@Override
		public void set(Int2LongMap.Entry e) { throw new UnsupportedOperationException(); }
		
		@Override
		public void add(Int2LongMap.Entry e) { throw new UnsupportedOperationException(); }
	}
	
	final class KeyIterator extends MapEntryIterator implements IntListIterator {
		
		public KeyIterator(boolean descending) {
			super(descending);
		}
		
		public KeyIterator(Entry entry) {
			super(entry);
		}
		
		@Override
		public int previousInt() { return previousEntry().key; }
		@Override
		public int nextInt() { return nextEntry().key; }
		
		@Override
		public void set(int e) { throw new UnsupportedOperationException(); }
		@Override
		public void add(int e) { throw new UnsupportedOperationException(); }
	}
	
	final class ValueIterator extends MapEntryIterator implements LongListIterator {
		
		public ValueIterator(boolean descending) {
			super(descending);
		}
		
		public ValueIterator(Entry entry) {
			super(entry);
		}
		
		@Override
		public long previousLong() {
			return previousEntry().value;
		}
		
		@Override
		public long nextLong() {
			return nextEntry().value;
		}
		
		@Override
		public void set(long e) { throw new UnsupportedOperationException(); }
		
		@Override
		public void add(long e) { throw new UnsupportedOperationException(); }
	}
	
	abstract class MapEntryIterator {
		Entry next;
		Entry previous;
		Entry current;
		int index = 0;
		
		public MapEntryIterator(boolean descending) {
			if(descending) previous = last;
			else next = first;
		}
		
		public MapEntryIterator(Entry entry) {
			next = entry;
			previous = entry.previous();
		}
		
		public boolean hasNext() {
			return next != null;
		}
		
		public boolean hasPrevious() {
			return previous != null;
		}
		
		public int nextIndex() {
			return index;
		}
		
		public int previousIndex() {
			return index - 1;
		}
		
		protected void updateNext() {
			next = current.next();
		}
		
		protected void updatePrevious() {
			previous = current.previous();
		}
		
		public Entry nextEntry() {
			if(!hasNext()) throw new NoSuchElementException();
			current = previous = next;
			updateNext();
			index++;
			return current;
		}
		
		public Entry previousEntry() {
			if(!hasPrevious()) throw new NoSuchElementException();
			current = next = previous;
			updatePrevious();
			index--;
			return current;
		}
		
		public void remove() {
			if(current == null) throw new IllegalStateException();
			if(current == previous) index--;
			updateNext();
			updatePrevious();
			if(current.needsSuccessor()) next = current;
			removeNode(current);
			current = null;
		}
	}
	
	private static final class Entry implements Int2LongMap.Entry
	{
		static final int BLACK = 1;
		
		int key;
		long value;
		int state;
		Entry parent;
		Entry left;
		Entry right;
		
		Entry(int key, long value, Entry parent) {
			this.key = key;
			this.value = value;
			this.parent = parent;
		}
		
		Entry copy() {
			Entry entry = new Entry(key, value, null);
			entry.state = state;
			if(left != null) {
				Entry newLeft = left.copy();
				entry.left = newLeft;
				newLeft.parent = entry;
			}
			if(right != null) {
				Entry newRight = right.copy();
				entry.right = newRight;
				newRight.parent = entry;
			}
			return entry;
		}
		
		@Override
		public int getIntKey() {
			return key;
		}
		
		@Override
		public long getLongValue() {
			return value;
		}
		
		@Override
		public long setValue(long value) {
			long oldValue = this.value;
			this.value = value;
			return oldValue;
		}
		
		long addTo(long value) {
			long oldValue = this.value;
			this.value += value;
			return oldValue;
		}
		
		@Override
		public boolean equals(Object obj) {
			if(obj instanceof Map.Entry) {
				if(obj instanceof Int2LongMap.Entry) {
					Int2LongMap.Entry entry = (Int2LongMap.Entry)obj;
					return key == entry.getIntKey() && value == entry.getLongValue();
				}
				Map.Entry<?, ?> entry = (Map.Entry<?, ?>)obj;
				Object otherKey = entry.getKey();
				if(otherKey == null) return false;
				Object otherValue = entry.getValue();
				return otherKey instanceof Integer && otherValue instanceof Long && key == ((Integer)otherKey).intValue() && value == ((Long)otherValue).longValue();
			}
			return false;
		}
		
		@Override
		public int hashCode() {
			return Integer.hashCode(key) ^ Long.hashCode(value);
		}
		
		@Override
		public String toString() {
			return Integer.toString(key) + "=" + Long.toString(value);
		}
		
		boolean isBlack() {
			return (state & BLACK) != 0;
		}
		
		void setBlack(boolean value) {
			if(value) state |= BLACK;
			else state &= ~BLACK;
		}
		
		boolean needsSuccessor() { return left != null && right != null; }
		
		boolean replace(Entry entry) {
			if(entry != null) entry.parent = parent;
			if(parent != null) {
				if(parent.left == this) parent.left = entry;
				else parent.right = entry;
			}
			return parent == null;
		}
		
		Entry next() {
			if(right != null) {
				Entry parent = right;
				while(parent.left != null) parent = parent.left;
				return parent;
			}
			Entry parent = this.parent;
			Entry control = this;
			while(parent != null && control == parent.right) {
				control = parent;
				parent = parent.parent;
			}
			return parent;
		}
		
		Entry previous() {
			if(left != null) {
				Entry parent = left;
				while(parent.right != null) parent = parent.right;
				return parent;
			}
			Entry parent = this.parent;
			Entry control = this;
			while(parent != null && control == parent.left) {
				control = parent;
				parent = parent.parent;
			}
			return parent;
		}
	}
}