package speiger.src.collections.ints.sets;

import java.util.Collection;
import java.util.Iterator;
import java.util.NoSuchElementException;
import java.util.Objects;
import speiger.src.collections.ints.collections.IntBidirectionalIterator;
import speiger.src.collections.ints.functions.IntComparator;
import speiger.src.collections.ints.functions.IntConsumer;
import speiger.src.collections.objects.functions.consumer.ObjectIntConsumer;
import speiger.src.collections.ints.functions.function.Int2BooleanFunction;
import speiger.src.collections.ints.functions.function.IntIntUnaryOperator;
import speiger.src.collections.ints.collections.IntCollection;
import speiger.src.collections.ints.collections.IntIterator;
import speiger.src.collections.ints.lists.IntListIterator;
import speiger.src.collections.ints.utils.IntIterators;
import speiger.src.collections.utils.SanityChecks;

/**
 * A Simple Type Specific AVL TreeSet implementation that reduces boxing/unboxing.
 * It is using a bit more memory then <a href="https://github.com/vigna/fastutil">FastUtil</a>,
 * but it saves a lot of Performance on the Optimized removal and iteration logic.
 * Which makes the implementation actually useable and does not get outperformed by Javas default implementation.
 */
public class IntAVLTreeSet extends AbstractIntSet implements IntNavigableSet
{
	/** The center of the Tree */
	protected transient Entry tree;
	/** The Lowest possible Node */
	protected transient Entry first;
	/** The Highest possible Node */
	protected transient Entry last;
	/** The amount of elements stored in the Set */
	protected int size = 0;
	/** The Sorter of the Tree */
	protected transient IntComparator comparator;
	/** the default return value for max searches */
	protected int defaultMaxNotFound = Integer.MIN_VALUE;
	/** the default return value for min searches */
	protected int defaultMinNotFound = Integer.MAX_VALUE;
	
	/**
	 * Default Constructor
	 */
	public IntAVLTreeSet() {
	}
	
	/**
	 * Constructor that allows to define the sorter
	 * @param comp the function that decides how the tree is sorted, can be null
	 */
	public IntAVLTreeSet(IntComparator comp) {
		comparator = comp;
	}
	
	/**
	 * Helper constructor that allow to create a set from an array
	 * @param array the elements that should be used
	 */
	public IntAVLTreeSet(int[] array) {
		this(array, 0, array.length);
	}
	
	/**
	 * Helper constructor that allow to create a set from an array
	 * @param array the elements that should be used
	 * @param offset the starting index within the array
	 * @param length the amount of elements that are within the array
	 * @throws IllegalStateException if offset and length causes to step outside of the arrays range
	 */
	public IntAVLTreeSet(int[] array, int offset, int length) {
		SanityChecks.checkArrayCapacity(array.length, offset, length);
		for(int i = 0;i<length;i++) add(array[offset+i]);
	}
	
	/**
	 * Helper constructor that allow to create a set from an array
	 * @param array the elements that should be used
	 * @param comp the sorter of the tree, can be null
	 */
	public IntAVLTreeSet(int[] array, IntComparator comp) {
		this(array, 0, array.length, comp);
	}
	
	/**
	 * Helper constructor that allow to create a set from an array
	 * @param array the elements that should be used
	 * @param offset the starting index within the array
	 * @param length the amount of elements that are within the array
	 * @param comp the sorter of the tree, can be null
	 * @throws IllegalStateException if offset and length causes to step outside of the arrays range
	 */
	public IntAVLTreeSet(int[] array, int offset, int length, IntComparator comp) {
		comparator = comp;
		SanityChecks.checkArrayCapacity(array.length, offset, length);
		for(int i = 0;i<length;i++) add(array[offset+i]);
	}
	
	/**
	 * A Helper constructor that allows to create a Set with exactly the same values as the provided SortedSet.
	 * @param sortedSet the set the elements should be added to the TreeSet
	 * @note this also includes the Comparator if present
	 */
	public IntAVLTreeSet(IntSortedSet sortedSet) {
		comparator = sortedSet.comparator();
		addAll(sortedSet);
	}
	
	/**
	 * A Helper constructor that allows to create a Set with exactly the same values as the provided collection.
	 * @param collection the set the elements should be added to the TreeSet
	 */
	@Deprecated
	public IntAVLTreeSet(Collection<? extends Integer> collection) {
		addAll(collection);
	}
	
	/**
	 * A Helper constructor that allows to create a Set with exactly the same values as the provided collection.
	 * @param collection the set the elements should be added to the TreeSet
	 * @param comp the sorter of the tree, can be null
	 */
	@Deprecated
	public IntAVLTreeSet(Collection<? extends Integer> collection, IntComparator comp) {
		comparator = comp;
		addAll(collection);
	}
	
	/**
	 * A Helper constructor that allows to create a Set with exactly the same values as the provided collection.
	 * @param collection the set the elements should be added to the TreeSet
	 */
	public IntAVLTreeSet(IntCollection collection) {
		addAll(collection);
	}
	
	/**
	 * A Helper constructor that allows to create a Set with exactly the same values as the provided collection.
	 * @param collection the set the elements should be added to the TreeSet
	 * @param comp the sorter of the tree, can be null
	 */
	public IntAVLTreeSet(IntCollection collection, IntComparator comp) {
		comparator = comp;
		addAll(collection);
	}
	
	/**
	 * A Helper constructor that allows to create a set from a iterator of an unknown size
	 * @param iterator the elements that should be added to the set
	 */
	public IntAVLTreeSet(Iterator<Integer> iterator) {
		this(IntIterators.wrap(iterator));
	}
	
	/**
	 * A Helper constructor that allows to create a set from a iterator of an unknown size
	 * @param iterator the elements that should be added to the set
	 * @param comp the sorter of the tree, can be null
	 */
	public IntAVLTreeSet(Iterator<Integer> iterator, IntComparator comp) {
		this(IntIterators.wrap(iterator), comp);
	}
	
	/**
	 * A Helper constructor that allows to create a set from a iterator of an unknown size
	 * @param iterator the elements that should be added to the set
	 */
	public IntAVLTreeSet(IntIterator iterator) {
		while(iterator.hasNext()) add(iterator.nextInt());
	}
	
	/**
	 * A Helper constructor that allows to create a set from a iterator of an unknown size
	 * @param iterator the elements that should be added to the set
	 * @param comp the sorter of the tree, can be null
	 */
	public IntAVLTreeSet(IntIterator iterator, IntComparator comp) {
		comparator = comp;
		while(iterator.hasNext()) add(iterator.nextInt());
	}
	
	@Override
	public void setDefaultMaxValue(int value) { defaultMaxNotFound = value; }
	@Override
	public int getDefaultMaxValue() { return defaultMaxNotFound; }
	@Override
	public void setDefaultMinValue(int value) { defaultMinNotFound = value; }
	@Override
	public int getDefaultMinValue() { return defaultMinNotFound; }
	
	@Override
	public boolean add(int o) {
		if(tree == null) {
			tree = first = last = new Entry(o, null);
			size++;
			return true;
		}
		int compare = 0;
		Entry parent = tree;
		while(true) {
			if((compare = compare(o, parent.key)) == 0) return false;
			if(compare < 0) {
				if(parent.left == null) break;
				parent = parent.left;
			}
			else if(compare > 0) {
				if(parent.right == null) break;
				parent = parent.right;
			}
		}
		Entry adding = new Entry(o, parent);
		if(compare < 0)  {
			parent.left = adding;
			if(parent == first)	first = adding;
		}
		else  {
			parent.right = adding;
			if(parent == last) last = adding;
		}
		fixAfterInsertion(adding);
		size++;
		return true;
	}
	
	@Override
	public boolean addAndMoveToFirst(int o) { throw new UnsupportedOperationException(); }
	
	@Override
	public boolean addAndMoveToLast(int o) { throw new UnsupportedOperationException(); }
	
	@Override
	public boolean moveToFirst(int o) { throw new UnsupportedOperationException(); }
	
	@Override
	public boolean moveToLast(int o) { throw new UnsupportedOperationException(); }

	@Override
	public int lower(int e) {
		Entry node = findLowerNode(e);
		return node != null ? node.key : defaultMinNotFound;
	}

	@Override
	public int floor(int e) {
		Entry node = findFloorNode(e);
		return node != null ? node.key : defaultMinNotFound;
	}
	
	@Override
	public int higher(int e) {
		Entry node = findHigherNode(e);
		return node != null ? node.key : defaultMaxNotFound;
	}

	@Override
	public int ceiling(int e) {
		Entry node = findCeilingNode(e);
		return node != null ? node.key : defaultMaxNotFound;
	}
	
	@Override
	public void forEach(IntConsumer action) {
		Objects.requireNonNull(action);
		for(Entry entry = first;entry != null;entry = entry.next()) {
			action.accept(entry.key);
		}
	}
	
	@Override
	public <E> void forEach(E input, ObjectIntConsumer<E> action) {
		Objects.requireNonNull(action);
		for(Entry entry = first;entry != null;entry = entry.next())
			action.accept(input, entry.key);		
	}
	
	@Override
	public boolean matchesAny(Int2BooleanFunction filter) {
		Objects.requireNonNull(filter);
		for(Entry entry = first;entry != null;entry = entry.next()) {
			if(filter.get(entry.key)) return true;
		}
		return false;
	}
	
	@Override
	public boolean matchesNone(Int2BooleanFunction filter) {
		Objects.requireNonNull(filter);
		for(Entry entry = first;entry != null;entry = entry.next()) {
			if(filter.get(entry.key)) return false;
		}
		return true;
	}
	
	@Override
	public boolean matchesAll(Int2BooleanFunction filter) {
		Objects.requireNonNull(filter);
		for(Entry entry = first;entry != null;entry = entry.next()) {
			if(!filter.get(entry.key)) return false;
		}
		return true;
	}
	
	@Override
	public int findFirst(Int2BooleanFunction filter) {
		Objects.requireNonNull(filter);
		for(Entry entry = first;entry != null;entry = entry.next()) {
			if(filter.get(entry.key)) return entry.key;
		}
		return 0;
	}
	
	@Override
	public int reduce(int identity, IntIntUnaryOperator operator) {
		Objects.requireNonNull(operator);
		int state = identity;
		for(Entry entry = first;entry != null;entry = entry.next()) {
			state = operator.applyAsInt(state, entry.key);
		}
		return state;
	}
	
	@Override
	public int reduce(IntIntUnaryOperator operator) {
		Objects.requireNonNull(operator);
		int state = 0;
		boolean empty = true;
		for(Entry entry = first;entry != null;entry = entry.next()) {
			if(empty) {
				empty = false;
				state = entry.key;
				continue;
			}
			state = operator.applyAsInt(state, entry.key);
		}
		return state;
	}
	
	@Override
	public int count(Int2BooleanFunction filter) {
		Objects.requireNonNull(filter);
		int result = 0;
		for(Entry entry = first;entry != null;entry = entry.next()) {
			if(filter.get(entry.key)) result++;
		}
		return result;
	}
	
	protected Entry findNode(int o) {
		Entry node = tree;
		int compare;
		while(node != null) {
			if((compare = compare(o, node.key)) == 0) return node;
			if(compare < 0) node = node.left;
			else node = node.right;
		}
		return null;
	}
	
	protected Entry findLowerNode(int e) {
		Entry entry = tree;
		while(entry != null) {
			if(compare(e, entry.key) > 0) {
				if(entry.right != null) entry = entry.right;
				else return entry;
			}
			else {
				if(entry.left != null) entry = entry.left;
				else {
					Entry parent = entry.parent;
					while(parent != null && parent.left == entry) {
						entry = parent;
						parent = parent.parent;
					}
					return parent;
				}
			}
		}
		return null;
	}
	
	protected Entry findFloorNode(int e) {
		Entry entry = tree;
		int compare;
		while(entry != null) {
			if((compare = compare(e, entry.key)) > 0) {
				if(entry.right == null) break;
				entry = entry.right;
				continue;
			}
			else if(compare < 0) {
				if(entry.left != null) entry = entry.left;
				else {
					Entry parent = entry.parent;
					while(parent != null && parent.left == entry) {
						entry = parent;
						parent = parent.parent;
					}
					return parent;
				}
				continue;
			}
			break;
		}
		return entry;
	}
	
	protected Entry findCeilingNode(int e) {
		Entry entry = tree;
		int compare;
		while(entry != null) {
			if((compare = compare(e, entry.key)) < 0) {
				if(entry.left == null) break;
				entry = entry.left;
				continue;
			}
			else if(compare > 0) {
				if(entry.right != null) entry = entry.right;
				else {
					Entry parent = entry.parent;
					while(parent != null && parent.right == entry) {
						entry = parent;
						parent = parent.parent;
					}
					return parent;
				}
				continue;
			}
			break;
		}
		return entry;
	}
	
	protected Entry findHigherNode(int e) {
		Entry entry = tree;
		while(entry != null) {
			if(compare(e, entry.key) < 0) {
				if(entry.left != null) entry = entry.left;
				else return entry;
			}
			else {
				if(entry.right != null) entry = entry.right;
				else {
					Entry parent = entry.parent;
					while(parent != null && parent.right == entry) {
						entry = parent;
						parent = parent.parent;
					}
					return parent;
				}
			}
		}
		return null;
	}
	
	@Override
	public boolean contains(int e) {
		return findNode(e) != null;
	}
	
	@Override
	public boolean contains(Object e) {
		return findNode(((Integer)e).intValue()) != null;
	}
	
	@Override
	public int firstInt() {
		if(tree == null) throw new NoSuchElementException();
		return first.key;
	}
	
	@Override
	public int lastInt() {
		if(tree == null) throw new NoSuchElementException();
		return last.key;
	}
	
	@Override
	public boolean remove(int o) {
		if(tree == null) return false;
		Entry entry = findNode(o);
		if(entry != null) {
			removeNode(entry);
			return true;
		}
		return false;
	}
	
	@Override
	public boolean remove(Object o) {
		if(tree == null) return false;
		Entry entry = findNode(((Integer)o).intValue());
		if(entry != null) {
			removeNode(entry);
			return true;
		}
		return false;
	}
	
	@Override
	public int pollFirstInt() {
		if(tree == null) throw new NoSuchElementException();
		int value = first.key;
		removeNode(first);
		return value;
	}
	
	@Override
	public int pollLastInt() {
		if(tree == null) throw new NoSuchElementException();
		int value = last.key;
		removeNode(last);
		return value;
	}
	
	@Override
	public int size() { return size; }
	
	@Override
	public void clear() {
		size = 0;
		first = null;
		last = null;
		tree = null;
	}
	
	public IntAVLTreeSet copy() {
		IntAVLTreeSet set = new IntAVLTreeSet();
		set.size = size;
		if(tree != null) {
			set.tree = tree.copy();
			Entry lastFound = null;
			for(Entry entry = tree;entry != null;entry = entry.left) lastFound = entry;
			set.first = lastFound;
			lastFound = null;
			for(Entry entry = tree;entry != null;entry = entry.right) lastFound = entry;
			set.last = lastFound;
		}
		return set;
	}
	
	@Override
	public IntComparator comparator() { return comparator; }
	
	@Override
	public IntBidirectionalIterator iterator() { return new SetIterator(false); }
	
	@Override
	public IntBidirectionalIterator iterator(int fromElement) {
		Entry entry = findNode(fromElement);
		return entry == null ? null : new SetIterator(entry);
	}
	
	@Override
	public IntBidirectionalIterator descendingIterator() { return new SetIterator(true); }
	
	@Override
	public IntNavigableSet subSet(int fromElement, boolean fromInclusive, int toElement, boolean toInclusive) {
		return new AscendingSubSet(this, false, fromElement, fromInclusive, false, toElement, toInclusive);
	}

	@Override
	public IntNavigableSet headSet(int toElement, boolean inclusive) {
		return new AscendingSubSet(this, true, 0, true, false, toElement, inclusive);
	}

	@Override
	public IntNavigableSet tailSet(int fromElement, boolean inclusive) {
		return new AscendingSubSet(this, false, fromElement, inclusive, true, 0, true);
	}
	
	@Override
	public IntNavigableSet descendingSet() {
		return new DescendingSubSet(this, true, 0, true, true, 0, true);
	}
	
	protected void removeNode(Entry entry) {
		size--;
		if(entry.needsSuccessor()) {
			Entry successor = entry.next();
			entry.key = successor.key;
			entry = successor;
		}
		if(entry.previous() == null) first = entry.next();
		if(entry.next() == null) last = entry.previous();
		Entry replacement = entry.left != null ? entry.left : entry.right;
		if(replacement != null) {
			if(entry.replace(replacement)) tree = replacement;
			entry.left = entry.right = entry.parent = null;
			fixAfterDeletion(replacement);
		}
		else if(entry.parent == null) tree = first = last = null;
		else {
			fixAfterDeletion(entry);
			entry.replace(null);
			entry.parent = null;
		}
	}
	
	protected void validate(int k) { compare(k, k); }
	protected int compare(int k, int v) { return comparator != null ? comparator.compare(k, v) : Integer.compare(k, v);}
	
	/** From CLR */
	protected void rotateLeft(Entry entry) {
		if(entry != null) {
			Entry right = entry.right;
			entry.right = right.left;
			if(right.left != null) right.left.parent = entry;
			right.parent = entry.parent;
			if(entry.parent == null) tree = right;
			else if(entry.parent.left == entry) entry.parent.left = right;
			else entry.parent.right = right;
			right.left = entry;
			entry.parent = right;
			entry.updateHeight();
			right.updateHeight();
		}
	}
	
	/** From CLR */
	protected void rotateRight(Entry entry) {
		if(entry != null) {
			Entry left = entry.left;
			entry.left = left.right;
			if(left.right != null) left.right.parent = entry;
			left.parent = entry.parent;
			if(entry.parent == null) tree = left;
			else if(entry.parent.right == entry) entry.parent.right = left;
			else entry.parent.left = left;
			left.right = entry;
			entry.parent = left;
			entry.updateHeight();
			left.updateHeight();
		}
	}
	
	/** From CLR */
	protected void fixAfterInsertion(Entry entry) {
		while(entry != null) {
			entry.updateHeight();
			int balance = entry.getBalance();
			if(balance > 1) {
				int compare = entry.left.getBalance();
				if(compare > 0) rotateRight(entry);
				else if(compare < 0) {
					rotateLeft(entry.left);
					rotateRight(entry);
				}
			}
			else if(balance < -1) {
				int compare = entry.right.getBalance();
				if(compare < 0) rotateLeft(entry);
				else if(compare > 0) {
					rotateRight(entry.right);
					rotateLeft(entry);
				}
			}
			entry = entry.parent;
		}
	}
	
	/** From CLR */
	protected void fixAfterDeletion(Entry entry) {
		if(entry != null) {
			entry.updateHeight();
			int balance = entry.getBalance();
			if(balance > 1) {
				int subBalance = entry.left.getBalance();
				if(subBalance >= 0) rotateRight(entry);
				else {
					rotateLeft(entry.left);
					rotateRight(entry);
				}
			}
			else if(balance < -1)
			{
				int subBalance = entry.right.getBalance();
				if(subBalance <= 0) rotateLeft(entry);
				else {
					rotateRight(entry.right);
					rotateLeft(entry);
				}
			}
			entry = entry.parent;
		}
	}
	
	private static class AscendingSubSet extends SubSet
	{
		AscendingSubSet(IntAVLTreeSet set, boolean fromStart, int start, boolean loInclusive, boolean toEnd, int end, boolean hiInclusive) {
			super(set, fromStart, start, loInclusive, toEnd, end, hiInclusive);
		}
		
		@Override
		public IntNavigableSet subSet(int fromElement, boolean fromInclusive, int toElement, boolean toInclusive) {
			if(!inRange(fromElement, fromInclusive)) throw new IllegalArgumentException("fromElement is not in Range");
			if(!inRange(toElement, toInclusive)) throw new IllegalArgumentException("toElement is not in Range");
			return new AscendingSubSet(set, false, fromElement, fromInclusive, false, toElement, toInclusive);
		}

		@Override
		public IntNavigableSet headSet(int toElement, boolean inclusive) {
			if(!inRange(toElement, inclusive)) throw new IllegalArgumentException("toElement is not in Range");
			return new AscendingSubSet(set, fromStart, start, loInclusive, false, toElement, inclusive);
		}

		@Override
		public IntNavigableSet tailSet(int fromElement, boolean inclusive) {
			if(!inRange(fromElement, inclusive)) throw new IllegalArgumentException("fromElement is not in Range");
			return new AscendingSubSet(set, false, fromElement, inclusive, toEnd, end, hiInclusive);
		}

		@Override
		public IntBidirectionalIterator iterator() { return new SubSetIterator(findLowest()); }
		
		@Override
		public IntBidirectionalIterator iterator(int fromElement) {
			Entry entry = set.findNode(fromElement);
			return entry == null || !inClosedRange(fromElement) ? null : new SubSetIterator(entry);
		}
		
		@Override
		public IntBidirectionalIterator descendingIterator() { return new SubSetIterator(findHighest()); }

		@Override
		public IntNavigableSet descendingSet() {
			return new DescendingSubSet(set, fromStart, start, loInclusive, toEnd, end, hiInclusive);
		}
	}
	
	private static class DescendingSubSet extends SubSet
	{
		IntComparator comparator;
		DescendingSubSet(IntAVLTreeSet set, boolean fromStart, int start, boolean loInclusive, boolean toEnd, int end, boolean hiInclusive) {
			super(set, fromStart, start, loInclusive, toEnd, end, hiInclusive);
			comparator = set.comparator();
			if(comparator != null) comparator = comparator.reversed();
		}
		
		@Override
		public IntComparator comparator() { return comparator; }
		
		@Override
		public IntNavigableSet subSet(int fromElement, boolean fromInclusive, int toElement, boolean toInclusive) {
			if(!inRange(fromElement, fromInclusive)) throw new IllegalArgumentException("fromElement is not in Range");
			if(!inRange(toElement, toInclusive)) throw new IllegalArgumentException("toElement is not in Range");
			return new DescendingSubSet(set, false, toElement, toInclusive, false, fromElement, fromInclusive);
		}

		@Override
		public IntNavigableSet headSet(int toElement, boolean inclusive) {
			if(!inRange(toElement, inclusive)) throw new IllegalArgumentException("toElement is not in Range");
			return new DescendingSubSet(set, false, toElement, inclusive, toEnd, end, hiInclusive);
		}

		@Override
		public IntNavigableSet tailSet(int fromElement, boolean inclusive) {
			if(!inRange(fromElement, inclusive)) throw new IllegalArgumentException("fromElement is not in Range");
			return new DescendingSubSet(set, fromStart, start, loInclusive, false, fromElement, inclusive);
		}
		
		protected Entry next(Entry entry) { return entry.previous(); }
		protected Entry start() { return findHighest(); }
		
		@Override
		public int firstInt() { return super.lastInt(); }
		
		@Override
		public int pollFirstInt() { return super.pollLastInt(); }
		
		@Override
		public int lastInt() { return super.firstInt(); }
		
		@Override
		public int pollLastInt() { return super.pollFirstInt(); }
		
		@Override
		public int lower(int e) { return super.higher(e); }

		@Override
		public int floor(int e) { return super.ceiling(e); }

		@Override
		public int ceiling(int e) { return super.floor(e); }

		@Override
		public int higher(int e) { return super.lower(e); }

		@Override
		public IntBidirectionalIterator iterator() { return new DescendingSubIterator(findHighest()); }
		
		@Override
		public IntBidirectionalIterator iterator(int fromElement) { 
			Entry entry = set.findNode(fromElement);
			return entry == null || !inClosedRange(fromElement) ? null : new DescendingSubIterator(entry);
		}
		
		@Override
		public IntBidirectionalIterator descendingIterator() { return new DescendingSubIterator(findLowest()); }
		
		@Override
		public IntNavigableSet descendingSet() {
			return new AscendingSubSet(set, fromStart, start, loInclusive, toEnd, end, hiInclusive);
		}
		
		private class DescendingSubIterator extends SubSetIterator
		{
			public DescendingSubIterator(Entry entry) {
				super(entry);
			}
			
			@Override
			protected void updateNext() {
				next = current.previous();
				if(!toEnd && next != null && bottomReached(next)) next = null;
			}
			
			@Override
			protected void updatePrevious() {
				previous = current.next();
				if(!fromStart && previous != null && topReached(previous)) previous = null;
			}
		}
	}
	
	private static abstract class SubSet extends AbstractIntSet implements IntNavigableSet
	{
		IntAVLTreeSet set;
		int start;
		int end;
		boolean fromStart;
		boolean toEnd;
		boolean loInclusive;
		boolean hiInclusive;
		
		SubSet(IntAVLTreeSet set, boolean fromStart, int start, boolean loInclusive, boolean toEnd, int end, boolean hiInclusive) {
			this.set = set;
			this.start = start;
			this.end = end;
			this.fromStart = fromStart;
			this.toEnd = toEnd;
			this.loInclusive = loInclusive;
			this.hiInclusive = hiInclusive;
		}
		
		@Override
		public void setDefaultMaxValue(int value) { throw new UnsupportedOperationException(); }
		
		@Override
		public int getDefaultMaxValue() { return set.getDefaultMaxValue(); }
		
		@Override
		public void setDefaultMinValue(int value) { throw new UnsupportedOperationException(); }
		
		@Override
		public int getDefaultMinValue() { return set.getDefaultMinValue(); }
		
		boolean tooLow(int key) { return !fromStart && (loInclusive ? set.compare(key, start) < 0 : set.compare(key, start) <= 0); }
		boolean tooHigh(int key) { return !toEnd && (hiInclusive ? set.compare(key, end) > 0 : set.compare(key, end) >= 0); }
		boolean inRange(int key) { return !tooLow(key) && !tooHigh(key); }
		boolean inClosedRange(int key) { return (fromStart || set.compare(key, start) >= 0) && (toEnd || set.compare(end, key) >= 0); }
		boolean inRange(int key, boolean inclusive) { return inclusive ? inRange(key) : inClosedRange(key); }
		protected Entry next(Entry entry) { return entry.next(); }
		protected Entry start() { return findLowest(); }
		
		@Override
		public boolean addAndMoveToFirst(int o) { throw new UnsupportedOperationException(); }
		
		@Override
		public boolean addAndMoveToLast(int o) { throw new UnsupportedOperationException(); }

		@Override
		public boolean moveToFirst(int o) { throw new UnsupportedOperationException(); }

		@Override
		public boolean moveToLast(int o) { throw new UnsupportedOperationException(); }

		@Override
		public IntComparator comparator() { return set.comparator(); }
		
		@Override
		public abstract IntBidirectionalIterator iterator();
		
		@Override
		public int firstInt() {
			Entry entry = findLowest();
			return entry == null ? getDefaultMaxValue() : entry.key;
		}
		
		protected Entry findLowest() {
			if(fromStart) return set.first;
			Entry entry = loInclusive ? set.findCeilingNode(start) : set.findHigherNode(start);
			return entry == null || tooHigh(entry.key) ? null : entry;
		}
		
		@Override
		public int pollFirstInt() {
			if(fromStart) return set.pollFirstInt();
			Entry entry = loInclusive ? set.findCeilingNode(start) : set.findHigherNode(start);
			if(entry != null && !tooHigh(entry.key)) {
				int value = entry.key;
				set.removeNode(entry);
				return value;
			}
			return getDefaultMaxValue();
		}

		@Override
		public int lastInt() {
			Entry entry = findHighest();
			return entry == null ? getDefaultMinValue() : entry.key;
		}
		
		protected Entry findHighest() {
			if(toEnd) return set.last;
			Entry entry = hiInclusive ? set.findFloorNode(end) : set.findLowerNode(end);
			return entry == null || tooLow(entry.key) ? null : entry;
		}

		@Override
		public int pollLastInt() {
			if(toEnd) return set.pollLastInt();
			Entry entry = hiInclusive ? set.findFloorNode(end) : set.findLowerNode(end);
			if(entry != null && !tooLow(entry.key)) {
				int value = entry.key;
				set.removeNode(entry);
				return value;
			}
			return getDefaultMinValue();
		}
		
		@Override
		public boolean add(int o) {
			if(!inRange(o)) throw new IllegalArgumentException("Key is out of range");
			return set.add(o);
		}
		
		@Override
		public boolean contains(int e) {
			return inRange(e) && set.contains(e);
		}
		
		@Override
		public boolean remove(int o) {
			return inRange(o) && set.remove(o);
		}
		
		@Override
		public boolean contains(Object e) {
			int o = ((Integer)e).intValue();
			return inRange(o) && set.contains(o);
		}

		@Override
		public boolean remove(Object e) {
			int o = ((Integer)e).intValue();
			return inRange(o) && set.remove(o);
		}
		
		@Override
		public int lower(int e) {
			if(tooHigh(e)) {
				Entry entry = findHighest();
				return entry == null ? getDefaultMinValue() : entry.key;
			}
			Entry entry = set.findLowerNode(e);
			return entry == null || tooHigh(entry.key) ? getDefaultMaxValue() : entry.key;
		}

		@Override
		public int floor(int e) {
			if(tooHigh(e)) {
				Entry entry = findHighest();
				return entry == null ? getDefaultMinValue() : entry.key;
			}
			Entry entry = set.findFloorNode(e);
			return entry == null || tooHigh(entry.key) ? getDefaultMaxValue() : entry.key;
		}

		@Override
		public int ceiling(int e) {
			if(tooLow(e)) {
				Entry entry = findLowest();
				return entry == null ? getDefaultMaxValue() : entry.key;
			}
			Entry entry = set.findCeilingNode(e);
			return entry == null || tooLow(entry.key) ? getDefaultMinValue() : entry.key;
		}

		@Override
		public int higher(int e) {
			if(tooLow(e)) {
				Entry entry = findLowest();
				return entry == null ? getDefaultMaxValue() : entry.key;
			}
			Entry entry = set.findHigherNode(e);
			return entry == null || tooLow(entry.key) ? getDefaultMinValue() : entry.key;
		}
		
		@Override
		public int size() {
			return fromStart && toEnd ? set.size() : iterator().skip(Integer.MAX_VALUE);
		}
		
		@Override
		public SubSet copy() { throw new UnsupportedOperationException(); }
		
		@Override
		public void forEach(IntConsumer action) {
			Objects.requireNonNull(action);
			for(Entry entry = start();entry != null && inRange(entry.key);entry = next(entry)) {
				action.accept(entry.key);
			}
		}
		
		@Override
		public <E> void forEach(E input, ObjectIntConsumer<E> action) {
			Objects.requireNonNull(action);
			for(Entry entry = start();entry != null && inRange(entry.key);entry = next(entry))
				action.accept(input, entry.key);		
		}
		
		@Override
		public boolean matchesAny(Int2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(Entry entry = start();entry != null && inRange(entry.key);entry = next(entry)) {
				if(filter.get(entry.key)) return true;
			}
			return false;
		}
		
		@Override
		public boolean matchesNone(Int2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(Entry entry = start();entry != null && inRange(entry.key);entry = next(entry)) {
				if(filter.get(entry.key)) return false;
			}
			return true;
		}
		
		@Override
		public boolean matchesAll(Int2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(Entry entry = start();entry != null && inRange(entry.key);entry = next(entry)) {
				if(!filter.get(entry.key)) return false;
			}
			return true;
		}
		
		@Override
		public int reduce(int identity, IntIntUnaryOperator operator) {
			Objects.requireNonNull(operator);
			int state = identity;
			for(Entry entry = start();entry != null && inRange(entry.key);entry = next(entry)) {
				state = operator.applyAsInt(state, entry.key);
			}
			return state;
		}
		
		@Override
		public int reduce(IntIntUnaryOperator operator) {
			Objects.requireNonNull(operator);
			int state = 0;
			boolean empty = true;
			for(Entry entry = start();entry != null && inRange(entry.key);entry = next(entry)) {
				if(empty) {
					empty = false;
					state = entry.key;
					continue;
				}
				state = operator.applyAsInt(state, entry.key);
			}
			return state;
		}
		
		@Override
		public int findFirst(Int2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(Entry entry = start();entry != null && inRange(entry.key);entry = next(entry)) {
				if(filter.get(entry.key)) return entry.key;
			}
			return 0;
		}
		
		@Override
		public int count(Int2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			int result = 0;
			for(Entry entry = start();entry != null && inRange(entry.key);entry = next(entry)) {
				if(filter.get(entry.key)) result++;
			}
			return result;
		}
		
		class SubSetIterator implements IntListIterator
		{
			Entry previous;
			Entry next;
			Entry current;
			int index = 0;
			
			public SubSetIterator(Entry entry) {
				next = entry;
				previous = entry == null ? null : entry.previous();
			}
			
			@Override
			public boolean hasNext() {
				return next != null;
			}
			
			@Override
			public boolean hasPrevious() {
				return previous != null;
			}
			
			@Override
			public int nextIndex() {
				return index;
			}
			
			@Override
			public int previousIndex() {
				return index - 1;
			}
			
			boolean topReached(Entry entry) {
				return hiInclusive ? set.compare(entry.key, end) >= 0 : set.compare(entry.key, end) > 0;
			}
			
			boolean bottomReached(Entry entry) {
				return loInclusive ? set.compare(entry.key, start) <= 0 : set.compare(entry.key, start) < 0;
			}
			
			protected void updateNext() {
				next = current.next();
				if(!toEnd && next != null && topReached(next)) next = null;
			}
			
			protected void updatePrevious() {
				previous = current.previous();
				if(!fromStart && previous != null && bottomReached(previous)) previous = null;
			}
			
			@Override
			public int nextInt() {
				if(!hasNext()) throw new NoSuchElementException();
				current = previous = next;
				updateNext();
				index++;
				return current.key;
			}
			
			@Override
			public int previousInt() {
				if(!hasPrevious()) throw new NoSuchElementException();
				current = next = previous;
				updatePrevious();
				index--;
				return current.key;
			}
			
			@Override
			public void remove() {
				if(current == null) throw new IllegalStateException();
				if(current == previous) index--;
				updateNext();
				updatePrevious();
				set.removeNode(current);
				current = null;
			}
			
			@Override
			public void set(int e){ throw new UnsupportedOperationException(); }
			
			@Override
			public void add(int e) { throw new UnsupportedOperationException(); }
		}
	}
	
	private class SetIterator implements IntListIterator
	{
		Entry previous;
		Entry next;
		Entry current;
		int index = 0;
		 
		public SetIterator(boolean descending) {
			if(descending) previous = last;
			else next = first;
		}
		
		public SetIterator(Entry entry) {
			next = entry;
			previous = entry.previous();
		}
		
		@Override
		public boolean hasNext() {
			return next != null;
		}
		
		@Override
		public boolean hasPrevious() {
			return previous != null;
		}
		
		@Override
		public int nextIndex() {
			return index;
		}
		
		@Override
		public int previousIndex() {
			return index - 1;
		}
		
		protected void updateNext() {
			next = current.next();
		}
		
		protected void updatePrevious() {
			previous = current.previous();
		}
		
		@Override
		public int nextInt() {
			if(!hasNext()) throw new NoSuchElementException();
			current = previous = next;
			updateNext();
			index++;
			return current.key;
		}
		
		@Override
		public int previousInt() {
			if(!hasPrevious()) throw new NoSuchElementException();
			current = next = previous;
			updatePrevious();
			index--;
			return current.key;
		}
		
		@Override
		public void remove() {
			if(current == null) throw new IllegalStateException();
			if(current == previous) index--;
			updateNext();
			updatePrevious();
			if(current.needsSuccessor()) next = current;
			removeNode(current);
			current = null;
		}
				
		@Override
		public void set(int e){ throw new UnsupportedOperationException(); }
		
		@Override
		public void add(int e) { throw new UnsupportedOperationException(); }
	}
	
	private static final class Entry
	{
		int key;
		int state;
		Entry parent;
		Entry left;
		Entry right;
		
		Entry(int key, Entry parent) {
			this.key = key;
			this.parent = parent;
		}
		
		Entry copy() {
			Entry entry = new Entry(key, null);
			entry.state = state;
			if(left != null) {
				Entry newLeft = left.copy();
				entry.left = newLeft;
				newLeft.parent = entry;
			}
			if(right != null) {
				Entry newRight = right.copy();
				entry.right = newRight;
				newRight.parent = entry;
			}
			return entry;
		}
		
		int getHeight() { return state; }
		
		void updateHeight() { state = (1 + Math.max(left == null ? -1 : left.getHeight(), right == null ? -1 : right.getHeight())); }
		
		int getBalance() { return (left == null ? -1 : left.getHeight()) - (right == null ? -1 : right.getHeight()); }
		
		boolean needsSuccessor() { return left != null && right != null; }
		
		boolean replace(Entry entry) {
			if(entry != null) entry.parent = parent;
			if(parent != null) {
				if(parent.left == this) parent.left = entry;
				else parent.right = entry;
			}
			return parent == null;
		}
		
		Entry next() {
			if(right != null) {
				Entry parent = right;
				while(parent.left != null) parent = parent.left;
				return parent;
			}
			Entry parent = this.parent;
			Entry control = this;
			while(parent != null && control == parent.right) {
				control = parent;
				parent = parent.parent;
			}
			return parent;
		}
		
		Entry previous() {
			if(left != null) {
				Entry parent = left;
				while(parent.right != null) parent = parent.right;
				return parent;
			}
			Entry parent = this.parent;
			Entry control = this;
			while(parent != null && control == parent.left) {
				control = parent;
				parent = parent.parent;
			}
			return parent;
		}
	}
}