package speiger.src.collections.objects.maps.impl.tree;

import java.util.Map;
import java.util.Comparator;
import java.util.function.Consumer;
import java.util.function.BiFunction;
import java.util.Objects;
import java.util.NoSuchElementException;

import speiger.src.collections.objects.collections.ObjectBidirectionalIterator;
import speiger.src.collections.objects.functions.consumer.ObjectLongConsumer;
import speiger.src.collections.objects.functions.function.Object2LongFunction;
import speiger.src.collections.objects.functions.function.ObjectLongUnaryOperator;
import speiger.src.collections.objects.functions.function.ObjectObjectUnaryOperator;
import speiger.src.collections.objects.lists.ObjectListIterator;
import speiger.src.collections.objects.maps.abstracts.AbstractObject2LongMap;
import speiger.src.collections.objects.maps.interfaces.Object2LongMap;
import speiger.src.collections.objects.maps.interfaces.Object2LongNavigableMap;
import speiger.src.collections.objects.sets.AbstractObjectSet;
import speiger.src.collections.objects.sets.ObjectNavigableSet;
import speiger.src.collections.objects.sets.ObjectSet;
import speiger.src.collections.objects.sets.ObjectSortedSet;
import speiger.src.collections.objects.utils.ObjectIterators;
import speiger.src.collections.objects.utils.maps.Object2LongMaps;
import speiger.src.collections.longs.collections.AbstractLongCollection;
import speiger.src.collections.longs.collections.LongCollection;
import speiger.src.collections.longs.collections.LongIterator;
import speiger.src.collections.longs.functions.LongSupplier;
import speiger.src.collections.longs.functions.function.LongLongUnaryOperator;
import speiger.src.collections.longs.lists.LongListIterator;
import speiger.src.collections.longs.utils.LongIterators;
import speiger.src.collections.longs.functions.LongConsumer;
import speiger.src.collections.objects.functions.consumer.ObjectObjectConsumer;
import speiger.src.collections.objects.functions.function.Object2BooleanFunction;
import speiger.src.collections.longs.functions.function.Long2BooleanFunction;
import speiger.src.collections.objects.collections.ObjectIterator;

/**
 * A Simple Type Specific AVL TreeMap implementation that reduces boxing/unboxing.
 * It is using a bit more memory then <a href="https://github.com/vigna/fastutil">FastUtil</a>,
 * but it saves a lot of Performance on the Optimized removal and iteration logic.
 * Which makes the implementation actually useable and does not get outperformed by Javas default implementation.
 * @param <T> the type of elements maintained by this Collection
 */
public class Object2LongAVLTreeMap<T> extends AbstractObject2LongMap<T> implements Object2LongNavigableMap<T>
{
	/** The center of the Tree */
	protected transient Entry<T> tree;
	/** The Lowest possible Node */
	protected transient Entry<T> first;
	/** The Highest possible Node */
	protected transient Entry<T> last;
	/** The amount of elements stored in the Map */
	protected int size = 0;
	/** The Sorter of the Tree */
	protected transient Comparator<T> comparator;
	
	
	/** KeySet Cache */
	protected ObjectNavigableSet<T> keySet;
	/** Values Cache */
	protected LongCollection values;
	/** EntrySet Cache */
	protected ObjectSet<Object2LongMap.Entry<T>> entrySet;
	
	/**
	 * Default Constructor
	 */
	public Object2LongAVLTreeMap() {
	}
	
	/**
	 * Constructor that allows to define the sorter
	 * @param comp the function that decides how the tree is sorted, can be null
	 */
	public Object2LongAVLTreeMap(Comparator<T> comp) {
		comparator = comp;
	}
	
	/**
	 * Helper constructor that allow to create a map from boxed values (it will unbox them)
	 * @param keys the keys that should be put into the map
	 * @param values the values that should be put into the map.
	 * @throws IllegalStateException if the keys and values do not match in lenght
	 */
	public Object2LongAVLTreeMap(T[] keys, Long[] values) {
		this(keys, values, null);
	}
	
	/**
	 * Helper constructor that has a custom sorter and allow to create a map from boxed values (it will unbox them)
	 * @param keys the keys that should be put into the map
	 * @param values the values that should be put into the map.
	 * @param comp the function that decides how the tree is sorted, can be null
	 * @throws IllegalStateException if the keys and values do not match in lenght
	 */
	public Object2LongAVLTreeMap(T[] keys, Long[] values, Comparator<T> comp) {
		comparator = comp;
		if(keys.length != values.length) throw new IllegalStateException("Input Arrays are not equal size");
		for(int i = 0,m=keys.length;i<m;i++) put(keys[i], values[i].longValue());
	}
	
	/**
	 * Helper constructor that allow to create a map from unboxed values
	 * @param keys the keys that should be put into the map
	 * @param values the values that should be put into the map.
	 * @throws IllegalStateException if the keys and values do not match in lenght
	 */
	public Object2LongAVLTreeMap(T[] keys, long[] values) {
		this(keys, values, null);
	}
	
	/**
	 * Helper constructor that has a custom sorter and allow to create a map from unboxed values
	 * @param keys the keys that should be put into the map
	 * @param values the values that should be put into the map.
	 * @param comp the function that decides how the tree is sorted, can be null
	 * @throws IllegalStateException if the keys and values do not match in lenght
	 */
	public Object2LongAVLTreeMap(T[] keys, long[] values, Comparator<T> comp) {
		comparator = comp;
		if(keys.length != values.length) throw new IllegalStateException("Input Arrays are not equal size");
		for(int i = 0,m=keys.length;i<m;i++) put(keys[i], values[i]);
	}
	
	/**
	 * A Helper constructor that allows to create a Map with exactly the same values as the provided map.
	 * @param map the values that should be present in the map
	 */
	public Object2LongAVLTreeMap(Map<? extends T, ? extends Long> map) {
		this(map, null);
	}
	
	/**
	 * A Helper constructor that has a custom sorter and allows to create a Map with exactly the same values as the provided map.
	 * @param map the values that should be present in the map
	 * @param comp the function that decides how the tree is sorted, can be null
	 */
	public Object2LongAVLTreeMap(Map<? extends T, ? extends Long> map, Comparator<T> comp) {
		comparator = comp;
		putAll(map);
	}
	
	/**
	 * A Type Specific Helper function that allows to create a new Map with exactly the same values as the provided map.
	 * @param map the values that should be present in the map
 	 */
	public Object2LongAVLTreeMap(Object2LongMap<T> map) {
		this(map, null);
	}
	
	/**
	 * A Type Specific Helper function that has a custom sorter and allows to create a new Map with exactly the same values as the provided map.
	 * @param map the values that should be present in the map
	 * @param comp the function that decides how the tree is sorted, can be null
 	 */
	public Object2LongAVLTreeMap(Object2LongMap<T> map, Comparator<T> comp) {
		comparator = comp;
		putAll(map);
	}
	
	/** only used for primitives 
	 * @return null 
	 */
	public T getDefaultMaxValue() { return null; }
	/** only used for primitives 
	 * @return null 
	 */
	public T getDefaultMinValue() { return null; }
	
	@Override
	public long put(T key, long value) {
		validate(key);
		if(tree == null) {
			tree = first = last = new Entry<>(key, value, null);
			size++;
			return getDefaultReturnValue();
		}
		int compare = 0;
		Entry<T> parent = tree;
		while(true) {
			if((compare = compare(key, parent.key)) == 0) return parent.setValue(value);
			if(compare < 0) {
				if(parent.left == null) break;
				parent = parent.left;
			}
			else if(compare > 0) {
				if(parent.right == null) break;
				parent = parent.right;
			}
		}
		Entry<T> adding = new Entry<>(key, value, parent);
		if(compare < 0)  {
			parent.left = adding;
			if(parent == first)	first = adding;
		}
		else {
			parent.right = adding;
			if(parent == last) last = adding;
		}
		fixAfterInsertion(adding);
		size++;
		return getDefaultReturnValue();
	}
	
	@Override
	public long putIfAbsent(T key, long value) {
		validate(key);
		if(tree == null) {
			tree = first = last = new Entry<>(key, value, null);
			size++;
			return getDefaultReturnValue();
		}
		int compare = 0;
		Entry<T> parent = tree;
		while(true) {
			if((compare = compare(key, parent.key)) == 0) return parent.value;
			if(compare < 0) {
				if(parent.left == null) break;
				parent = parent.left;
			}
			else if(compare > 0) {
				if(parent.right == null) break;
				parent = parent.right;
			}
		}
		Entry<T> adding = new Entry<>(key, value, parent);
		if(compare < 0)  {
			parent.left = adding;
			if(parent == first)	first = adding;
		}
		else  {
			parent.right = adding;
			if(parent == last) last = adding;
		}
		fixAfterInsertion(adding);
		size++;
		return getDefaultReturnValue();
	}
	
	@Override
	public long addTo(T key, long value) {
		validate(key);
		if(tree == null) {
			tree = first = last = new Entry<>(key, value, null);
			size++;
			return getDefaultReturnValue();
		}
		int compare = 0;
		Entry<T> parent = tree;
		while(true) {
			if((compare = compare(key, parent.key)) == 0) return parent.addTo(value);
			if(compare < 0) {
				if(parent.left == null) break;
				parent = parent.left;
			}
			else if(compare > 0) {
				if(parent.right == null) break;
				parent = parent.right;
			}
		}
		Entry<T> adding = new Entry<>(key, value, parent);
		if(compare < 0)  {
			parent.left = adding;
			if(parent == first)	first = adding;
		}
		else  {
			parent.right = adding;
			if(parent == last) last = adding;
		}
		fixAfterInsertion(adding);
		size++;
		return getDefaultReturnValue();
	}
	
	@Override
	public long putAndMoveToFirst(T key, long value) { throw new UnsupportedOperationException(); }
	@Override
	public long putAndMoveToLast(T key, long value) { throw new UnsupportedOperationException(); }
	@Override
	public boolean moveToFirst(T key) { throw new UnsupportedOperationException(); }
	@Override
	public boolean moveToLast(T key) { throw new UnsupportedOperationException(); }
	@Override
	public long getAndMoveToFirst(T key) { throw new UnsupportedOperationException(); }
	@Override
	public long getAndMoveToLast(T key) { throw new UnsupportedOperationException(); }
	
	@Override
	public Comparator<T> comparator() { return comparator; }
	
	@Override
	public boolean containsKey(Object key) {
		return findNode((T)key) != null;
	}
	

	@Override
	public long getLong(T key) {
		Entry<T> node = findNode(key);
		return node == null ? getDefaultReturnValue() : node.value;
	}
	
	@Override
	public long getOrDefault(T key, long defaultValue) {
		Entry<T> node = findNode(key);
		return node == null ? defaultValue : node.value;
	}
	
	@Override
	public T firstKey() {
		if(tree == null) throw new NoSuchElementException();
		return first.key;
	}
	
	@Override
	public T pollFirstKey() {
		if(tree == null) throw new NoSuchElementException();
		T result = first.key;
		removeNode(first);
		return result;
	}
	
	@Override
	public T lastKey() {
		if(tree == null) throw new NoSuchElementException();
		return last.key;
	}
	
	@Override
	public T pollLastKey() {
		if(tree == null) throw new NoSuchElementException();
		T result = last.key;
		removeNode(last);
		return result;
	}
	
	@Override
	public Object2LongMap.Entry<T> firstEntry() {
		if(tree == null) throw new NoSuchElementException();
		return new BasicEntry<>(first.key, first.value);
	}
	
	@Override
	public Object2LongMap.Entry<T> lastEntry() {
		if(tree == null) throw new NoSuchElementException();
		return new BasicEntry<>(last.key, last.value);
	}
	
	@Override
	public Object2LongMap.Entry<T> pollFirstEntry() {
		if(tree == null) throw new NoSuchElementException();
		BasicEntry<T> entry = new BasicEntry<>(first.key, first.value);
		removeNode(first);
		return entry;
	}
	
	@Override
	public Object2LongMap.Entry<T> pollLastEntry() {
		if(tree == null) throw new NoSuchElementException();
		BasicEntry<T> entry = new BasicEntry<>(last.key, last.value);
		removeNode(last);
		return entry;
	}
	
	@Override
	public long firstLongValue() {
		if(tree == null) throw new NoSuchElementException();
		return first.value;
	}
	
	@Override
	public long lastLongValue() {
		if(tree == null) throw new NoSuchElementException();
		return last.value;
	}
	
	@Override
	public long rem(T key) {
		Entry<T> entry = findNode(key);
		if(entry == null) return getDefaultReturnValue();
		long value = entry.value;
		removeNode(entry);
		return value;
	}
	
	@Override
	public long remOrDefault(T key, long defaultValue) {
		Entry<T> entry = findNode(key);
		if(entry == null) return defaultValue;
		long value = entry.value;
		removeNode(entry);
		return value;
	}
	
	@Override
	public boolean remove(T key, long value) {
		Entry<T> entry = findNode(key);
		if(entry == null || entry.value != value) return false;
		removeNode(entry);
		return true;
	}
	
	@Override
	public boolean replace(T key, long oldValue, long newValue) {
		Entry<T> entry = findNode(key);
		if(entry == null || entry.value != oldValue) return false;
		entry.value = newValue;
		return true;
	}
	
	@Override
	public long replace(T key, long value) {
		Entry<T> entry = findNode(key);
		if(entry == null) return getDefaultReturnValue();
		long oldValue = entry.value;
		entry.value = value;
		return oldValue;
	}
	
	@Override
	public long computeLong(T key, ObjectLongUnaryOperator<T> mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		validate(key);
		Entry<T> entry = findNode(key);
		if(entry == null) {
			long newValue = mappingFunction.applyAsLong(key, getDefaultReturnValue());
			if(newValue == getDefaultReturnValue()) return newValue;
			put(key, newValue);
			return newValue;
		}
		long newValue = mappingFunction.applyAsLong(key, entry.value);
		if(newValue == getDefaultReturnValue()) {
			removeNode(entry);
			return newValue;
		}
		entry.value = newValue;
		return newValue;
	}
	
	@Override
	public long computeLongIfAbsent(T key, Object2LongFunction<T> mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		validate(key);
		Entry<T> entry = findNode(key);
		if(entry == null) {
			long newValue = mappingFunction.getLong(key);
			if(newValue == getDefaultReturnValue()) return newValue;
			put(key, newValue);
			return newValue;
		}
		if(entry.value == getDefaultReturnValue()) {
			long newValue = mappingFunction.getLong(key);
			if(newValue == getDefaultReturnValue()) return newValue;
			entry.value = newValue;
		}
		return entry.value;
	}
	
	@Override
	public long supplyLongIfAbsent(T key, LongSupplier valueProvider) {
		Objects.requireNonNull(valueProvider);
		validate(key);
		Entry<T> entry = findNode(key);
		if(entry == null) {
			long newValue = valueProvider.getLong();
			if(newValue == getDefaultReturnValue()) return newValue;
			put(key, newValue);
			return newValue;
		}
		if(entry.value == getDefaultReturnValue()) {
			long newValue = valueProvider.getLong();
			if(newValue == getDefaultReturnValue()) return newValue;
			entry.value = newValue;
		}
		return entry.value;
	}
	
	@Override
	public long computeLongIfPresent(T key, ObjectLongUnaryOperator<T> mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		validate(key);
		Entry<T> entry = findNode(key);
		if(entry == null || entry.value == getDefaultReturnValue()) return getDefaultReturnValue();
		long newValue = mappingFunction.applyAsLong(key, entry.value);
		if(newValue == getDefaultReturnValue()) {
			removeNode(entry);
			return newValue;
		}
		entry.value = newValue;
		return newValue;
	}
	
	@Override
	public long mergeLong(T key, long value, LongLongUnaryOperator mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		validate(key);
		Entry<T> entry = findNode(key);
		long newValue = entry == null || entry.value == getDefaultReturnValue() ? value : mappingFunction.applyAsLong(entry.value, value);
		if(newValue == getDefaultReturnValue()) {
			if(entry != null)
				removeNode(entry);
		}
		else if(entry == null) put(key, newValue);
		else entry.value = newValue;
		return newValue;
	}
	
	@Override
	public void mergeAllLong(Object2LongMap<T> m, LongLongUnaryOperator mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		for(Object2LongMap.Entry<T> entry : Object2LongMaps.fastIterable(m)) {
			T key = entry.getKey();
			Entry<T> subEntry = findNode(key);
			long newValue = subEntry == null || subEntry.value == getDefaultReturnValue() ? entry.getLongValue() : mappingFunction.applyAsLong(subEntry.value, entry.getLongValue());
			if(newValue == getDefaultReturnValue()) {
				if(subEntry != null)
					removeNode(subEntry);
			}
			else if(subEntry == null) put(key, newValue);
			else subEntry.value = newValue;
		}
	}
	
	@Override
	public void forEach(ObjectLongConsumer<T> action) {
		for(Entry<T> entry = first;entry != null;entry = entry.next())
			action.accept(entry.key, entry.value);
	}
	
	@Override
	public int size() { return size; }
	
	@Override
	public void clear() {
		size = 0;
		first = null;
		last = null;
		tree = null;
	}
	
	ObjectListIterator<T> keyIterator(boolean descending) {
		ObjectListIterator<T> iter = new KeyIterator(descending);
		return descending ? ObjectIterators.invert(iter) : iter;
	}
	
	ObjectListIterator<T> keyIterator(T key) {
		return new KeyIterator(findNode(key));
	}
	
	public Object2LongAVLTreeMap<T> copy() {
		Object2LongAVLTreeMap<T> set = new Object2LongAVLTreeMap<>();
		set.size = size;
		if(tree != null) {
			set.tree = tree.copy();
			Entry<T> lastFound = null;
			for(Entry<T> entry = tree;entry != null;entry = entry.left) lastFound = entry;
			set.first = lastFound;
			lastFound = null;
			for(Entry<T> entry = tree;entry != null;entry = entry.right) lastFound = entry;
			set.last = lastFound;
		}
		return set;
	}
	
	@Override
	public ObjectSortedSet<T> keySet() {
		return navigableKeySet();
	}
	
	@Override
	public ObjectSet<Object2LongMap.Entry<T>> object2LongEntrySet() {
		if(entrySet == null) entrySet = new EntrySet();
		return entrySet;
	}
	
	@Override
	public LongCollection values() {
		if(values == null) values = new Values();
		return values;
	}
	
	@Override
	public ObjectNavigableSet<T> navigableKeySet() {
		if(keySet == null) keySet = new KeySet<>(this);
		return keySet;
	}
	
	@Override
	public Object2LongNavigableMap<T> descendingMap() {
		return new DescendingSubMap<>(this, true, null, true, true, null, true);
	}
	
	@Override
	public ObjectNavigableSet<T> descendingKeySet() {
		return descendingMap().navigableKeySet();
	}
	
	@Override
	public Object2LongNavigableMap<T> subMap(T fromKey, boolean fromInclusive, T toKey, boolean toInclusive) {
		return new AscendingSubMap<>(this, false, fromKey, fromInclusive, false, toKey, toInclusive);
	}
	
	@Override
	public Object2LongNavigableMap<T> headMap(T toKey, boolean inclusive) {
		return new AscendingSubMap<>(this, true, null, true, false, toKey, inclusive);
	}
	
	@Override
	public Object2LongNavigableMap<T> tailMap(T fromKey, boolean inclusive) {
		return new AscendingSubMap<>(this, false, fromKey, inclusive, true, null, true);
	}
	
	@Override
	public T lowerKey(T e) {
		Entry<T> node = findLowerNode(e);
		return node != null ? node.key : getDefaultMinValue();
	}

	@Override
	public T floorKey(T e) {
		Entry<T> node = findFloorNode(e);
		return node != null ? node.key : getDefaultMinValue();
	}
	
	@Override
	public T higherKey(T e) {
		Entry<T> node = findHigherNode(e);
		return node != null ? node.key : getDefaultMaxValue();
	}

	@Override
	public T ceilingKey(T e) {
		Entry<T> node = findCeilingNode(e);
		return node != null ? node.key : getDefaultMaxValue();
	}
	
	@Override
	public Object2LongMap.Entry<T> lowerEntry(T key) {
		Entry<T> node = findLowerNode(key);
		return node != null ? new BasicEntry<>(node.key, node.value) : null;
	}
	
	@Override
	public Object2LongMap.Entry<T> higherEntry(T key) {
		Entry<T> node = findHigherNode(key);
		return node != null ? new BasicEntry<>(node.key, node.value) : null;
	}
	
	@Override
	public Object2LongMap.Entry<T> floorEntry(T key) {
		Entry<T> node = findFloorNode(key);
		return node != null ? new BasicEntry<>(node.key, node.value) : null;
	}
	
	@Override
	public Object2LongMap.Entry<T> ceilingEntry(T key) {
		Entry<T> node = findCeilingNode(key);
		return node != null ? new BasicEntry<>(node.key, node.value) : null;
	}
	
	protected Entry<T> findLowerNode(T key) {
		Entry<T> entry = tree;
		while(entry != null) {
			if(compare(key, entry.key) > 0) {
				if(entry.right != null) entry = entry.right;
				else return entry;
			}
			else {
				if(entry.left != null) entry = entry.left;
				else {
					Entry<T> parent = entry.parent;
					while(parent != null && parent.left == entry) {
						entry = parent;
						parent = parent.parent;
					}
					return parent;
				}
			}
		}
		return null;
	}
	
	protected Entry<T> findFloorNode(T key) {
		Entry<T> entry = tree;
		int compare;
		while(entry != null) {
			if((compare = compare(key, entry.key)) > 0) {
				if(entry.right == null) break;
				entry = entry.right;
				continue;
			}
			else if(compare < 0) {
				if(entry.left != null) entry = entry.left;
				else {
					Entry<T> parent = entry.parent;
					while(parent != null && parent.left == entry) {
						entry = parent;
						parent = parent.parent;
					}
					return parent;
				}
				continue;
			}
			break;
		}
		return entry;
	}
	
	protected Entry<T> findCeilingNode(T key) {
		Entry<T> entry = tree;
		int compare;
		while(entry != null) {
			if((compare = compare(key, entry.key)) < 0) {
				if(entry.left == null) break;
				entry = entry.left;
				continue;
			}
			else if(compare > 0) {
				if(entry.right != null) entry = entry.right;
				else {
					Entry<T> parent = entry.parent;
					while(parent != null && parent.right == entry) {
						entry = parent;
						parent = parent.parent;
					}
					return parent;
				}
				continue;
			}
			break;
		}
		return entry;
	}
	
	protected Entry<T> findHigherNode(T key) {
		Entry<T> entry = tree;
		while(entry != null) {
			if(compare(key, entry.key) < 0) {
				if(entry.left != null) entry = entry.left;
				else return entry;
			}
			else {
				if(entry.right != null) entry = entry.right;
				else {
					Entry<T> parent = entry.parent;
					while(parent != null && parent.right == entry) {
						entry = parent;
						parent = parent.parent;
					}
					return parent;
				}
			}
		}
		return null;
	}
	
	protected Entry<T> findNode(T key) {
		Entry<T> node = tree;
		int compare;
		while(node != null) {
			if((compare = compare(key, node.key)) == 0) return node;
			if(compare < 0) node = node.left;
			else node = node.right;
		}
		return null;
	}
	
	protected void removeNode(Entry<T> entry) {
		size--;
		if(entry.needsSuccessor()) {
			Entry<T> successor = entry.next();
			entry.key = successor.key;
			entry.value = successor.value;
			entry = successor;
		}
		if(entry.previous() == null) first = entry.next();
		if(entry.next() == null) last = entry.previous();
		Entry<T> replacement = entry.left != null ? entry.left : entry.right;
		if(replacement != null) {
			if(entry.replace(replacement)) tree = replacement;
			entry.left = entry.right = entry.parent = null;
			fixAfterDeletion(replacement);
		}
		else if(entry.parent == null) tree = first = last = null;
		else {
			fixAfterDeletion(entry);
			entry.replace(null);
			entry.parent = null;
		}
	}
	
	protected void validate(T k) { compare(k, k); }
	protected int compare(T k, T v) { return comparator != null ? comparator.compare(k, v) : ((Comparable<T>)k).compareTo((T)v);}
	
	/** From CLR */
	protected void rotateLeft(Entry<T> entry) {
		if(entry != null) {
			Entry<T> right = entry.right;
			entry.right = right.left;
			if(right.left != null) right.left.parent = entry;
			right.parent = entry.parent;
			if(entry.parent == null) tree = right;
			else if(entry.parent.left == entry) entry.parent.left = right;
			else entry.parent.right = right;
			right.left = entry;
			entry.parent = right;
			entry.updateHeight();
			right.updateHeight();
		}
	}
	
	/** From CLR */
	protected void rotateRight(Entry<T> entry) {
		if(entry != null) {
			Entry<T> left = entry.left;
			entry.left = left.right;
			if(left.right != null) left.right.parent = entry;
			left.parent = entry.parent;
			if(entry.parent == null) tree = left;
			else if(entry.parent.right == entry) entry.parent.right = left;
			else entry.parent.left = left;
			left.right = entry;
			entry.parent = left;
			entry.updateHeight();
			left.updateHeight();
		}
	}
	
	/** From CLR */
	protected void fixAfterInsertion(Entry<T> entry) {
		while(entry != null) {
			entry.updateHeight();
			int balance = entry.getBalance();
			if(balance > 1) {
				int compare = entry.left.getBalance();
				if(compare > 0) rotateRight(entry);
				else if(compare < 0) {
					rotateLeft(entry.left);
					rotateRight(entry);
				}
			}
			else if(balance < -1) {
				int compare = entry.right.getBalance();
				if(compare < 0) rotateLeft(entry);
				else if(compare > 0) {
					rotateRight(entry.right);
					rotateLeft(entry);
				}
			}
			entry = entry.parent;
		}
	}
	
	/** From CLR */
	protected void fixAfterDeletion(Entry<T> entry) {
		if(entry != null) {
			entry.updateHeight();
			int balance = entry.getBalance();
			if(balance > 1) {
				int subBalance = entry.left.getBalance();
				if(subBalance >= 0) rotateRight(entry);
				else {
					rotateLeft(entry.left);
					rotateRight(entry);
				}
			}
			else if(balance < -1)
			{
				int subBalance = entry.right.getBalance();
				if(subBalance <= 0) rotateLeft(entry);
				else {
					rotateRight(entry.right);
					rotateLeft(entry);
				}
			}
			entry = entry.parent;
		}
	}
	
	static final class AscendingSubMap<T> extends NavigableSubMap<T> {

		public AscendingSubMap(Object2LongAVLTreeMap<T> m, boolean fromStart, T low, boolean loInclusive, boolean toEnd, T high, boolean hiInclusive) {
			super(m, fromStart, low, loInclusive, toEnd, high, hiInclusive);
		}
		
		@Override
		ObjectListIterator<T> keyIterator(boolean descending) { return new SubMapKeyIterator(descending); }
		@Override
		ObjectListIterator<T> keyIterator(T key) {
			Object2LongAVLTreeMap.Entry<T> entry = m.findNode(key);
			return entry == null || !inClosedRange(key) ? null : new SubMapKeyIterator(entry, fromStart ? null : findLowest(), toEnd ? null : findHighest());
		}
		@Override
		LongListIterator valueIterator() { return new SubMapValueIterator(false); }
		@Override
		ObjectListIterator<Object2LongMap.Entry<T>> entryIterator() { return new SubMapEntrySetIterator(false); }
		
		@Override
		public Object2LongNavigableMap<T> descendingMap() {
			if(descendingMap == null) descendingMap = new DescendingSubMap<>(m, fromStart, low, loInclusive, toEnd, high, hiInclusive);
			return descendingMap;
		}
		
		Object2LongAVLTreeMap.Entry<T> next(Object2LongAVLTreeMap.Entry<T> entry) {
			return entry.next();
		}
		
		Object2LongAVLTreeMap.Entry<T> previous(Object2LongAVLTreeMap.Entry<T> entry) {
			return entry.previous();
		}
		
		@Override
		public Object2LongNavigableMap<T> subMap(T fromKey, boolean fromInclusive, T toKey, boolean toInclusive) {
			if (!inRange(fromKey, fromInclusive)) throw new IllegalArgumentException("fromKey out of range");
			if (!inRange(toKey, toInclusive)) throw new IllegalArgumentException("toKey out of range");
			return new AscendingSubMap<>(m, false, fromKey, fromInclusive, false, toKey, toInclusive);
		}
		
		@Override
		public Object2LongNavigableMap<T> headMap(T toKey, boolean inclusive) {
			if (!inRange(toKey, inclusive)) throw new IllegalArgumentException("toKey out of range");
			return new AscendingSubMap<>(m, fromStart, low, loInclusive, false, toKey, inclusive);
		}
		
		@Override
		public Object2LongNavigableMap<T> tailMap(T fromKey, boolean inclusive) {
			if (!inRange(fromKey, inclusive)) throw new IllegalArgumentException("fromKey out of range");
			return new AscendingSubMap<>(m, false, fromKey, inclusive, toEnd, high, hiInclusive);
		}
	}
	
	static final class DescendingSubMap<T> extends NavigableSubMap<T> {

		public DescendingSubMap(Object2LongAVLTreeMap<T> m, boolean fromStart, T low, boolean loInclusive, boolean toEnd, T high, boolean hiInclusive) {
			super(m, fromStart, low, loInclusive, toEnd, high, hiInclusive);
		}
		
		@Override
		public T firstKey() { return super.lastKey(); }
		
		@Override
		public T pollFirstKey() { return super.pollLastKey(); }
		
		@Override
		public T lastKey() { return super.firstKey(); }
		
		@Override
		public T pollLastKey() { return super.pollFirstKey(); }
		
		@Override
		public long firstLongValue() { return super.lastLongValue(); }
		
		@Override
		public long lastLongValue() { return super.firstLongValue(); }
		
		@Override
		public Object2LongMap.Entry<T> firstEntry() { return super.lastEntry(); }
		
		@Override
		public Object2LongMap.Entry<T> lastEntry() { return super.firstEntry(); }
		
		@Override
		public Object2LongMap.Entry<T> pollFirstEntry() { return super.pollLastEntry(); }
		
		@Override
		public Object2LongMap.Entry<T> pollLastEntry() { return super.pollFirstEntry(); }
		
		@Override
		public T lowerKey(T e) { return super.higherKey(e); }
		
		@Override
		public T floorKey(T e) { return super.ceilingKey(e); }
		
		@Override
		public T ceilingKey(T e) { return super.floorKey(e); }
		
		@Override
		public T higherKey(T e) { return super.lowerKey(e); }
		
		@Override
		public Object2LongMap.Entry<T> lowerEntry(T e) { return super.higherEntry(e); }
		
		@Override
		public Object2LongMap.Entry<T> floorEntry(T e) { return super.ceilingEntry(e); }
		
		@Override
		public Object2LongMap.Entry<T> ceilingEntry(T e) { return super.floorEntry(e); }
		
		@Override
		public Object2LongMap.Entry<T> higherEntry(T e) { return super.lowerEntry(e); }
		
		protected Object2LongAVLTreeMap.Entry<T> findLowest() {
			if(fromStart) return m.first;
			Object2LongAVLTreeMap.Entry<T> entry = loInclusive ? m.findFloorNode(high) : m.findLowerNode(high);
			return entry == null || tooHigh(entry.key) ? null : entry;
		}
		
		protected Object2LongAVLTreeMap.Entry<T> findHighest() {
			if(toEnd) return m.last;
			Object2LongAVLTreeMap.Entry<T> entry = hiInclusive ? m.findCeilingNode(low) : m.findHigherNode(low);
			return entry == null || tooLow(entry.key) ? null : entry;
		}
		
		@Override
		ObjectListIterator<T> keyIterator(boolean descending) {
			ObjectListIterator<T> iter = new SubMapKeyIterator(!descending);
			return descending ? iter : ObjectIterators.invert(iter);
		}

		@Override
		ObjectListIterator<T> keyIterator(T key) {
			Object2LongAVLTreeMap.Entry<T> entry = m.findNode(key);
			return entry == null || !inClosedRange(key) ? null : ObjectIterators.invert(new SubMapKeyIterator(entry, fromStart ? null : findLowest(), toEnd ? null : findHighest()));
		}
		
		Object2LongAVLTreeMap.Entry<T> next(Object2LongAVLTreeMap.Entry<T> entry) {
			return entry.previous();
		}
		
		Object2LongAVLTreeMap.Entry<T> previous(Object2LongAVLTreeMap.Entry<T> entry) {
			return entry.next();
		}
		
		@Override
		public Comparator<T> comparator() { return m.comparator() == null ? null : m.comparator().reversed(); }
		
		@Override
		LongListIterator valueIterator() { return LongIterators.invert(new SubMapValueIterator(true)); }
		
		@Override
		ObjectListIterator<Object2LongMap.Entry<T>> entryIterator() { return ObjectIterators.invert(new SubMapEntrySetIterator(true)); }
		
		@Override
		public Object2LongNavigableMap<T> descendingMap() {
			if(descendingMap == null) descendingMap = new AscendingSubMap<>(m, fromStart, low, loInclusive, toEnd, high, hiInclusive);
			return descendingMap;
		}

		@Override
		public Object2LongNavigableMap<T> subMap(T fromKey, boolean fromInclusive, T toKey, boolean toInclusive) {
			if (!inRange(fromKey, fromInclusive)) throw new IllegalArgumentException("fromKey out of range");
			if (!inRange(toKey, toInclusive)) throw new IllegalArgumentException("toKey out of range");
			return new DescendingSubMap<>(m, false, fromKey, fromInclusive, false, toKey, toInclusive);
		}
		
		@Override
		public Object2LongNavigableMap<T> headMap(T toKey, boolean inclusive) {
			if (!inRange(toKey, inclusive)) throw new IllegalArgumentException("toKey out of range");
			return new DescendingSubMap<>(m, fromStart, low, loInclusive, false, toKey, inclusive);
		}
		
		@Override
		public Object2LongNavigableMap<T> tailMap(T fromKey, boolean inclusive) {
			if (!inRange(fromKey, inclusive)) throw new IllegalArgumentException("fromKey out of range");
			return new DescendingSubMap<>(m, false, fromKey, inclusive, toEnd, high, hiInclusive);
		}
	}
	
	static abstract class NavigableSubMap<T> extends AbstractObject2LongMap<T> implements Object2LongNavigableMap<T> {
		final Object2LongAVLTreeMap<T> m;
		final T low;
		final T high;
		final boolean fromStart;
		final boolean toEnd;
		final boolean loInclusive;
		final boolean hiInclusive;
		transient Object2LongNavigableMap<T> descendingMap;
		transient SubMapEntrySet entrySet;
		transient KeySet<T> keySet;
		transient SubMapValues values;
		
		NavigableSubMap(Object2LongAVLTreeMap<T> m, boolean fromStart, T low, boolean loInclusive, boolean toEnd, T high, boolean hiInclusive) {
			this.m = m;
			this.low = low;
			this.high = high;
			this.fromStart = fromStart;
			this.toEnd = toEnd;
			this.loInclusive = loInclusive;
			this.hiInclusive = hiInclusive;
		}
		
		abstract ObjectListIterator<T> keyIterator(boolean descending);
		abstract ObjectListIterator<T> keyIterator(T key);
		abstract LongListIterator valueIterator();
		abstract ObjectListIterator<Object2LongMap.Entry<T>> entryIterator();
		abstract Object2LongAVLTreeMap.Entry<T> next(Object2LongAVLTreeMap.Entry<T> entry);
		abstract Object2LongAVLTreeMap.Entry<T> previous(Object2LongAVLTreeMap.Entry<T> entry);
		
		@Override
		public NavigableSubMap<T> copy() { throw new UnsupportedOperationException(); }
		@Override
		public abstract Object2LongNavigableMap<T> descendingMap();
		@Override
		public abstract Object2LongNavigableMap<T> subMap(T fromKey, boolean fromInclusive, T toKey, boolean toInclusive);
		@Override
		public abstract Object2LongNavigableMap<T> headMap(T toKey, boolean inclusive);
		@Override
		public abstract Object2LongNavigableMap<T> tailMap(T fromKey, boolean inclusive); 
		
		boolean tooLow(T key) { return !fromStart && (loInclusive ? m.compare(key, low) < 0 : m.compare(key, low) <= 0); }
		boolean tooHigh(T key) { return !toEnd && (hiInclusive ? m.compare(key, high) > 0 : m.compare(key, high) >= 0); }
		boolean inRange(T key) { return !tooLow(key) && !tooHigh(key); }
		boolean inClosedRange(T key) { return (fromStart || m.compare(key, low) >= 0) && (toEnd || m.compare(high, key) >= 0); }
		boolean inRange(T key, boolean inclusive) { return inclusive ? inRange(key) : inClosedRange(key); }
		
		public T getDefaultMaxValue() { return m.getDefaultMaxValue(); }
		public T getDefaultMinValue() { return m.getDefaultMinValue(); }
		@Override
		public AbstractObject2LongMap<T> setDefaultReturnValue(long v) { 
			m.setDefaultReturnValue(v);
			return this;
		}
		
		@Override
		public long getDefaultReturnValue() { return m.getDefaultReturnValue(); }
		
		@Override
		public long putAndMoveToFirst(T key, long value) { throw new UnsupportedOperationException(); }
		@Override
		public long putAndMoveToLast(T key, long value) { throw new UnsupportedOperationException(); }
		@Override
		public boolean moveToFirst(T key) { throw new UnsupportedOperationException(); }
		@Override
		public boolean moveToLast(T key) { throw new UnsupportedOperationException(); }
		@Override
		public long getAndMoveToFirst(T key) { throw new UnsupportedOperationException(); }
		@Override
		public long getAndMoveToLast(T key) { throw new UnsupportedOperationException(); }
		@Override
		public Comparator<T> comparator() { return m.comparator(); }
		@Override
		public boolean containsKey(Object key) { return inRange((T)key) && m.containsKey(key); }
		@Override
		public T firstKey() { 
			Object2LongAVLTreeMap.Entry<T> entry = findLowest();
			return entry == null ? getDefaultMaxValue() : entry.key;
		}
		
		@Override
		public T pollFirstKey() {
			if(fromStart) return m.pollFirstKey();
			Object2LongAVLTreeMap.Entry<T> entry = loInclusive ? m.findCeilingNode(low) : m.findHigherNode(low);
			if(entry != null && !tooHigh(entry.key)) {
				T value = entry.key;
				m.removeNode(entry);
				return value;
			}
			return getDefaultMaxValue();
		}
		
		@Override
		public T lastKey() {
			Object2LongAVLTreeMap.Entry<T> entry = findHighest();
			return entry == null ? getDefaultMinValue() : entry.key;	
		}
		
		@Override
		public T pollLastKey() {
			if(toEnd) return m.pollLastKey();
			Object2LongAVLTreeMap.Entry<T> entry = hiInclusive ? m.findFloorNode(high) : m.findLowerNode(high);
			if(entry != null && !tooLow(entry.key)) {
				T value = entry.key;
				m.removeNode(entry);
				return value;
			}
			return getDefaultMinValue();
		}
		
		@Override
		public long firstLongValue() {
			Object2LongAVLTreeMap.Entry<T> entry = findLowest();
			return entry == null ? getDefaultReturnValue() : entry.value;
		}
		
		@Override
		public long lastLongValue() {
			Object2LongAVLTreeMap.Entry<T> entry = findHighest();
			return entry == null ? getDefaultReturnValue() : entry.value;
		}
		
		protected Object2LongAVLTreeMap.Entry<T> findLowest() {
			if(fromStart) return m.first;
			Object2LongAVLTreeMap.Entry<T> entry = loInclusive ? m.findCeilingNode(low) : m.findHigherNode(low);
			return entry == null || tooHigh(entry.key) ? null : entry;
		}
		
		protected Object2LongAVLTreeMap.Entry<T> findHighest() {
			if(toEnd) return m.last;
			Object2LongAVLTreeMap.Entry<T> entry = hiInclusive ? m.findFloorNode(high) : m.findLowerNode(high);
			return entry == null || tooLow(entry.key) ? null : entry;
		}
		
		@Override
		public long put(T key, long value) {
			if(!inRange(key)) throw new IllegalArgumentException("key out of range");
			return m.put(key, value);
		}
		
		@Override
		public long putIfAbsent(T key, long value) {
			if(!inRange(key)) throw new IllegalArgumentException("key out of range");
			return m.putIfAbsent(key, value);
		}
		
		@Override
		public long addTo(T key, long value) {
			if(!inRange(key)) throw new IllegalArgumentException("key out of range");
			return m.addTo(key, value);
		}
		
		@Override
		public long getLong(T key) {
			return inRange(key) ? m.getLong(key) : getDefaultReturnValue();
		}
		
		@Override
		public long getOrDefault(T key, long defaultValue) {
			return inRange(key) ? m.getOrDefault(key, defaultValue) : getDefaultReturnValue();
		}
		
		@Override
		public long rem(T key) {
			return inRange(key) ? m.rem(key) : getDefaultReturnValue();
		}
		
		@Override
		public long remOrDefault(T key, long defaultValue) {
			return inRange(key) ? m.rem(key) : defaultValue;
		}
		
		@Override
		public boolean remove(T key, long value) {
			return inRange(key) && m.remove(key, value);
		}
		
		@Override
		public ObjectSet<Object2LongMap.Entry<T>> object2LongEntrySet() {
			if(entrySet == null) entrySet = new SubMapEntrySet();
			return entrySet;
		}
		
		@Override
		public ObjectSet<T> keySet() { return navigableKeySet(); }
		
		@Override
		public ObjectNavigableSet<T> navigableKeySet() {
			if(keySet == null) keySet = new KeySet<>(this);
			return keySet;
		}
		
		@Override
		public ObjectNavigableSet<T> descendingKeySet() {
			return descendingMap().navigableKeySet();
		}
		
		@Override
		public LongCollection values() {
			if(values == null) values = new SubMapValues();
			return values;
		}
		
		@Override
		public void forEach(ObjectLongConsumer<T> action) {
			for(Object2LongAVLTreeMap.Entry<T> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
				action.accept(entry.key, entry.value);
		}
		
		@Override
		public int size() {
			return (fromStart && toEnd) ? m.size() : entrySet().size();
		}
		
		@Override
		public Object2LongMap.Entry<T> firstEntry() {
			Object2LongAVLTreeMap.Entry<T> entry = findLowest();
			return entry == null ? null : new BasicEntry<>(entry.key, entry.value);
		}
		
		@Override
		public Object2LongMap.Entry<T> lastEntry() {
			Object2LongAVLTreeMap.Entry<T> entry = findHighest();
			return entry == null ? null : new BasicEntry<>(entry.key, entry.value);
		}
		
		@Override
		public Object2LongMap.Entry<T> pollFirstEntry() {
			Object2LongAVLTreeMap.Entry<T> entry = findLowest();
			if(entry == null) return null;
			BasicEntry<T> result = new BasicEntry<>(entry.key, entry.value);
			m.removeNode(entry);
			return result;
		}
		
		@Override
		public Object2LongMap.Entry<T> pollLastEntry() {
			Object2LongAVLTreeMap.Entry<T> entry = findHighest();
			if(entry == null) return null;
			BasicEntry<T> result = new BasicEntry<>(entry.key, entry.value);
			m.removeNode(entry);
			return result;
		}
		
		@Override
		public T lowerKey(T e) {
			if(tooHigh(e)) {
				Object2LongAVLTreeMap.Entry<T> entry = findHighest();
				return entry == null ? getDefaultMinValue() : entry.key;
			}
			Object2LongAVLTreeMap.Entry<T> entry = m.findLowerNode(e);
			return entry == null || tooHigh(entry.key) ? getDefaultMaxValue() : entry.key;
		}
		
		@Override
		public T floorKey(T e) {
			if(tooHigh(e)) {
				Object2LongAVLTreeMap.Entry<T> entry = findHighest();
				return entry == null ? getDefaultMinValue() : entry.key;
			}
			Object2LongAVLTreeMap.Entry<T> entry = m.findFloorNode(e);
			return entry == null || tooHigh(entry.key) ? getDefaultMaxValue() : entry.key;
		}
		
		@Override
		public T ceilingKey(T e) {
			if(tooLow(e)) {
				Object2LongAVLTreeMap.Entry<T> entry = findLowest();
				return entry == null ? getDefaultMaxValue() : entry.key;
			}
			Object2LongAVLTreeMap.Entry<T> entry = m.findCeilingNode(e);
			return entry == null || tooLow(entry.key) ? getDefaultMinValue() : entry.key;
		}
		
		@Override
		public T higherKey(T e) {
			if(tooLow(e)) {
				Object2LongAVLTreeMap.Entry<T> entry = findLowest();
				return entry == null ? getDefaultMaxValue() : entry.key;
			}
			Object2LongAVLTreeMap.Entry<T> entry = m.findHigherNode(e);
			return entry == null || tooLow(entry.key) ? getDefaultMinValue() : entry.key;
		}
		
		@Override
		public Object2LongMap.Entry<T> lowerEntry(T e) {
			if(tooHigh(e)) {
				Object2LongAVLTreeMap.Entry<T> entry = findHighest();
				return entry == null ? null : new BasicEntry<>(entry.key, entry.value);
			}
			Object2LongAVLTreeMap.Entry<T> entry = m.findLowerNode(e);
			return entry == null || tooHigh(entry.key) ? null : new BasicEntry<>(entry.key, entry.value);
		}
		
		@Override
		public Object2LongMap.Entry<T> floorEntry(T e) {
			if(tooHigh(e)) {
				Object2LongAVLTreeMap.Entry<T> entry = findHighest();
				return entry == null ? null : new BasicEntry<>(entry.key, entry.value);
			}
			Object2LongAVLTreeMap.Entry<T> entry = m.findFloorNode(e);
			return entry == null || tooHigh(entry.key) ? null : new BasicEntry<>(entry.key, entry.value);
		}
		
		@Override
		public Object2LongMap.Entry<T> ceilingEntry(T e) {
			if(tooLow(e)) {
				Object2LongAVLTreeMap.Entry<T> entry = findLowest();
				return entry == null ? null : new BasicEntry<>(entry.key, entry.value);
			}
			Object2LongAVLTreeMap.Entry<T> entry = m.findCeilingNode(e);
			return entry == null || tooLow(entry.key) ? null : new BasicEntry<>(entry.key, entry.value);
		}
		
		@Override
		public Object2LongMap.Entry<T> higherEntry(T e) {
			if(tooLow(e)) {
				Object2LongAVLTreeMap.Entry<T> entry = findLowest();
				return entry == null ? null : new BasicEntry<>(entry.key, entry.value);
			}
			Object2LongAVLTreeMap.Entry<T> entry = m.findHigherNode(e);
			return entry == null || tooLow(entry.key) ? null : new BasicEntry<>(entry.key, entry.value);
		}
		
		class SubMapEntrySet extends AbstractObjectSet<Object2LongMap.Entry<T>> implements ObjectSet<Object2LongMap.Entry<T>> {
			@Override
			@Deprecated
			public boolean contains(Object o) {
				if(o instanceof Map.Entry) {
					if(o instanceof Object2LongMap.Entry) {
						Object2LongMap.Entry<T> entry = (Object2LongMap.Entry<T>)o;
						if(entry.getKey() == null && comparator() == null) return false;
						Object2LongAVLTreeMap.Entry<T> subEntry = m.findNode(entry.getKey());
						if(subEntry != null) return entry.getLongValue() == subEntry.value;
					}
					else {
						Map.Entry<?, ?> entry = (Map.Entry<?, ?>)o;
						if(entry.getKey() == null && comparator() == null) return false;
						Object2LongAVLTreeMap.Entry<T> subEntry = m.findNode((T)entry.getKey());
						if(subEntry != null) return Objects.equals(entry.getValue(), Long.valueOf(subEntry.value));
					}
				}
				return false;
			}
			
			@Override
			@Deprecated
			public boolean remove(Object o) {
				if(o instanceof Map.Entry) {
					if(o instanceof Object2LongMap.Entry) {
						Object2LongMap.Entry<T> entry = (Object2LongMap.Entry<T>)o;
						return NavigableSubMap.this.remove(entry.getKey(), entry.getLongValue());
					}
					Map.Entry<?, ?> entry = (Map.Entry<?, ?>)o;
					return NavigableSubMap.this.remove(entry.getKey(), entry.getValue());
				}
				return false;
			}
			@Override
			public ObjectIterator<Object2LongMap.Entry<T>> iterator() { return entryIterator(); }
			@Override
			public int size() { return NavigableSubMap.this.size(); }
			@Override
			public void clear() { NavigableSubMap.this.clear(); }
			
			@Override
			public void forEach(Consumer<? super Object2LongMap.Entry<T>> action) {
				Objects.requireNonNull(action);
				for(Object2LongAVLTreeMap.Entry<T> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					action.accept(new BasicEntry<>(entry.key, entry.value));
			}
			
			@Override
			public <E> void forEach(E input, ObjectObjectConsumer<E, Object2LongMap.Entry<T>> action) {
				Objects.requireNonNull(action);
				for(Object2LongAVLTreeMap.Entry<T> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					action.accept(input, new BasicEntry<>(entry.key, entry.value));
			}
			
			@Override
			public boolean matchesAny(Object2BooleanFunction<Object2LongMap.Entry<T>> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return false;
				BasicEntry<T> subEntry = new BasicEntry<>();
				for(Object2LongAVLTreeMap.Entry<T> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					subEntry.set(entry.key, entry.value);
					if(filter.getBoolean(subEntry)) return true;
				}
				return false;
			}
			
			@Override
			public boolean matchesNone(Object2BooleanFunction<Object2LongMap.Entry<T>> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return true;
				BasicEntry<T> subEntry = new BasicEntry<>();
				for(Object2LongAVLTreeMap.Entry<T> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					subEntry.set(entry.key, entry.value);
					if(filter.getBoolean(subEntry)) return false;
				}
				return true;
			}
			
			@Override
			public boolean matchesAll(Object2BooleanFunction<Object2LongMap.Entry<T>> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return true;
				BasicEntry<T> subEntry = new BasicEntry<>();
				for(Object2LongAVLTreeMap.Entry<T> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					subEntry.set(entry.key, entry.value);
					if(!filter.getBoolean(subEntry)) return false;
				}
				return true;
			}
			
			@Override
			public <E> E reduce(E identity, BiFunction<E, Object2LongMap.Entry<T>, E> operator) {
				Objects.requireNonNull(operator);
				E state = identity;
				for(Object2LongAVLTreeMap.Entry<T> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					state = operator.apply(state, new BasicEntry<>(entry.key, entry.value));
				}
				return state;
			}
			
			@Override
			public Object2LongMap.Entry<T> reduce(ObjectObjectUnaryOperator<Object2LongMap.Entry<T>, Object2LongMap.Entry<T>> operator) {
				Objects.requireNonNull(operator);
				Object2LongMap.Entry<T> state = null;
				boolean empty = true;
				for(Object2LongAVLTreeMap.Entry<T> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					if(empty) {
						empty = false;
						state = new BasicEntry<>(entry.key, entry.value);
						continue;
					}
					state = operator.apply(state, new BasicEntry<>(entry.key, entry.value));
				}
				return state;
			}
			
			@Override
			public Object2LongMap.Entry<T> findFirst(Object2BooleanFunction<Object2LongMap.Entry<T>> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return null;
				BasicEntry<T> subEntry = new BasicEntry<>();
				for(Object2LongAVLTreeMap.Entry<T> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					subEntry.set(entry.key, entry.value);
					if(filter.getBoolean(subEntry)) return subEntry;
				}
				return null;
			}
			
			@Override
			public int count(Object2BooleanFunction<Object2LongMap.Entry<T>> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return 0;
				int result = 0;
				BasicEntry<T> subEntry = new BasicEntry<>();
				for(Object2LongAVLTreeMap.Entry<T> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					subEntry.set(entry.key, entry.value);
					if(filter.getBoolean(subEntry)) result++;
				}
				return result;
			}
		}
		
		final class SubMapValues extends AbstractLongCollection {
			@Override
			public boolean add(long o) { throw new UnsupportedOperationException(); }
			
			@Override
			public boolean contains(long e) {
				return containsValue(e);
			}
			
			@Override
			public LongIterator iterator() { return valueIterator(); }
			
			@Override
			public int size() {
				return NavigableSubMap.this.size();
			}
			
			@Override
			public void clear() {
				NavigableSubMap.this.clear();
			}
			
			@Override
			public void forEach(LongConsumer action) {
				Objects.requireNonNull(action);
				for(Object2LongAVLTreeMap.Entry<T> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					action.accept(entry.value);
			}
			
			@Override
			public <E> void forEach(E input, ObjectLongConsumer<E> action) {
				Objects.requireNonNull(action);
				for(Object2LongAVLTreeMap.Entry<T> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					action.accept(input, entry.value);
			}
			
			@Override
			public boolean matchesAny(Long2BooleanFunction filter) {
				Objects.requireNonNull(filter);
				for(Object2LongAVLTreeMap.Entry<T> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					if(filter.get(entry.value)) return true;
				return false;
			}
			
			@Override
			public boolean matchesNone(Long2BooleanFunction filter) {
				Objects.requireNonNull(filter);
				for(Object2LongAVLTreeMap.Entry<T> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					if(filter.get(entry.value)) return false;
				return true;
			}
			
			@Override
			public boolean matchesAll(Long2BooleanFunction filter) {
				Objects.requireNonNull(filter);
				for(Object2LongAVLTreeMap.Entry<T> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					if(!filter.get(entry.value)) return false;
				return true;
			}
			
			@Override
			public long reduce(long identity, LongLongUnaryOperator operator) {
				Objects.requireNonNull(operator);
				long state = identity;
				for(Object2LongAVLTreeMap.Entry<T> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					state = operator.applyAsLong(state, entry.value);
				return state;
			}
			
			@Override
			public long reduce(LongLongUnaryOperator operator) {
				Objects.requireNonNull(operator);
				long state = 0L;
				boolean empty = true;
				for(Object2LongAVLTreeMap.Entry<T> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					if(empty) {
						empty = false;
						state = entry.value;
						continue;
					}
					state = operator.applyAsLong(state, entry.value);
				}
				return state;
			}
			
			@Override
			public long findFirst(Long2BooleanFunction filter) {
				Objects.requireNonNull(filter);
				for(Object2LongAVLTreeMap.Entry<T> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					if(filter.get(entry.value)) return entry.value;
				return 0L;
			}
			
			@Override
			public int count(Long2BooleanFunction filter) {
				Objects.requireNonNull(filter);
				int result = 0;
				for(Object2LongAVLTreeMap.Entry<T> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					if(filter.get(entry.value)) result++;
				return result;
			}
		}
		
		final class SubMapEntrySetIterator extends SubMapEntryIterator implements ObjectListIterator<Object2LongMap.Entry<T>> {
			public SubMapEntrySetIterator(boolean descending) {
				super(descending);
			}
			
			public SubMapEntrySetIterator(Object2LongAVLTreeMap.Entry<T> entry, Object2LongAVLTreeMap.Entry<T> lowerFence, Object2LongAVLTreeMap.Entry<T> upperFence) {
				super(entry, lowerFence, upperFence);
			}
			
			@Override
			public Object2LongMap.Entry<T> next() { return nextEntry(); }
			@Override
			public Object2LongMap.Entry<T> previous() { return previousEntry(); }
			
			@Override
			public void set(Object2LongMap.Entry<T> e) { throw new UnsupportedOperationException(); }
			@Override
			public void add(Object2LongMap.Entry<T> e) { throw new UnsupportedOperationException(); }
		}
		
		final class SubMapKeyIterator extends SubMapEntryIterator implements ObjectListIterator<T> {
			public SubMapKeyIterator(boolean descending) {
				super(descending);
			}
			
			public SubMapKeyIterator(Object2LongAVLTreeMap.Entry<T> entry, Object2LongAVLTreeMap.Entry<T> lowerFence, Object2LongAVLTreeMap.Entry<T> upperFence) {
				super(entry, lowerFence, upperFence);
			}
			
			@Override
			public T next() { return nextEntry().key; }
			@Override
			public T previous() { return previousEntry().key; }
			
			@Override
			public void set(T e) { throw new UnsupportedOperationException(); }
			@Override
			public void add(T e) { throw new UnsupportedOperationException(); }
		}
		
		final class SubMapValueIterator extends SubMapEntryIterator implements LongListIterator {
			public SubMapValueIterator(boolean descending) {
				super(descending);
			}
			
			public SubMapValueIterator(Object2LongAVLTreeMap.Entry<T> entry, Object2LongAVLTreeMap.Entry<T> lowerFence, Object2LongAVLTreeMap.Entry<T> upperFence) {
				super(entry, lowerFence, upperFence);
			}
			
			@Override
			public long nextLong() { return nextEntry().value; }
			@Override
			public long previousLong() { return previousEntry().value; }
			
			@Override
			public void set(long e) { throw new UnsupportedOperationException(); }
			@Override
			public void add(long e) { throw new UnsupportedOperationException(); }
		}
		
		abstract class SubMapEntryIterator {
			T lowerFence;
			T upperFence;
			Object2LongAVLTreeMap.Entry<T> next;
			Object2LongAVLTreeMap.Entry<T> previous;
			Object2LongAVLTreeMap.Entry<T> current;
			int index = 0;
			
			public SubMapEntryIterator(boolean descending) {
				this(descending ? findHighest() : findLowest(), fromStart ? null : findLowest(), toEnd ? null : findHighest());
			}
			
			public SubMapEntryIterator(Object2LongAVLTreeMap.Entry<T> entry, Object2LongAVLTreeMap.Entry<T> lowerFence, Object2LongAVLTreeMap.Entry<T> upperFence) {
				next = entry;
				previous = entry.previous();
				this.lowerFence = lowerFence != null ? lowerFence.key : null;
				this.upperFence = upperFence != null ? upperFence.key : null;
			}
			
			public boolean hasNext() {
				return next != null && (upperFence == null || Objects.equals(next.key, upperFence));
			}
			
			public boolean hasPrevious() {
				return previous != null && (lowerFence == null || Objects.equals(next.key, lowerFence));
			}
			
			public int nextIndex() {
				return index;
			}
			
			public int previousIndex() {
				return index - 1;
			}
			
			protected void updateNext() {
				next = current.next();
			}
			
			protected void updatePrevious() {
				previous = current.previous();
			}
			
			public Object2LongAVLTreeMap.Entry<T> nextEntry() {
				if(!hasNext()) throw new NoSuchElementException();
				current = previous = next;
				updateNext();
				index++;
				return current;
			}
			
			public Object2LongAVLTreeMap.Entry<T> previousEntry() {
				if(!hasPrevious()) throw new NoSuchElementException();
				current = next = previous;
				updatePrevious();
				index--;
				return current;
			}
			
			public void remove() {
				if(current == null) throw new IllegalStateException();
				if(current == previous) index--;
				updateNext();
				updatePrevious();
				if(current.needsSuccessor()) next = current;
				m.removeNode(current);
				current = null;
			}
		}
	}
	
	private class EntrySet extends AbstractObjectSet<Object2LongMap.Entry<T>> {
		
		@Override
		public ObjectIterator<Object2LongMap.Entry<T>> iterator() {
			return new EntryIterator(false);
		}
		
		@Override
		@Deprecated
		public boolean contains(Object o) {
			if(o instanceof Map.Entry) {
				if(o instanceof Object2LongMap.Entry) {
					Object2LongMap.Entry<T> entry = (Object2LongMap.Entry<T>)o;
					if(entry.getKey() == null && comparator() == null) return false;
					Object2LongAVLTreeMap.Entry<T> subEntry = Object2LongAVLTreeMap.this.findNode(entry.getKey());
					if(subEntry != null) return entry.getLongValue() == subEntry.value;
				}
				else {
					Map.Entry<?, ?> entry = (Map.Entry<?, ?>)o;
					if(entry.getKey() == null && comparator() == null) return false;
					Object2LongAVLTreeMap.Entry<T> subEntry = Object2LongAVLTreeMap.this.findNode((T)entry.getKey());
					if(subEntry != null) return Objects.equals(entry.getValue(), Long.valueOf(subEntry.value));
				}
			}
			return false;
		}
		
		@Override
		@Deprecated
		public boolean remove(Object o) {
			if(o instanceof Map.Entry) {
				if(o instanceof Object2LongMap.Entry) {
					Object2LongMap.Entry<T> entry = (Object2LongMap.Entry<T>)o;
					return Object2LongAVLTreeMap.this.remove(entry.getKey(), entry.getLongValue());
				}
				Map.Entry<?, ?> entry = (Map.Entry<?, ?>)o;
				return Object2LongAVLTreeMap.this.remove(entry.getKey(), entry.getValue());
			}
			return false;
		}
		
		@Override
		public int size() {
			return Object2LongAVLTreeMap.this.size();
		}
		
		@Override
		public void clear() {
			Object2LongAVLTreeMap.this.clear();
		}
		
		@Override
		public void forEach(Consumer<? super Object2LongMap.Entry<T>> action) {
			Objects.requireNonNull(action);
			for(Object2LongAVLTreeMap.Entry<T> entry = first;entry != null;entry = entry.next())
				action.accept(new BasicEntry<>(entry.key, entry.value));
		}
		
		@Override
		public <E> void forEach(E input, ObjectObjectConsumer<E, Object2LongMap.Entry<T>> action) {
			Objects.requireNonNull(action);
			for(Object2LongAVLTreeMap.Entry<T> entry = first;entry != null;entry = entry.next())
				action.accept(input, new BasicEntry<>(entry.key, entry.value));
		}
		
		@Override
		public boolean matchesAny(Object2BooleanFunction<Object2LongMap.Entry<T>> filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return false;
			BasicEntry<T> subEntry = new BasicEntry<>();
			for(Object2LongAVLTreeMap.Entry<T> entry = first;entry != null;entry = entry.next()) {
				subEntry.set(entry.key, entry.value);
				if(filter.getBoolean(subEntry)) return true;
			}
			return false;
		}
		
		@Override
		public boolean matchesNone(Object2BooleanFunction<Object2LongMap.Entry<T>> filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return true;
			BasicEntry<T> subEntry = new BasicEntry<>();
			for(Object2LongAVLTreeMap.Entry<T> entry = first;entry != null;entry = entry.next()) {
				subEntry.set(entry.key, entry.value);
				if(filter.getBoolean(subEntry)) return false;
			}
			return true;
		}
		
		@Override
		public boolean matchesAll(Object2BooleanFunction<Object2LongMap.Entry<T>> filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return true;
			BasicEntry<T> subEntry = new BasicEntry<>();
			for(Object2LongAVLTreeMap.Entry<T> entry = first;entry != null;entry = entry.next()) {
				subEntry.set(entry.key, entry.value);
				if(!filter.getBoolean(subEntry)) return false;
			}
			return true;
		}
		
		@Override
		public <E> E reduce(E identity, BiFunction<E, Object2LongMap.Entry<T>, E> operator) {
			Objects.requireNonNull(operator);
			E state = identity;
			for(Object2LongAVLTreeMap.Entry<T> entry = first;entry != null;entry = entry.next()) {
				state = operator.apply(state, new BasicEntry<>(entry.key, entry.value));
			}
			return state;
		}
		
		@Override
		public Object2LongMap.Entry<T> reduce(ObjectObjectUnaryOperator<Object2LongMap.Entry<T>, Object2LongMap.Entry<T>> operator) {
			Objects.requireNonNull(operator);
			Object2LongMap.Entry<T> state = null;
			boolean empty = true;
			for(Object2LongAVLTreeMap.Entry<T> entry = first;entry != null;entry = entry.next()) {
				if(empty) {
					empty = false;
					state = new BasicEntry<>(entry.key, entry.value);
					continue;
				}
				state = operator.apply(state, new BasicEntry<>(entry.key, entry.value));
			}
			return state;
		}
		
		@Override
		public Object2LongMap.Entry<T> findFirst(Object2BooleanFunction<Object2LongMap.Entry<T>> filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return null;
			BasicEntry<T> subEntry = new BasicEntry<>();
			for(Object2LongAVLTreeMap.Entry<T> entry = first;entry != null;entry = entry.next()) {
				subEntry.set(entry.key, entry.value);
				if(filter.getBoolean(subEntry)) return subEntry;
			}
			return null;
		}
	}
	
	static final class KeySet<T> extends AbstractObjectSet<T> implements ObjectNavigableSet<T> {
		Object2LongNavigableMap<T> m;
		
		KeySet(Object2LongNavigableMap<T> m) {
			this.m = m;
		}
		
		@Override
		public boolean add(T o) { throw new UnsupportedOperationException(); }
		@Override
		public boolean addAndMoveToFirst(T o) { throw new UnsupportedOperationException(); }
		@Override
		public boolean addAndMoveToLast(T o) { throw new UnsupportedOperationException(); }
		@Override
		public boolean moveToFirst(T o) { throw new UnsupportedOperationException(); }
		@Override
		public boolean moveToLast(T o) { throw new UnsupportedOperationException(); }
		@Override
		public Comparator<T> comparator() { return m.comparator(); }
		@Override
		public T lower(T e) { return m.lowerKey(e); }
		@Override
		public T floor(T e) { return m.floorKey(e); }
		@Override
		public T ceiling(T e) { return m.ceilingKey(e); }
		@Override
		public T higher(T e) { return m.higherKey(e); }
		@Override
		public T first() { return m.firstKey(); }
		@Override
		public T pollFirst() { return m.pollFirstKey(); }
		@Override
		public T last() { return m.lastKey(); }
		@Override
		public T pollLast() { return m.pollLastKey(); }		
		@Override
		public boolean remove(Object o) { 
			int oldSize = m.size();
			m.remove(o); 
			return oldSize != m.size();
		}
		
		@Override
		public KeySet<T> copy() { throw new UnsupportedOperationException(); }
		
		@Override
		public ObjectNavigableSet<T> subSet(T fromElement, boolean fromInclusive, T toElement, boolean toInclusive) {
			return new KeySet<>(m.subMap(fromElement, fromInclusive, toElement, toInclusive));
		}
		
		@Override
		public ObjectNavigableSet<T> headSet(T toElement, boolean inclusive) {
			return new KeySet<>(m.headMap(toElement, inclusive));
		}
		
		@Override
		public ObjectNavigableSet<T> tailSet(T fromElement, boolean inclusive) {
			return new KeySet<>(m.tailMap(fromElement, inclusive));
		}
		
		@Override
		public ObjectNavigableSet<T> descendingSet() {
			return new KeySet<>(m.descendingMap());
		}
		
		@Override
		public ObjectBidirectionalIterator<T> iterator() {
			if(m instanceof Object2LongAVLTreeMap) return ((Object2LongAVLTreeMap<T>)m).keyIterator(false);
			return ((NavigableSubMap<T>)m).keyIterator(false);
		}
		
		@Override
		public ObjectBidirectionalIterator<T> iterator(T fromElement) {
			if(m instanceof Object2LongAVLTreeMap) return ((Object2LongAVLTreeMap<T>)m).keyIterator(fromElement);
			return ((NavigableSubMap<T>)m).keyIterator(fromElement);
		}
		
		@Override
		public ObjectBidirectionalIterator<T> descendingIterator() {
			if(m instanceof Object2LongAVLTreeMap) return ((Object2LongAVLTreeMap<T>)m).keyIterator(true);
			return ((NavigableSubMap<T>)m).keyIterator(true);
		}
		
		@Override
		public int size() {
			return m.size();
		}
		
		@Override
		public void clear() {
			m.clear();
		}
		
		protected Object2LongAVLTreeMap.Entry<T> start() {
			if(m instanceof Object2LongAVLTreeMap) return ((Object2LongAVLTreeMap<T>)m).first;
			return ((NavigableSubMap<T>)m).findLowest();
		}
		
		protected Object2LongAVLTreeMap.Entry<T> end() {
			if(m instanceof Object2LongAVLTreeMap) return null;
			return ((NavigableSubMap<T>)m).findLowest();
		}
		
		protected Object2LongAVLTreeMap.Entry<T> next(Object2LongAVLTreeMap.Entry<T> entry) {
			if(m instanceof Object2LongAVLTreeMap) return entry.next();
			return ((NavigableSubMap<T>)m).next(entry);
		}
		
		protected Object2LongAVLTreeMap.Entry<T> previous(Object2LongAVLTreeMap.Entry<T> entry) {
			if(m instanceof Object2LongAVLTreeMap) return entry.previous();
			return ((NavigableSubMap<T>)m).previous(entry);
		}
		
		@Override
		public void forEach(Consumer<? super T> action) {
			Objects.requireNonNull(action);
			for(Object2LongAVLTreeMap.Entry<T> entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				action.accept(entry.key);
		}
		
		@Override
		public <E> void forEach(E input, ObjectObjectConsumer<E, T> action) {
			Objects.requireNonNull(action);
			for(Object2LongAVLTreeMap.Entry<T> entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				action.accept(input, entry.key);
		}
		
		@Override
		public boolean matchesAny(Object2BooleanFunction<T> filter) {
			Objects.requireNonNull(filter);
			for(Object2LongAVLTreeMap.Entry<T> entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				if(filter.getBoolean(entry.key)) return true;
			return false;
		}
		
		@Override
		public boolean matchesNone(Object2BooleanFunction<T> filter) {
			Objects.requireNonNull(filter);
			for(Object2LongAVLTreeMap.Entry<T> entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				if(filter.getBoolean(entry.key)) return false;
			return true;
		}
		
		@Override
		public boolean matchesAll(Object2BooleanFunction<T> filter) {
			Objects.requireNonNull(filter);
			for(Object2LongAVLTreeMap.Entry<T> entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				if(!filter.getBoolean(entry.key)) return false;
			return true;
		}
		
		@Override
		public <E> E reduce(E identity, BiFunction<E, T, E> operator) {
			Objects.requireNonNull(operator);
			E state = identity;
			for(Object2LongAVLTreeMap.Entry<T> entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				state = operator.apply(state, entry.key);
			return state;
		}
		
		@Override
		public T reduce(ObjectObjectUnaryOperator<T, T> operator) {
			Objects.requireNonNull(operator);
			T state = null;
			boolean empty = true;
			for(Object2LongAVLTreeMap.Entry<T> entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry)) {
				if(empty) {
					empty = false;
					state = entry.key;
					continue;
				}
				state = operator.apply(state, entry.key);
			}
			return state;
		}
		
		@Override
		public T findFirst(Object2BooleanFunction<T> filter) {
			Objects.requireNonNull(filter);
			for(Object2LongAVLTreeMap.Entry<T> entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				if(filter.getBoolean(entry.key)) return entry.key;
			return null;
		}
		
		@Override
		public int count(Object2BooleanFunction<T> filter) {
			Objects.requireNonNull(filter);
			int result = 0;
			for(Object2LongAVLTreeMap.Entry<T> entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				if(filter.getBoolean(entry.key)) result++;
			return result;
		}
	}
	
	final class Values extends AbstractLongCollection {

		@Override
		public boolean add(long o) { throw new UnsupportedOperationException(); }
		
		@Override
		public boolean contains(long e) {
			return containsValue(e);
		}
		
		@Override
		public LongIterator iterator() { return new ValueIterator(false); }
		
		@Override
		public int size() {
			return Object2LongAVLTreeMap.this.size();
		}
		
		@Override
		public void clear() {
			Object2LongAVLTreeMap.this.clear();
		}
		
		@Override
		public void forEach(LongConsumer action) {
			Objects.requireNonNull(action);
			for(Object2LongAVLTreeMap.Entry<T> entry = first;entry != null;entry = entry.next())
				action.accept(entry.value);
		}
		
		@Override
		public <E> void forEach(E input, ObjectLongConsumer<E> action) {
			Objects.requireNonNull(action);
			for(Object2LongAVLTreeMap.Entry<T> entry = first;entry != null;entry = entry.next())
				action.accept(input, entry.value);
		}
		
		@Override
		public boolean matchesAny(Long2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(Object2LongAVLTreeMap.Entry<T> entry = first;entry != null;entry = entry.next())
				if(filter.get(entry.value)) return true;
			return false;
		}
		
		@Override
		public boolean matchesNone(Long2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(Object2LongAVLTreeMap.Entry<T> entry = first;entry != null;entry = entry.next())
				if(filter.get(entry.value)) return false;
			return true;
		}
		
		@Override
		public boolean matchesAll(Long2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(Object2LongAVLTreeMap.Entry<T> entry = first;entry != null;entry = entry.next())
				if(!filter.get(entry.value)) return false;
			return true;
		}
		
		@Override
		public long reduce(long identity, LongLongUnaryOperator operator) {
			Objects.requireNonNull(operator);
			long state = identity;
			for(Object2LongAVLTreeMap.Entry<T> entry = first;entry != null;entry = entry.next())
				state = operator.applyAsLong(state, entry.value);
			return state;
		}
		
		@Override
		public long reduce(LongLongUnaryOperator operator) {
			Objects.requireNonNull(operator);
			long state = 0L;
			boolean empty = true;
			for(Object2LongAVLTreeMap.Entry<T> entry = first;entry != null;entry = entry.next()) {
				if(empty) {
					empty = false;
					state = entry.value;
					continue;
				}
				state = operator.applyAsLong(state, entry.value);
			}
			return state;
		}
		
		@Override
		public long findFirst(Long2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(Object2LongAVLTreeMap.Entry<T> entry = first;entry != null;entry = entry.next())
				if(filter.get(entry.value)) return entry.value;
			return 0L;
		}
		
		@Override
		public int count(Long2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			int result = 0;
			for(Object2LongAVLTreeMap.Entry<T> entry = first;entry != null;entry = entry.next())
				if(filter.get(entry.value)) result++;
			return result;
		}
	}
	
	final class EntryIterator extends MapEntryIterator implements ObjectListIterator<Object2LongMap.Entry<T>> {
		
		public EntryIterator(boolean descending) {
			super(descending);
		}
		
		public EntryIterator(Entry<T> entry) {
			super(entry);
		}
		
		@Override
		public Object2LongMap.Entry<T> next() {
			return nextEntry();
		}
		
		@Override
		public Object2LongMap.Entry<T> previous() {
			return previousEntry();
		}
		
		@Override
		public void set(Object2LongMap.Entry<T> e) { throw new UnsupportedOperationException(); }
		
		@Override
		public void add(Object2LongMap.Entry<T> e) { throw new UnsupportedOperationException(); }
	}
	
	final class KeyIterator extends MapEntryIterator implements ObjectListIterator<T> {
		
		public KeyIterator(boolean descending) {
			super(descending);
		}
		
		public KeyIterator(Entry<T> entry) {
			super(entry);
		}
		
		@Override
		public T previous() { return previousEntry().key; }
		@Override
		public T next() { return nextEntry().key; }
		
		@Override
		public void set(T e) { throw new UnsupportedOperationException(); }
		@Override
		public void add(T e) { throw new UnsupportedOperationException(); }
	}
	
	final class ValueIterator extends MapEntryIterator implements LongListIterator {
		
		public ValueIterator(boolean descending) {
			super(descending);
		}
		
		public ValueIterator(Entry<T> entry) {
			super(entry);
		}
		
		@Override
		public long previousLong() {
			return previousEntry().value;
		}
		
		@Override
		public long nextLong() {
			return nextEntry().value;
		}
		
		@Override
		public void set(long e) { throw new UnsupportedOperationException(); }
		
		@Override
		public void add(long e) { throw new UnsupportedOperationException(); }
	}
	
	abstract class MapEntryIterator {
		Entry<T> next;
		Entry<T> previous;
		Entry<T> current;
		int index = 0;
		
		public MapEntryIterator(boolean descending) {
			if(descending) previous = last;
			else next = first;
		}
		
		public MapEntryIterator(Entry<T> entry) {
			next = entry;
			previous = entry.previous();
		}
		
		public boolean hasNext() {
			return next != null;
		}
		
		public boolean hasPrevious() {
			return previous != null;
		}
		
		public int nextIndex() {
			return index;
		}
		
		public int previousIndex() {
			return index - 1;
		}
		
		protected void updateNext() {
			next = current.next();
		}
		
		protected void updatePrevious() {
			previous = current.previous();
		}
		
		public Entry<T> nextEntry() {
			if(!hasNext()) throw new NoSuchElementException();
			current = previous = next;
			updateNext();
			index++;
			return current;
		}
		
		public Entry<T> previousEntry() {
			if(!hasPrevious()) throw new NoSuchElementException();
			current = next = previous;
			updatePrevious();
			index--;
			return current;
		}
		
		public void remove() {
			if(current == null) throw new IllegalStateException();
			if(current == previous) index--;
			updateNext();
			updatePrevious();
			if(current.needsSuccessor()) next = current;
			removeNode(current);
			current = null;
		}
	}
	
	private static final class Entry<T> implements Object2LongMap.Entry<T>
	{
		T key;
		long value;
		int state;
		Entry<T> parent;
		Entry<T> left;
		Entry<T> right;
		
		Entry(T key, long value, Entry<T> parent) {
			this.key = key;
			this.value = value;
			this.parent = parent;
		}
		
		Entry<T> copy() {
			Entry<T> entry = new Entry<>(key, value, null);
			entry.state = state;
			if(left != null) {
				Entry<T> newLeft = left.copy();
				entry.left = newLeft;
				newLeft.parent = entry;
			}
			if(right != null) {
				Entry<T> newRight = right.copy();
				entry.right = newRight;
				newRight.parent = entry;
			}
			return entry;
		}
		
		@Override
		public T getKey() {
			return key;
		}
		
		@Override
		public long getLongValue() {
			return value;
		}
		
		@Override
		public long setValue(long value) {
			long oldValue = this.value;
			this.value = value;
			return oldValue;
		}
		
		long addTo(long value) {
			long oldValue = this.value;
			this.value += value;
			return oldValue;
		}
		
		@Override
		public boolean equals(Object obj) {
			if(obj instanceof Map.Entry) {
				if(obj instanceof Object2LongMap.Entry) {
					Object2LongMap.Entry<T> entry = (Object2LongMap.Entry<T>)obj;
					if(entry.getKey() == null) return false;
					return Objects.equals(key, entry.getKey()) && value == entry.getLongValue();
				}
				Map.Entry<?, ?> entry = (Map.Entry<?, ?>)obj;
				Object otherKey = entry.getKey();
				if(otherKey == null) return false;
				Object otherValue = entry.getValue();
				return otherValue instanceof Long && Objects.equals(key, otherKey) && value == ((Long)otherValue).longValue();
			}
			return false;
		}
		
		@Override
		public int hashCode() {
			return Objects.hashCode(key) ^ Long.hashCode(value);
		}
		
		@Override
		public String toString() {
			return Objects.toString(key) + "=" + Long.toString(value);
		}
		
		int getHeight() { return state; }
		
		void updateHeight() { state = (1 + Math.max(left == null ? -1 : left.getHeight(), right == null ? -1 : right.getHeight())); }
		
		int getBalance() { return (left == null ? -1 : left.getHeight()) - (right == null ? -1 : right.getHeight()); }
		
		boolean needsSuccessor() { return left != null && right != null; }
		
		boolean replace(Entry<T> entry) {
			if(entry != null) entry.parent = parent;
			if(parent != null) {
				if(parent.left == this) parent.left = entry;
				else parent.right = entry;
			}
			return parent == null;
		}
		
		Entry<T> next() {
			if(right != null) {
				Entry<T> parent = right;
				while(parent.left != null) parent = parent.left;
				return parent;
			}
			Entry<T> parent = this.parent;
			Entry<T> control = this;
			while(parent != null && control == parent.right) {
				control = parent;
				parent = parent.parent;
			}
			return parent;
		}
		
		Entry<T> previous() {
			if(left != null) {
				Entry<T> parent = left;
				while(parent.right != null) parent = parent.right;
				return parent;
			}
			Entry<T> parent = this.parent;
			Entry<T> control = this;
			while(parent != null && control == parent.left) {
				control = parent;
				parent = parent.parent;
			}
			return parent;
		}
	}
}