package speiger.src.collections.objects.maps.impl.tree;

import java.util.Map;
import java.util.Comparator;
import java.util.function.Consumer;
import java.util.function.BiFunction;
import java.util.Objects;
import java.util.NoSuchElementException;

import speiger.src.collections.objects.collections.ObjectBidirectionalIterator;
import speiger.src.collections.objects.functions.consumer.ObjectObjectConsumer;
import speiger.src.collections.objects.functions.function.Object2ObjectFunction;
import speiger.src.collections.objects.functions.function.ObjectObjectUnaryOperator;
import speiger.src.collections.objects.lists.ObjectListIterator;
import speiger.src.collections.objects.maps.abstracts.AbstractObject2ObjectMap;
import speiger.src.collections.objects.maps.interfaces.Object2ObjectMap;
import speiger.src.collections.objects.maps.interfaces.Object2ObjectNavigableMap;
import speiger.src.collections.objects.sets.AbstractObjectSet;
import speiger.src.collections.objects.sets.ObjectNavigableSet;
import speiger.src.collections.objects.sets.ObjectSet;
import speiger.src.collections.objects.sets.ObjectSortedSet;
import speiger.src.collections.objects.utils.ObjectIterators;
import speiger.src.collections.objects.utils.maps.Object2ObjectMaps;
import speiger.src.collections.objects.collections.AbstractObjectCollection;
import speiger.src.collections.objects.collections.ObjectCollection;
import speiger.src.collections.objects.collections.ObjectIterator;
import speiger.src.collections.objects.functions.ObjectSupplier;
import speiger.src.collections.objects.functions.function.Object2BooleanFunction;

/**
 * A Simple Type Specific AVL TreeMap implementation that reduces boxing/unboxing.
 * It is using a bit more memory then <a href="https://github.com/vigna/fastutil">FastUtil</a>,
 * but it saves a lot of Performance on the Optimized removal and iteration logic.
 * Which makes the implementation actually useable and does not get outperformed by Javas default implementation.
 * @param <T> the type of elements maintained by this Collection
 * @param <V> the type of elements maintained by this Collection
 */
public class Object2ObjectAVLTreeMap<T, V> extends AbstractObject2ObjectMap<T, V> implements Object2ObjectNavigableMap<T, V>
{
	/** The center of the Tree */
	protected transient Entry<T, V> tree;
	/** The Lowest possible Node */
	protected transient Entry<T, V> first;
	/** The Highest possible Node */
	protected transient Entry<T, V> last;
	/** The amount of elements stored in the Map */
	protected int size = 0;
	/** The Sorter of the Tree */
	protected transient Comparator<T> comparator;
	
	
	/** KeySet Cache */
	protected ObjectNavigableSet<T> keySet;
	/** Values Cache */
	protected ObjectCollection<V> values;
	/** EntrySet Cache */
	protected ObjectSet<Object2ObjectMap.Entry<T, V>> entrySet;
	
	/**
	 * Default Constructor
	 */
	public Object2ObjectAVLTreeMap() {
	}
	
	/**
	 * Constructor that allows to define the sorter
	 * @param comp the function that decides how the tree is sorted, can be null
	 */
	public Object2ObjectAVLTreeMap(Comparator<T> comp) {
		comparator = comp;
	}
	
	/**
	 * Helper constructor that allow to create a map from unboxed values
	 * @param keys the keys that should be put into the map
	 * @param values the values that should be put into the map.
	 * @throws IllegalStateException if the keys and values do not match in lenght
	 */
	public Object2ObjectAVLTreeMap(T[] keys, V[] values) {
		this(keys, values, null);
	}
	
	/**
	 * Helper constructor that has a custom sorter and allow to create a map from unboxed values
	 * @param keys the keys that should be put into the map
	 * @param values the values that should be put into the map.
	 * @param comp the function that decides how the tree is sorted, can be null
	 * @throws IllegalStateException if the keys and values do not match in lenght
	 */
	public Object2ObjectAVLTreeMap(T[] keys, V[] values, Comparator<T> comp) {
		comparator = comp;
		if(keys.length != values.length) throw new IllegalStateException("Input Arrays are not equal size");
		for(int i = 0,m=keys.length;i<m;i++) put(keys[i], values[i]);
	}
	
	/**
	 * A Helper constructor that allows to create a Map with exactly the same values as the provided map.
	 * @param map the values that should be present in the map
	 */
	public Object2ObjectAVLTreeMap(Map<? extends T, ? extends V> map) {
		this(map, null);
	}
	
	/**
	 * A Helper constructor that has a custom sorter and allows to create a Map with exactly the same values as the provided map.
	 * @param map the values that should be present in the map
	 * @param comp the function that decides how the tree is sorted, can be null
	 */
	public Object2ObjectAVLTreeMap(Map<? extends T, ? extends V> map, Comparator<T> comp) {
		comparator = comp;
		putAll(map);
	}
	
	/**
	 * A Type Specific Helper function that allows to create a new Map with exactly the same values as the provided map.
	 * @param map the values that should be present in the map
 	 */
	public Object2ObjectAVLTreeMap(Object2ObjectMap<T, V> map) {
		this(map, null);
	}
	
	/**
	 * A Type Specific Helper function that has a custom sorter and allows to create a new Map with exactly the same values as the provided map.
	 * @param map the values that should be present in the map
	 * @param comp the function that decides how the tree is sorted, can be null
 	 */
	public Object2ObjectAVLTreeMap(Object2ObjectMap<T, V> map, Comparator<T> comp) {
		comparator = comp;
		putAll(map);
	}
	
	/** only used for primitives 
	 * @return null 
	 */
	public T getDefaultMaxValue() { return null; }
	/** only used for primitives 
	 * @return null 
	 */
	public T getDefaultMinValue() { return null; }
	
	@Override
	public V put(T key, V value) {
		validate(key);
		if(tree == null) {
			tree = first = last = new Entry<>(key, value, null);
			size++;
			return getDefaultReturnValue();
		}
		int compare = 0;
		Entry<T, V> parent = tree;
		while(true) {
			if((compare = compare(key, parent.key)) == 0) return parent.setValue(value);
			if(compare < 0) {
				if(parent.left == null) break;
				parent = parent.left;
			}
			else if(compare > 0) {
				if(parent.right == null) break;
				parent = parent.right;
			}
		}
		Entry<T, V> adding = new Entry<>(key, value, parent);
		if(compare < 0)  {
			parent.left = adding;
			if(parent == first)	first = adding;
		}
		else {
			parent.right = adding;
			if(parent == last) last = adding;
		}
		fixAfterInsertion(adding);
		size++;
		return getDefaultReturnValue();
	}
	
	@Override
	public V putIfAbsent(T key, V value) {
		validate(key);
		if(tree == null) {
			tree = first = last = new Entry<>(key, value, null);
			size++;
			return getDefaultReturnValue();
		}
		int compare = 0;
		Entry<T, V> parent = tree;
		while(true) {
			if((compare = compare(key, parent.key)) == 0) return parent.value;
			if(compare < 0) {
				if(parent.left == null) break;
				parent = parent.left;
			}
			else if(compare > 0) {
				if(parent.right == null) break;
				parent = parent.right;
			}
		}
		Entry<T, V> adding = new Entry<>(key, value, parent);
		if(compare < 0)  {
			parent.left = adding;
			if(parent == first)	first = adding;
		}
		else  {
			parent.right = adding;
			if(parent == last) last = adding;
		}
		fixAfterInsertion(adding);
		size++;
		return getDefaultReturnValue();
	}
	
	@Override
	public V putAndMoveToFirst(T key, V value) { throw new UnsupportedOperationException(); }
	@Override
	public V putAndMoveToLast(T key, V value) { throw new UnsupportedOperationException(); }
	@Override
	public boolean moveToFirst(T key) { throw new UnsupportedOperationException(); }
	@Override
	public boolean moveToLast(T key) { throw new UnsupportedOperationException(); }
	@Override
	public V getAndMoveToFirst(T key) { throw new UnsupportedOperationException(); }
	@Override
	public V getAndMoveToLast(T key) { throw new UnsupportedOperationException(); }
	
	@Override
	public Comparator<T> comparator() { return comparator; }
	
	@Override
	public boolean containsKey(Object key) {
		return findNode((T)key) != null;
	}
	

	@Override
	public V getObject(T key) {
		Entry<T, V> node = findNode(key);
		return node == null ? getDefaultReturnValue() : node.value;
	}
	
	@Override
	public V getOrDefault(Object key, V defaultValue) {
		Entry<T, V> node = findNode((T)key);
		return node == null ? defaultValue : node.value;
	}
	
	@Override
	public T firstKey() {
		if(tree == null) throw new NoSuchElementException();
		return first.key;
	}
	
	@Override
	public T pollFirstKey() {
		if(tree == null) throw new NoSuchElementException();
		T result = first.key;
		removeNode(first);
		return result;
	}
	
	@Override
	public T lastKey() {
		if(tree == null) throw new NoSuchElementException();
		return last.key;
	}
	
	@Override
	public T pollLastKey() {
		if(tree == null) throw new NoSuchElementException();
		T result = last.key;
		removeNode(last);
		return result;
	}
	
	@Override
	public Object2ObjectMap.Entry<T, V> firstEntry() {
		if(tree == null) throw new NoSuchElementException();
		return new BasicEntry<>(first.key, first.value);
	}
	
	@Override
	public Object2ObjectMap.Entry<T, V> lastEntry() {
		if(tree == null) throw new NoSuchElementException();
		return new BasicEntry<>(last.key, last.value);
	}
	
	@Override
	public Object2ObjectMap.Entry<T, V> pollFirstEntry() {
		if(tree == null) throw new NoSuchElementException();
		BasicEntry<T, V> entry = new BasicEntry<>(first.key, first.value);
		removeNode(first);
		return entry;
	}
	
	@Override
	public Object2ObjectMap.Entry<T, V> pollLastEntry() {
		if(tree == null) throw new NoSuchElementException();
		BasicEntry<T, V> entry = new BasicEntry<>(last.key, last.value);
		removeNode(last);
		return entry;
	}
	
	@Override
	public V firstValue() {
		if(tree == null) throw new NoSuchElementException();
		return first.value;
	}
	
	@Override
	public V lastValue() {
		if(tree == null) throw new NoSuchElementException();
		return last.value;
	}
	
	@Override
	public V rem(T key) {
		Entry<T, V> entry = findNode(key);
		if(entry == null) return getDefaultReturnValue();
		V value = entry.value;
		removeNode(entry);
		return value;
	}
	
	@Override
	public V remOrDefault(T key, V defaultValue) {
		Entry<T, V> entry = findNode(key);
		if(entry == null) return defaultValue;
		V value = entry.value;
		removeNode(entry);
		return value;
	}
	
	@Override
	public boolean remove(Object key, Object value) {
		Entry<T, V> entry = findNode((T)key);
		if(entry == null || !Objects.equals(value, entry.value)) return false;
		removeNode(entry);
		return true;
	}
	
	@Override
	public boolean replace(T key, V oldValue, V newValue) {
		Entry<T, V> entry = findNode(key);
		if(entry == null || entry.value != oldValue) return false;
		entry.value = newValue;
		return true;
	}
	
	@Override
	public V replace(T key, V value) {
		Entry<T, V> entry = findNode(key);
		if(entry == null) return getDefaultReturnValue();
		V oldValue = entry.value;
		entry.value = value;
		return oldValue;
	}
	
	@Override
	public V compute(T key, ObjectObjectUnaryOperator<T, V> mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		validate(key);
		Entry<T, V> entry = findNode(key);
		if(entry == null) {
			V newValue = mappingFunction.apply(key, getDefaultReturnValue());
			if(Objects.equals(newValue, getDefaultReturnValue())) return newValue;
			put(key, newValue);
			return newValue;
		}
		V newValue = mappingFunction.apply(key, entry.value);
		if(Objects.equals(newValue, getDefaultReturnValue())) {
			removeNode(entry);
			return newValue;
		}
		entry.value = newValue;
		return newValue;
	}
	
	@Override
	public V computeIfAbsent(T key, Object2ObjectFunction<T, V> mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		validate(key);
		Entry<T, V> entry = findNode(key);
		if(entry == null) {
			V newValue = mappingFunction.getObject(key);
			if(Objects.equals(newValue, getDefaultReturnValue())) return newValue;
			put(key, newValue);
			return newValue;
		}
		if(Objects.equals(entry.value, getDefaultReturnValue())) {
			V newValue = mappingFunction.getObject(key);
			if(Objects.equals(newValue, getDefaultReturnValue())) return newValue;
			entry.value = newValue;
		}
		return entry.value;
	}
	
	@Override
	public V supplyIfAbsent(T key, ObjectSupplier<V> valueProvider) {
		Objects.requireNonNull(valueProvider);
		validate(key);
		Entry<T, V> entry = findNode(key);
		if(entry == null) {
			V newValue = valueProvider.get();
			if(Objects.equals(newValue, getDefaultReturnValue())) return newValue;
			put(key, newValue);
			return newValue;
		}
		if(Objects.equals(entry.value, getDefaultReturnValue())) {
			V newValue = valueProvider.get();
			if(Objects.equals(newValue, getDefaultReturnValue())) return newValue;
			entry.value = newValue;
		}
		return entry.value;
	}
	
	@Override
	public V computeIfPresent(T key, ObjectObjectUnaryOperator<T, V> mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		validate(key);
		Entry<T, V> entry = findNode(key);
		if(entry == null || Objects.equals(entry.value, getDefaultReturnValue())) return getDefaultReturnValue();
		V newValue = mappingFunction.apply(key, entry.value);
		if(Objects.equals(newValue, getDefaultReturnValue())) {
			removeNode(entry);
			return newValue;
		}
		entry.value = newValue;
		return newValue;
	}
	
	@Override
	public V merge(T key, V value, ObjectObjectUnaryOperator<V, V> mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		validate(key);
		Entry<T, V> entry = findNode(key);
		V newValue = entry == null || Objects.equals(entry.value, getDefaultReturnValue()) ? value : mappingFunction.apply(entry.value, value);
		if(Objects.equals(newValue, getDefaultReturnValue())) {
			if(entry != null)
				removeNode(entry);
		}
		else if(entry == null) put(key, newValue);
		else entry.value = newValue;
		return newValue;
	}
	
	@Override
	public void mergeAll(Object2ObjectMap<T, V> m, ObjectObjectUnaryOperator<V, V> mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		for(Object2ObjectMap.Entry<T, V> entry : Object2ObjectMaps.fastIterable(m)) {
			T key = entry.getKey();
			Entry<T, V> subEntry = findNode(key);
			V newValue = subEntry == null || Objects.equals(subEntry.value, getDefaultReturnValue()) ? entry.getValue() : mappingFunction.apply(subEntry.value, entry.getValue());
			if(Objects.equals(newValue, getDefaultReturnValue())) {
				if(subEntry != null)
					removeNode(subEntry);
			}
			else if(subEntry == null) put(key, newValue);
			else subEntry.value = newValue;
		}
	}
	
	@Override
	public void forEach(ObjectObjectConsumer<T, V> action) {
		for(Entry<T, V> entry = first;entry != null;entry = entry.next())
			action.accept(entry.key, entry.value);
	}
	
	@Override
	public int size() { return size; }
	
	@Override
	public void clear() {
		size = 0;
		first = null;
		last = null;
		tree = null;
	}
	
	ObjectListIterator<T> keyIterator(boolean descending) {
		ObjectListIterator<T> iter = new KeyIterator(descending);
		return descending ? ObjectIterators.invert(iter) : iter;
	}
	
	ObjectListIterator<T> keyIterator(T key) {
		return new KeyIterator(findNode(key));
	}
	
	public Object2ObjectAVLTreeMap<T, V> copy() {
		Object2ObjectAVLTreeMap<T, V> set = new Object2ObjectAVLTreeMap<>();
		set.size = size;
		if(tree != null) {
			set.tree = tree.copy();
			Entry<T, V> lastFound = null;
			for(Entry<T, V> entry = tree;entry != null;entry = entry.left) lastFound = entry;
			set.first = lastFound;
			lastFound = null;
			for(Entry<T, V> entry = tree;entry != null;entry = entry.right) lastFound = entry;
			set.last = lastFound;
		}
		return set;
	}
	
	@Override
	public ObjectSortedSet<T> keySet() {
		return navigableKeySet();
	}
	
	@Override
	public ObjectSet<Object2ObjectMap.Entry<T, V>> object2ObjectEntrySet() {
		if(entrySet == null) entrySet = new EntrySet();
		return entrySet;
	}
	
	@Override
	public ObjectCollection<V> values() {
		if(values == null) values = new Values();
		return values;
	}
	
	@Override
	public ObjectNavigableSet<T> navigableKeySet() {
		if(keySet == null) keySet = new KeySet<>(this);
		return keySet;
	}
	
	@Override
	public Object2ObjectNavigableMap<T, V> descendingMap() {
		return new DescendingSubMap<>(this, true, null, true, true, null, true);
	}
	
	@Override
	public ObjectNavigableSet<T> descendingKeySet() {
		return descendingMap().navigableKeySet();
	}
	
	@Override
	public Object2ObjectNavigableMap<T, V> subMap(T fromKey, boolean fromInclusive, T toKey, boolean toInclusive) {
		return new AscendingSubMap<>(this, false, fromKey, fromInclusive, false, toKey, toInclusive);
	}
	
	@Override
	public Object2ObjectNavigableMap<T, V> headMap(T toKey, boolean inclusive) {
		return new AscendingSubMap<>(this, true, null, true, false, toKey, inclusive);
	}
	
	@Override
	public Object2ObjectNavigableMap<T, V> tailMap(T fromKey, boolean inclusive) {
		return new AscendingSubMap<>(this, false, fromKey, inclusive, true, null, true);
	}
	
	@Override
	public T lowerKey(T e) {
		Entry<T, V> node = findLowerNode(e);
		return node != null ? node.key : getDefaultMinValue();
	}

	@Override
	public T floorKey(T e) {
		Entry<T, V> node = findFloorNode(e);
		return node != null ? node.key : getDefaultMinValue();
	}
	
	@Override
	public T higherKey(T e) {
		Entry<T, V> node = findHigherNode(e);
		return node != null ? node.key : getDefaultMaxValue();
	}

	@Override
	public T ceilingKey(T e) {
		Entry<T, V> node = findCeilingNode(e);
		return node != null ? node.key : getDefaultMaxValue();
	}
	
	@Override
	public Object2ObjectMap.Entry<T, V> lowerEntry(T key) {
		Entry<T, V> node = findLowerNode(key);
		return node != null ? new BasicEntry<>(node.key, node.value) : null;
	}
	
	@Override
	public Object2ObjectMap.Entry<T, V> higherEntry(T key) {
		Entry<T, V> node = findHigherNode(key);
		return node != null ? new BasicEntry<>(node.key, node.value) : null;
	}
	
	@Override
	public Object2ObjectMap.Entry<T, V> floorEntry(T key) {
		Entry<T, V> node = findFloorNode(key);
		return node != null ? new BasicEntry<>(node.key, node.value) : null;
	}
	
	@Override
	public Object2ObjectMap.Entry<T, V> ceilingEntry(T key) {
		Entry<T, V> node = findCeilingNode(key);
		return node != null ? new BasicEntry<>(node.key, node.value) : null;
	}
	
	protected Entry<T, V> findLowerNode(T key) {
		Entry<T, V> entry = tree;
		while(entry != null) {
			if(compare(key, entry.key) > 0) {
				if(entry.right != null) entry = entry.right;
				else return entry;
			}
			else {
				if(entry.left != null) entry = entry.left;
				else {
					Entry<T, V> parent = entry.parent;
					while(parent != null && parent.left == entry) {
						entry = parent;
						parent = parent.parent;
					}
					return parent;
				}
			}
		}
		return null;
	}
	
	protected Entry<T, V> findFloorNode(T key) {
		Entry<T, V> entry = tree;
		int compare;
		while(entry != null) {
			if((compare = compare(key, entry.key)) > 0) {
				if(entry.right == null) break;
				entry = entry.right;
				continue;
			}
			else if(compare < 0) {
				if(entry.left != null) entry = entry.left;
				else {
					Entry<T, V> parent = entry.parent;
					while(parent != null && parent.left == entry) {
						entry = parent;
						parent = parent.parent;
					}
					return parent;
				}
				continue;
			}
			break;
		}
		return entry;
	}
	
	protected Entry<T, V> findCeilingNode(T key) {
		Entry<T, V> entry = tree;
		int compare;
		while(entry != null) {
			if((compare = compare(key, entry.key)) < 0) {
				if(entry.left == null) break;
				entry = entry.left;
				continue;
			}
			else if(compare > 0) {
				if(entry.right != null) entry = entry.right;
				else {
					Entry<T, V> parent = entry.parent;
					while(parent != null && parent.right == entry) {
						entry = parent;
						parent = parent.parent;
					}
					return parent;
				}
				continue;
			}
			break;
		}
		return entry;
	}
	
	protected Entry<T, V> findHigherNode(T key) {
		Entry<T, V> entry = tree;
		while(entry != null) {
			if(compare(key, entry.key) < 0) {
				if(entry.left != null) entry = entry.left;
				else return entry;
			}
			else {
				if(entry.right != null) entry = entry.right;
				else {
					Entry<T, V> parent = entry.parent;
					while(parent != null && parent.right == entry) {
						entry = parent;
						parent = parent.parent;
					}
					return parent;
				}
			}
		}
		return null;
	}
	
	protected Entry<T, V> findNode(T key) {
		Entry<T, V> node = tree;
		int compare;
		while(node != null) {
			if((compare = compare(key, node.key)) == 0) return node;
			if(compare < 0) node = node.left;
			else node = node.right;
		}
		return null;
	}
	
	protected void removeNode(Entry<T, V> entry) {
		size--;
		if(entry.needsSuccessor()) {
			Entry<T, V> successor = entry.next();
			entry.key = successor.key;
			entry.value = successor.value;
			entry = successor;
		}
		if(entry.previous() == null) first = entry.next();
		if(entry.next() == null) last = entry.previous();
		Entry<T, V> replacement = entry.left != null ? entry.left : entry.right;
		if(replacement != null) {
			if(entry.replace(replacement)) tree = replacement;
			entry.left = entry.right = entry.parent = null;
			fixAfterDeletion(replacement);
		}
		else if(entry.parent == null) tree = first = last = null;
		else {
			fixAfterDeletion(entry);
			entry.replace(null);
			entry.parent = null;
		}
	}
	
	protected void validate(T k) { compare(k, k); }
	protected int compare(T k, T v) { return comparator != null ? comparator.compare(k, v) : ((Comparable<T>)k).compareTo((T)v);}
	
	/** From CLR */
	protected void rotateLeft(Entry<T, V> entry) {
		if(entry != null) {
			Entry<T, V> right = entry.right;
			entry.right = right.left;
			if(right.left != null) right.left.parent = entry;
			right.parent = entry.parent;
			if(entry.parent == null) tree = right;
			else if(entry.parent.left == entry) entry.parent.left = right;
			else entry.parent.right = right;
			right.left = entry;
			entry.parent = right;
			entry.updateHeight();
			right.updateHeight();
		}
	}
	
	/** From CLR */
	protected void rotateRight(Entry<T, V> entry) {
		if(entry != null) {
			Entry<T, V> left = entry.left;
			entry.left = left.right;
			if(left.right != null) left.right.parent = entry;
			left.parent = entry.parent;
			if(entry.parent == null) tree = left;
			else if(entry.parent.right == entry) entry.parent.right = left;
			else entry.parent.left = left;
			left.right = entry;
			entry.parent = left;
			entry.updateHeight();
			left.updateHeight();
		}
	}
	
	/** From CLR */
	protected void fixAfterInsertion(Entry<T, V> entry) {
		while(entry != null) {
			entry.updateHeight();
			int balance = entry.getBalance();
			if(balance > 1) {
				int compare = entry.left.getBalance();
				if(compare > 0) rotateRight(entry);
				else if(compare < 0) {
					rotateLeft(entry.left);
					rotateRight(entry);
				}
			}
			else if(balance < -1) {
				int compare = entry.right.getBalance();
				if(compare < 0) rotateLeft(entry);
				else if(compare > 0) {
					rotateRight(entry.right);
					rotateLeft(entry);
				}
			}
			entry = entry.parent;
		}
	}
	
	/** From CLR */
	protected void fixAfterDeletion(Entry<T, V> entry) {
		if(entry != null) {
			entry.updateHeight();
			int balance = entry.getBalance();
			if(balance > 1) {
				int subBalance = entry.left.getBalance();
				if(subBalance >= 0) rotateRight(entry);
				else {
					rotateLeft(entry.left);
					rotateRight(entry);
				}
			}
			else if(balance < -1)
			{
				int subBalance = entry.right.getBalance();
				if(subBalance <= 0) rotateLeft(entry);
				else {
					rotateRight(entry.right);
					rotateLeft(entry);
				}
			}
			entry = entry.parent;
		}
	}
	
	static final class AscendingSubMap<T, V> extends NavigableSubMap<T, V> {

		public AscendingSubMap(Object2ObjectAVLTreeMap<T, V> m, boolean fromStart, T low, boolean loInclusive, boolean toEnd, T high, boolean hiInclusive) {
			super(m, fromStart, low, loInclusive, toEnd, high, hiInclusive);
		}
		
		@Override
		ObjectListIterator<T> keyIterator(boolean descending) { return new SubMapKeyIterator(descending); }
		@Override
		ObjectListIterator<T> keyIterator(T key) {
			Object2ObjectAVLTreeMap.Entry<T, V> entry = m.findNode(key);
			return entry == null || !inClosedRange(key) ? null : new SubMapKeyIterator(entry, fromStart ? null : findLowest(), toEnd ? null : findHighest());
		}
		@Override
		ObjectListIterator<V> valueIterator() { return new SubMapValueIterator(false); }
		@Override
		ObjectListIterator<Object2ObjectMap.Entry<T, V>> entryIterator() { return new SubMapEntrySetIterator(false); }
		
		@Override
		public Object2ObjectNavigableMap<T, V> descendingMap() {
			if(descendingMap == null) descendingMap = new DescendingSubMap<>(m, fromStart, low, loInclusive, toEnd, high, hiInclusive);
			return descendingMap;
		}
		
		Object2ObjectAVLTreeMap.Entry<T, V> next(Object2ObjectAVLTreeMap.Entry<T, V> entry) {
			return entry.next();
		}
		
		Object2ObjectAVLTreeMap.Entry<T, V> previous(Object2ObjectAVLTreeMap.Entry<T, V> entry) {
			return entry.previous();
		}
		
		@Override
		public Object2ObjectNavigableMap<T, V> subMap(T fromKey, boolean fromInclusive, T toKey, boolean toInclusive) {
			if (!inRange(fromKey, fromInclusive)) throw new IllegalArgumentException("fromKey out of range");
			if (!inRange(toKey, toInclusive)) throw new IllegalArgumentException("toKey out of range");
			return new AscendingSubMap<>(m, false, fromKey, fromInclusive, false, toKey, toInclusive);
		}
		
		@Override
		public Object2ObjectNavigableMap<T, V> headMap(T toKey, boolean inclusive) {
			if (!inRange(toKey, inclusive)) throw new IllegalArgumentException("toKey out of range");
			return new AscendingSubMap<>(m, fromStart, low, loInclusive, false, toKey, inclusive);
		}
		
		@Override
		public Object2ObjectNavigableMap<T, V> tailMap(T fromKey, boolean inclusive) {
			if (!inRange(fromKey, inclusive)) throw new IllegalArgumentException("fromKey out of range");
			return new AscendingSubMap<>(m, false, fromKey, inclusive, toEnd, high, hiInclusive);
		}
	}
	
	static final class DescendingSubMap<T, V> extends NavigableSubMap<T, V> {

		public DescendingSubMap(Object2ObjectAVLTreeMap<T, V> m, boolean fromStart, T low, boolean loInclusive, boolean toEnd, T high, boolean hiInclusive) {
			super(m, fromStart, low, loInclusive, toEnd, high, hiInclusive);
		}
		
		@Override
		public T firstKey() { return super.lastKey(); }
		
		@Override
		public T pollFirstKey() { return super.pollLastKey(); }
		
		@Override
		public T lastKey() { return super.firstKey(); }
		
		@Override
		public T pollLastKey() { return super.pollFirstKey(); }
		
		@Override
		public V firstValue() { return super.lastValue(); }
		
		@Override
		public V lastValue() { return super.firstValue(); }
		
		@Override
		public Object2ObjectMap.Entry<T, V> firstEntry() { return super.lastEntry(); }
		
		@Override
		public Object2ObjectMap.Entry<T, V> lastEntry() { return super.firstEntry(); }
		
		@Override
		public Object2ObjectMap.Entry<T, V> pollFirstEntry() { return super.pollLastEntry(); }
		
		@Override
		public Object2ObjectMap.Entry<T, V> pollLastEntry() { return super.pollFirstEntry(); }
		
		@Override
		public T lowerKey(T e) { return super.higherKey(e); }
		
		@Override
		public T floorKey(T e) { return super.ceilingKey(e); }
		
		@Override
		public T ceilingKey(T e) { return super.floorKey(e); }
		
		@Override
		public T higherKey(T e) { return super.lowerKey(e); }
		
		@Override
		public Object2ObjectMap.Entry<T, V> lowerEntry(T e) { return super.higherEntry(e); }
		
		@Override
		public Object2ObjectMap.Entry<T, V> floorEntry(T e) { return super.ceilingEntry(e); }
		
		@Override
		public Object2ObjectMap.Entry<T, V> ceilingEntry(T e) { return super.floorEntry(e); }
		
		@Override
		public Object2ObjectMap.Entry<T, V> higherEntry(T e) { return super.lowerEntry(e); }
		
		protected Object2ObjectAVLTreeMap.Entry<T, V> findLowest() {
			if(fromStart) return m.first;
			Object2ObjectAVLTreeMap.Entry<T, V> entry = loInclusive ? m.findFloorNode(high) : m.findLowerNode(high);
			return entry == null || tooHigh(entry.key) ? null : entry;
		}
		
		protected Object2ObjectAVLTreeMap.Entry<T, V> findHighest() {
			if(toEnd) return m.last;
			Object2ObjectAVLTreeMap.Entry<T, V> entry = hiInclusive ? m.findCeilingNode(low) : m.findHigherNode(low);
			return entry == null || tooLow(entry.key) ? null : entry;
		}
		
		@Override
		ObjectListIterator<T> keyIterator(boolean descending) {
			ObjectListIterator<T> iter = new SubMapKeyIterator(!descending);
			return descending ? iter : ObjectIterators.invert(iter);
		}

		@Override
		ObjectListIterator<T> keyIterator(T key) {
			Object2ObjectAVLTreeMap.Entry<T, V> entry = m.findNode(key);
			return entry == null || !inClosedRange(key) ? null : ObjectIterators.invert(new SubMapKeyIterator(entry, fromStart ? null : findLowest(), toEnd ? null : findHighest()));
		}
		
		Object2ObjectAVLTreeMap.Entry<T, V> next(Object2ObjectAVLTreeMap.Entry<T, V> entry) {
			return entry.previous();
		}
		
		Object2ObjectAVLTreeMap.Entry<T, V> previous(Object2ObjectAVLTreeMap.Entry<T, V> entry) {
			return entry.next();
		}
		
		@Override
		public Comparator<T> comparator() { return m.comparator() == null ? null : m.comparator().reversed(); }
		
		@Override
		ObjectListIterator<V> valueIterator() { return ObjectIterators.invert(new SubMapValueIterator(true)); }
		
		@Override
		ObjectListIterator<Object2ObjectMap.Entry<T, V>> entryIterator() { return ObjectIterators.invert(new SubMapEntrySetIterator(true)); }
		
		@Override
		public Object2ObjectNavigableMap<T, V> descendingMap() {
			if(descendingMap == null) descendingMap = new AscendingSubMap<>(m, fromStart, low, loInclusive, toEnd, high, hiInclusive);
			return descendingMap;
		}

		@Override
		public Object2ObjectNavigableMap<T, V> subMap(T fromKey, boolean fromInclusive, T toKey, boolean toInclusive) {
			if (!inRange(fromKey, fromInclusive)) throw new IllegalArgumentException("fromKey out of range");
			if (!inRange(toKey, toInclusive)) throw new IllegalArgumentException("toKey out of range");
			return new DescendingSubMap<>(m, false, fromKey, fromInclusive, false, toKey, toInclusive);
		}
		
		@Override
		public Object2ObjectNavigableMap<T, V> headMap(T toKey, boolean inclusive) {
			if (!inRange(toKey, inclusive)) throw new IllegalArgumentException("toKey out of range");
			return new DescendingSubMap<>(m, fromStart, low, loInclusive, false, toKey, inclusive);
		}
		
		@Override
		public Object2ObjectNavigableMap<T, V> tailMap(T fromKey, boolean inclusive) {
			if (!inRange(fromKey, inclusive)) throw new IllegalArgumentException("fromKey out of range");
			return new DescendingSubMap<>(m, false, fromKey, inclusive, toEnd, high, hiInclusive);
		}
	}
	
	static abstract class NavigableSubMap<T, V> extends AbstractObject2ObjectMap<T, V> implements Object2ObjectNavigableMap<T, V> {
		final Object2ObjectAVLTreeMap<T, V> m;
		final T low;
		final T high;
		final boolean fromStart;
		final boolean toEnd;
		final boolean loInclusive;
		final boolean hiInclusive;
		transient Object2ObjectNavigableMap<T, V> descendingMap;
		transient SubMapEntrySet entrySet;
		transient KeySet<T, V> keySet;
		transient SubMapValues values;
		
		NavigableSubMap(Object2ObjectAVLTreeMap<T, V> m, boolean fromStart, T low, boolean loInclusive, boolean toEnd, T high, boolean hiInclusive) {
			this.m = m;
			this.low = low;
			this.high = high;
			this.fromStart = fromStart;
			this.toEnd = toEnd;
			this.loInclusive = loInclusive;
			this.hiInclusive = hiInclusive;
		}
		
		abstract ObjectListIterator<T> keyIterator(boolean descending);
		abstract ObjectListIterator<T> keyIterator(T key);
		abstract ObjectListIterator<V> valueIterator();
		abstract ObjectListIterator<Object2ObjectMap.Entry<T, V>> entryIterator();
		abstract Object2ObjectAVLTreeMap.Entry<T, V> next(Object2ObjectAVLTreeMap.Entry<T, V> entry);
		abstract Object2ObjectAVLTreeMap.Entry<T, V> previous(Object2ObjectAVLTreeMap.Entry<T, V> entry);
		
		@Override
		public NavigableSubMap<T, V> copy() { throw new UnsupportedOperationException(); }
		@Override
		public abstract Object2ObjectNavigableMap<T, V> descendingMap();
		@Override
		public abstract Object2ObjectNavigableMap<T, V> subMap(T fromKey, boolean fromInclusive, T toKey, boolean toInclusive);
		@Override
		public abstract Object2ObjectNavigableMap<T, V> headMap(T toKey, boolean inclusive);
		@Override
		public abstract Object2ObjectNavigableMap<T, V> tailMap(T fromKey, boolean inclusive); 
		
		boolean tooLow(T key) { return !fromStart && (loInclusive ? m.compare(key, low) < 0 : m.compare(key, low) <= 0); }
		boolean tooHigh(T key) { return !toEnd && (hiInclusive ? m.compare(key, high) > 0 : m.compare(key, high) >= 0); }
		boolean inRange(T key) { return !tooLow(key) && !tooHigh(key); }
		boolean inClosedRange(T key) { return (fromStart || m.compare(key, low) >= 0) && (toEnd || m.compare(high, key) >= 0); }
		boolean inRange(T key, boolean inclusive) { return inclusive ? inRange(key) : inClosedRange(key); }
		
		public T getDefaultMaxValue() { return m.getDefaultMaxValue(); }
		public T getDefaultMinValue() { return m.getDefaultMinValue(); }
		@Override
		public AbstractObject2ObjectMap<T, V> setDefaultReturnValue(V v) { 
			m.setDefaultReturnValue(v);
			return this;
		}
		
		@Override
		public V getDefaultReturnValue() { return m.getDefaultReturnValue(); }
		
		@Override
		public V putAndMoveToFirst(T key, V value) { throw new UnsupportedOperationException(); }
		@Override
		public V putAndMoveToLast(T key, V value) { throw new UnsupportedOperationException(); }
		@Override
		public boolean moveToFirst(T key) { throw new UnsupportedOperationException(); }
		@Override
		public boolean moveToLast(T key) { throw new UnsupportedOperationException(); }
		@Override
		public V getAndMoveToFirst(T key) { throw new UnsupportedOperationException(); }
		@Override
		public V getAndMoveToLast(T key) { throw new UnsupportedOperationException(); }
		@Override
		public Comparator<T> comparator() { return m.comparator(); }
		@Override
		public boolean containsKey(Object key) { return inRange((T)key) && m.containsKey(key); }
		@Override
		public T firstKey() { 
			Object2ObjectAVLTreeMap.Entry<T, V> entry = findLowest();
			return entry == null ? getDefaultMaxValue() : entry.key;
		}
		
		@Override
		public T pollFirstKey() {
			if(fromStart) return m.pollFirstKey();
			Object2ObjectAVLTreeMap.Entry<T, V> entry = loInclusive ? m.findCeilingNode(low) : m.findHigherNode(low);
			if(entry != null && !tooHigh(entry.key)) {
				T value = entry.key;
				m.removeNode(entry);
				return value;
			}
			return getDefaultMaxValue();
		}
		
		@Override
		public T lastKey() {
			Object2ObjectAVLTreeMap.Entry<T, V> entry = findHighest();
			return entry == null ? getDefaultMinValue() : entry.key;	
		}
		
		@Override
		public T pollLastKey() {
			if(toEnd) return m.pollLastKey();
			Object2ObjectAVLTreeMap.Entry<T, V> entry = hiInclusive ? m.findFloorNode(high) : m.findLowerNode(high);
			if(entry != null && !tooLow(entry.key)) {
				T value = entry.key;
				m.removeNode(entry);
				return value;
			}
			return getDefaultMinValue();
		}
		
		@Override
		public V firstValue() {
			Object2ObjectAVLTreeMap.Entry<T, V> entry = findLowest();
			return entry == null ? getDefaultReturnValue() : entry.value;
		}
		
		@Override
		public V lastValue() {
			Object2ObjectAVLTreeMap.Entry<T, V> entry = findHighest();
			return entry == null ? getDefaultReturnValue() : entry.value;
		}
		
		protected Object2ObjectAVLTreeMap.Entry<T, V> findLowest() {
			if(fromStart) return m.first;
			Object2ObjectAVLTreeMap.Entry<T, V> entry = loInclusive ? m.findCeilingNode(low) : m.findHigherNode(low);
			return entry == null || tooHigh(entry.key) ? null : entry;
		}
		
		protected Object2ObjectAVLTreeMap.Entry<T, V> findHighest() {
			if(toEnd) return m.last;
			Object2ObjectAVLTreeMap.Entry<T, V> entry = hiInclusive ? m.findFloorNode(high) : m.findLowerNode(high);
			return entry == null || tooLow(entry.key) ? null : entry;
		}
		
		@Override
		public V put(T key, V value) {
			if(!inRange(key)) throw new IllegalArgumentException("key out of range");
			return m.put(key, value);
		}
		
		@Override
		public V putIfAbsent(T key, V value) {
			if(!inRange(key)) throw new IllegalArgumentException("key out of range");
			return m.putIfAbsent(key, value);
		}
		
		@Override
		public V getObject(T key) {
			return inRange(key) ? m.getObject(key) : getDefaultReturnValue();
		}
		
		@Override
		public V getOrDefault(Object key, V defaultValue) {
			return inRange((T)key) ? m.getOrDefault(key, defaultValue) : getDefaultReturnValue();
		}
		
		@Override
		public V rem(T key) {
			return inRange(key) ? m.rem(key) : getDefaultReturnValue();
		}
		
		@Override
		public V remOrDefault(T key, V defaultValue) {
			return inRange(key) ? m.rem(key) : defaultValue;
		}
		
		@Override
		public boolean remove(Object key, Object value) {
			return inRange((T)key) && m.remove(key, value);
		}
		
		@Override
		public ObjectSet<Object2ObjectMap.Entry<T, V>> object2ObjectEntrySet() {
			if(entrySet == null) entrySet = new SubMapEntrySet();
			return entrySet;
		}
		
		@Override
		public ObjectSet<T> keySet() { return navigableKeySet(); }
		
		@Override
		public ObjectNavigableSet<T> navigableKeySet() {
			if(keySet == null) keySet = new KeySet<>(this);
			return keySet;
		}
		
		@Override
		public ObjectNavigableSet<T> descendingKeySet() {
			return descendingMap().navigableKeySet();
		}
		
		@Override
		public ObjectCollection<V> values() {
			if(values == null) values = new SubMapValues();
			return values;
		}
		
		@Override
		public void forEach(ObjectObjectConsumer<T, V> action) {
			for(Object2ObjectAVLTreeMap.Entry<T, V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
				action.accept(entry.key, entry.value);
		}
		
		@Override
		public int size() {
			return (fromStart && toEnd) ? m.size() : entrySet().size();
		}
		
		@Override
		public Object2ObjectMap.Entry<T, V> firstEntry() {
			Object2ObjectAVLTreeMap.Entry<T, V> entry = findLowest();
			return entry == null ? null : new BasicEntry<>(entry.key, entry.value);
		}
		
		@Override
		public Object2ObjectMap.Entry<T, V> lastEntry() {
			Object2ObjectAVLTreeMap.Entry<T, V> entry = findHighest();
			return entry == null ? null : new BasicEntry<>(entry.key, entry.value);
		}
		
		@Override
		public Object2ObjectMap.Entry<T, V> pollFirstEntry() {
			Object2ObjectAVLTreeMap.Entry<T, V> entry = findLowest();
			if(entry == null) return null;
			BasicEntry<T, V> result = new BasicEntry<>(entry.key, entry.value);
			m.removeNode(entry);
			return result;
		}
		
		@Override
		public Object2ObjectMap.Entry<T, V> pollLastEntry() {
			Object2ObjectAVLTreeMap.Entry<T, V> entry = findHighest();
			if(entry == null) return null;
			BasicEntry<T, V> result = new BasicEntry<>(entry.key, entry.value);
			m.removeNode(entry);
			return result;
		}
		
		@Override
		public T lowerKey(T e) {
			if(tooHigh(e)) {
				Object2ObjectAVLTreeMap.Entry<T, V> entry = findHighest();
				return entry == null ? getDefaultMinValue() : entry.key;
			}
			Object2ObjectAVLTreeMap.Entry<T, V> entry = m.findLowerNode(e);
			return entry == null || tooHigh(entry.key) ? getDefaultMaxValue() : entry.key;
		}
		
		@Override
		public T floorKey(T e) {
			if(tooHigh(e)) {
				Object2ObjectAVLTreeMap.Entry<T, V> entry = findHighest();
				return entry == null ? getDefaultMinValue() : entry.key;
			}
			Object2ObjectAVLTreeMap.Entry<T, V> entry = m.findFloorNode(e);
			return entry == null || tooHigh(entry.key) ? getDefaultMaxValue() : entry.key;
		}
		
		@Override
		public T ceilingKey(T e) {
			if(tooLow(e)) {
				Object2ObjectAVLTreeMap.Entry<T, V> entry = findLowest();
				return entry == null ? getDefaultMaxValue() : entry.key;
			}
			Object2ObjectAVLTreeMap.Entry<T, V> entry = m.findCeilingNode(e);
			return entry == null || tooLow(entry.key) ? getDefaultMinValue() : entry.key;
		}
		
		@Override
		public T higherKey(T e) {
			if(tooLow(e)) {
				Object2ObjectAVLTreeMap.Entry<T, V> entry = findLowest();
				return entry == null ? getDefaultMaxValue() : entry.key;
			}
			Object2ObjectAVLTreeMap.Entry<T, V> entry = m.findHigherNode(e);
			return entry == null || tooLow(entry.key) ? getDefaultMinValue() : entry.key;
		}
		
		@Override
		public Object2ObjectMap.Entry<T, V> lowerEntry(T e) {
			if(tooHigh(e)) {
				Object2ObjectAVLTreeMap.Entry<T, V> entry = findHighest();
				return entry == null ? null : new BasicEntry<>(entry.key, entry.value);
			}
			Object2ObjectAVLTreeMap.Entry<T, V> entry = m.findLowerNode(e);
			return entry == null || tooHigh(entry.key) ? null : new BasicEntry<>(entry.key, entry.value);
		}
		
		@Override
		public Object2ObjectMap.Entry<T, V> floorEntry(T e) {
			if(tooHigh(e)) {
				Object2ObjectAVLTreeMap.Entry<T, V> entry = findHighest();
				return entry == null ? null : new BasicEntry<>(entry.key, entry.value);
			}
			Object2ObjectAVLTreeMap.Entry<T, V> entry = m.findFloorNode(e);
			return entry == null || tooHigh(entry.key) ? null : new BasicEntry<>(entry.key, entry.value);
		}
		
		@Override
		public Object2ObjectMap.Entry<T, V> ceilingEntry(T e) {
			if(tooLow(e)) {
				Object2ObjectAVLTreeMap.Entry<T, V> entry = findLowest();
				return entry == null ? null : new BasicEntry<>(entry.key, entry.value);
			}
			Object2ObjectAVLTreeMap.Entry<T, V> entry = m.findCeilingNode(e);
			return entry == null || tooLow(entry.key) ? null : new BasicEntry<>(entry.key, entry.value);
		}
		
		@Override
		public Object2ObjectMap.Entry<T, V> higherEntry(T e) {
			if(tooLow(e)) {
				Object2ObjectAVLTreeMap.Entry<T, V> entry = findLowest();
				return entry == null ? null : new BasicEntry<>(entry.key, entry.value);
			}
			Object2ObjectAVLTreeMap.Entry<T, V> entry = m.findHigherNode(e);
			return entry == null || tooLow(entry.key) ? null : new BasicEntry<>(entry.key, entry.value);
		}
		
		class SubMapEntrySet extends AbstractObjectSet<Object2ObjectMap.Entry<T, V>> implements ObjectSet<Object2ObjectMap.Entry<T, V>> {
			@Override
			@Deprecated
			public boolean contains(Object o) {
				if(o instanceof Map.Entry) {
					if(o instanceof Object2ObjectMap.Entry) {
						Object2ObjectMap.Entry<T, V> entry = (Object2ObjectMap.Entry<T, V>)o;
						if(entry.getKey() == null && comparator() == null) return false;
						Object2ObjectAVLTreeMap.Entry<T, V> subEntry = m.findNode(entry.getKey());
						if(subEntry != null) return Objects.equals(entry.getValue(), subEntry.value);
					}
					else {
						Map.Entry<?, ?> entry = (Map.Entry<?, ?>)o;
						if(entry.getKey() == null && comparator() == null) return false;
						Object2ObjectAVLTreeMap.Entry<T, V> subEntry = m.findNode((T)entry.getKey());
						if(subEntry != null) return Objects.equals(entry.getValue(), subEntry.value);
					}
				}
				return false;
			}
			
			@Override
			@Deprecated
			public boolean remove(Object o) {
				if(o instanceof Map.Entry) {
					if(o instanceof Object2ObjectMap.Entry) {
						Object2ObjectMap.Entry<T, V> entry = (Object2ObjectMap.Entry<T, V>)o;
						return NavigableSubMap.this.remove(entry.getKey(), entry.getValue());
					}
					Map.Entry<?, ?> entry = (Map.Entry<?, ?>)o;
					return NavigableSubMap.this.remove(entry.getKey(), entry.getValue());
				}
				return false;
			}
			@Override
			public ObjectIterator<Object2ObjectMap.Entry<T, V>> iterator() { return entryIterator(); }
			@Override
			public int size() { return NavigableSubMap.this.size(); }
			@Override
			public void clear() { NavigableSubMap.this.clear(); }
			
			@Override
			public void forEach(Consumer<? super Object2ObjectMap.Entry<T, V>> action) {
				Objects.requireNonNull(action);
				for(Object2ObjectAVLTreeMap.Entry<T, V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					action.accept(new BasicEntry<>(entry.key, entry.value));
			}
			
			@Override
			public <E> void forEach(E input, ObjectObjectConsumer<E, Object2ObjectMap.Entry<T, V>> action) {
				Objects.requireNonNull(action);
				for(Object2ObjectAVLTreeMap.Entry<T, V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					action.accept(input, new BasicEntry<>(entry.key, entry.value));
			}
			
			@Override
			public boolean matchesAny(Object2BooleanFunction<Object2ObjectMap.Entry<T, V>> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return false;
				BasicEntry<T, V> subEntry = new BasicEntry<>();
				for(Object2ObjectAVLTreeMap.Entry<T, V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					subEntry.set(entry.key, entry.value);
					if(filter.getBoolean(subEntry)) return true;
				}
				return false;
			}
			
			@Override
			public boolean matchesNone(Object2BooleanFunction<Object2ObjectMap.Entry<T, V>> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return true;
				BasicEntry<T, V> subEntry = new BasicEntry<>();
				for(Object2ObjectAVLTreeMap.Entry<T, V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					subEntry.set(entry.key, entry.value);
					if(filter.getBoolean(subEntry)) return false;
				}
				return true;
			}
			
			@Override
			public boolean matchesAll(Object2BooleanFunction<Object2ObjectMap.Entry<T, V>> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return true;
				BasicEntry<T, V> subEntry = new BasicEntry<>();
				for(Object2ObjectAVLTreeMap.Entry<T, V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					subEntry.set(entry.key, entry.value);
					if(!filter.getBoolean(subEntry)) return false;
				}
				return true;
			}
			
			@Override
			public <E> E reduce(E identity, BiFunction<E, Object2ObjectMap.Entry<T, V>, E> operator) {
				Objects.requireNonNull(operator);
				E state = identity;
				for(Object2ObjectAVLTreeMap.Entry<T, V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					state = operator.apply(state, new BasicEntry<>(entry.key, entry.value));
				}
				return state;
			}
			
			@Override
			public Object2ObjectMap.Entry<T, V> reduce(ObjectObjectUnaryOperator<Object2ObjectMap.Entry<T, V>, Object2ObjectMap.Entry<T, V>> operator) {
				Objects.requireNonNull(operator);
				Object2ObjectMap.Entry<T, V> state = null;
				boolean empty = true;
				for(Object2ObjectAVLTreeMap.Entry<T, V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					if(empty) {
						empty = false;
						state = new BasicEntry<>(entry.key, entry.value);
						continue;
					}
					state = operator.apply(state, new BasicEntry<>(entry.key, entry.value));
				}
				return state;
			}
			
			@Override
			public Object2ObjectMap.Entry<T, V> findFirst(Object2BooleanFunction<Object2ObjectMap.Entry<T, V>> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return null;
				BasicEntry<T, V> subEntry = new BasicEntry<>();
				for(Object2ObjectAVLTreeMap.Entry<T, V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					subEntry.set(entry.key, entry.value);
					if(filter.getBoolean(subEntry)) return subEntry;
				}
				return null;
			}
			
			@Override
			public int count(Object2BooleanFunction<Object2ObjectMap.Entry<T, V>> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return 0;
				int result = 0;
				BasicEntry<T, V> subEntry = new BasicEntry<>();
				for(Object2ObjectAVLTreeMap.Entry<T, V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					subEntry.set(entry.key, entry.value);
					if(filter.getBoolean(subEntry)) result++;
				}
				return result;
			}
		}
		
		final class SubMapValues extends AbstractObjectCollection<V> {
			@Override
			public boolean add(V o) { throw new UnsupportedOperationException(); }
			
			@Override
			public boolean contains(Object e) {
				return containsValue(e);
			}
			
			@Override
			public ObjectIterator<V> iterator() { return valueIterator(); }
			
			@Override
			public int size() {
				return NavigableSubMap.this.size();
			}
			
			@Override
			public void clear() {
				NavigableSubMap.this.clear();
			}
			
			@Override
			public void forEach(Consumer<? super V> action) {
				Objects.requireNonNull(action);
				for(Object2ObjectAVLTreeMap.Entry<T, V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					action.accept(entry.value);
			}
			
			@Override
			public <E> void forEach(E input, ObjectObjectConsumer<E, V> action) {
				Objects.requireNonNull(action);
				for(Object2ObjectAVLTreeMap.Entry<T, V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					action.accept(input, entry.value);
			}
			
			@Override
			public boolean matchesAny(Object2BooleanFunction<V> filter) {
				Objects.requireNonNull(filter);
				for(Object2ObjectAVLTreeMap.Entry<T, V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					if(filter.getBoolean(entry.value)) return true;
				return false;
			}
			
			@Override
			public boolean matchesNone(Object2BooleanFunction<V> filter) {
				Objects.requireNonNull(filter);
				for(Object2ObjectAVLTreeMap.Entry<T, V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					if(filter.getBoolean(entry.value)) return false;
				return true;
			}
			
			@Override
			public boolean matchesAll(Object2BooleanFunction<V> filter) {
				Objects.requireNonNull(filter);
				for(Object2ObjectAVLTreeMap.Entry<T, V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					if(!filter.getBoolean(entry.value)) return false;
				return true;
			}
			
			@Override
			public <E> E reduce(E identity, BiFunction<E, V, E> operator) {
				Objects.requireNonNull(operator);
				E state = identity;
				for(Object2ObjectAVLTreeMap.Entry<T, V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					state = operator.apply(state, entry.value);
				return state;
			}
			
			@Override
			public V reduce(ObjectObjectUnaryOperator<V, V> operator) {
				Objects.requireNonNull(operator);
				V state = null;
				boolean empty = true;
				for(Object2ObjectAVLTreeMap.Entry<T, V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					if(empty) {
						empty = false;
						state = entry.value;
						continue;
					}
					state = operator.apply(state, entry.value);
				}
				return state;
			}
			
			@Override
			public V findFirst(Object2BooleanFunction<V> filter) {
				Objects.requireNonNull(filter);
				for(Object2ObjectAVLTreeMap.Entry<T, V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					if(filter.getBoolean(entry.value)) return entry.value;
				return null;
			}
			
			@Override
			public int count(Object2BooleanFunction<V> filter) {
				Objects.requireNonNull(filter);
				int result = 0;
				for(Object2ObjectAVLTreeMap.Entry<T, V> entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					if(filter.getBoolean(entry.value)) result++;
				return result;
			}
		}
		
		final class SubMapEntrySetIterator extends SubMapEntryIterator implements ObjectListIterator<Object2ObjectMap.Entry<T, V>> {
			public SubMapEntrySetIterator(boolean descending) {
				super(descending);
			}
			
			public SubMapEntrySetIterator(Object2ObjectAVLTreeMap.Entry<T, V> entry, Object2ObjectAVLTreeMap.Entry<T, V> lowerFence, Object2ObjectAVLTreeMap.Entry<T, V> upperFence) {
				super(entry, lowerFence, upperFence);
			}
			
			@Override
			public Object2ObjectMap.Entry<T, V> next() { return nextEntry(); }
			@Override
			public Object2ObjectMap.Entry<T, V> previous() { return previousEntry(); }
			
			@Override
			public void set(Object2ObjectMap.Entry<T, V> e) { throw new UnsupportedOperationException(); }
			@Override
			public void add(Object2ObjectMap.Entry<T, V> e) { throw new UnsupportedOperationException(); }
		}
		
		final class SubMapKeyIterator extends SubMapEntryIterator implements ObjectListIterator<T> {
			public SubMapKeyIterator(boolean descending) {
				super(descending);
			}
			
			public SubMapKeyIterator(Object2ObjectAVLTreeMap.Entry<T, V> entry, Object2ObjectAVLTreeMap.Entry<T, V> lowerFence, Object2ObjectAVLTreeMap.Entry<T, V> upperFence) {
				super(entry, lowerFence, upperFence);
			}
			
			@Override
			public T next() { return nextEntry().key; }
			@Override
			public T previous() { return previousEntry().key; }
			
			@Override
			public void set(T e) { throw new UnsupportedOperationException(); }
			@Override
			public void add(T e) { throw new UnsupportedOperationException(); }
		}
		
		final class SubMapValueIterator extends SubMapEntryIterator implements ObjectListIterator<V> {
			public SubMapValueIterator(boolean descending) {
				super(descending);
			}
			
			public SubMapValueIterator(Object2ObjectAVLTreeMap.Entry<T, V> entry, Object2ObjectAVLTreeMap.Entry<T, V> lowerFence, Object2ObjectAVLTreeMap.Entry<T, V> upperFence) {
				super(entry, lowerFence, upperFence);
			}
			
			@Override
			public V next() { return nextEntry().value; }
			@Override
			public V previous() { return previousEntry().value; }
			
			@Override
			public void set(V e) { throw new UnsupportedOperationException(); }
			@Override
			public void add(V e) { throw new UnsupportedOperationException(); }
		}
		
		abstract class SubMapEntryIterator {
			T lowerFence;
			T upperFence;
			Object2ObjectAVLTreeMap.Entry<T, V> next;
			Object2ObjectAVLTreeMap.Entry<T, V> previous;
			Object2ObjectAVLTreeMap.Entry<T, V> current;
			int index = 0;
			
			public SubMapEntryIterator(boolean descending) {
				this(descending ? findHighest() : findLowest(), fromStart ? null : findLowest(), toEnd ? null : findHighest());
			}
			
			public SubMapEntryIterator(Object2ObjectAVLTreeMap.Entry<T, V> entry, Object2ObjectAVLTreeMap.Entry<T, V> lowerFence, Object2ObjectAVLTreeMap.Entry<T, V> upperFence) {
				next = entry;
				previous = entry.previous();
				this.lowerFence = lowerFence != null ? lowerFence.key : null;
				this.upperFence = upperFence != null ? upperFence.key : null;
			}
			
			public boolean hasNext() {
				return next != null && (upperFence == null || Objects.equals(next.key, upperFence));
			}
			
			public boolean hasPrevious() {
				return previous != null && (lowerFence == null || Objects.equals(next.key, lowerFence));
			}
			
			public int nextIndex() {
				return index;
			}
			
			public int previousIndex() {
				return index - 1;
			}
			
			protected void updateNext() {
				next = current.next();
			}
			
			protected void updatePrevious() {
				previous = current.previous();
			}
			
			public Object2ObjectAVLTreeMap.Entry<T, V> nextEntry() {
				if(!hasNext()) throw new NoSuchElementException();
				current = previous = next;
				updateNext();
				index++;
				return current;
			}
			
			public Object2ObjectAVLTreeMap.Entry<T, V> previousEntry() {
				if(!hasPrevious()) throw new NoSuchElementException();
				current = next = previous;
				updatePrevious();
				index--;
				return current;
			}
			
			public void remove() {
				if(current == null) throw new IllegalStateException();
				if(current == previous) index--;
				updateNext();
				updatePrevious();
				if(current.needsSuccessor()) next = current;
				m.removeNode(current);
				current = null;
			}
		}
	}
	
	private class EntrySet extends AbstractObjectSet<Object2ObjectMap.Entry<T, V>> {
		
		@Override
		public ObjectIterator<Object2ObjectMap.Entry<T, V>> iterator() {
			return new EntryIterator(false);
		}
		
		@Override
		@Deprecated
		public boolean contains(Object o) {
			if(o instanceof Map.Entry) {
				if(o instanceof Object2ObjectMap.Entry) {
					Object2ObjectMap.Entry<T, V> entry = (Object2ObjectMap.Entry<T, V>)o;
					if(entry.getKey() == null && comparator() == null) return false;
					Object2ObjectAVLTreeMap.Entry<T, V> subEntry = Object2ObjectAVLTreeMap.this.findNode(entry.getKey());
					if(subEntry != null) return Objects.equals(entry.getValue(), subEntry.value);
				}
				else {
					Map.Entry<?, ?> entry = (Map.Entry<?, ?>)o;
					if(entry.getKey() == null && comparator() == null) return false;
					Object2ObjectAVLTreeMap.Entry<T, V> subEntry = Object2ObjectAVLTreeMap.this.findNode((T)entry.getKey());
					if(subEntry != null) return Objects.equals(entry.getValue(), subEntry.value);
				}
			}
			return false;
		}
		
		@Override
		@Deprecated
		public boolean remove(Object o) {
			if(o instanceof Map.Entry) {
				if(o instanceof Object2ObjectMap.Entry) {
					Object2ObjectMap.Entry<T, V> entry = (Object2ObjectMap.Entry<T, V>)o;
					return Object2ObjectAVLTreeMap.this.remove(entry.getKey(), entry.getValue());
				}
				Map.Entry<?, ?> entry = (Map.Entry<?, ?>)o;
				return Object2ObjectAVLTreeMap.this.remove(entry.getKey(), entry.getValue());
			}
			return false;
		}
		
		@Override
		public int size() {
			return Object2ObjectAVLTreeMap.this.size();
		}
		
		@Override
		public void clear() {
			Object2ObjectAVLTreeMap.this.clear();
		}
		
		@Override
		public void forEach(Consumer<? super Object2ObjectMap.Entry<T, V>> action) {
			Objects.requireNonNull(action);
			for(Object2ObjectAVLTreeMap.Entry<T, V> entry = first;entry != null;entry = entry.next())
				action.accept(new BasicEntry<>(entry.key, entry.value));
		}
		
		@Override
		public <E> void forEach(E input, ObjectObjectConsumer<E, Object2ObjectMap.Entry<T, V>> action) {
			Objects.requireNonNull(action);
			for(Object2ObjectAVLTreeMap.Entry<T, V> entry = first;entry != null;entry = entry.next())
				action.accept(input, new BasicEntry<>(entry.key, entry.value));
		}
		
		@Override
		public boolean matchesAny(Object2BooleanFunction<Object2ObjectMap.Entry<T, V>> filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return false;
			BasicEntry<T, V> subEntry = new BasicEntry<>();
			for(Object2ObjectAVLTreeMap.Entry<T, V> entry = first;entry != null;entry = entry.next()) {
				subEntry.set(entry.key, entry.value);
				if(filter.getBoolean(subEntry)) return true;
			}
			return false;
		}
		
		@Override
		public boolean matchesNone(Object2BooleanFunction<Object2ObjectMap.Entry<T, V>> filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return true;
			BasicEntry<T, V> subEntry = new BasicEntry<>();
			for(Object2ObjectAVLTreeMap.Entry<T, V> entry = first;entry != null;entry = entry.next()) {
				subEntry.set(entry.key, entry.value);
				if(filter.getBoolean(subEntry)) return false;
			}
			return true;
		}
		
		@Override
		public boolean matchesAll(Object2BooleanFunction<Object2ObjectMap.Entry<T, V>> filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return true;
			BasicEntry<T, V> subEntry = new BasicEntry<>();
			for(Object2ObjectAVLTreeMap.Entry<T, V> entry = first;entry != null;entry = entry.next()) {
				subEntry.set(entry.key, entry.value);
				if(!filter.getBoolean(subEntry)) return false;
			}
			return true;
		}
		
		@Override
		public <E> E reduce(E identity, BiFunction<E, Object2ObjectMap.Entry<T, V>, E> operator) {
			Objects.requireNonNull(operator);
			E state = identity;
			for(Object2ObjectAVLTreeMap.Entry<T, V> entry = first;entry != null;entry = entry.next()) {
				state = operator.apply(state, new BasicEntry<>(entry.key, entry.value));
			}
			return state;
		}
		
		@Override
		public Object2ObjectMap.Entry<T, V> reduce(ObjectObjectUnaryOperator<Object2ObjectMap.Entry<T, V>, Object2ObjectMap.Entry<T, V>> operator) {
			Objects.requireNonNull(operator);
			Object2ObjectMap.Entry<T, V> state = null;
			boolean empty = true;
			for(Object2ObjectAVLTreeMap.Entry<T, V> entry = first;entry != null;entry = entry.next()) {
				if(empty) {
					empty = false;
					state = new BasicEntry<>(entry.key, entry.value);
					continue;
				}
				state = operator.apply(state, new BasicEntry<>(entry.key, entry.value));
			}
			return state;
		}
		
		@Override
		public Object2ObjectMap.Entry<T, V> findFirst(Object2BooleanFunction<Object2ObjectMap.Entry<T, V>> filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return null;
			BasicEntry<T, V> subEntry = new BasicEntry<>();
			for(Object2ObjectAVLTreeMap.Entry<T, V> entry = first;entry != null;entry = entry.next()) {
				subEntry.set(entry.key, entry.value);
				if(filter.getBoolean(subEntry)) return subEntry;
			}
			return null;
		}
	}
	
	static final class KeySet<T, V> extends AbstractObjectSet<T> implements ObjectNavigableSet<T> {
		Object2ObjectNavigableMap<T, V> m;
		
		KeySet(Object2ObjectNavigableMap<T, V> m) {
			this.m = m;
		}
		
		@Override
		public boolean add(T o) { throw new UnsupportedOperationException(); }
		@Override
		public boolean addAndMoveToFirst(T o) { throw new UnsupportedOperationException(); }
		@Override
		public boolean addAndMoveToLast(T o) { throw new UnsupportedOperationException(); }
		@Override
		public boolean moveToFirst(T o) { throw new UnsupportedOperationException(); }
		@Override
		public boolean moveToLast(T o) { throw new UnsupportedOperationException(); }
		@Override
		public Comparator<T> comparator() { return m.comparator(); }
		@Override
		public T lower(T e) { return m.lowerKey(e); }
		@Override
		public T floor(T e) { return m.floorKey(e); }
		@Override
		public T ceiling(T e) { return m.ceilingKey(e); }
		@Override
		public T higher(T e) { return m.higherKey(e); }
		@Override
		public T first() { return m.firstKey(); }
		@Override
		public T pollFirst() { return m.pollFirstKey(); }
		@Override
		public T last() { return m.lastKey(); }
		@Override
		public T pollLast() { return m.pollLastKey(); }		
		@Override
		public boolean remove(Object o) { 
			int oldSize = m.size();
			m.remove(o); 
			return oldSize != m.size();
		}
		
		@Override
		public KeySet<T, V> copy() { throw new UnsupportedOperationException(); }
		
		@Override
		public ObjectNavigableSet<T> subSet(T fromElement, boolean fromInclusive, T toElement, boolean toInclusive) {
			return new KeySet<>(m.subMap(fromElement, fromInclusive, toElement, toInclusive));
		}
		
		@Override
		public ObjectNavigableSet<T> headSet(T toElement, boolean inclusive) {
			return new KeySet<>(m.headMap(toElement, inclusive));
		}
		
		@Override
		public ObjectNavigableSet<T> tailSet(T fromElement, boolean inclusive) {
			return new KeySet<>(m.tailMap(fromElement, inclusive));
		}
		
		@Override
		public ObjectNavigableSet<T> descendingSet() {
			return new KeySet<>(m.descendingMap());
		}
		
		@Override
		public ObjectBidirectionalIterator<T> iterator() {
			if(m instanceof Object2ObjectAVLTreeMap) return ((Object2ObjectAVLTreeMap<T, V>)m).keyIterator(false);
			return ((NavigableSubMap<T, V>)m).keyIterator(false);
		}
		
		@Override
		public ObjectBidirectionalIterator<T> iterator(T fromElement) {
			if(m instanceof Object2ObjectAVLTreeMap) return ((Object2ObjectAVLTreeMap<T, V>)m).keyIterator(fromElement);
			return ((NavigableSubMap<T, V>)m).keyIterator(fromElement);
		}
		
		@Override
		public ObjectBidirectionalIterator<T> descendingIterator() {
			if(m instanceof Object2ObjectAVLTreeMap) return ((Object2ObjectAVLTreeMap<T, V>)m).keyIterator(true);
			return ((NavigableSubMap<T, V>)m).keyIterator(true);
		}
		
		@Override
		public int size() {
			return m.size();
		}
		
		@Override
		public void clear() {
			m.clear();
		}
		
		protected Object2ObjectAVLTreeMap.Entry<T, V> start() {
			if(m instanceof Object2ObjectAVLTreeMap) return ((Object2ObjectAVLTreeMap<T, V>)m).first;
			return ((NavigableSubMap<T, V>)m).findLowest();
		}
		
		protected Object2ObjectAVLTreeMap.Entry<T, V> end() {
			if(m instanceof Object2ObjectAVLTreeMap) return null;
			return ((NavigableSubMap<T, V>)m).findLowest();
		}
		
		protected Object2ObjectAVLTreeMap.Entry<T, V> next(Object2ObjectAVLTreeMap.Entry<T, V> entry) {
			if(m instanceof Object2ObjectAVLTreeMap) return entry.next();
			return ((NavigableSubMap<T, V>)m).next(entry);
		}
		
		protected Object2ObjectAVLTreeMap.Entry<T, V> previous(Object2ObjectAVLTreeMap.Entry<T, V> entry) {
			if(m instanceof Object2ObjectAVLTreeMap) return entry.previous();
			return ((NavigableSubMap<T, V>)m).previous(entry);
		}
		
		@Override
		public void forEach(Consumer<? super T> action) {
			Objects.requireNonNull(action);
			for(Object2ObjectAVLTreeMap.Entry<T, V> entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				action.accept(entry.key);
		}
		
		@Override
		public <E> void forEach(E input, ObjectObjectConsumer<E, T> action) {
			Objects.requireNonNull(action);
			for(Object2ObjectAVLTreeMap.Entry<T, V> entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				action.accept(input, entry.key);
		}
		
		@Override
		public boolean matchesAny(Object2BooleanFunction<T> filter) {
			Objects.requireNonNull(filter);
			for(Object2ObjectAVLTreeMap.Entry<T, V> entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				if(filter.getBoolean(entry.key)) return true;
			return false;
		}
		
		@Override
		public boolean matchesNone(Object2BooleanFunction<T> filter) {
			Objects.requireNonNull(filter);
			for(Object2ObjectAVLTreeMap.Entry<T, V> entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				if(filter.getBoolean(entry.key)) return false;
			return true;
		}
		
		@Override
		public boolean matchesAll(Object2BooleanFunction<T> filter) {
			Objects.requireNonNull(filter);
			for(Object2ObjectAVLTreeMap.Entry<T, V> entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				if(!filter.getBoolean(entry.key)) return false;
			return true;
		}
		
		@Override
		public <E> E reduce(E identity, BiFunction<E, T, E> operator) {
			Objects.requireNonNull(operator);
			E state = identity;
			for(Object2ObjectAVLTreeMap.Entry<T, V> entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				state = operator.apply(state, entry.key);
			return state;
		}
		
		@Override
		public T reduce(ObjectObjectUnaryOperator<T, T> operator) {
			Objects.requireNonNull(operator);
			T state = null;
			boolean empty = true;
			for(Object2ObjectAVLTreeMap.Entry<T, V> entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry)) {
				if(empty) {
					empty = false;
					state = entry.key;
					continue;
				}
				state = operator.apply(state, entry.key);
			}
			return state;
		}
		
		@Override
		public T findFirst(Object2BooleanFunction<T> filter) {
			Objects.requireNonNull(filter);
			for(Object2ObjectAVLTreeMap.Entry<T, V> entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				if(filter.getBoolean(entry.key)) return entry.key;
			return null;
		}
		
		@Override
		public int count(Object2BooleanFunction<T> filter) {
			Objects.requireNonNull(filter);
			int result = 0;
			for(Object2ObjectAVLTreeMap.Entry<T, V> entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				if(filter.getBoolean(entry.key)) result++;
			return result;
		}
	}
	
	final class Values extends AbstractObjectCollection<V> {

		@Override
		public boolean add(V o) { throw new UnsupportedOperationException(); }
		
		@Override
		public boolean contains(Object e) {
			return containsValue(e);
		}
		
		@Override
		public ObjectIterator<V> iterator() { return new ValueIterator(false); }
		
		@Override
		public int size() {
			return Object2ObjectAVLTreeMap.this.size();
		}
		
		@Override
		public void clear() {
			Object2ObjectAVLTreeMap.this.clear();
		}
		
		@Override
		public void forEach(Consumer<? super V> action) {
			Objects.requireNonNull(action);
			for(Object2ObjectAVLTreeMap.Entry<T, V> entry = first;entry != null;entry = entry.next())
				action.accept(entry.value);
		}
		
		@Override
		public <E> void forEach(E input, ObjectObjectConsumer<E, V> action) {
			Objects.requireNonNull(action);
			for(Object2ObjectAVLTreeMap.Entry<T, V> entry = first;entry != null;entry = entry.next())
				action.accept(input, entry.value);
		}
		
		@Override
		public boolean matchesAny(Object2BooleanFunction<V> filter) {
			Objects.requireNonNull(filter);
			for(Object2ObjectAVLTreeMap.Entry<T, V> entry = first;entry != null;entry = entry.next())
				if(filter.getBoolean(entry.value)) return true;
			return false;
		}
		
		@Override
		public boolean matchesNone(Object2BooleanFunction<V> filter) {
			Objects.requireNonNull(filter);
			for(Object2ObjectAVLTreeMap.Entry<T, V> entry = first;entry != null;entry = entry.next())
				if(filter.getBoolean(entry.value)) return false;
			return true;
		}
		
		@Override
		public boolean matchesAll(Object2BooleanFunction<V> filter) {
			Objects.requireNonNull(filter);
			for(Object2ObjectAVLTreeMap.Entry<T, V> entry = first;entry != null;entry = entry.next())
				if(!filter.getBoolean(entry.value)) return false;
			return true;
		}
		
		@Override
		public <E> E reduce(E identity, BiFunction<E, V, E> operator) {
			Objects.requireNonNull(operator);
			E state = identity;
			for(Object2ObjectAVLTreeMap.Entry<T, V> entry = first;entry != null;entry = entry.next())
				state = operator.apply(state, entry.value);
			return state;
		}
		
		@Override
		public V reduce(ObjectObjectUnaryOperator<V, V> operator) {
			Objects.requireNonNull(operator);
			V state = null;
			boolean empty = true;
			for(Object2ObjectAVLTreeMap.Entry<T, V> entry = first;entry != null;entry = entry.next()) {
				if(empty) {
					empty = false;
					state = entry.value;
					continue;
				}
				state = operator.apply(state, entry.value);
			}
			return state;
		}
		
		@Override
		public V findFirst(Object2BooleanFunction<V> filter) {
			Objects.requireNonNull(filter);
			for(Object2ObjectAVLTreeMap.Entry<T, V> entry = first;entry != null;entry = entry.next())
				if(filter.getBoolean(entry.value)) return entry.value;
			return null;
		}
		
		@Override
		public int count(Object2BooleanFunction<V> filter) {
			Objects.requireNonNull(filter);
			int result = 0;
			for(Object2ObjectAVLTreeMap.Entry<T, V> entry = first;entry != null;entry = entry.next())
				if(filter.getBoolean(entry.value)) result++;
			return result;
		}
	}
	
	final class EntryIterator extends MapEntryIterator implements ObjectListIterator<Object2ObjectMap.Entry<T, V>> {
		
		public EntryIterator(boolean descending) {
			super(descending);
		}
		
		public EntryIterator(Entry<T, V> entry) {
			super(entry);
		}
		
		@Override
		public Object2ObjectMap.Entry<T, V> next() {
			return nextEntry();
		}
		
		@Override
		public Object2ObjectMap.Entry<T, V> previous() {
			return previousEntry();
		}
		
		@Override
		public void set(Object2ObjectMap.Entry<T, V> e) { throw new UnsupportedOperationException(); }
		
		@Override
		public void add(Object2ObjectMap.Entry<T, V> e) { throw new UnsupportedOperationException(); }
	}
	
	final class KeyIterator extends MapEntryIterator implements ObjectListIterator<T> {
		
		public KeyIterator(boolean descending) {
			super(descending);
		}
		
		public KeyIterator(Entry<T, V> entry) {
			super(entry);
		}
		
		@Override
		public T previous() { return previousEntry().key; }
		@Override
		public T next() { return nextEntry().key; }
		
		@Override
		public void set(T e) { throw new UnsupportedOperationException(); }
		@Override
		public void add(T e) { throw new UnsupportedOperationException(); }
	}
	
	final class ValueIterator extends MapEntryIterator implements ObjectListIterator<V> {
		
		public ValueIterator(boolean descending) {
			super(descending);
		}
		
		public ValueIterator(Entry<T, V> entry) {
			super(entry);
		}
		
		@Override
		public V previous() {
			return previousEntry().value;
		}
		
		@Override
		public V next() {
			return nextEntry().value;
		}
		
		@Override
		public void set(V e) { throw new UnsupportedOperationException(); }
		
		@Override
		public void add(V e) { throw new UnsupportedOperationException(); }
	}
	
	abstract class MapEntryIterator {
		Entry<T, V> next;
		Entry<T, V> previous;
		Entry<T, V> current;
		int index = 0;
		
		public MapEntryIterator(boolean descending) {
			if(descending) previous = last;
			else next = first;
		}
		
		public MapEntryIterator(Entry<T, V> entry) {
			next = entry;
			previous = entry.previous();
		}
		
		public boolean hasNext() {
			return next != null;
		}
		
		public boolean hasPrevious() {
			return previous != null;
		}
		
		public int nextIndex() {
			return index;
		}
		
		public int previousIndex() {
			return index - 1;
		}
		
		protected void updateNext() {
			next = current.next();
		}
		
		protected void updatePrevious() {
			previous = current.previous();
		}
		
		public Entry<T, V> nextEntry() {
			if(!hasNext()) throw new NoSuchElementException();
			current = previous = next;
			updateNext();
			index++;
			return current;
		}
		
		public Entry<T, V> previousEntry() {
			if(!hasPrevious()) throw new NoSuchElementException();
			current = next = previous;
			updatePrevious();
			index--;
			return current;
		}
		
		public void remove() {
			if(current == null) throw new IllegalStateException();
			if(current == previous) index--;
			updateNext();
			updatePrevious();
			if(current.needsSuccessor()) next = current;
			removeNode(current);
			current = null;
		}
	}
	
	private static final class Entry<T, V> implements Object2ObjectMap.Entry<T, V>
	{
		T key;
		V value;
		int state;
		Entry<T, V> parent;
		Entry<T, V> left;
		Entry<T, V> right;
		
		Entry(T key, V value, Entry<T, V> parent) {
			this.key = key;
			this.value = value;
			this.parent = parent;
		}
		
		Entry<T, V> copy() {
			Entry<T, V> entry = new Entry<>(key, value, null);
			entry.state = state;
			if(left != null) {
				Entry<T, V> newLeft = left.copy();
				entry.left = newLeft;
				newLeft.parent = entry;
			}
			if(right != null) {
				Entry<T, V> newRight = right.copy();
				entry.right = newRight;
				newRight.parent = entry;
			}
			return entry;
		}
		
		@Override
		public T getKey() {
			return key;
		}
		
		@Override
		public V getValue() {
			return value;
		}
		
		@Override
		public V setValue(V value) {
			V oldValue = this.value;
			this.value = value;
			return oldValue;
		}
		
		@Override
		public boolean equals(Object obj) {
			if(obj instanceof Map.Entry) {
				if(obj instanceof Object2ObjectMap.Entry) {
					Object2ObjectMap.Entry<T, V> entry = (Object2ObjectMap.Entry<T, V>)obj;
					if(entry.getKey() == null) return false;
					return Objects.equals(key, entry.getKey()) && Objects.equals(value, entry.getValue());
				}
				Map.Entry<?, ?> entry = (Map.Entry<?, ?>)obj;
				Object otherKey = entry.getKey();
				if(otherKey == null) return false;
				Object otherValue = entry.getValue();
				return Objects.equals(key, otherKey) && Objects.equals(value, otherValue);
			}
			return false;
		}
		
		@Override
		public int hashCode() {
			return Objects.hashCode(key) ^ Objects.hashCode(value);
		}
		
		@Override
		public String toString() {
			return Objects.toString(key) + "=" + Objects.toString(value);
		}
		
		int getHeight() { return state; }
		
		void updateHeight() { state = (1 + Math.max(left == null ? -1 : left.getHeight(), right == null ? -1 : right.getHeight())); }
		
		int getBalance() { return (left == null ? -1 : left.getHeight()) - (right == null ? -1 : right.getHeight()); }
		
		boolean needsSuccessor() { return left != null && right != null; }
		
		boolean replace(Entry<T, V> entry) {
			if(entry != null) entry.parent = parent;
			if(parent != null) {
				if(parent.left == this) parent.left = entry;
				else parent.right = entry;
			}
			return parent == null;
		}
		
		Entry<T, V> next() {
			if(right != null) {
				Entry<T, V> parent = right;
				while(parent.left != null) parent = parent.left;
				return parent;
			}
			Entry<T, V> parent = this.parent;
			Entry<T, V> control = this;
			while(parent != null && control == parent.right) {
				control = parent;
				parent = parent.parent;
			}
			return parent;
		}
		
		Entry<T, V> previous() {
			if(left != null) {
				Entry<T, V> parent = left;
				while(parent.right != null) parent = parent.right;
				return parent;
			}
			Entry<T, V> parent = this.parent;
			Entry<T, V> control = this;
			while(parent != null && control == parent.left) {
				control = parent;
				parent = parent.parent;
			}
			return parent;
		}
	}
}