package speiger.src.collections.objects.sets;

import java.util.Collection;
import java.util.Comparator;
import java.util.function.Consumer;
import java.util.function.BiFunction;
import java.util.Iterator;
import java.util.NoSuchElementException;
import java.util.Objects;
import speiger.src.collections.objects.collections.ObjectBidirectionalIterator;
import speiger.src.collections.objects.functions.consumer.ObjectObjectConsumer;
import speiger.src.collections.objects.functions.function.Object2BooleanFunction;
import speiger.src.collections.objects.functions.function.ObjectObjectUnaryOperator;
import speiger.src.collections.objects.collections.ObjectCollection;
import speiger.src.collections.objects.collections.ObjectIterator;
import speiger.src.collections.objects.lists.ObjectListIterator;
import speiger.src.collections.utils.SanityChecks;

/**
 * A Simple Type Specific AVL TreeSet implementation that reduces boxing/unboxing.
 * It is using a bit more memory then <a href="https://github.com/vigna/fastutil">FastUtil</a>,
 * but it saves a lot of Performance on the Optimized removal and iteration logic.
 * Which makes the implementation actually useable and does not get outperformed by Javas default implementation.
 * @param <T> the type of elements maintained by this Collection
 */
public class ObjectAVLTreeSet<T> extends AbstractObjectSet<T> implements ObjectNavigableSet<T>
{
	/** The center of the Tree */
	protected transient Entry<T> tree;
	/** The Lowest possible Node */
	protected transient Entry<T> first;
	/** The Highest possible Node */
	protected transient Entry<T> last;
	/** The amount of elements stored in the Set */
	protected int size = 0;
	/** The Sorter of the Tree */
	protected transient Comparator<T> comparator;
	
	/**
	 * Default Constructor
	 */
	public ObjectAVLTreeSet() {
	}
	
	/**
	 * Constructor that allows to define the sorter
	 * @param comp the function that decides how the tree is sorted, can be null
	 */
	public ObjectAVLTreeSet(Comparator<T> comp) {
		comparator = comp;
	}
	
	/**
	 * Helper constructor that allow to create a set from an array
	 * @param array the elements that should be used
	 */
	public ObjectAVLTreeSet(T[] array) {
		this(array, 0, array.length);
	}
	
	/**
	 * Helper constructor that allow to create a set from an array
	 * @param array the elements that should be used
	 * @param offset the starting index within the array
	 * @param length the amount of elements that are within the array
	 * @throws IllegalStateException if offset and length causes to step outside of the arrays range
	 */
	public ObjectAVLTreeSet(T[] array, int offset, int length) {
		SanityChecks.checkArrayCapacity(array.length, offset, length);
		for(int i = 0;i<length;i++) add(array[offset+i]);
	}
	
	/**
	 * Helper constructor that allow to create a set from an array
	 * @param array the elements that should be used
	 * @param comp the sorter of the tree, can be null
	 */
	public ObjectAVLTreeSet(T[] array, Comparator<T> comp) {
		this(array, 0, array.length, comp);
	}
	
	/**
	 * Helper constructor that allow to create a set from an array
	 * @param array the elements that should be used
	 * @param offset the starting index within the array
	 * @param length the amount of elements that are within the array
	 * @param comp the sorter of the tree, can be null
	 * @throws IllegalStateException if offset and length causes to step outside of the arrays range
	 */
	public ObjectAVLTreeSet(T[] array, int offset, int length, Comparator<T> comp) {
		comparator = comp;
		SanityChecks.checkArrayCapacity(array.length, offset, length);
		for(int i = 0;i<length;i++) add(array[offset+i]);
	}
	
	/**
	 * A Helper constructor that allows to create a Set with exactly the same values as the provided SortedSet.
	 * @param sortedSet the set the elements should be added to the TreeSet
	 * @note this also includes the Comparator if present
	 */
	public ObjectAVLTreeSet(ObjectSortedSet<T> sortedSet) {
		comparator = sortedSet.comparator();
		addAll(sortedSet);
	}
	
	/**
	 * A Helper constructor that allows to create a Set with exactly the same values as the provided collection.
	 * @param collection the set the elements should be added to the TreeSet
	 */
	public ObjectAVLTreeSet(Collection<? extends T> collection) {
		addAll(collection);
	}
	
	/**
	 * A Helper constructor that allows to create a Set with exactly the same values as the provided collection.
	 * @param collection the set the elements should be added to the TreeSet
	 * @param comp the sorter of the tree, can be null
	 */
	public ObjectAVLTreeSet(Collection<? extends T> collection, Comparator<T> comp) {
		comparator = comp;
		addAll(collection);
	}
	
	/**
	 * A Helper constructor that allows to create a Set with exactly the same values as the provided collection.
	 * @param collection the set the elements should be added to the TreeSet
	 */
	public ObjectAVLTreeSet(ObjectCollection<T> collection) {
		addAll(collection);
	}
	
	/**
	 * A Helper constructor that allows to create a Set with exactly the same values as the provided collection.
	 * @param collection the set the elements should be added to the TreeSet
	 * @param comp the sorter of the tree, can be null
	 */
	public ObjectAVLTreeSet(ObjectCollection<T> collection, Comparator<T> comp) {
		comparator = comp;
		addAll(collection);
	}
	
	/**
	 * A Helper constructor that allows to create a set from a iterator of an unknown size
	 * @param iterator the elements that should be added to the set
	 */
	public ObjectAVLTreeSet(Iterator<T> iterator) {
		while(iterator.hasNext()) add(iterator.next());
	}
	
	/**
	 * A Helper constructor that allows to create a set from a iterator of an unknown size
	 * @param iterator the elements that should be added to the set
	 * @param comp the sorter of the tree, can be null
	 */
	public ObjectAVLTreeSet(Iterator<T> iterator, Comparator<T> comp) {
		comparator = comp;
		while(iterator.hasNext()) add(iterator.next());
	}
	
	/**
	 * A Helper constructor that allows to create a set from a iterator of an unknown size
	 * @param iterator the elements that should be added to the set
	 */
	public ObjectAVLTreeSet(ObjectIterator<T> iterator) {
		while(iterator.hasNext()) add(iterator.next());
	}
	
	/**
	 * A Helper constructor that allows to create a set from a iterator of an unknown size
	 * @param iterator the elements that should be added to the set
	 * @param comp the sorter of the tree, can be null
	 */
	public ObjectAVLTreeSet(ObjectIterator<T> iterator, Comparator<T> comp) {
		comparator = comp;
		while(iterator.hasNext()) add(iterator.next());
	}
	
	@Override
	public boolean add(T o) {
		validate(o);
		if(tree == null) {
			tree = first = last = new Entry<>(o, null);
			size++;
			return true;
		}
		int compare = 0;
		Entry<T> parent = tree;
		while(true) {
			if((compare = compare(o, parent.key)) == 0) return false;
			if(compare < 0) {
				if(parent.left == null) break;
				parent = parent.left;
			}
			else if(compare > 0) {
				if(parent.right == null) break;
				parent = parent.right;
			}
		}
		Entry<T> adding = new Entry<>(o, parent);
		if(compare < 0)  {
			parent.left = adding;
			if(parent == first)	first = adding;
		}
		else  {
			parent.right = adding;
			if(parent == last) last = adding;
		}
		fixAfterInsertion(adding);
		size++;
		return true;
	}
	
	@Override
	public boolean addAndMoveToFirst(T o) { throw new UnsupportedOperationException(); }
	
	@Override
	public boolean addAndMoveToLast(T o) { throw new UnsupportedOperationException(); }
	
	@Override
	public boolean moveToFirst(T o) { throw new UnsupportedOperationException(); }
	
	@Override
	public boolean moveToLast(T o) { throw new UnsupportedOperationException(); }

	@Override
	public T lower(T e) {
		Entry<T> node = findLowerNode(e);
		return node != null ? node.key : null;
	}

	@Override
	public T floor(T e) {
		Entry<T> node = findFloorNode(e);
		return node != null ? node.key : null;
	}
	
	@Override
	public T higher(T e) {
		Entry<T> node = findHigherNode(e);
		return node != null ? node.key : null;
	}

	@Override
	public T ceiling(T e) {
		Entry<T> node = findCeilingNode(e);
		return node != null ? node.key : null;
	}
	
	@Override
	public void forEach(Consumer<? super T> action) {
		Objects.requireNonNull(action);
		for(Entry<T> entry = first;entry != null;entry = entry.next()) {
			action.accept(entry.key);
		}
	}
	
	@Override
	public <E> void forEach(E input, ObjectObjectConsumer<E, T> action) {
		Objects.requireNonNull(action);
		for(Entry<T> entry = first;entry != null;entry = entry.next())
			action.accept(input, entry.key);		
	}
	
	@Override
	public boolean matchesAny(Object2BooleanFunction<T> filter) {
		Objects.requireNonNull(filter);
		for(Entry<T> entry = first;entry != null;entry = entry.next()) {
			if(filter.getBoolean(entry.key)) return true;
		}
		return false;
	}
	
	@Override
	public boolean matchesNone(Object2BooleanFunction<T> filter) {
		Objects.requireNonNull(filter);
		for(Entry<T> entry = first;entry != null;entry = entry.next()) {
			if(filter.getBoolean(entry.key)) return false;
		}
		return true;
	}
	
	@Override
	public boolean matchesAll(Object2BooleanFunction<T> filter) {
		Objects.requireNonNull(filter);
		for(Entry<T> entry = first;entry != null;entry = entry.next()) {
			if(!filter.getBoolean(entry.key)) return false;
		}
		return true;
	}
	
	@Override
	public T findFirst(Object2BooleanFunction<T> filter) {
		Objects.requireNonNull(filter);
		for(Entry<T> entry = first;entry != null;entry = entry.next()) {
			if(filter.getBoolean(entry.key)) return entry.key;
		}
		return null;
	}
	
	@Override
	public <E> E reduce(E identity, BiFunction<E, T, E> operator) {
		Objects.requireNonNull(operator);
		E state = identity;
		for(Entry<T> entry = first;entry != null;entry = entry.next()) {
			state = operator.apply(state, entry.key);
		}
		return state;
	}
	
	@Override
	public T reduce(ObjectObjectUnaryOperator<T, T> operator) {
		Objects.requireNonNull(operator);
		T state = null;
		boolean empty = true;
		for(Entry<T> entry = first;entry != null;entry = entry.next()) {
			if(empty) {
				empty = false;
				state = entry.key;
				continue;
			}
			state = operator.apply(state, entry.key);
		}
		return state;
	}
	
	@Override
	public int count(Object2BooleanFunction<T> filter) {
		Objects.requireNonNull(filter);
		int result = 0;
		for(Entry<T> entry = first;entry != null;entry = entry.next()) {
			if(filter.getBoolean(entry.key)) result++;
		}
		return result;
	}
	
	protected Entry<T> findNode(T o) {
		Entry<T> node = tree;
		int compare;
		while(node != null) {
			if((compare = compare(o, node.key)) == 0) return node;
			if(compare < 0) node = node.left;
			else node = node.right;
		}
		return null;
	}
	
	protected Entry<T> findLowerNode(T e) {
		Entry<T> entry = tree;
		while(entry != null) {
			if(compare(e, entry.key) > 0) {
				if(entry.right != null) entry = entry.right;
				else return entry;
			}
			else {
				if(entry.left != null) entry = entry.left;
				else {
					Entry<T> parent = entry.parent;
					while(parent != null && parent.left == entry) {
						entry = parent;
						parent = parent.parent;
					}
					return parent;
				}
			}
		}
		return null;
	}
	
	protected Entry<T> findFloorNode(T e) {
		Entry<T> entry = tree;
		int compare;
		while(entry != null) {
			if((compare = compare(e, entry.key)) > 0) {
				if(entry.right == null) break;
				entry = entry.right;
				continue;
			}
			else if(compare < 0) {
				if(entry.left != null) entry = entry.left;
				else {
					Entry<T> parent = entry.parent;
					while(parent != null && parent.left == entry) {
						entry = parent;
						parent = parent.parent;
					}
					return parent;
				}
				continue;
			}
			break;
		}
		return entry;
	}
	
	protected Entry<T> findCeilingNode(T e) {
		Entry<T> entry = tree;
		int compare;
		while(entry != null) {
			if((compare = compare(e, entry.key)) < 0) {
				if(entry.left == null) break;
				entry = entry.left;
				continue;
			}
			else if(compare > 0) {
				if(entry.right != null) entry = entry.right;
				else {
					Entry<T> parent = entry.parent;
					while(parent != null && parent.right == entry) {
						entry = parent;
						parent = parent.parent;
					}
					return parent;
				}
				continue;
			}
			break;
		}
		return entry;
	}
	
	protected Entry<T> findHigherNode(T e) {
		Entry<T> entry = tree;
		while(entry != null) {
			if(compare(e, entry.key) < 0) {
				if(entry.left != null) entry = entry.left;
				else return entry;
			}
			else {
				if(entry.right != null) entry = entry.right;
				else {
					Entry<T> parent = entry.parent;
					while(parent != null && parent.right == entry) {
						entry = parent;
						parent = parent.parent;
					}
					return parent;
				}
			}
		}
		return null;
	}
	
	@Override
	public boolean contains(Object e) {
		return findNode(((T)e)) != null;
	}
	
	@Override
	public T first() {
		if(tree == null) throw new NoSuchElementException();
		return first.key;
	}
	
	@Override
	public T last() {
		if(tree == null) throw new NoSuchElementException();
		return last.key;
	}
	
	@Override
	public boolean remove(Object o) {
		if(tree == null) return false;
		Entry<T> entry = findNode(((T)o));
		if(entry != null) {
			removeNode(entry);
			return true;
		}
		return false;
	}
	
	@Override
	public T pollFirst() {
		if(tree == null) throw new NoSuchElementException();
		T value = first.key;
		removeNode(first);
		return value;
	}
	
	@Override
	public T pollLast() {
		if(tree == null) throw new NoSuchElementException();
		T value = last.key;
		removeNode(last);
		return value;
	}
	
	@Override
	public int size() { return size; }
	
	@Override
	public void clear() {
		size = 0;
		first = null;
		last = null;
		tree = null;
	}
	
	public ObjectAVLTreeSet<T> copy() {
		ObjectAVLTreeSet<T> set = new ObjectAVLTreeSet<>();
		set.size = size;
		if(tree != null) {
			set.tree = tree.copy();
			Entry<T> lastFound = null;
			for(Entry<T> entry = tree;entry != null;entry = entry.left) lastFound = entry;
			set.first = lastFound;
			lastFound = null;
			for(Entry<T> entry = tree;entry != null;entry = entry.right) lastFound = entry;
			set.last = lastFound;
		}
		return set;
	}
	
	@Override
	public Comparator<T> comparator() { return comparator; }
	
	@Override
	public ObjectBidirectionalIterator<T> iterator() { return new SetIterator(false); }
	
	@Override
	public ObjectBidirectionalIterator<T> iterator(T fromElement) {
		Entry<T> entry = findNode(fromElement);
		return entry == null ? null : new SetIterator(entry);
	}
	
	@Override
	public ObjectBidirectionalIterator<T> descendingIterator() { return new SetIterator(true); }
	
	@Override
	public ObjectNavigableSet<T> subSet(T fromElement, boolean fromInclusive, T toElement, boolean toInclusive) {
		return new AscendingSubSet<>(this, false, fromElement, fromInclusive, false, toElement, toInclusive);
	}

	@Override
	public ObjectNavigableSet<T> headSet(T toElement, boolean inclusive) {
		return new AscendingSubSet<>(this, true, null, true, false, toElement, inclusive);
	}

	@Override
	public ObjectNavigableSet<T> tailSet(T fromElement, boolean inclusive) {
		return new AscendingSubSet<>(this, false, fromElement, inclusive, true, null, true);
	}
	
	@Override
	public ObjectNavigableSet<T> descendingSet() {
		return new DescendingSubSet<>(this, true, null, true, true, null, true);
	}
	
	protected void removeNode(Entry<T> entry) {
		size--;
		if(entry.needsSuccessor()) {
			Entry<T> successor = entry.next();
			entry.key = successor.key;
			entry = successor;
		}
		if(entry.previous() == null) first = entry.next();
		if(entry.next() == null) last = entry.previous();
		Entry<T> replacement = entry.left != null ? entry.left : entry.right;
		if(replacement != null) {
			if(entry.replace(replacement)) tree = replacement;
			entry.left = entry.right = entry.parent = null;
			fixAfterDeletion(replacement);
		}
		else if(entry.parent == null) tree = first = last = null;
		else {
			fixAfterDeletion(entry);
			entry.replace(null);
			entry.parent = null;
		}
	}
	
	protected void validate(T k) { compare(k, k); }
	protected int compare(T k, T v) { return comparator != null ? comparator.compare(k, v) : ((Comparable<T>)k).compareTo((T)v);}
	
	/** From CLR */
	protected void rotateLeft(Entry<T> entry) {
		if(entry != null) {
			Entry<T> right = entry.right;
			entry.right = right.left;
			if(right.left != null) right.left.parent = entry;
			right.parent = entry.parent;
			if(entry.parent == null) tree = right;
			else if(entry.parent.left == entry) entry.parent.left = right;
			else entry.parent.right = right;
			right.left = entry;
			entry.parent = right;
			entry.updateHeight();
			right.updateHeight();
		}
	}
	
	/** From CLR */
	protected void rotateRight(Entry<T> entry) {
		if(entry != null) {
			Entry<T> left = entry.left;
			entry.left = left.right;
			if(left.right != null) left.right.parent = entry;
			left.parent = entry.parent;
			if(entry.parent == null) tree = left;
			else if(entry.parent.right == entry) entry.parent.right = left;
			else entry.parent.left = left;
			left.right = entry;
			entry.parent = left;
			entry.updateHeight();
			left.updateHeight();
		}
	}
	
	/** From CLR */
	protected void fixAfterInsertion(Entry<T> entry) {
		while(entry != null) {
			entry.updateHeight();
			int balance = entry.getBalance();
			if(balance > 1) {
				int compare = entry.left.getBalance();
				if(compare > 0) rotateRight(entry);
				else if(compare < 0) {
					rotateLeft(entry.left);
					rotateRight(entry);
				}
			}
			else if(balance < -1) {
				int compare = entry.right.getBalance();
				if(compare < 0) rotateLeft(entry);
				else if(compare > 0) {
					rotateRight(entry.right);
					rotateLeft(entry);
				}
			}
			entry = entry.parent;
		}
	}
	
	/** From CLR */
	protected void fixAfterDeletion(Entry<T> entry) {
		if(entry != null) {
			entry.updateHeight();
			int balance = entry.getBalance();
			if(balance > 1) {
				int subBalance = entry.left.getBalance();
				if(subBalance >= 0) rotateRight(entry);
				else {
					rotateLeft(entry.left);
					rotateRight(entry);
				}
			}
			else if(balance < -1)
			{
				int subBalance = entry.right.getBalance();
				if(subBalance <= 0) rotateLeft(entry);
				else {
					rotateRight(entry.right);
					rotateLeft(entry);
				}
			}
			entry = entry.parent;
		}
	}
	
	private static class AscendingSubSet<T> extends SubSet<T>
	{
		AscendingSubSet(ObjectAVLTreeSet<T> set, boolean fromStart, T start, boolean loInclusive, boolean toEnd, T end, boolean hiInclusive) {
			super(set, fromStart, start, loInclusive, toEnd, end, hiInclusive);
		}
		
		@Override
		public ObjectNavigableSet<T> subSet(T fromElement, boolean fromInclusive, T toElement, boolean toInclusive) {
			if(!inRange(fromElement, fromInclusive)) throw new IllegalArgumentException("fromElement is not in Range");
			if(!inRange(toElement, toInclusive)) throw new IllegalArgumentException("toElement is not in Range");
			return new AscendingSubSet<>(set, false, fromElement, fromInclusive, false, toElement, toInclusive);
		}

		@Override
		public ObjectNavigableSet<T> headSet(T toElement, boolean inclusive) {
			if(!inRange(toElement, inclusive)) throw new IllegalArgumentException("toElement is not in Range");
			return new AscendingSubSet<>(set, fromStart, start, loInclusive, false, toElement, inclusive);
		}

		@Override
		public ObjectNavigableSet<T> tailSet(T fromElement, boolean inclusive) {
			if(!inRange(fromElement, inclusive)) throw new IllegalArgumentException("fromElement is not in Range");
			return new AscendingSubSet<>(set, false, fromElement, inclusive, toEnd, end, hiInclusive);
		}

		@Override
		public ObjectBidirectionalIterator<T> iterator() { return new SubSetIterator(findLowest()); }
		
		@Override
		public ObjectBidirectionalIterator<T> iterator(T fromElement) {
			Entry<T> entry = set.findNode(fromElement);
			return entry == null || !inClosedRange(fromElement) ? null : new SubSetIterator(entry);
		}
		
		@Override
		public ObjectBidirectionalIterator<T> descendingIterator() { return new SubSetIterator(findHighest()); }

		@Override
		public ObjectNavigableSet<T> descendingSet() {
			return new DescendingSubSet<>(set, fromStart, start, loInclusive, toEnd, end, hiInclusive);
		}
	}
	
	private static class DescendingSubSet<T> extends SubSet<T>
	{
		Comparator<T> comparator;
		DescendingSubSet(ObjectAVLTreeSet<T> set, boolean fromStart, T start, boolean loInclusive, boolean toEnd, T end, boolean hiInclusive) {
			super(set, fromStart, start, loInclusive, toEnd, end, hiInclusive);
			comparator = set.comparator();
			if(comparator != null) comparator = comparator.reversed();
		}
		
		@Override
		public Comparator<T> comparator() { return comparator; }
		
		@Override
		public ObjectNavigableSet<T> subSet(T fromElement, boolean fromInclusive, T toElement, boolean toInclusive) {
			if(!inRange(fromElement, fromInclusive)) throw new IllegalArgumentException("fromElement is not in Range");
			if(!inRange(toElement, toInclusive)) throw new IllegalArgumentException("toElement is not in Range");
			return new DescendingSubSet<>(set, false, toElement, toInclusive, false, fromElement, fromInclusive);
		}

		@Override
		public ObjectNavigableSet<T> headSet(T toElement, boolean inclusive) {
			if(!inRange(toElement, inclusive)) throw new IllegalArgumentException("toElement is not in Range");
			return new DescendingSubSet<>(set, false, toElement, inclusive, toEnd, end, hiInclusive);
		}

		@Override
		public ObjectNavigableSet<T> tailSet(T fromElement, boolean inclusive) {
			if(!inRange(fromElement, inclusive)) throw new IllegalArgumentException("fromElement is not in Range");
			return new DescendingSubSet<>(set, fromStart, start, loInclusive, false, fromElement, inclusive);
		}
		
		protected Entry<T> next(Entry<T> entry) { return entry.previous(); }
		protected Entry<T> start() { return findHighest(); }
		
		@Override
		public T first() { return super.last(); }
		
		@Override
		public T pollFirst() { return super.pollLast(); }
		
		@Override
		public T last() { return super.first(); }
		
		@Override
		public T pollLast() { return super.pollFirst(); }
		
		@Override
		public T lower(T e) { return super.higher(e); }

		@Override
		public T floor(T e) { return super.ceiling(e); }

		@Override
		public T ceiling(T e) { return super.floor(e); }

		@Override
		public T higher(T e) { return super.lower(e); }

		@Override
		public ObjectBidirectionalIterator<T> iterator() { return new DescendingSubIterator(findHighest()); }
		
		@Override
		public ObjectBidirectionalIterator<T> iterator(T fromElement) { 
			Entry<T> entry = set.findNode(fromElement);
			return entry == null || !inClosedRange(fromElement) ? null : new DescendingSubIterator(entry);
		}
		
		@Override
		public ObjectBidirectionalIterator<T> descendingIterator() { return new DescendingSubIterator(findLowest()); }
		
		@Override
		public ObjectNavigableSet<T> descendingSet() {
			return new AscendingSubSet<>(set, fromStart, start, loInclusive, toEnd, end, hiInclusive);
		}
		
		private class DescendingSubIterator extends SubSetIterator
		{
			public DescendingSubIterator(Entry<T> entry) {
				super(entry);
			}
			
			@Override
			protected void updateNext() {
				next = current.previous();
				if(!toEnd && next != null && bottomReached(next)) next = null;
			}
			
			@Override
			protected void updatePrevious() {
				previous = current.next();
				if(!fromStart && previous != null && topReached(previous)) previous = null;
			}
		}
	}
	
	private static abstract class SubSet<T> extends AbstractObjectSet<T> implements ObjectNavigableSet<T>
	{
		ObjectAVLTreeSet<T> set;
		T start;
		T end;
		boolean fromStart;
		boolean toEnd;
		boolean loInclusive;
		boolean hiInclusive;
		
		SubSet(ObjectAVLTreeSet<T> set, boolean fromStart, T start, boolean loInclusive, boolean toEnd, T end, boolean hiInclusive) {
			this.set = set;
			this.start = start;
			this.end = end;
			this.fromStart = fromStart;
			this.toEnd = toEnd;
			this.loInclusive = loInclusive;
			this.hiInclusive = hiInclusive;
		}
		
		public T getDefaultMaxValue() { return null; }
		
		public T getDefaultMinValue() { return null; }
		
		boolean tooLow(T key) { return !fromStart && (loInclusive ? set.compare(key, start) < 0 : set.compare(key, start) <= 0); }
		boolean tooHigh(T key) { return !toEnd && (hiInclusive ? set.compare(key, end) > 0 : set.compare(key, end) >= 0); }
		boolean inRange(T key) { return !tooLow(key) && !tooHigh(key); }
		boolean inClosedRange(T key) { return (fromStart || set.compare(key, start) >= 0) && (toEnd || set.compare(end, key) >= 0); }
		boolean inRange(T key, boolean inclusive) { return inclusive ? inRange(key) : inClosedRange(key); }
		protected Entry<T> next(Entry<T> entry) { return entry.next(); }
		protected Entry<T> start() { return findLowest(); }
		
		@Override
		public boolean addAndMoveToFirst(T o) { throw new UnsupportedOperationException(); }
		
		@Override
		public boolean addAndMoveToLast(T o) { throw new UnsupportedOperationException(); }

		@Override
		public boolean moveToFirst(T o) { throw new UnsupportedOperationException(); }

		@Override
		public boolean moveToLast(T o) { throw new UnsupportedOperationException(); }

		@Override
		public Comparator<T> comparator() { return set.comparator(); }
		
		@Override
		public abstract ObjectBidirectionalIterator<T> iterator();
		
		@Override
		public T first() {
			Entry<T> entry = findLowest();
			return entry == null ? getDefaultMaxValue() : entry.key;
		}
		
		protected Entry<T> findLowest() {
			if(fromStart) return set.first;
			Entry<T> entry = loInclusive ? set.findCeilingNode(start) : set.findHigherNode(start);
			return entry == null || tooHigh(entry.key) ? null : entry;
		}
		
		@Override
		public T pollFirst() {
			if(fromStart) return set.pollFirst();
			Entry<T> entry = loInclusive ? set.findCeilingNode(start) : set.findHigherNode(start);
			if(entry != null && !tooHigh(entry.key)) {
				T value = entry.key;
				set.removeNode(entry);
				return value;
			}
			return getDefaultMaxValue();
		}

		@Override
		public T last() {
			Entry<T> entry = findHighest();
			return entry == null ? getDefaultMinValue() : entry.key;
		}
		
		protected Entry<T> findHighest() {
			if(toEnd) return set.last;
			Entry<T> entry = hiInclusive ? set.findFloorNode(end) : set.findLowerNode(end);
			return entry == null || tooLow(entry.key) ? null : entry;
		}

		@Override
		public T pollLast() {
			if(toEnd) return set.pollLast();
			Entry<T> entry = hiInclusive ? set.findFloorNode(end) : set.findLowerNode(end);
			if(entry != null && !tooLow(entry.key)) {
				T value = entry.key;
				set.removeNode(entry);
				return value;
			}
			return getDefaultMinValue();
		}
		
		@Override
		public boolean add(T o) {
			if(!inRange(o)) throw new IllegalArgumentException("Key is out of range");
			return set.add(o);
		}
		
		@Override
		public boolean contains(Object e) {
			T o = ((T)e);
			return inRange(o) && set.contains(o);
		}

		@Override
		public boolean remove(Object e) {
			T o = ((T)e);
			return inRange(o) && set.remove(o);
		}
		
		@Override
		public T lower(T e) {
			if(tooHigh(e)) {
				Entry<T> entry = findHighest();
				return entry == null ? getDefaultMinValue() : entry.key;
			}
			Entry<T> entry = set.findLowerNode(e);
			return entry == null || tooHigh(entry.key) ? getDefaultMaxValue() : entry.key;
		}

		@Override
		public T floor(T e) {
			if(tooHigh(e)) {
				Entry<T> entry = findHighest();
				return entry == null ? getDefaultMinValue() : entry.key;
			}
			Entry<T> entry = set.findFloorNode(e);
			return entry == null || tooHigh(entry.key) ? getDefaultMaxValue() : entry.key;
		}

		@Override
		public T ceiling(T e) {
			if(tooLow(e)) {
				Entry<T> entry = findLowest();
				return entry == null ? getDefaultMaxValue() : entry.key;
			}
			Entry<T> entry = set.findCeilingNode(e);
			return entry == null || tooLow(entry.key) ? getDefaultMinValue() : entry.key;
		}

		@Override
		public T higher(T e) {
			if(tooLow(e)) {
				Entry<T> entry = findLowest();
				return entry == null ? getDefaultMaxValue() : entry.key;
			}
			Entry<T> entry = set.findHigherNode(e);
			return entry == null || tooLow(entry.key) ? getDefaultMinValue() : entry.key;
		}
		
		@Override
		public int size() {
			return fromStart && toEnd ? set.size() : iterator().skip(Integer.MAX_VALUE);
		}
		
		@Override
		public SubSet<T> copy() { throw new UnsupportedOperationException(); }
		
		@Override
		public void forEach(Consumer<? super T> action) {
			Objects.requireNonNull(action);
			for(Entry<T> entry = start();entry != null && inRange(entry.key);entry = next(entry)) {
				action.accept(entry.key);
			}
		}
		
		@Override
		public <E> void forEach(E input, ObjectObjectConsumer<E, T> action) {
			Objects.requireNonNull(action);
			for(Entry<T> entry = start();entry != null && inRange(entry.key);entry = next(entry))
				action.accept(input, entry.key);		
		}
		
		@Override
		public boolean matchesAny(Object2BooleanFunction<T> filter) {
			Objects.requireNonNull(filter);
			for(Entry<T> entry = start();entry != null && inRange(entry.key);entry = next(entry)) {
				if(filter.getBoolean(entry.key)) return true;
			}
			return false;
		}
		
		@Override
		public boolean matchesNone(Object2BooleanFunction<T> filter) {
			Objects.requireNonNull(filter);
			for(Entry<T> entry = start();entry != null && inRange(entry.key);entry = next(entry)) {
				if(filter.getBoolean(entry.key)) return false;
			}
			return true;
		}
		
		@Override
		public boolean matchesAll(Object2BooleanFunction<T> filter) {
			Objects.requireNonNull(filter);
			for(Entry<T> entry = start();entry != null && inRange(entry.key);entry = next(entry)) {
				if(!filter.getBoolean(entry.key)) return false;
			}
			return true;
		}
		
		@Override
		public <E> E reduce(E identity, BiFunction<E, T, E> operator) {
			Objects.requireNonNull(operator);
			E state = identity;
			for(Entry<T> entry = start();entry != null && inRange(entry.key);entry = next(entry)) {
				state = operator.apply(state, entry.key);
			}
			return state;
		}
		
		@Override
		public T reduce(ObjectObjectUnaryOperator<T, T> operator) {
			Objects.requireNonNull(operator);
			T state = null;
			boolean empty = true;
			for(Entry<T> entry = start();entry != null && inRange(entry.key);entry = next(entry)) {
				if(empty) {
					empty = false;
					state = entry.key;
					continue;
				}
				state = operator.apply(state, entry.key);
			}
			return state;
		}
		
		@Override
		public T findFirst(Object2BooleanFunction<T> filter) {
			Objects.requireNonNull(filter);
			for(Entry<T> entry = start();entry != null && inRange(entry.key);entry = next(entry)) {
				if(filter.getBoolean(entry.key)) return entry.key;
			}
			return null;
		}
		
		@Override
		public int count(Object2BooleanFunction<T> filter) {
			Objects.requireNonNull(filter);
			int result = 0;
			for(Entry<T> entry = start();entry != null && inRange(entry.key);entry = next(entry)) {
				if(filter.getBoolean(entry.key)) result++;
			}
			return result;
		}
		
		class SubSetIterator implements ObjectListIterator<T>
		{
			Entry<T> previous;
			Entry<T> next;
			Entry<T> current;
			int index = 0;
			
			public SubSetIterator(Entry<T> entry) {
				next = entry;
				previous = entry == null ? null : entry.previous();
			}
			
			@Override
			public boolean hasNext() {
				return next != null;
			}
			
			@Override
			public boolean hasPrevious() {
				return previous != null;
			}
			
			@Override
			public int nextIndex() {
				return index;
			}
			
			@Override
			public int previousIndex() {
				return index - 1;
			}
			
			boolean topReached(Entry<T> entry) {
				return hiInclusive ? set.compare(entry.key, end) >= 0 : set.compare(entry.key, end) > 0;
			}
			
			boolean bottomReached(Entry<T> entry) {
				return loInclusive ? set.compare(entry.key, start) <= 0 : set.compare(entry.key, start) < 0;
			}
			
			protected void updateNext() {
				next = current.next();
				if(!toEnd && next != null && topReached(next)) next = null;
			}
			
			protected void updatePrevious() {
				previous = current.previous();
				if(!fromStart && previous != null && bottomReached(previous)) previous = null;
			}
			
			@Override
			public T next() {
				if(!hasNext()) throw new NoSuchElementException();
				current = previous = next;
				updateNext();
				index++;
				return current.key;
			}
			
			@Override
			public T previous() {
				if(!hasPrevious()) throw new NoSuchElementException();
				current = next = previous;
				updatePrevious();
				index--;
				return current.key;
			}
			
			@Override
			public void remove() {
				if(current == null) throw new IllegalStateException();
				if(current == previous) index--;
				updateNext();
				updatePrevious();
				set.removeNode(current);
				current = null;
			}
			
			@Override
			public void set(T e){ throw new UnsupportedOperationException(); }
			
			@Override
			public void add(T e) { throw new UnsupportedOperationException(); }
		}
	}
	
	private class SetIterator implements ObjectListIterator<T>
	{
		Entry<T> previous;
		Entry<T> next;
		Entry<T> current;
		int index = 0;
		 
		public SetIterator(boolean descending) {
			if(descending) previous = last;
			else next = first;
		}
		
		public SetIterator(Entry<T> entry) {
			next = entry;
			previous = entry.previous();
		}
		
		@Override
		public boolean hasNext() {
			return next != null;
		}
		
		@Override
		public boolean hasPrevious() {
			return previous != null;
		}
		
		@Override
		public int nextIndex() {
			return index;
		}
		
		@Override
		public int previousIndex() {
			return index - 1;
		}
		
		protected void updateNext() {
			next = current.next();
		}
		
		protected void updatePrevious() {
			previous = current.previous();
		}
		
		@Override
		public T next() {
			if(!hasNext()) throw new NoSuchElementException();
			current = previous = next;
			updateNext();
			index++;
			return current.key;
		}
		
		@Override
		public T previous() {
			if(!hasPrevious()) throw new NoSuchElementException();
			current = next = previous;
			updatePrevious();
			index--;
			return current.key;
		}
		
		@Override
		public void remove() {
			if(current == null) throw new IllegalStateException();
			if(current == previous) index--;
			updateNext();
			updatePrevious();
			if(current.needsSuccessor()) next = current;
			removeNode(current);
			current = null;
		}
				
		@Override
		public void set(T e){ throw new UnsupportedOperationException(); }
		
		@Override
		public void add(T e) { throw new UnsupportedOperationException(); }
	}
	
	private static final class Entry<T>
	{
		T key;
		int state;
		Entry<T> parent;
		Entry<T> left;
		Entry<T> right;
		
		Entry(T key, Entry<T> parent) {
			this.key = key;
			this.parent = parent;
		}
		
		Entry<T> copy() {
			Entry<T> entry = new Entry<>(key, null);
			entry.state = state;
			if(left != null) {
				Entry<T> newLeft = left.copy();
				entry.left = newLeft;
				newLeft.parent = entry;
			}
			if(right != null) {
				Entry<T> newRight = right.copy();
				entry.right = newRight;
				newRight.parent = entry;
			}
			return entry;
		}
		
		int getHeight() { return state; }
		
		void updateHeight() { state = (1 + Math.max(left == null ? -1 : left.getHeight(), right == null ? -1 : right.getHeight())); }
		
		int getBalance() { return (left == null ? -1 : left.getHeight()) - (right == null ? -1 : right.getHeight()); }
		
		boolean needsSuccessor() { return left != null && right != null; }
		
		boolean replace(Entry<T> entry) {
			if(entry != null) entry.parent = parent;
			if(parent != null) {
				if(parent.left == this) parent.left = entry;
				else parent.right = entry;
			}
			return parent == null;
		}
		
		Entry<T> next() {
			if(right != null) {
				Entry<T> parent = right;
				while(parent.left != null) parent = parent.left;
				return parent;
			}
			Entry<T> parent = this.parent;
			Entry<T> control = this;
			while(parent != null && control == parent.right) {
				control = parent;
				parent = parent.parent;
			}
			return parent;
		}
		
		Entry<T> previous() {
			if(left != null) {
				Entry<T> parent = left;
				while(parent.right != null) parent = parent.right;
				return parent;
			}
			Entry<T> parent = this.parent;
			Entry<T> control = this;
			while(parent != null && control == parent.left) {
				control = parent;
				parent = parent.parent;
			}
			return parent;
		}
	}
}