package speiger.src.collections.shorts.maps.impl.tree;

import java.util.Map;
import java.util.Objects;
import java.util.NoSuchElementException;
import java.util.function.Consumer;
import java.util.function.BiFunction;

import speiger.src.collections.shorts.collections.ShortBidirectionalIterator;
import speiger.src.collections.shorts.functions.ShortComparator;
import speiger.src.collections.shorts.functions.ShortConsumer;
import speiger.src.collections.objects.functions.consumer.ObjectShortConsumer;
import speiger.src.collections.shorts.functions.consumer.ShortBooleanConsumer;
import speiger.src.collections.shorts.functions.function.Short2BooleanFunction;
import speiger.src.collections.shorts.functions.function.ShortBooleanUnaryOperator;
import speiger.src.collections.shorts.functions.function.ShortShortUnaryOperator;
import speiger.src.collections.shorts.lists.ShortListIterator;
import speiger.src.collections.shorts.maps.abstracts.AbstractShort2BooleanMap;
import speiger.src.collections.shorts.maps.interfaces.Short2BooleanMap;
import speiger.src.collections.shorts.maps.interfaces.Short2BooleanNavigableMap;
import speiger.src.collections.shorts.sets.AbstractShortSet;
import speiger.src.collections.shorts.sets.ShortNavigableSet;
import speiger.src.collections.shorts.sets.ShortSet;
import speiger.src.collections.shorts.sets.ShortSortedSet;
import speiger.src.collections.shorts.utils.ShortIterators;
import speiger.src.collections.shorts.utils.maps.Short2BooleanMaps;
import speiger.src.collections.booleans.collections.AbstractBooleanCollection;
import speiger.src.collections.booleans.collections.BooleanCollection;
import speiger.src.collections.booleans.collections.BooleanIterator;
import speiger.src.collections.booleans.functions.BooleanSupplier;
import speiger.src.collections.booleans.functions.function.BooleanBooleanUnaryOperator;
import speiger.src.collections.booleans.lists.BooleanListIterator;
import speiger.src.collections.booleans.utils.BooleanIterators;
import speiger.src.collections.booleans.functions.BooleanConsumer;
import speiger.src.collections.objects.functions.consumer.ObjectObjectConsumer;
import speiger.src.collections.objects.functions.function.Object2BooleanFunction;
import speiger.src.collections.objects.functions.consumer.ObjectBooleanConsumer;
import speiger.src.collections.booleans.functions.function.Boolean2BooleanFunction;
import speiger.src.collections.objects.lists.ObjectListIterator;
import speiger.src.collections.objects.utils.ObjectIterators;
import speiger.src.collections.objects.collections.ObjectIterator;
import speiger.src.collections.objects.functions.function.ObjectObjectUnaryOperator;
import speiger.src.collections.objects.sets.AbstractObjectSet;
import speiger.src.collections.objects.sets.ObjectSet;

/**
 * A Simple Type Specific RB TreeMap implementation that reduces boxing/unboxing.
 * It is using a bit more memory then <a href="https://github.com/vigna/fastutil">FastUtil</a>,
 * but it saves a lot of Performance on the Optimized removal and iteration logic.
 * Which makes the implementation actually useable and does not get outperformed by Javas default implementation.
 */
public class Short2BooleanRBTreeMap extends AbstractShort2BooleanMap implements Short2BooleanNavigableMap
{
	/** The center of the Tree */
	protected transient Entry tree;
	/** The Lowest possible Node */
	protected transient Entry first;
	/** The Highest possible Node */
	protected transient Entry last;
	/** The amount of elements stored in the Map */
	protected int size = 0;
	/** The Sorter of the Tree */
	protected transient ShortComparator comparator;
	
	/** the default return value for max searches */
	protected short defaultMaxNotFound = Short.MIN_VALUE;
	/** the default return value for min searches */
	protected short defaultMinNotFound = Short.MAX_VALUE;
	
	/** KeySet Cache */
	protected ShortNavigableSet keySet;
	/** Values Cache */
	protected BooleanCollection values;
	/** EntrySet Cache */
	protected ObjectSet<Short2BooleanMap.Entry> entrySet;
	
	/**
	 * Default Constructor
	 */
	public Short2BooleanRBTreeMap() {
	}
	
	/**
	 * Constructor that allows to define the sorter
	 * @param comp the function that decides how the tree is sorted, can be null
	 */
	public Short2BooleanRBTreeMap(ShortComparator comp) {
		comparator = comp;
	}
	
	/**
	 * Helper constructor that allow to create a map from boxed values (it will unbox them)
	 * @param keys the keys that should be put into the map
	 * @param values the values that should be put into the map.
	 * @throws IllegalStateException if the keys and values do not match in lenght
	 */
	public Short2BooleanRBTreeMap(Short[] keys, Boolean[] values) {
		this(keys, values, null);
	}
	
	/**
	 * Helper constructor that has a custom sorter and allow to create a map from boxed values (it will unbox them)
	 * @param keys the keys that should be put into the map
	 * @param values the values that should be put into the map.
	 * @param comp the function that decides how the tree is sorted, can be null
	 * @throws IllegalStateException if the keys and values do not match in lenght
	 */
	public Short2BooleanRBTreeMap(Short[] keys, Boolean[] values, ShortComparator comp) {
		comparator = comp;
		if(keys.length != values.length) throw new IllegalStateException("Input Arrays are not equal size");
		for(int i = 0,m=keys.length;i<m;i++) put(keys[i].shortValue(), values[i].booleanValue());
	}
	
	/**
	 * Helper constructor that allow to create a map from unboxed values
	 * @param keys the keys that should be put into the map
	 * @param values the values that should be put into the map.
	 * @throws IllegalStateException if the keys and values do not match in lenght
	 */
	public Short2BooleanRBTreeMap(short[] keys, boolean[] values) {
		this(keys, values, null);
	}
	
	/**
	 * Helper constructor that has a custom sorter and allow to create a map from unboxed values
	 * @param keys the keys that should be put into the map
	 * @param values the values that should be put into the map.
	 * @param comp the function that decides how the tree is sorted, can be null
	 * @throws IllegalStateException if the keys and values do not match in lenght
	 */
	public Short2BooleanRBTreeMap(short[] keys, boolean[] values, ShortComparator comp) {
		comparator = comp;
		if(keys.length != values.length) throw new IllegalStateException("Input Arrays are not equal size");
		for(int i = 0,m=keys.length;i<m;i++) put(keys[i], values[i]);
	}
	
	/**
	 * A Helper constructor that allows to create a Map with exactly the same values as the provided map.
	 * @param map the values that should be present in the map
	 */
	public Short2BooleanRBTreeMap(Map<? extends Short, ? extends Boolean> map) {
		this(map, null);
	}
	
	/**
	 * A Helper constructor that has a custom sorter and allows to create a Map with exactly the same values as the provided map.
	 * @param map the values that should be present in the map
	 * @param comp the function that decides how the tree is sorted, can be null
	 */
	public Short2BooleanRBTreeMap(Map<? extends Short, ? extends Boolean> map, ShortComparator comp) {
		comparator = comp;
		putAll(map);
	}
	
	/**
	 * A Type Specific Helper function that allows to create a new Map with exactly the same values as the provided map.
	 * @param map the values that should be present in the map
 	 */
	public Short2BooleanRBTreeMap(Short2BooleanMap map) {
		this(map, null);
	}
	
	/**
	 * A Type Specific Helper function that has a custom sorter and allows to create a new Map with exactly the same values as the provided map.
	 * @param map the values that should be present in the map
	 * @param comp the function that decides how the tree is sorted, can be null
 	 */
	public Short2BooleanRBTreeMap(Short2BooleanMap map, ShortComparator comp) {
		comparator = comp;
		putAll(map);
	}
	
	@Override
	public void setDefaultMaxValue(short value) { defaultMaxNotFound = value; }
	@Override
	public short getDefaultMaxValue() { return defaultMaxNotFound; }
	@Override
	public void setDefaultMinValue(short value) { defaultMinNotFound = value; }
	@Override
	public short getDefaultMinValue() { return defaultMinNotFound; }
	
	@Override
	public boolean put(short key, boolean value) {
		if(tree == null) {
			tree = first = last = new Entry(key, value, null);
			size++;
			return getDefaultReturnValue();
		}
		int compare = 0;
		Entry parent = tree;
		while(true) {
			if((compare = compare(key, parent.key)) == 0) return parent.setValue(value);
			if(compare < 0) {
				if(parent.left == null) break;
				parent = parent.left;
			}
			else if(compare > 0) {
				if(parent.right == null) break;
				parent = parent.right;
			}
		}
		Entry adding = new Entry(key, value, parent);
		if(compare < 0)  {
			parent.left = adding;
			if(parent == first)	first = adding;
		}
		else {
			parent.right = adding;
			if(parent == last) last = adding;
		}
		fixAfterInsertion(adding);
		size++;
		return getDefaultReturnValue();
	}
	
	@Override
	public boolean putIfAbsent(short key, boolean value) {

		if(tree == null) {
			tree = first = last = new Entry(key, value, null);
			size++;
			return getDefaultReturnValue();
		}
		int compare = 0;
		Entry parent = tree;
		while(true) {
			if((compare = compare(key, parent.key)) == 0) return parent.value;
			if(compare < 0) {
				if(parent.left == null) break;
				parent = parent.left;
			}
			else if(compare > 0) {
				if(parent.right == null) break;
				parent = parent.right;
			}
		}
		Entry adding = new Entry(key, value, parent);
		if(compare < 0)  {
			parent.left = adding;
			if(parent == first)	first = adding;
		}
		else  {
			parent.right = adding;
			if(parent == last) last = adding;
		}
		fixAfterInsertion(adding);
		size++;
		return getDefaultReturnValue();
	}
	
	@Override
	public boolean putAndMoveToFirst(short key, boolean value) { throw new UnsupportedOperationException(); }
	@Override
	public boolean putAndMoveToLast(short key, boolean value) { throw new UnsupportedOperationException(); }
	@Override
	public boolean moveToFirst(short key) { throw new UnsupportedOperationException(); }
	@Override
	public boolean moveToLast(short key) { throw new UnsupportedOperationException(); }
	@Override
	public boolean getAndMoveToFirst(short key) { throw new UnsupportedOperationException(); }
	@Override
	public boolean getAndMoveToLast(short key) { throw new UnsupportedOperationException(); }
	
	@Override
	public ShortComparator comparator() { return comparator; }
	
	@Override
	public boolean containsKey(short key) {
		return findNode(key) != null;
	}
	
	@Override
	public boolean get(short key) {
		Entry node = findNode(key);
		return node == null ? getDefaultReturnValue() : node.value;
	}
	
	@Override
	public boolean getOrDefault(short key, boolean defaultValue) {
		Entry node = findNode(key);
		return node == null ? defaultValue : node.value;
	}
	
	@Override
	public short firstShortKey() {
		if(tree == null) throw new NoSuchElementException();
		return first.key;
	}
	
	@Override
	public short pollFirstShortKey() {
		if(tree == null) throw new NoSuchElementException();
		short result = first.key;
		removeNode(first);
		return result;
	}
	
	@Override
	public short lastShortKey() {
		if(tree == null) throw new NoSuchElementException();
		return last.key;
	}
	
	@Override
	public short pollLastShortKey() {
		if(tree == null) throw new NoSuchElementException();
		short result = last.key;
		removeNode(last);
		return result;
	}
	
	@Override
	public Short2BooleanMap.Entry firstEntry() {
		if(tree == null) throw new NoSuchElementException();
		return new BasicEntry(first.key, first.value);
	}
	
	@Override
	public Short2BooleanMap.Entry lastEntry() {
		if(tree == null) throw new NoSuchElementException();
		return new BasicEntry(last.key, last.value);
	}
	
	@Override
	public Short2BooleanMap.Entry pollFirstEntry() {
		if(tree == null) throw new NoSuchElementException();
		BasicEntry entry = new BasicEntry(first.key, first.value);
		removeNode(first);
		return entry;
	}
	
	@Override
	public Short2BooleanMap.Entry pollLastEntry() {
		if(tree == null) throw new NoSuchElementException();
		BasicEntry entry = new BasicEntry(last.key, last.value);
		removeNode(last);
		return entry;
	}
	
	@Override
	public boolean firstBooleanValue() {
		if(tree == null) throw new NoSuchElementException();
		return first.value;
	}
	
	@Override
	public boolean lastBooleanValue() {
		if(tree == null) throw new NoSuchElementException();
		return last.value;
	}
	
	@Override
	public boolean remove(short key) {
		Entry entry = findNode(key);
		if(entry == null) return getDefaultReturnValue();
		boolean value = entry.value;
		removeNode(entry);
		return value;
	}
	
	@Override
	public boolean removeOrDefault(short key, boolean defaultValue) {
		Entry entry = findNode(key);
		if(entry == null) return defaultValue;
		boolean value = entry.value;
		removeNode(entry);
		return value;
	}
	
	@Override
	public boolean remove(short key, boolean value) {
		Entry entry = findNode(key);
		if(entry == null || entry.value != value) return false;
		removeNode(entry);
		return true;
	}
	
	@Override
	public boolean replace(short key, boolean oldValue, boolean newValue) {
		Entry entry = findNode(key);
		if(entry == null || entry.value != oldValue) return false;
		entry.value = newValue;
		return true;
	}
	
	@Override
	public boolean replace(short key, boolean value) {
		Entry entry = findNode(key);
		if(entry == null) return getDefaultReturnValue();
		boolean oldValue = entry.value;
		entry.value = value;
		return oldValue;
	}
	
	@Override
	public boolean computeBoolean(short key, ShortBooleanUnaryOperator mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		Entry entry = findNode(key);
		if(entry == null) {
			boolean newValue = mappingFunction.applyAsBoolean(key, getDefaultReturnValue());
			if(newValue == getDefaultReturnValue()) return newValue;
			put(key, newValue);
			return newValue;
		}
		boolean newValue = mappingFunction.applyAsBoolean(key, entry.value);
		if(newValue == getDefaultReturnValue()) {
			removeNode(entry);
			return newValue;
		}
		entry.value = newValue;
		return newValue;
	}
	
	@Override
	public boolean computeBooleanIfAbsent(short key, Short2BooleanFunction mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		Entry entry = findNode(key);
		if(entry == null) {
			boolean newValue = mappingFunction.get(key);
			if(newValue == getDefaultReturnValue()) return newValue;
			put(key, newValue);
			return newValue;
		}
		if(entry.value == getDefaultReturnValue()) {
			boolean newValue = mappingFunction.get(key);
			if(newValue == getDefaultReturnValue()) return newValue;
			entry.value = newValue;
		}
		return entry.value;
	}
	
	@Override
	public boolean supplyBooleanIfAbsent(short key, BooleanSupplier valueProvider) {
		Objects.requireNonNull(valueProvider);
		Entry entry = findNode(key);
		if(entry == null) {
			boolean newValue = valueProvider.getBoolean();
			if(newValue == getDefaultReturnValue()) return newValue;
			put(key, newValue);
			return newValue;
		}
		if(entry.value == getDefaultReturnValue()) {
			boolean newValue = valueProvider.getBoolean();
			if(newValue == getDefaultReturnValue()) return newValue;
			entry.value = newValue;
		}
		return entry.value;
	}
	
	@Override
	public boolean computeBooleanIfPresent(short key, ShortBooleanUnaryOperator mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		Entry entry = findNode(key);
		if(entry == null || entry.value == getDefaultReturnValue()) return getDefaultReturnValue();
		boolean newValue = mappingFunction.applyAsBoolean(key, entry.value);
		if(newValue == getDefaultReturnValue()) {
			removeNode(entry);
			return newValue;
		}
		entry.value = newValue;
		return newValue;
	}
	
	@Override
	public boolean mergeBoolean(short key, boolean value, BooleanBooleanUnaryOperator mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		Entry entry = findNode(key);
		boolean newValue = entry == null || entry.value == getDefaultReturnValue() ? value : mappingFunction.applyAsBoolean(entry.value, value);
		if(newValue == getDefaultReturnValue()) {
			if(entry != null)
				removeNode(entry);
		}
		else if(entry == null) put(key, newValue);
		else entry.value = newValue;
		return newValue;
	}
	
	@Override
	public void mergeAllBoolean(Short2BooleanMap m, BooleanBooleanUnaryOperator mappingFunction) {
		Objects.requireNonNull(mappingFunction);
		for(Short2BooleanMap.Entry entry : Short2BooleanMaps.fastIterable(m)) {
			short key = entry.getShortKey();
			Entry subEntry = findNode(key);
			boolean newValue = subEntry == null || subEntry.value == getDefaultReturnValue() ? entry.getBooleanValue() : mappingFunction.applyAsBoolean(subEntry.value, entry.getBooleanValue());
			if(newValue == getDefaultReturnValue()) {
				if(subEntry != null)
					removeNode(subEntry);
			}
			else if(subEntry == null) put(key, newValue);
			else subEntry.value = newValue;
		}
	}
	
	@Override
	public void forEach(ShortBooleanConsumer action) {
		for(Entry entry = first;entry != null;entry = entry.next())
			action.accept(entry.key, entry.value);
	}
	
	@Override
	public int size() { return size; }
	
	@Override
	public void clear() {
		size = 0;
		first = null;
		last = null;
		tree = null;
	}
	
	ShortListIterator keyIterator(boolean descending) {
		ShortListIterator iter = new KeyIterator(descending);
		return descending ? ShortIterators.invert(iter) : iter;
	}
	
	ShortListIterator keyIterator(short key) {
		return new KeyIterator(findNode(key));
	}
	
	public Short2BooleanRBTreeMap copy() {
		Short2BooleanRBTreeMap set = new Short2BooleanRBTreeMap();
		set.size = size;
		if(tree != null) {
			set.tree = tree.copy();
			Entry lastFound = null;
			for(Entry entry = tree;entry != null;entry = entry.left) lastFound = entry;
			set.first = lastFound;
			lastFound = null;
			for(Entry entry = tree;entry != null;entry = entry.right) lastFound = entry;
			set.last = lastFound;
		}
		return set;
	}
	
	@Override
	public ShortSortedSet keySet() {
		return navigableKeySet();
	}
	
	@Override
	public ObjectSet<Short2BooleanMap.Entry> short2BooleanEntrySet() {
		if(entrySet == null) entrySet = new EntrySet();
		return entrySet;
	}
	
	@Override
	public BooleanCollection values() {
		if(values == null) values = new Values();
		return values;
	}
	
	@Override
	public ShortNavigableSet navigableKeySet() {
		if(keySet == null) keySet = new KeySet(this);
		return keySet;
	}
	
	@Override
	public Short2BooleanNavigableMap descendingMap() {
		return new DescendingSubMap(this, true, (short)0, true, true, (short)0, true);
	}
	
	@Override
	public ShortNavigableSet descendingKeySet() {
		return descendingMap().navigableKeySet();
	}
	
	@Override
	public Short2BooleanNavigableMap subMap(short fromKey, boolean fromInclusive, short toKey, boolean toInclusive) {
		return new AscendingSubMap(this, false, fromKey, fromInclusive, false, toKey, toInclusive);
	}
	
	@Override
	public Short2BooleanNavigableMap headMap(short toKey, boolean inclusive) {
		return new AscendingSubMap(this, true, (short)0, true, false, toKey, inclusive);
	}
	
	@Override
	public Short2BooleanNavigableMap tailMap(short fromKey, boolean inclusive) {
		return new AscendingSubMap(this, false, fromKey, inclusive, true, (short)0, true);
	}
	
	@Override
	public short lowerKey(short e) {
		Entry node = findLowerNode(e);
		return node != null ? node.key : getDefaultMinValue();
	}

	@Override
	public short floorKey(short e) {
		Entry node = findFloorNode(e);
		return node != null ? node.key : getDefaultMinValue();
	}
	
	@Override
	public short higherKey(short e) {
		Entry node = findHigherNode(e);
		return node != null ? node.key : getDefaultMaxValue();
	}

	@Override
	public short ceilingKey(short e) {
		Entry node = findCeilingNode(e);
		return node != null ? node.key : getDefaultMaxValue();
	}
	
	@Override
	public Short2BooleanMap.Entry lowerEntry(short key) {
		Entry node = findLowerNode(key);
		return node != null ? new BasicEntry(node.key, node.value) : null;
	}
	
	@Override
	public Short2BooleanMap.Entry higherEntry(short key) {
		Entry node = findHigherNode(key);
		return node != null ? new BasicEntry(node.key, node.value) : null;
	}
	
	@Override
	public Short2BooleanMap.Entry floorEntry(short key) {
		Entry node = findFloorNode(key);
		return node != null ? new BasicEntry(node.key, node.value) : null;
	}
	
	@Override
	public Short2BooleanMap.Entry ceilingEntry(short key) {
		Entry node = findCeilingNode(key);
		return node != null ? new BasicEntry(node.key, node.value) : null;
	}
	
	protected Entry findLowerNode(short key) {
		Entry entry = tree;
		while(entry != null) {
			if(compare(key, entry.key) > 0) {
				if(entry.right != null) entry = entry.right;
				else return entry;
			}
			else {
				if(entry.left != null) entry = entry.left;
				else {
					Entry parent = entry.parent;
					while(parent != null && parent.left == entry) {
						entry = parent;
						parent = parent.parent;
					}
					return parent;
				}
			}
		}
		return null;
	}
	
	protected Entry findFloorNode(short key) {
		Entry entry = tree;
		int compare;
		while(entry != null) {
			if((compare = compare(key, entry.key)) > 0) {
				if(entry.right == null) break;
				entry = entry.right;
				continue;
			}
			else if(compare < 0) {
				if(entry.left != null) entry = entry.left;
				else {
					Entry parent = entry.parent;
					while(parent != null && parent.left == entry) {
						entry = parent;
						parent = parent.parent;
					}
					return parent;
				}
				continue;
			}
			break;
		}
		return entry;
	}
	
	protected Entry findCeilingNode(short key) {
		Entry entry = tree;
		int compare;
		while(entry != null) {
			if((compare = compare(key, entry.key)) < 0) {
				if(entry.left == null) break;
				entry = entry.left;
				continue;
			}
			else if(compare > 0) {
				if(entry.right != null) entry = entry.right;
				else {
					Entry parent = entry.parent;
					while(parent != null && parent.right == entry) {
						entry = parent;
						parent = parent.parent;
					}
					return parent;
				}
				continue;
			}
			break;
		}
		return entry;
	}
	
	protected Entry findHigherNode(short key) {
		Entry entry = tree;
		while(entry != null) {
			if(compare(key, entry.key) < 0) {
				if(entry.left != null) entry = entry.left;
				else return entry;
			}
			else {
				if(entry.right != null) entry = entry.right;
				else {
					Entry parent = entry.parent;
					while(parent != null && parent.right == entry) {
						entry = parent;
						parent = parent.parent;
					}
					return parent;
				}
			}
		}
		return null;
	}
	
	protected Entry findNode(short key) {
		Entry node = tree;
		int compare;
		while(node != null) {
			if((compare = compare(key, node.key)) == 0) return node;
			if(compare < 0) node = node.left;
			else node = node.right;
		}
		return null;
	}
	
	protected void removeNode(Entry entry) {
		size--;
		if(entry.needsSuccessor()) {
			Entry successor = entry.next();
			entry.key = successor.key;
			entry.value = successor.value;
			entry = successor;
		}
		Entry replacement = entry.left != null ? entry.left : entry.right;
		if(replacement != null) {
			if(entry.replace(replacement)) tree = replacement;
			if(entry == first) first = replacement;
			if(entry == last) last = entry.right != null ? entry.right : replacement;
			entry.left = entry.right = entry.parent = null;
			if(entry.isBlack()) fixAfterDeletion(replacement);
		}
		else if(entry.parent == null) tree = first = last = null;
		else {
			if(entry.isBlack())
				fixAfterDeletion(entry);
			entry.replace(null);
			if(entry.parent != null) {
				Entry parent = entry.parent;
				if(entry == first) first = parent.left != null ? parent.left : parent;
				if(entry == last) last = entry.right != null ? parent.right : parent;
			}
			entry.parent = null;
		}
	}
	
	protected void validate(short k) { compare(k, k); }
	protected int compare(short k, short v) { return comparator != null ? comparator.compare(k, v) : Short.compare(k, v);}
	protected static boolean isBlack(Entry p) { return p == null || p.isBlack(); }
	protected static Entry parentOf(Entry p) { return (p == null ? null : p.parent); }
	protected static void setBlack(Entry p, boolean c) { if(p != null) p.setBlack(c); }
	protected static Entry leftOf(Entry p) { return p == null ? null : p.left; }
	protected static Entry rightOf(Entry p) { return (p == null) ? null : p.right; }
	
	protected void rotateLeft(Entry entry) {
		if(entry != null) {
			Entry right = entry.right;
			entry.right = right.left;
			if(right.left != null) right.left.parent = entry;
			right.parent = entry.parent;
			if(entry.parent == null) tree = right;
			else if(entry.parent.left == entry) entry.parent.left = right;
			else entry.parent.right = right;
			right.left = entry;
			entry.parent = right;
		}
	}
	
	protected void rotateRight(Entry entry) {
		if(entry != null) {
			Entry left = entry.left;
			entry.left = left.right;
			if(left.right != null) left.right.parent = entry;
			left.parent = entry.parent;
			if(entry.parent == null) tree = left;
			else if(entry.parent.right == entry) entry.parent.right = left;
			else entry.parent.left = left;
			left.right = entry;
			entry.parent = left;
		}
	}
	
	protected void fixAfterInsertion(Entry entry) {
		entry.setBlack(false);
		while(entry != null && entry != tree && !entry.parent.isBlack()) {
			if(parentOf(entry) == leftOf(parentOf(parentOf(entry)))) {
				Entry y = rightOf(parentOf(parentOf(entry)));
				if(!isBlack(y)) {
					setBlack(parentOf(entry), true);
					setBlack(y, true);
					setBlack(parentOf(parentOf(entry)), false);
					entry = parentOf(parentOf(entry));
				}
				else {
					if(entry == rightOf(parentOf(entry))) {
						entry = parentOf(entry);
						rotateLeft(entry);
					}
					setBlack(parentOf(entry), true);
					setBlack(parentOf(parentOf(entry)), false);
					rotateRight(parentOf(parentOf(entry)));
				}
			}
			else {
				Entry y = leftOf(parentOf(parentOf(entry)));
				if(!isBlack(y)) {
					setBlack(parentOf(entry), true);
					setBlack(y, true);
					setBlack(parentOf(parentOf(entry)), false);
					entry = parentOf(parentOf(entry));
				}
				else {
					if(entry == leftOf(parentOf(entry))) {
						entry = parentOf(entry);
						rotateRight(entry);
					}
					setBlack(parentOf(entry), true);
					setBlack(parentOf(parentOf(entry)), false);
					rotateLeft(parentOf(parentOf(entry)));
				}
			}
		}
		tree.setBlack(true);
	}
	
	protected void fixAfterDeletion(Entry entry) {
		while(entry != tree && isBlack(entry)) {
			if(entry == leftOf(parentOf(entry))) {
				Entry sib = rightOf(parentOf(entry));
				if(!isBlack(sib)) {
					setBlack(sib, true);
					setBlack(parentOf(entry), false);
					rotateLeft(parentOf(entry));
					sib = rightOf(parentOf(entry));
				}
				if(isBlack(leftOf(sib)) && isBlack(rightOf(sib))) {
					setBlack(sib, false);
					entry = parentOf(entry);
				}
				else {
					if(isBlack(rightOf(sib))) {
						setBlack(leftOf(sib), true);
						setBlack(sib, false);
						rotateRight(sib);
						sib = rightOf(parentOf(entry));
					}
					setBlack(sib, isBlack(parentOf(entry)));
					setBlack(parentOf(entry), true);
					setBlack(rightOf(sib), true);
					rotateLeft(parentOf(entry));
					entry = tree;
				}
			}
			else {
				Entry sib = leftOf(parentOf(entry));
				if(!isBlack(sib)) {
					setBlack(sib, true);
					setBlack(parentOf(entry), false);
					rotateRight(parentOf(entry));
					sib = leftOf(parentOf(entry));
				}
				if(isBlack(rightOf(sib)) && isBlack(leftOf(sib))) {
					setBlack(sib, false);
					entry = parentOf(entry);
				}
				else {
					if(isBlack(leftOf(sib))) {
						setBlack(rightOf(sib), true);
						setBlack(sib, false);
						rotateLeft(sib);
						sib = leftOf(parentOf(entry));
					}
					setBlack(sib, isBlack(parentOf(entry)));
					setBlack(parentOf(entry), true);
					setBlack(leftOf(sib), true);
					rotateRight(parentOf(entry));
					entry = tree;
				}
			}
		}
		setBlack(entry, true);
	}
	
	static final class AscendingSubMap extends NavigableSubMap {

		public AscendingSubMap(Short2BooleanRBTreeMap m, boolean fromStart, short low, boolean loInclusive, boolean toEnd, short high, boolean hiInclusive) {
			super(m, fromStart, low, loInclusive, toEnd, high, hiInclusive);
		}
		
		@Override
		ShortListIterator keyIterator(boolean descending) { return new SubMapKeyIterator(descending); }
		@Override
		ShortListIterator keyIterator(short key) {
			Short2BooleanRBTreeMap.Entry entry = m.findNode(key);
			return entry == null || !inClosedRange(key) ? null : new SubMapKeyIterator(entry, fromStart ? null : findLowest(), toEnd ? null : findHighest());
		}
		@Override
		BooleanListIterator valueIterator() { return new SubMapValueIterator(false); }
		@Override
		ObjectListIterator<Short2BooleanMap.Entry> entryIterator() { return new SubMapEntrySetIterator(false); }
		
		@Override
		public Short2BooleanNavigableMap descendingMap() {
			if(descendingMap == null) descendingMap = new DescendingSubMap(m, fromStart, low, loInclusive, toEnd, high, hiInclusive);
			return descendingMap;
		}
		
		Short2BooleanRBTreeMap.Entry next(Short2BooleanRBTreeMap.Entry entry) {
			return entry.next();
		}
		
		Short2BooleanRBTreeMap.Entry previous(Short2BooleanRBTreeMap.Entry entry) {
			return entry.previous();
		}
		
		@Override
		public Short2BooleanNavigableMap subMap(short fromKey, boolean fromInclusive, short toKey, boolean toInclusive) {
			if (!inRange(fromKey, fromInclusive)) throw new IllegalArgumentException("fromKey out of range");
			if (!inRange(toKey, toInclusive)) throw new IllegalArgumentException("toKey out of range");
			return new AscendingSubMap(m, false, fromKey, fromInclusive, false, toKey, toInclusive);
		}
		
		@Override
		public Short2BooleanNavigableMap headMap(short toKey, boolean inclusive) {
			if (!inRange(toKey, inclusive)) throw new IllegalArgumentException("toKey out of range");
			return new AscendingSubMap(m, fromStart, low, loInclusive, false, toKey, inclusive);
		}
		
		@Override
		public Short2BooleanNavigableMap tailMap(short fromKey, boolean inclusive) {
			if (!inRange(fromKey, inclusive)) throw new IllegalArgumentException("fromKey out of range");
			return new AscendingSubMap(m, false, fromKey, inclusive, toEnd, high, hiInclusive);
		}
	}
	
	static final class DescendingSubMap extends NavigableSubMap {

		public DescendingSubMap(Short2BooleanRBTreeMap m, boolean fromStart, short low, boolean loInclusive, boolean toEnd, short high, boolean hiInclusive) {
			super(m, fromStart, low, loInclusive, toEnd, high, hiInclusive);
		}
		
		@Override
		public short firstShortKey() { return super.lastShortKey(); }
		
		@Override
		public short pollFirstShortKey() { return super.pollLastShortKey(); }
		
		@Override
		public short lastShortKey() { return super.firstShortKey(); }
		
		@Override
		public short pollLastShortKey() { return super.pollFirstShortKey(); }
		
		@Override
		public boolean firstBooleanValue() { return super.lastBooleanValue(); }
		
		@Override
		public boolean lastBooleanValue() { return super.firstBooleanValue(); }
		
		@Override
		public Short2BooleanMap.Entry firstEntry() { return super.lastEntry(); }
		
		@Override
		public Short2BooleanMap.Entry lastEntry() { return super.firstEntry(); }
		
		@Override
		public Short2BooleanMap.Entry pollFirstEntry() { return super.pollLastEntry(); }
		
		@Override
		public Short2BooleanMap.Entry pollLastEntry() { return super.pollFirstEntry(); }
		
		@Override
		public short lowerKey(short e) { return super.higherKey(e); }
		
		@Override
		public short floorKey(short e) { return super.ceilingKey(e); }
		
		@Override
		public short ceilingKey(short e) { return super.floorKey(e); }
		
		@Override
		public short higherKey(short e) { return super.lowerKey(e); }
		
		@Override
		public Short2BooleanMap.Entry lowerEntry(short e) { return super.higherEntry(e); }
		
		@Override
		public Short2BooleanMap.Entry floorEntry(short e) { return super.ceilingEntry(e); }
		
		@Override
		public Short2BooleanMap.Entry ceilingEntry(short e) { return super.floorEntry(e); }
		
		@Override
		public Short2BooleanMap.Entry higherEntry(short e) { return super.lowerEntry(e); }
		
		protected Short2BooleanRBTreeMap.Entry findLowest() {
			if(fromStart) return m.first;
			Short2BooleanRBTreeMap.Entry entry = loInclusive ? m.findCeilingNode(low) : m.findHigherNode(low);
			return entry == null || tooHigh(entry.key) ? null : entry;
		}

		protected Short2BooleanRBTreeMap.Entry findHighest() {
			if(toEnd) return m.last;
			Short2BooleanRBTreeMap.Entry entry = hiInclusive ? m.findFloorNode(high) : m.findLowerNode(high);
			return entry == null || tooLow(entry.key) ? null : entry;
		}
		
		@Override
		ShortListIterator keyIterator(boolean descending) {
			ShortListIterator iter = new SubMapKeyIterator(!descending);
			return descending ? iter : ShortIterators.invert(iter);
		}
		
		@Override
		ShortListIterator keyIterator(short key) {
			Short2BooleanRBTreeMap.Entry entry = m.findNode(key);
			return entry == null || !inClosedRange(key) ? null : ShortIterators.invert(new SubMapKeyIterator(entry, fromStart ? null : findLowest(), toEnd ? null : findHighest()));
		}
		
		Short2BooleanRBTreeMap.Entry next(Short2BooleanRBTreeMap.Entry entry) {
			return entry.previous();
		}
		
		Short2BooleanRBTreeMap.Entry previous(Short2BooleanRBTreeMap.Entry entry) {
			return entry.next();
		}
		
		@Override
		public ShortComparator comparator() { return m.comparator() == null ? null : m.comparator().reversed(); }
		
		@Override
		BooleanListIterator valueIterator() { return BooleanIterators.invert(new SubMapValueIterator(true)); }
		
		@Override
		ObjectListIterator<Short2BooleanMap.Entry> entryIterator() { return ObjectIterators.invert(new SubMapEntrySetIterator(true)); }
		
		@Override
		public Short2BooleanNavigableMap descendingMap() {
			if(descendingMap == null) descendingMap = new AscendingSubMap(m, fromStart, low, loInclusive, toEnd, high, hiInclusive);
			return descendingMap;
		}

		@Override
		public Short2BooleanNavigableMap subMap(short fromKey, boolean fromInclusive, short toKey, boolean toInclusive) {
			if (!inRange(fromKey, fromInclusive)) throw new IllegalArgumentException("fromKey out of range");
			if (!inRange(toKey, toInclusive)) throw new IllegalArgumentException("toKey out of range");
			return new DescendingSubMap(m, false, fromKey, fromInclusive, false, toKey, toInclusive);
		}
		
		@Override
		public Short2BooleanNavigableMap headMap(short toKey, boolean inclusive) {
			if (!inRange(toKey, inclusive)) throw new IllegalArgumentException("toKey out of range");
			return new DescendingSubMap(m, fromStart, low, loInclusive, false, toKey, inclusive);
		}
		
		@Override
		public Short2BooleanNavigableMap tailMap(short fromKey, boolean inclusive) {
			if (!inRange(fromKey, inclusive)) throw new IllegalArgumentException("fromKey out of range");
			return new DescendingSubMap(m, false, fromKey, inclusive, toEnd, high, hiInclusive);
		}
	}
	
	static abstract class NavigableSubMap extends AbstractShort2BooleanMap implements Short2BooleanNavigableMap {
		final Short2BooleanRBTreeMap m;
		final short low;
		final short high;
		final boolean fromStart;
		final boolean toEnd;
		final boolean loInclusive;
		final boolean hiInclusive;
		transient Short2BooleanNavigableMap descendingMap;
		transient SubMapEntrySet entrySet;
		transient KeySet keySet;
		transient SubMapValues values;
		
		NavigableSubMap(Short2BooleanRBTreeMap m, boolean fromStart, short low, boolean loInclusive, boolean toEnd, short high, boolean hiInclusive) {
			this.m = m;
			this.low = low;
			this.high = high;
			this.fromStart = fromStart;
			this.toEnd = toEnd;
			this.loInclusive = loInclusive;
			this.hiInclusive = hiInclusive;
		}
		
		abstract ShortListIterator keyIterator(boolean descending);
		abstract ShortListIterator keyIterator(short key);
		abstract BooleanListIterator valueIterator();
		abstract ObjectListIterator<Short2BooleanMap.Entry> entryIterator();
		abstract Short2BooleanRBTreeMap.Entry next(Short2BooleanRBTreeMap.Entry entry);
		abstract Short2BooleanRBTreeMap.Entry previous(Short2BooleanRBTreeMap.Entry entry);
		
		@Override
		public NavigableSubMap copy() { throw new UnsupportedOperationException(); }
		@Override
		public abstract Short2BooleanNavigableMap descendingMap();
		@Override
		public abstract Short2BooleanNavigableMap subMap(short fromKey, boolean fromInclusive, short toKey, boolean toInclusive);
		@Override
		public abstract Short2BooleanNavigableMap headMap(short toKey, boolean inclusive);
		@Override
		public abstract Short2BooleanNavigableMap tailMap(short fromKey, boolean inclusive);
		
		boolean tooLow(short key) { return !fromStart && (loInclusive ? m.compare(key, low) < 0 : m.compare(key, low) <= 0); }
		boolean tooHigh(short key) { return !toEnd && (hiInclusive ? m.compare(key, high) > 0 : m.compare(key, high) >= 0); }
		boolean inRange(short key) { return !tooLow(key) && !tooHigh(key); }
		boolean inClosedRange(short key) { return (fromStart || m.compare(key, low) >= 0) && (toEnd || m.compare(high, key) >= 0); }
		boolean inRange(short key, boolean inclusive) { return inclusive ? inRange(key) : inClosedRange(key); }
		
		@Override
		public void setDefaultMaxValue(short e) { m.setDefaultMaxValue(e); }
		@Override
		public short getDefaultMaxValue() { return m.getDefaultMaxValue(); }
		@Override
		public void setDefaultMinValue(short e) { m.setDefaultMinValue(e); }
		@Override
		public short getDefaultMinValue() { return m.getDefaultMinValue(); }
		
		@Override
		public AbstractShort2BooleanMap setDefaultReturnValue(boolean v) { 
			m.setDefaultReturnValue(v);
			return this;
		}
		
		@Override
		public boolean getDefaultReturnValue() { return m.getDefaultReturnValue(); }
		
		@Override
		public boolean putAndMoveToFirst(short key, boolean value) { throw new UnsupportedOperationException(); }
		@Override
		public boolean putAndMoveToLast(short key, boolean value) { throw new UnsupportedOperationException(); }
		@Override
		public boolean moveToFirst(short key) { throw new UnsupportedOperationException(); }
		@Override
		public boolean moveToLast(short key) { throw new UnsupportedOperationException(); }
		@Override
		public boolean getAndMoveToFirst(short key) { throw new UnsupportedOperationException(); }
		@Override
		public boolean getAndMoveToLast(short key) { throw new UnsupportedOperationException(); }
		@Override
		public ShortComparator comparator() { return m.comparator(); }
		@Override
		public boolean containsKey(short key) { return inRange(key) && m.containsKey(key); }
		
		@Override
		public short firstShortKey() { 
			Short2BooleanRBTreeMap.Entry entry = findLowest();
			return entry == null ? getDefaultMaxValue() : entry.key;
		}
		
		@Override
		public short pollFirstShortKey() {
			if(fromStart) return m.pollFirstShortKey();
			Short2BooleanRBTreeMap.Entry entry = loInclusive ? m.findCeilingNode(low) : m.findHigherNode(low);
			if(entry != null && !tooHigh(entry.key)) {
				short value = entry.key;
				m.removeNode(entry);
				return value;
			}
			return getDefaultMaxValue();
		}
		
		@Override
		public short lastShortKey() {
			Short2BooleanRBTreeMap.Entry entry = findHighest();
			return entry == null ? getDefaultMinValue() : entry.key;	
		}
		
		@Override
		public short pollLastShortKey() {
			if(toEnd) return m.pollLastShortKey();
			Short2BooleanRBTreeMap.Entry entry = hiInclusive ? m.findFloorNode(high) : m.findLowerNode(high);
			if(entry != null && !tooLow(entry.key)) {
				short value = entry.key;
				m.removeNode(entry);
				return value;
			}
			return getDefaultMinValue();
		}
		
		@Override
		public boolean firstBooleanValue() {
			Short2BooleanRBTreeMap.Entry entry = findLowest();
			return entry == null ? getDefaultReturnValue() : entry.value;
		}
		
		@Override
		public boolean lastBooleanValue() {
			Short2BooleanRBTreeMap.Entry entry = findHighest();
			return entry == null ? getDefaultReturnValue() : entry.value;
		}
		
		protected Short2BooleanRBTreeMap.Entry findLowest() {
			if(fromStart) return m.first;
			Short2BooleanRBTreeMap.Entry entry = loInclusive ? m.findCeilingNode(low) : m.findHigherNode(low);
			return entry == null || tooHigh(entry.key) ? null : entry;
		}
		
		protected Short2BooleanRBTreeMap.Entry findHighest() {
			if(toEnd) return m.last;
			Short2BooleanRBTreeMap.Entry entry = hiInclusive ? m.findFloorNode(high) : m.findLowerNode(high);
			return entry == null || tooLow(entry.key) ? null : entry;
		}
		
		@Override
		public boolean put(short key, boolean value) {
			if(!inRange(key)) throw new IllegalArgumentException("key out of range");
			return m.put(key, value);
		}
		
		@Override
		public boolean putIfAbsent(short key, boolean value) {
			if(!inRange(key)) throw new IllegalArgumentException("key out of range");
			return m.putIfAbsent(key, value);
		}
		
		@Override
		public boolean get(short key) {
			return inRange(key) ? m.get(key) : getDefaultReturnValue();
		}
		
		@Override
		public boolean getOrDefault(short key, boolean defaultValue) {
			return inRange(key) ? m.getOrDefault(key, defaultValue) : getDefaultReturnValue();
		}
		
		@Override
		public boolean remove(short key) {
			return inRange(key) ? m.remove(key) : getDefaultReturnValue();
		}
		
		@Override
		public boolean removeOrDefault(short key, boolean defaultValue) {
			return inRange(key) ? m.remove(key) : defaultValue;
		}
		
		@Override
		public boolean remove(short key, boolean value) {
			return inRange(key) && m.remove(key, value);
		}
		
		@Override
		public ObjectSet<Short2BooleanMap.Entry> short2BooleanEntrySet() {
			if(entrySet == null) entrySet = new SubMapEntrySet();
			return entrySet;
		}
		
		@Override
		public ShortSet keySet() { return navigableKeySet(); }
		
		@Override
		public ShortNavigableSet navigableKeySet() {
			if(keySet == null) keySet = new KeySet(this);
			return keySet;
		}
		
		@Override
		public ShortNavigableSet descendingKeySet() {
			return descendingMap().navigableKeySet();
		}
		
		@Override
		public BooleanCollection values() {
			if(values == null) values = new SubMapValues();
			return values;
		}
		
		@Override
		public void forEach(ShortBooleanConsumer action) {
			for(Short2BooleanRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
				action.accept(entry.key, entry.value);
		}
		
		@Override
		public int size() {
			return (fromStart && toEnd) ? m.size() : entrySet().size();
		}
		
		@Override
		public Short2BooleanMap.Entry firstEntry() {
			Short2BooleanRBTreeMap.Entry entry = findLowest();
			return entry == null ? null : new BasicEntry(entry.key, entry.value);
		}
		
		@Override
		public Short2BooleanMap.Entry lastEntry() {
			Short2BooleanRBTreeMap.Entry entry = findHighest();
			return entry == null ? null : new BasicEntry(entry.key, entry.value);
		}
		
		@Override
		public Short2BooleanMap.Entry pollFirstEntry() {
			Short2BooleanRBTreeMap.Entry entry = findLowest();
			if(entry == null) return null;
			BasicEntry result = new BasicEntry(entry.key, entry.value);
			m.removeNode(entry);
			return result;
		}
		
		@Override
		public Short2BooleanMap.Entry pollLastEntry() {
			Short2BooleanRBTreeMap.Entry entry = findHighest();
			if(entry == null) return null;
			BasicEntry result = new BasicEntry(entry.key, entry.value);
			m.removeNode(entry);
			return result;
		}
		
		@Override
		public short lowerKey(short e) {
			if(tooHigh(e)) {
				Short2BooleanRBTreeMap.Entry entry = findHighest();
				return entry == null ? getDefaultMinValue() : entry.key;
			}
			Short2BooleanRBTreeMap.Entry entry = m.findLowerNode(e);
			return entry == null || tooHigh(entry.key) ? getDefaultMaxValue() : entry.key;
		}
		
		@Override
		public short floorKey(short e) {
			if(tooHigh(e)) {
				Short2BooleanRBTreeMap.Entry entry = findHighest();
				return entry == null ? getDefaultMinValue() : entry.key;
			}
			Short2BooleanRBTreeMap.Entry entry = m.findFloorNode(e);
			return entry == null || tooHigh(entry.key) ? getDefaultMaxValue() : entry.key;
		}
		
		@Override
		public short ceilingKey(short e) {
			if(tooLow(e)) {
				Short2BooleanRBTreeMap.Entry entry = findLowest();
				return entry == null ? getDefaultMaxValue() : entry.key;
			}
			Short2BooleanRBTreeMap.Entry entry = m.findCeilingNode(e);
			return entry == null || tooLow(entry.key) ? getDefaultMinValue() : entry.key;
		}
		
		@Override
		public short higherKey(short e) {
			if(tooLow(e)) {
				Short2BooleanRBTreeMap.Entry entry = findLowest();
				return entry == null ? getDefaultMaxValue() : entry.key;
			}
			Short2BooleanRBTreeMap.Entry entry = m.findHigherNode(e);
			return entry == null || tooLow(entry.key) ? getDefaultMinValue() : entry.key;
		}
		
		@Override
		public Short2BooleanMap.Entry lowerEntry(short e) {
			if(tooHigh(e)) {
				Short2BooleanRBTreeMap.Entry entry = findHighest();
				return entry == null ? null : new BasicEntry(entry.key, entry.value);
			}
			Short2BooleanRBTreeMap.Entry entry = m.findLowerNode(e);
			return entry == null || tooHigh(entry.key) ? null : new BasicEntry(entry.key, entry.value);
		}
		
		@Override
		public Short2BooleanMap.Entry floorEntry(short e) {
			if(tooHigh(e)) {
				Short2BooleanRBTreeMap.Entry entry = findHighest();
				return entry == null ? null : new BasicEntry(entry.key, entry.value);
			}
			Short2BooleanRBTreeMap.Entry entry = m.findFloorNode(e);
			return entry == null || tooHigh(entry.key) ? null : new BasicEntry(entry.key, entry.value);
		}
		
		@Override
		public Short2BooleanMap.Entry ceilingEntry(short e) {
			if(tooLow(e)) {
				Short2BooleanRBTreeMap.Entry entry = findLowest();
				return entry == null ? null : new BasicEntry(entry.key, entry.value);
			}
			Short2BooleanRBTreeMap.Entry entry = m.findCeilingNode(e);
			return entry == null || tooLow(entry.key) ? null : new BasicEntry(entry.key, entry.value);
		}
		
		@Override
		public Short2BooleanMap.Entry higherEntry(short e) {
			if(tooLow(e)) {
				Short2BooleanRBTreeMap.Entry entry = findLowest();
				return entry == null ? null : new BasicEntry(entry.key, entry.value);
			}
			Short2BooleanRBTreeMap.Entry entry = m.findHigherNode(e);
			return entry == null || tooLow(entry.key) ? null : new BasicEntry(entry.key, entry.value);
		}
		
		class SubMapEntrySet extends AbstractObjectSet<Short2BooleanMap.Entry> implements ObjectSet<Short2BooleanMap.Entry> {
			@Override
			@Deprecated
			public boolean contains(Object o) {
				if(o instanceof Map.Entry) {
					if(o instanceof Short2BooleanMap.Entry) {
						Short2BooleanMap.Entry entry = (Short2BooleanMap.Entry)o;
						if(entry.getKey() == null && comparator() == null) return false;
						Short2BooleanRBTreeMap.Entry subEntry = m.findNode(entry.getShortKey());
						if(subEntry != null) return entry.getBooleanValue() == subEntry.value;
					}
					else {
						Map.Entry<?, ?> entry = (Map.Entry<?, ?>)o;
						if(entry.getKey() == null && comparator() == null) return false;
						if(!(entry.getKey() instanceof Short)) return false;
						Short2BooleanRBTreeMap.Entry subEntry = m.findNode((Short)entry.getKey());
						if(subEntry != null) return Objects.equals(entry.getValue(), Boolean.valueOf(subEntry.value));
					}
				}
				return false;
			}
			
			@Override
			@Deprecated
			public boolean remove(Object o) {
				if(o instanceof Map.Entry) {
					if(o instanceof Short2BooleanMap.Entry) {
						Short2BooleanMap.Entry entry = (Short2BooleanMap.Entry)o;
						return NavigableSubMap.this.remove(entry.getShortKey(), entry.getBooleanValue());
					}
					Map.Entry<?, ?> entry = (Map.Entry<?, ?>)o;
					return NavigableSubMap.this.remove(entry.getKey(), entry.getValue());
				}
				return false;
			}
			@Override
			public ObjectIterator<Short2BooleanMap.Entry> iterator() { return entryIterator(); }
			@Override
			public int size() { return NavigableSubMap.this.size(); }
			@Override
			public void clear() { NavigableSubMap.this.clear(); }
			
			@Override
			public void forEach(Consumer<? super Short2BooleanMap.Entry> action) {
				Objects.requireNonNull(action);
				for(Short2BooleanRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					action.accept(new BasicEntry(entry.key, entry.value));
			}
			
			@Override
			public <E> void forEach(E input, ObjectObjectConsumer<E, Short2BooleanMap.Entry> action) {
				Objects.requireNonNull(action);
				for(Short2BooleanRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					action.accept(input, new BasicEntry(entry.key, entry.value));
			}
			
			@Override
			public boolean matchesAny(Object2BooleanFunction<Short2BooleanMap.Entry> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return false;
				BasicEntry subEntry = new BasicEntry();
				for(Short2BooleanRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					subEntry.set(entry.key, entry.value);
					if(filter.getBoolean(subEntry)) return true;
				}
				return false;
			}
			
			@Override
			public boolean matchesNone(Object2BooleanFunction<Short2BooleanMap.Entry> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return true;
				BasicEntry subEntry = new BasicEntry();
				for(Short2BooleanRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					subEntry.set(entry.key, entry.value);
					if(filter.getBoolean(subEntry)) return false;
				}
				return true;
			}
			
			@Override
			public boolean matchesAll(Object2BooleanFunction<Short2BooleanMap.Entry> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return true;
				BasicEntry subEntry = new BasicEntry();
				for(Short2BooleanRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					subEntry.set(entry.key, entry.value);
					if(!filter.getBoolean(subEntry)) return false;
				}
				return true;
			}
			
			@Override
			public <E> E reduce(E identity, BiFunction<E, Short2BooleanMap.Entry, E> operator) {
				Objects.requireNonNull(operator);
				E state = identity;
				for(Short2BooleanRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					state = operator.apply(state, new BasicEntry(entry.key, entry.value));
				}
				return state;
			}
			
			@Override
			public Short2BooleanMap.Entry reduce(ObjectObjectUnaryOperator<Short2BooleanMap.Entry, Short2BooleanMap.Entry> operator) {
				Objects.requireNonNull(operator);
				Short2BooleanMap.Entry state = null;
				boolean empty = true;
				for(Short2BooleanRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					if(empty) {
						empty = false;
						state = new BasicEntry(entry.key, entry.value);
						continue;
					}
					state = operator.apply(state, new BasicEntry(entry.key, entry.value));
				}
				return state;
			}
			
			@Override
			public Short2BooleanMap.Entry findFirst(Object2BooleanFunction<Short2BooleanMap.Entry> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return null;
				BasicEntry subEntry = new BasicEntry();
				for(Short2BooleanRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					subEntry.set(entry.key, entry.value);
					if(filter.getBoolean(subEntry)) return subEntry;
				}
				return null;
			}
			
			@Override
			public int count(Object2BooleanFunction<Short2BooleanMap.Entry> filter) {
				Objects.requireNonNull(filter);
				if(size() <= 0) return 0;
				int result = 0;
				BasicEntry subEntry = new BasicEntry();
				for(Short2BooleanRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					subEntry.set(entry.key, entry.value);
					if(filter.getBoolean(subEntry)) result++;
				}
				return result;
			}
		}
		
		final class SubMapValues extends AbstractBooleanCollection {
			@Override
			public boolean add(boolean o) { throw new UnsupportedOperationException(); }
			
			@Override
			public boolean contains(boolean e) {
				return containsValue(e);
			}
			
			@Override
			public BooleanIterator iterator() { return valueIterator(); }
			
			@Override
			public int size() {
				return NavigableSubMap.this.size();
			}
			
			@Override
			public void clear() {
				NavigableSubMap.this.clear();
			}
			
			@Override
			public void forEach(BooleanConsumer action) {
				Objects.requireNonNull(action);
				for(Short2BooleanRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					action.accept(entry.value);
			}
			
			@Override
			public <E> void forEach(E input, ObjectBooleanConsumer<E> action) {
				Objects.requireNonNull(action);
				for(Short2BooleanRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					action.accept(input, entry.value);
			}
			
			@Override
			public boolean matchesAny(Boolean2BooleanFunction filter) {
				Objects.requireNonNull(filter);
				for(Short2BooleanRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					if(filter.get(entry.value)) return true;
				return false;
			}
			
			@Override
			public boolean matchesNone(Boolean2BooleanFunction filter) {
				Objects.requireNonNull(filter);
				for(Short2BooleanRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					if(filter.get(entry.value)) return false;
				return true;
			}
			
			@Override
			public boolean matchesAll(Boolean2BooleanFunction filter) {
				Objects.requireNonNull(filter);
				for(Short2BooleanRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					if(!filter.get(entry.value)) return false;
				return true;
			}
			
			@Override
			public boolean reduce(boolean identity, BooleanBooleanUnaryOperator operator) {
				Objects.requireNonNull(operator);
				boolean state = identity;
				for(Short2BooleanRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					state = operator.applyAsBoolean(state, entry.value);
				return state;
			}
			
			@Override
			public boolean reduce(BooleanBooleanUnaryOperator operator) {
				Objects.requireNonNull(operator);
				boolean state = false;
				boolean empty = true;
				for(Short2BooleanRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry)) {
					if(empty) {
						empty = false;
						state = entry.value;
						continue;
					}
					state = operator.applyAsBoolean(state, entry.value);
				}
				return state;
			}
			
			@Override
			public boolean findFirst(Boolean2BooleanFunction filter) {
				Objects.requireNonNull(filter);
				for(Short2BooleanRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					if(filter.get(entry.value)) return entry.value;
				return false;
			}
			
			@Override
			public int count(Boolean2BooleanFunction filter) {
				Objects.requireNonNull(filter);
				int result = 0;
				for(Short2BooleanRBTreeMap.Entry entry = findLowest(), last = findHighest();entry != null && (last == null || last != previous(entry));entry = next(entry))
					if(filter.get(entry.value)) result++;
				return result;
			}
		}
		
		final class SubMapEntrySetIterator extends SubMapEntryIterator implements ObjectListIterator<Short2BooleanMap.Entry> {
			public SubMapEntrySetIterator(boolean descending) {
				super(descending);
			}
			
			public SubMapEntrySetIterator(Short2BooleanRBTreeMap.Entry entry, Short2BooleanRBTreeMap.Entry lowerFence, Short2BooleanRBTreeMap.Entry upperFence) {
				super(entry, lowerFence, upperFence);
			}
			
			@Override
			public Short2BooleanMap.Entry next() { return nextEntry(); }
			@Override
			public Short2BooleanMap.Entry previous() { return previousEntry(); }
			
			@Override
			public void set(Short2BooleanMap.Entry e) { throw new UnsupportedOperationException(); }
			@Override
			public void add(Short2BooleanMap.Entry e) { throw new UnsupportedOperationException(); }
		}
		
		final class SubMapKeyIterator extends SubMapEntryIterator implements ShortListIterator {
			public SubMapKeyIterator(boolean descending) {
				super(descending);
			}
			
			public SubMapKeyIterator(Short2BooleanRBTreeMap.Entry entry, Short2BooleanRBTreeMap.Entry lowerFence, Short2BooleanRBTreeMap.Entry upperFence) {
				super(entry, lowerFence, upperFence);
			}
			
			@Override
			public short nextShort() { return nextEntry().key; }
			@Override
			public short previousShort() { return previousEntry().key; }
			
			@Override
			public void set(short e) { throw new UnsupportedOperationException(); }
			@Override
			public void add(short e) { throw new UnsupportedOperationException(); }
		}
		
		final class SubMapValueIterator extends SubMapEntryIterator implements BooleanListIterator {
			public SubMapValueIterator(boolean descending) {
				super(descending);
			}
			
			public SubMapValueIterator(Short2BooleanRBTreeMap.Entry entry, Short2BooleanRBTreeMap.Entry lowerFence, Short2BooleanRBTreeMap.Entry upperFence) {
				super(entry, lowerFence, upperFence);
			}
			
			@Override
			public boolean nextBoolean() { return nextEntry().value; }
			@Override
			public boolean previousBoolean() { return previousEntry().value; }
			
			@Override
			public void set(boolean e) { throw new UnsupportedOperationException(); }
			@Override
			public void add(boolean e) { throw new UnsupportedOperationException(); }
		}
		
		abstract class SubMapEntryIterator {
			Short lowerFence;
			Short upperFence;
			Short2BooleanRBTreeMap.Entry next;
			Short2BooleanRBTreeMap.Entry previous;
			Short2BooleanRBTreeMap.Entry current;
			int index = 0;
			
			public SubMapEntryIterator(boolean descending) {
				this(descending ? findHighest() : findLowest(), fromStart ? null : findLowest(), toEnd ? null : findHighest());
			}
			
			public SubMapEntryIterator(Short2BooleanRBTreeMap.Entry entry, Short2BooleanRBTreeMap.Entry lowerFence, Short2BooleanRBTreeMap.Entry upperFence) {
				next = entry;
				previous = entry.previous();
				this.lowerFence = lowerFence != null ? Short.valueOf(lowerFence.key) : null;
				this.upperFence = upperFence != null ? Short.valueOf(upperFence.key) : null;
			}
			
			public boolean hasNext() {
				return next != null && (upperFence == null || next.key == upperFence.shortValue());
			}
			
			public boolean hasPrevious() {
				return previous != null && (lowerFence == null || next.key == lowerFence.shortValue());
			}
			
			public int nextIndex() {
				return index;
			}
			
			public int previousIndex() {
				return index - 1;
			}
			
			protected void updateNext() {
				next = current.next();
			}
			
			protected void updatePrevious() {
				previous = current.previous();
			}
			
			public Short2BooleanRBTreeMap.Entry nextEntry() {
				if(!hasNext()) throw new NoSuchElementException();
				current = previous = next;
				updateNext();
				index++;
				return current;
			}
			
			public Short2BooleanRBTreeMap.Entry previousEntry() {
				if(!hasPrevious()) throw new NoSuchElementException();
				current = next = previous;
				updatePrevious();
				index--;
				return current;
			}
			
			public void remove() {
				if(current == null) throw new IllegalStateException();
				if(current == previous) index--;
				updateNext();
				updatePrevious();
				if(current.needsSuccessor()) next = current;
				m.removeNode(current);
				current = null;
			}
		}
	}
	
	private class EntrySet extends AbstractObjectSet<Short2BooleanMap.Entry> {
		
		@Override
		public ObjectIterator<Short2BooleanMap.Entry> iterator() {
			return new EntryIterator(false);
		}
		
		@Override
		@Deprecated
		public boolean contains(Object o) {
			if(o instanceof Map.Entry) {
				if(o instanceof Short2BooleanMap.Entry) {
					Short2BooleanMap.Entry entry = (Short2BooleanMap.Entry)o;
					if(entry.getKey() == null && comparator() == null) return false;
					Short2BooleanRBTreeMap.Entry subEntry = Short2BooleanRBTreeMap.this.findNode(entry.getShortKey());
					if(subEntry != null) return entry.getBooleanValue() == subEntry.value;
				}
				else {
					Map.Entry<?, ?> entry = (Map.Entry<?, ?>)o;
					if(entry.getKey() == null && comparator() == null) return false;
					if(!(entry.getKey() instanceof Short)) return false;
					Short2BooleanRBTreeMap.Entry subEntry = Short2BooleanRBTreeMap.this.findNode((Short)entry.getKey());
					if(subEntry != null) return Objects.equals(entry.getValue(), Boolean.valueOf(subEntry.value));
				}
			}
			return false;
		}
		
		@Override
		@Deprecated
		public boolean remove(Object o) {
			if(o instanceof Map.Entry) {
				if(o instanceof Short2BooleanMap.Entry) {
					Short2BooleanMap.Entry entry = (Short2BooleanMap.Entry)o;
					return Short2BooleanRBTreeMap.this.remove(entry.getShortKey(), entry.getBooleanValue());
				}
				Map.Entry<?, ?> entry = (Map.Entry<?, ?>)o;
				return Short2BooleanRBTreeMap.this.remove(entry.getKey(), entry.getValue());
			}
			return false;
		}
		
		@Override
		public int size() {
			return Short2BooleanRBTreeMap.this.size();
		}
		
		@Override
		public void clear() {
			Short2BooleanRBTreeMap.this.clear();
		}
		
		@Override
		public void forEach(Consumer<? super Short2BooleanMap.Entry> action) {
			Objects.requireNonNull(action);
			for(Short2BooleanRBTreeMap.Entry entry = first;entry != null;entry = entry.next())
				action.accept(new BasicEntry(entry.key, entry.value));
		}
		
		@Override
		public <E> void forEach(E input, ObjectObjectConsumer<E, Short2BooleanMap.Entry> action) {
			Objects.requireNonNull(action);
			for(Short2BooleanRBTreeMap.Entry entry = first;entry != null;entry = entry.next())
				action.accept(input, new BasicEntry(entry.key, entry.value));
		}
		
		@Override
		public boolean matchesAny(Object2BooleanFunction<Short2BooleanMap.Entry> filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return false;
			BasicEntry subEntry = new BasicEntry();
			for(Short2BooleanRBTreeMap.Entry entry = first;entry != null;entry = entry.next()) {
				subEntry.set(entry.key, entry.value);
				if(filter.getBoolean(subEntry)) return true;
			}
			return false;
		}
		
		@Override
		public boolean matchesNone(Object2BooleanFunction<Short2BooleanMap.Entry> filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return true;
			BasicEntry subEntry = new BasicEntry();
			for(Short2BooleanRBTreeMap.Entry entry = first;entry != null;entry = entry.next()) {
				subEntry.set(entry.key, entry.value);
				if(filter.getBoolean(subEntry)) return false;
			}
			return true;
		}
		
		@Override
		public boolean matchesAll(Object2BooleanFunction<Short2BooleanMap.Entry> filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return true;
			BasicEntry subEntry = new BasicEntry();
			for(Short2BooleanRBTreeMap.Entry entry = first;entry != null;entry = entry.next()) {
				subEntry.set(entry.key, entry.value);
				if(!filter.getBoolean(subEntry)) return false;
			}
			return true;
		}
		
		@Override
		public <E> E reduce(E identity, BiFunction<E, Short2BooleanMap.Entry, E> operator) {
			Objects.requireNonNull(operator);
			E state = identity;
			for(Short2BooleanRBTreeMap.Entry entry = first;entry != null;entry = entry.next()) {
				state = operator.apply(state, new BasicEntry(entry.key, entry.value));
			}
			return state;
		}
		
		@Override
		public Short2BooleanMap.Entry reduce(ObjectObjectUnaryOperator<Short2BooleanMap.Entry, Short2BooleanMap.Entry> operator) {
			Objects.requireNonNull(operator);
			Short2BooleanMap.Entry state = null;
			boolean empty = true;
			for(Short2BooleanRBTreeMap.Entry entry = first;entry != null;entry = entry.next()) {
				if(empty) {
					empty = false;
					state = new BasicEntry(entry.key, entry.value);
					continue;
				}
				state = operator.apply(state, new BasicEntry(entry.key, entry.value));
			}
			return state;
		}
		
		@Override
		public Short2BooleanMap.Entry findFirst(Object2BooleanFunction<Short2BooleanMap.Entry> filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return null;
			BasicEntry subEntry = new BasicEntry();
			for(Short2BooleanRBTreeMap.Entry entry = first;entry != null;entry = entry.next()) {
				subEntry.set(entry.key, entry.value);
				if(filter.getBoolean(subEntry)) return subEntry;
			}
			return null;
		}
		
		@Override
		public int count(Object2BooleanFunction<Short2BooleanMap.Entry> filter) {
			Objects.requireNonNull(filter);
			if(size() <= 0) return 0;
			int result = 0;
			BasicEntry subEntry = new BasicEntry();
			for(Short2BooleanRBTreeMap.Entry entry = first;entry != null;entry = entry.next()) {
				subEntry.set(entry.key, entry.value);
				if(filter.getBoolean(subEntry)) result++;
			}
			return result;
		}
	}
	
	static final class KeySet extends AbstractShortSet implements ShortNavigableSet {
		Short2BooleanNavigableMap m;
		
		KeySet(Short2BooleanNavigableMap m) {
			this.m = m;
		}
		
		@Override
		public void setDefaultMaxValue(short e) { m.setDefaultMaxValue(e); }
		@Override
		public short getDefaultMaxValue() { return m.getDefaultMaxValue(); }
		@Override
		public void setDefaultMinValue(short e) { m.setDefaultMinValue(e); }
		@Override
		public short getDefaultMinValue() { return m.getDefaultMinValue(); }
		@Override
		public boolean add(short o) { throw new UnsupportedOperationException(); }
		@Override
		public boolean addAndMoveToFirst(short o) { throw new UnsupportedOperationException(); }
		@Override
		public boolean addAndMoveToLast(short o) { throw new UnsupportedOperationException(); }
		@Override
		public boolean moveToFirst(short o) { throw new UnsupportedOperationException(); }
		@Override
		public boolean moveToLast(short o) { throw new UnsupportedOperationException(); }
		@Override
		public ShortComparator comparator() { return m.comparator(); }
		@Override
		public short lower(short e) { return m.lowerKey(e); }
		@Override
		public short floor(short e) { return m.floorKey(e); }
		@Override
		public short ceiling(short e) { return m.ceilingKey(e); }
		@Override
		public short higher(short e) { return m.higherKey(e); }
		@Override
		public short firstShort() { return m.firstShortKey(); }
		@Override
		public short pollFirstShort() { return m.pollFirstShortKey(); }
		@Override
		public short lastShort() { return m.lastShortKey(); }
		@Override
		public short pollLastShort() { return m.pollLastShortKey(); }		
		@Override
		public boolean remove(short o) { 
			int oldSize = m.size();
			m.remove(o); 
			return oldSize != m.size();
		}
		
		@Override
		public KeySet copy() { throw new UnsupportedOperationException(); }
		
		@Override
		public ShortNavigableSet subSet(short fromElement, boolean fromInclusive, short toElement, boolean toInclusive) {
			return new KeySet(m.subMap(fromElement, fromInclusive, toElement, toInclusive));
		}
		
		@Override
		public ShortNavigableSet headSet(short toElement, boolean inclusive) {
			return new KeySet(m.headMap(toElement, inclusive));
		}
		
		@Override
		public ShortNavigableSet tailSet(short fromElement, boolean inclusive) {
			return new KeySet(m.tailMap(fromElement, inclusive));
		}
		
		@Override
		public ShortNavigableSet descendingSet() {
			return new KeySet(m.descendingMap());
		}
		
		@Override
		public ShortBidirectionalIterator iterator() {
			if(m instanceof Short2BooleanRBTreeMap) return ((Short2BooleanRBTreeMap)m).keyIterator(false);
			return ((NavigableSubMap)m).keyIterator(false);
		}
		
		@Override
		public ShortBidirectionalIterator iterator(short fromElement) {
			if(m instanceof Short2BooleanRBTreeMap) return ((Short2BooleanRBTreeMap)m).keyIterator(fromElement);
			return ((NavigableSubMap)m).keyIterator(fromElement);
		}
		
		@Override
		public ShortBidirectionalIterator descendingIterator() {
			if(m instanceof Short2BooleanRBTreeMap) return ((Short2BooleanRBTreeMap)m).keyIterator(true);
			return ((NavigableSubMap)m).keyIterator(true);
		}
		
		@Override
		public int size() {
			return m.size();
		}
		
		@Override
		public void clear() {
			m.clear();
		}
		
		protected Short2BooleanRBTreeMap.Entry start() {
			if(m instanceof Short2BooleanRBTreeMap) return ((Short2BooleanRBTreeMap)m).first;
			return ((NavigableSubMap)m).findLowest();
		}
		
		protected Short2BooleanRBTreeMap.Entry end() {
			if(m instanceof Short2BooleanRBTreeMap) return null;
			return ((NavigableSubMap)m).findLowest();
		}
		
		protected Short2BooleanRBTreeMap.Entry next(Short2BooleanRBTreeMap.Entry entry) {
			if(m instanceof Short2BooleanRBTreeMap) return entry.next();
			return ((NavigableSubMap)m).next(entry);
		}
		
		protected Short2BooleanRBTreeMap.Entry previous(Short2BooleanRBTreeMap.Entry entry) {
			if(m instanceof Short2BooleanRBTreeMap) return entry.previous();
			return ((NavigableSubMap)m).previous(entry);
		}
		
		@Override
		public void forEach(ShortConsumer action) {
			Objects.requireNonNull(action);
			for(Short2BooleanRBTreeMap.Entry entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				action.accept(entry.key);
		}
		
		@Override
		public <E> void forEach(E input, ObjectShortConsumer<E> action) {
			Objects.requireNonNull(action);
			for(Short2BooleanRBTreeMap.Entry entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				action.accept(input, entry.key);
		}
		
		@Override
		public boolean matchesAny(Short2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(Short2BooleanRBTreeMap.Entry entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				if(filter.get(entry.key)) return true;
			return false;
		}
		
		@Override
		public boolean matchesNone(Short2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(Short2BooleanRBTreeMap.Entry entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				if(filter.get(entry.key)) return false;
			return true;
		}
		
		@Override
		public boolean matchesAll(Short2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(Short2BooleanRBTreeMap.Entry entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				if(!filter.get(entry.key)) return false;
			return true;
		}
		
		@Override
		public short reduce(short identity, ShortShortUnaryOperator operator) {
			Objects.requireNonNull(operator);
			short state = identity;
			for(Short2BooleanRBTreeMap.Entry entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				state = operator.applyAsShort(state, entry.key);
			return state;
		}
		
		@Override
		public short reduce(ShortShortUnaryOperator operator) {
			Objects.requireNonNull(operator);
			short state = (short)0;
			boolean empty = true;
			for(Short2BooleanRBTreeMap.Entry entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry)) {
				if(empty) {
					empty = false;
					state = entry.key;
					continue;
				}
				state = operator.applyAsShort(state, entry.key);
			}
			return state;
		}
		
		@Override
		public short findFirst(Short2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(Short2BooleanRBTreeMap.Entry entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				if(filter.get(entry.key)) return entry.key;
			return (short)0;
		}
		
		@Override
		public int count(Short2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			int result = 0;
			for(Short2BooleanRBTreeMap.Entry entry = start(), end = end();entry != null && (end == null || (end != previous(entry)));entry = next(entry))
				if(filter.get(entry.key)) result++;
			return result;
		}
	}
	
	final class Values extends AbstractBooleanCollection {

		@Override
		public boolean add(boolean o) { throw new UnsupportedOperationException(); }
		
		@Override
		public boolean contains(boolean e) {
			return containsValue(e);
		}
		
		@Override
		public BooleanIterator iterator() { return new ValueIterator(false); }
		
		@Override
		public int size() {
			return Short2BooleanRBTreeMap.this.size();
		}
		
		@Override
		public void clear() {
			Short2BooleanRBTreeMap.this.clear();
		}
		
		@Override
		public void forEach(BooleanConsumer action) {
			Objects.requireNonNull(action);
			for(Short2BooleanRBTreeMap.Entry entry = first;entry != null;entry = entry.next())
				action.accept(entry.value);
		}
		
		@Override
		public <E> void forEach(E input, ObjectBooleanConsumer<E> action) {
			Objects.requireNonNull(action);
			for(Short2BooleanRBTreeMap.Entry entry = first;entry != null;entry = entry.next())
				action.accept(input, entry.value);
		}
		
		@Override
		public boolean matchesAny(Boolean2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(Short2BooleanRBTreeMap.Entry entry = first;entry != null;entry = entry.next())
				if(filter.get(entry.value)) return true;
			return false;
		}
		
		@Override
		public boolean matchesNone(Boolean2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(Short2BooleanRBTreeMap.Entry entry = first;entry != null;entry = entry.next())
				if(filter.get(entry.value)) return false;
			return true;
		}
		
		@Override
		public boolean matchesAll(Boolean2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(Short2BooleanRBTreeMap.Entry entry = first;entry != null;entry = entry.next())
				if(!filter.get(entry.value)) return false;
			return true;
		}
		
		@Override
		public boolean reduce(boolean identity, BooleanBooleanUnaryOperator operator) {
			Objects.requireNonNull(operator);
			boolean state = identity;
			for(Short2BooleanRBTreeMap.Entry entry = first;entry != null;entry = entry.next())
				state = operator.applyAsBoolean(state, entry.value);
			return state;
		}
		
		@Override
		public boolean reduce(BooleanBooleanUnaryOperator operator) {
			Objects.requireNonNull(operator);
			boolean state = false;
			boolean empty = true;
			for(Short2BooleanRBTreeMap.Entry entry = first;entry != null;entry = entry.next()) {
				if(empty) {
					empty = false;
					state = entry.value;
					continue;
				}
				state = operator.applyAsBoolean(state, entry.value);
			}
			return state;
		}
		
		@Override
		public boolean findFirst(Boolean2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			for(Short2BooleanRBTreeMap.Entry entry = first;entry != null;entry = entry.next())
				if(filter.get(entry.value)) return entry.value;
			return false;
		}
		
		@Override
		public int count(Boolean2BooleanFunction filter) {
			Objects.requireNonNull(filter);
			int result = 0;
			for(Short2BooleanRBTreeMap.Entry entry = first;entry != null;entry = entry.next())
				if(filter.get(entry.value)) result++;
			return result;
		}
	}
	
	final class EntryIterator extends MapEntryIterator implements ObjectListIterator<Short2BooleanMap.Entry> {
		
		public EntryIterator(boolean descending) {
			super(descending);
		}
		
		public EntryIterator(Entry entry) {
			super(entry);
		}
		
		@Override
		public Short2BooleanMap.Entry next() {
			return nextEntry();
		}
		
		@Override
		public Short2BooleanMap.Entry previous() {
			return previousEntry();
		}
		
		@Override
		public void set(Short2BooleanMap.Entry e) { throw new UnsupportedOperationException(); }
		
		@Override
		public void add(Short2BooleanMap.Entry e) { throw new UnsupportedOperationException(); }
	}
	
	final class KeyIterator extends MapEntryIterator implements ShortListIterator {
		
		public KeyIterator(boolean descending) {
			super(descending);
		}
		
		public KeyIterator(Entry entry) {
			super(entry);
		}
		
		@Override
		public short previousShort() { return previousEntry().key; }
		@Override
		public short nextShort() { return nextEntry().key; }
		
		@Override
		public void set(short e) { throw new UnsupportedOperationException(); }
		@Override
		public void add(short e) { throw new UnsupportedOperationException(); }
	}
	
	final class ValueIterator extends MapEntryIterator implements BooleanListIterator {
		
		public ValueIterator(boolean descending) {
			super(descending);
		}
		
		public ValueIterator(Entry entry) {
			super(entry);
		}
		
		@Override
		public boolean previousBoolean() {
			return previousEntry().value;
		}
		
		@Override
		public boolean nextBoolean() {
			return nextEntry().value;
		}
		
		@Override
		public void set(boolean e) { throw new UnsupportedOperationException(); }
		
		@Override
		public void add(boolean e) { throw new UnsupportedOperationException(); }
	}
	
	abstract class MapEntryIterator {
		Entry next;
		Entry previous;
		Entry current;
		int index = 0;
		
		public MapEntryIterator(boolean descending) {
			if(descending) previous = last;
			else next = first;
		}
		
		public MapEntryIterator(Entry entry) {
			next = entry;
			previous = entry.previous();
		}
		
		public boolean hasNext() {
			return next != null;
		}
		
		public boolean hasPrevious() {
			return previous != null;
		}
		
		public int nextIndex() {
			return index;
		}
		
		public int previousIndex() {
			return index - 1;
		}
		
		protected void updateNext() {
			next = current.next();
		}
		
		protected void updatePrevious() {
			previous = current.previous();
		}
		
		public Entry nextEntry() {
			if(!hasNext()) throw new NoSuchElementException();
			current = previous = next;
			updateNext();
			index++;
			return current;
		}
		
		public Entry previousEntry() {
			if(!hasPrevious()) throw new NoSuchElementException();
			current = next = previous;
			updatePrevious();
			index--;
			return current;
		}
		
		public void remove() {
			if(current == null) throw new IllegalStateException();
			if(current == previous) index--;
			updateNext();
			updatePrevious();
			if(current.needsSuccessor()) next = current;
			removeNode(current);
			current = null;
		}
	}
	
	private static final class Entry implements Short2BooleanMap.Entry
	{
		static final int BLACK = 1;
		
		short key;
		boolean value;
		int state;
		Entry parent;
		Entry left;
		Entry right;
		
		Entry(short key, boolean value, Entry parent) {
			this.key = key;
			this.value = value;
			this.parent = parent;
		}
		
		Entry copy() {
			Entry entry = new Entry(key, value, null);
			entry.state = state;
			if(left != null) {
				Entry newLeft = left.copy();
				entry.left = newLeft;
				newLeft.parent = entry;
			}
			if(right != null) {
				Entry newRight = right.copy();
				entry.right = newRight;
				newRight.parent = entry;
			}
			return entry;
		}
		
		@Override
		public short getShortKey() {
			return key;
		}
		
		@Override
		public boolean getBooleanValue() {
			return value;
		}
		
		@Override
		public boolean setValue(boolean value) {
			boolean oldValue = this.value;
			this.value = value;
			return oldValue;
		}
		
		@Override
		public boolean equals(Object obj) {
			if(obj instanceof Map.Entry) {
				if(obj instanceof Short2BooleanMap.Entry) {
					Short2BooleanMap.Entry entry = (Short2BooleanMap.Entry)obj;
					return key == entry.getShortKey() && value == entry.getBooleanValue();
				}
				Map.Entry<?, ?> entry = (Map.Entry<?, ?>)obj;
				Object otherKey = entry.getKey();
				if(otherKey == null) return false;
				Object otherValue = entry.getValue();
				return otherKey instanceof Short && otherValue instanceof Boolean && key == ((Short)otherKey).shortValue() && value == ((Boolean)otherValue).booleanValue();
			}
			return false;
		}
		
		@Override
		public int hashCode() {
			return Short.hashCode(key) ^ Boolean.hashCode(value);
		}
		
		@Override
		public String toString() {
			return Short.toString(key) + "=" + Boolean.toString(value);
		}
		
		boolean isBlack() {
			return (state & BLACK) != 0;
		}
		
		void setBlack(boolean value) {
			if(value) state |= BLACK;
			else state &= ~BLACK;
		}
		
		boolean needsSuccessor() { return left != null && right != null; }
		
		boolean replace(Entry entry) {
			if(entry != null) entry.parent = parent;
			if(parent != null) {
				if(parent.left == this) parent.left = entry;
				else parent.right = entry;
			}
			return parent == null;
		}
		
		Entry next() {
			if(right != null) {
				Entry parent = right;
				while(parent.left != null) parent = parent.left;
				return parent;
			}
			Entry parent = this.parent;
			Entry control = this;
			while(parent != null && control == parent.right) {
				control = parent;
				parent = parent.parent;
			}
			return parent;
		}
		
		Entry previous() {
			if(left != null) {
				Entry parent = left;
				while(parent.right != null) parent = parent.right;
				return parent;
			}
			Entry parent = this.parent;
			Entry control = this;
			while(parent != null && control == parent.left) {
				control = parent;
				parent = parent.parent;
			}
			return parent;
		}
	}
}